!******************************************************************************
!
! File : btree_support.f90
!   by : Alan Tackett
!   on : 08/06/98
!  for : WAE program
!
!  Support routines and data structures for the binary tree module
!
!******************************************************************************

Module btree_support

Use btree_data

Implicit none

Type (BinaryTreeNode), PRIVATE, Pointer  :: Check_Prev
Integer, PRIVATE :: Node_Count

!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
!  PrintPoint - Prints the information
!
!******************************************************************************

Logical Function PrintPoint(Item)
  Type (BinaryTreeData), Intent(IN) :: Item

  Write(*,*) '|G|=',Item%Gmag, ' * N=',Item%Npnt

  PrintPoint = .TRUE.
  Return
End Function

!******************************************************************************
!
!  Compare - Compares two data items and determines their relationship.
!              The ordering is determined as |G|,Z,Y,X.
!
!    P1, P2  - Data items to compare
!    
!    Return Values
!         -1 -- P1<P2
!          0 -- P1==P2
!          1 -- P1>P2
!
!******************************************************************************

Integer Function Compare(pd1, pd2)
  Type (BinaryTreeData), Intent(IN) :: pd1, pd2

  Real, PARAMETER :: TOL = 1E-6
  Integer :: Result
  Integer :: P1(3), P2(3)
  Real    :: diff

  P1 = PD1%Npnt
  P2 = PD2%Npnt

  Result = 0

  diff = PD1%Gmag - PD2%Gmag 
  If (diff < -TOL) then
     Result = -1
  else If (diff > TOL) then
     Result = 1
  Else 
     If (P1(3) < P2(3)) then
        Result = -1
     else if (P1(3) > P2(3)) then
        Result = 1
     else
        If (P1(2) < P2(2)) then
           Result = -1
        else If (P1(2) > P2(2)) then
           Result = 1
        else
           If (P1(1) < P2(1)) then
              Result = -1
           else If (P1(1) > P2(1)) then
              Result = 1
           else        
              Result = 0
           End If
        End If
     End If
  End If

  Compare = Result

  Return
End Function

!******************************************************************************
!
!  InOrderTraversal - Performs an in order traversal of the binary tree
!
!     Root     - Binary tree root
!     FuncCall - Function to call at each grid point.  The function returns
!                true if you should continue and false otherwise.
!           
!
!******************************************************************************

Recursive Subroutine InOrderTraversal(Root, FuncCall)
  Type (BinaryTreeNode), Pointer  :: Root
  Logical,               EXTERNAL :: FuncCall

  Logical :: Ok

  If (Associated(Root)) then
     Call InOrderTraversal(Root%LeftChild, FuncCall)

     Ok = FuncCall(Root%Info)

     If (Ok) Call InOrderTraversal(Root%RightChild, FuncCall)
  EndIf

  Return
End Subroutine


!******************************************************************************
!
!  CheckOrder - Performs an in order traversal of the binary tree to do 
!               an integrity check.
!
!     Root     - Binary tree root
!
!******************************************************************************

Recursive Subroutine CheckOrder(Root)
  Type (BinaryTreeNode), Pointer  :: Root

  Integer :: Result

  If (Associated(Root)) then
     Call CheckOrder(Root%LeftChild)

     Node_Count = Node_Count + 1
     If (Associated(Check_Prev)) then        
        Result = Compare(Check_Prev%Info, Root%Info)
        IF (Result==-1) then
           Write(*,*) 'CO: P=',Root%Info%Gmag, ' * CMP=',Result, ' * N=',Node_Count
        else
           Write(*,*) 'CO: P=',Root%Info%Gmag, ' * CMP=',Result, ' * N=',&
                Node_Count, ' ERR!!!!!!!!!!!!!!!'
        End if
     else
        Write(*,*) 'CO: P=',Root%Info%Gmag, ' * N=',Node_Count, &
             ' ******* START ******'
     End If
     
     Check_Prev => Root

     Call CheckOrder(Root%RightChild)
  EndIf

  Return
End Subroutine

!******************************************************************************
!
!  CheckIntegrity - Performs an in order traversal of the binary tree to do 
!               an integrity check.
!
!     Root     - Binary tree root
!
!******************************************************************************

Subroutine CheckIntegrity(Root, N)
  Type (BinaryTreeNode), Pointer  :: Root
  Integer, Intent(IN) :: N

  Nullify(Check_Prev)
  node_Count = 0

  Call CheckOrder(Root)

  If (N==Node_Count) then
     Write(*,*) 'CheckIntegrity:  N=Node_Count!'
  else
     Write(*,*) 'CheckIntegrity: Error!!!!! N=',N, ' * Node_Count=',Node_Count
  End If

  Return
End Subroutine


End Module


