!*****************************************************************************
!
!  File : bzint.f90
!    by : Alan Tackett
!    on : 7/27/95
!   for : PAW Method
!
!  bzint - Brillouin Zone Integration
!
!  This module contains generic routines to perform BZ integration using 
!  several different methods. The MEthods currently implemented are :
!
!     - Linear Tetrahedron (Analytic)
!     - Quadratic Tetrahedron (Numerical, DOS not implemented)
!
!  There is also a routine for constructing a 3-D grid of lattice points
!  that can then be used for creating the tetrahedrons
!
!*****************************************************************************

Module bzint

Use ltbzi
Use qtbzi
Use gaussbzi
Use mathlib
Use bz_data

Implicit NONE

           !*** Define Method Constants ***
Integer, Parameter :: bz_NONE         = 0        !** No Method defined
Integer, Parameter :: bz_LINEAR_TETRA      = 1   !** Use linear Tetra w/o corr
Integer, Parameter :: bz_LINEAR_TETRA_BASE = 1   !** Use linear Tetra w/o corr
Integer, Parameter :: bz_LINEAR_TETRA_CORR = 2   !** Use linear Tetra w/ Corr
Integer, Parameter :: bz_QUAD_TETRA        = 3   !** USe quadratic tetrahedron
Integer, Parameter :: bz_GAUSS             = 4   !** Use gaussian smearing

Integer, Parameter :: bz_SET_METHOD    = 0       !** Set Method
Integer, Parameter :: bz_SET_FERMI     = 1       !** Set the fermi energy
Integer, Parameter :: bz_SET_ElECTRONS = 2       !** # of electrons
Integer, Parameter :: bz_SET_LOW       = 3       !** Set lower bound for ef
Integer, Parameter :: bz_SET_HIGH      = 4       !** Set high bound for ef
Integer, Parameter :: bz_SET_ACCURACY  = 5       !** Set accuracy
Integer, Parameter :: bz_SET_MAXITER   = 6       !** Set Max iterations
Integer, Parameter :: bz_SET_GAUSS     = 7       !** Set Gaussian Width

Type (bz_struct),  PRIVATE :: local_bz         !** Private bz for root finding

Contains

!*****************************************************************************
!
! bz_Init - Initialize the BZ method for use
!
!   BZ    - BZ Data Structure to hold all the info
!   Bands - Number of Bands
!   TotalKPnts - Total Number of K-points created
!   Kp    - K-points of grid - (ALLOCATED)
!   Ke    - Energy at K points - (ALLOCATED, DIMS=(Kpnt,Band))
!   Wt    - List of Weights - (ALLOCATED, DIMS=(Kpnt, Band))
!   TotalCubes - TotalNumber of Cubes created
!   Cube  - List of Cubes - (ALLOCATED, DIMS=(cube, point))
!   CubeVol  - Cube Volume.
!   UCellVol - Unit Cell Volume.
!
!*****************************************************************************

Subroutine bz_Init(bz, Method, Bands, TotalKPnts, Kp, Ke, Wt, TotalCubes, &
                   Cube, CubeVol, UCellVol)
  Type (bz_struct), Intent(OUT) :: bz
  Integer,           Intent(IN) :: Method
  Integer,           Intent(IN) :: Bands
  Integer,           Intent(IN) :: TotalKpnts
  Real,    Pointer              :: Kp(:,:)
  Real,    Pointer              :: Ke(:,:)
  Real,    Pointer              :: Wt(:,:)
  Integer,           Intent(IN) :: TotalCubes
  Integer, Pointer              :: Cube(:,:)
  Real,              Intent(IN) :: CubeVol  
  Real,              Intent(IN) :: UCellVol  

  write(log_unit,*) 'Entered bz_init which no longer works'
  stop
  bz%Method = Method
  bz%Bands = Bands
  bz%TotalKpnts = TotalKpnts
  bz%Kp => Kp
  bz%Ke => Ke
  bz%Weight => Wt
  bz%TotalCubes = TotalCubes
  bz%Cube => Cube
  bz%CubeVol = CubeVol
  bz%UnitCellVol = UCellVol
  bz%TotalUniq = 0

  bz%bz_High = 1
  bz%bz_Low = 0

write(*,*) 'bz_Init: CV = ', CubeVol, ' UCV=', UCellVol, ' * Bands=',bands, ' * TotalKPnts=',TotalKPnts, ' Cubes=',TotalCubes
  Select Case (Method)
    Case (BZ_LINEAR_TETRA_BASE)
      Call InitLTBZI(Bands, 1, TotalKpnts, Kp, KE, Wt, TotalCubes, 0, &
               Cube, CubeVol ,UCellVol)
    Case (BZ_LINEAR_TETRA_CORR)
      Call InitLTBZI(Bands, 1, TotalKpnts, Kp, KE, Wt, TotalCubes,  0, &
               Cube, CubeVol ,UCellVol)
      Call t_DetKUsage 
    Case (BZ_QUAD_TETRA)
      Call InitQTBZI(Bands, 1, TotalKpnts, Kp, KE, Wt, TotalCubes, &
               Cube, CubeVol ,UCellVol)
    Case (BZ_GAUSS)
      Call InitGaussBZI(bz)
  End Select

  Return
End Subroutine


!*****************************************************************************
!
!  bz_Remap - Remaps the energy from the Unique Kpoint list to the full
!             list of K points
!
!*****************************************************************************

Subroutine bz_Remap(bz)
  Type (bz_struct), Intent(OUT) :: bz

  Integer :: i

  If (Size(bz%Ku) <= 0) RETURN
 
  Do i=1, Bz%TotalKpnts
     bz%KE(:, bz%KMap(i)) = bz%KEu(:, i)
  End Do

  Return
End Subroutine


!*****************************************************************************
!
!  bz_Sort - Sorts the energy Levels based on the particular method.
!            
!*****************************************************************************

Subroutine bz_Sort(bz)
  Type (bz_struct), Intent(INOUT) :: bz

  Select Case (bz%Method)
    Case (BZ_LINEAR_TETRA_BASE)
      Call t_Sort
    Case (BZ_LINEAR_TETRA_CORR)
      Call t_Sort
  End Select

End Subroutine

!*****************************************************************************
!
! bz_PrepEnergy - ReMaps the Uniq Kpoint energy levels to the full set
!     of Kpoint energys and also sorts them if required.
!
!*****************************************************************************

Subroutine bz_PrepEnergy(bz)
  Type (bz_struct), Intent(OUT) :: bz

   Call bz_Remap(bz)
   Call bz_Sort(bz)

   Return
End Subroutine


!*****************************************************************************
!
! BZ_MakeGrid - Creates a 3-D grid of lattice points based on the supplied
!               Basis.
!
!   BZ    - Data structure to store all the info
!   Basis - Basis to use for lattice points
!   Rmin  - Array on Min values in units of BASIS vectors
!   Rmax  - Array of Max values in units of BASIS vectors
!   Bands - Number of Bands
!   NCubes - Array containing the number of cubes to create along each axis
!   NCubeEdge - Array containing the number of points each cube dimension needs
!   DoInitMethod - Calls the appropriate init routine based on the 
!                  method selected.  Also sets the method.
!                  You still need to call bz_Set or bz_Set_All to configure
!                  the remaining parameters.(OPTIONAL)
!   UCellVol - Unit Cell Volume required if DoInitMethod is used.
!   
! Return Values
!    This routine allocates space for various arrays needed for performing
!    the BZ integration and returns numerous items all stored in bz. 
!    The following is a list of the parameters returned in bz:
!
!      KPe and Wt - Allocated only, not used.
!      Kp, Kuniq, Cube, TotalKpnts, TotalCubes - Allocated and used 
!
!*****************************************************************************

Subroutine BZ_MakeGrid(BZ,Basis, Rmin, Rmax, Bands, NCubes, &
                       NCubeEdge, DoInitMethod, UCellVol)
  Implicit NONE
  Type (bz_struct), Intent(OUT) :: bz
  Real,             Intent(IN)  :: Basis(3,3)
  Real,             Intent(IN)  :: Rmin(3)
  Real,             Intent(IN)  :: Rmax(3)
  Integer,          Intent(IN)  :: Bands
  Integer,          Intent(IN)  :: NCubes(3)
  Integer,          Intent(IN)  :: NCubeEdge(3)
  Integer, OPTIONAL,Intent(IN)  :: DoInitMethod
  Real,    OPTIONAL,Intent(IN)  :: UCellVol



  Integer           :: i,j,k, n, m, sy, sz, sx0, sy0, sz0, Npnts(3)
  Integer           :: ix, iy, iz, base, offset, bz_method, TotKpnts, TotCubes
  Integer,  Pointer :: Corner(:), Cube(:,:)
  Real,     Pointer :: Kp(:,:), Ke(:,:), Wt(:,:)
  Real              :: x, y, z, delta(3), CVol


  write(log_unit,*) 'Entered BZ_MakeGrid which no longer works'
  stop
  Npnts = nCubes*(nCubeEdge-1) + 1
  n = Product(NPnts)

  Allocate(Kp(3,n), bz%Ku(3,n), STAT=i)
  If (i/=0) then
     Write(*,*) 'bz_MakeGrid: Error during Kp,Ku allocate. Size:',n
     STOP
  End If
  Allocate(Ke(Bands, n), BZ%KeU(Bands, n), STAT=i)
  If (i/=0) then
     Write(*,*) 'bz_MakeGrid: Error during KE, KEU allocate. Size:',n
     STOP
  End If
  Allocate(Wt(Bands, n), BZ%WtUniq(Bands,n), STAT=i)
  If (i/=0) then
     Write(*,*) 'bz_MakeGrid: Error during Wt,WtUniq allocate. Size:',n
     STOP
  End If
  Allocate(Cube(Product(nCubes), Product(nCubeEdge)), STAT=i)
  If (i/=0) then
     Write(*,*) 'bz_MakeGrid: Error during Cube allocate. Size:',n
     STOP
  End If

  delta = (Rmax - Rmin)/(Npnts-1)

  n = 0
  !Do z = Rmin(3), Rmax(3), delta(3)
    !Do y = Rmin(2), Rmax(2), delta(2)
      !Do x = Rmin(1), Rmax(1), delta(1)
   Do k=1,Npnts(3)
      z=Rmin(3)+delta(3)*(k-1)
      Do j=1,Npnts(2)
         y=Rmin(2)+delta(2)*(j-1)
         Do i=1,Npnts(1)
            x=Rmin(1)+delta(1)*(i-1)
        n = n + 1
        Kp(1:3, n) = MATMUL(Basis, (/ x, y, z /))
      End Do
    End Do
  End Do

  TotKPnts = n
  
  sy = Npnts(1)
  sz = sy * Npnts(2)
  sx0 = nCubeEdge(1)-1
  sy0 = sy * (nCubeEdge(1)-1)
  sz0 = sz * (nCubeEdge(2) - 1)
  n = 0
  Do k=0, nCubes(3)-1
    Do j=0, nCubes(2)-1
      Do i=0, nCubes(1)-1
        n = n + 1
        Corner => Cube(n,:)
  
        m = 0
        base = 1 + sx0*i + sy0*j + sz0*k
        Do iz=0, nCubeEdge(3)-1
          Do iy=0, nCubeEdge(2)-1
            Do ix=0, nCubeEdge(1)-1
              m = m + 1
              Corner(m) = base + ix + sy*iy + sz*iz 
            End Do
          End Do
        End Do
      End Do
    End Do
  End Do

  TotCubes = n

  if (PRESENT(DoInitMethod) .AND. PRESENT(UCellVol)) then
    bz_Method = DoInitMethod
    CVol = Product(Delta) * Product(nCubeEdge-1)

!Subroutine bz_Init(bz, Method, Bands, TotalKPnts, Kp, Ke, Wt, TotalCubes, &
!                   Cube, CubeVol, UCellVol)
    Call bz_Init(bz,bz_Method,Bands,TotKpnts,KP,KE,Wt, &
                 TotCubes,Cube,CVol,UCellVol)
  End IF

  Return
End Subroutine

!*****************************************************************************
!
! bz_UniqueKpnts - Takes a list of Kpoints and determines which points in
!        the list are unique. IF Uniq_Size >= 0 then it is assumed that
!        the following arrays have the sizees: BZ%Ku(3, Uniq_Size) and
!        BZ%KEu(BZ%TotalBands, Uniq_Size)
!
! BZ        - Data structure to store all the info
! Basis     - Real Space Cartesian Crystal Lattice Basis
! RotMatrix - List of Rotation Matrices - RotMatrix(3,3,TotRot)
! TotRot    - Number of Rotation Matrices 
! Uniq_Size - Number of Unique K points.  If this value is <= 0 then 
!             the number of unique K-points are calculated and the
!             value is returned. The unique list is NOT created in this case.
!
!
! Algorithm
!   The method used to determine if a K point is unique is based on the 
!   following relation
!
!          B . V = 2*Pi*n
!           i            i
! 
!   where B(i) = Real space basis vector
!         A(i) = Reciprocal space basis vector
!         V = n(1)*A(1) + n(2)*A(2) + n(3)*A(3) is an arbirtary vector
! 
!   If n(1),n(2), and n(3) are integers then V is a reciprocal lattice vector.
!   Define V = Rk-k', where R=rotation matrix, k=K point in question, and
!   k' is an already defined unique k-point. If V is NOT a reciprocal lattice
!   vector for SOME choice of R and k' currently specified then k is a unique
!   k-point.   
!  
! Return Values
!   Kmap, Kuniq, and KEuniq, are allocated arrays returned along with TotUniq
!
!*****************************************************************************

Subroutine bz_UniqueKpnts(bz, Basis, RotMatrix, TotRot, Uniq_Size)
   Implicit NONE
   Type (bz_struct), Intent(INOUT) :: bz
   Real,             Intent(IN)    :: Basis(3,3)
   Real,             Intent(IN)    :: RotMAtrix(:,:,:)
   Integer,          Intent(IN)    :: TotRot
   Integer,          Intent(INOUT) :: Uniq_Size

   Real, PARAMETER :: err = 0.0001
   Real    :: V(3), n1, n2, n3
   Integer :: i,j, k, Num_Uniq
   Logical :: Used


   write(log_unit,*) 'Entered bz_UniqueKpnts which no longer works'
   stop
   if (Uniq_Size > 0) then
     bz%Ku(:, 1) = bz%Kp(:,1)
     bz%KMap(1) = 1
   End If

   Num_Uniq = 1

   Do i=2, bz%TotalKpnts
     Used = .FALSE.
     j = 1
     Do While ((.NOT. Used) .AND. (j <= Num_Uniq))      !** Loop over Uniq pnts
       k = 1
       Do While ((.NOT. Used) .AND. (k <= TotRot))     !** Loop over RotMat
         V = MATMUL(RotMatrix(1:3,1:3, k), bz%Kp(1:3,i)) - bz%Ku(1:3,j)

         n1 = DOT_PRODUCT(Basis(1:3,1), V) / (2*PI)
         n2 = DOT_PRODUCT(Basis(1:3,2), V) / (2*PI)
         n3 = DOT_PRODUCT(Basis(1:3,3), V) / (2*PI)

         if (IsInteger(n1,err) .AND. IsInteger(n2,err) &
            .AND. IsInteger(n3,err) .AND. (Uniq_Size>0)) then 
            Used = .TRUE.
	    bz%KMap(i) = k
         End If
       
         k = k + 1
       End Do

       j = j + 1
     End Do

     if (.NOT. Used) then
        Num_Uniq = Num_Uniq + 1
        if (Uniq_Size > 0) bz%Ku(1:3,Num_Uniq) = bz%Kp(1:3, i)
     End If
   End Do

   i = bz_ReallocUnique(bz, Uniq_Size)

   If (i /= 0) then
      Write(*,*) 'bz_UniqueKPnts: Error during bz_ReallocaUnique, Uniq_size=',&
          Uniq_Size       
   End If
   Return     
End Subroutine

!*****************************************************************************
!
! bz_AllocUnique - Allocates the space for the unique K-points list 
!           and associated data structures required
!
! Uniq_Size - Size of the unique K-points list
!
! Return Values
!   The allocation error status is returned
!
!*****************************************************************************

Integer Function bz_ReallocUnique(bz, Uniq_Size)
  Implicit NONE
  Type (bz_struct), Intent(INOUT) :: bz
  Integer,          Intent(INOUT) :: Uniq_Size

  Integer :: ierr
!  Real, Target :: A(bz%Bands, Uniq_Size)
  Real, Target , allocatable :: A(:,:)

  Allocate(A(bz%Bands, Uniq_Size))

  bz%TotalUniq = Uniq_Size

  If (Uniq_Size > 0) then
     
     A = bz%Keu(:,1:Uniq_Size)
     Deallocate(BZ%Keu)
     Allocate(bz%Keu(Bz%BAnds, Uniq_Size), STAT=ierr)
     bz%Keu(:,1:Uniq_Size) = A

     A = bz%WtUniq(:,1:Uniq_Size)
     Deallocate(BZ%WtUniq)
     Allocate(bz%WtUniq(Bz%BAnds, Uniq_Size), STAT=ierr)
     bz%WtUniq(:,1:Uniq_Size) = A

     A(1:3, 1:Uniq_Size) = Bz%Ku(1:3, 1:Uniq_Size)
     DeAllocate(BZ%Ku)
     Allocate(bz%Ku(3, Uniq_Size), STAT=ierr)
     Bz%Ku(1:3, 1:Uniq_Size) = A(1:3, 1:Uniq_Size) 
  else
     DeAllocate(Bz%Ku, Bz%KEu, bz%WtUniq, STAT=ierr)
  End If

  bz_ReallocUnique = ierr
  DeAllocate(A)
  Return
End Function

!*****************************************************************************
!
! bz_Set_Weights - Simply sets the pointer to the integration weights
!
! BZ - Data structure to store all the info
! Wt - Array containing integration weights. Same size as Ke
!
!*****************************************************************************

Subroutine bz_Set_Weights(bz, Wt)
   Implicit NONE
   Type (bz_struct), Intent(INOUT) :: bz
   Real,                   Pointer ::  Wt(:,:)

   bz%Weight => Wt

   Return
End Subroutine


!*****************************************************************************
!
!  bz_Set - Sets a BZ integration value
!
!  BZ  - Data structure to store all the info
!  Set - BZ Value to set
!  Val - Value to store
!
!*****************************************************************************

Subroutine bz_Set(bz, Set, Val)
   Implicit NONE
   Type (bz_struct), Intent(INOUT) :: bz
   Integer,          Intent(IN)    :: Set
   Real,             Intent(IN)    :: Val


   Select Case (Set)
     Case (bz_SET_METHOD)
       bz%Method = Val
     Case (bz_SET_FERMI)
       bz%Fermi = Val
     Case (bz_SET_ELECTRONS)
       bz%Electrons = Val
     Case (bz_SET_LOW)
       bz%bz_Low = Val
     Case (bz_SET_HIGH)
       bz%bz_High = Val
     Case (bz_SET_ACCURACY)
       bz%Accuracy = VAl
     Case (bz_SET_MAXITER)
       bz%MaxIter = VAl
     Case (bz_SET_GAUSS)
       bz%Sigma = Val
   End Select

   Return 
End Subroutine

!*****************************************************************************
!
!  bz_Get - Returns a BZ integration value
!
!  BZ  - Data structure to store all the info
!  Get - BZ Value to Get
!
!*****************************************************************************

Real Function bz_Get(bz, Get)
   Implicit NONE
   Type (bz_struct), Intent(INOUT) :: bz
   Integer,          Intent(IN)    :: Get
  
   Select Case (Get)
     Case (bz_SET_METHOD)
       bz_Get = bz%Method
     Case (bz_SET_FERMI)
       bz_Get = bz%Fermi
     Case (bz_SET_ELECTRONS)
       bz_Get = bz%Electrons
     Case (bz_SET_LOW)
       bz_Get = bz%bz_Low
     Case (bz_SET_HIGH)
       bz_Get = bz%bz_High
     Case (bz_SET_ACCURACY)
       bz_Get = bz%Accuracy
     Case (bz_SET_MAXITER)
       bz_Get = bz%MaxIter
     Case (bz_SET_GAUSS)
       bz_Get = bz%Sigma
   End Select

   Return 
End Function


!*****************************************************************************
!
!  bz_Set_All - Sets all BZ integration values
!
!  BZ       - Data structure to store all the info
!  Method   - BZ method to use
!  Fermi    - Fermi Energy
!  elec     - # of electrons
!  ef_Low   - Lower bound on Fermi energy
!  ef_High  - Upper bound on Fermi Energy
!  Accuracy - Fermi Energy Accuracy
!  MaxIter  - Max iteration for fincing ef
!
!*****************************************************************************

Subroutine bz_Set_All(bz,Method,Fermi,elec, ef_Low, ef_High, Accuracy, &
                      MaxIter, Sigma)
  Type (bz_struct),  Intent(INOUT) :: bz
  Integer,           Intent(IN)    :: Method
  Real,              Intent(IN)    :: Fermi
  Real,              Intent(IN)    :: Elec
  Real,              Intent(IN)    :: ef_Low
  Real,              Intent(IN)    :: ef_High
  Real,              Intent(IN)    :: Accuracy
  Integer,           Intent(IN)    :: MaxIter
  Real,  OPTIONAL,   Intent(IN)    :: Sigma

  bz%Method = Method
  bz%Fermi = Fermi
  bz%Electrons = Elec
  bz%bz_Low = ef_Low
  bz%bz_High = ef_High
  bz%Accuracy = Accuracy
  bz%MaxIter = MaxIter

  If (Present(Sigma)) bz%Sigma = Sigma

  Return
End Subroutine


!*****************************************************************************
!
!  bz_Get_All - Gets all BZ integration values
!
!  BZ       - Data structure to store all the info
!  Method   - BZ method
!  Fermi    - Fermi Energy
!  Elec     - # of electrons
!  ef_Low   - Lower bound on Fermi energy
!  ef_High  - Upper bound on Fermi Energy
!  Accuracy - Fermi Energy Accuracy
!
!*****************************************************************************

Subroutine bz_Get_All(bz,Method, Fermi,Elec, ef_Low,ef_High, Accuracy, &
                      MaxIter, Sigma)
  Implicit NONE
  Type (bz_struct),  Intent(IN) :: bz
  Integer,          Intent(OUT) :: Method
  Real,             Intent(OUT) :: Fermi
  Real,             Intent(OUT) :: Elec
  Real,             Intent(OUT) :: ef_Low
  Real,             Intent(OUT) :: ef_High
  Real,             Intent(OUT) :: Accuracy
  Integer,          Intent(OUT) :: MaxIter
  Real,  OPTIONAL,  Intent(OUT) :: Sigma

  Method = bz%Method
  Fermi = bz%Fermi
  Elec = bz%Electrons
  ef_Low = bz%bz_Low
  ef_High = bz%bz_High
  Accuracy = bz%Accuracy
  MaxIter = bz%MaxIter

  If (Present(Sigma)) Sigma = bz%Sigma

  Return
End Subroutine


!*****************************************************************************
!
! bz_DOS - Calculates the Density of States
!
!  BZ  - Data structure to store all the info
!
!*****************************************************************************

Real Function bz_DOS(bz)
  Implicit NONE
  Type (bz_struct),  Intent(IN) :: bz

  Real  :: Energy(1), DOS(1)

  Select Case (bz%Method)
    Case (BZ_LINEAR_TETRA_BASE)
       Energy(1) = bz%Fermi
       Call t_DOS(1, Energy, DOS)
    Case (BZ_LINEAR_TETRA_CORR)
       Energy(1) = bz%Fermi
       Call t_DOS(1, Energy, DOS)
    Case (BZ_QUAD_TETRA)
       write(*,*) 'bz_DOS: DOS not currently implemented with BZ_QUAD_TETRA!!'
       DOS(1) = 0
    Case (BZ_GAUSS)
       write(*,*) 'bz_DOS: DOS not currently implemented with BZ_GAUSS!!'
       DOS(1) = 0
  End Select

  bz_DOS = DOS(1)
  Return
End Function


!*****************************************************************************
!
! bz_NOS - Calculates the Number of States
!
!  BZ  - Data structure to store all the info
!
!*****************************************************************************

Real Function bz_NOS(bz)
  Implicit NONE
  Type (bz_struct),  Intent(INOUT) :: bz

  Real :: Energy(1), NOS(1)

!Write(*,*) 'bz_NOS: Fermi=',bz%Fermi
  Select Case (bz%Method)
    Case (BZ_LINEAR_TETRA_BASE)
       Energy(1) = bz%Fermi
       Call t_NOS(1, Energy, NOS)
    Case (BZ_LINEAR_TETRA_CORR)
       Energy(1) = bz%Fermi
       Call t_NOS(1, Energy, NOS)
    Case (BZ_QUAD_TETRA)
       NOS(1) = qt_NOS(bz%Fermi)
!write(*,*) ' bz_NOS: qt_NOS = ',NOS(1)
    Case (BZ_GAUSS)
       NOS(1) = Gauss_NOS(bz)
  End Select

  bz_NOS = NOS(1)
  Return
End Function

!*****************************************************************************
!
!  bz_Occupy - Calculates the occupancies of the system
!  
!  BZ  - Data structure containing BZ info
!  Occ - Array containing the Unique occupancies
!
!*****************************************************************************

Subroutine BZ_Occupy(BZ, Occ)
  Type (bz_struct),  Intent(INOUT) :: bz
  Real,              Intent(OUT)   :: Occ(:,:)

  Real :: nos

  Occ = 0

  Select Case (bz%Method)
    Case (BZ_LINEAR_TETRA_BASE)
       Write(*,*) 'BZ_Occupy: Not implemented for BZ_LINEAR_TETRA_BASE!'
       STOP
    Case (BZ_LINEAR_TETRA_CORR)
       Write(*,*) 'BZ_Occupy: Not implemented for BZ_LINEAR_TETRA_CORR!'
       STOP
    Case (BZ_QUAD_TETRA)
       Write(*,*) 'BZ_Occupy: Not implemented for BZ_QUAD_TETRA!'
       STOP
    Case (BZ_GAUSS)
       nos = Gauss_NOS(BZ, Occ)
  End Select

  Return
End Subroutine

!*****************************************************************************
!
! bz_Weights - Calculates the Integration Weights
!
!  BZ  - Data structure to store all the info
!
!*****************************************************************************

Subroutine bz_Weights(bz)
  Implicit NONE
  Type (bz_struct),  Intent(INOUT) :: bz
  
  Select Case (bz%Method)
    Case (BZ_LINEAR_TETRA_BASE)
       Call t_CalcWeights(bz%Fermi, TETRA_BASE_ONLY)
    Case (BZ_LINEAR_TETRA_CORR)
       Call t_CalcWeights(bz%Fermi,TETRA_BOTH)
    Case (BZ_QUAD_TETRA)
       Call qt_Weights(bz%fermi)
    Case (BZ_GAUSS)
       Call Gauss_Weights(bz)
  End Select

  Return
End Subroutine

!*****************************************************************************
!
!  bz_FuncDer - Function for calculating the generic NOS and DOS.
!               Internal Use Only.
!  
!  x     - Energy
!  fn    - NOS
!  deriv - DOS
!
!*****************************************************************************

Real Function bz_FuncDer(x, fn, deriv)
   Implicit NONE
   Real, Intent(IN)  :: x
   Real, Intent(OUT) :: fn
   Real, Intent(OUT) :: deriv

   local_bz%Fermi = x
   fn = bz_NOS(local_bz) - local_bz%Electrons
   deriv = bz_DOS(local_bz)

   bz_FuncDer = fn
   Return
End Function

!*****************************************************************************
!
!  bz_Func - Function for calculating the generic NOS and DOS.
!            Internal Use Only.
!  
!  x - Fermi Energy
!
!*****************************************************************************

Real Function bz_Func(x)
   Implicit NONE
   Real, Intent(IN)  :: x

   
   local_bz%Fermi = x
   bz_Func = bz_NOS(local_bz) - local_bz%Electrons

   Return
End Function

!*****************************************************************************
!
!  bz_CalcFermi - Calculates the Fermi Energy Surface given the number
!                 of Electrons.
!
!  BZ      - Data structure to store all the info
!  NumFunc - (OPTIONAL) Returns the number of Function evaluations
!
!  Return Values
!     Returns the Fermi Energy AND sets the Fermi Energy for subsequent
!     calculations
! 
!*****************************************************************************

Real Function bz_CalcFermi(bz, NumFunc)
  Implicit NONE
  Type (bz_struct),  Intent(INOUT) :: bz
  Integer, Optional, Intent(OUT)   :: NumFunc

  Integer :: i, j, k


  local_bz = bz

  bz%bz_Low = MinVal(Bz%Keu);   bz%bz_High = MaxVal(Bz%Keu);
  i = brak_grow(bz_Func, bz%bz_Low, bz%bz_High, j)

  if (i==0) then
     write(*,*) 'bz_CalcFermi: Could not bracket root!' 
     write(*,*) 'bz_CalcFermi: After Brak_Grow----------------------'
     write(*,*) 'bz_CalcFermi: Range:',bz%bz_Low, ' * ',bz%bz_High, ' * ',bz%Accuracy
     Write(*,*) 'bz_CalcFermi: MaxITer = ', bz%MaxIter
  End If

  Select Case (bz%Method)
    Case (BZ_LINEAR_TETRA_BASE)
      bz%Fermi =rt_Brent(bz_Func,bz%bz_Low,bz%bz_High,bz%Accuracy,bz%MaxIter,k)
    Case (BZ_LINEAR_TETRA_CORR)
      bz%Fermi =rt_Brent(bz_Func,bz%bz_Low,bz%bz_High,bz%Accuracy,bz%MaxIter,k)
    Case (BZ_QUAD_TETRA)
      bz%Fermi =rt_Brent(bz_Func,bz%bz_Low,bz%bz_High,bz%Accuracy,bz%MaxIter,k)
    Case (BZ_GAUSS)
      bz%Fermi =rt_Brent(bz_Func,bz%bz_Low,bz%bz_High,bz%Accuracy,bz%MaxIter,k)
  End Select

  bz_CalcFermi = bz%Fermi

  if (PRESENT(NumFunc)) NumFunc = j+k

  Return
End Function


!*****************************************************************************
!
!  bz_Integrate - Calculates the weighted integral of the function over the 
!                 BZ.  Assumes the weights have already been calculated
!
!  BZ - Data structure to store all the info
!  Fn - Function evaluated at each k pnt and band, must have same dim as Ke.  
!
!*****************************************************************************

Real Function bz_Integrate(bz, Fn)
   Implicit NONE
   Type (bz_struct),  Intent(IN) :: bz
   Real,              Intent(IN) :: Fn(:,:)

   bz_Integrate = Sum(Fn*bz%Weight)
 
   Return
End Function

End Module
