!******************************************************************************
!
! File : charges.f90
!   by : Natalie Holzwarth
!   on : 9/2/99
!  for : PAW program
!
! Module for calculating grids and integrals of charge density
!
! Updated for spin by Ping Tang and N. A. W. Holzwarth
!   Last modified on 5/23/05
!******************************************************************************

Module charges

Use mathlib
Use paw_inout
Use atom_data
Use crystal_data
Use doijmatrix
Use gpoints
Use options_data
Use projectors
Use mem_data
Use memmgr
Use word
Use relaxsys
Use strings

Implicit NONE!!!!!!!!!!!!!!

!******************************************************************************
Contains
!******************************************************************************



!******************************************************************************
!
!  Bandcharge - Calculates charge in spheres for partial density of  
!           states analysis.  If sphere contains an atomic site, radius
!           must be at least rc.  (Should be checked before calling this
!           program.) 
!    
!  Input:
!    Nsphere -- number of spheres  (0 for DOS output only)
!    Iter -- number of iterations for RelaxElectron call
!    Tol -- tolerance of error for RelaxElectron call
!   For Nsphere > 0,  the following variables are read in:
!    Asite -- logical array; true if sphere is on atom site
!    Map  -- integer array; if Asite(i)=.true. then Map(i)=atom index       
!    Centers -- real array; Center(1:3,i) = sphere center in fractional coord
!                           Center(4,i) = sphere radius 
!             Note: If Asite(i)=.true., Center(4,i)>=Rc
!                   If Asite(i)=.false., only plane wave charge is calculated
!     In the input paw file after the CALCULATE PDOS (Nsphere #), the file
!      should contain the following Nsphere entries:  (Note -- this program
!      is not designed to combine charges for more than one center.  If
!      crystal has more than one equivalent site, one should calculate
!      the charge within each and add them together in the analysis pgm.)
!
!    For each sphere, list:
!     Keyword
!      If Keyword=atom, then list:
!        Name, radius    -- where Name corresponds to Atom_list(i)%Name
!                      and where radius >= AtomType_Info(k)%Rc
!      If Keyword/=atom, then list:
!        Fractional position, radius  (no checking done)
!    
!   Input should end with "End"
!
!  Output:
!    Results written to file
!
!******************************************************************************

Subroutine Bandcharge(Nsphere,WC)
   Integer,          Intent(IN)  :: Nsphere
   Type (Word_Context), Intent(INOUT)    :: WC

!   Integer             :: Iter,Map(Nsphere)
!   Logical             :: Asite(Nsphere) , tmplog, EConverged
!   Real                :: Center(4,Nsphere),Tol,Etol
!   Real                :: Charge(Nsphere),wgt
!   Real                :: G(3), GM, BlochK(3), con,arg ,R(3,Nsphere)        
!   Integer             :: Kpnt, PsiB, PsiK,  Psi_toUse(Mem_Mapsize)
   Integer             :: Iter
   Integer,allocatable  :: Map(:)
   Logical             :: tmplog, EConverged
   Logical,allocatable  :: Asite(:)
   Real                :: Tol,Etol,wgt,G(3),GM,BlochK(3),con,arg
   Real   ,allocatable  :: Center(:,:),Charge(:),R(:,:)        
   Integer             :: Kpnt, PsiB, PsiK
   Integer,allocatable  :: Psi_toUse(:)
   Integer          :: gi, N, j,k, t, tlen ,i , ib
   Integer          :: atom,nili,njlj,li,lj,mi,mj
   Complex          :: zz
   Type (Mem_handle) , pointer :: psi 
   Complex, pointer :: WK(:),PDOT(:) ,psi1(:)
   Character*100    :: keyword,token

   Allocate(Map(Nsphere),Asite(Nsphere),Center(4,Nsphere),R(3,Nsphere))
   Allocate(Charge(Nsphere),Psi_ToUse(Mem_Mapsize))


   If (Bandstructure_mode) then
      Write(Error_Unit,*) &
       'Error in Bandcharges -- not yet programmed for Band Structure mode'
       stop
   Endif

     Bandstructure_mode=.true.
     DoLCAO=0
     Call GetNumber(WC,Iter)
     If (Iter.lt.3) Iter=10
     Call GetNumber(WC,Tol)
     If (Tol.lt.1.e-13) Tol=0.001

     write(Log_Unit,*) 'Calling RelaxElectrons with iter,tol,minbands=',&
            Iter, Tol, MinBands
     write(Log_Unit,*) 'Eigen_MAX =', Eigen_MAX

     Call RelaxElectrons(Iter, Tol, Etol, EConverged)         

   t = 0
   Do Kpnt=1,NumKpnts
      Do i=1,Mem_MapSize
        If (Kpnt==PsiInfo(i)%Kpnt.and.PsiInfo(i)%Energy.le.Eigen_Max) t=t+1
      Enddo
   Enddo

   write(CHARGEOUT_UNIT,*) t    ! list total number of bands in output

   If (Nsphere.gt.0) then

     Asite=.true.; Center=0;  Map=0


 ! Input for partial density information
     i=0
     Do
       Call GetNextWord(WC,keyword,tlen)
       Call UpperCase(keyword)

       If (trim(keyword) == "END") exit

        i=i+1
        if (i.gt.Nsphere) then
         write(Error_Unit,*) 'error in bandcharge -- too many spheres',i,Nsphere
         stop
        endif
       If (trim(keyword) == "ATOM") then
         Asite(i)=.true.
         Call GetNextWord(WC,keyword,tlen)
           if (tlen<=0) then
            write(Error_Unit,*)& 
               'error in bandcharge -- next item should be atom label'
            stop
           endif
         do j=1,Specific_Atoms
           if (trim(keyword)==trim(Atom_list(j)%Name)) then
           Map(i)=j
           Center(1:3,i)=Atom_list(j)%Frac_Pos
           Call GetNumber(WC, con)
           Center(4,i)=MAX(con,AtomType_Info(Atom_list(j)%TypeIndex)%Rc)
           write(CHARGEOUT_UNIT,'(i4," atom = ",a5,5x,"radius = ",1pe15.7)') &
            i, trim(keyword),Center(4,i)
           write(LOG_UNIT,'(i4," atom = ",a5,5x,"radius = ",1pe15.7)') &
            i, trim(keyword),Center(4,i)
           Exit
           Endif
         Enddo

         if (Map(i).eq.0) then
           write(Error_Unit,*)&
             'error in bandcharge -- rc not correct',i,con
           stop
         endif
       else
         Asite(i)=.false.
         Call GetNumbers(WC,Center(1:4,i))
         write(CHARGEOUT_UNIT,'(i4," interstial site at ",1p3e15.7," radius = ",1pe15.7)') i,Center(1:4,i)
     Endif

    Enddo
    if (i.ne.Nsphere) then
      write(error_unit,*) 'input error in bandcharge', i, Nsphere
      stop
    endif

     Allocate(WK(FFT_Grid(4,G_HIGH)),stat=j)
      if (j.ne.0) then
       write(Error_Unit,*) 'error allocating work array in bandcharge', &
            FFT_Grid(4,G_HIGH)
       stop
      endif

     Do i=1,Nsphere
     R(1:3,i)=MATMUL(Xtal%Basis,Center(1:3,i)) !Change to Cartesian Coord.
     Enddo
   Endif


   
   N=Gpnt_Size(G_HIGH)-1


   Do Kpnt=1,NumKpnts

      BlochK = BZ%Ku(:,Kpnt)
      wgt=bz%WtUniq(1,Kpnt)

      Psi_toUse=MH_Skip
      Where (PsiInfo(:)%Kpnt==Kpnt) Psi_toUse=MH_toProcess
      Call Phase_Generic(Psi_toUse,Kpnt)
      Call GetNextPsi(ib,Psi)

      Do While (ib>0)

        i=Psi%Index
        PsiInfo(i)%DoSave=1
        PsiInfo(i)%Available=Mem_Used

        If(PsiInfo(i)%Energy.le.Eigen_Max) then
        If(.not.spindependence) then
          write(CHARGEOUT_UNIT,'(1p5e15.7)') BlochK,wgt,PsiInfo(i)%Energy
        else
          if(PsiInfo(i)%spinup) &
           write(CHARGEOUT_UNIT,'(1p5e15.7,"  up")') BlochK,wgt,PsiInfo(i)%Energy
          if(.not.PsiInfo(i)%spinup) &
           write(CHARGEOUT_UNIT,'(1p5e15.7,"  dn")') BlochK,wgt,PsiInfo(i)%Energy
        endif
        If (Nsphere.gt.0) then

           Psi1=>Psi%Ptr
           Call FilterValue(Psi1)

           WK = 0

           WK(FFTMap_High(1:Gpnt_Size(G_Low))) = Psi1(1:Gpnt_Size(G_Low))
             t = Gpnt_Size(G_Low)+1
             j = Gpnt_Size(G_HIGH)+1;   k = j + (Gpnt_Size(G_Low)-1) - 1
           WK(FFTMap_High(j:k)) = Psi1(t:)
         
           Call PerformFFT(FFT_TO_R, G_HIGH, WK)
         
           WK =  CONJG(WK)*WK
         
           Call PerformFFT(FFT_TO_G, G_HIGH, WK)

           WK = Four_Pi * WK / xtal%volume

      ! calculate plane wave contributions to charges
         
           Do j=1,Nsphere
             charge(j)=(center(4,j)**3)*WK(FFTMap_High(1))/3
           Enddo
             

           Do gi=2,Gpnt_Size(G_HIGH)
               GM=Gpnt(4,gi)
               G=Gpnt(1:3,gi)
               do j=1,Nsphere
                  arg=GM*center(4,j)
                  con=(sin(arg)-arg*cos(arg))/(GM**3)
                  arg=Dot_product(G,R(1:3,j))
                  zz=cmplx(cos(arg),sin(arg))
                  charge(j)=charge(j) + con*zz*WK(FFTmap_High(gi)) &
                              + con*conjg(zz)*WK(FFTmap_High(N+gi))
               Enddo
             Enddo

      ! write(CHARGEOUT_UNIT,'(1p5e15.7)') (charge(j),j=1,Nsphere)  !remove this

       ! Correct for charges within atomic spheres
           
           PDOT =>PsiInfo(i)%PDOT

           Call CalcProjProducts(psi1,PDOT)

             Do j=1,Nsphere

             If (Asite(j)) then
              atom = Map(j) 

              do t=1,PLM_MAX
                 if (atom.eq.PLM(1,t))  then
                    nili=PLM(2,t)
                    li  =PLM(3,t)
                    mi  =PLM(4,t)

                    zz=0
                    Do k=PLM_AtomRange(1,atom),PLM_AtomRange(2,atom)
                    njlj=PLM(2,k)
                    lj  =PLM(3,k)
                    mj  =PLM(4,k)

                    
                    zz=zz+OijMatrix(atom,nili,njlj,li,lj,mi,mj)*PDOT(k)
                    Enddo

                    charge(j)=charge(j)+conjg(PDOT(t))*zz
                  Endif
              Enddo
            EndIf
                   
            Enddo

        write(CHARGEOUT_UNIT,'(1p5e15.7)') (charge(j),j=1,Nsphere)

        Endif  ! Nsphere > 0
 
       Endif  ! Energy<=Eigen_Max

     Call GetNextPsi(ib,Psi)

    Enddo    !ib
  Enddo    !Kpnt

 Bandstructure_mode = .false.
 DeAllocate(WK)
 DeAllocate(Map,Asite,Center,R,Psi_ToUse,charge)

 Return
End Subroutine

!*****************************************************************************
!
! Partialden
!
!  Input:
!    Nranges -- number of energy ranges
!    Iter -- number of iterations for RelaxElectron call
!    Tol -- tolerance of error for RelaxElectron call
!
!    For each range, list:
!      OCC 
!    /      filename     ebegin   eend 
!    \
!      NOT   
!     where OCC (or occupied) means program should use stored occupancy
!        and NOT (or any other keyword (no spaces) means prgram uses
!        BZ weight factor times full occupancy
!    
!   Input should end with "End"
!
!  Output:
!    Results written to file --
!     If spindependence, write spinup results then spindown results
!
!******************************************************************************
Subroutine Partialden(Nranges,WC)
   Integer , Intent(IN) :: Nranges
   Type (Word_Context), Intent(INOUT) :: WC

   Integer       :: Iter
   Integer       :: Kpnt, PsiB, PsiK,  Psi_toUse(Mem_Mapsize)
   Integer       :: gi, N, j,k, t, tlen ,i , ib ,irange ,a , Base
   Real          :: Tol,Etol, Ebegin(Nranges),Eend(Nranges)
   Real          :: wgt,charge
   Real          :: G(3), GM, BlochK(3), con,arg
   Complex          :: zz
   Type (Mem_handle) , pointer :: psi 
   Complex, pointer :: WK(:),DEN(:),PDOT(:) ,psi1(:), Wij(:,:)
   Complex, pointer :: DENspin(:),Wijspin(:,:)
   Character*100 :: Filename(Nranges)
   Character*100 :: token,keyword
   Logical       :: Useocc(Nranges),tmplog, EConverged


     Call GetNumber(WC,Iter)
     If (Iter.lt.3) Iter=10
     Call GetNumber(WC,Tol)
     If (Tol.lt.1.e-13) Tol=0.001

     write(Log_Unit,*) 'Entering Partialden with Nranges,Iter,tol',Nranges,Iter,tol

!Input for partial density ranges
     i=0
     Do
       Call GetNextWord(WC,keyword,tlen)
       Call UpperCase(keyword)

       If (trim(keyword) == "END") exit

        i=i+1
        if (i.gt.Nranges) then
         write(Error_Unit,*) 'error in partialden -- too many ranges',i,Nranges
         stop
        endif

        Useocc(i)=.false.
        If (keyword(1:3) == "OCC")  Useocc(i)=.true.
        Call GetNextWord(WC,keyword,tlen)
        If (tlen.le.0) then
         write(error_unit,*) 'error in partialden -- no file name'
         stop
        Endif
        Filename(i)=keyword
        Call GetNumber(WC,Ebegin(i))
        Call GetNumber(WC,Eend(i))

        write(Log_Unit,'("Range ",i3," file = ",a20," range = ",1p2e15.7)') &
           i, trim(Filename(i)),Ebegin(i),Eend(i)
        write(Output_Unit,'("Range ",i3," file = ",a20," range = ",1p2e15.7)') &
           i, trim(Filename(i)),Ebegin(i),Eend(i)
        If (Eend(i) < Ebegin(i)) then
           write(error_unit,*) 'error in partialden -- eend < ebegin'
           stop
        Endif
        If (Eend(i) > Eigen_Max) Eigen_Max=Eend(i)

    Enddo
    if (i.ne.Nranges) then
      write(error_unit,*) 'input error in partialden', i, Nranges
      stop
    endif


   If (Bandstructure_mode) then
      Write(Error_Unit,*) &
       'Error in Bandcharges -- not yet programmed for Band Structure mode'
       stop
   Endif

     Bandstructure_mode=.true.
     DoLCAO=0

     write(Log_Unit,*) 'Calling RelaxElectrons with iter,tol,minbands=',&
            Iter, Tol, MinBands
     write(Log_Unit,*) ' Eigen_Max = ',Eigen_Max

     Call RelaxElectrons(Iter, Tol, Etol, EConverged)         


     Allocate(WK(FFT_Grid(4,G_HIGH)),DEN(FFT_Grid(4,G_HIGH)),stat=j)
     If (spindependence) Allocate(DENspin(FFT_Grid(4,G_HIGH)),stat=j)
      if (j.ne.0) then
       write(Error_Unit,*) 'error allocating work array in partialden', &
            FFT_Grid(4,G_HIGH)
       stop
      endif


   
   N=Gpnt_Size(G_HIGH)-1

 Do irange=1,Nranges
   Open(unit=CHARGEOUT_UNIT,file=trim(Filename(irange)),form='unformatted')
   Write(CHARGEOUT_UNIT) Ebegin(irange),Eend(irange),Useocc(irange)
  If (spindependence) then
   Open(unit=SCRATCH_UNIT,file=trim(Filename(irange))//'spindn',form='unformatted')
   Write(SCRATCH_UNIT) Ebegin(irange),Eend(irange),Useocc(irange)
  endif

   DEN=0
   if (spindependence) DENspin=0
   Do a = 1 ,Specific_Atoms
     Atom_List(a)%Wij = 0
     if (spindependence) Atom_List(a)%Wijspin = 0
   Enddo

   Do Kpnt=1,NumKpnts

      Psi_toUse=MH_Skip
      If (Useocc(irange)) then
        Where ((PsiInfo(:)%Kpnt == Kpnt) .AND. (PsiInfo(:)%Occupancy >1E-10))
         Psi_toUse = MH_toProcess
        End Where
      Else                         
         wgt=bz%WtUniq(1,Kpnt)
         Where (PsiInfo(:)%Kpnt==Kpnt) Psi_toUse=MH_toProcess
      Endif
      Call Phase_Generic(Psi_toUse,Kpnt)
      Call GetNextPsi(ib,Psi)

      Do While (ib>0)

        i=Psi%Index
        PsiInfo(i)%DoSave=1
        PsiInfo(i)%Available=Mem_Used

        If (Useocc(irange)) wgt=PsiInfo(i)%Occupancy

        If(PsiInfo(i)%Energy.le.Eend(irange).AND.  & 
          PsiInfo(i)%Energy.ge.Ebegin(irange).AND. wgt>1.e-10) then


           Psi1=>Psi%Ptr
           Call FilterValue(Psi1)

           WK = 0

           WK(FFTMap_High(1:Gpnt_Size(G_Low))) = Psi1(1:Gpnt_Size(G_Low))
             t = Gpnt_Size(G_Low)+1
             j = Gpnt_Size(G_HIGH)+1;   k = j + (Gpnt_Size(G_Low)-1) - 1
           WK(FFTMap_High(j:k)) = Psi1(t:)
         
           Call PerformFFT(FFT_TO_R, G_HIGH, WK)
         
           If(.not.spindependence.or.(spindependence.and.PsiInfo(i)%spinup)) then
             DEN = DEN + wgt*CONJG(WK)*WK
           else
             DENspin=DENspin + wgt*CONJG(WK)*WK
           endif

             PDOT =>PsiInfo(i)%PDOT
             Call CalcProjProducts(psi1,PDOT)

             Do a=1, Specific_Atoms
              If(.not.spindependence.or.(spindependence.and.PsiInfo(i)%spinup)) then
                Wij => Atom_List(a)%Wij
              else
                Wij => Atom_List(a)%Wijspin
              endif
                Base = PLM_AtomRange(1,a) - 1
                Do t=PLM_AtomRange(1,a), PLM_AtomRange(2,a)
                   Do j=PLM_AtomRange(1,a), PLM_AtomRange(2,a)
                      Wij(t-Base, j-Base) = Wij(t-Base, j-Base) &
                        + CONJG(PDot(t))*PDot(j)*wgt
                   End Do
                End Do
             End Do
 
       Endif  ! Energy range       

     Call GetNextPsi(ib,Psi)

    Enddo    !ib
  Enddo    !Kpnt

  If (xtal%Rot_Size > 0) Call SymRho( Den, WK)
  Call PerformFFT(FFT_TO_G, G_HIGH, Den)
  WK=0
  WK(1:Gpnt_Size(G_High))=Den(FFTmap_High(1:Gpnt_Size(G_High)))
  If (xtal%Rot_Size > 0)  Call SymWij

  write(CHARGEOUT_UNIT) (WK(i),i=1,Gpnt_Size(G_High))
  do a=1,Specific_Atoms
    Wij => Atom_List(a)%Wij
    t= PLM_AtomRange(2,a) - PLM_AtomRange(1,a) + 1
    write(CHARGEOUT_UNIT) ((Wij(i,j),i=1,t),j=1,t)
  enddo
  Close(CHARGEOUT_UNIT)

  if (spindependence) then
    If (xtal%Rot_Size > 0) Call SymRho( Denspin, WK)
    Call PerformFFT(FFT_TO_G, G_HIGH, Denspin)
    WK=0
    WK(1:Gpnt_Size(G_High))=Denspin(FFTmap_High(1:Gpnt_Size(G_High)))
    !If (xtal%Rot_Size > 0)  Call SymWij

    write(SCRATCH_UNIT) (WK(i),i=1,Gpnt_Size(G_High))
    do a=1,Specific_Atoms
      Wij => Atom_List(a)%Wijspin
      t= PLM_AtomRange(2,a) - PLM_AtomRange(1,a) + 1
      write(SCRATCH_UNIT) ((Wij(i,j),i=1,t),j=1,t)
    enddo
  endif
  Close(SCRATCH_UNIT)

 Enddo   !irange
 Bandstructure_mode = .false.
 DeAllocate(WK,DEN)
 If (spindependence) Deallocate(DENspin)

End Subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!  Subroutine Ld3ddata
!
!   Read in axis and grid information for 3-d plots
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

Subroutine Ld3ddata(WC)
  Type (Word_Context), Intent(INOUT) :: WC
 
  Character*100 :: token,token2
  Integer       :: Tlen,icount
  Real          :: Scale,center(3)
  Logical       :: Readcenter

  Scale = 1 
  icount=0
  Readcenter = .false.
  Volumeplot%Origin = 0


   Loop : Do 
    Call GetNextWord(WC, token, Tlen)
    Call UpperCase(Token)
    If (Trim(token)=="SCALE") then
       Call GetNumbers(WC,Scale)
       If (Scale<1.e-8) Scale = 1
    Else if (token(1:1)=="X") then
       Call GetNumbers(WC,Volumeplot%X)  !x*Scale = Cartesian Bohr Units
       icount=icount+1
    Else If (token(1:1)=="Y") then
       Call GetNumbers(WC,Volumeplot%Y)  !y*Scale = Cartesian Bohr Units
       icount=icount+1
    Else If (token(1:1)=="Z") then
       Call GetNumbers(WC,Volumeplot%Z)  !z*Scale = Cartesian Bohr Units
       icount=icount+1
    Else If (token(1:1)=="O") then
       Call GetNumbers(WC,Volumeplot%ORIGIN)  !o*Scale = Cartesian Bohr Units
       icount=icount+1
    Else If (Trim(token)=="CENTER") then  ! specify center of plotting cell
       Call GetNumbers(WC,center)  !center*Scale = Cartesian Bohr Units
       icount=icount+1
       Readcenter=.true.
    Else If (Trim(token)=="BOND") then
       Call GetNumbers(WC,Volumeplot%maxbond)  !b*Scale = Cartesian Bohr Units
       icount=icount+1
    Else If (Trim(token)=="BONDTOL") then
       Call GetNumbers(WC,Volumeplot%bondtol)  !% bond length for drawing bond
    Else If (Trim(token)=="GRID") then
       Call GetNumbers(WC,Volumeplot%grid(1:3)) !Number of grid points
       icount=icount+10
    Else If (Trim(token)=="PLOTNAME") then
       Call GetNextWord(WC, token2, tlen)
       If (tlen>0) Plot_Name=Trim(token2)
    Else If (Trim(token)=="END") then
       exit loop
    Else
      Write(Error_unit,*) 'Error in plot input -- ',Token
      stop
    Endif
 
  Enddo  Loop

  Volumeplot%query=Max(icount,Volumeplot%query)

  If (Readcenter) then
      Volumeplot%Origin=Center-0.5*(Volumeplot%X+Volumeplot%Y+Volumeplot%Z)
  EndIf

  Volumeplot%X=Volumeplot%X*Scale
  Volumeplot%Y=Volumeplot%Y*Scale
  Volumeplot%Z=Volumeplot%Z*Scale
  Volumeplot%Origin=Volumeplot%Origin*Scale
  Volumeplot%maxbond=Volumeplot%maxbond

End subroutine Ld3ddata

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!  Subroutine Ld2ddata
!
!   Read in axis and grid information for 2-d plots
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

Subroutine Ld2ddata(WC)
  Type (Word_Context), Intent(INOUT) :: WC
 
  Character*100 :: token,token2
  Integer       :: Tlen,icount
  Real          :: Scale

  Scale = 1
  icount=0


   Loop : Do 
    Call GetNextWord(WC, token, Tlen)
    Call UpperCase(Token)
    If (Trim(token)=="SCALE") then
       Call GetNumbers(WC,Scale)
       If (Scale<1.e-8) Scale = 1
    Else if (token(1:1)=="X") then
       Call GetNumbers(WC,Planeplot%X)  !x*Scale = Cartesian Bohr Units
       icount=icount+1
    Else If (token(1:1)=="Y") then
       Call GetNumbers(WC,Planeplot%Y)  !y*Scale = Cartesian Bohr Units
       icount=icount+1
    Else If (token(1:1)=="O") then
       Call GetNumbers(WC,Planeplot%ORIGIN)  !o*Scale = Cartesian Bohr Units
       icount=icount+1
    Else If (Trim(token)=="GRID") then
       Call GetNumbers(WC,Planeplot%grid(1:2)) !Number of grid points
       icount=icount+10
    Else If (Trim(token)=="PLOTNAME") then
       Call GetNextWord(WC, token2, tlen)
       If (tlen>0) Plot_Name=Trim(token2)
    Else If (Trim(token)=="END") then
       exit loop
    Else
      Write(Error_unit,*) 'Error in plot input -- ',Token
      stop
    Endif
 
  Enddo  Loop

  Planeplot%query=Max(icount,Planeplot%query)
  Planeplot%X=Planeplot%X*Scale
  Planeplot%Y=Planeplot%Y*Scale
  Planeplot%Origin=Planeplot%Origin*Scale

End subroutine Ld2ddata

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! Subroutine Atomplot3d
!
!   For given unit cell -- specified by Volumeplot%X,Y,Z,Origin,maxbond
!                             (input from read_input)
!    Calculate atomic positions and bonds for Ball and Stick model
!         (Data Explorer output and XCrysden output)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

Subroutine Atomplot3d

 Real, allocatable :: ballpoints(:,:)
 Real :: resx, resy, resz ,plotmat(3,3),solvmat(3,3),v(3),vin(3),vout(3)
 Real :: chk,rtol,tx,ty,tz,XA,YA,ZA
 Real, parameter :: atol=1.e-3
 Integer, allocatable :: noballs(:),z(:)
 Integer :: ncount,ncountmax,many,ierr,npair,i,j,k,n

! file for plotting axes
  If (volumeplot%query.lt.5) then
    write(error_unit) 'error in 3d plot -- need to read in data'
    stop
  endif

  rtol=volumeplot%bondtol  !Tolerance of including atoms

  rtol=MAX(atol,rtol)

! Determine length of plotting cell boundaries
    resx=SQRT(DOT_Product(volumeplot%X,volumeplot%X))
    resy=SQRT(DOT_Product(volumeplot%Y,volumeplot%Y))
    resz=SQRT(DOT_Product(volumeplot%Z,volumeplot%Z))
    XA=resx*0.52917720859
    YA=resy*0.52917720859
    ZA=resz*0.52917720859


  Open(unit=CHARGEOUT_UNIT,&
   file=Trim(PLOT_Name)//".axes.dx",form="formatted")

   write(CHARGEOUT_UNIT, &
     '("object 1 class array type float rank 0 items 8 data follows")')
   write(CHARGEOUT_UNIT,'("  1  1  1  1  1  1  1  1")')
   write(CHARGEOUT_UNIT,'("attribute ""dep"" string ""positions""")')
   write(CHARGEOUT_UNIT,'("#")')
   write(CHARGEOUT_UNIT,&
    '("object 2 class array type float rank 1 shape 3 items 8 data follows")')

   v=0
   write(CHARGEOUT_UNIT,'(1p3e15.7)') v(1:3)
   write(CHARGEOUT_UNIT,'(1p3e15.7)') XA,v(2),v(3)
   write(CHARGEOUT_UNIT,'(1p3e15.7)') v(1),YA,v(3)
   write(CHARGEOUT_UNIT,'(1p3e15.7)') XA,YA,v(3)
   write(CHARGEOUT_UNIT,'(1p3e15.7)') v(1),v(2),ZA
   write(CHARGEOUT_UNIT,'(1p3e15.7)') XA,V(2),ZA
   write(CHARGEOUT_UNIT,'(1p3e15.7)') v(1),YA,ZA
   write(CHARGEOUT_UNIT,'(1p3e15.7)') XA,YA,ZA

   write(CHARGEOUT_UNIT,'("attribute ""dep"" string ""positions""")')
   write(CHARGEOUT_UNIT,'("#")')

   write(CHARGEOUT_UNIT, &
    '("object 3 class array type int rank 1 shape 2 items 12 data follows")')
   write(CHARGEOUT_UNIT, '("  0  1")')
   write(CHARGEOUT_UNIT, '("  0  2")')
   write(CHARGEOUT_UNIT, '("  1  3")')
   write(CHARGEOUT_UNIT, '("  2  3")')
   write(CHARGEOUT_UNIT, '("  4  5")')
   write(CHARGEOUT_UNIT, '("  4  6")')
   write(CHARGEOUT_UNIT, '("  5  7")')
   write(CHARGEOUT_UNIT, '("  6  7")')
   write(CHARGEOUT_UNIT, '("  0  4")')
   write(CHARGEOUT_UNIT, '("  1  5")')
   write(CHARGEOUT_UNIT, '("  2  6")')
   write(CHARGEOUT_UNIT, '("  3  7")')

   write(CHARGEOUT_UNIT,'("attribute ""element type"" string ""lines""")')
   write(CHARGEOUT_UNIT,'("attribute ""ref"" string ""positions""")')
   write(CHARGEOUT_UNIT,'("#")')
   write(CHARGEOUT_UNIT,'("object ""molecule"" class field")')
   write(CHARGEOUT_UNIT,'("component ""data"" value 1")')
   write(CHARGEOUT_UNIT,'("component ""positions"" value 2")')
   write(CHARGEOUT_UNIT,'("component ""connections"" value 3")')
   write(CHARGEOUT_UNIT,'("attribute ""name"" string ""molecule""")')
   write(CHARGEOUT_UNIT,'("#")')
   write(CHARGEOUT_UNIT,'("end")')

   close (CHARGEOUT_UNIT)

! file for atomic positions
  Open(unit=CHARGEOUT_UNIT,&
    file=Trim(PLOT_Name)//".atom.dx",form="formatted")


! Rough estimate of number atoms in plot

    many=resx*resy*resz/xtal%Volume+1
    ncountmax=8*(many**3)*Specific_Atoms 

  Allocate(noballs(ncountmax),ballpoints(3,ncountmax),z(ncountmax))

  noballs=0
  ballpoints=0


   ncount=0

   If (volumeplot%maxbond.gt.0.) then
     tx=rtol*volumeplot%maxbond/resx
     ty=rtol*volumeplot%maxbond/resy
     tz=rtol*volumeplot%maxbond/resz
   Else
     chk=xtal%volume**(0.33333)
     tx=rtol*chk/resx
     ty=rtol*chk/resy
     tz=rtol*chk/resz
   EndIf


   do n=1,Specific_Atoms      ! Some assumptions here about shape of cell

     do i = -many, many
        do j = -many, many
           do k = -many, many

           v = Atom_list(n)%Pos - Volumeplot%origin &
                 + MATMUL(xtal%Basis,(/i,j,k/))

           vout(1) = DOT_Product(Volumeplot%X,v)/resx
           vout(2) = DOT_Product(Volumeplot%Y,v)/resy
           vout(3) = DOT_Product(Volumeplot%Z,v)/resz

           vin(1)=vout(1)/resx
           vin(2)=vout(2)/resy
           vin(3)=vout(3)/resz

!  Check if atom is in plotting cell
           if (vin(1).gt.-tx.and.vin(1).lt.1+tx &
               .and.vin(2).gt.-ty.and.vin(2).lt.1+ty &
               .and.vin(3).gt.-tz.and.vin(3).lt.1+tz) then
               
              ncount=ncount+1 
              noballs(ncount)=Atom_list(n)%TypeIndex
              z(ncount)=AtomType_Info(noballs(ncount))%Atomic_Charge
              ballpoints(:,ncount)=vout

           Endif

           Enddo
         Enddo     
       Enddo
    Enddo

! Change atom positions to Angstrom Units

    ballpoints(:,1:ncount) = ballpoints(:,1:ncount)*0.52917720859

!  output atomic coordinates for Data Explorer
 
    write(CHARGEOUT_UNIT, &
     '("object 1 class array type float rank 0 items",i4," data follows")') &
      ncount
    write(CHARGEOUT_UNIT,*) (noballs(i),i=1,ncount)   

    write(CHARGEOUT_UNIT,'("attribute ""dep"" string ""positions""")')
    write(CHARGEOUT_UNIT,'("#")')
    write(CHARGEOUT_UNIT,&
     '("object 2 class array type float rank 1 shape 3 items",i4," data follows")') &
      ncount

         do i=1,ncount
          write(CHARGEOUT_UNIT,*) ballpoints(1:3,i)
          enddo

        write(CHARGEOUT_UNIT,'("attribute ""dep"" string ""positions""")')
        write(CHARGEOUT_UNIT,'("#")')


         chk=(Volumeplot%maxbond*0.52917720859)**2
         npair=0  
          do i=1,ncount-1
           do j=i+1,ncount
            v=ballpoints(:,i)-ballpoints(:,j)
            if(DOT_Product(v,v).le.chk) then
            npair=npair+1
            ! write(CHARGEOUT_UNIT,*) i,j
            endif
           enddo
          enddo

   write(CHARGEOUT_UNIT, &
   '("object 3 class array type int rank 1 shape 2 items",i4," data follows")')&
   npair 
          do i=1,ncount-1
           do j=i+1,ncount
            v=ballpoints(:,i)-ballpoints(:,j)
            if(DOT_Product(v,v).le.chk) then
             write(CHARGEOUT_UNIT,*) i-1,j-1
            endif
           enddo
          enddo

         write(CHARGEOUT_UNIT,'("attribute ""element type"" string ""lines""")')
         write(CHARGEOUT_UNIT,'("attribute ""ref"" string ""positions""")')
         write(CHARGEOUT_UNIT,'("#")')
         write(CHARGEOUT_UNIT,'("object ""molecule"" class field")')
         write(CHARGEOUT_UNIT,'("component ""data"" value 1")')
         write(CHARGEOUT_UNIT,'("component ""positions"" value 2")')
         write(CHARGEOUT_UNIT,'("component ""connections"" value 3")')
         write(CHARGEOUT_UNIT,'("attribute ""name"" string ""molecule""")')
         write(CHARGEOUT_UNIT,'("#")')
         write(CHARGEOUT_UNIT,'("end")')

      Close(CHARGEOUT_UNIT)



!  XCrysDen output
  Open(unit=CHARGEOUT_UNIT,&
    file=Trim(PLOT_Name)//".xsf",form="formatted")
  Write(CHARGEOUT_UNIT,'("CRYSTAL")')
  Write(CHARGEOUT_UNIT,'("PRIMVEC   1")')
  write(CHARGEOUT_UNIT,'(1p3e15.7)') Volumeplot%X(1:3)*0.52917720859
  write(CHARGEOUT_UNIT,'(1p3e15.7)') Volumeplot%Y(1:3)*0.52917720859
  write(CHARGEOUT_UNIT,'(1p3e15.7)') Volumeplot%Z(1:3)*0.52917720859
  Write(CHARGEOUT_UNIT,'("PRIMCOORD   1")')
  Write(CHARGEOUT_UNIT,'(i10  "   1")')  ncount

     do i=1,ncount
        j=z(i)+0.0001
        write(CHARGEOUT_UNIT,'(i8,2x,1p3e15.7)') j, ballpoints(1:3,i)
     enddo



End Subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! Subroutine Atomplot2d(ncount)
!
!   For given unit cell -- specified by Planeplot%X,Y,Origin
!                             (input from read_input)
!    Calculate atomic positions for Contour plot
!         (IDL format)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

Subroutine Atomplot2d(ncount)
 Integer, intent(OUT) :: ncount

 Real, allocatable :: ballpoints(:,:)
 Real :: resx, resy,plotmat(2,2),solvmat(2,2),v(3),vin(2),vout(2)
 Real :: chk,rtol,tx,ty,tz
 Real, parameter :: atol=1.e-3
 Integer, allocatable :: noballs(:)
 Integer :: ncountmax,many,ierr,npair,i,j,k,n

! file for plotting axes
  If (planeplot%query.lt.3) then
    write(error_unit) 'error in 2d plot -- need to read in data'
    stop
  endif

! file for atomic positions
  Open(unit=CHARGEOUT_UNIT,&
    file=Trim(PLOT_Name)//".atom.idl",form="formatted")


! Determine length of plotting cell boundaries
    resx=SQRT(DOT_Product(planeplot%X,planeplot%X))
    resy=SQRT(DOT_Product(planeplot%Y,planeplot%Y))

! Rough estimate of number atoms in plot

    many=resx*resy/xtal%Volume**0.66 +1
    many=10*many
    ncountmax=4*(many**2)*Specific_Atoms 

  Allocate(noballs(ncountmax),ballpoints(2,ncountmax))

  noballs=0
  ballpoints=0

! Form plot cell matrix
   plotmat=0
   plotmat(1,1)=DOT_Product(planeplot%X,planeplot%X)
   plotmat(2,2)=DOT_Product(planeplot%Y,planeplot%Y)
   plotmat(1,2)=DOT_Product(planeplot%X,planeplot%Y)
   plotmat(2,1)=plotmat(1,2)

   chk=plotmat(1,1)*plotmat(2,2)-plotmat(1,2)*plotmat(2,1)
   solvmat(1,1)=plotmat(2,2)/chk
   solvmat(2,2)=plotmat(1,1)/chk
   solvmat(1,2)=-plotmat(1,2)/chk
   solvmat(2,1)=-plotmat(2,1)/chk

   tx=atol*resx 
   ty=atol*resy

   ncount=0

   do n=1,Specific_Atoms      ! Some assumptions here about shape of cell

     do i = -many, many
        do j = -many, many
           do k = -many, many

           v = Atom_list(n)%Pos - Planeplot%origin &
                 + MATMUL(xtal%Basis,(/i,j,k/))

           vout(1) = DOT_Product(Planeplot%X,v)
           vout(2) = DOT_Product(Planeplot%Y,v)

           vin = MATMUL(solvmat,vout)

!  Check if atom is in plotting cell
           if (vin(1).gt.-tx.and.vin(1).lt.1+tx &
               .and.vin(2).gt.-ty.and.vin(2).lt.1+ty) then
               
              ncount=ncount+1 
              noballs(ncount)=Atom_list(n)%TypeIndex
              ballpoints(:,ncount)=vin

           Endif

           Enddo
         Enddo     
       Enddo
    Enddo


!  output atomic coordinates for IDL
 
    write(CHARGEOUT_UNIT, '(1p5e15.7)') (ballpoints(1,i),i=1,ncount)
    write(CHARGEOUT_UNIT, '(1p5e15.7)') (ballpoints(2,i),i=1,ncount)

      Close(CHARGEOUT_UNIT)
     Return


End Subroutine Atomplot2d

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! Subroutine Planegrid
!
!   For given unit cell -- specified by Volumeplot%X,Y,Origin
!                             (input from read_input)
!    Calculate density in 2 dimensional grid for Plot_Name file
!      (previously generated by partialden)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
Subroutine Planegrid

 Integer             :: nx,ny
 Integer             ::  many=10   ! how many possible unit cells to test
 Real, allocatable   ::  gridden(:,:)
 Real                ::  Ebegin, Eend,xleng,yleng,arg,CRC,dden,Rmag
 Real                ::  Xu(3),Yu(3),R(3),D(3),xm,ym,zm
 Complex, allocatable::  DEN(:),px(:),py(:)
 Complex             ::  zzz,zz1
 Integer             ::  i,j,k,a,ix,iy,lx,ly,t,natom
 Logical             ::  Useocc

   If (planeplot%query.lt.13.or.SUM(planeplot%grid(1:2)).lt.2) then
    write(error_unit,*) 'error in 2d plot -- need data'
    stop
   endif
   nx=planeplot%grid(1)
   ny=planeplot%grid(2)

! Check coordinates
   Xleng=sqrt(Dot_Product(planeplot%x,planeplot%x))
   Yleng=sqrt(Dot_Product(planeplot%y,planeplot%y))
   Write(Log_Unit,'(" Plot origin ",1p3E15.7)')planeplot%Origin
   Write(Log_Unit,'(" X-axis&length ",1p4E15.7)')planeplot%X,Xleng
   Write(Log_Unit,'(" Y-axis&length ",1p4E15.7)')planeplot%Y,Yleng
   Write(Log_Unit,'(" Nx, Ny        ",3i10    )')planeplot%grid(1:2)
   Write(Output_Unit,'(" Plot origin ",1p3E15.7)')planeplot%Origin
   Write(Output_Unit,'(" X-axis&length ",1p4E15.7)')planeplot%X,Xleng
   Write(Output_Unit,'(" Y-axis&length ",1p4E15.7)')planeplot%Y,Yleng
   Write(Output_Unit,'(" Nx, Ny        ",3i10    )')planeplot%grid(1:2)
   
   If (abs(Dot_Product(planeplot%x,planeplot%y)).gt.1.e-4) then
    write(Error_Unit,*) 'Error in planeplot- axes are not orthogonal'
    stop
   endif

   many= (Xleng*Yleng)/xtal%volume**0.66 + 3
   many= MAX(3,many)

   Write(Log_Unit,'(" Unit cell repeat factor = ",i10)') many
   Write(Output_Unit,'(" Unit cell repeat factor = ",i10)') many

   lx=nx-1
   ly=ny-1

   Xu=planeplot%X/lx ; xm=SQRT(Dot_Product(Xu,Xu))
   Yu=planeplot%Y/ly ; ym=SQRT(Dot_Product(Yu,Yu))
 

  Allocate(DEN(FFT_Grid(4,G_HIGH)),gridden(nx,ny),&
   px(nx),py(ny),stat=j)
   if (j.ne.0) then
    write(Error_Unit,*) 'error allocating work array in Planegrid', &
         FFT_Grid(4,G_HIGH),nx,ny
    stop
   endif

  Open(unit=CHARGEOUT_UNIT,file=Trim(PLOT_Name),form="unformatted")

  write(Log_Unit,*) ' Reading from partialden file ',Trim(PLOT_Name)
  write(Output_Unit,*) ' Reading from partialden file ',Trim(PLOT_Name)
  read(CHARGEOUT_UNIT) Ebegin,Eend,Useocc
  write(Log_Unit,*) 'Ebegin,Eend,Useocc= ', Ebegin,Eend,Useocc
  write(Output_Unit,*) 'Ebegin,Eend,Useocc= ', Ebegin,Eend,Useocc

  write(6,*) 'gpnt',Gpnt_size(G_HIGH)
  read(CHARGEOUT_UNIT) (DEN(i),i=1,Gpnt_Size(G_HIGH))
  do a=1,Specific_Atoms
   t= PLM_AtomRange(2,a) - PLM_AtomRange(1,a) + 1
   read(CHARGEOUT_UNIT) ((Atom_List(a)%Wij(i,j),i=1,t),j=1,t)
  enddo

 Close(CHARGEOUT_UNIT)


  Call Atomplot2d(natom)

  Open(unit=CHARGEOUT_UNIT,&
   file=Trim(PLOT_Name)//".2d.idl",form="formatted")

! DEN ==> DEN*exp(i*G*origin)

   do i=2,Gpnt_Size(G_HIGH)
     arg= Dot_Product(Gpnt(1:3,i),planeplot%origin)
     den(i)=den(i)*CMPLX(cos(arg),sin(arg))
   enddo


    Do iy=1,ny
     Do ix=1,nx
      gridden(ix,iy)=den(1)
     Enddo
    Enddo

  do i=2,Gpnt_Size(G_High)
   arg=Dot_Product(Gpnt(1:3,i),Xu(1:3))
   zzz=CMPLX(cos(arg),sin(arg))
   px(1)=1
    do ix=2,nx
     px(ix)=px(ix-1)*zzz
    enddo
   arg=Dot_Product(Gpnt(1:3,i),Yu(1:3))
   zzz=CMPLX(cos(arg),sin(arg))
   py(1)=1
    do iy=2,ny
     py(iy)=py(iy-1)*zzz
    enddo

   Do iy=1,ny
    zzz=2*DEN(i)*py(iy)
    Do ix=1,nx
      gridden(ix,iy)=gridden(ix,iy) + zzz*px(ix)
    Enddo
   Enddo

  Enddo  ! Gpnt
! Plane wave component completed
!
! Correct density inside each atomic sphere


    Do iy=1,ny
     Do ix=1,nx
       R=planeplot%origin + (ix-1)*Xu + (iy-1)*Yu
        
       loop:  do

       do a = 1,Specific_Atoms
         CRC=AtomType_Info(Atom_list(a)%TypeIndex)%Rc
         do i= -many,many
          do j= -many,many
           do k= -many,many
            D = R -  Atom_list(a)%Pos - MATMUL(xtal%Basis,(/i,j,k/))
            Rmag=SQRT(Dot_Product(D,D))

            If (Rmag.le.CRC) then
              Call Correctden(a,D,Rmag,dden)
              gridden(ix,iy)=gridden(ix,iy)+dden*xtal%volume
              exit loop
            EndIf
           Enddo
          Enddo
         Enddo
       Enddo
        exit loop
      Enddo loop

    Enddo  !ix
   Enddo  !iy

! output density grid data
 write(CHARGEOUT_UNIT,'(1p5e15.7)') ((gridden(ix,iy),ix=1,nx), &
    iy=1,ny)
  write(Log_Unit,*) 'maximum density=', maxval(gridden(1:nx,1:ny))
  write(Log_Unit,*) 'minimum density=', minval(gridden(1:nx,1:ny))
 close(CHARGEOUT_UNIT)

!Special information files for idl plots       

 open(unit=CHARGEOUT_UNIT,file=Trim(PLOT_Name)//".title",form='formatted')
  Write(CHARGEOUT_UNIT,*) Trim(PLOT_Name)
  Write(CHARGEOUT_UNIT,*) nx, ny
  Write(CHARGEOUT_UNIT,*) 1   !scale factor
  Write(CHARGEOUT_UNIT,*) Xleng,Yleng
  Write(CHARGEOUT_UNIT,*) natom
 close(CHARGEOUT_UNIT)

  DeAllocate(DEN,gridden,px,py)

End Subroutine Planegrid


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! Subroutine Volplot
!
!   For given unit cell -- specified by Volumeplot%X,Y,Z,Origin,maxbond
!                             (input from read_input)
!    Calculate density in 3 dimensional grid for Plot_Name file
!      (previously generated by partialden)
!
!    Grid must by >= 1 in all directions
!
!    Result given in electrons/volume in A^3
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
Subroutine Volplot

 Integer             :: nx,ny,nz
 Integer             ::  many=10   ! how many possible unit cells to test
 Real, allocatable   ::  gridden(:,:,:)
 Real                ::  Ebegin, Eend,xleng,yleng,zleng,arg,CRC,dden,Rmag
 Real                ::  Xu(3),Yu(3),Zu(3),R(3),D(3),xm,ym,zm , vol
 Real                ::  AA(3),BB(3),CC(3),DI(3),DIJ(3)	
 Complex, allocatable::  DEN(:),px(:),py(:),pz(:)
 Complex             ::  zzz,zz1
 Integer             ::  i,j,k,a,ix,iy,iz,lx,ly,lz,t
 Logical             ::  Useocc

   If (volumeplot%query.lt.15.or.SUM(volumeplot%grid(1:3)).lt.3) then
    write(error_unit,*) 'error in 3d plot -- need data'
    stop
   endif
   nx=volumeplot%grid(1)
   ny=volumeplot%grid(2)
   nz=volumeplot%grid(3)

! Check coordinates
   Xleng=sqrt(Dot_Product(volumeplot%x,volumeplot%x))
   Yleng=sqrt(Dot_Product(volumeplot%y,volumeplot%y))
   Zleng=sqrt(Dot_Product(volumeplot%z,volumeplot%z))
   Write(Log_Unit,'(" Plot origin ",1p3E15.7)')volumeplot%Origin
   Write(Log_Unit,'(" X-axis&length ",1p4E15.7)')volumeplot%X,Xleng
   Write(Log_Unit,'(" Y-axis&length ",1p4E15.7)')volumeplot%Y,Yleng
   Write(Log_Unit,'(" Z-axis&length ",1p4E15.7)')volumeplot%Z,Zleng
   Write(Log_Unit,'(" Nx, Ny, Nz    ",3i10    )')volumeplot%grid(1:3)
   Write(Output_Unit,'(" Plot origin ",1p3E15.7)')volumeplot%Origin
   Write(Output_Unit,'(" X-axis&length ",1p4E15.7)')volumeplot%X,Xleng
   Write(Output_Unit,'(" Y-axis&length ",1p4E15.7)')volumeplot%Y,Yleng
   Write(Output_Unit,'(" Z-axis&length ",1p4E15.7)')volumeplot%Z,Zleng
   Write(Output_Unit,'(" Nx, Ny, Nz    ",3i10    )')volumeplot%grid(1:3)
   
   If (abs(Dot_Product(volumeplot%x,volumeplot%y)).gt.1.e-4 &
    .or.abs(Dot_Product(volumeplot%x,volumeplot%z)).gt.1.e-4 &
    .or.abs(Dot_Product(volumeplot%z,volumeplot%y)).gt.1.e-4) then
    write(Error_Unit,*) 'Error in volplot- axes are not orthogonal'
    stop
   endif

   many= 4*(Xleng*Yleng*Zleng)/xtal%volume + 1
   many= MAX(1,many)

   Write(Log_Unit,'(" Unit cell repeat factor = ",i10)') many
   Write(Output_Unit,'(" Unit cell repeat factor = ",i10)') many


   vol=xtal%volume*(0.14818470952904492571d0)
   Write(Log_Unit,'(" Unit cell volume in A^3 = ",1pe15.7)') vol
   Write(Output_Unit,'(" Unit cell volume in A^3 = ",1pe15.7)') vol

   lx=nx-1
   ly=ny-1
   lz=nz-1

   Xu=0;Yu=0;Zu=0;Xm=0;Ym=0;Zm=0
   If (lx.gt.0) Xu=volumeplot%X/lx ; xm=SQRT(Dot_Product(Xu,Xu))
   If (ly.gt.0) Yu=volumeplot%Y/ly ; ym=SQRT(Dot_Product(Yu,Yu))
   If (lz.gt.0) Zu=volumeplot%Z/lz ; zm=SQRT(Dot_Product(Zu,Zu))

   AA=xtal%Basis(1:3,1)
   BB=xtal%Basis(1:3,2)
   CC=xtal%Basis(1:3,3)
 
  Allocate(DEN(Gpnt_size(G_HIGH)),gridden(nx,ny,nz),&
   px(nx),py(ny),pz(nz),stat=j)
   if (j.ne.0) then
    write(Error_Unit,*) 'error allocating work array in Volplot', &
         FFT_Grid(4,G_HIGH),nx,ny,nz
    stop
   endif

  Open(unit=CHARGEOUT_UNIT,file=Trim(PLOT_Name),form="unformatted")

  write(Log_Unit,*) ' Reading from partialden file ',Trim(PLOT_Name)
  write(Output_Unit,*) ' Reading from partialden file ',Trim(PLOT_Name)
  read(CHARGEOUT_UNIT) Ebegin,Eend,Useocc
  write(Log_Unit,*) 'Ebegin,Eend,Useocc= ', Ebegin,Eend,Useocc
  write(Output_Unit,*) 'Ebegin,Eend,Useocc= ', Ebegin,Eend,Useocc

  write(6,*) 'gpnt',Gpnt_size(G_HIGH)
  read(CHARGEOUT_UNIT) (DEN(i),i=1,Gpnt_Size(G_HIGH))
  do a=1,Specific_Atoms
   t= PLM_AtomRange(2,a) - PLM_AtomRange(1,a) + 1
   write(6,*) 'specific atom = ', a, PLM_AtomRange(2,a), PLM_AtomRange(1,a)
   read(CHARGEOUT_UNIT) ((Atom_List(a)%Wij(i,j),i=1,t),j=1,t)
  enddo

 Close(CHARGEOUT_UNIT)


  Open(unit=CHARGEOUT_UNIT,&
   file=Trim(PLOT_Name)//".3d.dx",form="formatted")

! DEN ==> DEN*exp(i*G*origin)

   do i=2,Gpnt_Size(G_HIGH)
     arg= Dot_Product(Gpnt(1:3,i),volumeplot%origin)
     den(i)=den(i)*CMPLX(cos(arg),sin(arg))
   enddo


   Do iz=1,nz
    Do iy=1,ny
     Do ix=1,nx
      gridden(ix,iy,iz)=den(1)
     Enddo
    Enddo
   Enddo

  do i=2,Gpnt_Size(G_High)
   arg=Dot_Product(Gpnt(1:3,i),Xu(1:3))
   zzz=CMPLX(cos(arg),sin(arg))
   px(1)=1
    If (ix.gt.1) then
      do ix=2,nx
         px(ix)=px(ix-1)*zzz
      enddo
    EndIf
   arg=Dot_Product(Gpnt(1:3,i),Yu(1:3))
   zzz=CMPLX(cos(arg),sin(arg))
   py(1)=1
    If (iy.gt.1) then
      do iy=2,ny
       py(iy)=py(iy-1)*zzz
      enddo
    EndIf
   arg=Dot_Product(Gpnt(1:3,i),Zu(1:3))
   zzz=CMPLX(cos(arg),sin(arg))
   pz(1)=1
    If (iz.gt.1) then
      do iz=2,nz
       pz(iz)=pz(iz-1)*zzz
      enddo
    EndIf
   Do iz=1,nz
    zzz=2*DEN(i)*pz(iz)
    Do iy=1,ny
     zz1=zzz*py(iy)
     Do ix=1,nx
      gridden(ix,iy,iz)=gridden(ix,iy,iz) + zz1*px(ix)
     Enddo
    Enddo
   Enddo

  Enddo  ! Gpnt
! Plane wave component completed
!
! Correct density inside each atomic sphere


   Do iz=1,nz
    Do iy=1,ny
     Do ix=1,nx
       R=volumeplot%origin + (ix-1)*Xu + (iy-1)*Yu + (iz-1)*Zu
        
       loop:  do

       do a = 1,Specific_Atoms
         CRC=AtomType_Info(Atom_list(a)%TypeIndex)%Rc
         do i= -many,many
            DI = R - Atom_list(a)%Pos - AA*i
          do j= -many,many
             DIJ = DI - BB*j
           do k= -many,many
            !D = R -  Atom_list(a)%Pos - MATMUL(xtal%Basis,(/i,j,k/))
            D = DIJ - CC*k
            Rmag=SQRT(Dot_Product(D,D))

            If (Rmag.le.CRC) then
              Call Correctden(a,D,Rmag,dden)
              gridden(ix,iy,iz)=gridden(ix,iy,iz)+dden*xtal%volume
              exit loop
            EndIf
           Enddo
          Enddo
         Enddo
       Enddo
        exit loop
      Enddo loop

    Enddo  !ix
   Enddo  !iy
  Enddo  !iz
 
!  change to electrons/A^3

  gridden=gridden/vol

! output density grid data

 write(CHARGEOUT_UNIT,'(1p5e15.7)') (((gridden(ix,iy,iz),ix=1,nx), &
    iy=1,ny),iz=1,nz)
  write(Log_Unit,*) 'maximum density=', maxval(gridden(1:nx,1:ny,1:nz))
  write(Log_Unit,*) 'minimum density=', minval(gridden(1:nx,1:ny,1:nz))
 close(CHARGEOUT_UNIT)

!Special header file Data-Explorer output

 open(unit=CHARGEOUT_UNIT,file=Trim(PLOT_Name)//".general",form='formatted')
  Write(CHARGEOUT_UNIT,*) 'file = ',Trim(PLOT_Name)//".3d.dx"
  Write(CHARGEOUT_UNIT,*) 'grid = ',nx,' x ',ny,' x ',nz
  Write(CHARGEOUT_UNIT,*) 'format = text'
  Write(CHARGEOUT_UNIT,*) 'interleaving = record'
  Write(CHARGEOUT_UNIT,*) 'majority = column'
  Write(CHARGEOUT_UNIT,*) 'field = density'
  Write(CHARGEOUT_UNIT,*) 'structure = scalar'
  Write(CHARGEOUT_UNIT,*) 'type = float'
  Write(CHARGEOUT_UNIT,*) 'dependency = positions'
   xm=0.52917720859*xm       ! length unit in Anstroms
   ym=0.52917720859*ym       ! length unit in Anstroms
   zm=0.52917720859*zm       ! length unit in Anstroms
  Write(CHARGEOUT_UNIT,'(" positions = 0.0,",f9.5,", 0.0,",f9.5,", &
     &0.0,",f9.5)')&
     xm,ym,zm
  Write(CHARGEOUT_UNIT,*) 'end'
 close(CHARGEOUT_UNIT)
 DeAllocate(DEN,gridden,px,py,pz)

End Subroutine Volplot

!******************************************************************************
!
! Subroutine correctden
!  Returns density correction within atomic sphere a
!
!******************************************************************************
Subroutine Correctden(atom,R,Rmag,dden)
Integer, Intent(IN)   :: atom             ! Specific_atom index
Real,    Intent(IN)   :: R(3),Rmag     ! position relative to atom center
Real,    Intent(OUT)  :: dden

Real, Parameter       :: rtol=1.e-4
Type (Specific_Atom), Pointer :: A
Type (Atom_Info_Fixed), Pointer :: AT
Integer              :: Basis_Size,Li,Lj,nili,njlj,mi,mj,t,Lmax,indexi,indexj
Integer              :: ibase,jbase,Nlist
Integer, pointer     :: L_Value(:), nl_Base(:), nlm_LUT(:,:)
Real                 :: aden,h,phi(20),tphi(20),r2,r3
Real, pointer        :: APhi(:,:),ATPhi(:,:)
Complex              :: Ylm(13,4),Ylmstar(13,4),ylmi,ylmj,zz
Complex, pointer     :: Wij(:,:)

 A => Atom_List(atom)
 AT => AtomType_Info(A%TypeIndex)
 Basis_size=AT%Basis_Size
 Lmax=Maxval(AT%L_Value)
 If (Lmax.gt.3) then
   write(error_unit,*) 'error in correctden -- lmax > 3',lmax
   stop
 endif
 If (Basis_size.gt.20) then
   write(error_unit,*) 'error in correctden -- Basis_size > 20',Basis_size
   stop
 endif

 h=AT%Mesh_Step
 Nlist=AT%Mesh_Size
 Wij => A%Wij
 APhi => AT%Phi
 ATPhi => AT%TPhi
 L_value => AT%L_Value
 nl_Base => AT%nl_Base
 nlm_LUT => AT%nlm_LUT

 dden=0

 Call YlmList(r(1),r(2),r(3),lmax,ylm,error_unit)
   
 ylmstar = CONJG(ylm)

 phi=0
 Tphi=0
 If (Rmag > rtol ) then
   do nili=1,Basis_Size
     call Lininterp(Nlist,h,APhi(1:Nlist,nili),Rmag,phi(nili),AT%RadGrid)
     call Lininterp(Nlist,h,ATPhi(1:Nlist,nili),Rmag,Tphi(nili),AT%RadGrid)
   enddo
   phi(1:Basis_Size)=phi(1:Basis_size)/Rmag
   Tphi(1:Basis_Size)=Tphi(1:Basis_size)/Rmag
 Else
   do nili=1,Basis_Size
     If (L_Value(nili).eq.0) then
      If (AT%RadGrid%islog) then
        r2=AT%RadGrid%drdu(2)-AT%RadGrid%drdu(1)
        r3=AT%RadGrid%drdu(3)-AT%RadGrid%drdu(1)
        phi(nili)=Extrapto0(APhi(2,nili)/r2,APhi(3,nili)/r3)
        Tphi(nili)=Extrapto0(ATPhi(2,nili)/r2,ATPhi(3,nili)/r3)
      else
        phi(nili)=Extrapto0(APhi(2,nili)/h,APhi(3,nili)/(2*h))
        Tphi(nili)=Extrapto0(ATPhi(2,nili)/h,ATPhi(3,nili)/(2*h))
      endif
     EndIf
   EndDo
 Endif

 do nili=1,Basis_Size
  Li=L_Value(nili)
  ibase=nl_Base(nili); ibase=ibase+nlm_LUT(2,ibase)
    do njlj=1,Basis_Size
     Lj=L_Value(njlj)
     jbase=nl_Base(njlj); jbase=jbase+nlm_LUT(2,jbase)
      
     ! aden=0
     ! Do mi = -Li, Li
     !   Ylmi = CONJG(ylm(Li+1+mi,Li+1))
     !    Do mj= -Lj, Lj
     !     Ylmj = ylm(Lj+1+mj,Lj+1)
     !     aden=aden + A%Wij(ibase+mi,jbase+mj) *Ylmi * Ylmj
     !    Enddo
     ! Enddo
     
       aden=0
       Do mj = -Lj, Lj
         indexj=jbase+mj
         zz=0
         Do mi = -Li, Li
           zz = zz + Ylmstar(Li+1+mi,Li+1)*Wij(ibase+mi,indexj)
         Enddo
         aden = aden + zz*Ylm(Lj+1+mj,Lj+1)
       Enddo

     dden = dden + aden*(phi(nili)*phi(njlj)-Tphi(nili)*Tphi(njlj))
   enddo
 Enddo

 write(6,*) 'correctden',rmag,dden
 call flush(6)

End Subroutine correctden


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! Subroutine GridPlot_3d(Garray)
!
!   For given unit cell -- specified by Volumeplot%X,Y,Z,Origin,maxbond
!                             (input from read_input)
!    Evaluate Garray(Gnpt) on  3 dimensional grid for Plot_Name file
!       where Garray is evaluated on linear Fourier space grid
!              Gpnt(:,1:Gpnt_Size(G_HIGH))
!
!    Result given in units of Garray
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
Subroutine GridPlot_3d(Garray,logic)
  Complex, INTENT(INOUT) :: Garray(:)
  Logical, INTENT(IN) :: logic
 Integer             :: nx,ny,nz
 Real, allocatable   ::  gridden(:,:,:)
 Real                ::  Ebegin, Eend,xleng,yleng,zleng,arg,CRC,dden,Rmag
 Real                ::  Xu(3),Yu(3),Zu(3),R(3),D(3),xm,ym,zm , vol
 Real                ::  AA(3),BB(3),CC(3),DI(3),DIJ(3)	
 Complex, allocatable::  px(:),py(:),pz(:)
 Complex             ::  zzz,zz1
 Integer             ::  many=10   ! how many possible unit cells to test
 Integer             ::  i,j,k,a,ix,iy,iz,lx,ly,lz,t
 Logical             ::  Useocc

   If (volumeplot%query.lt.15.or.SUM(volumeplot%grid(1:3)).lt.3) then
    write(error_unit,*) 'error in 3d plot -- need data'
    stop
   endif
   nx=volumeplot%grid(1)
   ny=volumeplot%grid(2)
   nz=volumeplot%grid(3)

! Check coordinates
   Xleng=sqrt(Dot_Product(volumeplot%x,volumeplot%x))
   Yleng=sqrt(Dot_Product(volumeplot%y,volumeplot%y))
   Zleng=sqrt(Dot_Product(volumeplot%z,volumeplot%z))
   Write(Log_Unit,'(" Plot origin ",1p3E15.7)')volumeplot%Origin
   Write(Log_Unit,'(" X-axis&length ",1p4E15.7)')volumeplot%X,Xleng
   Write(Log_Unit,'(" Y-axis&length ",1p4E15.7)')volumeplot%Y,Yleng
   Write(Log_Unit,'(" Z-axis&length ",1p4E15.7)')volumeplot%Z,Zleng
   Write(Log_Unit,'(" Nx, Ny, Nz    ",3i10    )')volumeplot%grid(1:3)
   Write(Output_Unit,'(" Plot origin ",1p3E15.7)')volumeplot%Origin
   Write(Output_Unit,'(" X-axis&length ",1p4E15.7)')volumeplot%X,Xleng
   Write(Output_Unit,'(" Y-axis&length ",1p4E15.7)')volumeplot%Y,Yleng
   Write(Output_Unit,'(" Z-axis&length ",1p4E15.7)')volumeplot%Z,Zleng
   Write(Output_Unit,'(" Nx, Ny, Nz    ",3i10    )')volumeplot%grid(1:3)
   
   If (abs(Dot_Product(volumeplot%x,volumeplot%y)).gt.1.e-4 &
    .or.abs(Dot_Product(volumeplot%x,volumeplot%z)).gt.1.e-4 &
    .or.abs(Dot_Product(volumeplot%z,volumeplot%y)).gt.1.e-4) then
    write(Error_Unit,*) 'Error in volplot- axes are not orthogonal'
    stop
   endif

   many= (Xleng*Yleng*Zleng)/xtal%volume + 1
   many= MAX(1,many)

   Write(Log_Unit,'(" Unit cell repeat factor = ",i10)') many
   Write(Output_Unit,'(" Unit cell repeat factor = ",i10)') many

   vol=xtal%volume*(0.14818470952904492571d0)
   Write(Log_Unit,'(" Unit cell volume in A^3 = ",1pe15.7)') vol
   Write(Output_Unit,'(" Unit cell volume in A^3 = ",1pe15.7)') vol

   lx=nx-1
   ly=ny-1
   lz=nz-1

   Xu=0;Yu=0;Zu=0;Xm=0;Ym=0;Zm=0
   If (lx.gt.0) Xu=volumeplot%X/lx ; xm=SQRT(Dot_Product(Xu,Xu))
   If (ly.gt.0) Yu=volumeplot%Y/ly ; ym=SQRT(Dot_Product(Yu,Yu))
   If (lz.gt.0) Zu=volumeplot%Z/lz ; zm=SQRT(Dot_Product(Zu,Zu))

   AA=xtal%Basis(1:3,1)
   BB=xtal%Basis(1:3,2)
   CC=xtal%Basis(1:3,3)
 
  Allocate(gridden(nx,ny,nz), px(nx),py(ny),pz(nz),stat=j)
   if (j.ne.0) then
    write(Error_Unit,*) 'error allocating work array in Volplot', &
         nx,ny,nz,j
    stop
   endif

  Open(unit=CHARGEOUT_UNIT,&
   file=Trim(PLOT_Name)//".3d.dx",form="formatted")

! Garray ==> Garray*exp(i*G*origin)

   do i=2,Gpnt_Size(G_HIGH)
     arg= Dot_Product(Gpnt(1:3,i),volumeplot%origin)
     Garray(i)=Garray(i)*CMPLX(cos(arg),sin(arg))
   enddo


   Do iz=1,nz
    Do iy=1,ny
     Do ix=1,nx
      gridden(ix,iy,iz)=Garray(1)
     Enddo
    Enddo
   Enddo

  do i=2,Gpnt_Size(G_High)
   arg=Dot_Product(Gpnt(1:3,i),Xu(1:3))
   zzz=CMPLX(cos(arg),sin(arg))
   px(1)=1
    If (ix.gt.1) then
      do ix=2,nx
         px(ix)=px(ix-1)*zzz
      enddo
    EndIf
   arg=Dot_Product(Gpnt(1:3,i),Yu(1:3))
   zzz=CMPLX(cos(arg),sin(arg))
   py(1)=1
    If (iy.gt.1) then
      do iy=2,ny
       py(iy)=py(iy-1)*zzz
      enddo
    EndIf
   arg=Dot_Product(Gpnt(1:3,i),Zu(1:3))
   zzz=CMPLX(cos(arg),sin(arg))
   pz(1)=1
    If (iz.gt.1) then
      do iz=2,nz
       pz(iz)=pz(iz-1)*zzz
      enddo
    EndIf
   Do iz=1,nz
    zzz=2*Garray(i)*pz(iz)
    Do iy=1,ny
     zz1=zzz*py(iy)
     Do ix=1,nx
      gridden(ix,iy,iz)=gridden(ix,iy,iz) + zz1*px(ix)
     Enddo
    Enddo
   Enddo

  Enddo  ! Gpnt
! Plane wave component completed

 If(.not.logic) then      ! if logic==.false., zero values inside atomic spheres


! Plane wave component completed
   Do iz=1,nz
    Do iy=1,ny
     Do ix=1,nx
       R=volumeplot%origin + (ix-1)*Xu + (iy-1)*Yu + (iz-1)*Zu

       loop:  do

       do a = 1,Specific_Atoms
         CRC=AtomType_Info(Atom_list(a)%TypeIndex)%Rc
         do i= -many,many
            DI = R - Atom_list(a)%Pos - AA*i
          do j= -many,many
             DIJ = DI - BB*j
           do k= -many,many
            !D = R -  Atom_list(a)%Pos - MATMUL(xtal%Basis,(/i,j,k/))
            D = DIJ - CC*k
            Rmag=SQRT(Dot_Product(D,D))

            If (Rmag.le.CRC) then
              gridden(ix,iy,iz)=0.d0
              exit loop
            EndIf
           Enddo
          Enddo
         Enddo
       Enddo
        exit loop
      Enddo loop

    Enddo  !ix
   Enddo  !iy
  Enddo  !iz

 Endif
!
! output grid data

 write(CHARGEOUT_UNIT,'(1p5e15.7)') (((gridden(ix,iy,iz),ix=1,nx), &
    iy=1,ny),iz=1,nz)
 close(CHARGEOUT_UNIT)

  write(Log_Unit,*) 'maximum Garray=', maxval(gridden(1:nx,1:ny,1:nz))
  write(Log_Unit,*) 'minimum Garray=', minval(gridden(1:nx,1:ny,1:nz))

!Special header file Data-Explorer output

 open(unit=CHARGEOUT_UNIT,file=Trim(PLOT_Name)//".general",form='formatted')
  Write(CHARGEOUT_UNIT,*) 'file = ',Trim(PLOT_Name)//".3d.dx"
  Write(CHARGEOUT_UNIT,*) 'grid = ',nx,' x ',ny,' x ',nz
  Write(CHARGEOUT_UNIT,*) 'format = text'
  Write(CHARGEOUT_UNIT,*) 'interleaving = record'
  Write(CHARGEOUT_UNIT,*) 'majority = column'
  Write(CHARGEOUT_UNIT,*) 'field = density'
  Write(CHARGEOUT_UNIT,*) 'structure = scalar'
  Write(CHARGEOUT_UNIT,*) 'type = float'
  Write(CHARGEOUT_UNIT,*) 'dependency = positions'
   xm=0.52917720859*xm       ! length unit in Anstroms
   ym=0.52917720859*ym       ! length unit in Anstroms
   zm=0.52917720859*zm       ! length unit in Anstroms
  Write(CHARGEOUT_UNIT,'(" positions = 0.0,",f9.5,", 0.0,",f9.5,", &
     &0.0,",f9.5)')&
     xm,ym,zm
  Write(CHARGEOUT_UNIT,*) 'end'
 close(CHARGEOUT_UNIT)
 DeAllocate(gridden,px,py,pz)

 End Subroutine GridPlot_3d

!  Note: it is not sufficient to load previous run before calling Estatic;
!     because only by calling the SCF step are SCFSmooth, SCFHatRho, etc.
!     are properly initialized!!!


  SUBROUTINE Estatic(logic)
    Logical, INTENT(IN) :: logic

    Complex, allocatable :: V_static(:)
    Integer   :: ig
    Real      :: con

   
   Allocate(V_static(Gpnt_size(G_HIGH)),stat=ig)
        if (ig.ne.0) then
           write(Error_Unit,*) 'allocation error in 3dElect', &
                         Gpnt_size(G_HIGH),ig
           stop
        endif

   V_static=0
   V_static=SCFvalues%RhoSmooth+SCFvalues%RhoHat+SCFvalues%CoreTail
   if (spindependence) V_static=V_static+SCFvalues%RhoSmoothspin

   write(Log_unit,*) 'Estatic total density', V_static(1)

   con=2*13.60569193d0*Four_Pi/xtal%Volume    ! eV units for V_static
   V_static(1)=0

 ! Note V_static == e*electrostatic potential relative to positive test charge
   do ig=2,Gpnt_size(G_HIGH)
      V_static(ig)=-con*V_static(ig)/Gpnt(4,ig)**2
   enddo

   Call GridPlot_3d(V_static,logic)

   deallocate(V_static)

  END SUBROUTINE Estatic
End Module
