!******************************************************************************
!
! File : coulomb_matrix.f90
!   by : Alan Tackett
!   on : 11/10/95
!  for : PAW Project
!
! Creates the coulomb matrix for a particular atom type.
! used in the atom_data module for the field Cijkl.
!
!                                               
!                                               
!                M ---   aL                  L-M        LM
! Cijkl(:) = (-1)  \    V                *  G          G      
!                  /     n l n l n l n l     l m l m    l m l m
!                  ---    i i j j k k l l     i i j j    k k l l
!                   L                                 
!                                               
!
! Where the G's are Gaunt coefficients and V is the fixed atom V_hartree data.
! The various l's amd m's are determined by the atomic orbitals defined
! for the specific atom. Capital L is a value in the range of
!
!      L    <= L <= L       
!       min          max
!
!
! where L  = MAX(| l - l |, | l - l |)  and  L  = MIN( l + l , l + l )
!        min        i   j      k   l          max       i   j   k   l
!
! Based on the symmetry properties of the Gaunt coefficients you know that
!
!    LM   
!   G        != 0    if (l + l + L) is EVEN and M = m - m
!    l m l m              i   j                      j   i
!     i i j j
!
!
!    *******  NOTE: No Error checking is done!!!!!! ***************
!                   The LUT_Cijkl must use 4 byte integers!
!    **************************************************************
!
!
!  Parameters:
!
!      Atom       - Atom type to use for initializing Cijkl
!      Array_Size - Size of Cijkl and LUT_Cijkl.
!                   if Array_Size = -1 then no data is stored and 
!                   the array size required to store all the non-zero
!                   coefficients is returned.
!
!******************************************************************************


Subroutine Coulomb_Matrix(Atom, Array_Size)
  Use atom_data
  Use spherical_harmonic
  Use coulomb_pack
  Use vhartree_pack

  Implicit None

  Type(Atom_Info_Fixed), Intent(INOUT) :: atom
  Integer,               Intent(INOUT) :: Array_Size

  Logical          :: Ok
  Integer          :: loop
  Integer          :: cnt, V_cnt, Num_Hartree
  Integer          :: Lmin, Lmax, Ltest
  Integer          :: nili, njlj, nklk, nlll
  Integer          :: nili2, njlj2, nklk2, nlll2
  Integer          :: Li2, Lj2, Lk2, Ll2
  Integer          :: li, lj, lk, ll, mi, mj, mk, ml, M, L, i, j
  Integer          :: mlmax, mlmin, Cap_L(100)
  Integer, Pointer :: L_Value(:), LUT(:), LUT_V_HARTREE(:)
  Real,    Pointer :: Vh(:), Cijkl(:)
  Real             :: msign, term


  
  Num_Hartree = Atom%Hartree_Size
  L_Value => Atom%L_Value
  Vh => Atom%V_Hartree
  Cijkl => Atom%Cijkl
  LUT => Atom%LUT_Cijkl
  LUT_V_HARTREE => Atom%LUT_V_HARTREE
  
  cnt = 0
  V_cnt = 1
  Do While (V_cnt <= Num_Hartree)
      Call VHartree_Decode(LUT_V_Hartree(V_cnt), nili, njlj, nklk, nlll, L)
   
      li = L_Value(nili);  lj = L_value(njlj);
      lk = L_Value(nklk);  ll = L_Value(nlll);

      Ltest = li + lj - lk - ll
      nili2 = nili; njlj2 = njlj; nklk2 = nklk; nlll2 = nlll;

      i = 0
      Do While ((nili2==nili) .AND. (njlj2==njlj) .AND. &
                (nklk2==nklk) .AND. (nlll2==nlll) .AND. (V_cnt+i<Num_Hartree))
        i = i + 1
        Cap_L(i) = L
        Call VHartree_Decode(LUT_V_Hartree(V_cnt+i),nili2,njlj2,nklk2,nlll2,L)
      End Do

      If (i==0) then
         i = 1
         Cap_L(1) = L
      End If

      nili2 = nili;  njlj2 = njlj;  nklk2 = nklk;  nlll2=nlll;
      li2 = li;      Lj2 = Lj;      Lk2 = Lk;      Ll2 = Ll;
      
      Do loop=1, 4
         Ok = .FALSE.

         If (loop==1) then
            Ok = .TRUE.        !** Always do this one
         Else If ((loop==2) .AND. (nili /= njlj)) then
            Ok = .TRUE.              !** Swap ij -> jikl
            nili = njlj2;  Li = lj2;
            njlj = nili2;  Lj = Li2;
            nklk = nklk2;  Lk = Lk2;
            nlll = nlll2;  Ll = Ll2;
         Else If ((loop==3) .AND. (nklk /= nlll)) then
            Ok = .TRUE.              !** Swap kl -> ijlk
            nili = nili2;  Li = li2;
            njlj = njlj2;  Lj = Lj2;
            nklk = nlll2;  Lk = Ll2;
            nlll = nklk2;  Ll = Lk2;
         Else If ((loop==4) .AND. (nili /= njlj) .AND. (nklk /= nlll)) then
            Ok = .TRUE.              !** Swap both -> jilk
            nili = njlj2;  Li = lj2;
            njlj = nili2;  Lj = Li2;
            nklk = nlll2;  Lk = Ll2;
            nlll = nklk2;  Ll = Lk2;
         End If

         If (Ok) then
            Do mi = -li, li
               Do mj = -lj, lj
                  M = mi - mj        
                  msign = (-1.0)**M
                  Do mk = -lk, lk
                     Do ml = -ll, ll
                        if (M==(ml-mk)) then    
                           term = 0
                           Do j = 1, i
                              if (abs(M) <= Cap_L(j)) then
                                 term = term+Vh(V_cnt+j-1) * &
                                      Gaunt(Cap_L(j),-M,li,mi,lj,mj) * &
                                      Gaunt(Cap_L(j), M, lk, mk, ll, ml)
                              End IF
                           End Do
                              
                           if (ABS(term)>1E-20) then
                              cnt = cnt + 1
                              if (Array_Size > 0) then
                                 Cijkl(cnt) = term*msign
                                 LUT(cnt) = Coulomb_Encode(nili,njlj,nklk, &
                                      nlll,mi,mj,mk)
                              End If
                           End If
                        End IF
                     End Do
                  End Do
               End Do
            End Do
         End If
      End Do

      V_cnt = V_Cnt + i
  End Do

  Array_Size = cnt

  Return      
End Subroutine Coulomb_Matrix


