!******************************************************************************
!
! File : coulomb_pack.f90
!   by : Alan Tackett
!   on : 11/10/95
!  for : PAW Method 
!
! Contains 2 routines to encode and decode the coulomb matrix indices that are
! used in the atom_data module for the field Cijkl.
!
!                M ---   aL                  L-M        LM
! Cijkl(:) = (-1)  \    V                *  G          G      
!                  /     n l n l n l n l     l m l m    l m l m
!                  ---    i i j j k k l l     i i j j    k k l l
!                   L                                 
!
! Where the G's are Gaunt coefficients and V is the fixed atom V_hartree data.
! The various l's amd m's are determined by the atomic orbitals defined
! for the specific atom. Capital L is a value in the range of
!
!      L    <= L <= L       
!       min          max
!
!
! where L  = MAX(| l - l |, | l - l |)  and  L  = MIN( l + l , l + l )
!        min        i   j      k   l          max       i   j   k   l
!
! Based on the symmetry properties of the Gaunt coefficients you know that
!
!    LM   
!   G        != 0    if (l + l + L) is EVEN and M = m - m
!    l m l m              i   j                      j   i
!     i i j j
!
!
! For this reason the indices that have to be stored are :
!
!               (n l ) (n l ) (n l ) (n l )  m    m    m
!                 i i    j j    k k    l l    i    j    k
!
! Bits Required:  (5)    (5)    (5)    (5)  (3)  (3)  (3)  = 29 bits total
!
! The l's can be determined from the (nl)'s and  m = m + m - m
!                                                 l   k   i   j
!
! Each l value is in the range on 0..3 and the corresponding m is in the 
! range of -l <= m <= l for 2*l + 1 possible values.  This gives the range
! for L = 0..6.  The (nl) indices are stored as pairs which correspond
! to the actual basis function index.  Since only 5 bits are used the max
! number of basis functions is 32 for each atom.  The l's are not stored 
! in order for the encoding to use less than 32 bits.  The m's are stored
! by adding an offset of 3 this makes all the m's positive and doesn't
! require the l values.
!
! Also note that since 29 bits are required each integer should 
! be 4 byte integer. The 29 LSB's are used all other bits are set to 0.
!
!    *******  NOTE: No Error checking is done!!!!!! ********
!
!******************************************************************************

Module coulomb_pack

implicit none
!******************************************************************************
Contains
!******************************************************************************


!******************************************************************************
!
!  Coulomb_Encode - Encodes the given nili,njlj, nklk, nlll, mi, mj, 
!                   and mk supplied assuming the above restrictions on 
!                   l's, m's, and L.
!
!
!  NOTE : NO error checking is done!!!!
!         The return value must be a 4 byte or greater integer!!
!
!******************************************************************************

Integer Function Coulomb_Encode(nili, njlj, nklk, nlll, mi, mj, mk)
   Integer, Intent(IN) :: nili
   Integer, Intent(IN) :: njlj
   Integer, Intent(IN) :: nklk
   Integer, Intent(IN) :: nlll
   Integer, Intent(IN) :: mi
   Integer, Intent(IN) :: mj
   Integer, Intent(IN) :: mk

   Integer :: N

   N = nili
   N = ISHFT(N, 5) + njlj
   N = ISHFT(N, 5) + nklk
   N = ISHFT(N, 5) + nlll
   N = ISHFT(N, 3) + mi+3
   N = ISHFT(N, 3) + mj+3
   N = ISHFT(N, 3) + mk+3

   Coulomb_Encode = N

   Return
End Function
         
!******************************************************************************
!
!  Coulomb_Decode - Decodes the given value(N) into the separate nili, njlj,
!                   nklk,nlll, mi, mj, and mk assuming the above restrictions
!                   on (nl)'s, l's, m's, and L.
!
!
!  NOTE : NO error checking is done!!!!
!         The default integer must be a 4 byte or greater integer!!
!
!******************************************************************************

Subroutine Coulomb_Decode(N, nili, njlj, nklk, nlll, mi, mj, mk)
   Integer, Intent(IN)  :: N
   Integer, Intent(OUT) :: nili
   Integer, Intent(OUT) :: njlj
   Integer, Intent(OUT) :: nklk
   Integer, Intent(OUT) :: nlll
   Integer, Intent(OUT) :: mi
   Integer, Intent(OUT) :: mj
   Integer, Intent(OUT) :: mk

   nili = ISHFT(IAND(31*(2**24), N), -24)
   njlj = ISHFT(IAND(31*(2**19), N), -19)
   nklk = ISHFT(IAND(31*(2**14), N), -14)
   nlll = ISHFT(IAND(31*(2**9) , N),  -9)

   mi = ISHFT(IAND(7*(2**6), N), -6) - 3
   mj = ISHFT(IAND(7*(2**3), N), -3) - 3
   mk = IAND(7, N) - 3

   Return
End Subroutine
    

End Module
