!****************************************************************************
!
!  File : crystal_symmetry.f
!    by : N.A.W. Holzwarth
!    on :
!   for : PAW Project
!
!  Module for generating the symmetry matrices for use by SymRho and
!  SymWij.
!
!  Modified on 2/13/95 by Alan Tackett to use the new data names
!
!****************************************************************************

module crystal_symmetry
  Use crystal_data
  Use atom_data
  Use paw_inout
  Use mathlib
  
  implicit none
  save
  
  integer :: nel, nkind, mxatom
  integer, pointer :: igrot(:,:,:),iatrans(:,:)
  Real, pointer :: gtrans(:,:),srot(:,:,:)
  Complex, pointer :: zarot(:,:,:,:)
  Real, parameter :: tol=1.d-6
!  Real, parameter :: Two_Pi = 2*PI     ! constant Two_Pi=two_pi set in mathlib
  Real :: b1(3),b2(3), b3(3), a1(3),a2(3),a3(3)
contains
  
  subroutine setsym(lmax)
    ! input rotation matrices in cartesian coordinates and translation
    !     vector in lattice coordinates (in units of a1,a2,a3)
    !
    ! uses sign & phase convension of  M. E. Rose, Elementary Theory of Angular
    !   Momentum, John Wiley & Sons,. inc. 1957)
    !   zalpha = exp(-i*alpha)   zgamma = exp (-i*gamma)
    !
    ! assumes each transformation  can be expressed in terms of 3 Euler angles
    !   with or without inversion
    
      !*******use lattice_data*****

    implicit none
    
    integer :: lmax, iapkind, iakind
    Real :: rot(3,3),trans(3),v(3),vp(3), tau(3), T(3)
    integer :: ier,jer,many,i,j,k,irot,it,is,isn,mxrl,mxrm,nread
    integer :: iap,index,ia,ikind,il,l,m,mp,l1,l2,l3
    Real :: t1,t2,t3,diff,check
    Real :: cosbeta,sinbeta,cosalp,sinalp,cosgam,singam
    Complex :: zalpha,zgamma

    nel = Xtal%Rot_Size          !**** The Rotation matrices and
    srot => Xtal%RotMatrix       !**** size are already loaded
    gtrans => Xtal%Trans         !****
    mxatom = Specific_Atoms
    nkind = Atom_Types

!**    allocate(gtrans(3,nel))

    a1 = Xtal%Basis(:,1);  a2 = Xtal%Basis(:,2);  a3 = Xtal%Basis(:,3)
    b1 = Xtal%Recip(:,1);  b2 = Xtal%Recip(:,2);  b3 = Xtal%Recip(:,3)

    mxrl=lmax+1
    mxrm=2*lmax+1
    allocate(zarot(mxrm,mxrm,mxrl,nel),igrot(3,3,nel),stat=ier)
    if (ier.ne.0) then
       write(6,*) 'error in allocating rotation matrices',mxrm,mxrl,nel
       stop
    endif
    allocate(iatrans(mxatom,nel),stat=ier)
    if (ier.ne.0) then
       write(6,*) 'error in allocating translation matrices',nel,mxatom
       stop
    endif
    
    zarot=0;igrot=0;iatrans=0
    write(6,*)
    
    write(Log_Unit,*) ' # symmetry elements ',nel,' :'
    write(Log_Unit,*)
    write(Log_Unit,*) 'Cartesian coordinates                Lattice coordinates'
    ! read in rotation matrices in cartesian coordinates
    !  and translations vectors in primitive lattice coordinates
    
    do irot=1,nel
       Rot = SRot(:,:,irot) !**** Changed
       trans = gtrans(:,irot)

       ! PW transformations
       igrot(1,1,irot)=ielement(a1,rot,b1)
       igrot(2,1,irot)=ielement(a2,rot,b1)
       igrot(3,1,irot)=ielement(a3,rot,b1)
       igrot(1,2,irot)=ielement(a1,rot,b2)
       igrot(2,2,irot)=ielement(a2,rot,b2)
       igrot(3,2,irot)=ielement(a3,rot,b2)
       igrot(1,3,irot)=ielement(a1,rot,b3)
       igrot(2,3,irot)=ielement(a2,rot,b3)
       igrot(3,3,irot)=ielement(a3,rot,b3)
       
       write(Log_Unit,*)
       write(Log_Unit,*) 'element #',irot
       do i=1,3
          write(Log_Unit,'(3f8.4,10x,3i5)') (rot(i,j),j=1,3),(igrot(i,j,irot),j=1,3)
       end do

       write(Log_Unit,'(a3,4x,3f10.4)') 't =', (gtrans(i,irot),i=1,3)

       ! set up atomic shifts
       do ia=1, Specific_Atoms
          iakind = Atom_List(ia)%TypeIndex
          Tau = Atom_List(ia)%Frac_Pos
          index=0
          bigloop: do iap=1, Specific_Atoms
             iapkind = Atom_List(iap)%TypeIndex

             v = Atom_List(iap)%Pos
             vp = MATMUL(rot, V)
             
 !write(Log_Unit,*) 'setsym: Vp=',Vp
             t1 = DOT_PRODUCT(XTal%Recip(:,1),vp)/Two_Pi + trans(1)
             t2 = DOT_PRODUCT(XTal%Recip(:,2),vp)/Two_Pi + trans(2)
             t3 = DOT_PRODUCT(XTal%Recip(:,3),vp)/Two_Pi + trans(3)
 !write(Log_Unit,*) 'setsym: T=',t1,t2,t3, ' * Tau=',Tau, 'trans=',trans
             if (iakind == iapkind) then
               small1: do l1=-7,7,1
                 small2: do l2=-7,7,1
                   small3: do l3=-7,7,1
                     diff=abs(t1-tau(1)+l1)+abs(t2-tau(2)+l2)+abs(t3-tau(3)+l3)
                     if (diff.lt.tol) then
                        index=iap
                        exit bigloop
                     end if
                   enddo small3
                 enddo small2
               enddo small1
             End If
          end do bigloop

          if (index.lt.1) then
             write(Log_Unit,*) 'error in symmetry set up ',index,ia,iakind, diff
             stop
          end if

          iatrans(ia,irot)=index
          
          write(Log_Unit,"(i3,'-->',i3)") ia,iatrans(ia,irot)

       enddo ! ia loop

       !   calculate rotation matrices
       ! l=0 case

       zarot(1,1,1,irot)=1

       ! l>0 case

       if (lmax.gt.0) then
          call makeeuler(rot,cosbeta,zalpha,zgamma,isn)


write(Log_Unit,*)'setsym: cosbeta=',cosbeta,' * zalpha=',zalpha, ' * zgamma=',zgamma
         do l=1,lmax
            il=(isn)**l

            do m=-L, L
               i=m+L+1
               do  mp=-L, L
                  j=mp+L+1
write(Log_Unit,*) 'setsym: m=',m, ' mp=',mp, ' dbeta=',dbeta(cosbeta,l,mp,m)
                  zarot(j,i,l+1,irot)=il*dbeta(cosbeta,l,mp,m)        &
                       *(zalpha**mp)*(zgamma**m)
               enddo
            enddo
         enddo

      endif   ! lmax case

      if( lmax.gt.0) then
         write(Log_Unit,*) ' Rotation matrices for l=1'
         do i=1,3
            write(Log_Unit,'(3(2f7.3,2x))') (zarot(i,j,2,irot),j=1,3)
         enddo
         if (lmax.gt.1) then
            write(Log_Unit,*) ' Rotation matrices for l=2'
            do i=1,5
               write(Log_Unit,'(5(2f7.3,2x))') (zarot(i,j,3,irot),j=1,5)
            enddo
         endif
      endif
   enddo     !nel loop


   return
 end subroutine setsym


 subroutine makeeuler(rot,cosbeta,zalpha,zgamma,isn)

   implicit none
   
   Real :: rot(3,3),cosbeta,sinbeta
   Real :: cosalp,sinalp,cosgam,singam,check
   Complex :: zalpha,zgamma
   integer :: ier,isn

   do isn= -1,1,2
      cosbeta=isn*rot(3,3)
      sinbeta=sqrt(1.d0-cosbeta*cosbeta)
      if (abs(sinbeta).gt.tol)  then
         zalpha=isn*cmplx(rot(3,1),-rot(3,2))/sinbeta
         zgamma=isn*cmplx(-rot(1,3),-rot(2,3))/sinbeta
      else
         zalpha=isn*cmplx(rot(1,1),-rot(1,2))/cosbeta
         zgamma=1.d0
      endif

      ! check matrix:
      ier=0
      cosalp=zalpha
      sinalp=-aimag(zalpha)
      cosgam=zgamma
      singam=-aimag(zgamma)
      check=cosalp*cosbeta*cosgam-sinalp*singam
      if (abs(check-isn*rot(1,1)).gt.tol) ier=ier+1
      check=sinalp*cosbeta*cosgam+cosalp*singam
      if (abs(check-isn*rot(1,2)).gt.tol) ier=ier+1
      check=-sinbeta*cosgam
      if (abs(check-isn*rot(1,3)).gt.tol) ier=ier+1
      check=-cosalp*cosbeta*singam-sinalp*cosgam
      if (abs(check-isn*rot(2,1)).gt.tol) ier=ier+1
      check=-sinalp*cosbeta*singam+cosalp*cosgam
      if (abs(check-isn*rot(2,2)).gt.tol) ier=ier+1
      check=sinbeta*singam
      if (abs(check-isn*rot(2,3)).gt.tol) ier=ier+1
      check=cosalp*sinbeta
      if (abs(check-isn*rot(3,1)).gt.tol) ier=ier+1
      check=sinalp*sinbeta
      if (abs(check-isn*rot(3,2)).gt.tol) ier=ier+1
      if (ier.eq.0) return
   enddo

   write(Log_Unit,*) 'error in rot ',ier,rot
   write(Log_Unit,*) 'alpha=',cosalp,sinalp
   write(Log_Unit,*) 'beta =',cosbeta,sinbeta
   write(Log_Unit,*) 'gamma = ',cosgam,singam
   isn=0
   
   stop
 end subroutine makeeuler
 
 function dbeta(cosbeta,l,mp,m)

   !  calculate the rotation matrix d^l_{m'm}(beta) using Eq. 4.14 of
   !  M. E. Rose, Elementary Theory of Angular Momentum, John Wiley & Sons
   !    New York, 1957
   !   Assume l relatively small so that factorials do not cause
   !    roundoff error
   
   implicit none

   integer , parameter :: mxterms=200
   Real :: dbeta,cosbeta
   integer :: l,mp,m,i
   Real :: cosbetab2,sinbetab2,pref,sum,t,arg
   integer :: ina,inb,inc,ml,ms,last
   
   dbeta=0
   
   !  special cases
   if (abs(cosbeta-1.d0).lt.tol) then
      if (mp.eq.m) dbeta=1
   else if (abs(cosbeta+1.d0).lt.tol) then
      if (mp.eq.-m) dbeta=(-1)**(l+m)
   else  ! general case
      cosbetab2=sqrt((1+cosbeta)*0.5d0)
      sinbetab2=sqrt((1-cosbeta)*0.5d0)
      ml=max(mp,m)
      ms=min(mp,m)
      if (ml.ne.mp) sinbetab2=-sinbetab2
      ! t=-(cosbetab2/sinbetab2)**2   !possible error found by A. Wright
      t=-(sinbetab2/cosbetab2)**2   !possible error found by A. Wright
      pref=sqrt((Factorial(l-ms)*Factorial(l+ml))/                        &
           (Factorial(l+ms)*Factorial(l-ml)))/Factorial(ml-ms)             &
           *(cosbetab2**(2*l+ms-ml))*((-sinbetab2)**(ml-ms))
      sum=1
      arg=1
      ina=ml-l
      inb=-ms-l
      inc=ml-ms+1
      do  i=1,mxterms
         if (ina.eq.0.or.inb.eq.0) exit
         arg=(arg*ina*inb*t)/(i*inc)
         sum=sum+arg
         ina=ina+1
         inb=inb+1
         inc=inc+1
      enddo
      
      dbeta=pref*sum
   endif
   
   return
 end function dbeta
 
 
 function ielement(tt,rot,gg)
   
   implicit none
   Real :: tt(3),gg(3),rot(3,3),sum,dsign
   integer :: ielement,i,j

   sum=0.d0
   do  j=1,3
      do  i=1,3
         sum=sum+tt(i)*rot(i,j)*gg(j)
      enddo
   enddo
   ielement=sum/Two_Pi+sign(0.00001,sum)
   return
 end function ielement
 
!*****************************************************************************
!
! Symmetrize_pos  -- inputs nominal atomic positions and outputs
!                        symmetrized positions in fractional units
!
!*****************************************************************************

Subroutine Symmetrize_Pos(Inpos,Outpos)
Real, intent(IN) :: Inpos(3,Specific_Atoms)
Real, intent(OUT) :: Outpos(3,Specific_Atoms)

Real,allocatable :: tmppos(:,:),frot(:,:,:)
Integer,allocatable :: Iswitch(:,:),shift(:,:)
Real :: v(3),vp(3),vpp(3)
Integer :: ia,iap,is,i,j

Outpos=InPos

If (Xtal%Rot_Size.le.1) Return

Allocate(tmppos(3,Specific_Atoms),frot(3,3,Xtal%Rot_Size), &
            shift(3,Specific_Atoms),Iswitch(Specific_atoms,Xtal%Rot_Size))

tmppos = 0
outpos = 0
frot=0
Iswitch=0

  Do is=1,Xtal%Rot_Size
    frot(:,:,is)= MATMUL(Xtal%InverseBasis,&
              MATMUL(Xtal%RotMatrix(:,:,is),Xtal%Basis))
     Do ia=1,Specific_Atoms
        Iswitch(iatrans(ia,is),is)=ia
     Enddo
  Enddo

  Do ia=1,Specific_Atoms
   vp=Inpos(:,ia)              
   v(1)=HalfMod(vp(1)); v(2)=HalfMod(vp(2)); v(3)=HalfMod(vp(3))
   shift(:,ia)= (Inpos(:,ia)   - v)*(1.00001)
    Do is=1,Xtal%Rot_Size
       iap=Iswitch(ia,is)
       vp = MATMUL(frot(:,:,is),v) + Xtal%trans(:,is)
       vpp(1)=HalfMod(vp(1)); vpp(2)=HalfMod(vp(2)); vpp(3)=HalfMod(vp(3))
       tmppos(:,iap)=tmppos(:,iap)+vpp
    Enddo
  Enddo

write(*,*) 'inpos                                 outpos'
  Do ia=1,Specific_Atoms
     v=tmppos(:,ia)/Xtal%Rot_Size
     Call FilterVec(v,Outpos(:,ia))
     Outpos(:,ia)=Outpos(:,ia) + shift(:,ia)
write(*,'(i5,1p3e15.7,2x,1p3e15.7)') ia,inpos(:,ia),outpos(:,ia)
     v=inpos(:,ia)-outpos(:,ia)
     if (DOT_PRODUCT(v,v).gt.0.1) then
      Write(Error_Unit,*) 'Warning from Symmetrize_Pos for atom #',ia,&
         ' inpos = ', inpos(:,ia),' outpos = ',outpos(:,ia)
     EndIf
  Enddo

Deallocate(tmppos,frot,Iswitch,shift)


Return

End Subroutine Symmetrize_Pos

!******************************************************************************
!
! Function HalfMod(x)
!
!    HalfMod sifts x between -.5 < HalfMod <= .5
!
!******************************************************************************
Function HalfMod(x)
Real, intent(IN) :: x
Real :: HalfMod

 HalfMod=x
 Do
    If (HalfMod > 0.5000000001d0) then
       HalfMod=HalfMod - 1
    Else
      Exit
    EndIf
 EndDo

 Do
    If (HalfMod <= -0.499999999d0) then
       HalfMod=HalfMod +1
    Else
      Exit
    EndIf
 Enddo

 Return

End Function HalfMod

!*****************************************************************************
!
! FilterVec(Vin,Vout)
!
!  Elements of Vin(3) are zeroed if their value is less than tolerance
!
!*****************************************************************************
Subroutine FilterVec(Vin,Vout)
  Real,intent(IN) :: Vin(3)
  Real,intent(OUT):: Vout(3)
  Real, parameter :: tol=1.e-10
  integer :: i

  do i=1,3
    If (ABS(Vin(i)) < tol) then
           Vout(i) = 0
    Else
           Vout(i)= Vin(i)
    EndIf
  Enddo

  Return
 
End subroutine FilterVec

 
end module crystal_symmetry
