!******************************************************************************
!
! File : denvhat_pack.f90
!   by : Alan Tackett
!   on : 11/15/95
!  for : PAW Method 
!
! Contains 2 routines to encode and decode the density matrix indices that are
! used in the atom_data module for the fields DENSITY and V_Hat.
!
!               aL                       ^aL
! Density(:) = N              V_Hat(:) = v
!               n l n l                   n l n l
!                i i j j                   i i j j
!
!
! The indices have to be stored are :
!
!                (n l )  (n l )    L
!                  i i     j j   
!
! Bits Required:   (5)     (5)    (3)   = 13 bits total
!
! Each l value is in the range on 0..3 and the corresponding m is in the 
! range of -l <= m <= l for 2*l + 1 possible values.  This gives the range
! for L = 0..6.  The (nl) indices are stored as pairs which correspond
! to the actual basis function index.  Since only 5 bits are used the max
! number of basis functions is 32 for each atom. The l's can be determined
! from the (nl)'s and hence are not stored.  
!
! The 13 LSB's are used all other bits are set to 0.
!
!    *******  NOTE: No Error checking is done!!!!!! ********
!
!******************************************************************************

Module denvhat_pack

implicit none
!******************************************************************************
Contains
!******************************************************************************


!******************************************************************************
!
!  DenVhat_Encode - Encodes the given nili, njlj,  and L supplied
!                   assuming the above restrictions on indices.
!
!
!  NOTE : NO error checking is done!!!!
!
!******************************************************************************

Integer Function DenVhat_Encode(nili, njlj, L)
   Integer, Intent(IN) :: nili
   Integer, Intent(IN) :: njlj
   Integer, Intent(IN) :: L

   Integer :: N

   N = nili
   N = ISHFT(N, 5) + njlj
   N = ISHFT(N, 3) + L

   DenVhat_Encode = N
   Return
End Function
         
!******************************************************************************
!
!  DenVhat_Decode - Decodes the given value(N) into the separate nili, njlj,
!                    L assuming the above restrictions on  (nl)'s, l's,
!                    m's, and L.
!
!
!  NOTE : NO error checking is done!!!!
!
!******************************************************************************

Subroutine DenVhat_Decode(N, nili, njlj, L)
   Integer, Intent(IN)  :: N
   Integer, Intent(OUT) :: nili
   Integer, Intent(OUT) :: njlj
   Integer, Intent(OUT) :: L

   nili = ISHFT(IAND(31*(2**8), N), -8)
   njlj = ISHFT(IAND(31*(2**3), N), -3)
   L  = IAND(7, N) 

   Return
End Subroutine
    
End Module
