!******************************************************************************
!
! File : doijmatrix.f90
!   by : Alan Tackett
!   on : 02/12/97
!  for : PAW Project
!
!  Contains 2 routines for assisting in the calculation of H*Psi and 
!  O*Psi. These routines are passed to AccumProj(projector module) 
!  from HPsi and OPsi in the hamiltonian module.
!
!         ~        ----   ~ a      ~ a
!  HPsi = H*Psi +  \     |P  > D  <P  |Psi>
!                  /       i    ij  j
!                  ----
!                  a,i,j
!
!  There is a similiar definition for OPsi but H~ is replaced with 1 and
!  Dij is replaced with Oij - which is very spase.
!
!  These routines are designed to return either Oij or Dij to AccumProj
!  so that it can determine the weighting factor for each projector in doing
!  the projector sum.  
!
! modified for spin by Ping Tang and N. A. W. Holzwarth
!   last modification 5/23/05
!******************************************************************************

Module doijmatrix

Use atom_data
Use options_data

Implicit NONE!!!!!!!!!!!!

!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
! OijMatrix - Returns the Oij elements for AccumProj
!
! AtomIndex   - The Number of the atom
! nili, njlj, - Projector basis indices
! Li, Lj      - Corresponding L values
! mi, mj      - M values
!
!******************************************************************************

Complex Function OijMatrix(AtomIndex, nili, njlj, li, lj, mi, mj)
  Integer, Intent(IN) :: AtomIndex
  Integer, Intent(IN) :: nili, njlj
  Integer, Intent(IN) :: li, lj
  Integer, Intent(IN) :: mi, mj

  Integer, Pointer :: nlm_base(:)
  Integer :: AT_Index

  AT_index = Atom_List(AtomIndex)%TypeIndex
  nlm_Base => AtomType_Info(AT_Index)%nl_Base
  OijMatrix = AtomType_Info(AT_Index)%Oij(nlm_base(nili)+Li+mi, &
                  nlm_base(njlj)+Lj+mj)

!OijMatrix = 0 !QWERT
  Return
End Function

!******************************************************************************
!
! DijMatrix - Returns the Dij elements for AccumProj
!
! AtomIndex   - The Number of the atom
! nili, njlj, - Projector basis indices
! Li, Lj      - Corresponding L values
! mi, mj      - M values
!
!******************************************************************************

Complex Function DijMatrix(AtomIndex, nili, njlj, li, lj, mi, mj)
  Integer, Intent(IN) :: AtomIndex
  Integer, Intent(IN) :: nili, njlj
  integer, Intent(IN) :: li, lj
  Integer, Intent(IN) :: mi, mj

  Integer, Pointer :: nlm_base(:)

  nlm_Base => AtomType_Info(Atom_List(AtomIndex)%TypeIndex)%nl_Base
  If(Global_spinup) then    !spin up or no spin
       DijMatrix = Atom_List(AtomIndex)%Dij(nlm_base(nili)+Li+mi, &
                    nlm_base(njlj)+Lj+mj)
  Else   !spin down
        DijMatrix = Atom_List(AtomIndex)%Dijspin(nlm_base(nili)+Li+mi, &
                    nlm_base(njlj)+Lj+mj)
  Endif

!DijMatrix = 1 !QWERT  
  Return
End Function

End Module
