!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Modified by Ping Tang and N. A. W. Holzwarth
!  to include spin (LSDA, SGGA)
!  Last change 5/23/05
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
MODULE exchange_corr

  USE mathlib

  IMPLICIT NONE

  ! 
  INTEGER :: XC_TYPE
  Logical :: XC_TYPE_SET

  INTEGER, PARAMETER :: XC_LDA_PW=1
  INTEGER, PARAMETER :: XC_GGA_PBE=2
  INTEGER, PARAMETER :: XC_LSDA_PW=3
  INTEGER, PARAMETER :: XC_SGGA_PBE=4


  ! Parameters for the Perdew-Wang (PRB 45,13244 (1992)) LDA correlation      
  REAL, PRIVATE, PARAMETER :: AA=0.0310907d0
  REAL, PRIVATE, PARAMETER :: a1=0.21370d0
  REAL, PRIVATE, PARAMETER :: b1=7.59570d0
  REAL, PRIVATE, PARAMETER :: b2=3.58760d0
  REAL, PRIVATE, PARAMETER :: b3=1.63820d0
  REAL, PRIVATE, PARAMETER :: b4=0.49294d0

! Parameters for the Perdew-Wang (PRB 45,13244 (1992)) LDA correlation zeta=0
  Real, private, parameter :: AAec0=0.0310907d0
  Real, private, parameter :: a1ec0=0.21370d0
  Real, private, parameter :: b1ec0=7.59570d0
  Real, private, parameter :: b2ec0=3.58760d0
  Real, private, parameter :: b3ec0=1.63820d0
  Real, private, parameter :: b4ec0=0.49294d0

! Parameters for the Perdew-Wang (PRB 45,13244 (1992)) LDA correlation zeta=1
  Real, private, parameter :: AAec1=0.015545d0
  Real, private, parameter :: a1ec1=0.20548d0
  Real, private, parameter :: b1ec1=14.1189d0
  Real, private, parameter :: b2ec1=6.19770d0
  Real, private, parameter :: b3ec1=3.36620d0
  Real, private, parameter :: b4ec1=0.62517d0

! Parameters for the Perdew-Wang (PRB 45,13244 (1992)) LDA correlation
  Real, private, parameter :: AAalp=0.016887d0
  Real, private, parameter :: a1alp=0.11125d0
  Real, private, parameter :: b1alp=10.3570d0
  Real, private, parameter :: b2alp=3.62310d0
  Real, private, parameter :: b3alp=0.88026d0
  Real, private, parameter :: b4alp=0.49671d0

! Parameters for the PBE (PRL 77, 3865 (1996)) GGA-PBE
  REAL, PRIVATE, PARAMETER :: kappa= 0.804d0
  REAL, PRIVATE, PARAMETER :: mu = 0.2195149727645171d0
  REAL, PRIVATE, PARAMETER :: beta = 0.06672455060314922d0
  REAL, PRIVATE, PARAMETER :: gamm = 0.03109069086965489503494086371273d0
  REAL, PRIVATE, PARAMETER :: betabygamm=beta/gamm

CONTAINS

  !**********************************************************
  ! Subroutine to calculate the LDA exchange correlation functionals
  !   using the form of Perdew and Wang (PRB 45, 13244 (1992)
  !   assuming no spin polarization
  !  Inside this routine, energies are in Hartree units

  SUBROUTINE pwldafunc(den,exc,vxc)
    REAL(8), INTENT(IN) :: den    !density
    REAL(8), INTENT(OUT) :: exc,vxc

    ! Variables depending on den
    REAL(8) :: n,g,kf,rs,ks

    REAL(8) :: ex,ec,pprs,decdrs
    REAL(8) :: term

    n=den
    IF (n < machine_zero)  THEN
       exc=0.d0; vxc=0.d0
       RETURN
    ENDIF

    kf=(3.d0*(pi**2)*n)**0.3333333333333333333333333333d0
    rs=(3.d0/(4.d0*pi*n))**0.3333333333333333333333333333d0
    ks=SQRT(4.d0*kf/pi)

    ex=-3.d0*kf/(4.d0*pi)
    pprs=SQRT(rs)*(b1+b3*rs)+rs*(b2+b4*rs)
    !ec=-2.d0*AA*(1.d0+a1*rs)*ddlog(1.d0+1.d0/(2.d0*AA*pprs))
    term=Logofterm(1.d0/(2.d0*AA*pprs))
    ec=-2.d0*AA*(1.d0+a1*rs)*term

    exc=ex+ec

    !decdrs=-(2.d0*AA*a1)*ddlog(1.d0+1.d0/(2*AA*pprs)) &
    decdrs=-(2.d0*AA*a1)*term &
         +((1.d0+a1*rs)*((b1+3*b3*rs)/(2.d0*SQRT(rs))+&
          b2+2*b4*rs))/(pprs*(pprs+1.d0/(2.d0*AA)))

    vxc = (4.d0/3.d0)*ex+ec-(decdrs*rs)/3.d0
   

    IF ((ABS(exc).GT.1.d65).OR.(ABS(vxc).GT.1.d65)) THEN
       WRITE(6,*) 'Problem in PW',n,rs,ec
    ENDIF

    exc=2*exc; vxc=2*vxc      ! change to Rydberg units

    RETURN
  END SUBROUTINE pwldafunc
  !**********************************************************
  ! Subroutine to calculate the exchange correlation functionals
  !   using the form of Perdew, Burke, and Ernzerhof (PRL 77, 3865 (1996))
  !   assuming no spin polarization
  !  Inside this routine, energies are in Hartree units

  SUBROUTINE pbefunc(den,grad,fxc,dfxcdn,dfxcdgbg)
    REAL(8), INTENT(IN) :: den,grad    !density, magnitude of grad(density)
    REAL(8), INTENT(OUT) :: fxc,dfxcdn,dfxcdgbg

    ! Variables depending on den & grad
    REAL(8) :: n,g,kf,rs,ks,s,t

    REAL(8) :: ex,ec,Fx,H,A,pprs,ppt,At2,dFds,dHdt,decdrs,dHdrs,dHdA,dAdrs
    REAL(8) :: term,dHdtbg,dFdsbg

    n=den
    IF (n < machine_zero)  THEN
       fxc=0.d0; dfxcdn=0.d0; dfxcdgbg=0.d0
       RETURN
    ENDIF
    g=grad
    IF (g < machine_zero) g=machine_zero

    kf=(3.d0*(pi**2)*n)**0.3333333333333333333333333333d0
    rs=(3.d0/(4.d0*pi*n))**0.3333333333333333333333333333d0
    ks=SQRT(4.d0*kf/pi)
    s=g/(2.d0*kf*n)
    t=g/(2.d0*ks*n)

    ex=-3.d0*kf/(4.d0*pi)
    pprs=SQRT(rs)*(b1+b3*rs)+rs*(b2+b4*rs)
    !ec=-2.d0*AA*(1.d0+a1*rs)*ddlog(1.d0+1.d0/(2.d0*AA*pprs))
    term=Logofterm(1.d0/(2.d0*AA*pprs))
    ec=-2.d0*AA*(1.d0+a1*rs)*term
    Fx=1.d0+kappa -kappa/(1.d0+(mu/kappa)*s*s)
    A=Aofec(ec)
    At2=A*t*t
    ppt=(1.d0+At2*(1.d0+At2))
    !H=gamm*ddlog(1.d0+(betabygamm)*(t*t)*((1.d0+At2)/ppt))
    H=gamm*Logofterm((betabygamm)*(t*t)*((1.d0+At2)/ppt))

    fxc=n*(ex*Fx+ec+H)

    dFds = (2.d0*mu*s)/(1.d0+(mu/kappa)*(s**2))**2
    dFdsbg = ((2.d0*mu)/(1.d0+(mu/kappa)*(s**2))**2)/(2.d0*kf*n)
    dHdt = (2.d0*t*beta*gamm*(1.d0+2.d0*At2))/&
           ((gamm*ppt+beta*t*t*(1.d0+At2))*ppt)
    dHdtbg = ((2.d0*beta*gamm*(1.d0+ & 
         2.d0*At2))/((gamm*ppt+beta*t*t*(1.d0+At2))*ppt))/(2.d0*ks*n)
    !decdrs=-(2.d0*AA*a1)*ddlog(1.d0+1.d0/(2*AA*pprs)) &
    decdrs=-(2.d0*AA*a1)*term &
         +((1.d0+a1*rs)*((b1+3*b3*rs)/(2.d0*SQRT(rs))+&
             b2+2*b4*rs))/(pprs*(pprs+1.d0/(2.d0*AA)))
    dHdA=-((2.d0+At2)*(At2*t*t*t*t*beta*gamm))/&
            ((gamm*ppt+beta*t*t*(1.d0+At2))*ppt)
    dAdrs=ddexp(-ec/gamm)*A*A*decdrs/beta
    dHdrs=dHdA*dAdrs

    dfxcdn = (4.d0/3.d0)*ex*(Fx-dFds*s)+ec-(decdrs*rs)/3.d0+&
             H-(dHdrs*rs)/3.d0 &
         - (7.d0/6.d0)*dHdt*t

    dfxcdgbg = ex*dFdsbg/(2.d0*kf) + dHdtbg/(2.d0*ks)
    !write(400,'(1p5e15.7)') fxc,dfxcdn,dfxcdgbg,dHdA,dAdrs
    !write(401,'(1p8e15.7)') ex,ec,dFdsbg,dHdtbg,kf,ks,n,dHdt
    
    IF ((ABS(fxc).GT.1.d65).OR.(ABS(dfxcdn).GT.1.d65)&
        .OR.(ABS(dfxcdgbg).GT.1.d65)) THEN
       WRITE(6,*) 'Problem in PBE',n,g,rs,s,t,ec,A,H
       stop
    ENDIF

  fxc=2*fxc; dfxcdn=2*dfxcdn; dfxcdgbg=2*dfxcdgbg   ! change to Rydberg units


    RETURN
  END SUBROUTINE pbefunc

  !**********************************************************
  ! Subroutine to calculate the exchange correlation functionals
  !   using the form of Perdew, Burke, and Ernzerhof (PRL 77, 3865 (1996))
  !   assuming the spin polarization
  !   gradient inputs are gradup==|grad denup|
  !                       graddn==|grad dendn|
  !                       grad  ==|grad (denup+dendn)|
  !  Inside this routine, energies are in Hartree units
  !  We use relationship of  Vxceven=(vxcup+vxcdn)/2
  !                          Vxcodd=(vxcup-vxcdn)/2
  !              Gradient exchange terms: dFxdgbgup,dFxdgbgdn
  !              Gradient correlation term: dcdgbg

 SUBROUTINE Sggapbefunc(denup,gradup,dendn,graddn,grad, &
         fxc, dfxcdneven,dfxcdnodd, dfxdgbgup,dfxdgbgdn,dcdgbg)
    REAL(8), INTENT(IN) :: denup,gradup,dendn,graddn,grad  
                         !density, magnitude of grad(density)
    REAL(8), INTENT(OUT) :: fxc,dfxcdneven,dfxcdnodd,dfxdgbgup,dfxdgbgdn,dcdgbg

    ! Variables depending on den & grad
    REAL(8) :: nup,ndn,n,g,gup,gdn,kf,kfup,kfdn,rs,&
             ks,s,sup,sdn,t,dtdnup,dtdndn,zeta
    REAL(8) :: c1,c2,c3,c4,c5,c6,c7,c8,phi,dphidzeta,&
             fzeta,dfdzeta,dzetadnup,dzetadndn
    REAL(8) :: pprsec0, pprsec1,pprsalp,termec0,termec1,&
             termalp,ec0,ec1,alpc,drsdnup,drsdndn,drsdn
    REAL(8) :: dpdrsec0,dpdrsec1,dpdrsalp,dec0drs,dec1drs,&
             dalpdrs,decdrs,decdzeta

    REAL(8) :: ex,exup,exdn,ec,Fx,Fxup,Fxdn,H,A,ppt,At2,&
           dFds,dFdsup,dFdsdn,dHdt,dHdrs,dHdA,dAdrs
    REAL(8) :: dHdphi,dAdphi,dHdnup,dHdndn,dfxdnup,dfxdndn,&
               decdnup,decdndn,dAdzeta,dHdzeta
    REAL(8) :: dHdtbg,dFdsbg,ddfxc,dFdsbgup,dFdsbgdn,dfxcdgbg   
                !,dfxdgbgup,dfxdgbgdn,dcdgbg

    ndn=dendn
    nup=denup
    IF (ndn < machine_zero) ndn=machine_zero
    IF (nup < machine_zero) nup=machine_zero
    n=nup+ndn
    !write(6,*) 'inside pbefunc,denup and dendn=',denup,dendn
    !IF (ndn < machine_zero.OR.nup < machine_zero)  THEN
    !   fxc=0.d0; dfxcdneven=0.d0; dfxcdnodd=0.d0; 
    !          dfxdgbgup=0.d0;dfxdgbgdn=0.d0;dcdgbg=0.d0;
    ! dfxcdgbgup=0;dfxcdgbgdn=0;
    !   RETURN
    !ENDIF
    gdn=graddn      !(already abs)ABS(graddn)
    gup=gradup      !(already abs)ABS(gradup)
    g=grad
    IF (g < machine_zero) g=machine_zero
    IF (gup < machine_zero) gup=machine_zero
    IF (gdn < machine_zero) gdn=machine_zero
    !write(6,*) 'next step'
    kf=(3.d0*(pi**2)*n)**0.3333333333333333333333333333d0
    kfup=(6.d0*(pi**2)*nup)**0.333333333333333333333333333d0 
    kfdn=(6.d0*(pi**2)*ndn)**0.333333333333333333333333333d0
    rs=(3.d0/(4.d0*pi*n))**0.3333333333333333333333333333d0
    ks=SQRT(4.d0*kf/pi)
    s=g/(2.d0*kf*n)
    sup=gup/(2.d0*kfup*nup)
    sdn=gdn/(2.d0*kfdn*ndn)
    drsdn=-rs/(3.d0*n)
    drsdnup=-rs/(3.d0*n)
    drsdndn=-rs/(3.d0*n)
    zeta=(nup-ndn)/n
    dzetadnup=2.d0*ndn/(n**2.d0)
    dzetadndn=-2.d0*nup/(n**2.d0)
    
    c1=(1+zeta)**1.33333333333333333333333333333d0
    c2=(1-zeta)**1.33333333333333333333333333333d0
    c3=(1+zeta)**0.66666666666666666666666666667d0
    c4=(1-zeta)**0.66666666666666666666666666667d0
    c5=(1/((1+zeta)**2+machine_zero))**0.1666666666666666666666666667d0
    c6=(1/((1-zeta)**2+machine_zero))**0.1666666666666666666666666667d0
    c7=(1+zeta)**0.33333333333333333333333333333d0
    c8=(1-zeta)**0.33333333333333333333333333333d0
    
    phi=(c3+c4)/2.d0
    t=g/(2.d0*ks*phi*n)
    dtdnup=-7.d0*t/(6.d0*n)-t*(c5-c6)*dzetadnup/(3.d0*phi)
    dtdndn=-7.d0*t/(6.d0*n)-t*(c5-c6)*dzetadndn/(3.d0*phi)
    dphidzeta=(c5-c6)/3.d0
    fzeta=(c1+c2-2.d0)/(2.d0**1.3333333333333333333333333333d0-2.d0)
    dfdzeta=4.d0*(c7-c8)/(3.d0*(2.d0**1.3333333333333333333333333333d0-2.d0))
    
    ex=-3.d0*kf/(4.d0*pi)
    exup=-3.d0*kfup/(4.d0*pi)
    exdn=-3.d0*kfdn/(4.d0*pi)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  first derivative of ec
!!!
    pprsec0=SQRT(rs)*(b1ec0+b3ec0*rs)+rs*(b2ec0+b4ec0*rs) 
    pprsec1=SQRT(rs)*(b1ec1+b3ec1*rs)+rs*(b2ec1+b4ec1*rs)
    pprsalp=SQRT(rs)*(b1alp+b3alp*rs)+rs*(b2alp+b4alp*rs)

    !ec=-2.d0*AA*(1.d0+a1*rs)*ddlog(1.d0+1.d0/(2.d0*AA*pprs))
    termec0=Logofterm(1.d0/(2.d0*AAec0*pprsec0))
    termec1=Logofterm(1.d0/(2.d0*AAec1*pprsec1))
    termalp=Logofterm(1.d0/(2.d0*AAalp*pprsalp))

    ec0=-2.d0*AAec0*(1.d0+a1ec0*rs)*termec0
    ec1=-2.d0*AAec1*(1.d0+a1ec1*rs)*termec1
    alpc=2.d0*AAalp*(1.d0+a1alp*rs)*termalp

    ec=ec0+alpc*fzeta*(1-zeta**4.d0)/1.709921d0+(ec1-ec0)*fzeta*zeta**4.d0
   
    dpdrsec0=(b1ec0+3.d0*b3ec0*rs)/(2.d0*sqrt(rs))+b2ec0+b4ec0*rs*2.d0
    dpdrsec1=(b1ec1+3.d0*b3ec1*rs)/(2.d0*sqrt(rs))+b2ec1+b4ec1*rs*2.d0
    dpdrsalp=(b1alp+3.d0*b3alp*rs)/(2.d0*sqrt(rs))+b2alp+b4alp*rs*2.d0
    dec0drs=-(2.d0*AAec0*a1ec0)*termec0+&
       2.d0*AAec0*(1.d0+a1ec0*rs)*dpdrsec0/(pprsec0*(1.d0+2.d0*AAec0*pprsec0))
    dec1drs=-(2.d0*AAec1*a1ec1)*termec1+&
       2.d0*AAec1*(1.d0+a1ec1*rs)*dpdrsec1/(pprsec1*(1.d0+2.d0*AAec1*pprsec1))
    dalpdrs=2.d0*AAalp*a1alp*termalp-&
       2.d0*AAalp*(1.d0+a1alp*rs)*dpdrsalp/(pprsalp*(1.d0+2.d0*AAalp*pprsalp))

    decdrs=dec0drs+dalpdrs*fzeta*(1-zeta**4.d0)/1.709921d0+&
          (dec1drs-dec0drs)*fzeta*zeta**4.d0
    decdzeta=alpc*(dfdzeta*(1-zeta**4.d0)-4.d0*fzeta*zeta**3.d0)/1.709921d0+&
       (ec1-ec0)*(dfdzeta*zeta**4.d0+4.d0*fzeta*zeta**3.d0)

    decdnup=decdrs*drsdnup+decdzeta*dzetadnup
    decdndn=decdrs*drsdndn+decdzeta*dzetadndn
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  calculat fxc here
    
    Fx=1.d0+kappa -kappa/(1.d0+(mu/kappa)*s*s)
    Fxup=1.d0+ kappa -kappa/(1.d0+(mu/kappa)*sup*sup)
    Fxdn=1.d0+ kappa -kappa/(1.d0+(mu/kappa)*sdn*sdn)
   
    A=Aofec(ec/phi**3)
    At2=A*t*t
    ppt=(1.d0+At2*(1.d0+At2))
    !H=gamm*ddlog(1.d0+(betabygamm)*(t*t)*((1.d0+At2)/ppt))
    H=gamm*(phi**3)*Logofterm((betabygamm)*(t*t)*((1.d0+At2)/ppt))

    !fxc=n*(ex*Fx+ec+H)
    fxc=nup*exup*Fxup+ndn*exdn*Fxdn+n*(ec+H)

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!  calculate dfxup/dnup

    dFds = (2.d0*mu*s)/(1.d0+(mu/kappa)*(s**2))**2
    dFdsup= (2.d0*mu*sup)/(1.d0+(mu/kappa)*(sup**2))**2
    dFdsdn= (2.d0*mu*sdn)/(1.d0+(mu/kappa)*(sdn**2))**2
    dfxdnup=-kfup*(Fxup-dFdsup*sup)/pi
    dfxdndn=-kfdn*(Fxdn-dFdsdn*sdn)/pi

    dFdsbg = ((2.d0*mu)/(1.d0+(mu/kappa)*(s**2))**2)/(2.d0*kf*n)
    dFdsbgup = ((2.d0*mu)/(1.d0+(mu/kappa)*(sup**2))**2)/(2.d0*kfup*nup)
    dFdsbgdn = ((2.d0*mu)/(1.d0+(mu/kappa)*(sdn**2))**2)/(2.d0*kfdn*ndn)
    dHdt = (2.d0*t*beta*(phi**3.d0)*gamm*(1.d0+2.d0*At2))/&
           ((gamm*ppt+beta*t*t*(1.d0+At2))*ppt)
    dHdtbg = ((2.d0*beta*(phi**2.d0)*gamm*(1.d0+ & 
         2.d0*At2))/((gamm*ppt+beta*t*t*(1.d0+At2))*ppt))/(2.d0*ks*n)
    dHdA=-((2.d0+At2)*(At2*t*t*t*t*beta*gamm*phi**3.d0))/&
              ((gamm*ppt+beta*t*t*(1.d0+At2))*ppt)
    dAdrs=ddexp(-ec/(gamm*(phi**3.d0)))*A*A*decdrs/(beta*phi**3.d0)
    dAdzeta=ddexp(-ec/(gamm*(phi**3.d0)))*A*A*decdzeta/(beta*phi**3.d0)
    dHdrs=dHdA*dAdrs
    dHdzeta=dHdA*dAdzeta

    dHdphi=3.d0*gamm*(phi**2)*Logofterm((betabygamm)*(t*t)*((1.d0+At2)/ppt))
    dAdphi=-3.d0*A*A*ec*exp(-ec/(gamm*(phi**3)))/(beta*(phi**4))
 
    dHdnup=dHdrs*drsdnup+dHdzeta*dzetadnup+dHdphi*dphidzeta*dzetadnup+&
          dHdA*dAdphi*dphidzeta*dzetadnup+dHdt*dtdnup
    dHdndn=dHdrs*drsdndn+dHdzeta*dzetadndn+dHdphi*dphidzeta*dzetadndn+&
          dHdA*dAdphi*dphidzeta*dzetadndn+dHdt*dtdndn
    
    dfxcdneven=(dfxdnup+dfxdndn)/2.d0+ec+H+n*(decdnup+dHdnup+decdndn+&
          dHdndn)/2.d0   !dfxceven=(dfxcup+dfxcdn)/2
    dfxcdnodd=(dfxdnup-dfxdndn)/2.d0+n*(decdnup+dHdnup-decdndn-dHdndn)/2.d0         !dfxcodd=(dfxup-dfxcdn)/2
   
    dfxdgbgup = exup*dFdsbgup/(2.d0*kfup)    
    !there are upspin and dnspin gradient parts for 
    dfxdgbgdn = exdn*dFdsbgdn/(2.d0*kfdn)    !dfxdgbg*grad(n)
    dcdgbg = dHdtbg/(2.d0*ks*phi)         !one gradient part for dHdtbg*grad(n)
    IF ((ABS(fxc).GT.1.d65).OR.(ABS(dfxcdneven).GT.1.d65).OR.&
       (ABS(dfxcdnodd).GT.1.d65).OR.(ABS(dfxdgbgup).GT.1.d65).OR.&
       (ABS(dfxdgbgdn).GT.1.d65).OR.ABS(dcdgbg).GT.1.d65) THEN
       WRITE(6,*) 'Problem in PBE',n,g,rs,s,t,ec,A,H
    ENDIF

  ! Change to Rydberg units
    fxc=2*fxc; dfxcdneven=2*dfxcdneven; dfxcdnodd=2*dfxcdnodd
    dfxdgbgup=2*dfxdgbgup; dfxdgbgdn=2*dfxdgbgdn; dcdgbg=2*dcdgbg

    RETURN
END SUBROUTINE Sggapbefunc
  !*******************************************************************
  !
  ! Function Logofterm -- needed to take care of behavior for small term 
  !  Evaluates log(1+term)

  FUNCTION Logofterm(term)

    REAL(8) :: term, Logofterm

    IF (ABS(term)>machine_precision) THEN
       Logofterm=ddlog(1.d0+term)
    ELSE
       Logofterm=term
    ENDIF

    RETURN
  END FUNCTION Logofterm

  !*******************************************************************
  !
  ! Function Aofec -- needed to take care of behavior for small ec

  FUNCTION Aofec(ec)

    REAL(8) :: ec, Aofec

    IF (ABS(ec)>machine_precision) THEN
       Aofec=betabygamm/(ddexp(-ec/gamm)-1.d0)
    ELSEIF (ABS(ec)>machine_zero) THEN
       Aofec=beta/(-ec)
    ELSE
       Aofec=-beta*DSIGN(machine_infinity,ec)
    ENDIF

    RETURN
  END FUNCTION Aofec

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! Exchange-correlational functions for LSDA according to
!               Perdew and Wang
! Program written by Ping Tang and NAWH
! Last modified 5/25/02
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!**********************************************************
! Subroutine to calculate the LSDA exchange correlation functionals
!   using the form of Perdew and Wang (PRB 45, 13244 (1992)
!   assuming spin polarization
!  Inside this routine, energies are in Hartree units
!  Input den=denup+dendown, mu=denup-dendown
!  Output Exc,Vexc(even), Vexc(odd)
!   Where Vxceven=(vxcup+vxcdn)/2
!         Vxcodd=(vxcup-vxcdn)/2

 Subroutine pwlsdafunc(den,mu,Exc,vxceven,vxcodd)
  Real(8), intent(IN) :: den,mu         !density
  Real(8), intent(OUT) :: Exc,vxceven,vxcodd

! Variables depending on den
  Real(8) :: n,g,kf,rs,zeta,nup,ndown

  Real(8) :: ex,ec,pprsec0,decdrs,dexdrs,dexdzeta,decdzeta,ec0,ec1,alpc
  Real(8) :: fzeta,pprsec1,pprsalp,dpdrsec0,dpdrsec1,dpdrsalp
  Real(8) :: dec0drs,dec1drs,dalpdrs,c1,c2,c3,c4,drsdn,dzetadnup,dzetadndn
  Real(8) :: dfdzeta,coef,termec0,termec1,termalp

  n=den
  If (n < machine_zero .or. &
       abs(mu) > n)  then
    Exc=0.d0; vxceven=0.d0
    vxcodd=0.d0
    return
  EndIf

  kf=(3.d0*(pi**2)*n)**0.3333333333333333333333333333d0
  rs=(3.d0/(4.d0*pi*n))**0.3333333333333333333333333333d0
  drsdn=-rs/(3*n)
  zeta=mu/n

  c1=(1.d0+zeta)**1.33333333333333333333333d0
  c2=(1.d0-zeta)**1.33333333333333333333333d0
  c3=(1.d0+zeta)**0.33333333333333333333333d0
  c4=(1.d0-zeta)**0.33333333333333333333333d0

  coef=3.d0/(8.d0*pi)*(9.d0*pi/4.d0)**0.3333333333333333333d0/rs
  ex=-coef*(c1+c2)
  dexdrs=-ex/rs
  dexdzeta=-4.d0*coef*(c3-c4)/3.d0

  fzeta=(c1+c2-2.d0)/(2.d0**1.33333333333333333333d0-2.d0)
  dfdzeta=4*(c3-c4)/(3.d0*(2.d0**1.333333333333333333333333d0-2.d0))

  pprsec0=sqrt(rs)*(b1ec0+b3ec0*rs)+rs*(b2ec0+b4ec0*rs)
  pprsec1=sqrt(rs)*(b1ec1+b3ec1*rs)+rs*(b2ec1+b4ec1*rs)
  pprsalp=sqrt(rs)*(b1alp+b3alp*rs)+rs*(b2alp+b4alp*rs)

  !ec=-2.d0*AA*(1.d0+a1*rs)*ddlog(1.d0+1.d0/(2.d0*AAec0*pprs))
  termec0=Logofterm(1.d0/(2.d0*AAec0*pprsec0))
  termec1=Logofterm(1.d0/(2.d0*AAec1*pprsec1))
  termalp=Logofterm(1.d0/(2.d0*AAalp*pprsalp))

  ec0=-2.d0*AAec0*(1.d0+a1ec0*rs)*termec0
  ec1=-2.d0*AAec1*(1.d0+a1ec1*rs)*termec1
  alpc=2.d0*AAalp*(1.d0+a1alp*rs)*termalp
  ec=ec0+alpc*fzeta*(1-zeta**4.d0)/1.709921d0+(ec1-ec0)*fzeta*zeta**4.d0
  
  dpdrsec0=(b1ec0+3.d0*b3ec0*rs)/(2.d0*sqrt(rs))+b2ec0+b4ec0*rs*2.d0
  dpdrsec1=(b1ec1+3.d0*b3ec1*rs)/(2.d0*sqrt(rs))+b2ec1+b4ec1*rs*2.d0
  dpdrsalp=(b1alp+3.d0*b3alp*rs)/(2.d0*sqrt(rs))+b2alp+b4alp*rs*2.d0
  dec0drs=-(2.d0*AAec0*a1ec0)*termec0+&
     2.d0*AAec0*(1.d0+a1ec0*rs)*dpdrsec0/(pprsec0*(1.d0+2.d0*AAec0*pprsec0))
  dec1drs=-2.d0*AAec1*a1ec1*termec1+&
     2.d0*AAec1*(1.d0+a1ec1*rs)*dpdrsec1/(pprsec1*(1.d0+2.d0*AAec0*pprsec1))
  dalpdrs=2.d0*AAalp*a1alp*termalp-&
     2.d0*AAalp*(1.d0+a1alp*rs)*dpdrsalp/(pprsalp*(1.d0+2.d0*AAalp*pprsalp))
  
   decdrs=dec0drs+dalpdrs*fzeta*(1-zeta**4.d0)/1.709921d0+&
     (dec1drs-dec0drs)*fzeta*zeta**4.d0
   decdzeta=alpc*(dfdzeta*(1-zeta**4.d0)-4.d0*fzeta*zeta**3.d0)/1.709921d0+&
     (ec1-ec0)*(dfdzeta*zeta**4.d0+4.d0*fzeta*zeta**3.d0)

  Exc=ex+ec
 
  vxceven=ex+ec-(decdrs+dexdrs)*rs/3-(decdzeta+dexdzeta)*zeta
  vxcodd =(decdzeta+dexdzeta)
  
  if ((abs(Exc).gt.1.d65).or.(abs(vxceven).gt.1.d65).or.(abs(vxcodd).gt.1.d65)) then
      write(6,*) 'Problem in PW',n,rs,ec
  endif

  Exc=2*Exc; vxceven=2*vxceven    ! change to Rydberg units
  vxcodd=2*vxcodd
  
  !do i=1,n
  !   write(155, '(1p3e15.7)') r(i),vxceven(i), vxcodd(i)
  !enddo
  !stop
  return
  end subroutine


end Module
