!******************************************************************************
!
! File : fileio.f90
!   by : Alan Tackett
!   on : 04/29/97
!  for : MG-PAW
!
!  Module contains definitions for allocating the File Units and Names
!
!******************************************************************************

Module fileio

Use atom_data
Use paw_inout
Use mem_data
Use misc       

Implicit NONE!!!!

Integer :: FD_Psi               !** Base unit for Psi
Integer :: FD_RadProj           !** Base unit for the Radial Projectors
Integer :: FD_BlochK            !** Base unit for Atom Phase factors
Integer :: FD_Ylm               !** Base unit for Ylm structure factors
Integer :: FD_Proj              !** Base unit for the complete projetors(G)
Integer :: FD_RS_Proj           !** Base unit for the complete projetors(RS)
Integer :: FD_Oinv              !** Base unit for the Oinv

Integer, PRIVATE :: MaxAtom


Interface WriteFile_LOW    !** Interface for WriteFile_LOW
  Module Procedure CMPLX_WriteFile_LOW
  Module Procedure REAL_WriteFile_LOW
End Interface

Interface ReadFile_LOW    !** Interface for ReadFile_LOW
  Module Procedure CMPLX_ReadFile_LOW
  Module Procedure REAL_ReadFile_LOW
End Interface

Type (Word_Context), PRIVATE, Pointer :: WC  !** Optional Context for IO errors
Integer, PRIVATE   :: Error_Out  !** Optional error unit for error messages

!******************************************************************************
Contains
!******************************************************************************
!******************************************************************************
!
!  AssignBaseFileUnits - Assigns the Base file units to all the devices
!
!  BaseFile - Base File unit to use
!
!******************************************************************************

Subroutine AssignBaseFileUnits(BaseUnit)
  Integer, Intent(IN) :: BaseUnit


  FD_Psi     = BaseUnit
  FD_Oinv    = FD_Psi + 1
  FD_BlochK  = FD_Oinv + 1
  FD_Ylm     = FD_BlochK + 1
  FD_RadProj = FD_Ylm + 1
  FD_Proj    = FD_RadProj + Atom_Types
  FD_RS_Proj = FD_PRoj + Specific_Atoms

  Write(Log_Unit,*) 'ABFU: FD_Psi=', FD_Psi
  Write(Log_Unit,*) 'ABFU: FD_Oinv=', FD_Oinv
  Write(Log_Unit,*) 'ABFU: FD_BlochK=', FD_BlochK
  Write(Log_Unit,*) 'ABFU: FD_Ylm=',FD_Ylm
  Write(Log_Unit,*) 'ABFU: FD_RadProj=',FD_RadProj
  Write(Log_Unit,*) 'ABFU: FD_Proj=',FD_Proj
  Write(Log_Unit,*) 'ABFU: FD_RS_Proj=',FD_RS_Proj

  Return
End Subroutine




!******************************************************************************
!
!  Init_FileIO - Initializes the file IO module for use
!
!******************************************************************************

Subroutine Init_FileIO

  Call Init_FileIO_Low(paw_WC, Error_Unit)

  Return
End Subroutine


!******************************************************************************
!
! CMPLX_OpenBuffer - Opens the set of buffer files.
!
! FD_Base - Base Unit
! IDText  - Text string containing where OpenBuffer is called from
!
!******************************************************************************

Subroutine CMPLX_OpenBuffer(FD_base, IDText)
  Integer,           Intent(IN) :: FD_Base
  Character*(*)                 :: IDText

  Integer       :: i, err, last
  Integer       :: RecSize
  Character*100 :: msg
  Complex :: c

     RecSize = SizeOf_C*PsiArraySize

     Write(Log_Unit, *) 'OpenBuffer: Unit=',FD_Base, &
          ' * Level=',i, ' * Size=',PsiArraySize   
     Open(FD_base, Form="UNFORMATTED", status="SCRATCH", &
          Access="DIRECT", RECL=REcSize, IOSTAT=err)
     Write(msg,*) ' Error opening file. Rec Size=',recSize
     Msg = TRIM(IDText) // Msg
     Call Check_Error(err, msg, Error_Out, .TRUE., PAW_wc, IDText)

  Return
End Subroutine

!******************************************************************************
!
! REAL_OpenBuffer - Opens the set of buffer files.
!
! FD_Base - Base Unit
! IDText  - Text string containing where OpenBuffer is called from
!
!******************************************************************************

Subroutine REAL_OpenBuffer(FD_base, IDText)
  Integer,           Intent(IN) :: FD_Base
  Character*(*)                 :: IDText

  Integer       :: i, err, last
  Integer       :: RecSize
  Character*100 :: msg


     RecSize = SizeOf_Real*PsiArraySize

     Write(Log_Unit, *) 'OpenBuffer: Unit=',FD_Base, &
           ' * Size=',PsiArraySize
     Open(FD_base+i, Form="UNFORMATTED", status="SCRATCH", &
          Access="DIRECT", RECL=REcSize, IOSTAT=err)
     Write(msg,*) ' Error opening file. Rec Size=',recSize
     Msg = TRIM(IDText) // Msg
     Call Check_Error(err, msg, Error_Out, .TRUE., PAW_wc, IDText)

  Return
End Subroutine

!******************************************************************************
!
!  Init_FileIO_Low - Initializes the module routines for use
!
!    WordC - Optional Word context
!    Err_Unit - Error unit number
!
!******************************************************************************

Subroutine Init_FileIO_Low(WordC, Err_Unit)
  Type (Word_Context), OPTIONAL, TARGET    :: WordC
  Integer,                       Intent(IN) :: Err_Unit

  Error_Out = Err_Unit
  
  Nullify(WC)
  If (Present(WordC)) WC => WordC

  Return
End Subroutine


!******************************************************************************
!
!  CMPLX_WriteFile_LOW - Performs a low level WRITE of a file from disk
!
!  FD_Unit - File unit
!  Rec     - Record Number
!  Ptr     - Ptr for data
!  Msg     - Error Text
!
!******************************************************************************

Subroutine CMPLX_WriteFile_LOW(FD_Unit, Rec,  Ptr, Msg)
  Integer,           Intent(IN)  :: FD_Unit
  Integer,           Intent(IN)  :: Rec
  Complex,   Intent(IN)  :: Ptr(:)
  Character*(*),     Intent(IN)  ::  Msg

  Integer       :: err, i, N

  !! write(Log_Unit,*)  'COMPLXwrite', FD_Unit,rec
  N = Size(Ptr)
  Write(FD_Unit, IOSTAT=err, REC=Rec) (Ptr(i), i=1, N)

  If (err /= 0) then
     Call Check_Error(err, msg, Error_Out, .TRUE., wc, "CMPLX_WriteFile_Low:")
     STOP
  End If
  
  Return
End Subroutine

!******************************************************************************
!
!  CMPLX_ReadFile_LOW - Performs a low level read of a file from disk
!
!  FD_Unit - File unit
!  Rec     - Record Number
!  Ptr     - Ptr for storing data
!  Msg     - Error Text
!
!******************************************************************************

Subroutine CMPLX_ReadFile_LOW(FD_Unit, Rec, Ptr, Msg)
  Integer,         Intent(IN)  :: FD_Unit
  Integer,         Intent(IN)  :: Rec
  Complex, Intent(OUT) :: Ptr(:)
  Character*(*),   Intent(IN)  ::  Msg

  Integer       :: err, i, N
  Real  :: tmp
  !! write(Log_Unit,*)  'COMPLXread', FD_Unit,rec
  N = Size(Ptr)
  Read(FD_Unit, IOSTAT=err, REC=rec) (Ptr(i), i=1, N)

  If (err /= 0) then
     Call Check_Error(err, msg, Error_Out, .TRUE., wc, "CMPLXReadFile_LOW:")
     tmp = 0.0
     tmp = 1.0/tmp
     STOP
  End If
  
  Return
End Subroutine


!******************************************************************************
!
!  REAL_WriteFile_LOW - Performs a low level WRITE of a file from disk
!
!  FD_Unit - File unit
!  Rec     - Record Number
!  Ptr     - Ptr for data
!  Msg     - Error Text
!
!******************************************************************************

Subroutine REAL_WriteFile_LOW(FD_Unit, Rec,  Ptr, Msg)
  Integer,           Intent(IN)  :: FD_Unit
  Integer,           Intent(IN)  :: Rec
  Real,              Intent(IN)  :: Ptr(:)
  Character*(*),     Intent(IN)  ::  Msg

  Integer       :: err, i, N

  N = Size(Ptr)
  Write(FD_Unit, IOSTAT=err, REC=Rec) (Ptr(i), i=1, N)

  If (err /= 0) then
     Call Check_Error(err, msg, Error_Out, .TRUE., wc, "REAL_WriteFile_Low:")
     STOP
  End If
  
  Return
End Subroutine

!******************************************************************************
!
!  REAL_ReadFile_LOW - Performs a low level read of a file from disk
!
!  FD_Unit - File unit
!  Rec     - Record Number
!  Ptr     - Ptr for storing data
!  Msg     - Error Text
!
!******************************************************************************

Subroutine REAL_ReadFile_LOW(FD_Unit, Rec, Ptr, Msg)
  Integer,         Intent(IN)  :: FD_Unit
  Integer,         Intent(IN)  :: Rec
  Real,            Intent(OUT) :: Ptr(:)
  Character*(*),   Intent(IN)  ::  Msg

  Integer       :: err, i, N
  Real  :: tmp

  N = Size(Ptr)
  Read(FD_Unit, IOSTAT=err, REC=rec) (Ptr(i), i=1, N)

  If (err /= 0) then
     Call Check_Error(err, msg, Error_Out, .TRUE., wc, "REAL_ReadFile_LOW:")
     tmp = 0.0
     tmp = 1.0/tmp
     STOP
  End If
  
  Return
End Subroutine

End Module
