!***************************************************************************
!
! File : gaussbzi.f90
!   by : Alan Tackett
!   on : 2/3/97
!  for : PAW Method
!
!
! gaussbzi - Gaussian Smearing Brillouin Zone Integration
!
! This file contains routines to perform BZ integration using a
! Gaussian smearing Method which is outlined in the articles below:
!
!   Fu, C.-L., Ho, K.-M., PRB 28, 5480 (1983).
!   Holzwarth, N.A.W., Zeng, Y., PRB 49, 2351 (1994).
!
!
!  NOTE: DOS is not implemented only NOS
!
! Modified for spin dependence by Ping Tang and N. A. W. Holzwarth
!   last modification 5/23/05
!**************************************************************************

Module gaussbzi

Use bz_data
Use options_data
Use paw_inout

implicit none
Contains

!****************************************************************************
!
! InitGaussBZI - Initialize the Gaussian routines for use.
!
! This routine is assumed to be called from bz_Init(...) contained in the
! BZINT module.  Check there to see what is set in bz.
!
!****************************************************************************

Subroutine InitGaussBZI(bz)
  Type (bz_Struct), Intent(INOUT) :: Bz

  Return
End Subroutine

!****************************************************************************
!
! Gauss_FreeUnused - Frees the unused arrays and makes them point to the
!                    unique versions.  Should not be called
!                    until after GAUSS_WEIGHTS has been called!
!
! The Following Arrays are deallocated in the BZ data structure:
!
!    Weight, Kp, Ke, Kmap, and Cube
!
!
!****************************************************************************

Subroutine Gauss_FreeUnused(BZ)
  Type (BZ_Struct), Intent(INOUT) :: bz

  DeAllocate(BZ%Weight, BZ%Kp, BZ%Ke, BZ%KMap, BZ%Cube)

  BZ%Weight => BZ%WtUniq
  BZ%Kp     => BZ%Ku
  BZ%Ke     => BZ%KEu

  BZ%TotalCubes = 0
  BZ%TotalKPnts = BZ%TotalUniq

  Return
End Subroutine

!****************************************************************************
!
! Gauss_NOS - Calculates the NOS for all bands
!
!   bz   - BZ data structure, uses Fermi energy contained
!
!  The NOS contribution for a particular (band,K-point) is defined as
!
!                 
!                 (         ( E  - E    ) )
!      NOS  =  d  (  1 + erf(  f    n,k ) )  where SUM(d ) = 1
!         n,k   k (         ( --------- ) )             k
!                 (         (   sigma   ) )
!
!
!  The d 's are contained in BZ%WtUniq array
!       k
!
! Return Values
!     Returns the NOS and optionally the NOS for each n,k.
!
! Note: Only uses the Unique K-points, KE's, and Weights!!
!
!****************************************************************************

Real Function gauss_NOS(BZ, Occ)
   type (BZ_Struct), Intent(INOUT) :: bz
   Real, OPTIONAL,   Intent(OUT)   :: Occ(:,:)

   Integer :: i, j ,n,s
   Real    :: NOS, fn, c, Fermi, DERF
   Real, Pointer :: KEu(:,:), Wtu(:,:)

   KEu => BZ%KEu
   Wtu => BZ%WtUniq

   c = 1.0/BZ%Sigma
   Fermi = BZ%Fermi

   NOS = 0
   s=1
   if (spindependence) s=2
   n=BZ%Bands*s    !BZ%Bands == Numbands
   !write(log_unit,*) 'gauss_NOS: begin', n,s
   !call flush(log_unit)

!Write(*,*) 'Gauss_NOS: Fermi=',Fermi   
!Write(Log_unit,*) 'Gauss_NOS: Fermi=',Fermi   
!call flush(log_unit)
   Do i=1, BZ%TotalUniq
      !Do j = 1, BZ%Bands
       Do j = 1, n
         fn = (1 + DERF((Fermi - KEu(j,i))*c))/s

!Write(*,*) 'Gauss_NOS: Kpnt=',i, ' * Band=',j, ' * KE=',KEu(j,i), ' * Occ=',Wtu(j,i)*Fn
!Write(log_unit,*) 'Gauss_NOS: Kpnt=',i, ' * Band=',j, ' * KE=',KEu(j,i), ' * Occ=',Wtu(j,i)*Fn
!call flush(log_unit)
         NOS = NOS + Wtu(j,i)*fn

         IF (Present(Occ)) Occ(j,i) = Wtu(j,i)*fn
      End Do
   End Do

   Gauss_NOS = NOS
   !write(log_unit,*) 'Gaussbzi: NOS',NOS
   !call flush(log_unit)

   Return
End Function

!****************************************************************************
!     
! Gauss_Weights - Calculates the Weights for all bands
!
! Return Values
!     Stores all the UNIQUE weights
!
!****************************************************************************

Subroutine Gauss_Weights(bz)
   type (BZ_Struct), Intent(INOUT) :: bz

   Integer :: i
   
   Real, Pointer :: Wtu(:,:)
   Integer, Pointer :: KMap(:)

   Wtu => Bz%WtUniq
   KMap => BZ%KMap

   Wtu = 0
   Do i = 1, Bz%TotalKpnts
     Wtu(1,KMap(i)) = Wtu(1,KMap(i)) + 1
   End Do

   Wtu(1,:) = Wtu(1,:) / Bz%TotalKpnts
  
   Do i=2, Bz%Bands
      Wtu(i,:) = wtu(1,:)
   End Do

   Return
End Subroutine

End Module
   
