!******************************************************************************
!
! File : gausslib.f90
!   by : Alan Tackett
!   on : 04/25/00
!  for : Convert_SEPM
!
!  Searches and manipulates a table of gaussian functions
!
!******************************************************************************

Module gausslib

Use gaussfunc_data

Implicit None !!!!

!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
!  Combine_Gauss - Combines two gaussian lists
!
!     A, B  - Lists to combine
!     C     - Result, C = alpha*A +beta*B
!     alpha, beta - Scale factors for A and B
!
!******************************************************************************

Subroutine Combine_Gauss(A,B,C, alpha, beta)
  Type (Gaussfunc), Intent(IN) :: A
  Type (Gaussfunc), Intent(IN) :: B
  Type (Gaussfunc), Intent(OUT):: C
  Real,             Intent(IN) :: alpha
  Real,             Intent(IN) :: beta

  Real, PARAMETER :: tol = 1E-8
  Integer :: i,j,k, n
  Type (Gaussfunc)   :: AB
  Real :: err, Diff(2)

  !** Make temp result **
  i = A%N+B%N
  AB%N = A%N
  Allocate(AB%Coeff(3,i))

  !** Copy initial gauss func **
  AB%Coeff(:,1:A%N) = A%Coeff
  AB%Coeff(GAUSS_SCALE, 1:A%N) = alpha*AB%Coeff(GAUSS_SCALE,1:A%N)
  n = A%N

  !** Now try to merge with the B function **
  Do i=1, B%N
     j = 0
     err = 2*tol
     Do While ((j<A%N) .AND. (err > tol))
        j = j + 1
        diff = A%Coeff(1:2,j) - B%Coeff(1:2,i)
        err = SQRT(DOT_PRODUCT(diff,diff))
     End do

     if (err < tol) then
        AB%Coeff(GAUSS_SCALE,j) = AB%Coeff(GAUSS_SCALE,j) + &
             beta*B%Coeff(GAUSS_SCALE,i)
     else
        AB%N = AB%N + 1
        n = AB%N
        AB%Coeff(:,n) = B%Coeff(:,i)
        AB%Coeff(GAUSS_SCALE,n) = beta*B%Coeff(GAUSS_SCALE, i)
     End if
  End do

  !** Now reallocate the final 'C' **
  C%N = AB%N
  Allocate(C%Coeff(3,C%N))
  C%Coeff = AB%Coeff(:,1:C%N)

  !** finally clean up by dealloc'ing AB **
  DeAllocate(AB%Coeff)

  Return
End Subroutine


!******************************************************************************
!
!  EvalGauss - Evaluates a gaussian function at the specified point
!
!      Func - Function
!      R    - Position
!  
!******************************************************************************

Real Function EvalGauss(Func, R)
  Type (GaussFunc), Intent(IN) :: Func
  Real,             Intent(IN) :: R

  Integer :: i
  Real    :: gsum, pow

  gsum = 0
  Do i=1, Func%N
     pow = (R - Func%Coeff(GAUSS_SHIFT, i))/Func%Coeff(GAUSS_SIGMA,i)
     pow = -pow*pow
     if (pow > -40) gsum = gsum + Func%Coeff(GAUSS_SCALE, i)*exp(pow)
  End do

  EvalGauss = gsum
  Return
End function

end Module


