!******************************************************************************
!
! File : gpoints.f90
!   by : Alan Tackett
!   on : 04/14/99
!  for : PAW 
!
! Module containing the G-points used for evaluation of the Wave functions
! and potentials.
!
! The arrays Gpnt(:,:) and Npnt(:,:) contain only the unique reciprocal
! lattice vectors, i.e. k and -k are not stored.  The vectors are sorted
! in ascending order with k=(0,0,0) in slot 1.
!
! Gpnt(:,:) => Gpnt(1:3, i) : Kx,Ky,Kz for the ith k-vector
!              Gpnt(4,i)    : |k|, magnitude of the K-vector
!
! NPnt(:,:) => NPnt(1:3,i)  : Nx,Ny,Nz(integers) for the ith K-vector
!
! FFTMap_Low(:) - FFT mapping for Psi functions(Gcut_Psi)
! FFTMap_High(:) - FFT mapping for functions with the larger Gcut_Den
!
!  modified 2/2/06 to filter wfn's so that |k + G| < Gcut_LOW
!
!******************************************************************************

Module gpoints

Use mathlib
Use word
Use misc
Use paw_inout
Use options_data
Use fftw
Use btree
Use crystal_data
Use bz_data

Implicit NONE!!!!!

Integer, PARAMETER :: G_LOW  = 1   !** Low PW cutoff index
Integer, PARAMETER :: G_HIGH = 2   !** High PW cutoff index
Integer, PARAMETER :: G_PROJ = 3   !** PW cutoff index for RS proj


Real,    Pointer :: Gpnt(:,:)        !** G-vector list
Integer, Pointer :: NPnt(:,:)        !** Coresponding integer G-vector
Logical, Pointer :: G_NewMag(:)     !** TRUE if Gpnt(4,i-1)/=Gpnt(4,i)

Integer :: NewMag_Size(3)   !** Size of Non-redundant G-mags
Integer :: Gpnt_Size(3)     !** Number of G-points for Low/High cutoffs <= 0
Integer :: Gall_Size(3)     !** Number of G-points for Low/High cutoffs (all)
Integer, Pointer :: FFTmap_Low(:)
Integer, Pointer :: FFTmap_High(:)
Integer, Pointer :: FFTmap_PROJ(:)

Integer :: FFT_Grid(4,3)       !** Grid Dimensions - (:,1)=Low

Real, PARAMETER  :: Filter_Min = 1E-30  !** for FilterValue pgm

Integer, PRIVATE :: Count !** Used for converting the btree to 1d array
Real,    PRIVATE :: G_Recip(3,3)   

! data structure needed for filtering
Type Gcutfilter_struct
   Integer :: nzeros
   Integer, pointer :: zeromap(:)
End type

Type (Gcutfilter_struct), allocatable :: Gcutfilter(:)

!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
!  ConvertGpoints - Converts the G-points list to the 1-D array
!
!******************************************************************************

Logical Function ConvertGpoints(Node)
  Type (BinaryTreeData) :: Node

  Count = Count + 1

!Write(Log_Unit,*) 'ConvertGpoints:  Count=',count
!Write(Log_Unit,*) 'ConvertGpoints:  Gmag=',Node%Gmag, ' * N=',Node%Npnt
!write(Log_Unit,*) 'ConvertGpoints : Recip=',G_recip
  Gpnt(1:3, Count) = MATMUL(G_Recip, Node%Npnt)
  Gpnt(4, Count) = Node%Gmag
  Npnt(:, Count) = Node%Npnt
  
  ConvertGpoints = .TRUE.

  Return
End Function

!******************************************************************************
!
! GenGPoints - Generates the list of G-points 
!
!   Recip  - Reciprocal Lattice Vectors
!   Gcut   - Curvilinear Plane wave cutoff
!   GSize  - Number of curvilinear Plane Waves generated
!   Gpnt   - List of Gpnts
!
!   NOTE: If Gsize <= 0 then Gpnts is not used but GSize contains the number
!         of planewaves needed. 
!******************************************************************************

Subroutine GenGpoints(Recip, Gcut, Gsize,WC)
  Real,    Intent(IN)    :: Recip(:,:)
  Real,    Intent(IN)    :: Gcut
  Integer, Intent(OUT) :: Gsize
   Type (Word_Context), Intent(INOUT) :: WC

  Integer :: i,j,k, NMax, t, Used ,m(3)
  Real    :: g, g2, gv(3)
  Character*100 :: Token
  Logical          :: Ok, Unique
  Type (BinaryTreeNode), Pointer     :: Root, Match
  Type (BinaryTreeData), Pointer     :: Item, Test
  
  Nullify(Root)


           !*** Estimate the Max indices for Gcut ****
  do i=1,3
     g=DOT_PRODUCT(Recip(:,i),Recip(:,i))
     j=i+1-((i)/3)*3
     k=i+2-((i+1)/3)*3
     g=MIN(g,DOT_PRODUCT((Recip(:,i)-Recip(:,j)),(Recip(:,i)-Recip(:,j))))
     g=MIN(g,DOT_PRODUCT((Recip(:,i)+Recip(:,j)),(Recip(:,i)+Recip(:,j))))
     g=MIN(g,DOT_PRODUCT((Recip(:,i)-Recip(:,k)),(Recip(:,i)-Recip(:,k))))
     g=MIN(g,DOT_PRODUCT((Recip(:,i)+Recip(:,k)),(Recip(:,i)+Recip(:,k))))
     m(i)=1.5*gcut/SQRT(g)
  enddo

  
  Used = 0

  !** First do the X=0, plane - Need to check for dups ****
  Allocate(Test)
  i = 0
  Do j=-m(2), m(2)
     Do k=-m(3), m(3)
        gv = MATMUL(Recip, (/i,j,k/))
        g = SQRT(DOT_PRODUCT(gv,gv))

        if (g < Gcut) then
           Test%Gmag = g
           Test%Npnt = - (/i,j,k/)

           Call SearchBTree(Root, Test, Match, Ok)
           If (.NOT. Ok) then
              Used = Used + 1
              Allocate(Item)
              Item%Gmag = Test%Gmag
              Item%Npnt = -Test%Npnt
              Call InsertNode(Root, Item, Ok)
           End If
        End If
     End Do
  End Do


  !*** Now add the rest. Don't have to check for dups ****
  Do i=1, m(1)
    Do j=-m(2), m(2)
       Do k=-m(3), m(3)
          gv = MATMUL(Recip, (/i,j,k/))
          g = SQRT(DOT_PRODUCT(gv,gv))

          if (g < Gcut) then
              Used = Used + 1

              Allocate(Item)
              Item%Gmag = g
              Item%Npnt = (/i,j,k/)
              Call InsertNode(Root, Item, Ok)
          End If
       End Do
    End Do
  End Do

  GSize = Used

  Allocate(Gpnt(4,GSize), Npnt(3,GSize),STAT=i)
  Write(Token,*) 'InitGpoints: Error allocating Recip array. Size=',GSize
  Call Check_Error(i, Token, Error_Unit, .TRUE., WC, 'GenGpoints:')
  
  Gpnt=0
  Npnt=0

  !*** Now store the list in the 1D array ***
  G_Recip = Recip
  Count = 0
  Call InOrderTraversal(Root, ConvertGpoints)
  Call FreeTree(Root, .TRUE.)
  
  Return
End Subroutine



!******************************************************************************
!
! GenGPoints - Generates the list of G-points 
!
!   Recip  - Reciprocal Lattice Vectors
!   Gcut   - Curvilinear Plane wave cutoff
!   GSize  - Number of curvilinear Plane Waves generated
!   Gpnt   - List of Gpnts
!
!   NOTE: If Gsize <= 0 then Gpnts is not used but GSize contains the number
!         of planewaves needed. 
!***OLD***
!******************************************************************************

Subroutine GenGpoints_OLD(Recip, Gcut, Gsize,WC)
  Real,    Intent(IN)    :: Recip(:,:)
  Real,    Intent(IN)    :: Gcut
  Integer, Intent(OUT) :: Gsize
   Type (Word_Context), Intent(INOUT) :: WC

  Integer :: i,j,k, NMax, t, Used ,m(3)
  Real    :: g, g2, gv(3)
  Character*100 :: Token
  Integer, Pointer :: LUT(:,:)
  Logical          :: Ok, Unique

  Unique = .TRUE.

           !*** Estimate the Max indices for Gcut ****
  do i=1,3
     g=DOT_PRODUCT(Recip(:,i),Recip(:,i))
     j=i+1-((i)/3)*3
     k=i+2-((i+1)/3)*3
     g=MIN(g,DOT_PRODUCT((Recip(:,i)-Recip(:,j)),(Recip(:,i)-Recip(:,j))))
     g=MIN(g,DOT_PRODUCT((Recip(:,i)+Recip(:,j)),(Recip(:,i)+Recip(:,j))))
     g=MIN(g,DOT_PRODUCT((Recip(:,i)-Recip(:,k)),(Recip(:,i)-Recip(:,k))))
     g=MIN(g,DOT_PRODUCT((Recip(:,i)+Recip(:,k)),(Recip(:,i)+Recip(:,k))))
     m(i)=1.5*gcut/SQRT(g)
  enddo
  
  t = (m(1)+1)*(m(2)+1)*(m(3)+1)
  Allocate(LUT(3,t))

  Used = 1  
  LUT(:,Used) = 0  !*** Store DC G-point


  Do i=-m(1), m(1)
    Do j=-m(2), m(2)
       Do k=-m(3), m(3)
          gv = MATMUL(Recip, (/i,j,k/))
          g = SQRT(DOT_PRODUCT(gv,gv))

          if ((g /= 0) .AND. (g < Gcut)) then
             Ok = .TRUE.
             If (Unique) then
               t = 1
               Do While ((t<=Used) .AND. Ok)
                 If ((i==-LUT(1,t)).AND.(j==-LUT(2,t)).AND.(k==-LUT(3,t))) then
                      Ok = .FALSE.
                 End If
                 t = t + 1
               End Do
             End If

             If (Ok) then
                Used = Used + 1
                LUT(:,Used) = (/i,j,k/)
             End IF
          End If
       End Do
    End Do
  End Do

  GSize = Used

  Allocate(Gpnt(4,GSize), Npnt(3,GSize),STAT=i)
  Write(Token,*) 'InitGpoints: Error allocating Recip array. Size=',GSize
  Call Check_Error(i, Token, Error_Unit, .TRUE., WC, 'GenGpoints:')
  
  Gpnt=0
  Npnt=0
  
  do i=1,GSize
     gv = MATMUL(Recip, LUT(:,i))
     g = SQRT(DOT_PRODUCT(gv,gv))
     Gpnt(1:3,i)=gv
     Gpnt(4,i)=g
     Npnt(:,i)=LUT(:,i)
  enddo
  
  DeAllocate(LUT)

  Return
End Subroutine

!******************************************************************************
!
!  FindGcut - Finds the Gcut cutoff index
!
!  Gcut  - Plane wave cutoff
!  Gpnt  - List of Gpoints. Gpnt(1:3,i)=INTEGER indices, Gpnt(4,i)=Actual G mag
!  GSize - Size of Kpoints list
!
!******************************************************************************

Integer Function FindGcut(Gcut, Gpnt, GSize)
  Real,    Intent(IN)    :: Gcut
  Real,    Intent(INOUT) :: Gpnt(:,:)
  Integer, Intent(IN)    :: GSize

  Integer :: i
  
  i=1
  Do While ((i<GSize) .AND. (Gpnt(4,i)<Gcut))
     i = i + 1
  End Do

  If (Gpnt(4,i) < Gcut) i = i + 1    !** Only used if i==Gsize
  
  FindGcut = i - 1

  Return
End Function


!******************************************************************************
!
! InitGPoints - Generates the G-points and initializes the buffers and
!               stores the result on disk.
!
!  Grid    - Approximate Grid dimensions. Actual value is returned
!  Recip   - Reciprocal Lattice Vectors
!  Qcut    - Density Plane wave cutoff -- (HIGH)
!  Gcut    - Psi Plane wave cutoff -- (LOW)
!  Pcut    - Proj Plane wave cutoff __ (PROJ)
!
!******************************************************************************

Subroutine InitGpoints(Grid, Recip, Qcut, GCut, Pcut,  WC) 
   Integer, Intent(INOUT) :: Grid(:)
   Real,    Intent(IN) :: Recip(:,:)
   Real,    Intent(IN) :: Gcut
   Real,    Intent(IN) :: Qcut
   Real,    Intent(IN) :: Pcut
   Type (Word_Context), Intent(INOUT) :: WC

   Integer :: i, j, last, ioerr, HowMany, n, k
   Character*100 :: token
   Real          :: Gr(3), A(3,3), err, Gcutmod,Pcutmod,extra
   Integer       :: Gn(3)
   Integer, Pointer :: FFTMap(:)


   Call GenGPoints(Recip, Qcut, Gpnt_Size(G_HIGH),WC)

!**   Call SortGpnts(Gpnt_Size(G_HIGH)) !** Sort in ascending mag
!  add extra amount to Psi cutoff so that |k + G|< Gcut can be
!    satisfied for all reasonable k    -- Proj cutoff is not
!    separate in this version

   extra=sqrt(Dot_Product(xtal%Recip(:,1),xtal%Recip(:,1))) + &
          sqrt(Dot_Product(xtal%Recip(:,2),xtal%Recip(:,2))) + &
          sqrt(Dot_Product(xtal%Recip(:,3),xtal%Recip(:,3)))
   Gcutmod=sqrt(Gcut**2+1.5*extra*Gcut)

   write(Log_unit,*) 'InitGpoints: Gcut, Gcutmod = ', Gcut,Gcutmod
   Pcutmod=Gcutmod

   Gpnt_Size(G_LOW) = FindGcut(Gcutmod,Gpnt,Gpnt_Size(G_HIGH)) !*** Psi PW cutoff
   Gpnt_Size(G_PROJ) = FindGcut(Pcutmod,Gpnt,Gpnt_Size(G_HIGH)) !** Proj PW cutoff

   Write(Log_Unit,*) 'InitGpoints: Gpnt_Size:',Gpnt_Size

   !*** Fill the G_NewMag array ****
   Allocate(G_NewMag(Gpnt_Size(G_HIGH)),STAT=i)
   Write(Token,*) 'InitGpoints: Error allocating G_NewMat  . Size=',&
        Gpnt_Size(G_HIGH)
   Call Check_Error(i, Token, Error_Unit, .TRUE., WC, 'InitKpoints:')

   G_NewMag(1) = .TRUE.
   j = 1
   Do i=2, Gpnt_Size(G_High) 
      err = Gpnt(4,i-1) - Gpnt(4,i)
      err = SQRT(err*err)

      If (err > 1E-9) then
         G_NewMag(i) = .TRUE.
         j = j + 1
         If (i<=Gpnt_Size(G_Low)) NewMag_Size(G_Low) = j
      else
         G_NewMag(i) = .FALSE.
      End if
   End Do
   NewMag_Size(G_High) = j

   Do n=1, 3
      Do i=1, 3      !** Determine min LOW/High grid dimensions
         j = 2*MaxVal(ABS(Npnt(i,1:Gpnt_Size(n)))) + 1
         FFT_Grid(i,n) = NearestIntWithFactors(j,(/2,3,5/), 3, +1)
      End Do

      FFT_Grid(4,n) = PRODUCT(FFT_Grid(1:3,n))
   End Do

   Write(Log_Unit,*) 'InitGpoints:  FFT_Grid(LOW):',FFT_Grid(:,G_LOW)
   Write(Log_Unit,*) 'InitGpoints: FFT_Grid(HIGH):',FFT_Grid(:,G_HIGH)
   Write(Log_Unit,*) 'InitGpoints: FFT_Grid(PROJ):',FFT_Grid(:,G_PROJ)

   Gall_Size = 2*Gpnt_Size - 1
   Allocate(FFTmap_Low(2*Gpnt_Size(G_LOW)-1),&
            FFTmap_High(2*Gpnt_Size(G_HIGH)-1), &
            FFTmap_PROJ(2*Gpnt_Size(G_PROJ)-1),stat=ioerr)
   Write(token,*)'InitGpoints: Error allocating FFTmap, Size =',&
      Gpnt_Size(G_HIGH)
   Call Check_Error(ioerr,token,Error_Unit, .TRUE.,WC,"InitKPOINTS:")

!   !*** Check to make sure grid is large enough ***
!   j=MaxVal(ABS(NPnt(1:3,1:Gpnt_Size(G_HIGH))))
!   k = MaxVal(FFT_Grid(1:3,G_HIGH))/2 + 1
!   if (j>k) then
!      Write(Error_Unit,*) 'InitKpoints: Npnt to large when adding Gn!'
!      Write(Error_Unit,*) 'InitKPoints: Max K=',k
!      STOP
!   End If

    !do i=1,30
    !write(Log_Unit,'(i4,3i3,1pe15.7)') i,Npnt(1:3,i),Gpnt(4,i)
    ! enddo

     
   !*** Lastly calculate the FFT mappings ***
   Do i=1, 3
      If (i==G_LOW) then
         FFTMap => FFTmap_Low
      else if (i==G_HIGH) then
         FFTMap => FFTmap_High
      else if (i==G_PROJ) then
         FFTMap => FFTmap_PROJ
      End If

      j = 1    !** G=0 vector
      Gn = Npnt(:,j) + 1  
      Where (Gn <= 0) Gn = Gn + FFT_Grid(1:3,i)
      FFTMap(j) = CalcLinearIndex(FFT_Grid(1:3,i), Gn) + 1

      Do j=2, Gpnt_Size(i)
         Gn = Npnt(:,j) + 1  !** +G vector
         Where (Gn <= 0) Gn = Gn + FFT_Grid(1:3,i)
         FFTMap(j) = CalcLinearIndex(FFT_Grid(1:3,i), Gn) + 1
         If (FFTMap(j)>FFT_Grid(4,i)) then
            Write(Log_Unit,*) 'InitGPoints:  FFTMap+Error!! j=',j, &
                 ' * NPnt=',Npnt(:,j), ' * Gn=',Gn, ' * FFTMap(j)=',FFTMAp(j)
            STOP
         End If

         Gn = -Npnt(:,j) + 1   !** -G vector
         Where (Gn <= 0) Gn = Gn + FFT_Grid(1:3,i)
         FFTMap(j+Gpnt_Size(i)-1) = CalcLinearIndex(FFT_Grid(1:3,i), Gn) + 1
         If (FFTMap(j)>FFT_Grid(4,i)) then
            Write(Log_Unit,*) 'InitGPoints:  FFTMap-Error!! j=',j, &
                 ' * NPnt=',Npnt(:,j), ' * Gn=',Gn, ' * FFTMap(j)=',FFTMAp(j)
            STOP
         End If
      End Do
   End Do

   Grid = FFT_Grid(1:3,G_LOW)

   !** Now create the FFT plans ***
   Call InitFFTW(3)
   Call FFT_CreatePlan(G_LOW, FFT_Grid(:,G_Low))
   Call FFT_CreatePlan(G_HIGH, FFT_Grid(:,G_High))
   Call FFT_CreatePlan(G_PROJ, FFT_Grid(:,G_PROJ))

   Return   
End Subroutine


!******************************************************************************
!
!  FilterValue - Filter's a complex array by zeroing the small values
!
!******************************************************************************

Subroutine FilterValue(Grid, MinReal)
   Complex, Intent(INOUT) :: Grid(:)
   Real, OPTIONAL, Intent(IN) :: MinReal

   Real :: TooSmall

   TooSmall = Filter_Min
   If (Present(MinReal)) TooSmall = MinReal

   Where (ABS(Real(Grid)) < TooSmall) Grid = CMPLX(0.0, AImag(Grid))
   Where (ABS(AImag(Grid)) < TooSmall) Grid = CMPLX(Real(Grid), 0.0)

   Return
End Subroutine

!******************************************************************************
!
!  CalcLinearIndex - Calculates the Linear Index of 3-d array
!
!  GridDim - Grid Dimensions
!  PM    - 3-d Mesh point index
!
!     ***** NOTE: Should only be used on unifrom grids!!!!! ******
!
!******************************************************************************

Function CalcLinearIndex(GridDim, PM)
  Integer             :: CalcLinearIndex
  Integer, Intent(IN) :: gridDim(:)
  Integer, Intent(IN) :: PM(:)

  CalcLinearIndex = GridDim(1)*GridDim(2)*(PM(3)-1) + &
                     GridDim(1)*(PM(2)-1) + PM(1)-1

  Return
End Function

!*************************************************************************
!
!    V_G_to_r
!       Takes V(G) (input) to V(r) (output) where
!          G defined on Gpnt_Size(INPUTGRID)
!          r defined on FFTmap(OUTPUTGRID)
!
!    Vr initialized to 0
!
!*************************************************************************
Subroutine V_G_to_r(VG,INPUTGRID,Vr,OUTPUTGRID)
 Complex, intent(IN) :: VG(:)
 Complex, intent(OUT) :: Vr(:)
 Integer, intent(IN) :: INPUTGRID,OUTPUTGRID

 Integer  :: i,j,k,l
 Integer, pointer :: FFTMap(:)


 If (OUTPUTGRID == G_LOW) then
   FFTMap => FFTMap_LOW
 Else if (OUTPUTGRID == G_PROJ) then
   FFTMap => FFTMap_PROJ
 Else if (OUTPUTGRID == G_HIGH) then
   FFTMap => FFTMap_HIGH
 Else
   Write(Error_Unit,*) " Error in V_G_to_r", INPUTGRID,OUTPUTGRID
   stop
 EndIf

 If (Gall_Size(OUTPUTGRID)<Gall_Size(INPUTGRID)) then
  write(Error_Unit,*) " Error in V_G_to_r", INPUTGRID,OUTPUTGRID
 Endif

 Vr=0
 Vr(FFTMap(1:Gpnt_Size(INPUTGRID)))=VG(1:Gpnt_Size(INPUTGRID))
 i=Gpnt_Size(INPUTGRID)+1;j=Gpnt_Size(OUTPUTGRID)+1;k=j+Gpnt_Size(INPUTGRID)-2
 Vr(FFTMap(j:k))=VG(i:Gall_Size(INPUTGRID))
 Call PerformFFT(FFT_TO_R, OUTPUTGRID, Vr)

Return

 End Subroutine

!*************************************************************************
!
!    V_r_to_G
!       Takes V(r) (input) to V(G) (output) where
!          r defined on FFTmap(INPUTGRID)
!          G defined on Gpnt_Size(OUTPUTGRID)
!
!    VG accumulated        
!
!*************************************************************************
Subroutine V_r_to_G(Vr,INPUTGRID,VG,OUTPUTGRID)
 Complex, intent(INOUT) :: Vr(:)
 Complex, intent(INOUT) :: VG(:)
 Integer, intent(IN) :: INPUTGRID,OUTPUTGRID

 Integer  :: i,j,k,l
 Integer, pointer :: FFTMap(:)



 If (INPUTGRID == G_LOW) then
   FFTMap => FFTMap_LOW
 Else if (INPUTGRID == G_PROJ) then
   FFTMap => FFTMap_PROJ
 Else if (INPUTGRID == G_HIGH) then
   FFTMap => FFTMap_HIGH
 Else
   Write(Error_Unit,*) " Error in V_r_to_G", INPUTGRID,OUTPUTGRID
   stop
 EndIf

 If (Gall_Size(OUTPUTGRID)>Gall_Size(INPUTGRID)) then
  write(Error_Unit,*) " Error in V_r_to_G", INPUTGRID,OUTPUTGRID
 Endif

 Call PerformFFT(FFT_TO_G, INPUTGRID, Vr)
 i=Gpnt_Size(OUTPUTGRID)+1;j=Gpnt_Size(INPUTGRID)+1;k=j+Gpnt_Size(OUTPUTGRID)-2
 VG(1:Gpnt_Size(OUTPUTGRID))=VG(1:Gpnt_Size(OUTPUTGRID)) + &
  Vr(FFTMap(1:Gpnt_Size(OUTPUTGRID)))
 VG(i:Gall_Size(OUTPUTGRID))=VG(i:Gall_Size(OUTPUTGRID)) + &
  Vr(FFTMap(j:k))

 Return
 End Subroutine

!*****************************************************************************
!  
! Subroutine LdPlaneWave
!
!******************************************************************************

Subroutine LdPlaneWave(WC)

  Type (Word_Context), Intent(INOUT) :: WC   

  Integer :: i, tlen
  Character*80 :: Token


  PW_Gcut = 0
  Call GetNextWord(WC, Token, tlen)
  Call UpperCase(token)

  Do While (Trim(token) /= "END")
    if (Trim(Token) == "GCUT_LOW") then    !** Set Psi PW Cutoff   
       Call GetNumbers(WC, PW_Gcut(G_LOW))
       if (Print_Level == PRINT_VERBOSE) then
          write(LOG_Unit, *) "  Gcut_Low = ", PW_Gcut(G_LOW)
       End IF 
    else if (Trim(Token) == "GCUT_HIGH") then    !** Set density PW Cutoff
       Call GetNumbers(WC, PW_Gcut(G_HIGH))
       if (Print_Level == PRINT_VERBOSE) then
          write(LOG_Unit, *) "  Gcut_High = ", PW_Gcut(G_HIGH)
       End IF 
    else if (Trim(Token) == "GCUT_PROJ") then    !** Set projector PW Cutoff
       Call GetNumbers(WC, PW_Gcut(G_PROJ))
       if (Print_Level == PRINT_VERBOSE) then
          write(LOG_Unit, *) "  Gcut_PROJ = ", PW_Gcut(G_PROJ)
       End IF 
    else
       Write(Error_Unit, *) "ldPlaneWave: Unknown Command : ", Trim(Token)
    End If         

    Call GetNextWord(WC, Token, tlen)
    if (W_ERROR /= W_OK) Call EOF_Error(WC, "ldPlaneWave:(next word): ")
    Call UpperCase(token)
  End Do

  If (PW_Gcut(G_PROJ) == 0) PW_Gcut(G_PROJ) = PW_GCut(G_LOW)

  Return  
End Subroutine

 Subroutine InitkplusGfilter(NoKpnts)
   Integer, INTENT(IN) :: NoKpnts
   Integer :: i,j,k
   Real :: Kvec(3),gmag2,ecut,v(3)

   If(allocated(Gcutfilter)) then
       k=size(Gcutfilter)
         do i=1,k
            Deallocate(Gcutfilter(i)%zeromap)
         enddo
         Deallocate(Gcutfilter)
   endif 

   write(Log_unit,*) 'InitkplusGfilter (re)starting with kpts = ', NoKpnts
   allocate(Gcutfilter(NoKpnts),stat=i)
      if (i /= 0) then
         write(error_unit,*) 'Error in InitkplusGfilter ', NoKpnts,i
         stop
      endif

   ecut=PW_Gcut(G_LOW)**2

   do k=1,NoKpnts
      j=0
      Kvec=BZ%Ku(:,k)
      do i=1,Gall_Size(G_LOW)
         if (i <= Gpnt_Size(G_LOW)) then
            v=Kvec+Gpnt(1:3,i)
            gmag2=v(1)*v(1)+v(2)*v(2)+v(3)*v(3)
            if (gmag2>ecut) j=j+1
         else
            v=Kvec-Gpnt(1:3,i-Gpnt_Size(G_LOW)+1)
            gmag2=v(1)*v(1)+v(2)*v(2)+v(3)*v(3)
            if (gmag2>ecut) j=j+1
         endif
      enddo

      write(Log_unit,*) 'In InitkplusGfilter -- kpnt, nzeros = ', k,j  
      Gcutfilter(k)%nzeros=j
      if (j>0) then
        allocate(Gcutfilter(k)%zeromap(j), stat=i)
         if (i /= 0) then
            write(error_unit,*) 'Error in InitkplusGfilter ', k,j
            stop
         endif
      else
        write(6,*) 'WARNING!!!!! no padding for kvec =', Kvec
      endif

      j=0
      do i=1,Gall_Size(G_LOW)
         if (i <= Gpnt_Size(G_LOW)) then
            v=Kvec+Gpnt(1:3,i)
            gmag2=v(1)*v(1)+v(2)*v(2)+v(3)*v(3)
            if (gmag2>ecut) then
               j=j+1
               Gcutfilter(k)%zeromap(j)=i
            endif
         else
            v=Kvec-Gpnt(1:3,i-Gpnt_Size(G_LOW)+1)
            gmag2=v(1)*v(1)+v(2)*v(2)+v(3)*v(3)
            if (gmag2>ecut) then
               j=j+1
               Gcutfilter(k)%zeromap(j)=i
            endif
         endif
      enddo
   enddo

 End Subroutine InitkplusGfilter

 Subroutine kplusGfilter(WF,kindex)
    COMPLEX, INTENT(INOUT) :: WF(:)
    INTEGER, INTENT(IN) :: kindex

    INTEGER :: many,i

    many=Gcutfilter(kindex)%nzeros
    if (many>0) then
       do i=1,many
          WF(Gcutfilter(kindex)%zeromap(i))=0
       enddo
    endif

 End subroutine kplusGfilter


End Module




