!******************************************************************************
!
! File : hamfunc.f90
!   originally from hamiltonian.f90
!   by : Alan Tackett
!   on : 02/07/97
!  for : PAW
!
! Module for calculating the Hamiltonian and Cohesive Energy of the PAW
! system.  Also contains routines to do H*Psi and O*Psi.
!
!  Contains relatively simple functions in construction of Hamiltonian
!      terms 
!
! Updated for spin by Ping Tang and N. A. W. Holzwarth
!      Last change 5/25/05
!******************************************************************************

Module hamfunc

Use atom_data
Use crystal_symmetry
Use gausslib
Use gpoints 
Use mathlib
Use options_data
Use paw_inout
Use spherical_harmonic
Use orbital_pack

Implicit NONE!!!!

!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
!  RestrictPot - Restricts the potential from the High planewave grid to the
!       low cutoff grid in FFT space.
!
!  Vin    - High PW grid in FFT space(Input)
!  Vout   - Low PW grid in FFT space(Returned) 
!
!******************************************************************************

Subroutine RestrictPot( Vin, Vout)
  Complex, Intent(IN) :: Vin(:)
  Complex, Intent(OUT) :: Vout(:)

  Integer :: i, N_low, N_high
  real    :: weight, Ec

  Ec = PW_Gcut(G_PROJ)*(1-V_Smooth_width)

  N_low = Gpnt_Size(G_PROJ) - 1
  N_High = Gpnt_Size(G_HIGH) - 1

  Vout = 0

Write(Log_Unit,*) 'RestrictPot: PW_Gcut=',PW_Gcut, ' * V_Smooth_Width=',V_Smooth_Width
  Vout(FFTmap_PROJ(1)) = Vin(FFTmap_High(1))
  Do i=2, Gpnt_size(G_PROJ)
!     If (V_Smooth_Width > 0.0) then
!        weight = abs(Gpnt(4,i)-PW_Gcut(G_PROJ))/(PW_Gcut(G_Proj)*V_Smooth_Width)
!        weight = 2.0/(exp(-weight) + 1.0) - 1
!     else
!        Weight = 1
!     End if

     Weight = 1
     If (Gpnt(4,i) > Ec) then
        Weight = Pi*(PW_Gcut(G_PROJ) - Gpnt(4,i)) / &
             (2*V_Smooth_Width*PW_Gcut(G_PROJ))
        Weight = sin(weight)**2
     End if
!Write(Log_Unit,*) 'RestrictPot: i=',i, ' * |G|=',Gpnt(4,i), ' * Wt=',weight 
!weight = 1
     Vout(FFTmap_HIGH(i)) = Vin(FFTmap_High(i)) * Weight
     Vout(FFTmap_HIGH(i+N_HIGH)) = Vin(FFTmap_High(i+N_High)) * Weight
  End Do

  Return
End Subroutine


!******************************************************************************
!                  
!  CalcVlocal - Calculates Vlocal                        
!                  
!******************************************************************************

Subroutine CalcVlocal(V)
  Complex,         Intent(INOUT) :: V(:)

  Real    :: theta, Gmag, RadG, G(3), Esum
  Complex :: Phase, Vsum
  Integer :: atom, atype, lut, gindex, RadIndex, fi
  Type (Atom_Info_Fixed), Pointer :: AT

  V = 0

  RadIndex = 0
  Do gindex=1, Gpnt_Size(G_HIGH)
     G =  Gpnt(1:3,gindex)
     Gmag = Gpnt(4,gindex)

     !*** If needed inc the RadIndex ***
     If (G_NewMag(gindex)) RadIndex = RadIndex + 1

     Vsum = 0
     Do atype=1, Atom_Types
        AT => AtomType_Info(atype)
        RadG = AT%Rad_VLocal(RadIndex)

        !*** Create the structure factors and accumulate vlocal***
        Do lut = AtomType_Range(1,atype), AtomType_range(2,atype)
           atom = Atomtype_map(lut)
        
           theta = DOT_PRODUCT(G, Atom_List(atom)%Pos)
           Phase = CMPLX(cos(theta), -sin(theta))
!Phase = CONJG(Phase)
           Vsum = Vsum + Phase*RadG
        End Do
     End do

     V(gindex) = V(gindex) + Vsum
  End Do

  Return
End Subroutine

!******************************************************************************
!
!  function spheresymFT
!
!******************************************************************************

Function spheresymFT(ngrid,radialfn,Gmag,dr,Grid)
  Real :: spheresymFT
  Integer, INTENT(IN) :: ngrid
  Real, INTENT(IN) :: radialfn(:),Gmag,dr
  Type(GridInfo) :: Grid
 
  Real, allocatable :: fn(:)
  Integer       :: i
  Real          :: r, con,x,R0

   con = Four_Pi

   Allocate(fn(ngrid))
   if (Grid%islog) R0=Grid%drdu(1)

   do i=1,ngrid
      if  (Grid%islog) then
         r= Grid%drdu(i) - Grid%drdu(1)
      else
         r=dr*(i-1)
      endif
      x=r
      if (Gmag > machine_precision) x=SIN(Gmag*r)/Gmag
      fn(i) = con*r*radialfn(i)*x
   End do
          
  spheresymFT=IntSimpson(ngrid,dr,fn,Grid)

  deallocate(fn)
End Function spheresymFT

Subroutine CalcCoreTail(V)
  Complex, INTENT(OUT) :: V(:)
  Real    :: theta, Gmag, RadG, G(3), Esum,dr
  Complex :: Phase, Vsum
  Integer :: atom, atype, lut, gindex, RadIndex, fi,ngrid,i,j
  Type (Atom_Info_Fixed), Pointer :: AT
 
  Write(Log_unit,*) 'In CalcCoreTail',size(V)
  Do atype=1, Atom_Types
     AT => AtomType_Info(atype)
     Allocate(AT%FTCoreTail(NewMag_Size(G_High)), STAT=j)
        if (j /= 0) then
           write(Error_unit,*) 'Error CalcCoreTail: allocation ', &
                    NewMag_Size(G_High),atype,j
           stop
        endif
     AT%FTCoreTail=0
     dr= AT%Mesh_Step
     ngrid=AT%CoreTail_Points
     Radindex=0
     Do gindex=1, Gpnt_Size(G_HIGH)
         If (G_NewMag(gindex)) then
            Gmag = Gpnt(4,gindex)
            RadG = spheresymFT(ngrid,AT%CoreTail_Density,Gmag,dr,AT%RadGrid)
            RadIndex=RadIndex+1
            AT%FTCoreTail(RadIndex)=RadG
         Endif
     Enddo
  Enddo
    
  V=0; Radindex=0
  Do gindex=1, Gpnt_Size(G_HIGH)
     G =  Gpnt(1:3,gindex)
     If (G_NewMag(gindex)) Radindex=Radindex+1
 
     Vsum = 0
     Do atype=1, Atom_Types
        AT => AtomType_Info(atype)
                                                                                
        !*** Create the structure factors and accumulate vlocal***
        Do lut = AtomType_Range(1,atype), AtomType_range(2,atype)
           atom = Atomtype_map(lut)
                                                                                
           theta = DOT_PRODUCT(G, Atom_List(atom)%Pos)
           Phase = CMPLX(cos(theta), -sin(theta))
           Vsum = Vsum + Phase*AT%FTCoreTail(RadIndex)
        End Do
     End do
      
     V(gindex)=Vsum
  End Do
End Subroutine

                                                                                
Subroutine UpdateCoreTail(V)
  Complex, INTENT(OUT) :: V(:)
  Real    :: theta, Gmag, RadG, G(3), Esum,dr
  Complex :: Phase, Vsum
  Integer :: atom, atype, lut, gindex, RadIndex, fi,ngrid,i,j
  Type (Atom_Info_Fixed), Pointer :: AT
                                                                                
  Write(Log_unit,*) 'In UpdateCoreTail',size(V)
                                                                                
  V=0; Radindex=0
  Do gindex=1, Gpnt_Size(G_HIGH)
     G =  Gpnt(1:3,gindex)
     If (G_NewMag(gindex)) Radindex=Radindex+1
                                                                                
     Vsum = 0
     Do atype=1, Atom_Types
        AT => AtomType_Info(atype)
                                                                                
        !*** Create the structure factors and accumulate coretail***
        Do lut = AtomType_Range(1,atype), AtomType_range(2,atype)
           atom = Atomtype_map(lut)
                                                                                
           theta = DOT_PRODUCT(G, Atom_List(atom)%Pos)
           Phase = CMPLX(cos(theta), -sin(theta))
           Vsum = Vsum + Phase*AT%FTCoreTail(RadIndex)
        End Do
     End do
                                                                                
     V(gindex)=Vsum
  End Do
 End subroutine
                                                                                

!******************************************************************************
!
!  CalcRadial_Vlocal - Calculates the radial Vlocal grids
!
!******************************************************************************

Subroutine CalcRadial_Vlocal
 
  Integer       :: i,j, k, G, Gmag, n, L,ngrid
  Character*200 :: Token
  Real, Pointer :: RadV(:),fn(:)
  Real          :: dr, r,  norm,xmax,gmax,con
  Type (Atom_Info_Fixed), Pointer :: AT

   con = Four_Pi/Xtal%volume
   L=0  ! L=0 for Vlocal

   ngrid=MAXVAL(AtomType_Info(:)%Mesh_Size)
   Allocate(fn(ngrid))

  Do i=1, Atom_Types
    AT => AtomType_Info(i)

    Allocate(AT%Rad_Vlocal(NewMag_Size(G_High)), STAT=j)
    Write(Token,*) 'CalcRadial_Vlocal: Error allocating array. Size=',&
         NewMag_Size(G_HIGH), ' * TypeIndex=',i
    Call Check_Error(j, Token, Error_Unit, .TRUE., PAW_WC,'CalcRadial_Vlocal:')

    RadV => AT%Rad_Vlocal

    RadV = 0

    If (Atomic_Mode == SIM_SEPM) then 
       k = 0
       con = 1.0 /xtal%volume
       Do j=1, Gpnt_Size(G_HIGH)
          If (G_NewMag(j)) then
             k = k + 1
             RadV(k) = EvalGauss(AT%V_sepm, Gpnt(4,j)) * con
          End If
       End Do
    else
      ! norm = con*AT%V_Local
       norm = con
       
       if (ABS(norm).gt.1.d-9) then
          
          DR=AT%Mesh_Step      
          write(6,*) 'CalcRadial_Vlocal: DR',DR
          
          !*** Tabulate the vlocal function ***
          Do j=1, AT%Mesh_Size
             if (AT%RadGrid%islog) then
                r = AT%RadGrid%drdu(j) - AT%RadGrid%drdu(1)
             else
                r = DR*(j-1)
             endif
             
             !Fn(j) = norm*r*AT%Shape_Func(j)
             Fn(j) = norm*r*AT%RadR_Vlocal(j)
          End do
          
          k = 0
          Do j=1, Gpnt_Size(G_HIGH)
             If (G_NewMag(j)) then
                k = k + 1
                !RadV(k) = RadialFourier(L, Gpnt(4,j), DR, AT%Mesh_Size, fn)
                RadV(k) = spheresymFT(AT%Mesh_Size,&
                    AT%RadR_Vlocal,Gpnt(4,j),dr,AT%RadGrid)/xtal%volume
             End If
          End Do
       Endif
    End If
  End Do

  Deallocate(fn)

  Return
End Subroutine


!******************************************************************************
!
!  CalcRadialHat - Calculates the radial n^ grids
!
!******************************************************************************

Subroutine CalcRadialHat

  Integer       :: i,j, k, G, Gmag, n, L, ngrid
  Character*200 :: Token
  Real, Pointer :: RadHat(:),fn(:)
  Real          ::  dr, r,  norm,xmax,gmax,R0
  Type (Atom_Info_Fixed), Pointer :: AT


   ngrid=MAXVAL(AtomType_Info(:)%Mesh_Size)
   Allocate(fn(ngrid))


  Do i=1, Atom_Types
    AT => AtomType_Info(i)

    AT%Hat_MaxL = 2*MaxVal(AT%L_Value)
  
    Allocate(AT%RadHat(NewMag_Size(G_High), AT%Hat_MaxL+1), STAT=j)
    Write(Token,*) 'CalcRadialHat: Error allocating array. Size=',&
         NewMag_Size(G_HIGH), ' * ', AT%Basis_Size, ' * TypeIndex=',i
    Call Check_Error(j, Token, Error_Unit, .TRUE., PAW_WC, 'CalcRadialHat:')
    
    DR=AT%Mesh_Step    
    if (AT%RadGrid%islog) R0=AT%RadGrid%drdu(1)
     !write(6,*) 'CalcRadHat: DR',DR

    Do L=0, AT%Hat_MaxL
       RadHat => AT%RadHat(:,L+1)

       n=AT%Mesh_Size
       do j=1,n            
         If (AT%RadGrid%islog) then
              r = AT%RadGrid%drdu(j) - AT%RadGrid%drdu(1)
         else
              r = DR*(j-1)
         endif
         fn(j)=(r**(2*L+2))*AT%Shape_Func(j)
       enddo

       norm = sqrt(Four_Pi)/IntSimpson(n, DR, fn,AT%RadGrid)

       !*** Tabulate the hat function ***
       Do j=1, n
         If (AT%RadGrid%islog) then
              r = AT%RadGrid%drdu(j) - AT%RadGrid%drdu(1)
         else
              r = DR*(j-1)
         endif
          Fn(j) = norm * AT%Shape_Func(j) * r**(L+1)
       End do

       k = 0
       Do j=1, Gpnt_Size(G_HIGH)
          If (G_NewMag(j)) then
             k = k + 1
             RadHat(k) = RadialFourier(L, Gpnt(4,j), DR, n, fn,AT%RadGrid)
            !write(6,'("radhat",2i4,1p2e15.7)') k,L,Gpnt(4,j),radHat(k)
          End If
       End Do
    End Do

  End Do

  Deallocate(fn)

  Return
End Subroutine

!******************************************************************************
!
! CalcKineticEnergy - Calculates the KE for the given Psi nad K-pnt
!
!   BlochK - K-Point
!   Psi    - Wave function
!
!******************************************************************************
                                                                                                             
Real Function CalcKineticEnergy(BlochK, Psi)
  Real,    Intent(IN) :: BlochK(:)
  Complex, Intent(IN) :: Psi(:)
                                                                                                             
  Integer :: gi, N
  Real    :: G(3), G2
  Complex :: KE
                                                                                                             
                                                                                                             
  N = Gpnt_Size(G_LOW) - 1
                                                                                                             
  KE = 0
  Do gi=1, Gpnt_Size(G_LOW)
     G = BlochK + Gpnt(1:3,gi)
     G2 = DOT_PRODUCT(G,G)
     KE = KE + CONJG(Psi(gi))*Psi(gi) * G2
  End Do
                                                                                                             
  Do gi=2, Gpnt_Size(G_LOW)
     G = BlochK - Gpnt(1:3,gi)
     G2 = DOT_PRODUCT(G,G)
     KE = KE + CONJG(Psi(gi+N))*Psi(gi+N) * G2
  End Do
                                                                                                             
  CalcKineticEnergy = KE
                                                                                                             
  Return
End function


!!*****************************************************
!! LoadHatij
!!*****************************************************
!Subroutine LoadHatij
! Type (Atom_Info_Fixed), pointer :: AT
! Integer :: a,NG,i,j,k,L
!
! !write(Log_Unit,*) 'LoadHatij start:', SAVE_HAT_UNIT
! !call flush(Log_unit)
!! Rewind(SAVE_HAT_UNIT)
! NG=Gpnt_Size(G_HIGH)
! Do a=1,Atom_Types
!    AT => AtomType_Info(a)
!    i=AT%nlm_Size
!    allocate(AT%Hatij(i,i,NG),STAT=j)
!       if (j /= 0) then
!          write(Error_unit,*) 'LoadHatij allocation error', j,i,NG
!          stop
!       endif
!       !write(Log_Unit,*) 'LoadHatij loading:', a,i
!       !call flush(Log_unit)
!    do k=1,NG
!       read(SAVE_HAT_UNIT) AT%Hatij(:,:,k)
!    enddo
!       !write(Log_Unit,*) 'LoadHatij loaded:', a,i
!       !write(Log_Unit,*) 'LoadHatij G=0 values:', AT%Hatij(:,:,1)
!       !call flush(Log_unit)
! Enddo
!
!End subroutine

!!*****************************************************
!! DumpHatij
!!*****************************************************
!Subroutine DumpHatij
! Type (Atom_Info_Fixed), pointer :: AT
!
! Integer :: a
!
! Do a=1,Atom_Types
!    AT => AtomType_Info(a)
!    deallocate(AT%Hatij)
! Enddo
!
!End subroutine

!!*****************************************************
!! MakeHatij  -- construct coefficients and write to disk
!!*****************************************************
!Subroutine MakeHatij
! Type (Atom_Info_Fixed), pointer :: AT
! Integer :: a,NG,i,j,k,HatIndex,gindex,Qlm_Size
! Integer        :: nili, njlj, mi, mj, M, L, Li, Lj, ibase, jbase
! Complex, parameter :: ai=(0,-1)
! Complex              :: ylm(13,7),term
! Real :: G(3)
! Integer, Pointer :: LUT_Orb(:), nl_Base(:), nlm_LUT(:,:)
! Real, Pointer  :: aQlm(:)
!
! !write(Log_Unit,*) 'MakeHatij start:', SAVE_HAT_UNIT
! !call flush(Log_unit)
! Open(SAVE_HAT_UNIT,Status='Scratch', form='unformatted')
! NG=Gpnt_Size(G_HIGH)
! Do a=1,Atom_Types
!    AT => AtomType_Info(a)
!    i=AT%nlm_Size
!    allocate(AT%Hatij(i,i,NG),STAT=j)
!       if (j /= 0) then
!          write(Error_unit,*) 'MakeHatij allocation error', j,i,NG
!          stop
!       endif
!    AT%Hatij=0
!
!    Qlm_Size=AT%QVlm_Size
!    !write(Log_Unit,*) 'MakeHatij loop:', a,i,Qlm_Size,AT%Hat_MaxL
!    !call flush(Log_unit)
!    nl_base => AT%nl_Base
!    nlm_LUT => AT%nlm_LUT
!    aQlm => AT%aQlm
!    LUT_Orb => AT%LUT_Orb
!
!    HatIndex=0
!    Do gindex=1, Gpnt_Size(G_HIGH)
!       If (G_NewMag(gindex)) HatIndex=HatIndex+1
!       G=Gpnt(1:3,gindex)
!       Do L=0,AT%Hat_MaxL
!          Ylm(:,L+1)=Spharm(G(1), G(2), G(3), L, .true.)
!          Ylm(:,L+1)=Ylm(:,L+1)*(ai**L)*AT%RadHat(HatIndex,L+1)
!       Enddo
!
!     Do j=1, Qlm_Size
!        Call Orbital_Decode(LUT_Orb(j), nili, njlj, mi, mj, L)
!        M = mj - mi
!        ibase = nl_base(nili);  ibase = ibase + nlm_LUT(2, ibase)
!        jbase = nl_base(njlj);  jbase = jbase + nlm_LUT(2, jbase)
!
!        AT%Hatij(ibase+mi,jbase+mj,gindex) = &
!              AT%Hatij(ibase+mi,jbase+mj,gindex) +&
!              aQlm(j)*Ylm(L+M+1,L+1)    
!     Enddo
!     write(SAVE_HAT_UNIT) AT%Hatij(:,:,gindex)
!    Enddo
!    !write(Log_Unit,*) 'MakeHatij G=0 values:', AT%Hatij(:,:,1)
!    !call flush(Log_unit)
! Enddo
!
! Call DumpHatij
!
!End subroutine

!*******************************************************************
!   Count_atoms
!********************************************************************
 Subroutine Count_atoms
     Integer :: i,j,k,n,a

     do a=1,Atom_Types
        AtomType_Info(a)%Nofthistype=0
     enddo

    do a=1,Specific_Atoms
       i=Atom_List(a)%TypeIndex
       AtomType_Info(i)%Nofthistype=AtomType_Info(i)%Nofthistype+1
    enddo

    do a=1,Atom_Types
       write(Log_Unit,*) 'Count_Atoms: ', a, AtomType_Info(a)%Atom_Name, &
               AtomType_Info(a)%Nofthistype
    enddo

 end subroutine Count_atoms

End Module
