!******************************************************************************
!
! File : hamiltonian.f90
!   by : Alan Tackett
!   on : 02/07/97
!  for : PAW
!
! Module for calculating the Hamiltonian and Cohesive Energy of the PAW
! system.  Also contains routines to do H*Psi and O*Psi.
!
! Contains CalcHam and supporting routines
!
! Updated for spin by Ping Tang and N. A. W. Holzwarth
!      Last change 5/23/05
!******************************************************************************

Module hamiltonian

Use anderson_mixing
Use atom_data
Use basis_lib
Use coulomb_pack
Use crystal_data
Use denvhat_pack
Use doijmatrix
Use gausslib
Use gpoints 
Use hamfunc
Use hamsym
Use hamvxc
Use laplacian
Use mathlib
Use mem_data
Use memmgr
Use misc
Use options_data
Use orbital_pack
Use paw_inout
Use projectors
Use spherical_harmonic
Use psilib
Use search_sort
Use vhartree_pack
Use word
Use work_mgr

!**Use atom_charge

Implicit NONE!!!!

Integer, Pointer :: BandMap(:,:)

Real  :: Diag_Energy       !** K+V_ion energy
Real  :: hat_Energy        !** V^ + E^ energy
Real  :: Hartree_Energy   !** Hartree  energy
!Real  :: Auxform_of_Hartree   !** For testing self-energy
Real  :: OneExc_Energy     !** Exchange-correlation energy in AccumVxc
Real  :: TildeKE_Energy    !** KE energy for smooth hamiltonian
Real  :: TildePot_Energy   !** Pot energy for smooth hamiltonian
Real  :: TildeXC_Energy    !** XC energy for smooth hamiltonian
Real  :: Hat_Hat_Energy    !** Coulomb interaction for Hat-Hat
Real  :: Prev_CohesiveEnergy(2)
!Real  :: Self_Energy_Correction  !** estimate of self-energy

Complex, Pointer ::  Work_Dij(:), Work_Dijspin(:)  !** Work array for mixing Dij's

!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
!  CalcWij - Calculates the Wij coefficients
!
!  PDot - Projector times Psi products. The indices of PDot are as follows:
!
!       PDOT(atom, +/-M, ProjIndex, Band, Kpnt)
!          
!       Where 1=+M and 2=-M
!      
!
!******************************************************************************

Subroutine CalcWij(PDot)
  Complex, Intent(IN) :: PDot(:,:,:)
  complex :: sum1,sum2 ! test
  Integer :: Base, i, j, a ,N,NN
  Complex, Pointer :: Wij(:,:)

  sum1=0  !test
  Do a=1, Specific_Atoms
     Wij => Atom_List(a)%Wij

     Wij = 0
     !write(Log_unit,*)'CalcWij: a = ',a,PLM_AtomRange(1,a),PLM_AtomRange(2,a)
     !call flush(log_unit)

     Base = PLM_AtomRange(1,a) - 1
     Do i=PLM_AtomRange(1,a), PLM_AtomRange(2,a)
        Do j=PLM_AtomRange(1,a), PLM_AtomRange(2,a)
           Wij(i-Base, j-Base) = &
             SUM(CONJG(PDot(i,1:Numbands,:))*PDot(j,1:Numbands,:)&
                   *Occupancy(1:Numbands,:))
             !write(log_unit,*) 'CalcWij', i,j,Wij(i-Base,j-Base)
             !call flush(log_unit)
        End Do
     End Do

!  i=AtomType_Info(Atom_List(a)%TypeIndex)%nlm_Size   !!testing
!  sum2=SUM(Wij(1:i,1:i)*AtomType_Info(Atom_List(a)%TypeIndex)%Oij(1:i,1:i))
 
!  write(Log_Unit,*) 'test Wij for atom',a,sum2   !!testing
!  sum1=sum1+sum2   !!testing


  End Do

  write(log_unit,*) 'CalcWij:  finished s=1'
  call flush(log_unit)
if(spindependence ) then
  N=Numbands+1;NN=2*Numbands
  write(log_unit,*) 'CalcWij: N,NN ', N,NN
  Do a=1, Specific_Atoms
     Wij => Atom_List(a)%Wijspin

     Wij = 0
     write(Log_unit,*)'CalcWij: a = ',a,PLM_AtomRange(1,a),PLM_AtomRange(2,a)
     call flush(log_unit)

     Base = PLM_AtomRange(1,a) - 1
     Do i=PLM_AtomRange(1,a), PLM_AtomRange(2,a)
        Do j=PLM_AtomRange(1,a), PLM_AtomRange(2,a)
           Wij(i-Base, j-Base) = SUM(CONJG(PDot(i,N:NN,:))*PDot(j,N:NN,:)&
                    *Occupancy(N:NN,:))
             write(log_unit,*) 'CalcWij', i,j,Wij(i-Base,j-Base)
             call flush(log_unit)
        End Do
     End Do
  End Do
end if

!  write(Log_Unit,*) 'total wij*Oij test', sum1 !! testing
  
  Return
End Subroutine

!******************************************************************************
!
!  CalcWijkl - Calculates the Wijkl coefficients for self-energy corr.
!
!  PDot - Projector times Psi products. The indices of PDot are as follows:
!
!
!******************************************************************************

!Subroutine CalcWijkl(PDot)
!  Complex, Intent(IN) :: PDot(:,:,:)
!  Integer :: Base, i, j, k, l, s ,p,N,NN,Cijkl_Size
!  Integer :: nili,njlj,nklk,nlll,mi,mj,mk,ml,ibase,jbase,kbase,lbase
!  Complex, Pointer :: Wijkl(:)
!  Integer, Pointer :: LUT(:), nl_base(:), nlm_LUT(:,:)
!  Type (Atom_Info_Fixed), pointer :: AT
!
! ! Initialize Wijkl
!  Do s=1,Atom_Types
!     AtomType_Info(s)%Wijkl = 0
!  enddo
!
!  NN=Numbands
!  If (spindependence) NN=2*NN   ! Twice as many bands stored in spin case
!  Do s=1, Specific_Atoms
!     AT => AtomType_Info(Atom_List(s)%TypeIndex)
!
!     Wijkl => AT%Wijkl
!     LUT =>AT%LUT_Cijkl
!     Cijkl_Size = AT%Cijkl_Size 
!     nl_base => AT%nl_base
!     nlm_LUT => AT%nlm_LUT
!
!     Base = PLM_AtomRange(1,s) - 1
!
!     Do p=1,Cijkl_Size
!        Call Coulomb_Decode(LUT(p), nili, njlj, nklk, nlll, mi, mj, mk)
!        ml = mi - mj + mk
!        ibase = nl_Base(nili); ibase = ibase + nlm_lut(2,ibase) + mi;
!        jbase = nl_Base(njlj); jbase = jbase + nlm_lut(2,jbase) + mj;
!        kbase = nl_Base(nklk); kbase = kbase + nlm_lut(2,kbase) + mk;
!        lbase = nl_Base(nlll); lbase = lbase + nlm_lut(2,lbase) + ml;
!
!        i=ibase+Base;j=jbase+Base;k=kbase+Base;l=lbase+Base;
!
!        Wijkl(p)=Wijkl(p)+SUM(Occupancy(1:NN,:)*&
!            CONJG(PDot(i,1:NN,:)*PDot(k,1:NN,:))*&
!             PDot(j,1:NN,:)*PDot(l,1:NN,:))
!     End Do
!
!  Enddo
!  
!End Subroutine
 
!******************************************************************************
!                            a
!  CalcQlm - Calculates the Q  coefficients according to eq. A22 and 
!                            LM 
!
!             also accumulates the Qlm term in eq. A28 to dEdQlm
!
!******************************************************************************

Subroutine CalcQlm(energy)
  Real, Intent(INOUT) :: Energy
  
  Integer        :: i, j, k, Qlm_Size, Offset_Q
  Integer        :: nili, njlj, mi, mj, M, L, Li, Lj, ibase, jbase
  Integer, Pointer :: LUT_Orb(:), nl_Base(:), nlm_LUT(:,:)
  Real, Pointer  :: aQlm(:), Hat(:)
  Complex, Pointer :: Qlm(:,:), Wij(:,:), dEdQlm(:,:), Wijspin(:,:)
  Complex :: Qlmvalue
  Real     :: fn,tt
  Type (Specific_Atom), Pointer :: A
  Type (Atom_Info_Fixed), Pointer :: AT
  

   write(log_unit,*) 'CalcQlm: Begin'
   call flush(log_unit)
     tt=0    !!!!!!
  Do i=1, Specific_Atoms
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)

     nl_base => AT%nl_Base
     nlm_LUT => AT%nlm_LUT
     aQlm => AT%aQlm
     LUT_Orb => AT%LUT_Orb
     Qlm_Size = AT%QVlm_Size
     Qlm => A%Qlm
     dEdQlm => A%dEdQlm
     Wij => A%Wij
     if(spindependence) Wijspin => A%Wijspin

     Offset_Q = 2*MaxVal(AT%L_Value) + 1
     !Qlm(1, Offset_Q) = Qlm(1,Offset_Q) + AT%Core_Charge - At%Atomic_Charge 
     Qlm(1, Offset_Q) = Qlm(1,Offset_Q) + AT%Qeffion
     Hat => AT%Hat_SelfEnergy

     Do j=1, Qlm_Size
        Call Orbital_Decode(LUT_Orb(j), nili, njlj, mi, mj, L)
        M = mj - mi
        ibase = nl_base(nili);  ibase = ibase + nlm_LUT(2, ibase)
        jbase = nl_base(njlj);  jbase = jbase + nlm_LUT(2, jbase)

        Qlm(L+1, M+Offset_Q) = Qlm(L+1, M+Offset_Q) + &
                 Wij(ibase+mi, jbase+mj) * aQlm(j)
        if(spindependence) then
           Qlm(L+1, M+Offset_Q) = Qlm(L+1, M+Offset_Q) + &
                 Wijspin(ibase+mi, jbase+mj) * aQlm(j)
        end if
        !write(log_unit,*) 'CalcQLM: L M QLM ', L, M, QLM(L+1,M+Offset_Q)
        !call flush(Log_unit)
     End Do

     !***** Now accum dEdQlm and the Energy *****
     Do L=0, Offset_Q - 1

        fn = Hat(L)

        Do M=-L, L
           Qlmvalue=Qlm(L+1,Offset_Q + M)
           dEdQlm(L+1, Offset_Q+M) = dEdQlm(L+1, Offset_Q+M) - &
                     2*fn * CONJG(Qlmvalue)

           Energy= Energy - &
                fn*CONJG(Qlmvalue)*Qlmvalue
           if (L==0.AND.M==0) Qlmvalue=Qlmvalue-AT%Qeffion
          !if (Estimate_Selfenergy) &
          ! Auxform_of_Hartree=Auxform_of_Hartree&
          !     + fn*CONJG(Qlmvalue)*Qlmvalue
        End Do
     End Do
   tt=tt+Qlm(1,Offset_Q) !!!!!!
  ! add coretail effects
  Energy=Energy-AT%CoreTail_SelfEnergy-Qlm(1,Offset_Q)*AT%CoreTail_HatEnergy
  dEdQlm(1,Offset_Q)=dEdQlm(1,Offset_Q)-AT%CoreTail_HatEnergy
  End Do

 ! if (Estimate_Selfenergy) &
 !          write(Log_unit,*) ' Auxform ', Auxform_of_Hartree
  write(Log_Unit,*) ' check Q00 ', tt   !!!!!
  call flush(log_unit)
  

!   Write(6,*) ' dEdQlm with self term'
!   call WriteQlm(2,6)


  Return
End Subroutine

!******************************************************************************
!                  
!  AccumVhartree - Accumulates the Vhartree terms in Dij based on eq A26
!                  
!******************************************************************************

Subroutine AccumVhartree(Energy)
  Real, Intent(INOUT) :: Energy
  
  Integer        :: i, j, k, Qlm_Size, Offset_Q,  Cijkl_Size
  Integer        :: nili, njlj, nlll,nklk, mi, mj, mk, ml
  Integer        :: ibase,jbase,lbase,kbase
  Integer, Pointer :: LUT_Cijkl(:), nl_base(:), nlm_LUT(:,:)
  Real, Pointer  :: Cijkl(:),TCijkl(:)
  Complex, Pointer :: Dij(:,:), Wij(:,:), Dijspin(:,:), Wijspin(:,:)
  Complex        :: fn,gn
  Real           :: Venergy
  Type (Specific_Atom), Pointer :: A
  Type (Atom_Info_Fixed), Pointer :: AT
  
  write(log_unit,*) 'AccumVhartree: Begin'
  call flush(log_unit)

  Venergy = 0

  Do i=1, Specific_Atoms
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)

     Offset_Q = 2*MaxVal(AT%L_Value) + 1

     Cijkl => AT%Cijkl
     !If (Estimate_selfenergy) TCijkl => AT%TCijkl
     LUT_Cijkl => AT%LUT_Cijkl
     Cijkl_Size = AT%Cijkl_Size
     nl_base => AT%nl_base
     nlm_LUT => AT%nlm_LUT
     Wij => A%Wij
     Dij => A%Dij

     if(spindependence) then
          Wijspin => A%Wijspin
          Dijspin => A%Dijspin
     end if

     Do j=1, Cijkl_Size
        Call Coulomb_Decode(LUT_Cijkl(j), nili, njlj, nklk, nlll, mi, mj, mk)
        ml = mi - mj + mk
        ibase = nl_Base(nili); ibase = ibase + nlm_lut(2,ibase) + mi;
        jbase = nl_Base(njlj); jbase = jbase + nlm_lut(2,jbase) + mj;
        kbase = nl_Base(nklk); kbase = kbase + nlm_lut(2,kbase) + mk;
        lbase = nl_Base(nlll); lbase = lbase + nlm_lut(2,lbase) + ml;
        if(.NOT.spindependence) then
            fn = Wij(kbase, lbase) * Cijkl(j)
            !If (Estimate_selfenergy)&
            !     gn = Wij(kbase, lbase) * TCijkl(j)/2
            Dij(ibase,jbase) = Dij(ibase,jbase) + fn             

            Venergy = Venergy + Wij(ibase,jbase) * fn
            !If (Estimate_selfenergy) &
            !         Auxform_of_Hartree = Auxform_of_Hartree&
            !     + Wij(ibase,jbase) * gn
        else
            fn = (Wij(kbase, lbase)+Wijspin(kbase, lbase)) * Cijkl(j)
            !If (Estimate_selfenergy) gn = (Wij(kbase, lbase)+Wijspin(kbase, lbase)) * TCijkl(j)/2
            Dij(ibase,jbase) = Dij(ibase,jbase) + fn             
            Dijspin(ibase,jbase) = Dijspin(ibase,jbase) + fn             

            Venergy = Venergy + (Wij(ibase,jbase)+Wijspin(ibase, jbase)) * fn
            !If (Estimate_selfenergy) Auxform_of_Hartree = Auxform_of_Hartree&
            !     + (Wij(ibase,jbase)+Wijspin(ibase, jbase)) * gn
        end if
        
     End Do
  End Do

  Energy = Energy + 0.5*Venergy
 
  !If (Estimate_selfenergy) write(log_unit,*) 'Auxform', Auxform_of_Hartree

  write(log_unit,*) 'AccumVhartree: End'
  call flush(log_unit)

  Return
End Subroutine

!****************************************************************
!  AccumOneSelf -- one-center contributions to self-energy
!   no longer used
!****************************************************************
!Subroutine AccumOneSelf
!    Integer :: a,i,j,k,Cijkl_Size,ii,jj,l
!    Real, Pointer  :: TCijkl(:)
!    Complex, Pointer :: Wijkl(:)
!    Type (Atom_Info_Fixed), Pointer :: AT
!    Type (Specific_atom) , Pointer :: AA
!    Complex :: arg
!    Real :: x
!
!    Do a=1,Atom_Types
!       AT=>AtomType_Info(a)
!
!       Cijkl_Size=AT%Cijkl_Size
!       TCijkl => AT%TCijkl
!       Wijkl => AT%Wijkl
!       x=0
!  
!       Do j=1,Cijkl_Size
!          !if (ABS(Wijkl(j))>0.1.and.ABS(TCijkl(j))>0.1) &
!          !write(log_unit,*) 'ijkl contrib: ',a,j,Real(Wijkl(j)),TCijkl(j)
!          x = x - Wijkl(j)*TCijkl(j)/2
!       Enddo
!          Self_Energy_Correction = Self_Energy_Correction + x
!
!       write(Log_unit,'(" Self Energy from ", i5, 2x, a5, 1pe15.7)') &
!                 a,AT%Atom_Name,x
!    Enddo
!
!    write(Log_unit, *) ' AccumOneSelf: ', Self_Energy_Correction
!
!!    do a=1,Specific_Atoms
!!       AA => Atom_List(a)
!!       AT => ATomType_Info(AA%TypeIndex)
!!       do i=1,AT%Basis_Size
!!          l=AT%L_Value(i)
!!          do j=1,AT%Basis_Size
!!             if (AT%L_Value(j)==l) then
!!                arg=0
!!                do ii=0,2*l
!!                   arg=arg+AA%Wij(AT%nl_base(i)+ii,AT%nl_base(j)+ii)
!!                   if (spindependence) then
!!                      arg=arg+AA%Wijspin(AT%nl_base(i)+ii,AT%nl_base(j)+ii)
!!                   endif
!!                enddo
!!                Self_Energy_Correction=Self_Energy_Correction &
!!                   -arg*AT%EXFij(i,j)/2
!!              endif
!!           enddo
!!        enddo
!!      enddo
!!
!!
!!    write(Log_unit, *) ' AccumOneSelf: -- part 2', Self_Energy_Correction
!End Subroutine
!******************************************************************************
!                  
!  AccumVhatdEdQlm - Accumulates the Vhat terms in Dij based on eq A23 
!                    and also accumulates the contributions to dEdQlm
!                  
!******************************************************************************

Subroutine AccumVhatdEdQlm(Energy)
  Real, Intent(INOUT) :: Energy
  
  Integer        :: i, j, k, Vlm_Size, Offset_Q, ibase,jbase
  Integer        :: nili, njlj, mi, mj, M, L
  Integer, Pointer :: LUT_Orb(:), nl_base(:), nlm_LUT(:,:)
  Real, Pointer  :: aVlm(:)
  Complex, Pointer :: dEdQlm(:,:), Qlm(:,:), Wij(:,:), Dij(:,:)
  Complex, Pointer :: Wijspin(:,:), Dijspin(:,:)
  Complex          :: fn,gn
  Type (Specific_Atom), Pointer :: A
  Type (Atom_Info_Fixed), Pointer :: AT

  write(log_unit,*) 'AccumVhatDEdQlm: Begin'
  call flush(log_unit)

  Do i=1, Specific_Atoms
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)

     Offset_Q = AT%Hat_MaxL + 1

     aVlm => AT%aVlm
     LUT_Orb => AT%LUT_Orb
     Vlm_Size = AT%QVlm_Size
     Qlm => A%Qlm
     dEdQlm => A%dEdQlm
     nl_base => AT%nl_base
     nlm_LUT => AT%nlm_LUT
     Wij => A%Wij
     Dij => A%Dij

     if(spindependence) then
          Wijspin => A%Wijspin
          Dijspin => A%Dijspin
     end if

     Do j=1, Vlm_Size
        Call Orbital_Decode(LUT_Orb(j), nili, njlj, mi, mj, L)
        M = -(mj - mi)           !** Change sign for aVlm
        
        ibase = nl_Base(nili); ibase = ibase + nlm_lut(2,ibase) + mi;
        jbase = nl_Base(njlj); jbase = jbase + nlm_lut(2,jbase) + mj;

        fn = aVlm(j) * Qlm(L+1, M+Offset_Q)   !** Reverse M-sign for Qlm
        gn = fn
        if (L==0.AND.M==0) gn=gn-aVlm(j)*AT%Qeffion
        Dij(ibase, jbase) = Dij(ibase,jbase) - fn
        
        Energy = Energy - Real(Wij(ibase,jbase) * fn)
        !If (Estimate_Selfenergy) Auxform_of_Hartree = Auxform_of_Hartree&
        !       + Real(Wij(ibase,jbase) * gn)

                    !** Reverse M-sign for dEdQlm
        dEdQlm(L+1, Offset_Q + M) = dEdQlm(L+1, Offset_Q + M) - &
             Wij(ibase,jbase) * aVlm(j)
        if(spindependence) then
            Dijspin(ibase, jbase) = Dijspin(ibase,jbase) - fn
            Energy = Energy - Real(Wijspin(ibase,jbase) * fn)
            !If (Estimate_Selfenergy) Auxform_of_Hartree = Auxform_of_Hartree&
            !   + Real(Wijspin(ibase,jbase) * gn)
            dEdQlm(L+1, Offset_Q + M) = dEdQlm(L+1, Offset_Q + M) - &
                 Wijspin(ibase,jbase) * aVlm(j)
        end if
     End Do
  End Do

!   Write(6,*) ' dEdQlm with vhat term'
!   call WriteQlm(2,6)
  !If (Estimate_Selfenergy) write(log_unit,*) 'Auxform', Auxform_of_Hartree
  write(log_unit,*) 'AccumVhatDEdQlm: complete'
  call flush(log_unit)

  Return
End Subroutine

!******************************************************************************
!                  
!  AccumVlocal - Accumulates the Vlocal                        
!                  
!******************************************************************************

Subroutine AccumVlocal(Energy, Den, Denspin)
  Complex,         Intent(IN) :: Den(:)
  Complex,         Intent(IN), OPTIONAL :: Denspin(:)
  Real,            Intent(INOUT) :: Energy !Accumulate Pot. Energy

  Complex,         Pointer :: V(:),Vspin(:)
  Real    :: theta, Gmag, RadG, G(3), Esum
  Complex :: Phase, Vsum
  Integer :: atom, atype, lut, gindex, RadIndex, fi
  Type (Atom_Info_Fixed), Pointer :: AT

  write(log_unit,*) 'AccumVlocal: begin'
  call flush(log_unit)

  V => SCFvalues%Ve
  if(spindependence) Vspin => SCFvalues%Vespin
  Esum = 0

  Do gindex=1, Gpnt_Size(G_HIGH)
     Vsum = SCFvalues%V_local(gindex)

     fi = FFTmap_High(gindex)
     V(fi) = V(fi) + Vsum
     !write(log_unit,*) 'AccumVlocal ', gindex,fi,Vsum
     !call flush(log_unit)
     Esum = Esum + Vsum*Conjg(Den(fi))
     if(spindependence) then
         Vspin(fi) = Vspin(fi) +Vsum
         Esum = Esum + Vsum*Conjg(Denspin(fi))
     endif
     If (gindex>1) then
        fi = FFTmap_High(Gindex + Gpnt_Size(G_HIGH)-1)
        V(fi) = V(fi) + CONJG(Vsum)
     !write(log_unit,*) 'AccumVlocal ', gindex,fi,CONJG(Vsum)
     !call flush(log_unit)
        Esum = Esum + Conjg(Vsum*Den(fi))
        if(spindependence) then
             Vspin(fi) = Vspin(fi) +CONJG(Vsum)
             Esum = Esum + Conjg(Vsum*Denspin(fi))
        endif
     End if
  End Do
 
  Energy = Energy + Esum

  write(log_unit,*) 'AccumVlocal: end'
  call flush(log_unit)


  Return
End Subroutine


!******************************************************************************
!                  
!  CalcDenhat - Accumulates the n^ density
!                  
!******************************************************************************

Subroutine CalcDenHat( DenHat)
  Complex, Intent(INOUT) :: DenHat(:)    ! input density is augmented with n^
  
  Real    :: theta, Gmag,  RadG, MinQlm, G(3)
  Complex :: Phase(Specific_Atoms), Ylm(13), Ylm_neg(13), den,tmp
  Complex, parameter :: ai=(0,-1)
  Integer :: atom, atype, L, m, lut, gindex, Offset_Q, HatIndex, fi, Gi(3)
  Logical :: WhichL(20, Atom_Types)
  Type (Atom_Info_Fixed), Pointer :: AT
  Type (Specific_Atom), Pointer :: A

  Call Start_Timer(Timer(DenHat_Timer))
  !** First create L filter using the Qlm **
  WhichL = .FALSE.
  !MinQlm = 1.E-30
  Do atom=1, Specific_Atoms
     A => Atom_List(atom)
     AT => AtomType_Info(A%TypeIndex)

     MinQlm = 1E-20*MAXVAL(REAL(CONJG(A%Qlm)*A%Qlm))

     Offset_Q = AT%Hat_MaxL + 1
     Do L=0, AT%Hat_MaxL
        Do m=-L, L
           theta = CONJG(A%Qlm(L+1, Offset_Q+m))*A%Qlm(L+1, Offset_Q+m)
           If (theta > MinQlm) WhichL(L+1, A%TypeIndex) = .TRUE.
        End Do
     End Do
  End Do

  !*** Now calculate the Hat density in G-space ***
  HatIndex = 0
  Do gindex=1, Gpnt_Size(G_HIGH)
     fi = FFTmap_High(gindex)
     G = Gpnt(1:3,gindex)
     Gmag = Gpnt(4,gindex)

     !*** If needed inc the HatIndex ***
     If (G_NewMag(gindex)) HatIndex = HatIndex + 1

     den = 0
     Do atype=1, Atom_Types
        AT => AtomType_Info(atype)

        Offset_Q = AT%Hat_MaxL + 1

        !*** Create and store the structure factors ***
        Do lut = AtomType_Range(1,atype), AtomType_range(2,atype)
           atom = Atomtype_map(lut)
        
           theta = DOT_PRODUCT(G, Atom_List(atom)%Pos)
           Phase(atom) = CMPLX(cos(theta), -sin(theta))
        End Do


        Do L=0, AT%Hat_MaxL
           If (WhichL(L+1, atype)) then
              Ylm = Spharm(G(1), G(2), G(3), L, .TRUE.)
              
              RadG = AT%RadHat(HatIndex, L+1)
              Do lut = AtomType_Range(1,atype), AtomType_range(2,atype)
                 atom = Atomtype_map(lut)

!If (gindex==1) then 
!  Write(Log_Unit,*) 'CalcDenHat: atype=',atype, ' * L=', L, ' * RadG=',RadG, ' * Phase=',Phase(atom)
!end If
                 tmp=0
                 Do m=-L, L
                    tmp=tmp+Atom_List(Atom)%Qlm(L+1, Offset_Q+m)*Ylm(L+m+1)
!If (gindex==1) then 
!  Write(Log_Unit,*) 'CalcDenHat: atype=',atype, ' * atom=',atom, ' * L=',L, ' * m=',m, ' * Qlm=',Atom_List(Atom)%Qlm(L+1, Offset_Q+m), ' * Term=',Atom_List(Atom)%Qlm(L+1, Offset_Q+m) * Phase(atom)*RadG*Ylm(L+m+1)
!end If
                 Enddo

                 tmp=tmp*(ai**L)
                 den = den + tmp*Phase(atom)*RadG
!If (gindex==1) then
!   Write(Log_Unit,*) 'CalcDenHat: atom=',atom, ' * L=',L, ' * SumDenHat=',den
!End If
             End Do
           End if
        End Do
     End Do

!If (gindex == 1) Write(Log_Unit,*) 'CalcDenHat: Qhat=',Den, ' * Den(1)=',DenHat(1)

     DenHat(fi) = DenHat(fi) + den

! Store RhoHat
     SCFvalues%RhoHat(gindex)=den

     !write(6,*) 'CalcDen',fi,gmag,den
     If (gindex>1) then
        fi = FFTmap_High(gindex+Gpnt_size(G_HIGH)-1)
        DenHat(fi) = DenHat(fi) + CONJG(den)
     End If
  End Do

  Call Stop_Timer(Timer(DenHat_Timer))
    !write(6,*) 'in denhat -', denHat(1),denHat(7),denHat(97),denHat(200)
  Return
End Subroutine

!******************************************************************************
!                  
!  CalcDen_dEdQlm - Accumulates the n~+n^ density term to dE/dQlm.
!                  
!******************************************************************************

Subroutine CalcDen_dEdQlm( Den)
  Complex, Intent(IN) :: Den(:)     !This must be n~ + n^(G)
  
  Real    :: theta, Gmag,  RadG, MinQlm, c1, G(3)
  Complex :: Phase(Specific_Atoms), Ylm(13), Ylm_neg(13), DenG, tmp
  Complex , parameter :: ai=(0,-1)
  Integer :: atom, atype, L, m, lut, gindex, Offset_Q, HatIndex,Gi(3)
  Logical :: WhichL(20, Atom_Types)
  Type (Atom_Info_Fixed), Pointer :: AT
  Type (Specific_Atom), Pointer :: A


  c1 = 2*Four_Pi/xtal%Volume


  HatIndex = 1
  Do gindex=2, Gpnt_Size(G_HIGH)
     G = Gpnt(1:3,gindex)
     Gmag = Gpnt(4,gindex)

     DenG = CONJG(Den(FFTmap_High(gindex))+SCFValues%CoreTail(gindex))/(Gmag*Gmag)

     !*** If needed inc the HatIndex ***
     If (G_NewMag(gindex)) HatIndex = HatIndex + 1

     Do atype=1, Atom_Types
        AT => AtomType_Info(atype)

        Offset_Q = AT%Hat_MaxL + 1

        !*** Create and store the structure factors ***
        Do lut = AtomType_Range(1,atype), AtomType_range(2,atype)
           atom = Atomtype_map(lut)
        
           theta = DOT_PRODUCT(G, Atom_List(atom)%Pos)
           Phase(atom) = CMPLX(cos(theta), -sin(theta))
        End Do

        Do L=0, AT%Hat_MaxL
           Ylm = Spharm(G(1), G(2), G(3), L, .TRUE.)
           !Ylm_neg = Spharm(-G(1), -G(2), -G(3), L, .TRUE.)
           Ylm_neg = Ylm; if (L>0) Ylm_neg=Ylm_neg*((-1)**L)
           
           RadG = AT%RadHat(HatIndex, L+1)
           tmp = ai**L
           Do lut = AtomType_Range(1,atype), AtomType_range(2,atype)
              atom = Atomtype_map(lut)

              Do m=-L, L
                 Atom_List(atom)%dEdQlm(L+1, Offset_Q+m) = &
                      Atom_List(atom)%dEdQlm(L+1, Offset_Q+m) + &
                      c1 * tmp * (DenG * Ylm(L+m+1)*Phase(atom) + &
                      Ylm_neg(L+m+1)*CONJG(DenG * Phase(atom)))*RadG
              End Do
           End Do
        End Do
     End Do
  End Do

!   Write(6,*) ' dEdQlm with den term'
!   call WriteQlm(2,6)

  Return
End Subroutine


!******************************************************************************
!                  
!  CalcDenhat_dEdQlm - Accumulates the  n^ density in Den to n~ and 
!                also accumulates the term in dEdQlm containing n^+n~.
!                  
!******************************************************************************

Subroutine CalcDenHat_dEdQlm( Den)
  Complex,         Intent(INOUT) :: Den(:)


  Call CalcDenHat( Den)
  Call CalcDen_dEdQlm( Den)

  Return
End Subroutine

!******************************************************************************
!
!  AccumPotential - Accumulates the potential energy due to n~+n^+ncore~.
!
!******************************************************************************

Subroutine AccumPotential( Energy, Rho)
  Real,    Intent(INOUT) :: Energy
  Complex, Intent(INOUT) :: Rho(:)
  Complex, Pointer :: V(:), Vspin(:)

  Real :: c1, c2 , Esum
  Complex :: Vpnt, Den
  Integer :: gi, N

  c1 = Four_Pi/xtal%Volume
  c2 = 2*c1

 !write(6,*) 'in AccumPotential',c1,c2

  N = Gpnt_Size(G_HIGH) - 1
  V => SCFValues%Ve
  if(spindependence) Vspin => SCFValues%Vespin
  Esum = 0
  Do gi=2, Gpnt_Size(G_High)
     Den = Rho(FFTmap_High(gi)) +SCFValues%CoreTail(Gi)
     Vpnt = c2*(Den)/ Gpnt(4,gi)**2
     Esum = Esum + REAL(Vpnt*CONJG(den))
     Hat_Hat_Energy=Hat_Hat_Energy+ &
       c2*CONJG(SCFvalues%RhoHat(gi))*SCFvalues%RhoHat(gi)/Gpnt(4,gi)**2
     !write(6,*) 'ACCPOT',gi,ESUM
    !V(FFTmap_High(gi)) = V(FFTmap_High(gi)) + FFT_Grid(4,G_High)*Vpnt
    !V(FFTmap_High(gi+N)) = V(FFTmap_High(gi+N)) + FFT_Grid(4,G_High)*CONJG(Vpnt)
    ! If (V_MixType /= MIX_V) then
        V(FFTmap_High(gi)) = V(FFTmap_High(gi)) + Vpnt      !** Orig
        V(FFTmap_High(gi+N)) = V(FFTmap_High(gi+N)) + CONJG(Vpnt)
        if(spindependence) then
            Vspin(FFTmap_High(gi)) = Vspin(FFTmap_High(gi)) + Vpnt      !** Orig
            Vspin(FFTmap_High(gi+N)) = Vspin(FFTmap_High(gi+N)) + CONJG(Vpnt)
        end if
    ! else
    !    Rho(FFTmap_High(gi)) = Vpnt
    !    Rho(FFTmap_High(gi+N)) = CONJG(Vpnt)
    !  End if
  End Do
  
  Energy = Energy + Esum

  Return
End Subroutine

!!******************************************************************************
!! Subroutine AccumSmoothSelf
!!******************************************************************************

!Subroutine AccumSmoothSelf(WK1)
!   Complex,  Intent(OUT) :: WK1(:)
!
!   Real     :: G(3), G2, BlochK(3), angle , prefac, prefack, Occ,   CurrE
!   Type (Atom_Info_Fixed), Pointer :: AT
!   Type (Specific_Atom), Pointer ::A
!   Integer          ::  Kpnt, PsiBand, PsiK,i,j,k, s, Band,Degen,Clusters
!   Integer          :: gi, N,  t,Flag,Base,start,finish,start1,finish1,spin
!   Integer  :: nili,njlj,nklk,nlll,mi,mj,mk,ml,p,in,jn,kn,ln
!   Integer, allocatable   :: Clusterinit(:),Clusterend(:)
!   Complex, allocatable   :: Den(:),Psi(:)
!   Complex, pointer      :: PDOT(:)
!   Complex  :: phase,term,term1,term2
!   Real, parameter :: Etol=1.e-3
!
!   Prefac= 2*Four_Pi/xtal%Volume   ! Summing over pos only
!   N=Gpnt_Size(G_HIGH)-1
!   s=1
!   if (spindependence) s=2
!   Allocate(Den(N+1),Psi(Gall_Size(G_LOW)),clusterinit(s*Numbands),&
!            clusterend(2*Numbands),stat=i)
!      if (i/=0) then
!         write(Error_unit,*) 'AccumSelf allocation error ', N+1,i
!         stop
!      endif
!
!   Call LoadHatij
!   Do t=1,Atom_Types
!      AtomType_Info(t)%Wijkl=0
!   enddo
!
!   Do k=1, NumKpnts
!      Clusters=0;clusterinit=0;clusterend=0
!   
!      Mem_Kpnt=k
!
!     Do spin=1,s
!      CurrE=1.0e30
!      Do Band=1+(spin-1)*Numbands,Numbands+(spin-1)*Numbands
!         i=BKMap(Band,k)
!         !write(Log_unit,'("BKMap ",10i5)') k,Band,Clusters,i
!         if (i>0 .and. PsiInfo(i)%Occupancy>1.e-6) then
!            !write(log_Unit,*) 'CHK: ', PsiInfo(i)%Kpnt,PsiInfo(i)%Energy,PsiInfo(i)%Occupancy
!            if (ABS(PsiInfo(i)%Energy-CurrE)<Etol) then
!               if (Clusters>0) then
!                  clusterend(Clusters)=Band
!                  CurrE=PsiInfo(i)%Energy
!               else
!                  write (Log_Unit,*) ' Error AccumSmoothSelf ', i,Clusters,Band,k
!                  stop
!               endif
!            else
!              Clusters=Clusters+1
!              clusterinit(Clusters)=Band
!              clusterend(Clusters)=Band 
!              CurrE=PsiInfo(i)%Energy
!            endif
!          endif
!       enddo ! Bands
!      enddo !spin
!       If (Clusters>0) then
!         Do j=1,Clusters
!            Degen=Clusterend(j)-Clusterinit(j)+1
!            Den=0;   Occ=0
!            !write(Log_Unit,*) 'Clusters - ', k,j,DeGen,Clusterinit(j),Clusterend(j)
!            do t=1,Specific_Atoms
!               Atom_List(t)%TWij=0
!            enddo
!            Do Band=Clusterinit(j),Clusterend(j)
!               i=BKMap(Band,k)
!            
!               Call GetPsi_fromBuffer(i,Psi)
!               Call FilterValue(Psi)
!               call kplusGfilter(Psi,k)
!
!               Occ=Occ+PsiInfo(i)%Occupancy
!
!               WK1=0
!               Call V_G_to_r(Psi,G_LOW,WK1,G_HIGH)
!               WK1=CONJG(WK1)*WK1
!               Call PerformFFT(FFT_TO_G, G_HIGH, WK1)
!
!               Den(1)=Den(1)+WK1(FFTmap_High(1))
!               do gi=2,N+1
!                  Den(gi)=Den(gi)+(WK1(FFTmap_High(gi))+&
!                          WK1(FFTmap_High(gi+N)))/2
!               enddo
!
!               If (.NOT. PsiInfo(i)%PDOT_Stored) then
!                  Call CalcProjProducts(Psi, PsiInfo(i)%PDOT)
!                  PsiInfo(i)%PDOT_Stored= .TRUE.
!               endif
!
!               PDOT => PsiInfo(i)%PDOT
!               Do t=1,Specific_Atoms
!                  A => Atom_List(t)
!                  AT => AtomType_Info(A%TypeIndex)
!                  Base = PLM_AtomRange(1,t) - 1
!
!                  start=PLM_AtomRange(1,t); finish=PLM_AtomRange(2,t)
!                  start1=start-Base; finish1=finish-Base
!                  do p=start1,finish1
!                     A%TWij(start1:finish1,p)=A%TWij(start1:finish1,p)+&
!                         CONJG(PDOT(start:finish))*PDOT(p+Base)
!                  enddo
!              enddo 
!             enddo !Band
!             Den=Den/Degen
!             Do t=1,Specific_Atoms
!                Atom_List(t)%TWij= Atom_List(t)%TWij/Degen
!             Enddo
!
!             do gi=1,Gpnt_Size(G_High)
!                  G=Gpnt(1:3,gi)
!                  term=0
!                  Do t=1,Specific_Atoms
!                     A => Atom_List(t)
!                     AT => AtomType_Info(A%TypeIndex)
!                     
!                     angle=Dot_Product(G,A%Pos)
!                     Phase= CMPLX(cos(angle),-sin(angle))
!  
!                     Base = AT%nlm_Size
!
!                     term=term+Phase*&
!                       SUM(A%TWij(1:Base,1:Base)*AT%Hatij(1:Base,1:Base,gi))
!                  enddo
!                  Den(gi)=Den(gi)+term
!              enddo
!   
!              ! Density term is now complete -- Accumulate Self-Energy
!              !write(Log_unit,*)'AccumSmoothSelf: Den(G=0) = ',degen,Occ,Den(1)
!               Prefack=Prefac*Occ
!               !write(Log_unit,*) 'Prefack',Prefack
!
!               Do gi=2,Gpnt_Size(G_High)
!                  Self_Energy_Correction=Self_Energy_Correction&
!                 - Prefack*CONJG(Den(gi))*Den(gi)/Gpnt(4,gi)**2
!               Enddo
!
!               ! Accumulate Wijkl terms
!               Do t=1,Specific_Atoms
!                  A => Atom_List(t)
!                  AT => AtomType_Info(A%TypeIndex)
!                  Do p=1,AT%Cijkl_Size
!                     Call Coulomb_Decode(AT%LUT_Cijkl(p),nili,njlj,nklk,nlll,&
!                          mi, mj, mk)
!                     ml = mi - mj + mk
!                     in=AT%nl_base(nili)+AT%L_Value(nili)+mi
!                     jn=AT%nl_base(njlj)+AT%L_Value(njlj)+mj
!                     kn=AT%nl_base(nklk)+AT%L_Value(nklk)+mk
!                     ln=AT%nl_base(nlll)+AT%L_Value(nlll)+ml
!                     AT%Wijkl(p)=AT%Wijkl(p)+Occ*A%TWij(in,jn)*A%TWij(kn,ln)
!                  Enddo
!               Enddo
!
!      enddo   !Clusters
!    endif
!   enddo   !Kpnt
!
!   write(Log_unit,*) 'AccumSmoothSelf:  Energy = ', Self_Energy_Correction
!
!   call DumpHatij
!   DeAllocate(Den,Psi ,Clusterinit,Clusterend)
!
!End Subroutine

!******************************************************************************
! Subroutine AccumWijkl
!   no longer used
!******************************************************************************

!Subroutine AccumWijkl
!
!   Real     :: G(3), G2, BlochK(3), angle,prefac,prefack,Occ,CurrE,x
!   Type (Atom_Info_Fixed), Pointer :: AT
!   Type (Specific_Atom), Pointer ::A
!   Integer          ::  Kpnt, PsiBand, PsiK,i,j,k, s, Band,Degen,Clusters
!   Integer          :: gi, N,  t,Flag,Base,start,finish,start1,finish1,spin
!   Integer  :: nili,njlj,nklk,nlll,mi,mj,mk,ml,p,in,jn,kn,ln
!   Integer, allocatable   :: Clusterinit(:),Clusterend(:)
!   Complex, allocatable   :: Psi(:)
!   Complex, pointer      :: PDOT(:)
!   Complex  :: phase,term,term1,term2
!   Real, parameter :: Etol=1.e-3
!
!   s=1
!   if (spindependence) s=2
!   Allocate(Psi(Gall_Size(G_LOW)),clusterinit(s*Numbands),&
!            clusterend(2*Numbands),stat=i)
!      if (i/=0) then
!         write(Error_unit,*) 'AccumSelf allocation error ', N+1,i
!         stop
!      endif
!
!   Do t=1,Atom_Types
!      AtomType_Info(t)%Wijkl=0
!   enddo
!
!   Do k=1, NumKpnts
!      Clusters=0;clusterinit=0;clusterend=0
!   
!      Mem_Kpnt=k
!
!     Do spin=1,s
!      CurrE=1.0e30
!      Do Band=1+(spin-1)*Numbands,Numbands+(spin-1)*Numbands
!         i=BKMap(Band,k)
!         !write(Log_unit,'("BKMap ",10i5)') k,Band,Clusters,i
!         if (i>0 .and. PsiInfo(i)%Occupancy>1.e-6) then
!            !write(log_Unit,*) 'CHK: ', PsiInfo(i)%Kpnt,PsiInfo(i)%Energy,PsiInfo(i)%Occupancy
!            if (ABS(PsiInfo(i)%Energy-CurrE)<Etol) then
!               if (Clusters>0) then
!                  clusterend(Clusters)=Band
!                  CurrE=PsiInfo(i)%Energy
!               else
!                  write (Log_Unit,*) ' Error AccumSmoothSelf ', i,Clusters,Band,k
!                  stop
!               endif
!            else
!              Clusters=Clusters+1
!              clusterinit(Clusters)=Band
!              clusterend(Clusters)=Band 
!              CurrE=PsiInfo(i)%Energy
!            endif
!          endif
!       enddo ! Bands
!      enddo !spin
!       If (Clusters>0) then
!         Do j=1,Clusters
!            Degen=Clusterend(j)-Clusterinit(j)+1
!            Occ=0
!           !write(Log_Unit,*) 'Clusters - ', k,j,DeGen,Clusterinit(j),Clusterend(j)
!            do t=1,Specific_Atoms
!               Atom_List(t)%TWij=0
!            enddo
!            Do Band=Clusterinit(j),Clusterend(j)
!               i=BKMap(Band,k)
!            
!               Call GetPsi_fromBuffer(i,Psi)
!               Call FilterValue(Psi)
!               call kplusGfilter(Psi,k)
!
!               Occ=Occ+PsiInfo(i)%Occupancy
!
!               If (.NOT. PsiInfo(i)%PDOT_Stored) then
!                  Call CalcProjProducts(Psi, PsiInfo(i)%PDOT)
!                  PsiInfo(i)%PDOT_Stored= .TRUE.
!               endif
!
!               PDOT => PsiInfo(i)%PDOT
!               Do t=1,Specific_Atoms
!                  A => Atom_List(t)
!                  AT => AtomType_Info(A%TypeIndex)
!                  Base = PLM_AtomRange(1,t) - 1
!
!                  start=PLM_AtomRange(1,t); finish=PLM_AtomRange(2,t)
!                  start1=start-Base; finish1=finish-Base
!                  do p=start1,finish1
!                     A%TWij(start1:finish1,p)=A%TWij(start1:finish1,p)+&
!                         CONJG(PDOT(start:finish))*PDOT(p+Base)
!                  enddo
!              enddo 
!             enddo !Band
!             !write(Log_unit,*) 'Completed cluster',j,degen,Occ
!             Do t=1,Specific_Atoms
!                Atom_List(t)%TWij= Atom_List(t)%TWij/Degen
!             Enddo
!
!
!               ! Accumulate Wijkl terms
!               Do t=1,Specific_Atoms
!                  A => Atom_List(t)
!                  AT => AtomType_Info(A%TypeIndex)
!                  Do p=1,AT%Cijkl_Size
!                     Call Coulomb_Decode(AT%LUT_Cijkl(p),nili,njlj,nklk,nlll,&
!                          mi, mj, mk)
!                     ml = mi - mj + mk
!                     in=AT%nl_base(nili)+AT%L_Value(nili)+mi
!                     jn=AT%nl_base(njlj)+AT%L_Value(njlj)+mj
!                     kn=AT%nl_base(nklk)+AT%L_Value(nklk)+mk
!                     ln=AT%nl_base(nlll)+AT%L_Value(nlll)+ml
!                     AT%Wijkl(p)=AT%Wijkl(p)+(Occ)*A%TWij(in,jn)*A%TWij(kn,ln)
!                     !write(Log_unit,*) 'Wijkl:',t,p,AT%Wijkl(p)
!                  Enddo
!               Enddo
!
!      enddo   !Clusters
!    endif
!   enddo   !Kpnt
!
!   Do t=1,Atom_Types
!      AtomType_Info(t)%Wijkl=AtomType_Info(t)%Nofthistype*AtomType_Info(t)%Wijkl
!      write(Log_Unit,*) ' Self energy weight for ', t,AtomType_Info(t)%Atom_Name
!      i=MaxLoc(ABS(AtomType_Info(t)%Wijkl(:)),1)  
!      Call Coulomb_Decode(AtomType_Info(t)%LUT_Cijkl(i),nili,njlj,nklk,nlll,&
!                          mi, mj, mk)
!      ml = mi - mj + mk
!      x=ABS(AtomType_Info(t)%Wijkl(i))
!      write(Log_unit,'(9i6,1pe15.7)') i,nili,njlj,nklk,nlll,mi,mj,mk,ml,x
!      Do j=1,AtomType_Info(t)%Cijkl_Size
!         if (j/=i.and.ABS(AtomType_Info(t)%Wijkl(j))>x/4) then
!            Call Coulomb_Decode(AtomType_Info(t)%LUT_Cijkl(j),&
!                   nili,njlj,nklk,nlll, mi, mj, mk)
!            ml = mi - mj + mk
!            write(Log_unit,'(9i6,1pe15.7)') j,nili,njlj,nklk,nlll,mi,mj,mk,ml,&
!                 ABS(AtomType_Info(t)%Wijkl(j))
!         endif
!     enddo
!   enddo
!
!   DeAllocate(Psi ,Clusterinit,Clusterend)
!
!End Subroutine  AccumWijkl
!
!******************************************************************************
!
!  AccumShift - Accumulates the coulomb shift term eq A27
!
!******************************************************************************

Subroutine AccumShift
  
  Integer        :: i, j, k, Qlm_Size, Offset_Q, ibase,jbase
  Integer        :: nili, njlj, mi, mj, M, L
  Integer, Pointer :: LUT_Orb(:), nlm_LUT(:,:), nl_Base(:)
  Real, Pointer  :: aQlm(:)
  Complex, Pointer :: dEdQlm(:,:), Dij(:,:), Dijspin(:,:)
  Type (Specific_Atom), Pointer :: A
  Type (Atom_Info_Fixed), Pointer :: AT
  
  Do i=1, Specific_Atoms
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)

     Offset_Q = AT%Hat_MaxL + 1

     aQlm => AT%aQlm
     LUT_Orb => AT%LUT_Orb
     Qlm_Size = AT%QVlm_Size
     dEdQlm => A%dEdQlm
     nl_base => AT%nl_Base
     nlm_LUT => AT%nlm_LUT
     Dij => A%Dij
     if(spindependence) Dijspin => A%Dijspin

     Do j=1, Qlm_Size
        Call Orbital_Decode(LUT_Orb(j), nili, njlj, mi, mj, L)
        M = mj - mi

        ibase = nl_Base(nili); ibase = ibase + nlm_lut(2,ibase) + mi;
        jbase = nl_Base(njlj); jbase = jbase + nlm_lut(2,jbase) + mj;

        Dij(ibase,jbase) = Dij(ibase,jbase) + aQlm(j) * dEdQlm(L+1, M+Offset_Q)
        if(spindependence) Dijspin(ibase, jbase) = Dijspin(ibase, jbase) + &
                           aQlm(j) * dEdQlm(L+1, M+Offset_Q)
     End Do
  End Do

  Return
End Subroutine

!******************************************************************************
!
!  AddDiagElements - Adds the Diagonal Elements to the Dij for each atom
!                    eq A10, A11
!
!******************************************************************************

Subroutine AddDiagElements(Energy)
   Real, Intent(INOUT) :: Energy

   Integer :: i, j, k, t, Basis_Size, Overlap_Size
   Integer :: nili, njlj, m, mmin, mmax, Li, Lj, ibase,jbase
   Complex, Pointer :: Dij(:,:), Wij(:,:),Dijspin(:,:), Wijspin(:,:)
   Real, Pointer :: Vion(:), KE(:)
   Integer, Pointer :: L_Values(:), nlm_LUT(:,:),nl_base(:)
   Complex       :: fn
   Integer, Pointer :: LUT(:)

   Type (Specific_Atom), Pointer :: A
   Type (Atom_Info_Fixed), Pointer :: AT
 
   Do i=1, Specific_Atoms
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)
     Dij => A%Dij
     Wij => A%Wij
   if(spindependence) then
     Dijspin => A%Dijspin
     Wijspin => A%Wijspin
   end if
     
     Vion => AT%V_ion
     KE => AT%Kinetic
     Basis_Size = AT%Basis_Size
     L_Values => AT%L_Value
     nl_base => AT%nl_Base
     nlm_LUT => AT%nlm_LUT

     t = 0
     Do nili=1, Basis_Size
        Li = L_Values(nili)
        ibase = nl_Base(nili); ibase = ibase + nlm_lut(2,ibase);

        Do njlj=1, nili
           Lj = L_Values(njlj)

           If (Li == Lj) then
              t = t + 1
              jbase = nl_Base(njlj); jbase = jbase + nlm_lut(2,jbase);

              fn = KE(t) + Vion(t)

              Do m=-Li, Li
                 Dij(ibase+m,jbase+m) = Dij(ibase+m,jbase+m) + Fn
                 Energy = Energy+Wij(ibase+m,jbase+m)*Fn
                 if(spindependence) then
                    Dijspin(ibase+m,jbase+m) = Dijspin(ibase+m,jbase+m) + Fn
                    Energy = Energy+Wijspin(ibase+m,jbase+m)*Fn
                 end if
                 
                 if (nili /= njlj) then
                    Dij(jbase+m,ibase+m) = Dij(jbase+m,ibase+m) + Fn
                    Energy = Energy+Wij(jbase+m,ibase+m)*Fn   
                    if(spindependence) then
                       Dijspin(jbase+m,ibase+m) = Dijspin(jbase+m,ibase+m) + Fn
                       Energy = Energy+Wijspin(jbase+m,ibase+m)*Fn
                    end if   
                 End If
              End Do
           End If
        End Do
     End Do
  End Do

  Return
End Subroutine

!******************************************************************************
! LoadWij - Temp for comparison
!******************************************************************************
! It is not updated for spin dependence

Subroutine LoadWij
   Integer :: i,j,k, nili,njlj,mi,mj,li,lj, ikind, ia, basis_size
   Complex :: c
   Integer, pointer :: nl_base(:), nlm_lut(:,:)
   Complex, Pointer :: Dij(:,:), Wij(:,:)
   Type (Specific_Atom), Pointer :: A
   Type (Atom_Info_Fixed), Pointer :: AT

   Open(50, file="wij.out", recl=1000)

write(*,*) 'LoadWij:!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
   Do i=1, Specific_Atoms
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)
     nl_base => AT%nl_Base
     nlm_LUT => AT%nlm_LUT
     Wij => A%Wij
     Dij => A%Dij

     Basis_Size = AT%Basis_Size

     Do j=1, AT%nlm_Size
        Do k=1, AT%nlm_Size
           Read(50,*) ikind,ia,nili,Li,mi,njlj,Lj,mj,c

           Wij(nl_base(nili)+mi+Li,nl_base(njlj)+mj+Lj) = c
        End Do
     End Do

  End Do

  Close(50)
  Return
End Subroutine

!******************************************************************************
! WriteWij - Temp for comparison
!******************************************************************************
!It is not updated for spin dependence
Subroutine WriteWij(fd)
   Integer, Intent(IN) :: fd

   Integer :: i,j,k, nili,njlj,mi,mj,li,lj, ikind, ia, Basis_size
   Complex :: c
   Integer, pointer :: nl_base(:), nlm_lut(:,:)
   Complex, Pointer :: Dij(:,:), Wij(:,:)
   Type (Specific_Atom), Pointer :: A
   Type (Atom_Info_Fixed), Pointer :: AT

!   Open(50, file="wij.out", recl=1000)

   Write(fd,*) 'Printing Wij (nili,Li,mi, njlj,Lj,mj, Wij)********************'
   Do i=1, Specific_Atoms
      Write(fd,*) 'ATOM:',i, '------------------------------------------------'
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)
     nl_base => AT%nl_Base
     nlm_LUT => AT%nlm_LUT
     Wij => A%Wij
     Dij => A%Dij

     Basis_Size = AT%Basis_Size

     Do j=1, AT%nlm_Size
        nili = nlm_LUT(1,j)
        Li = nlm_LUT(2,j)
        mi = nlm_LUT(3,j)

        Do k=1, AT%nlm_Size
           njlj = nlm_LUT(1,k)
           Lj = nlm_LUT(2,k)
           mj = nlm_LUT(3,k)

           c = Wij(nl_base(nili)+mi+Li, nl_base(njlj)+Mj+Lj)
!**           c = c/(4*PI/Xtal%Volume)
           If (ABS(c*COnjg(c))>1E-20) then
              Write(fd,'(6i3)', ADVANCE="NO") nili,Li,mi,njlj,Lj,mj
              Write(fd,*) '(i,j)',nl_base(nili)+mi+Li, nl_base(njlj)+mj+Lj
              Write(fd,*) c
           End If

        End Do
     End Do
  End Do
   Write(fd,*) '******************************************************'
!**     Close(50)
  Return
End Subroutine

!******************************************************************************
! WriteDij - Temp for comparison
!******************************************************************************

Subroutine WriteDij(fd)
   Integer, Intent(IN) :: fd

   Integer :: i,j,k, nili,njlj,mi,mj,li,lj, ikind, ia, Basis_size
   Complex :: c
   Integer, pointer :: nl_base(:), nlm_lut(:,:)
   Complex, Pointer :: Dij(:,:), Wij(:,:)
   Type (Specific_Atom), Pointer :: A
   Type (Atom_Info_Fixed), Pointer :: AT

!   Open(50, file="wij.out", recl=1000)

  Write(fd,*) 'Printing Dij atom (nili,Li,mi, njlj,Lj,mj, Dij)****************'
   Do i=1, Specific_Atoms
      Write(fd,*) 'ATOM:',i, '------------------------------------------------'
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)
     nl_base => AT%nl_Base
     nlm_LUT => AT%nlm_LUT
     Wij => A%Wij
     Dij => A%Dij

     Basis_Size = AT%Basis_Size

     Do j=1, AT%nlm_Size
        nili = nlm_LUT(1,j)
        Li = nlm_LUT(2,j)
        mi = nlm_LUT(3,j)

        Do k=1, AT%nlm_Size
           njlj = nlm_LUT(1,k)
           Lj = nlm_LUT(2,k)
           mj = nlm_LUT(3,k)

           c = Dij(nl_base(nili)+mi+Li, nl_base(njlj)+Mj+Lj)
!**           c = c/(4*PI/Xtal%Volume)
           If (ABS(c*COnjg(c))>1E-20) then
              Write(fd,*) 'Dij=',c
              Write(fd,*) '(i,j)',nl_base(nili)+mi+Li, nl_base(njlj)+mj+Lj
              Write(fd,*) '     ',nili,Li,mi,njlj,Lj,mj
           End If

        End Do
     End Do
  End Do
   Write(fd,*) '******************************************************'
!**     Close(50)
  Return
End Subroutine

!******************************************************************************
! WriteQlm - Temp for comparison
!******************************************************************************

Subroutine WriteQlm(Which, fd)
  Integer, Intent(IN) :: Which
   Integer, Intent(IN) :: fd

   Integer :: i,j,k, M,l, ikind, ia, Offset_Q
   Complex :: c, QlmSum, totalQlm
   Complex, Pointer :: Qlm(:,:)
   Integer, pointer :: nl_base(:), nlm_lut(:,:)
   Complex, Pointer :: Dij(:,:), Wij(:,:)
   Type (Specific_Atom), Pointer :: A
   Type (Atom_Info_Fixed), Pointer :: AT

!   Open(50, file="wij.out", recl=1000)


   totalQlm = 0
   If (Which==1) then
      Write(fd,*) 'Printing Qlm (atom, L, M, Qlm)********************'
   Else
      Write(fd,*) 'Printing dEdQlm (atom, L, M, Qlm)********************'
   End If

   Do i=1, Specific_Atoms
      Write(fd,*) 'ATOM:',i, '------------------------------------------------'
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)
     nl_base => AT%nl_Base
     nlm_LUT => AT%nlm_LUT

     If (Which==1) then
        Qlm => A%Qlm
     else
        Qlm => A%dEdQlm
     End If

     Offset_Q = 2*MaxVal(AT%L_Value) + 1

     QlmSum = 0
     Do L=0, Offset_Q-1
        Do m=-L, L
           c = Qlm(L+1, Offset_Q + m)
           Qlmsum = QlmSum + c
           If (ABS(c*COnjg(c))>1E-5) then
              Write(fd,*) L,M,c
           End If
        End Do
     End Do

     Write(fd,*) ' SUM(Qlm)=',QlmSum
     TotalQlm = TotalQlm + QlmSum
  End Do
  Write(fd,*) '******************************************************'
  Write(fd,*) ' Total SUM(Qlm)=',TotalQlm
  Write(fd,*) '******************************************************'

!**     Close(50)
  Return
End Subroutine


!******************************************************************************
!
!  MakeBandMap - Creates the mapping from diskindex to (Band,Kpnt)
!
!******************************************************************************

Subroutine MakeBandMap
  !Integer :: BandsPerKpnt(NumKpnts), i, j, k
  !Integer :: LUT(Mem_MapSize),  ListSize, SortLUT(Mem_MapSize)
  !Real :: Eigen(Mem_MapSize)
  Integer ::  i, j, k , ListSize
  Integer ,allocatable :: LUT(:), SortLUT(:), BandsPerKpnt(:)
  Real ,allocatable:: Eigen(:)

  !write(log_unit,*) 'MakeBandMap: start', NumKpnts,Numbands
  !call flush(log_unit)

  Allocate(BandsPerKpnt(NumKpnts),LUT(Mem_MapSize),SortLUT(Mem_MapSize), &
            Eigen(Mem_MapSize))

  BandMap = 0
  BKMap = 0
if(.NOT.spindependence) then
  !write(log_unit,*) 'MakeBandMap: NOT spin dependent'
  !call flush(log_unit)
  BandsPerKpnt = 0;LUT=0;SortLUT=0;Eigen=1.e30

  ListSize = 0         !*** Create LUT and Eigenvalue list
  Do i=1, Mem_MapSize
     If (PsiInfo(i)%Kpnt > 0) then
        ListSize = ListSize + 1
        LUT(ListSize) = i
        Eigen(ListSize) = PsiInfo(i)%Energy
     End if
  End Do

           !** Sort the eigenvalues ***
  Call Insertion_Sort(Eigen(1:ListSize), SortLUT(1:ListSize), .TRUE.)  


  Do i=1, ListSize
     j = LUT(SortLUT(i))
     If (PsiInfo(j)%Kpnt > 0) then
        k = PsiInfo(j)%Kpnt
        If (BandsPerKpnt(k) < NumBands) then
           BandsPerKpnt(k) = BandsPerKpnt(k) + 1
           BandMap(1,j) = BandsPerKpnt(k)
           BandMap(2,j) = k
           BKMap(BandsPerKpnt(k),k)=j
           
           !write(Log_unit,*) 'MakeBandMap', j,BandMap(1,j),BandMap(2,j)
           !call flush(log_unit)
        End If
     End If
  End Do
else 
  !write(log_unit,*) 'MakeBandMap: spin dependent'
  !call flush(log_unit)

  BandsPerKpnt = 0;LUT=0;SortLUT=0;Eigen=1.e30
  ListSize = 0         !*** Create LUT and Eigenvalue list
  Do i=1, Mem_MapSize
     If (PsiInfo(i)%Kpnt > 0 .and. PsiInfo(i)%spinup) then
        ListSize = ListSize + 1
        LUT(ListSize) = i
        Eigen(ListSize) = PsiInfo(i)%Energy
     End if
  End Do

           !** Sort the eigenvalues ***
  Call Insertion_Sort(Eigen(1:ListSize), SortLUT(1:ListSize), .TRUE.)  


  Do i=1, ListSize
     j = LUT(SortLUT(i))
     If (PsiInfo(j)%Kpnt > 0 .and. PsiInfo(j)%spinup) then
        k = PsiInfo(j)%Kpnt
        If (BandsPerKpnt(k) < NumBands) then
           BandsPerKpnt(k) = BandsPerKpnt(k) + 1
           BandMap(1,j) = BandsPerKpnt(k)
           BandMap(2,j) = k
           BKMap(BandsPerKpnt(k),k)=j
           !write(Log_unit,*) 'MakeBandMap', j,BandMap(1,j),BandMap(2,j)
           !call flush(log_unit)
        End If
     End If
  End Do
  BandsPerKpnt = 0;LUT=0;SortLUT=0;Eigen=1.e30

  ListSize = 0         !*** Create LUT and Eigenvalue list
  Do i=1, Mem_MapSize
     If (PsiInfo(i)%Kpnt > 0 .and. .NOT.PsiInfo(i)%spinup) then
        ListSize = ListSize + 1
        LUT(ListSize) = i
        Eigen(ListSize) = PsiInfo(i)%Energy
     End if
  End Do

           !** Sort the eigenvalues ***
  Call Insertion_Sort(Eigen(1:ListSize), SortLUT(1:ListSize), .TRUE.)  


  Do i=1, ListSize
     j = LUT(SortLUT(i))
     If (PsiInfo(j)%Kpnt > 0 .and. .NOT.PsiInfo(j)%spinup) then
        k = PsiInfo(j)%Kpnt
        If (BandsPerKpnt(k) < NumBands) then
           BandsPerKpnt(k) = BandsPerKpnt(k) + 1
           BandMap(1,j) = BandsPerKpnt(k)+NumBands
           BandMap(2,j) = k
           BKMap(BandsPerKpnt(k)+NumBands,k)=j
           !write(Log_unit,*) 'MakeBandMap', j,BandMap(1,j),BandMap(2,j)
           !call flush(log_unit)
        End If
     End If
  End Do
end if
  DeAllocate(BandsPerKpnt,LUT,SortLUT,Eigen)

  Return
End Subroutine

!******************************************************************************
!
!  CalcNewOccAndFermi - Calculates the New occupancies and Fermi Energy
!            for the system.
!
! ***  NOTE:  This routine assumes the MakeBandMap has been called! ***
!
! Also update virtual memory list
!******************************************************************************

Subroutine CalcNewOccAndFermi
  Real :: OldFermi, NewFermi, NewWgt,x
  Integer :: i, band, Kpnt, temp,ind(1),s

  If (Atomic_Mode == SIM_SEPM) RETURN
  s=1
  if(spindependence) s=2

  !write(log_unit,*) 'CalcNewOccAndFermi: start', 10*MaxVal(PsiInfo(:)%Energy)
  !call flush(log_unit)

  BZ%KEu(:,:) = 10*MaxVal(PsiInfo(:)%Energy) + 150

  Do i=1, Mem_MapSize
     Band = BandMap(1,i)
     Kpnt = BandMap(2,i)

     If ((Kpnt /= 0) .AND. (Band <= s*NumBands)) then
        BZ%KEu(Band, Kpnt) = PsiInfo(i)%Energy
        write(log_unit,*) 'CalcNewOccAndFermi: loading', Band,Kpnt,BZ%KEu(Band, Kpnt)
        call flush(log_unit)
        PsiInfo(i)%DoSave=1
        PsiInfo(i)%Available=Mem_Used
     Else
        PsiInfo(i)%Kpnt=0
        PsiInfo(i)%spinup=.true.
        PsiInfo(i)%DoSave=0
        PsiInfo(i)%Available=Mem_Available
        temp=PsiInfo(i)%MemBufIndex
        if (temp>0) then
           Globalmap%Psi_Handle(temp)%Index=0
           PsiInfo(i)%MemBufIndex=0
        EndIf
     End If
  End Do

  If  (.not.bandstructure_mode) then  !Keep Fermi level if bandstructure calc
    OldFermi = BZ%Fermi
    NewFermi = BZ_CalcFermi(BZ)       !*** Calculate the new fermi energy
    Write(Log_Unit,*) &
         'CalcNewOccAndFermi: NewFermi = ',NewFermi,' OldFermi = ',OldFermi
    call flush(log_unit)
  EndIf

    If (bandstructure_mode) &
          Write(Log_Unit,*) 'CalcNewOccAndFermi: Keeping OldFermi = ',BZ%Fermi
    
  If  (ReCalculate_Occupancies) then  !Recalculate occupancies                     
    Call BZ_Occupy(BZ, Occupancy)     !*** Occupy the bands
  
    NewWgt = 1.0
    x=0
    Do i=1, Mem_MapSize      !*** Now Map the new occupations back ***
       If (PsiInfo(i)%Available==Mem_Used) then
          Band = BandMap(1,i)
          Kpnt = BandMap(2,i)
          PsiInfo(i)%Occupancy = (1.0-NewWgt)*PsiInfo(i)%Occupancy + &
                                 NewWgt*Occupancy(Band, Kpnt)
          !write(log_unit,*) 'CalcNewOccAndFermi:',Band,Kpnt,PsiInfo(i)%Energy,PsiInfo(i)%Occupancy
          x=x+PsiInfo(i)%Occupancy
       else
          PsiInfo(i)%Occupancy = 0
       End If
    End Do
  EndIf
  write(log_unit,*) 'CalcNewOccAndFermi: ', x
  call flush(log_unit)
  Return
End Subroutine

!******************************************************************************
!
!  CalcSmoothH - Calculates the smooth density and stores it in V%FAS
!           and also accumulates the energy for the smooth psi to H.
!
!           The KE and Exc for psi~ are ONLY calculated. The terms involving
!           n^ can't be calculated yet since no Wij's and Qlm's are known. 
!
!
!  Both Den and WK1 are LARGE arrays(G_HIGH)
!
!******************************************************************************

Subroutine CalcSmoothH( ProjDotPsi, Den, WK1, Denspin)
   Complex,  Intent(OUT) :: ProjDotPsi(:,:,:)
   Complex,  Intent(OUT), OPTIONAL ::Denspin(:)
   Complex,  Intent(OUT) :: Den(:)
   Complex,  Intent(OUT) :: WK1(:)

   Real     :: G(3), G2, BlochK(3), KE, TotOcc, norm1,norm2
   Type (Mem_handle), Pointer :: Psi
  !Integer          :: Psi_toUse(Mem_MapSize), Kpnt, PsiBand, PsiK,  i
   Integer          ::  Kpnt, PsiBand, PsiK,  i
   Integer  ,allocatable  :: Psi_toUse(:)
   Integer          :: gi, N, j,k, t,Flag
   Complex, Pointer :: Comp(:), Ocomp(:), Work(:), Ox(:)
  !Complex          :: C_PDOT(PLM_MAx)
   Complex  ,allocatable :: C_PDOT(:)
   Logical          :: Stored

   write(log_unit,*) 'CalcSmoothH: begin'
   call flush(log_unit)
   Call Start_Timer(Timer(CalcSmoothH_Timer))

   Allocate(Psi_toUse(Mem_MapSize),C_PDOT(PLM_MAx))
   N = Gpnt_Size(G_LOW) - 1
   Flag=1
   
   ProjDotPSi = 0.0

   den = 0
   if(spindependence .and. .NOT.present(Denspin)) then 
      write(error_unit,*) 'Denspin is not correctly input'
      stop
   endif

   if(present(Denspin)) Denspin=0
 
!Write(Log_Unit,*) 'CalcSmoothH:  * Occ=',PsiInfo(:)%Occupancy
!call flush(log_unit)
!   PsiInfo(:)%PDOT_Stored = .FALSE.  !** Clear PDOT Info

!--- This part taken out by NAWH 6-24-00
!--   !*** Add neutralizing charge if needed ***
!--   If (ABS(NetCharge)>1E-6) then
!--      Stored = .FALSE.
!--      Call GetBuffer( Comp)
!--      Call GetBuffer( OComp)
!--      Comp = 0
!--      Comp(1) = 1
!--      Call OPsi( Comp, Ocomp, C_PDOT, 1, Stored)
!--      ke = DOT_PRODUCT(Comp,Ocomp)
!--Write(Error_Unit, *) 'CalcSmoothH: C_DOT=',ke
!--      Comp = Comp / SQRT(ke)
!--      WK1 = 0
!--
!--      WK1(FFTMap_High(1:Gpnt_Size(G_Low))) = Comp(1:Gpnt_Size(G_Low))
!--      t = Gpnt_Size(G_Low)+1
!--      j = Gpnt_Size(G_HIGH)+1;   k = j + (Gpnt_Size(G_Low)-1) - 1
!--      WK1(FFTMap_High(j:k)) = Comp(t:)
!--
!--WK1 = 0
!--WK1(FFTMap_High(1)) = NetCharge         
!--      Call PerformFFT(FFT_TO_R, G_HIGH, WK1)
!--         
!--      Den = Den + NetCharge*CONJG(WK1)*WK1
!--      
!--      Call FreeBuffer(Comp)
!--      Call FreeBuffer(Ocomp)
!--   End If

   TotOcc = 0
   Do i=1, NumKpnts
      BlochK = BZ%Ku(:,i)
      !write(log_unit,*) 'CalcSmoothH: BlochK = ',i,BlochK

      Psi_toUse = MH_Skip
      Where ((PsiInfo(:)%Kpnt == i) .AND. (PsiInfo(:)%Occupancy >1E-10))
         Psi_toUse = MH_toProcess
      End Where
      !write(log_unit,*) 'CalcSmoothH: i',i
      ! do j=1,Mem_mapsize
      !    if (Psi_toUse(j) /= 0) then
      !       write(log_unit,*) 'wfn ',j,PsiInfo(j)%Energy ,PsiInfo(j)%Occupancy 
      !    endif
      ! enddo       
      Call Phase_Generic( Psi_toUse,  i)
      Call GetNextPsi(Flag, Psi)

      Do While (Flag > 0)
!Write(Log_Unit,*) 'CalcSmoothH: Flag=',Flag, ' * DI=',Psi%Index, ' * Occ=', PsiInfo(Psi%Index)%Occupancy
!call flush(log_unit)
         TotOcc = TotOcc + PsiInfo(Psi%Index)%Occupancy
         !write(log_unit,*) 'Occ ', TotOcc
         !call flush(log_unit)

         Call FilterValue(Psi%Ptr)
         call kplusGfilter(Psi%Ptr,i)

         !*** Calculate the KE for this Psi ***
         KE = CalcKineticEnergy(BlochK, Psi%Ptr)
         PsiInfo(Psi%Index)%KE = KE
         PsiInfo(Psi%Index)%KE_Stored = .TRUE. 

         If (Atomic_Mode == SIM_TIME) then
            Call GetBuffer( Work)
            Call  CalcDEL2_TIME( BlochK, Psi%Ptr, Work)
            KE = DOT_PRODUCT(Psi%Ptr,Work)
            Call FreeBuffer(Work)
         End if

         TildeKE_Energy = TildeKE_Energy + PsiInfo(Psi%Index)%Occupancy*KE

         WK1=0
         !*** Now accumulate the density ***
         Call V_G_to_r(Psi%Ptr,G_LOW,WK1,G_HIGH)
         !write(log_unit,*) 'SmoothH: wfn = ', Psi%Index,PsiInfo(Psi%Index)%spinup,PsiInfo(Psi%Index)%Occupancy
         !call flush(log_unit)
         if(spindependence .and. .NOT.PsiInfo(Psi%Index)%spinup) then
           Denspin = Denspin + PsiInfo(Psi%Index)%Occupancy * CONJG(WK1)*WK1
         else
           Den = Den + PsiInfo(Psi%Index)%Occupancy * CONJG(WK1)*WK1
         endif

!write(Log_Unit,*)'CalcSmoothH',PsiInfo(Psi%Index)%Occupancy
!write(Log_Unit,*) 'den',den(1),den(2),den(3)
!call flush(log_unit)
!write(Log_Unit,*) 'CalcSmoothH: sum(den) = ',sum(den)*xtal%Volume/PRODUCT(FFT_Grid(1:3,G_HIGH))
         
         !*** Finally calculate the PDOTS for Wij ***
         PsiBand = BandMap(1, Psi%Index)
         PsiK = BandMap(2,Psi%Index)
!Write(Log_Unit,*) 'PsiBand, PsiK=',PsiBand, PsiK, ' * NumBands=',NumBands
!call flush(Log_unit)
         If (.NOT. PsiInfo(Psi%Index)%PDOT_Stored) then 
            If (Proj_Mode == PROJ_RS) then
               WK1 = 0
               Call RS_toR(Psi%Ptr, WK1)
               Call CalcProjProducts(WK1, &
                    PsiInfo(Psi%Index)%PDOT)
            else
               Call CalcProjProducts(Psi%Ptr, PsiInfo(Psi%Index)%PDOT)
            End If

            PsiInfo(Psi%Index)%PDOT_Stored = .TRUE.
         End If


         ProjDotPsi(:,PsiBand, PsiK) = PsiInfo(Psi%Index)%PDOT

!write(log_unit,*) 'CalcSmoothH: PDOT index  = ', PsiBand,PsiK
!call flush(log_unit)
!Write(Log_Unit,*) 'CalcSmoothH: PDOT=',ProjDotPsi(:,PsiBand, PsiK)
!call flush(log_unit)
         
         Call GetNextPsi(Flag, Psi)
      End Do
   End Do

  Write(Log_Unit,*) 'CalcSmoothH: SUM(Occ)=',TotOcc 
  call flush(log_unit)
  !Den = Den * FFT_Grid(4,G_HIGH)   ! correct for FFT normalization

   DeAllocate(Psi_toUse,C_PDOT)
   Call Stop_Timer(Timer(CalcSmoothH_Timer))
!STOP
   Return
End Subroutine


!******************************************************************************
!
! CalcHam - Calculates the Hamiltonian
!
!******************************************************************************

Subroutine CalcHam( Energy)
   Real,     Intent(OUT) :: Energy

   Integer :: i,j,k, PsiBand, PsiK,  n, t
   Integer          :: den_handle, psi_R_Handle, Psi_G_Handle, KE_Handle
   Complex, Pointer :: BlochK(:), Wk1(:), Wk2(:)
   Real             :: P(3), rmag, Offset(3), energytemp, toosmall
   Real     :: Qt, Qh, c1, c2, gmag
   Complex, Pointer :: OldDen(:), Work1(:), Work2(:), Fn(:), DenTemp(:)
   Complex, Allocatable :: ProjDotPsi(:,:,:)
   Type (Mem_handle), Pointer :: Psi, FAS
  !Integer          :: Psi_toUse(Mem_MapSize), Kpnt, atom, LL, LG, Pin(3)
   Integer          :: Kpnt, atom, LL, LG, Pin(3)
   Integer ,allocatable  :: Psi_toUse(:)
   Complex, Pointer :: Den(:), Vwork(:), Denspin(:)
   Real :: Vxc, Exc, ExcSum
   Logical :: Ok, DoZero, DoPrint, DoLoadWij
   Character*100 :: msg
   
   write(log_unit,*) 'CalcHam: begin'
   call flush(log_unit)

   Allocate(Psi_toUse(Mem_MapSize))
   energytemp = 0
   If (Prev_cohesiveEnergy(1) < 1E20) then
      energytemp = Prev_CohesiveEnergy(2) - Prev_CohesiveEnergy(1)
      gmag = 0.5*(abs(Prev_CohesiveEnergy(2)) + abs(Prev_CohesiveEnergy(1)))
      energyTemp = abs(energytemp / gmag)
!      energyTemp = abs(energytemp / Prev_CohesiveEnergy(2))
      If (energytemp > Mix_DampRelative) then
Write(Log_Unit,*) 'DAMP'
         SCFvalues%AC%NewMix = V_DampNewMix
         If(spindependence)SCFvalues%ACspin%NewMix = V_DampNewMix
         Do i=1, Specific_Atoms
            Atom_List(i)%Dij_AC%NewMix = Dij_DampNewMix
            If(spindependence)Atom_List(i)%Dijspin_AC%NewMix = Dij_DampNewMix
         End Do
      else
Write(Log_Unit,*) 'RESTORE'
         SCFvalues%AC%NewMix = V_NewMix
         If(spindependence)SCFvalues%ACspin%NewMix = V_NewMix
         Do i=1, Specific_Atoms
            Atom_list(i)%Dij_AC%NewMix = Dij_NewMix
            If(spindependence)Atom_list(i)%Dijspin_AC%NewMix = Dij_NewMix
         End do
      End if
   End If

   Write(Log_unit,*) 'CalcHam: Echange=',energytemp   
   Write(Log_unit,*) 'CalcHam: V_NewMix=',SCFvalues%AC%NewMix
   Write(Log_unit,*) 'CalcHam: Dij_NewMix=',Atom_list(1)%Dij_AC%NewMix, ' *',Dij_NewMix
   call flush(Log_unit)

   Call Start_Timer(Timer(CalcHam_Timer))
   Call PrintDate(Log_Unit, 'CalcHam: Start of Routine!')   
!Energy = 0
!Do i=1, Specific_Atoms
! Energy = Energy +0.5*SUM(Atom_List(i)%Pos**2)
!End Do
!RETURN

DoPrint = .FALSE.
DoLoadWij = .FALSE.
DoZero = .FALSE.


   Diag_Energy = 0
   Hat_Energy = 0
   OneExc_Energy = 0
   Hartree_Energy = 0
   !Auxform_of_Hartree = 0
   TildeKE_Energy = 0
   TildePot_Energy = 0
   TildeXC_Energy = 0
   Hat_Hat_Energy = 0
   !Self_Energy_Correction = 0

   Energy = 0

  !!!In order to save memory for self-energy correction
  !!!  ProjDotPsi allocated later 
   !*** Allocate the Large FFT and ProjDotPsi arrays ***
   i = FFT_Grid(4,G_HIGH)
   if(.NOT.spindependence) then
      !Allocate(Den(i), ProjDotPsi(PLM_MAX, NumBands,NumKpnts), STAT=j)
      Allocate(Den(i),  STAT=j)
      msg = 'CalcHam: Could Not allocate LARGE buffers!'
      Call Check_Error(j, msg, Error_Unit, .TRUE.,paw_wc, "CalcHam:")
   else
      !Allocate(Den(i),Denspin(i), ProjDotPsi(PLM_MAX, 2*NumBands,NumKpnts), STAT=j)
      Allocate(Den(i),Denspin(i),  STAT=j)
      msg = 'CalcHam: Could Not allocate LARGE buffers!'
      Call Check_Error(j, msg, Error_Unit, .TRUE.,paw_wc, "CalcHam:")
   endif
   !****************

   Vwork =>SCFvalues%Work
If (DoPrint) Write(*,*) 'CalcHam: Start of routine !!!!!!!!!!!!!!!!!!!!!!!!!!'
   !if(Estimate_selfenergy) Call AccumSmoothSelf(Vwork)
   !if(Estimate_scselfenergy) Call AccumWijkl
     

 ! Large arrays for smooth self-energy  no longer needed -- allocate ProjDotPsi
   if(.NOT.spindependence) then
      Allocate(ProjDotPsi(PLM_MAX, NumBands,NumKpnts), STAT=j)
      msg = 'CalcHam: Could Not allocate LARGE buffers!'
      Call Check_Error(j, msg, Error_Unit, .TRUE.,paw_wc, "CalcHam:")
   else
      Allocate(ProjDotPsi(PLM_MAX, 2*NumBands,NumKpnts), STAT=j)
      msg = 'CalcHam: Could Not allocate LARGE buffers!'
      Call Check_Error(j, msg, Error_Unit, .TRUE.,paw_wc, "CalcHam:")
   endif
   !****************

  if(.NOT.spindependence) Call CalcSmoothH( ProjDotPsi, Den, Vwork)
  if(spindependence) Call CalcSmoothH( ProjDotPsi, Den, Vwork, Denspin)

write(Log_Unit,*) 'totaldensity after calcsmooth',sum(den)
call flush(log_unit)
If (xtal%Rot_Size > 0) Call SymRho( Den, Vwork)
If (xtal%Rot_Size > 0 .and. spindependence) Call SymRho( Denspin, Vwork)
write(Log_Unit,*) 'totaldensity after symrho ',sum(den)
call flush(log_unit)

   If (Atomic_Mode /= SIM_AE) then
If (DoPrint) Write(*,*) 'CalcHam: Before Wij!!!!!!!!!!!!!!!!!!!'
      Do i=1, Specific_Atoms
         Atom_List(i)%Wij_Old = Atom_List(i)%Wij
         Atom_List(i)%Dij = 0        !** Init Dij matrix
         Atom_List(i)%Qlm = 0
         Atom_List(i)%dEdQlm = 0
      End Do
    if(spindependence) then
      Do i=1, Specific_Atoms
         Atom_List(i)%Wij_Oldspin = Atom_List(i)%Wijspin
         Atom_List(i)%Dijspin = 0        !** Init Dij matrix
      End Do
   end if

      Call CalcWij(ProjDotPsi)    !** Calculate and the Wij's (eq A19)
!!! The following is now performed in AccumSmoothSelf
     !If (Estimate_selfenergy) &
     ! Call CalcWijkl(ProjDotPsi)  !** Calculate weights for self energy

!Call WriteWij(Log_Unit)
If (DoPrint) then
call writewij(6)
End If
If (DoLoadWij) Call LoadWij

!STOP

If (DoPrint) then
   Write(Log_Unit,*) 'CalcHam: Before SymWij.......................'
   Call WriteWij(Log_Unit)
End If
      If (xtal%Rot_Size > 0)  Call SymWij        !** Symmetrize the Wij's
      !If (Estimate_selfenergy.and.xtal%Rot_Size > 0)  Call SymWijkl 

write(log_unit,*) 'CalcHam: After SymWij'
call flush(log_unit)

If (DoPrint) then
   Write(*,*) 'CAlcHam: Before WriteSymWij!!!!!!!!!!!!!!!!!!!!!!!!!!'
   call writewij(6)
   Write(*,*) 'CalcHam: After Wij!!!!!!!!!!!!!!!!!!!'
   !STOP
End If

energytemp = hat_energy
     Call CalcQlm(Hat_Energy)           !** Calculate the Qlm's     (eq A22)
energytemp = hat_energy - energytemp
If (DoPrint) then
  Write(*,*) 'CalcHam: E^ energy =',energytemp
  Write(*,*) 'CalcHam: After CalcQLM'
  Call writeQlm(1,6)
  Call WriteQlm(2,6)
!RETURN
End If

!Call writeQlm(1,Log_Unit)
!Call WriteQlm(2,Log_Unit)

!Do i=1, Specific_atoms
!  If (DoZero) Atom_List(i)%Dij = CMPLX(0,0)        !** TESTING
!End Do
!If (DoPrint) Write(*,*) 'CalcHam: Before AddDiag!!!'

      Call AddDiagElements(Diag_Energy)  !** Adds the Diagonal Elements to Dij

!If (DoPrint) then 
!  Call WriteDij(6)
   
!  Do i=1, Specific_atoms
!    If (DoZero) Atom_List(i)%Dij = CMPLX(0,0)        !** TESTING
!  End Do
!End If
       !** Accum Vhat terms in Dij and dEdQlm(eq A23) **
energytemp = hat_energy
     Call AccumVhatdEdQlm(Hat_Energy) 
If (DoPrint) then
   energytemp = hat_energy - energytemp
   Write(*,*) 'CalcHam: V^ energy =',energytemp
   Write(*,*) 'CalcHam: Hat energy =',Hat_energy

   Write(*,*) 'CalcHam: After AccumVhatdEdQlm - dEdQlm'
   Call WriteQlm(2,6)
   Write(*,*) 'CalcHam: After AccumVhatdEdQlm - Dij'
   Call WriteDij(6)

!   Do i=1, Specific_atoms
!      if (DoZero) Atom_List(i)%Dij = 0        !** TESTING
!   End Do
End If
             !** Accumulate Vhartree terms in Dij (eq A26)
      Call AccumVhartree(Hartree_Energy)  
      write(log_unit,*) 'after accumVhartree'
      call flush(log_unit)
      !If (Estimate_selfenergy) Call AccumOneSelf
      !If (Estimate_scselfenergy) Call AccumOneSelf
If (DoPrint) then
Write(*,*) 'CalcHam: After AccumVhartree'
Call WriteDij(6)

Write(*,*) 'CalcHam: Before AccumVxc!!!'
!Do i=1, Specific_atoms
!  If (DoZero) Atom_List(i)%Dij = 0        !** TESTING
!End Do
End If
      If (XC_TYPE == XC_LDA_PW) Call AccumVxc_LDA( OneExc_Energy)
      If (XC_TYPE == XC_LSDA_PW) Call AccumVxc_LSDA( OneExc_Energy)
      If (XC_TYPE == XC_GGA_PBE) Call AccumVxc_GGA( OneExc_Energy)
      If (XC_TYPE == XC_SGGA_PBE) Call AccumVxc_SGGA( OneExc_Energy)
      write(log_unit,*) 'after accumVXC'
      call flush(log_unit)
If (DoPRint) then
Write(*,*) 'CalcHam: After AccumVxc!!!!!'
Call WriteDij(6)
End If

   Vwork = 0

   If (XC_Type == XC_LDA_PW) then
       Call CalcVxc_LDA( Den,TildeXC_Energy)   !** Calculate Vxc and Exc 

       !** Convert Density to FFT space ***
       Call PerformFFT(FFT_TO_G, G_HIGH, Den)

       !*** Store the Smooth density *** -- average over n(G)+Conjg(n(-G))
       !SCFvalues%RhoSmooth = Den(FFTmap_High(1:Gpnt_Size(G_High)))
       SCFvalues%RhoSmooth(1) = Den(FFTmap_High(1))
       SCFvalues%RhoSmooth(2:Gpnt_Size(G_High)) = &
            (Den(FFTmap_High(2:Gpnt_Size(G_High))) + &
            Conjg(Den(FFTmap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High)))))/2
   Else if (XC_Type == XC_LSDA_PW) then
       Call CalcVxc_LSDA( Den, Denspin,TildeXC_Energy)   !** Calculate Vxc and Exc 

       !** Convert Density to FFT space ***
       Call PerformFFT(FFT_TO_G, G_HIGH, Den)
       Call PerformFFT(FFT_TO_G, G_HIGH, Denspin)
       

       !*** Store the Smooth density *** -- average over n(G)+Conjg(n(-G))
       !SCFvalues%RhoSmooth = Den(FFTmap_High(1:Gpnt_Size(G_High)))
       SCFvalues%RhoSmooth(1) = Den(FFTmap_High(1))
       SCFvalues%RhoSmoothspin(1) = Denspin(FFTmap_High(1))
       
       SCFvalues%RhoSmooth(2:Gpnt_Size(G_High)) = &
            (Den(FFTmap_High(2:Gpnt_Size(G_High))) + &
            Conjg(Den(FFTmap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High)))))/2
       SCFvalues%RhoSmoothspin(2:Gpnt_Size(G_High)) = &
            (Denspin(FFTmap_High(2:Gpnt_Size(G_High))) + &
            Conjg(Denspin(FFTmap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High)))))/2
    Else if (XC_Type == XC_GGA_PBE) then
       !** Convert Density to FFT space ***
       Call PerformFFT(FFT_TO_G, G_HIGH, Den)

       !*** Store the Smooth density *** -- average over n(G)+Conjg(n(-G))
       SCFvalues%RhoSmooth(1) = Den(FFTmap_High(1))
       SCFvalues%RhoSmooth(2:Gpnt_Size(G_High)) = &
            (Den(FFTmap_High(2:Gpnt_Size(G_High))) + &
            Conjg(Den(FFTmap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High)))))/2
       
       Call CalcVxc_GGA(Den,TildeXC_Energy)
    Else if (XC_Type == XC_SGGA_PBE) then
       !** Convert Density to FFT space ***
       Call PerformFFT(FFT_TO_G, G_HIGH, Den)
       Call PerformFFT(FFT_TO_G, G_HIGH, Denspin)
       !*** Store the Smooth density *** -- average over n(G)+Conjg(n(-G))
       SCFvalues%RhoSmooth(1) = Den(FFTmap_High(1))
       SCFvalues%RhoSmooth(2:Gpnt_Size(G_High)) = &
            (Den(FFTmap_High(2:Gpnt_Size(G_High))) + &
            Conjg(Den(FFTmap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High)))))/2
       SCFvalues%RhoSmoothspin(1) = Denspin(FFTmap_High(1))
       SCFvalues%RhoSmoothspin(2:Gpnt_Size(G_High)) = &
            (Denspin(FFTmap_High(2:Gpnt_Size(G_High))) + &
            Conjg(Denspin(FFTmap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High)))))/2
       
       Call CalcVxc_SGGA(Den, Denspin,TildeXC_Energy)
    Else
       Write(Error_Unit,*) 'CalcHam:    Error in XC_Type ', XC_Type
       stop
    EndIf

!    write(6,*) 'after FFT -- den(1)',den(1)

   !** Accumulates the Vlocal term to Vwork
   if(spindependence) then                        
     Call AccumVlocal(TildePot_Energy, Den, Denspin)
     !Den = Den + Denspin
     write(Log_unit,*) 'CalcHam: Den,Denspin', Size(Den),Size(Denspin)
     call flush(log_unit)
     Do i=1, FFT_Grid(4,G_HIGH)
       Den(i)=Den(i)+Denspin(i)
     enddo
     write(Log_unit,*) 'CalcHam: Den,Denspin after:', FFT_Grid(4,G_HIGH)
     call flush(log_unit)
    
   else  
     Call AccumVlocal(TildePot_Energy, Den)
   end if

      Qt = Den(1)
     write(log_unit,*) 'Qt = ', Qt
     call flush(log_unit)

     Call CalcDenHat_dEdQlm( Den)     !** Calc N^ 

!!!!!!!!!
!     write(Log_Unit,*)'test \hat{n}', SCFvalues%rhohat(1)

!---------------------
!   If (V_MixType == MIX_DENSITY) then
!      If (V_NewMix < 0.9999) then 
!         Allocate(DenTemp(Gpnt_Size(G_HIGH)))
         !DenTemp = Den(FFTmap_High(1:Gpnt_Size(G_HIGH)))
!  average over n(G) and n(-G)*
!          DenTemp(1)=Den(FFTmap_High(1))
!          DenTemp(2:Gpnt_Size(G_High)) = &
!               (Den(FFTmap_High(2:Gpnt_Size(G_High))) + &
!               Conjg(Den(FFTmap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High)))))/2
!          DenTemp = DenTemp - &
!              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_HIGH))
!
!         Call Anderson_Mix(SCFvalues%AC, &
!              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_HIGH)), &
!              DenTemp)
!
!         Den = 0
!         Den(FFTmap_High(1:Gpnt_Size(G_HIGH))) = &
!              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_High))
!         Den(FFTmap_High(Gpnt_Size(G_High)+1:)) = &
!              CONJG(SCFvalues%Ve_Fourier(2:Gpnt_Size(G_high)))
!
!         DeAllocate(DenTemp)
!      else
!         SCFvalues%Ve_Fourier(1) = Den(FFTmap_High(1))
!         SCFvalues%Ve_Fourier(2:Gpnt_Size(G_HIGH)) =  &
!              (Den(FFTmap_HIGH(2:Gpnt_Size(G_HIGH))) + &
!               Conjg(Den(FFTmap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High)))))/2
!      End If
!   End If


    
if (DoPrint) Call WriteQlm(2,6)
   End IF

   Qh = Den(1) +SCFValues%CoreTail(1) - Qt

   Call AccumShift
   
   Call AccumPotential( TildePot_Energy, Den)

!*** New Mix V w/o Vxc and Vlocal
!   If (V_MixType == MIX_V) then
!      If (V_NewMix < 0.9999) then 
!         Allocate(DenTemp(Gpnt_Size(G_HIGH)))
!         DenTemp(1) = Den(FFTmap_HIGH(1))
!         DenTemp(2:Gpnt_Size(G_HIGH)) = &
!             (Den(FFTmap_HIGH(2:Gpnt_Size(G_HIGH))) + &
!             Conjg(Den(FFTmap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High)))))/2
!         DenTemp(1:Gpnt_Size(G_HIGH)) = DenTemp(1:Gpnt_Size(G_HIGH)) - &
!              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_HIGH))
!      
!         Call Anderson_Mix(SCFvalues%AC, &
!              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_HIGH)), DenTemp)
!         
!         Den = 0
!         Den(FFTmap_HIGH(1:Gpnt_Size(G_HIGH))) = &
!              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_HIGH))
!         n = Gpnt_Size(G_High)+Gpnt_size(G_HIGH)-1
!         Den(FFTmap_HIGH(Gpnt_Size(G_HIGH)+1:n)) = &
!              CONJG(SCFvalues%Ve_Fourier(2:Gpnt_Size(G_HIGH)))
!         Vwork = Vwork + Den
!         DeAllocate(DenTemp)
!      else
!         SCFvalues%Ve_Fourier(1) = Den(FFTmap_High(1))
!         SCFvalues%Ve_Fourier(2:Gpnt_Size(G_HIGH)) =  &
!              (Den(FFTmap_HIGH(2:Gpnt_Size(G_HIGH))) + &
!               Conjg(Den(FFTmap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High)))))/2
!
!         Vwork(FFTmap_HIGH(1:Gpnt_Size(G_HIGH))) = &
!              Vwork(FFTmap_HIGH(1:Gpnt_Size(G_HIGH))) + &
!              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_HIGH))
!         n = Gpnt_Size(G_High)+Gpnt_size(G_HIGH)-1
!         Vwork(FFTmap_HIGH(Gpnt_Size(G_HIGH)+1:n)) = &
!              Vwork(FFTmap_HIGH(Gpnt_Size(G_HIGH)+1:n)) + &
!              CONJG(SCFvalues%Ve_Fourier(2:Gpnt_Size(G_HIGH)))
!      End If
!
!   End If

!! Store currect Dij coefficients
  Do i=1,Specific_atoms
     Do j=1, AtomType_Info(Atom_list(i)%TypeIndex)%nlm_Size
        Do k=1, j                        
           Atom_List(i)%cDij(j,k)=(Atom_List(i)%Dij(j,k) + &
                  CONJG(Atom_List(i)%Dij(k,j)))/2
           Atom_List(i)%cDij(k,j)=Atom_List(i)%cDij(j,k) 
        Enddo
     Enddo
  Enddo
  If (spindependence) then
  Do i=1,Specific_atoms
     Do j=1, AtomType_Info(Atom_list(i)%TypeIndex)%nlm_Size
        Do k=1, j                        
           Atom_List(i)%cDijspin(j,k)=(Atom_List(i)%Dijspin(j,k) + &
                  CONJG(Atom_List(i)%Dijspin(k,j)))/2
           Atom_List(i)%cDijspin(k,j)=Atom_List(i)%cDijspin(j,k) 
        Enddo
     Enddo
  Enddo
  Endif

   If (Dij_NewMix < 0.9999) then                                        
      Do i=1, Specific_Atoms    !** Mix each atom's Dij separately   
         j = AtomType_Info(Atom_list(i)%TypeIndex)%nlm_Size             
         n = (j*(j+1))/2              !**  Determine unique Dij's

         t = 0    !** Copy new Dij's to 1-D mixing array ***         
         Do j=1, AtomType_Info(Atom_list(i)%TypeIndex)%nlm_Size
            Do k=1, j                        
               t = t + 1                                     
               Work_Dij(t) = (Atom_List(i)%Dij(j,k)   + &   !** Work_Dij = F(X)
                  CONJG(Atom_List(i)%Dij(k,j)))/2
            End Do
         End Do
                                                                              
         !** Remember that Anderson mixing wants X and F(X)-X ** 
         !** X = Atom_list(i)%Dij_Old and F(X)-X = Work_Dij    
         Work_Dij(1:n) = Work_Dij(1:n) - Atom_list(i)%Dij_old !** Form F(X)-X

         write(Log_Unit,*)'Mix DeltaDij = ',i,Dot_Product(Work_Dij,Work_Dij)
                                                                              
         !*** Perform the mixing ***                             
         Call Anderson_Mix(Atom_List(i)%Dij_AC, Atom_list(i)%Dij_Old, &
              Work_Dij(1:n))                                       
                                                                              
         !*** Now store the results back in the Dij arrays ***      
         !*** Upon exiting Anderson_Mix Atom_List(i)%Dij_old contains the
         !*** mixed or new Dij's to use                              
         t = 0                                                          
         Do j=1, AtomType_Info(Atom_list(i)%TypeIndex)%nlm_Size     
            Do k=1, j                                       
               t = t + 1                                      
               Atom_List(i)%Dij(j,k) = Atom_list(i)%Dij_old(t)
               If (j/=k) Atom_List(i)%Dij(k,j) = CONJG(Atom_list(i)%Dij_old(t))
            End Do
         End Do

       if(spindependence) then
         t = 0    !** Copy new Dij's to 1-D mixing array ***         
         Do j=1, AtomType_Info(Atom_list(i)%TypeIndex)%nlm_Size
            Do k=1, j                        
               t = t + 1                                     
               Work_Dijspin(t) = (Atom_List(i)%Dijspin(j,k)   + &   !** Work_Dij = F(X)
                  CONJG(Atom_List(i)%Dijspin(k,j)))/2
            End Do
         End Do
                                                                              
         !** Remember that Anderson mixing wants X and F(X)-X ** 
         !** X = Atom_list(i)%Dij_Old and F(X)-X = Work_Dij    
         Work_Dijspin(1:n) = Work_Dijspin(1:n) - Atom_list(i)%Dij_oldspin !** Form F(X)-X

         write(Log_Unit,*)'Mix DeltaDij = ',i,Dot_Product(Work_Dijspin,Work_Dijspin)
                                                                              
         !*** Perform the mixing ***                             
         Call Anderson_Mix(Atom_List(i)%Dijspin_AC, Atom_list(i)%Dij_Oldspin, &
              Work_Dijspin(1:n))                                       
                                                                              
         !*** Now store the results back in the Dij arrays ***      
         !*** Upon exiting Anderson_Mix Atom_List(i)%Dij_old contains the
         !*** mixed or new Dij's to use                              
         t = 0                                                          
         Do j=1, AtomType_Info(Atom_list(i)%TypeIndex)%nlm_Size     
            Do k=1, j                                       
               t = t + 1                                      
               Atom_List(i)%Dijspin(j,k) = Atom_list(i)%Dij_oldspin(t)
               If (j/=k) Atom_List(i)%Dijspin(k,j) = CONJG(Atom_list(i)%Dij_oldspin(t))
            End Do
         End Do
       endif
      End Do
   else
      Do i=1, Specific_Atoms    !** Mix each atom's Dij separately
         t = 0    !** Copy new Dij's to 1-D mixing array ***            
         Do j=1, AtomType_Info(Atom_list(i)%TypeIndex)%nlm_Size
            Do k=1, j                                       
               t = t + 1                          
               Atom_list(i)%Dij_old(t) = (Atom_List(i)%Dij(j,k) + &
                        CONJG(Atom_List(i)%Dij(k,j)))/2
               if(spindependence) then
                   Atom_list(i)%Dij_oldspin(t) = (Atom_List(i)%Dijspin(j,k) + &
                        CONJG(Atom_List(i)%Dijspin(k,j)))/2
               endif
 
            End Do
         End Do 
      EndDo
   End If
                                                                              
!*********  

If (DoPrint) then
   Call WriteWij(Log_Unit)
   Call WriteDij(Log_Unit)
   Call WriteQlm(1,Log_Unit)
   Call WriteQlm(2,Log_Unit)
End If

   If (Atomic_Mode == SIM_AE) then
      Energy = AtomicEnergy - NuclearEnergy - TildePot_Energy &
           - TildeKE_Energy - TildeXC_Energy  

      Write(Log_Unit,*) 'CalcHam: Cohesive Energy = ',Energy
      Write(Log_Unit,*) 'CalcHam:         CalcHam = ',Energy-AtomicEnergy 
      Write(Log_Unit,*) 'CalcHam:   Atomic Energy = ',AtomicEnergy
      Write(Log_Unit,*) 'CalcHam:  Nuclear Energy = ',NuclearEnergy
      Write(Log_Unit,*) 'CalcHam:              KE = ',TildeKE_Energy
      Write(Log_Unit,*) 'CalcHam:      Pot(uncor) = ',TildePot_Energy
      Write(Log_Unit,*) 'CalcHam:        Pot + En = ',TildePot_Energy &
                                                   + NuclearEnergy
      Write(Log_Unit,*) 'CalcHam:              XC = ',TildeXC_Energy
      Write(Log_Unit,*) 'CalcHam:       Hat - Hat = ',Hat_Hat_Energy
   else
      Energy = TildePot_Energy + TildeKE_Energy + TildeXC_Energy  &
           + OneExc_Energy + Hartree_Energy + Diag_Energy + Hat_Energy

      Write(Log_Unit,*) 'CalcHam: '
      Write(Log_Unit,*) 'CalcHam: Cohesive Energy = ',AtomicEnergy-Energy
      Write(Log_Unit,*) 'CalcHam:         CalcHam = ',Energy 
      Write(Log_Unit,*) 'CalcHam:          Atomic = ',AtomicEnergy 
    !If (Estimate_selfenergy) &
    !  Write(Log_Unit,*) 'CalcHam: Self_Energy_Cor = ', &
    !                 Self_Energy_Correction
    !If (Estimate_scselfenergy) &
    !  Write(Log_Unit,*) 'CalcHam: Self_Energy_Cor = ', &
    !                 Self_Energy_Correction
      Write(Log_Unit,*) 'CalcHam:            Diag = ',Diag_Energy
      Write(Log_Unit,*) 'CalcHam:         Hartree = ',Hartree_Energy
    ! If (Estimate_selfenergy) &
    !  Write(Log_Unit,*) 'CalcHam: Auxform_of_Hart = ',&
    !                 Auxform_of_Hartree
      Write(Log_Unit,*) 'CalcHam:             Hat = ',Hat_Energy
      If (XC_Type == XC_LDA_PW) &
       Write(Log_Unit,*) 'CalcHam:          OneExc = ',OneExc_Energy,' (LDA-PW)'
      If (XC_Type == XC_LSDA_PW) &
       Write(Log_Unit,*) 'CalcHam:          OneExc = ',OneExc_Energy,' (LSDA-PW)'
      If (XC_Type == XC_GGA_PBE) &
       Write(Log_Unit,*) 'CalcHam:          OneExc = ',OneExc_Energy,' (GGA-PBE)'
      If (XC_Type == XC_SGGA_PBE) &
       Write(Log_Unit,*) 'CalcHam:          OneExc = ',OneExc_Energy,' (SGGA-PBE)'
      Write(Log_Unit,*) 'CalcHam:             KE~ = ',TildeKE_Energy
      Write(Log_Unit,*) 'CalcHam:            Pot~ = ',TildePot_Energy
      If (XC_Type == XC_LDA_PW) &
       Write(Log_Unit,*) 'CalcHam:             XC~ = ',TildeXC_Energy,' (LDA-PW)'
      If (XC_Type == XC_LSDA_PW) &
       Write(Log_Unit,*) 'CalcHam:             XC~ = ',TildeXC_Energy,' (LSDA-PW)'
      If (XC_Type == XC_GGA_PBE) &
       Write(Log_Unit,*) 'CalcHam:             XC~ = ',TildeXC_Energy,' (GGA-PBE)'
      If (XC_Type == XC_SGGA_PBE) &
       Write(Log_Unit,*) 'CalcHam:             XC~ = ',TildeXC_Energy,' (SGGA-PBE)'
      Write(Log_Unit,*) 'CalcHam:       Hat - Hat = ',Hat_Hat_Energy

      Energy = AtomicEnergy - Energy
   End If 

   Write(Log_Unit,*)    'CalcHam:   Smooth Charge = ',Qt

   If (Atomic_Mode /= SIM_AE) then
      Write(Log_Unit,*) 'CalcHam:              Q^ = ',Qh

      rmag = Qt+Qh
      Write(Log_Unit,*) 'CalcHam:           Q~+Q^ =',rmag

      If (ABS(Rmag)>1E-3) then
         Write(Log_Unit,*) 'CalcHam: ERROR!!! Qt+Qh=',rmag
         Write(Error_Unit,*) 'CalcHam: ERROR!!! Qt+Qh=',rmag
      End If
   End If

!   !*** Restrict the potential ***
!    If (Filter_potential) then
!          Call RestrictPot( Vwork, Den)
!       Else
!          Den=Vwork
!    EndIf

!! store current Veff
   SCFValues%cVe_F(1) = SCFValues%Ve(FFTmap_HIGH(1))
   SCFValues%cVe_F(2:Gpnt_Size(G_HIGH)) = &
     (SCFValues%Ve(FFTmap_HIGH(2:Gpnt_Size(G_HIGH))) + &
     CONJG(SCFValues%Ve(FFTmap_HIGH(Gpnt_Size(G_HIGH)+1:Gall_Size(G_HIGH)))))/2

  If (spindependence) then
   SCFValues%cVe_Fspin(1) = SCFValues%Vespin(FFTmap_HIGH(1))
   SCFValues%cVe_Fspin(2:Gpnt_Size(G_HIGH)) = &
     (SCFValues%Vespin(FFTmap_HIGH(2:Gpnt_Size(G_HIGH))) + &
     CONJG(SCFValues%Vespin(FFTmap_HIGH(Gpnt_Size(G_HIGH)+1:Gall_Size(G_HIGH)))))/2
  Endif

!*** Old V_Mix w/ Vxc and Vlocal ***
   If (V_MixType == MIX_VEFF) then
      If (V_NewMix < 0.9999) then 
         Vwork(1) = SCFValues%Ve(FFTmap_HIGH(1))
         Vwork(2:Gpnt_Size(G_HIGH)) = &
           (SCFValues%Ve(FFTmap_HIGH(2:Gpnt_Size(G_HIGH))) + &
           CONJG(SCFValues%Ve(FFTmap_HIGH(Gpnt_Size(G_HIGH)+1:Gall_Size(G_HIGH)))))/2
         Den(1:Gpnt_Size(G_HIGH)) = Vwork(1:Gpnt_Size(G_HIGH)) - &
              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_HIGH))
      
         Write(Log_Unit,*)'Mix DeltaVeff',Dot_Product(Den(1:Gpnt_Size(G_HIGH)),&
             Den(1:Gpnt_Size(G_HIGH)))

         Call Anderson_Mix(SCFvalues%AC, &
              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_HIGH)), &
              Den(1:Gpnt_Size(G_HIGH)))
         
         SCFValues%Ve = 0
         SCFValues%Ve(FFTmap_HIGH(1:Gpnt_Size(G_HIGH))) = &
              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_HIGH))
         SCFValues%Ve(FFTmap_HIGH(Gpnt_Size(G_HIGH)+1:)) = &
              CONJG(SCFvalues%Ve_Fourier(2:Gpnt_Size(G_HIGH)))
      else
         SCFvalues%Ve_Fourier(1:) = SCFValues%Ve(FFTmap_HIGH(1))
         SCFvalues%Ve_Fourier(2:Gpnt_Size(G_HIGH)) =  &
              (SCFValues%Ve(FFTmap_HIGH(2:Gpnt_Size(G_HIGH))) + &
              CONJG(SCFValues%Ve(FFTmap_HIGH(Gpnt_Size(G_HIGH)+1:Gall_Size(G_HIGH)))))/2
      End If
   End If

   !*** Calc the change in the old and new potentials ***  
   Call PerformFFT(FFT_TO_R, G_HIGH, SCFValues%Ve)
!**     write(6,*) 'potential sum' , sum(den)

 if(spindependence) then
!*** Old V_Mix w/ Vxc and Vlocal ***
   If (V_MixType == MIX_VEFF) then
      If (V_NewMix < 0.9999) then 
         Vwork(1) = SCFValues%Vespin(FFTmap_HIGH(1))
         Vwork(2:Gpnt_Size(G_HIGH)) = &
           (SCFValues%Vespin(FFTmap_HIGH(2:Gpnt_Size(G_HIGH))) + &
           CONJG(SCFValues%Vespin(FFTmap_HIGH(Gpnt_Size(G_HIGH)+1:Gall_Size(G_HIGH)))))/2
         Den(1:Gpnt_Size(G_HIGH)) = Vwork(1:Gpnt_Size(G_HIGH)) - &
              SCFvalues%Ve_Fourierspin(1:Gpnt_Size(G_HIGH))
      
         Write(Log_Unit,*)'Mix DeltaVeff',Dot_Product(Den(1:Gpnt_Size(G_HIGH)),&
             Den(1:Gpnt_Size(G_HIGH)))

         Call Anderson_Mix(SCFvalues%ACspin, &
              SCFvalues%Ve_Fourierspin(1:Gpnt_Size(G_HIGH)), &
              Den(1:Gpnt_Size(G_HIGH)))
         
         SCFValues%Vespin  = 0
         SCFValues%Vespin(FFTmap_HIGH(1:Gpnt_Size(G_HIGH))) = &
              SCFvalues%Ve_Fourierspin(1:Gpnt_Size(G_HIGH))
         SCFValues%Vespin(FFTmap_HIGH(Gpnt_Size(G_HIGH)+1:)) = &
              CONJG(SCFvalues%Ve_Fourierspin(2:Gpnt_Size(G_HIGH)))
      else
         SCFvalues%Ve_Fourierspin(1:) = SCFValues%Vespin(FFTmap_HIGH(1))
         SCFvalues%Ve_Fourierspin(2:Gpnt_Size(G_HIGH)) =  &
              (SCFValues%Vespin(FFTmap_HIGH(2:Gpnt_Size(G_HIGH))) + &
              CONJG(SCFValues%Vespin(FFTmap_HIGH(Gpnt_Size(G_HIGH)+1:Gall_Size(G_HIGH)))))/2
      End If
   End If

   !*** Calc the change in the old and new potentials ***  
   Call PerformFFT(FFT_TO_R, G_HIGH, SCFValues%Vespin)
!**     write(6,*) 'potential sum' , sum(den)
  endif


   Write(Log_Unit,*) ' '
   Write(Log_Unit,*) 'CalcHam:  Change in Wij (atom, change):'
   Qt = 0
   Do i=1, Specific_Atoms
      Atom_List(i)%Wij_Old = Atom_List(i)%Wij - Atom_List(i)%Wij_Old
      gmag = SUM(CONJG(Atom_List(i)%Wij_Old) * Atom_List(i)%Wij_Old)
      Qt = Qt + gmag
      Write(Log_Unit,*) 'CalcHam:CWij      ',i, gmag
   End Do
 if(spindependence) then
   Do i=1, Specific_Atoms
      Atom_List(i)%Wij_Oldspin = Atom_List(i)%Wijspin - Atom_List(i)%Wij_Oldspin
      gmag = SUM(CONJG(Atom_List(i)%Wij_Oldspin) * Atom_List(i)%Wij_Oldspin)
      Qt = Qt + gmag
      Write(Log_Unit,*) 'CalcHam:CWijdn      ',i, gmag
   End Do
 endif
   Write(Log_Unit,*) 'CalcHam:CWij  Total Change in Wij:', Qt

   DeAllocate(Den,ProjDotPsi)  !*** Free the big arrays ***
   if(spindependence) DeAllocate(Denspin) 
Call PrintDate(Log_Unit, 'CalcHam: Finished Routine!')   
   Call Stop_Timer(Timer(CalcHam_Timer))

   Prev_CohesiveEnergy(1) = Prev_CohesiveEnergy(2)
   Prev_CohesiveEnergy(2) = energy

   HamLoaded=.true.
   DeAllocate(Psi_toUse)

  Return
End Subroutine


!******************************************************************************
!
!  CalcBandMerit - Calculates the band structure merit function.
!
!******************************************************************************

Subroutine CalcBandMerit( Energy)
   Real,     Intent(OUT) :: Energy

   Integer :: i, Kpnt

   Call Start_Timer(Timer(CalcHam_Timer))

   !!!! Call CalcEigenValues  -- eigenvalues already determined

   Energy = 0
   Do Kpnt=1, NumKpnts
      Do i=1, Mem_MapSize
         If (PsiInfo(i)%Kpnt == Kpnt) then
            If (PsiInfo(i)%Energy <= Eigen_Max) then
               Energy = Energy + ABS(PSiInfo(i)%Energy)
            End If
         End If
      End Do
   end Do

   Call Stop_Timer(Timer(CalcHam_Timer))

   Return
End Subroutine

!******************************************************************************
!
! CalcMeritFunction - Calculates the Merit function
!     Normally CalcHam is called unless BandStructure_Mode = .TRUE. then
!     H is assumed to be fixed and the Band structure merit function is called
! 
!******************************************************************************

Subroutine CalcMeritFunction( Energy)
   Real,     Intent(OUT) :: Energy


   If (BandStructure_Mode) then
      Call CalcBandMerit(Energy)
   else
      Call CalcHam( Energy)
   End If

   Return
End Subroutine

!******************************************************************************
!
! InitHam - Initializes the HAmiltonian module
!
!******************************************************************************

Subroutine InitHam
  Integer :: i,j, Rad_Size
  Type (atom_info_fixed), Pointer :: AT
  Complex, Pointer :: Work(:), Work2(:)

  Prev_CohesiveEnergy = 1E30

  write(Log_unit,*) 'Entering InitHam'
  call flush(Log_Unit)
  Allocate(BandMap(2,Mem_MapSize),stat=i)
    If (i /= 0) then
      Write(Error_Unit,*) 'InitHam:  Failed to allocate BandMap',i
      Call Flush(Error_Unit)
      stop
    EndIf

  If (Atomic_Mode == SIM_AE) RETURN   !** Ignore the rest if an AE calc

!RETURN !**REMOVE** Debug Only

       !** Calculate the radial hat functions
  If (Atomic_Mode /= SIM_SEPM) Call CalcRadialHat      
  Call CalcRadial_Vlocal  !** Calculate the radial V_local functions
 !If (Estimate_selfenergy )&
 ! Call MakeHatij    !** Initialize smooth Self-energy terms
  Call CalcVlocal(SCFvalues%V_local)
  Call CalcCoreTail(SCFvalues%CoreTail)

  write(Log_unit,*) 'InitHam: SCFvalues%CoreTail', SCFvalues%CoreTail(1)
  call flush(Log_Unit)
  If (Atomic_Mode == SIM_SEPM) then
     Call GetBuffer( Work2)

     Work => SCFvalues%V_Local
     Work2(1:Gpnt_Size(G_LOW)) = Work(1:Gpnt_Size(G_Low))
     SCFvalues%Ve_Fourier =  Work(1:Gpnt_Size(G_LOW))

     Work2(Gpnt_Size(G_LOW)+1:) = &
          Work(Gpnt_Size(G_high)+1:Gpnt_Size(G_High)+Gpnt_size(G_low)-1)     

     Work => SCFvalues%Work
     Work = 0
     Call RS_toR(Work2, Work)
     SCFvalues%Ve = Work

!SCFvalues%Ve = 0
     Call FreeBuffer(Work2)

     RETURN  !*** EXIT EARLY ***
  End IF

  Do i=1, Atom_Types
     AT => AtomType_Info(i)
     Allocate(AT%Rad_Size, AT%Rad_Skip)

!AT%Mesh_Size = At%Mesh_Size - 2   !*** DEBUG ONLY ****
!AT%Mesh_Size = At%Mesh_Size-1  !*** DEBUG ONLY * for li


     Rad_Size = 0
     Do j=1, AT%Mesh_Size, AtomicVxc_Step
        Rad_Size = Rad_Size + 1
     End Do
     AT%Rad_Size = Rad_Size
     AT%Rad_Skip = AtomicVxc_Step     !note this is always 1 unless
                                      !  we fix the Core_Density part!!!
     Write(Log_Unit,*) 'InitHam: ',' * Atom=',i, &
          ' * SIZE=',AT%Rad_Size, &
          ' * SKIP=',AT%Rad_Skip     
     call flush(Log_unit)

  End Do


  !**** Prepare the Vxc data structures *****
     !************ Generate the Angular Integration points ***************
     !Call InitSpharm
     Num_AngPoints = Angular_Points**2
     AngPoints => AngInit(Angular_Points,Angular_Points,Xtal%Volume,Error_Unit)
  Do i=1, Atom_Types
     AT => AtomType_Info(i)
     Call PrepareVxc(AT)
  End Do

  !*** Allocate space for Dij mixing ***
  i = MaxVal(AtomType_Info(:)%nlm_size)
  i = i*(i+1)/2
  Allocate(Work_Dij(i),stat=j)
  if(spindependence) Allocate(Work_Dijspin(i), stat=j)
    If (j /= 0) then
      Write(Error_Unit,*) 'InitHam:  Failed to allocate BandMap',j
      Call Flush(Error_Unit)
      stop
    EndIf

  !*** Init the Dij Mixing ***
  Do i=1, Specific_Atoms
     j = AtomType_Info(Atom_List(i)%TypeIndex)%nlm_Size
     j = j*(j+1)/2

     Call InitAnderson(Atom_List(i)%Dij_AC, Error_Unit, Mix_Size, &
          j, 1E-5, Dij_NewMix, AndersonConditionNo)
     if(spindependence) then
         Call InitAnderson(Atom_List(i)%Dijspin_AC, Error_Unit, Mix_Size, &
          j, 1E-5, Dij_NewMix, AndersonConditionNo)
     endif
  End Do

  Call Count_Atoms

  Return
End Subroutine

End Module
