!******************************************************************************
!
! File : hamop.f90
!   originally in hamiltonian.f90
!   by : Alan Tackett
!   on : 02/07/97
!  for : PAW
!
! Module for calculating the Hamiltonian and Cohesive Energy of the PAW
! system.  Also contains routines to do H*Psi and O*Psi.
!
! Contains routines of the type H*Psi
!
! Updated for spin by Ping Tang and N. A. W. Holzwarth
!      Last change 5/25/05
!******************************************************************************

Module hamop

Use atom_data
Use basis_lib
Use crystal_data
Use gpoints 
Use laplacian
Use memmgr
Use oinverse
Use options_data
Use paw_inout
Use psilib
Use projectors
Use spherical_harmonic

Implicit NONE!!!!

!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
!  Hpsi - Multiplies the Hamiltonian times the Given Wave vector
!         and returns the result
!
!  Psi   - Wave vector to multiply. NO Bloch K and DetJ!!!!
!  Hx    - Output result. Hx is initialized to 0!
!  Kpnt  - K-point array index
!
!  KE      - Work Array
!
!******************************************************************************

Subroutine Hpsi( Psi, Hx, Ve, ProjDotPsi, Kpnt, PDOT_Stored)
  Complex,          Intent(IN)    :: Psi(:)
  Complex,          Intent(IN)    :: Ve(:)
  Complex,          Intent(INOUT) :: Hx(:)
  Complex,          Intent(INOUT) :: ProjDotPsi(:)
  Integer,          Intent(IN)    :: Kpnt
  Logical,          Intent(INOUT) :: PDOT_Stored


  Complex, Pointer :: Work(:)
  Integer :: i,j,k

  Call Start_Timer(Timer(HPsi_Timer))

  If (Atomic_Mode == SIM_TIME) then
     Call CalcDEL2_TIME( BZ%Ku(:,Kpnt), Psi, Hx)
  else
     Call CalcDEL2_Phase( BZ%Ku(:,Kpnt), Psi, Hx)
     Hx = -Hx
  End If

  Work => SCFvalues%Work

  If (Atomic_Mode == SIM_SEPM) then
     Call RS_toR(Psi, Work) 
     Work = V_Weight*Ve*Work
     Call RS_toG(Work, Hx)
  else if (Proj_Mode == PROJ_RS) then
     Call RS_toR(Psi, Work) 
     If (.NOT. PDOT_Stored) then
        Call CalcProjProducts(Work, ProjDotPsi)
        PDOT_Stored = .TRUE.
     End If

     Work = V_Weight*Ve*Work
     Call AccumProj(Work, ProjDotPsi, DijMatrix)

     Call RS_toG(Work, Hx)
  else 

     Call V_G_to_r(Psi,G_LOW,Work,G_HIGH)

     Work = V_Weight*Ve*Work
     Call V_r_to_G(Work,G_HIGH,Hx,G_LOW)

     If (.NOT. PDOT_Stored) then
        Call CalcProjProducts(Psi, ProjDotPsi)
        PDOT_Stored = .TRUE.
     End If

     Call AccumProj(Hx, ProjDotPsi, DijMatrix)
  End If

  Call Stop_Timer(Timer(HPsi_Timer))

  Return
End Subroutine


!******************************************************************************
!
!  Opsi - Multiplies the Overlap matrix times the Given Wave vector
!         and returns the result
!
!  Psi   - Wave vector to multiply.
!  Ox    - Output result. Ox is initialized to 0!
!  ProjDotPsi - Array containing the projector psi dot products
!  Kpnt - Kpnt array index
!
!******************************************************************************

Subroutine Opsi( Psi, Ox, ProjDotPsi, Kpnt, PDOT_Stored)
  Complex,          Intent(IN)  :: Psi(:)
  Complex,          Intent(OUT) :: Ox(:)
  Complex,          Intent(INOUT) :: ProjDotPsi(:)
  Integer,          Intent(IN)  :: Kpnt
  Logical,          Intent(INOUT) :: PDOT_Stored

  Complex, Pointer :: Work(:)
  COmplex :: xOx
  Real :: temp

!Ox = Psi
!RETURN   !***REMOVE** Debug

  Call Start_Timer(Timer(OPsi_Timer))

  Ox = Psi

  If (Atomic_Mode == SIM_SEPM) then
     !** Don't do anything **
  else If (Proj_Mode == PROJ_RS) then
     Work => SCFvalues%Work
     If (.NOT. PDOT_Stored) then
        Call RS_toR(Psi, Work) 

        Call CalcProjProducts(Work, ProjDotPsi)
        PDOT_Stored = .TRUE.
     End If

     Work = 0
     Call AccumProj(Work, ProjDotPsi, OijMatrix)

     Call RS_toG(Work, Ox)
  else
     If (.NOT. PDOT_Stored) then
        Call CalcProjProducts(Psi, ProjDotPsi)
        PDOT_Stored = .TRUE.
     End If

     Call AccumProj(Ox, ProjDotPsi, OijMatrix)
  End If

  Call Stop_Timer(Timer(OPsi_Timer))

  Return
End Subroutine

!******************************************************************************
!
!  CalcHxandOx - Calculates H and O times the given Psi
!
!  Psi   - Wave vector to multiply. NO Bloch K and DetJ!!!!
!  Hx    - Output result. Hx is initialized to 0!
!  Ox    - Output result. Ox is initialized to 0!
!  Kpnt  - K-point array index
!
!******************************************************************************

Subroutine CalcHxandOx( Psi, Hx, Ox, Ve, PDOT, Kpnt, PDOT_Stored)
  Complex,          Intent(IN)  :: Psi(:)
  Complex,          Intent(IN)  :: Ve(:)
  Complex,          Intent(OUT) :: Hx(:)
  Complex,          Intent(OUT) :: Ox(:)
  Complex,          Intent(INOUT) :: PDOT(:) 
  Integer,          Intent(IN)  :: Kpnt
  Logical,          Intent(INOUT) :: PDOT_Stored

!*  Complex :: PDOT(PLM_Max)
  Complex, Pointer :: KE(:), Work(:)
  Integer :: i,j


!-----
  Call Start_Timer(Timer(HandOPsi_Timer))

  Call CalcDEL2_Phase( BZ%Ku(:,Kpnt), Psi, Hx)
  Hx = -Hx

  Work => SCFvalues%Work


  If (Atomic_Mode == SIM_SEPM) then
     Call RS_toR(Psi, Work) 
     Work = V_Weight*Ve*Work
     Call RS_toG(Work, Hx)
     Ox = Psi
  else If (Proj_Mode == PROJ_RS) then
     Call RS_toR(Psi, Work) 
     Ox = Psi
     If (.NOT. PDOT_Stored) then
        Call CalcProjProducts(Work, PDot)
        PDOT_Stored = .TRUE.
     End If

     RS_Work1 = V_Weight*Ve*Work
     Work = 0
     Call AccumBothProj(RS_Work1, Work, PDot, DijMatrix, OijMatrix)

     Call RS_toG(RS_Work1, Hx)
     Call RS_toG(Work, Ox)
  else
     Call V_G_to_r(Psi,G_LOW,Work,G_HIGH) 
     Work = V_Weight*Ve*Work
     Call V_r_to_G(Work,G_HIGH,Hx,G_LOW)
  

     If (.NOT. PDOT_Stored) then
        Call CalcProjProducts(Psi, PDot)
        PDOT_Stored = .TRUE.
     End If

     Ox = Psi
     Call AccumBothProj(Hx, Ox, PDot, DijMatrix, OijMatrix)
  End If

  Call Stop_Timer(Timer(HandOPsi_Timer))
  Return
End subroutine


!******************************************************************************
!
!  Calc_HOiH_Hx_Ox - Calculates HOiH, H and O times the given Psi
!
!  Psi   - Wave vector to multiply. NO Bloch K and DetJ!!!!
!  HOiHx - Output result. Initialized to 0!
!  Hx    - Output result. Hx is initialized to 0!
!  Ox    - Output result. Ox is initialized to 0!
!  Kpnt  - K-point array index
!
!******************************************************************************

Subroutine Calc_HOiHx_Hx_Ox( Psi, HOiHx, Hx, Ox, Ve, PDOT, Kpnt, PDOT_Stored, Eo, Ewidth)
  Complex,          Intent(IN)  :: Psi(:)
  Complex,          Intent(IN)  :: Ve(:)
  Complex,          Intent(OUT) :: HOiHx(:)
  Complex,          Intent(OUT) :: Hx(:)
  Complex,          Intent(OUT) :: Ox(:)
  Complex,          Intent(INOUT) :: PDOT(:) 
  Integer,          Intent(IN)  :: Kpnt
  Logical,          Intent(INOUT) :: PDOT_Stored
  Real,             Intent(IN)  :: Eo
  Real,             Intent(IN)  :: Ewidth

 !Complex :: PDOT_OiHx(PLM_Max)
  Complex ,allocatable:: PDOT_OiHx(:)
  Complex, Pointer :: KE(:), Work(:), W2(:), W3(:)
  Integer :: i,j


  Allocate(PDOT_OiHx(PLM_Max))
  Call CalcDEL2_Phase( BZ%Ku(:,Kpnt), Psi, Hx)
  Hx = -Hx

  Work => SCFvalues%Work


  If (Proj_Mode == PROJ_RS) then
     Call RS_toR(Psi, Work) 
     If (.NOT. PDOT_Stored) then
        Call CalcProjProducts(Work, PDot)
        PDOT_Stored = .TRUE.
     End If

     RS_Work1 = V_Weight*Ve*Work
     Call AccumBothProj(RS_Work1, Work, PDot, DijMatrix, OijMatrix)

     RS_Work2 = Work   !** Copy Ox into a temp space
     Ox = 0
     Call RS_toG(Work, Ox)

     Work = 0
     Call RS_toR(Hx, Work) !** Make sure we add the KE in RS  
     Work = Work + RS_Work1 - Eo*RS_Work2  !** Start the shift

     Call RS_toG(RS_Work1, Hx)
!HOiHx = 0
!Call RS_toG(Work, HOiHx)
!RETURN

     !*** Now calculate HOiHx ***
     Call CalcOinverse(Work, kpnt)   !** Work = OiHx after call
 
     RS_Work1 = Work    !** Need OiHx in both RS and GS

     Call Getbuffer( W2)
     W2 = 0
     Call RS_toG(Work, W2)
     Call CalcDEL2_Phase( BZ%Ku(:,Kpnt), W2, HOiHx)
     HOiHx = -HOiHx
     Call Freebuffer(W2)

     Call CalcProjProducts(RS_Work1, PDot_OiHx)

     RS_Work1 = V_Weight*Ve*RS_Work1 - Eo*RS_Work1
     Work = 0
     Call AccumBothProj(RS_Work1, Work, PDot_OiHx, DijMatrix, OijMatrix)

     RS_Work1 = RS_Work1 - Eo*Work
     Call RS_toG(RS_Work1, HOiHx)

     HOiHx = HOiHx/Ewidth**2   !** Re-scale the answer     
  else
     Call V_G_to_r(Psi,G_LOW,Work,G_HIGH)
     
     Work = V_Weight*Ve*Work
     Call V_r_to_G(Work,G_HIGH,Hx,G_LOW)

     If (.NOT. PDOT_Stored) then
        Call CalcProjProducts(Psi, PDot)
        PDOT_Stored = .TRUE.
     End If

     Ox = Psi
     Call AccumBothProj(Hx, Ox, PDot, DijMatrix, OijMatrix)

     !*** Now calculate HOiHx ***
     HOiHx = Hx - Eo*Ox   !** Shift the spectrum
     Call CalcOinverse(HOiHx, kpnt)

!     Call GetBuffer( W3)
!     W3 = HOiHx
     Call CalcProjProducts(HOiHx, PDot_OiHx)

     Call GetBuffer( W2)
     Call CalcDEL2_Phase( BZ%Ku(:,Kpnt), HOiHx, W2)

     Call V_G_to_r(HOiHx,G_LOW,Work,G_HIGH)
     Work = V_Weight*Ve*Work
     HOiHx = -W2 - Eo*HOiHx
     Call V_r_to_G(Work,G_HIGH,HOiHx,G_LOW)
     Call PerformFFT(FFT_TO_G, G_HIGH, Work)

     W2 = 0
     Call AccumBothProj(HOiHx, W2, PDot_OiHx, DijMatrix, OijMatrix)

     HOiHx = HOiHx - Eo*W2

     HOiHx = HOiHx/Ewidth**2   !** Re-scale the answer

     Call FreeBuffer(W2)
  End If

  DeAllocate(PDOT_OiHx)

  Return
End subroutine

!******************************************************************************
!
!  CalcMagAndEnergy - Calculates the energy and magnitude of the given band
!
!  PsiBand - Wave vector to multiplies band
!  Energy  - Wave function energy
!  Mag     - Magnitude of the wave function
!
!******************************************************************************
! This routine is not compatable with spin 

!Subroutine CalcMagAndEnergy( PsiBand, Energy, Mag)
!  Integer,          Intent(IN)    :: PsiBand
!  Real,     Intent(OUT)   :: Energy
!  Real,     Intent(OUT)   :: Mag

!  Complex, Pointer :: Hx(:), Psi(:), KE(:), Ox(:)
 !Complex, TARGET  :: PDot(PLM_Max)
!  Complex, TARGET , allocatable  :: PDot(:)
!  Integer :: Kpnt
  

!  Allocate(PDot(PLM_Max))
!  Call GetBuffer( Hx)
!  Call GetBuffer( Psi)
!  Call GetBuffer( Ox)

!  Kpnt = PsiInfo(PsiBand)%Kpnt

!  Call GetPsi_fromBuffer( PsiBand, Psi)


!  Call CalcHxandOx( Psi, Hx, Ox, &
!       PsiInfo(PsiBand)%PDOT, Kpnt, PSiInfo(PsiBand)%PDOT_Stored)

!  Energy = Basis_DotProd( Psi,Hx, .FALSE.)
!  Mag = Basis_DotProd( Psi,Ox, .FALSE.)

!  Call Freebuffer(Psi)
!  Call FreeBuffer(Hx)
!  Call FreeBuffer(Ox)
!  DeAllocate(PDot)

!  Return
!End Subroutine

!******************************************************************************
!
!  CalcEigenValues - Calculates the Energy eigenvalues for each 
!              wave function 
!
!******************************************************************************
Subroutine CalcEigenValues

 !Integer :: Psi_toUse(Mem_MapSize), i, Flag, Kpnt
  Integer :: i, Flag, Kpnt
  Integer ,allocatable:: Psi_toUse(:)
  Complex, Pointer  :: Hx(:)
  Complex, Pointer  :: Ve(:)
  Real              :: DV
  Type(Mem_Handle), Pointer :: Psi
 !Complex           :: PDot(PLM_Max)
  Complex ,allocatable :: PDot(:)


  Allocate(Psi_toUse(Mem_MapSize),PDot(PLM_Max))
  Call Getbuffer( Hx)
if(.NOT.spindependence) then
  Do Kpnt=1, NumKpnts
     Psi_toUse = MH_Skip
     Where (PsiInfo(:)%Kpnt == Kpnt) Psi_toUse = MH_toProcess
     Call Phase_Generic( Psi_toUse,  Kpnt)

     Call GetNextPsi(Flag, Psi)
     Do While (Flag > 0) 
        i = Psi%Index
        PsiInfo(i)%DoSave=1
        PsiInfo(i)%Available=Mem_Used
        Call Hpsi( Psi%Ptr, Hx, SCFValues%Ve, PsiInfo(i)%PDot, Kpnt, &
             PsiInfo(i)%PDOT_Stored)
        PsiInfo(Psi%Index)%Energy = Basis_DotProd(Psi%Ptr, Hx, .FALSE.)
        
!write(Log_Unit,*) 'CalcEigenValues: DI=',Psi%Index, ' * New=',PsiInfo(Psi%Index)%Energy
        Call GetNextPsi(Flag, Psi)
     End Do
  End do
else 
  Do Kpnt=1, NumKpnts
     Psi_toUse = MH_Skip
     Where (PsiInfo(:)%Kpnt == Kpnt ) Psi_toUse = MH_toProcess
     Call Phase_Generic( Psi_toUse,  Kpnt)

     Call GetNextPsi(Flag, Psi)
     Do While (Flag > 0) 
        i = Psi%Index
        PsiInfo(i)%DoSave=1
        PsiInfo(i)%Available=Mem_Used
        If(PsiInfo(i)%spinup) then
           Ve=>SCFvalues%Ve
        else
           Ve=>SCFvalues%Vespin
        endif
        Call Hpsi( Psi%Ptr, Hx, Ve, PsiInfo(i)%PDot, Kpnt, &
             PsiInfo(i)%PDOT_Stored)
        PsiInfo(Psi%Index)%Energy = Basis_DotProd(Psi%Ptr, Hx, .FALSE.)
        
!write(Log_Unit,*) 'CalcEigenValues: DI=',Psi%Index, ' * New=',PsiInfo(Psi%Index)%Energy
        Call GetNextPsi(Flag, Psi)
     End Do
  End do
endif

  Call FreeBuffer(Hx)
  DeAllocate(Psi_toUse,PDot)

  Return
End Subroutine

End Module
