!******************************************************************************
!
! File : hamsym.f90
!  originally from hamiltonian.f90
!   by : Alan Tackett
!   on : 02/07/97
!  for : PAW
!
! Module for calculating the Hamiltonian and Cohesive Energy of the PAW
! system.  Also contains routines to do H*Psi and O*Psi.
!
!   contains symmetry related functions
!
! Updated for spin by Ping Tang and N. A. W. Holzwarth
!      Last change 5/25/05
!******************************************************************************

Module hamsym

Use atom_data
Use basis_lib
Use crystal_data
Use crystal_symmetry
Use gpoints 
Use mathlib
Use options_data
Use paw_inout
Use spherical_harmonic
Use coulomb_pack

Implicit NONE!!!!


!******************************************************************************
Contains
!******************************************************************************
!
!  SymRho - Symmetrizes the density based on the crystal symmetry operations
!
!  Rho   - The Density to symmetrize
!
!  ** Uses plane waves **
!******************************************************************************

Subroutine SymRho( Den, Work)
  Complex , Intent(INOUT) :: Den(:)
  Complex , Intent(INOUT) :: Work(:)

  Integer :: i,j,k1,s
  Real  :: R(3), Rp(3), gcut, mag, arg
  Complex  :: zsum, zterm
  Integer :: Nx, Ny, Nz,  Gi(3), N(3), M1(3), M2(3)

  Call Start_Timer(Timer(SymRho_Timer))

  write(log_unit,*) 'SymRho:  Begin'
  call flush(log_unit)
  Nx = FFT_Grid(1,G_HIGH)
  Ny = FFT_Grid(2,G_HIGH)
  Nz = FFT_Grid(3,G_HIGH)

  !Work = 0
  !Call PerformFFT(FFT_TO_G, G_HIGH, Den, Work)
  !write(6,*) 'after PerformFFT', Work(1)

  Work = Den
  Call PerformFFT(FFT_TO_G, G_HIGH,  Work)
  write(Log_Unit,*) 'SymRho: Start Q=', Work(1)
  call flush(log_unit)

  Den = 0

  Den(1) = Work(1)

  Do i=2, Gpnt_Size(G_HIGH)
        Gi = Npnt(:,i)

        Zsum = 0 
        Do j=1, Xtal%Rot_Size
           N = MatMul(Igrot(:,:,j),Gi)
           arg = DOT_PRODUCT(N, xtal%Trans(:,j))
           If (ABS(Arg) > 1E-8) then
              arg = Two_PI*arg
              zterm = CMPLX(cos(arg),sin(arg))
           else
              zterm = CMPLX(1,0)
           End If
           
           M1 = N + 1
           Where (M1<=0) M1 = M1 + FFT_Grid(1:3,G_High)
           IF ((M1(1)<=Nx) .AND. (M1(1)>0) .AND. &
                (M1(2)<=Ny) .AND. (M1(2)>0) .AND. &
                (M1(3)<=Nz) .AND. (M1(3)>0)) then
                 k1 = CalcLinearIndex(FFT_Grid(:,G_HIGH), M1)+1
                 zsum = zsum + Work(k1)*zterm
             !write(6,'("SymRho",12i4)') i,j,Gi,N,M1,k1
            else
              write(error_unit,*) 'Symrho: bad rotation',j,Gi,N
              STOP
           End If
        End Do

           zsum = zsum/xtal%Rot_Size
           Den(FFTmap_High(i)) = zsum
             !write(6,*) 'sum',FFTmap_High(i),zsum
           Den(FFTmap_High(i+Gpnt_Size(G_High)-1)) = CONJG(zsum)
             !write(6,*) 'sum*',FFTmap_High(i+Gpnt_Size(G_HIGH)-1),CONJG(zsum)
  End Do

  write(Log_Unit,*) 'SymRho: After SymRho Q=', Den(1)
  Call PerformFFT(FFT_TO_R, G_HIGH, Den)

!  write(6,*) 'in SymRho sum', sum(den)
  write(log_unit,*) 'in SymRho sum', sum(den)
  call flush(log_unit)

  Call Stop_Timer(Timer(SymRho_Timer))

  Return
End Subroutine


!******************************************************************************
!
!  SymWij - Symmetrizes the Wij coefficients
!
!******************************************************************************

Subroutine SymWij
  type Wij_Atom
     Complex, Pointer :: Wij(:,:)
  End Type

  Integer :: i,j,k, nili, njlj, Li, Lj,mi1, mj1, mi2, mj2, Basis_Size
  Integer :: ibase, jbase
  Real    :: c
  Complex :: zsum
  Complex, Pointer :: Wij(:,:), Wij_Copy(:,:)
  Type (Wij_Atom) :: Wij_List(Specific_Atoms)
  Type (Specific_Atom), Pointer :: A
  Type (Atom_Info_Fixed), Pointer :: AT
  Character*100 :: Msg

  write(log_unit,*) 'SymWij: Begin'
  call flush(log_unit)

  c = 1.0/XTal%Rot_Size

  Do i=1, Specific_Atoms          !**** Make a temp copy of Wij ****
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)
     
     j=AT%nlm_Size
     Allocate(Wij_List(i)%Wij(j, j), STAT=k)
     Write(MSg,*) "SymWij: Error Allocating Temp array for atom:",i
     Call Check_Error(k, Msg, Error_Unit, .TRUE., IDTEXT="SymWij:")
     Wij_List(i)%Wij = A%Wij
  End Do

  Do i=1, Specific_Atoms        !*** Symmetrize Wij
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)
     
     Basis_Size = AT%Basis_Size     
     Wij => A%Wij

     Do nili=1, Basis_Size
        Li = AT%L_Value(nili)
        ibase = AT%nl_Base(nili)+Li
        Do mi1=-Li, Li
           Do njlj=1, Basis_Size
              Lj = AT%L_Value(njlj)
              jbase = AT%nl_Base(njlj)+Lj
              Do mj1=-Lj, Lj

                 zsum = 0
                 Do j=1, XTal%Rot_Size
                    Wij_Copy => Wij_List(iatrans(i,j))%Wij

                    Do mi2=-Li, Li
                       Do mj2=-Lj, Lj
                         zsum= zsum + (zarot(mi2+Li+1,mi1+Li+1,Li+1,j)) &
                               *CONJG(zarot(mj2+Lj+1,mj1+Lj+1,Lj+1,j)) &
                               *Wij_copy(ibase+mi2, jbase+mj2)
                       End Do
                    End Do

                 End Do 
                 Wij(ibase+mi1, jbase+mj1) = zsum*c
              End Do
           End Do    !**njlj
        End Do
     End Do    !**nili
  End Do

  write(Log_unit,*) 'SymWij:  Finish non spin part'
  call flush(log_unit)

  write(Log_unit,*) 'spindependence', spindependence
  call flush(log_unit)
if(spindependence) then
  Do i=1, Specific_Atoms          !**** Make a temp copy of Wij ****
     A => Atom_List(i)
     Wij_List(i)%Wij = A%Wijspin
  End Do

  Do i=1, Specific_Atoms        !*** Symmetrize Wij
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)
     
     Basis_Size = AT%Basis_Size     
     Wij => A%Wijspin

     Do nili=1, Basis_Size
        Li = AT%L_Value(nili)
        ibase = AT%nl_Base(nili)+Li
        Do mi1=-Li, Li
           Do njlj=1, Basis_Size
              Lj = AT%L_Value(njlj)
              jbase = AT%nl_Base(njlj)+Lj
              Do mj1=-Lj, Lj

                 zsum = 0
                 Do j=1, XTal%Rot_Size
                    Wij_Copy => Wij_List(iatrans(i,j))%Wij

                    Do mi2=-Li, Li
                       Do mj2=-Lj, Lj
                         zsum= zsum + (zarot(mi2+Li+1,mi1+Li+1,Li+1,j)) &
                               *CONJG(zarot(mj2+Lj+1,mj1+Lj+1,Lj+1,j)) &
                               *Wij_copy(ibase+mi2, jbase+mj2)
                       End Do
                    End Do

                 End Do 
                 Wij(ibase+mi1, jbase+mj1) = zsum*c
              End Do
           End Do    !**njlj
        End Do
     End Do    !**nili
  End Do
  write(Log_unit,*) 'SymWij:  Finish spin part'
  call flush(log_unit)
End if

  Do i=1, Specific_Atoms        !*** Free Wij Copy
     DeAllocate(Wij_List(i)%Wij)
  End Do

  Return
End Subroutine

!******************************************************************************
!
!  SymWijkl - Symmetrizes the Wijkl coefficients
!         for self_energy terms
!      not used
!******************************************************************************

!Subroutine SymWijkl
!
!  Integer :: i,j,k, nili, njlj, Li, Lj,mi, mj, mi2, mj2
!  Integer :: nklk,nlll,mk,ml,mk2,ml2,Lk,Ll,p,pp
!  Integer :: count
!  Integer, pointer :: LUT(:)
!  Real    :: c
!  Complex :: zsum
!  Complex, Pointer :: Wijkl(:), Wijkl_Copy(:)
!  Type (Atom_Info_Fixed), Pointer :: AT
!  Character*100 :: Msg
!
!  write(log_unit,*) 'SymWijkl: Begin'
!  call flush(log_unit)
!
!  c = 1.0/XTal%Rot_Size
!
!  Do i=1, Atom_Types          !**** Make a temp copy of Wijkl ****
!     AT => AtomType_Info(i)
!     
!     j=AT%Cijkl_Size
!     Allocate(Wijkl_Copy(j), STAT=k)
!         if (k/=0) then
!            write(Error_unit,*) "SymWijkl: Error Allocating array for atom:",i
!            stop
!         endif
!     Wijkl =>AT%Wijkl
!     Wijkl_Copy = Wijkl
!     Wijkl=0
!     LUT => AT%LUT_Cijkl
!
!     do k=1,AT%Cijkl_Size
!        write(Log_Unit,*) k,LUT(k)
!     enddo
!     do p=1,AT%Cijkl_Size
!        Call Coulomb_Decode(LUT(p), nili, njlj, nklk, nlll, mi, mj, mk)
!        ml = mi - mj + mk
!        li=AT%L_Value(nili)
!        lj=AT%L_Value(njlj)
!        lk=AT%L_Value(nklk)
!        ll=AT%L_Value(nlll)
!        if (abs(ml)<=ll) then
!
!        zsum=0
!           Do  mi2=-li,li
!               Do mj2=-lj,lj
!                  Do mk2=-lk,lk
!                     ml2=mi2-mj2+mk2
!                     if (abs(ml2)<=ll) then
!                     k=Coulomb_Encode(nili,njlj,nklk,nlll,mi2,mj2,mk2)
!                     pp=0
!                     do count=1,AT%Cijkl_Size
!                        if (LUT(count)==k) then
!                           pp=count
!                           exit
!                        endif
!                     enddo
!                     if (pp>0.and.pp<=AT%Cijkl_Size) then
!                       Do j=1, XTal%Rot_Size
!                         zsum= zsum + (zarot(mi2+Li+1,mi+Li+1,Li+1,j)) &
!                               *CONJG(zarot(mj2+Lj+1,mj+Lj+1,Lj+1,j)) &
!                               *(zarot(mk2+Lk+1,mk+Lk+1,Lk+1,j)) &
!                               *CONJG(zarot(ml2+Ll+1,ml+Ll+1,Ll+1,j)) &
!                               *Wijkl_Copy(pp)
!                       enddo
!                     else
!                        write(Log_unit,*) 'Error in SymWijkl', li,lj,lk,ll,&
!                              mi2,mj2,mk2,ml2,j,k,pp
!                        stop
!                     endif
!                   endif
!                  enddo
!               enddo
!             enddo
!         Wijkl(p)=zsum*c
!         endif
!     Enddo
!     
!     DeAllocate(Wijkl_Copy)
!  Enddo
!
! End subroutine SymWijkl
End module
