!******************************************************************************
!
! File : initsystem.f90
!   by : Alan Tackett
!   on : 7/30/98
!  for : PAW Project
!
! InitSystem - Performs the One time initialization for the PAW system
!
! Modified for spin by Ping Tang and N. A. W. Holzwarth
!      Last change 5/23/05
!******************************************************************************

Subroutine InitSystem
  Use paw_inout
  Use misc
  Use bz_data
  Use psilib
  Use options_data
  Use crystal_data
  Use atom_data
  Use spherical_harmonic
  Use mathlib
  Use hamiltonian
  Use crystal_symmetry
  Use mem_data
  Use memmgr
  Use solver
  Use relaxsys
  Use gpoints
  Use fileio
  Use structfact
  USe ylm_fact
  Use oinverse
  Use ldatom_info
  Use local_criteria_lib
  Use exchange_corr

  Implicit NONE!!!!!!!

  Character*100 :: Msg
  Integer       :: KGrid(3), FinalGrid(3), i, aerr, Lmax, j,Npts,a,many
  Real          :: t , s
  Real    :: PC(3), R(3), G(3), theta, error, t1, Exc, Vxc, den,grad
  Logical       :: Ok
  Integer :: k,n, Oldk,bufloaded, m, u,v, x,y,z,PM(3), mindex, h1,h2,h3,p
  Type (Atom_Info_Fixed), Pointer :: AT



  Call Start_Timer(Timer(Setup_timer))

  Call PrintDate(Log_Unit, 'InitSystem: Paw Start!')

  Call Init_FIleIO    !*** Initialize IO routines ***

  Vxc_MinSize = 40

  PAW_FirstTime = .TRUE.

  if (Print_Level == PRINT_VERBOSE) then
     write(LOG_Unit, *) 'InitSystem:  Starting System Initialization.***************'
  End If

  If (Atomic_Mode == SIM_SEPM) V_MixType = MIX_VEFF

  !** Generate all the atoms and their positions **
  Call Clone_Cell(xtal%CloneCell) 
  
  If (UseShells) then
     Write(Log_Unit,*) 'Initializing Shell Parmameters'
    
      Do p = 1,Shell%NPARAMS
      n=0
      t = 0       !Shell%C(p) = 0
      Do i=1,Specific_Atoms
         If (Shell%AtomMap(i)%MapParams>0) then
            Do j=1,Shell%AtomMap(i)%MapParams
               If (Shell%AtomMap(i)%Map(j)==p) then
                  n=n+1
                  t=t+Dot_Product(Shell%AtomMap(i)%IV(:,j), &
                     (Atom_List(i)%Pos-Shell%AtomMap(i)%Origin))
               EndIf
            Enddo
         EndIf
       Enddo

       If (n < 1) then
          Write(Error_Unit,*) 'Error in InitSystem -- Shell parameter',&
               p
          Stop
       EndIf

       Shell%C(p) = t/n
      
       Write(Log_Unit,'(" Shell param ", i5," = ", 1pe15.7)')p,Shell%C(p)
      EndDo       

    ! check positions
    Do a=1,Specific_Atoms
       many = Shell%AtomMap(a)%MapParams
       If (many > 0) then
          R = Shell%AtomMap(a)%Origin
          Do i=1,many
             p=Shell%AtomMap(a)%Map(i)
             R = R + Shell%C(p)*Shell%AtomMap(a)%V(:,i)
          Enddo
          G=R-Atom_List(a)%Pos
          s=Dot_Product(G,G)
          if (s>1.e-8) then
            write(Error_unit,*) 'Error in Shell for atom', a
            write(Error_unit,*) 'Correct position:', Atom_List(a)%Pos
            write(Error_unit,*) 'Fract   position:', Atom_List(a)%Frac_Pos
            write(Error_unit,*) 'Shell   position:', R
            stop
          endif
        EndIf
        Write(Log_Unit,*) 'Shell parameters check for atom ', a
     Enddo


  EndIf

  Call CheckAtomOverlap

  !*** Next make the AtomType_LUT and AtomType_Range arrays ***
  Allocate(AtomType_MAP(Max_Specific_Atoms), AtomType_Range(2,Max_Atom_Types))
  k = 0
  Do i=1, Atom_Types
     AtomType_Range(1,i) = k + 1
     Do j=1, Specific_Atoms
        If (Atom_List(j)%TypeIndex == i) then
           k = k + 1
           AtomType_MAP(k) = j
        End If
     End Do
     AtomType_Range(2,i) = k
  End Do

  Call CountBands                !*** Count the Bands

  If (NumBands < (User_MinPsi+1)) then 
     write(LOG_Unit, *) 'InitSystem:  NumBands < MinBands! Changing ', &
          ' NumBands form', NumBands, ' to ', User_MinPsi+1
     NumBands = User_MinPsi+1
  end If

  if (Print_Level == PRINT_VERBOSE) then
     write(LOG_Unit, *) 'InitSystem:  Total Number of Bands :', NumBands
     write(LOG_Unit, *) 'InitSystem:  Total Number of Electrons :', TotalElectrons
  End If


      !************* BZ Setup ****************
  If (.NOT. Xtal_Defined) then
     Write(Error_Unit, *) 'InitSystem: Crystal not Defined! Exiting!'
     Call Word_GetAndPrint(PAW_WC, Error_Unit, 'InitSystem:')
     !Call SendQuit
     STOP
  End If

  if (Print_Level == PRINT_VERBOSE) then
     write(LOG_Unit, *) 'InitSystem:  Generating the BZ K-point grid.'
  End If

  BZ%Bands = NumBands
  Call bz_Set_All(bz,Xtal%BZ_Method,0.5,TotalElectrons, 0.0, 1.0, 1E-15, &
                      200, Xtal%Sigma)

  If (Associated(BZ%Ku)) then         !** Ignoring Kpnts grid using the user
                                      !** supplied list
    if(.not.spindependence) then
     Allocate(BZ%Keu(NumBands,BZ%TotalUniq), &
              BZ%WtUniq(NumBands,BZ%TotalUniq), &  
              STAT=aerr)   
    else
     Allocate(BZ%Keu(2*NumBands,BZ%TotalUniq), &
              BZ%WtUniq(2*NumBands,BZ%TotalUniq), &
              STAT=aerr)  
    end if
     if (aerr /= 0) then
        Write(Error_Unit, *) "InitSystem: Error Allocating BZ info! Error = ", aerr
        Call Word_GetAndPrint(PAW_wc, Error_Unit, 'InitSystem(BZ): ')
        !Call SendQuit
        STOP
     End If

     t = SUM(XTal%Wt)              !** Make sure the weights are normalized
     xtal%Wt = Xtal%wt/ t
     Do i=1, NumBands
        BZ%WtUniq(i,:) = Xtal%Wt
        If (spindependence) BZ%WtUniq(i+NumBands,:) = Xtal%Wt
     End Do

     DeAllocate(Xtal%Wt)      !*** Free the Temporary Wt buffer

     NumKpnts = BZ%TotalKpnts
  else                     !*** Generate the List of BZ K-Points
     write(Log_unit,*)' K-point generation not working'
     stop
     BZ%Weight => BZ%WtUniq;
     Kgrid = Xtal%Kpnts_Grid - 1
     Call BZ_MakeGrid(BZ,Xtal%Recip,(/0.0,0.0,0.0/),(/1.0,1.0,1.0/),NumBands, &
          KGrid, (/2,2,2/), 1, Xtal%Volume)

     Call BZ_Set(BZ, BZ_SET_ELECTRONS, TotalElectrons)

     if (Print_Level == PRINT_VERBOSE) then
        write(LOG_Unit, *) 'InitSystem:  Determining Unique BZ K-points.'
     End If
     i = 0 !** Must call Twice first time det uniq size and second stores them 
     Call BZ_UniqueKpnts(BZ, Xtal%Basis, Xtal%RotMatrix, Xtal%Rot_Size, i)
     Call BZ_UniqueKpnts(BZ, Xtal%Basis, Xtal%RotMatrix, Xtal%Rot_Size, i)
     NumKpnts = i
  End If

  if (Print_Level == PRINT_VERBOSE) then
     write(LOG_Unit,*) 'InitSystem: Number of Unique BZ K-points:',NumKpnts
     Write(LOG_Unit,*)'InitSustem: Generating the Angular integration points'
  End If

  if(.not.spindependence) Allocate(Occupancy(NumBands, NumKpnts), &
               BKMap(NumBands,NumKpnts),STAT=i)
  if(spindependence) Allocate(Occupancy(2*NumBands, NumKpnts), &
               BKMap(2*NumBands,NumKpnts),STAT=i)

  Write(Msg,*) "InitSystem: Can't alloc space Occupancies!"
  Call Check_Error(i, MSg, Error_Unit, .TRUE., PAW_WC, "InitSystem:")

 ! Check for spindependence and if necessary, reset exchange-correlation
 !    parameters
  if (.not.XC_TYPE_SET) then
     write(Error_Unit,*) 'InitSystem: Warning XC_Type not set -- default used'
     XC_TYPE_SET=.true.
     XC_TYPE=XC_LDA_PW
  endif

  if (Spindependence) then
     If (XC_TYPE == XC_LDA_PW) then
         XC_TYPE =  XC_LSDA_PW
         write(Log_Unit,*) 'InitSystem: Spin restricted calculation'
         write(Log_Unit,*) 'InitSystem: XC_TYPE = LSDA_PW'
     Endif
     If (XC_TYPE == XC_GGA_PBE) then
         XC_TYPE =  XC_SGGA_PBE
         write(Log_Unit,*) 'InitSystem: Spin restricted calculation'
         write(Log_Unit,*) 'InitSystem: XC_TYPE = SGGA_PBE'
     Endif
  endif

 ! If (Atomic_Mode /= SIM_SEPM) then - this part moved to PrepareVxc-- ********
 ! !**** Create the Core XC-- this part moved to PrepareVxc-- ********
    !    If (XC_Type == XC_LDA_PW) then
    !     Do i=1, Atom_Types
    !       AT => AtomType_Info(i)
    !       Do j=2, AT%Mesh_Size
    !          den = AT%Core_Density(j)
    !          !!Call ExchangeCorr(den, Vxc, Exc, .FALSE., XC_Type)
    !          If (XC_Type == XC_LDA_PW) Call pwldafunc(den,Exc,Vxc)
    !          AT%Core_Xchange(j) = Exc
    !       End Do
    !       AT%Core_Xchange(1) = 2*AT%Core_Xchange(2)-AT%Core_Xchange(3)
    !     End Do
    !    Else If (XC_Type == XC_GGA_PBE) then
    !     Do i=1,Atom_Types
    !       AT => AtomType_Info(i)
    !       Npts=AT%Mesh_Size
    !       write(6,*) 'in initsystem',Npts,SIZE(AT%Core_Density),SIZE(AT%Core_dDendr)
    !       call flush(6)
    !       Call nderiv(AT%Mesh_Step,AT%Core_Density(:),AT%Core_dDendr(:),Npts)
    !       Do j=1,Npts
    !        grad=ABS(AT%Core_dDendr(j))     ! stores gradient
    !        Call  pbefunc(AT%Core_Density(j),grad,AT%Core_Xchange(j),s,t)
    !                  !! Remember: for GGA Core_Xchange is fxc
    !       Enddo
    !     Enddo
    !    Else
    !      Write(Error_Unit, * ) 'Initsystem:  Error in XC_Type', XC_Type
    !      Stop
    !    EndIf
    !
!
 !    !************ Generate the Angular Integration points ***************
  !   !Call InitSpharm
   !  Num_AngPoints = Angular_Points**2
    ! AngPoints => AngInit(Angular_Points,Angular_Points,Xtal%Volume,Error_Unit)
 ! End If


      !******* Initialize the Crystal_Symmetry routines **********
  Lmax = -1
  Do i=1, Atom_Types
     j = MaxVal(AtomType_Info(i)%L_Value)
     Lmax = MAX(Lmax, j)
  End Do

  If (Print_Level <= PRINT_NORMAL) then
     Write(Log_Unit,*) 'InitSystem: Calling SetSym with Lmax=',Lmax
  End If

  If (xtal%Auto_Symmetry) Call StoreSymmetry(Sym_Name)
  If (xtal%Rot_Size > 0) Call SetSym(Lmax)
 

  !** Adjust the projector Gcut based on the mode ***
!  PW_Gcut(G_PROJ) = PW_Gcut(G_LOW)
  If (Proj_Mode == 0) then
     Proj_Mode = Run_Mode
     PW_Gcut(G_PROJ) = PW_Gcut(G_LOW)
  End If
  !*** Initialize the G-points **** 
  Call InitGpoints(Grid, xtal%Recip, PW_Gcut(G_HIGH), PW_Gcut(G_LOW), &
         PW_Gcut(G_PROJ), paw_WC)

  call InitkplusGfilter(Numkpnts)
  PsiArraySize=Gall_Size(G_LOW)


  Call AssignBaseFileUnits(FD_Base)               


  !***** Initialize the Memory Management routines ******
  Call InitWorkArrays(10)
  Call ReAllocate_Work(Gall_Size(G_LOW))

  !***** Initialize the Memory Management routines ******
  Call InitMemMgr(Psi_Memory)

  !*** Initialize the Local selection criteria map ***
  Call InitLocalCriteria


  !**** Initialize the  solver routines *******
  Call InitSolver

  !***** Initialize the Structure factor routines ******
  Call InitStructFactors(FD_BlochK, Bloch_Memory)

  If (Atomic_Mode /= SIM_SEPM) then
     !***** Initialize the Ylm factor routines ******
     j = MAXVal(AtomType_Info(1)%L_Value)
     Do i=2, Atom_Types
        j = Max(j,MAXVal(AtomType_Info(i)%L_Value))
     End Do
     
     Call InitYlmFact(FD_Ylm, Ylm_Memory, j)
     
     !**** Initialize the Radial projectors *****
     If (Proj_Mode == 0) Proj_mode = Run_Mode
     Call InitProjectors(FD_RadProj, FD_Proj, Proj_Memory)

     If (Eigen_Mode == EIGEN_ENERGY) Call InitOinverse(FD_Oinv, 50.0)
!Call InitOinverse(FD_Oinv, 50.0)
  end If

  !********* Setup Wave Function Buffers *********
  if (Print_Level == PRINT_VERBOSE) then
     write(LOG_Unit, *) 'InitSystem:  Initializing the PSI routines.'
  End If

  write(Log_unit,*) 'InitSysem:  Before InitPsiLib'
  call flush(Log_Unit)
  Call InitPsiLib
  write(Log_unit,*) 'InitSysem:  After  InitPsiLib'
  call flush(Log_Unit)
 

  If (Atomic_Mode /= SIM_SEPM) then
      !********** Calculate the Base Energy **********
     CohesiveEnergy = 0
     TotalEnergy = 0
     AtomicEnergy = 0
     Do i=1, Specific_Atoms
        AtomicEnergy = AtomicEnergy + & 
             AtomType_Info(Atom_List(i)%TypeIndex)%Atom_Energy
     End Do
  
     if (Print_Level == PRINT_VERBOSE) then
        write(LOG_Unit, *) 'InitSystem:  Total Atomic Energy :', AtomicEnergy
     End If
  End If

      !******* Initialize the Hamiltonian Routines **********
  If (Print_Level <= PRINT_NORMAL) then
     Write(Log_Unit,*) 'InitSystem: Initializing Hamiltonian Module.'
     call flush(Log_Unit)
  End If
  Call InitHam

  write(log_unit,*) 'InitSytem: Finished InitHam -- start alloc PDOT'
  call flush(log_unit)

  !!!Call Init_relax  !** Initialize relax routines **


  If (Atomic_Mode == SIM_SEPM) PLM_Max = 0
  !*** Allocate Space for The PDOT's for each Psi ***

  write(Log_unit,*) 'PLM_Max = ', PLM_Max
  call flush(log_unit)

  Do i=1, Mem_MapSize
     PsiInfo(i)%PDot_Stored = .FALSE.
     Allocate(PsiInfo(i)%PDOT(PLM_Max))
  End Do
  
  PAW_FirstTime = .FALSE.

!  Do i=1, Atom_types    !*** DEBUG ONLY ***
!     ATomType_Info(i)%Oij = 0
!  End Do

!  if (Print_Level == PRINT_VERBOSE) then
!     write(LOG_Unit, *) 'InitSystem:  Transferring data to nodes'
!  End If  

  Matrix_Range = RANGE_DO
  
  !***** Initialize ploting data****
   volumeplot%query=0
   volumeplot%bondtol=0
   planeplot%query=0

  !** Tell everybody else it's ok to continue **
!**  Call MPI_BCAST(CMD_NEXT, 1, DATA_INTEGER, Root_Rank, PAW_COMM, i)
!**  Call InitSystem_MPI

  if (Print_Level == PRINT_VERBOSE) then
     write(LOG_Unit, *) 'InitSystem:  Finished System Initializiation!***************'
  End If  
  
     write(LOG_Unit, *) 'InitSystem:  Finished System Initializiation!***************'
    call flush(Log_unit)
  Call Stop_Timer(Timer(Setup_timer))
  Return
End Subroutine



