!******************************************************************************
!
! File : lcao.f90
!   by : Alan Tackett
!   on : 02/18/97
!  for : PAW project
!
!  This module contains routines to generate the initial wave function for
!  use in the PAW program.  The wave function is constructed from a linear
!  combination of atomic orbitals and Bloch wave vectors.  The inital 
!  occupancies are also calculated.
!
!******************************************************************************

Module lcao

Use paw_inout
Use word
Use atom_data
Use misc
Use crystal_data
Use bz_data
Use psilib
Use options_data
Use spherical_harmonic
Use mem_data
USe memmgr
Use projectors
Use debug
Use hamiltonian
Use work_mgr
Use gpoints

Implicit NONE!!!!!

!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
!  LCAO_Interpolate - Interpolates the Radial LCAO function with a third
!                     order polynomial around the requested point and
!                     returns the iterpolated value
!
!  TPhi      - Radial Basis function
!  Mesh_Size - Number of radial grid points
!  Mesh_Step - Real space distance between mesh points
!  R         - R-value to interpolate
!
! Not properly fixed for loggrid
!******************************************************************************

Real Function LCAO_Interpolate(Tphi, Mesh_size, Mesh_Step, R)
  Real,    Intent(IN) :: tphi(:)
  Integer, Intent(IN) :: Mesh_Size
  Real,    Intent(IN) :: Mesh_Step
  Real,    Intent(IN) :: R

  Real    :: fn, a,b,c,d, h1,h2,h3, y0,y1,y2,y3, x
  Integer :: i,j,k

  write(Log_unit,*) 'Function LCAO_Interpolate needs fixing '
   stop
  i = R/Mesh_Step + 1 - 1        !*** We want 2 pnts to the left and right of R

  If (i>=Mesh_Size) then          !*** If too far away then return 0
     fn = 0
  else 
     IF (i<=0) then              !*** Make sure Index is good
        i = 1
     else if (i>Mesh_Size-3) then
        i = Mesh_Size - 3
     End If

     h1=Mesh_Step; h2=h1*h1;     h3=h1*h2;
     y0 = TPhi(i); y1=TPhi(i+1); y2=TPhi(i+2); y3=TPhi(i+3);

     d = y0
     c = (18*y1 - 9*y2 + 2*y3 - 11*y0)/(6*h1)
     b = (-y3 + 2*y0 - 5*y1 + 4*y2)/(2*h2)
     a = (-3*y2 + y3 - y0 + 3*y1)/(6*h3)

     x = R - (i-1)*Mesh_Step
     Fn = d + x*(c + x*(b + a*x))
  End If
  
  LCAO_Interpolate = fn

  Return
End Function

!******************************************************************************
!
!  CalcLCAO - Generates an LCAO wave function given the radial basis function
!             and associated spherical harmonic
!
!  psi       - Wave function calculated(Returned)
!  tphi      - Radial Basis function
!  Mesh_Size - Number of radial grid points
!  Mesh_Step - Real space distance between mesh points
!  L,M       - Spherical harmonic L and M values
!  Kvec      - Bloch K point
!  Ra        - Atom position
!  Cells     - Array determining which cells to sum over
!  G_Size    - Max G index to use
!
!*OLD*
!******************************************************************************

Subroutine CalcLCAO_OLD(psi,tphi, mesh_size, mesh_step,L, M, Kvec, Ra, &
     Cells, G_Size)
  Complex, Intent(OUT) :: psi(:)
  Real,    Intent(IN)  :: tphi(:)
  Integer, Intent(IN)  :: Mesh_Size
  Real,    Intent(IN)  :: Mesh_Step
  Integer, Intent(IN)  :: L
  Integer, Intent(IN)  :: M
  Real,    Intent(IN)  :: Kvec(:)
  Real,    Intent(IN)  :: Ra(:)
  Integer, Intent(IN)  :: Cells(:)  
  Integer, Intent(IN)  :: G_Size

  Integer :: cx, cy, cz, px, py, pz, t, i, j, N
  Real    :: G(3), RadG, theta, Gmag
  Complex :: Phase
  Complex :: Ylm(13)

  Write(Log_Unit,*) "CalcLCAO_OLD is no longer working"
  stop

!!  write(Log_Unit,*) 'CalcLCAO: Kvec=',Kvec, ' Ra=',Ra
  Psi = CMPLX(0,0)

  N = Gpnt_Size(G_LOW) - 1
  Do i=1, G_Size
     G = Kvec+Gpnt(1:3,i)
     Gmag = SQRT(DOT_PRODUCT(G,G))
     theta = DOT_PRODUCT(Gpnt(1:3,i), Ra)
     Phase = CMPLX(cos(theta), -sin(theta))
     
     Ylm = Spharm(G(1),G(2),G(3), L, .TRUE.)
     !RadG = RadialFourier(L, Gmag, Mesh_Step, Mesh_Size, TPhi)
     Psi(i) = RadG*Ylm(L+m+1)*Phase
     
     If (i>1) then
        G = Kvec-Gpnt(1:3,i)
        Gmag = SQRT(DOT_PRODUCT(G,G))
        Ylm = Spharm(G(1),G(2),G(3), L, .TRUE.)
        !RadG = RadialFourier(L, Gmag, Mesh_Step, Mesh_Size, TPhi)
        Psi(i+N) = RadG*Ylm(L+m+1)*Conjg(Phase)
     Endif
  End Do


  Return
End Subroutine


!******************************************************************************
!
!  CalcLCAO - Generates an LCAO wave function given the radial basis function
!             and associated spherical harmonic
!
!  psi       - Wave function calculated(Returned)
!  tphi      - Radial Basis function in Rad-G
!  Mesh_Size - Number of radial grid points
!  Mesh_Step - Real space distance between mesh points
!  L,M       - Spherical harmonic L and M values
!  Kvec      - Bloch K point
!  Ra        - Atom position
!  Cells     - Array determining which cells to sum over
!  G_Size    - Max G index to use
!*NEW*
!******************************************************************************

Subroutine CalcLCAO(psi,tphi, mesh_size, mesh_step,L, M, Kvec, Ra, &
     Cells, Glist, G_Size)
  Complex, Intent(OUT) :: psi(:)
  Real,    Intent(IN)  :: tphi(:)
  Integer, Intent(IN)  :: Mesh_Size
  Real,    Intent(IN)  :: Mesh_Step
  Integer, Intent(IN)  :: L
  Integer, Intent(IN)  :: M
  Real,    Intent(IN)  :: Kvec(:)
  Real,    Intent(IN)  :: Ra(:)
  Integer, Intent(IN)  :: Cells(:)  
  Real,    Intent(IN)  :: GList(:,:)
  Integer, Intent(IN)  :: G_Size

  Integer :: cx, cy, cz, px, py, pz, t, i, j, N, Ng
  Real    :: G(3), RadG, theta, Gmag
  Complex :: Phase
  Complex :: Ylm(13)


!!  write(Log_Unit,*) 'CalcLCAO: Kvec=',Kvec, ' Ra=',Ra
  Psi = CMPLX(0,0)

  N = Gpnt_Size(G_LOW) - 1
  Ng = G_Size - 1

  Do i=1, G_Size
     G = Glist(1:3,i)
     Gmag = Glist(4,i)

     theta = DOT_PRODUCT(Gpnt(1:3,i), Ra)
     Phase = CMPLX(cos(theta), -sin(theta))
     
     Ylm = Spharm(G(1),G(2),G(3), L, .TRUE.)
     RadG = TPhi(i)
     Psi(i) = RadG*Ylm(L+m+1)*Phase
     
     If (i>1) then
        G = Glist(1:3,i+Ng)
        Gmag = Glist(4,i+Ng)

        Ylm = Spharm(G(1),G(2),G(3), L, .TRUE.)
        RadG = TPhi(i+Ng)
        Psi(i+N) = RadG*Ylm(L+m+1)*Conjg(Phase)
     Endif
  End Do

  Return
End Subroutine


!******************************************************************************
!
!  CalcRadG_LCAO - Calculates the radial FFT forms of the LCAO fns
!
!     G_Size - Max +G 
!     Kvec   - K-Point
!
!******************************************************************************

Subroutine CalcRadG_LCAO(G_Size, Kvec, Glist)
  Integer, Intent(IN) :: G_Size
  Real,    Intent(IN) :: Kvec(:)
  Real,    Intent(OUT) :: Glist(:,:)

  Integer :: a, i, nl, j, L, Mesh_size, N
  Real, Pointer :: RadG(:), TPhi(:)
  Real    :: G(3), Gmag, Mesh_Step
  Type (Atom_Info_Fixed), Pointer :: AT

  N = G_Size - 1

  Do a=1, Atom_Types
     AT => AtomType_Info(a)
     Mesh_Step = AT%LCAO_Step
     Mesh_Size = AT%LCAO_Size

     Do nl=1, AT%Basis_Size
        RadG => AT%RadG_LCAO(:,nl)
        TPhi => AT%TPhi_LCAO(:,nl)
        L = AT%L_Value(nl)

        Do i=1, G_Size
           G = Kvec+Gpnt(1:3,i)
           Gmag = SQRT(DOT_PRODUCT(G,G))
           Glist(1:3,i) = G
           Glist(4,i) = Gmag
           RadG(i) = RadialFourier(L, Gmag, Mesh_Step, Mesh_Size, TPhi,&
               AT%RadGrid)
     
           If (i>1) then
              G = Kvec-Gpnt(1:3,i)
              Gmag = SQRT(DOT_PRODUCT(G,G))
              Glist(1:3,i+N) = G
              Glist(4,i+N) = Gmag
              RadG(i+N) = RadialFourier(L, Gmag, Mesh_Step, Mesh_Size, TPhi,&
                AT%RadGrid)
           Endif
        End Do
        
     End Do
  End Do

  Return
end Subroutine

!******************************************************************************
!
!  FreeLCAO - Frees the RadG LCAO storage
!
!******************************************************************************

Subroutine FreeLCAO
  Integer :: i

  Do i=1, Atom_Types
     DeAllocate(AtomType_Info(i)%RadG_LCAO)
  End Do
  
  Return
End Subroutine


!******************************************************************************
!
!  AllocLCAO - Allocates the RadG LCAO storage
!
!     Array_Size - Size of the array
!
!******************************************************************************

Subroutine AllocLCAO(Array_Size)
  Integer, Intent(IN) :: Array_Size

  Integer :: i

  !** Allocate space for the RadG versions of each LCAO type **
  Do i=1, Atom_Types
     Allocate(AtomType_Info(i)%RadG_LCAO(Array_Size, &
          AtomType_Info(i)%basis_Size))
  End Do
  
  Return
End Subroutine

!******************************************************************************
!
!  InitLCAO - Initializes the LCAO routines for use
!
!******************************************************************************

Subroutine InitLCAO

  Return
End Subroutine

End Module

