!******************************************************************************
!
! File : ldsupercell.f90
!   by : Alan Tackett
!   on : 7/27/98
!  for : PAW Project
!
! LdSuperCell - Loads the super cell information.
!
! Definition Format :
!
!   SuperCell
!      SCALE scale_factor     # Optional Scale factor for A,B,C
!      A   (Xa,Ya,Za) UNIT       # Unit cell vectors in xyz coords 
!      B   (Xb,Yb,Zb) UNIT
!      C   (Xc,Yc,Zc) UNIT
!   
!
!      K-POINTS_GRID (nA, nB, nC) # Defines the K-points grid
!                                 # where nA, nB, and nC are the recip lattice
!                                 # subdivisions along each recip vector
!
!      K-POINTS_LIST List_Size      # Or you can specifty the list yourself
!          (Kx1, Ky1, Kz1)  Weight1 # where each line is a kpoint followed
!            . . . . . . . . . .    # by a weight
!      END
!
!      BZ_METHOD method          # BZ integration method. Current options are:
!                                #    LINEAR_TETRA - Linear tetrahedron method
!                                #    QUAD_TETRA   - Quadratic tetrahedron
!                                #    GAUSS        - Gaussian Smearing
!
!      GAUSS_WIDTH  sigma        # Gaussian Smearing Sigma. Only used if 
!                                # BZ_Method is GAUSS. Ignored otherwise
!
!      AUTO_SYMMETRY             # Automatically determine the symmetry
!
!         #  OR
!
!      ROT_SIZE  RotMatSize      # Number of rotation matrices      
!
!      TRANSLATION index (x,y,z) # Rotation Matrix symmetry Translation
!
!      MATRIX index              # Load the 'index'th matrix  
!          (X1i,Y1i, Z1i)
!          (X2i,Y2i, Z2i)
!          (X3i,Y3i, Z3i)
!      End
!      .
!      .
!      .
!   End 
!      
!
!******************************************************************************

Subroutine LDSuperCell(WC)
  Use crystal_data
  Use paw_inout
  Use mathlib
  Use word
  Use units
  Use strings
  Use storedata
  Use ldatom_info

  Implicit NONE!!!!!

  Type (Word_Context), Intent(INOUT) :: WC
  Real, Pointer :: Ku(:,:), Wt(:)  
  Character(132) :: token
  Integer      :: i, aerr, tlen, j
  Logical      :: NewToken, ListLoaded
  Real         :: Scale


  scale = 1
  xtal%CloneCell =  0
  xtal%Basis = -100
  xtal%Rot_Size = -1
  xtal%BZ_Method = -1
  xtal%Kpnts_grid = -1
  xtal%Sigma = 0.1
  xtal%Auto_symmetry = .FALSE.

  Allocate(xtal%Trans(3,48), xtal%RotMatrix(3,3,48))
 
  ListLoaded = .FALSE.

  If (Print_Level <= PRINT_COMMANDS) then
     Write(Log_Unit, *) ">SuperCell"
  End IF
  
  Call GetNextWord(WC, token, Tlen)
  Call UpperCase(Token)
  
  Do While (token /= "END")
    NewToken = .TRUE.

!***write(*,*) 'ldsupercell: Token=',trim(Token)

    if (Trim(token) == "A") then
       NewToken= .NOT. GetRealsWithUnit(WC,xtal%Basis(:,1),Length_Unit,token,3)
       if (Print_Level == PRINT_VERBOSE) then
          write(LOG_Unit, *) "ldSuperCell:    A = ", xtal%Basis(:,1)
       End IF 
    else if (Trim(token) == "B") then
       NewToken =.NOT. GetRealsWithUnit(WC,xtal%Basis(:,2),Length_Unit,token,3)
       if (Print_Level == PRINT_VERBOSE) then
          write(LOG_Unit, *) "ldSuperCell:    B = ", xtal%Basis(:,2)
       End IF 
    else if (Trim(token) == "C") then
       NewToken= .NOT. GetRealsWithUnit(WC,xtal%Basis(:,3),Length_Unit,token,3)
       if (Print_Level == PRINT_VERBOSE) then
          write(LOG_Unit, *) "ldSuperCell:    C = ", xtal%Basis(:,3)
       End IF 
    else if (Trim(token) == "SCALE") then !** Lattice scale factor
       Call GetNumber(WC, Scale)
       if (Print_Level == PRINT_VERBOSE) then
          write(LOG_Unit, *) "ldSuperCell:    Scale = ", Scale
       End IF        
    else if (Trim(token) == "AUTO_SYMMETRY") then
       xtal%Auto_Symmetry = .TRUE.
    else if (Trim(token) == "SYM_OUTFILE") then
       Call GetNextWord(WC, token, Tlen)
       If (Tlen > 0) Sym_Name=Trim(token)
    else if (Trim(token) == "ROT_SIZE") then
       Call GetNumber(WC, xtal%Rot_Size)
       if (Print_Level == PRINT_VERBOSE) then
          write(LOG_Unit, *) "ldSuperCell:    Rot_Size = ", xtal%Rot_Size
       End IF 
    else if (Trim(token) == "MATRIX") then
        Call GetNumber(WC, j)
        Call GetNumbers(WC, xtal%RotMatrix(1,:, j), 3)
        Call GetNumbers(WC, xtal%RotMatrix(2,:, j), 3)
        Call GetNumbers(WC, xtal%RotMatrix(3,:, j), 3)
        Call GetNextWord(WC, Token, tlen)  !** Get the end statement
       if (Print_Level == PRINT_VERBOSE) then
          write(LOG_Unit, *) "ldSuperCell:    Ld RotMatrix = ", j
       End IF 
    else if (Trim(token) == "TRANSLATION") then
        Call GetNumber(WC, j)
        Call GetNumbers(WC, xtal%Trans(:,j), 3)
       if (Print_Level == PRINT_VERBOSE) then
          write(LOG_Unit, *) "ldSuperCell:    Ld Translation = ", j
       End IF 
    else if (Trim(token) == "K-POINTS_GRID") then
        Call GetNumbers(WC, Xtal%Kpnts_Grid)
        if (Print_Level == PRINT_VERBOSE) then
           write(LOG_UNIT, *) "ldSuperCell:   K-Points Grid = ",Xtal%Kpnts_grid
        End If
    else if (Trim(token) == "GAUSS_WIDTH") then
        Call GetNumber(WC, xtal%Sigma)
       if (Print_Level == PRINT_VERBOSE) then
          write(LOG_Unit, *) "ldSuperCell:    GAUSS_Width = ", xtal%Sigma
       End IF 
    else if (Trim(token) == "BZ_METHOD") then
        Call GetNextWord(WC, token, tlen)
        if (W_ERROR /= W_OK) Call EOF_Error(WC, "ldsupercell(bz_method):")

        Call UpperCase(token)
        if (trim(token) == "LINEAR_TETRA") then
           Xtal%BZ_Method = BZ_LINEAR_TETRA
           if (Print_Level == PRINT_VERBOSE) then
              Write(Log_Unit, *) "ldSuperCell:   BZ_Method = BZ_LINEAR_TETRA"
           End If
        else if (Trim(token) == "QUAD_TETRA") then
           Xtal%BZ_Method = BZ_QUAD_TETRA
           if (Print_Level == PRINT_VERBOSE) then
              Write(Log_Unit, *) "ldSuperCell:   BZ_Method = BZ_QUAD_TETRA"
           End If
        else if (Trim(token) == "GAUSS") then
           Xtal%BZ_Method = BZ_GAUSS
           if (Print_Level == PRINT_VERBOSE) then
              Write(Log_Unit, *) "ldSuperCell:   BZ_Method = BZ_GAUSS"
           End If
        else 
           Write(Error_Unit, *) "ldSuperCell: Unknown BZ_Method: ",Trim(token)
        End If
    else if (Trim(token) == "K-POINTS_LIST") then
        Call GetNumber(WC,BZ%TotalUniq)
        if (W_ERROR /= W_OK) Call EOF_Error(WC, "ldatom(K-Points_List):")
        BZ%TotalKpnts = BZ%TotalUniq

        Allocate(BZ%Ku(3,BZ%TotalUniq), xtal%Wt(BZ%TotalUniq), &
             User_Kpoints(3,BZ%TotalUniq), STAT=aerr)
        if (aerr /= 0) then
           Write(Error_Unit, *) "ldSuperCell: Error Allocating K-Points List! Error = ", aerr
           Call Word_GetAndPrint(WC, Error_Unit, 'ldSuperCell: ')
           STOP
        End If

        If (Print_LEvel == PRINT_VERBOSE) then
           Write(Log_Unit, *) 'LdSuperCell: K-Points_List Size = ',BZ%TotalUniq
        End If

        ListLoaded = .TRUE.

        BZ%Kp => BZ%Ku;  
        i = 1
        Do While ((i<=BZ%TotalUniq) .AND. (W_Error == W_OK))
           Call GetNumbers(WC, BZ%Ku(:,i))
           Call GetNumber(WC,xtal%Wt(i))
           i = i + 1
        End Do

        If (i-1 /= BZ%TotalUniq) then
          write(Error_unit,*) 'Error in K-point list',i-1,BZ%TotalUniq
          stop
        Endif

        User_Kpoints = BZ%Ku

        Call GetNExtWord(WC,Token,TLen)
        IF (W_Error/=W_Ok) Call EOF_Error(WC, "ldatom(K-Points_List-LOADING):")
    else if (Trim(token) == "CLONE_CELL") then
       Call GetNumbers(WC, xtal%CloneCell)
       Write(Log_Unit,*) 'LdSuperCell: Cloning Cell:',xtal%CloneCell
    else
       Write(Error_Unit, *) "ldSuperCell: Unknown Option: ", Trim(Token)
    End If

    Xtal%Kpnts_Grid = 0   !** Set to 0

    If (NewToken) then
       Call GetNextWord(WC, token, tlen)
       if (W_ERROR /= W_OK) Call EOF_Error(WC, "ldatom(next word): ")
    End If

    Call Uppercase(token)
  End Do

  
  !
  ! Calculate SuperCell Volume and Reciprocal lattice vectors
  !

  xtal%Basis = Scale * xtal%Basis

  xtal%Volume = ABS(DOT_PRODUCT(xtal%Basis(:,1), &
                       CrossProduct(xtal%Basis(:,2), xtal%Basis(:,3))))
  xtal%Recip = RecipBasis(xtal%Basis)
  xtal%InverseBasis = TRANSPOSE(Xtal%Recip) / (2*Pi)
  xtal%RecipVol = ABS(DOT_PRODUCT(xtal%Recip(:,1), &
                       CrossProduct(xtal%Recip(:,2), xtal%Recip(:,3))))

  If (Print_Level <= PRINT_NORMAL) then
     Write(Log_Unit, *) 'ldSuperCell: Xtal Volume : ', Xtal%Volume, &
          ' * Recip Volume :',Xtal%RecipVol
     Write(Log_Unit, *) 'ldSuperCell: Scaled Lattice Vectors:' 
     Write(Log_Unit, *) 'ldSuperCell: A :',Xtal%Basis(:,1)
     Write(Log_Unit, *) 'ldSuperCell: B :',Xtal%Basis(:,2)
     Write(Log_Unit, *) 'ldSuperCell: C :',Xtal%basis(:,3)
     Write(Log_Unit,*)  ' '
     Write(Log_Unit, *) 'ldSuperCell: Recip(:,1) :',Xtal%Recip(:,1)
     Write(Log_Unit, *) 'ldSuperCell: Recip(:,2) :',Xtal%Recip(:,2)
     Write(Log_Unit, *) 'ldSuperCell: Recip(:,3) :',Xtal%Recip(:,3)
  end If

!***  If (Auto_Symmetry) Call StoreSymmetry('RUNTIME')

  !
  !  Do a quck check on the BZ info to make sure its valid
  !  
  If (Product(Xtal%Kpnts_Grid) < 0) then
     Write(Error_Unit, *) "ldSuperCell: Invalid K-Points_Grid size: ", &
           Xtal%KPnts_Grid
     Write(Error_Unit, *) "ldSuperCell: Program Halted!"
     Call Word_GetAndPrint(WC, Error_unit, "ldSuperCell:")
     STOP
  End If

  If (xtal%BZ_Method == -1) then
     Write(Error_Unit, *) "ldSuperCell: Invalid BZ_Method! Program Halted!"
     Call Word_GetAndPrint(WC, Error_Unit, "ldSuperCell:")
     STOP
  End If


  bloch_zero = 1  !** Quick Fix should be removed if possible

  if (ListLoaded) then       !** Scale the Kpoints list loaded
     Do i=1, BZ%TotalUniq
        call fixhalfrange(BZ%Ku(1,i))
        call fixhalfrange(BZ%Ku(2,i))
        call fixhalfrange(BZ%Ku(3,i))
        BZ%Ku(:,i) = MATMUL(xtal%Recip, BZ%Ku(:,i))
     End Do
  End If


  xtal_Defined = .TRUE.

  Return
End Subroutine
