!******************************************************************************
!
! File : local_criteria_lib.f90
!   by : Alan Tackett
!   on : 05/04/00
!  for : PWPAW
!
!  Module for defining, loading, and using the local selection criteria 
!  for targeting wave functions in a particular region in space.
!
!******************************************************************************

Module local_criteria_lib

Use atom_data
Use crystal_data
Use paw_inout
Use word
Use gpoints

Implicit NONE!!!!!!!

Integer, PARAMETER :: CRIT_ATOM = 1   !** Atom-centered selection

!** Data structure containing the spatial or localized selection criteria
Type Local_Criteria_Type   
   Integer :: Criteria_Type
   Integer :: index
   Real    :: Width(3)
End Type      

Type (Local_Criteria_Type), Pointer :: Local_Criteria(:)
Integer Local_Criteria_Size
Real, Pointer :: Local_Criteria_Map(:)

!******************************************************************************
Contains
!******************************************************************************

!**********************************************************************
!
! AddLocalCrit_Sphere - Adds a sphere shaped selection to the criteria map
!
!   Map    - Map to add to
!   Center - Sphere center
!   Radius - Sphere radius
!   R_xtal - Different Lattice Offsets for periodic images
!
!**********************************************************************

Subroutine AddLocalCrit_Sphere(Map, Center, Radius, R_xtal)
  Real,    Intent(INOUT) :: Map(:)
  Real,    Intent(IN)  :: Center(:)
  Real,    Intent(IN)  :: Radius  
  Real,    Intent(IN)  :: R_xtal(:,:)

  Integer :: x,y,z,t, PM(3), offset_xtal
  Real    :: R(3), DR(3), mag, Rc2
  Logical :: Ok

  Rc2 = Radius*Radius
  Do z=1, FFT_Grid(3, G_PROJ)
     Do y=1, FFT_Grid(2, G_PROJ)
        Do x=1, FFT_Grid(1, G_PROJ)
           t = t + 1
           
           PM = (/x,y,z/)
           R = MATMUL(xtal%Basis, (PM-1.0)/(FFT_grid(1:3,G_PROJ)))


           offset_xtal = 0
           Ok = .FALSE.
           Do While ((.NOT. Ok) .AND. (offset_xtal<27))
              Offset_xtal = offset_xtal + 1

              DR = R - R_xtal(:,offset_xtal) - Center
              mag = DOT_PRODUCT(DR,DR)
              If (mag <= Rc2) Ok = .TRUE.
           end Do

           If (Ok) then
              Map(t) = 1
           End If
        End Do !*x
     End Do !*y
  End Do !*z

  Return
end subroutine

!******************************************************************************
!
!  CreateLocalCriteriaMap - Creates the local criteria map
!
!******************************************************************************

Subroutine CreateLocalCriteriaMap
  Integer :: i,x,y,z
  Real    :: R(3), R_xtal(3,27), Radius
  Type (Local_Criteria_Type), Pointer :: LC

  i = 0
  Do z=-1, 1
     Do y= -1, 1
        Do x=-1, 1
           i = i + 1
           R_xtal(:,i) = MatMul(xtal%Basis, (/x,y,z/))
        End Do
     End Do
  End Do
  R_xtal(:,14) = R_xtal(:,1);  R_xtal(:,1) = 0  !** Move R=0 to front for speed


  Local_Criteria_Map = 0
  Do i=1, Local_Criteria_Size
     LC => Local_Criteria(i)
     R = Atom_List(LC%index)%Pos
     Call AddLocalCrit_Sphere(Local_Criteria_Map, R, LC%Width(1), R_xtal)
  End Do

  Return
end subroutine

!******************************************************************************
!
!  ldLocalCriteria - Loads the local criteria
!
!******************************************************************************

Subroutine LdLocalCriteria(WC)
  Type (Word_Context), Intent(INOUT) :: WC

  Integer :: i, index, tlen
  Character*100 :: token, token2
  Type (Local_Criteria_Type), Pointer :: LC

  Call GetNumber(WC, i)
  Write(Log_Unit,*) 'LdLocalCriteria: Size=',i

  Local_Criteria_Size = i
  Allocate(Local_Criteria(i))

  Call GetNextWord(WC, Token, tlen)
!Write(*,*) 'Read_Input : Token: !',trim(token),'! * Error=',error
  Call UpperCase(token)

  index = 0
  Do While ((W_Error /= W_EOF) .AND. (TRIM(token) /= 'END'))
     index = index + 1
     LC => Local_Criteria(index)

     If (TRIM(token) == 'ATOM') then  !** Atom-centered sphere
        Call GetNextWord(WC, token, tlen)       
        Call GetNumber(WC, LC%Width(1))
               
        i = SearchList(Token, Atom_LUT)
        LC%Index = i
        if (i<=0) then     !** Name not used
           Write(Error_Unit,*) "LdLocalCriteria: Can't find atom! Name =",&
                Trim(Token)
           Call Word_GetAndPrint(WC, Error_Unit, 'LdLocalCriteria:')
           STOP
        End If        
     else
        index = index - 1
        Write(Error_Unit, *) "LdLocalCriteria: Unknown Command : ", Trim(Token)
     End If

    
     Call GetNextWord(WC, Token, tlen)
     Call UpperCase(token)
  End Do

  Local_Criteria_Size = index
  
  Return
End Subroutine

!******************************************************************************
!
!  InitLocalCriteria - Initializes the local selection criteria
!
!******************************************************************************

Subroutine InitLocalCriteria

  Allocate(Local_Criteria_Map(FFT_Grid(4,G_PROJ)))
  
  Local_Criteria_Map = 1  !** Defaults to all space **

  Call CreateLocalCriteriaMap

  Return
End Subroutine

End Module
