!*****************************************************************************
!
! File : ltbzi.f90
!   on : 5/8/95
!   by : Alan Tackett
!  for : DOS calculation using tetrahedron method
!
!*****************************************************************************

Module ltbzi

Implicit NONE!!!!
             !********* Global Constants ************
Integer, PARAMETER :: TETRA_BASE_ONLY = 0    
Integer, PARAMETER :: TETRA_CORR_ONLY = 1    
Integer, PARAMETER :: TETRA_BOTH      = 2    


Type KUsageInfo
   Integer :: NumCubes           ! Number of Cubes containing this Kpnt
   Integer :: CubeList(2,8)      ! List of Cube/Corner pairs
End Type

Type TetraVertices
   Integer         :: Corners(4)       !** Tetrahedron Corners
End Type

Type Corner2TetraRec                   !** Corner to Tetrahedron Record Type
   Integer         :: NumTetra             ! Number of Tetra point is in
   Integer         :: Tetra(6)             ! List of Tetra point is in
End Type

Type (TetraVertices), Private, Target  :: TetraCorners(6)
Type (Corner2TetraRec), Private,Target :: Corner2Tetra(8)
Integer,              Private          :: CornerMap(256)
Integer,              Private          :: NextMap(256)
Integer,              Private          :: Power2(8)

Integer,  Private, Allocatable, TARGET :: CornerOrder(:,:,:) !**Corner order
Integer,  Private, Allocatable, TARGET :: InvOrder(:,:,:)    !**Inv Crnr order
Integer,  Private, Allocatable, TARGET :: TetraOrder(:,:,:)  !**Tetra order
Real,     Private, Pointer             :: Klist(:,:)         !** K points list
Type (KUsageInfo), Private, Allocatable, Target :: Kusage(:) !** K points list
Real,                 Private, Pointer :: Kenergy(:,:)
Integer,              Private          :: TotalKpnts         !** Num K points
Integer,              Private, Pointer :: Cube(:,:)          !** Cube list
Integer,              Private          :: TotalCubes  !** Num Cubes
Integer,              Private          :: EdgeCubes   !** Num Edge Cubes
Integer,              Private          :: TotalBands  !** Num Energy Bands
Integer,              Private          :: TotalElectrons !** # electrons
Real,                 Private          :: TetraVolume !** Tetrahedron Volume
Real,                 Private          :: FermiEnergy !** Fermi Energy Level
Real,                 Private          :: TetraConst  !** Volume Const
Real,                 Private, Pointer :: Weight(:,:) !** Integration Weights 

Contains

!******************************************************************************
!
! InitDOSTetra - Initializes the Tetrahedron DOS calculation routines.
!                Must be called before ANY other dostetra functions!
!
!  Num_Bands- # of Energy Bands
!  Num_Electrons - # of electrons
!  KSize    - # of K points
!  KPnts    - Pointer to K Points table
!  KE       - Pointer to Energy Table for K points
!  Wt       - Pointer to the Integration Weights Table
!  Cubesize - # of Cubes
!  CubeEdge - # of Edge cubes(1st portion on Cubelist)
!  CubeList - Pointer to Cube info
!  CubeVol  - Volume of a cube in reciprocal space
!  UnitCellVol  - Volume of a Unit Cell
!
!******************************************************************************

Subroutine InitLTBZI(Num_Bands, Num_Electrons, KSize, KPnts, Ke, Wt, &
                        CubeSize, CubeEdge, CubeList, CubeVol, UnitCellVol)
  Integer,                 Intent(IN) :: Num_Bands
  Integer,                 Intent(IN) :: Num_Electrons  
  Integer,                 Intent(IN) :: KSize
  Real,            TARGET, Intent(IN) :: KPnts(:,:)
  Real,            TARGET, Intent(IN) :: Ke(:,:)
  Real,            TARGET, Intent(IN) :: Wt(:,:)
  Integer,                 Intent(IN) :: CubeSize
  Integer,                 Intent(IN) :: CubeEdge
  Integer,         TARGET, Intent(IN) :: CubeList(:, :)
  Real,                    Intent(IN) :: CubeVol
  Real,                    Intent(IN) :: UnitCellVol


  Integer :: i, j, k, n, Sum
  Integer, Pointer :: Corner(:)

                !******** Initialize Tetrahedron Corner LUT ***********
  TetraCorners(1)%Corners = (/ 1, 2, 3, 6 /)
  TetraCorners(2)%Corners = (/ 1, 3, 5, 6 /)
  TetraCorners(3)%Corners = (/ 2, 3, 4, 6 /)
  TetraCorners(4)%Corners = (/ 3, 5, 6, 7 /)
  TetraCorners(5)%Corners = (/ 3, 4, 6, 8 /)
  TetraCorners(6)%Corners = (/ 3, 6, 7, 8 /)

                !******** Initialize Corner to Tetra LUT ************
  Corner2Tetra(1)%NumTetra = 2
  Corner2Tetra(1)%Tetra = (/ 1, 2, 0, 0, 0, 0 /)
  Corner2Tetra(2)%NumTetra = 2
  Corner2Tetra(2)%Tetra = (/ 1, 3, 0, 0, 0, 0 /)
  Corner2Tetra(3)%NumTetra = 6
  Corner2Tetra(3)%Tetra = (/ 1, 2, 3, 4, 5, 6 /)
  Corner2Tetra(4)%NumTetra = 2
  Corner2Tetra(4)%Tetra = (/ 3, 5, 0, 0, 0, 0 /)
  Corner2Tetra(5)%NumTetra = 2
  Corner2Tetra(5)%Tetra = (/ 2, 4, 0, 0, 0, 0 /)
  Corner2Tetra(6)%NumTetra = 6
  Corner2Tetra(6)%Tetra = (/ 1, 2, 3, 4, 5, 6 /)
  Corner2Tetra(7)%NumTetra = 2
  Corner2Tetra(7)%Tetra = (/ 4, 6, 0, 0, 0, 0 /)
  Corner2Tetra(8)%NumTetra = 2
  Corner2Tetra(8)%Tetra = (/ 5, 6, 0, 0, 0, 0 /)

                !*********** Initialize CornerMap LUT **************
  k = 1
  n = 1
  do i=1, 8
    Power2(i) = n

    do j = n, 2*n-1
      CornerMap(j) = i
      NextMap(j) = n
    End Do

    k = n + 1
    n = 2 * n
  End Do

            !*********** Initialize the tables ***********
  TotalKpnts = KSize
  Klist => Kpnts
  KEnergy => Ke
  Weight => Wt

  TotalCubes = CubeSize
  EdgeCubes = CubeEdge
  Cube => CubeList

  TotalBands = Num_Bands
  TetraConst = CubeVol / UnitCellVol / 6.0

  TotalElectrons = Num_Electrons

  Allocate(CornerOrder(8, TotalBands, TotalCubes))
  Allocate(InvOrder(8, TotalBands, TotalCubes))
  Allocate(TetraOrder(6, TotalBands, Totalcubes))
  Allocate(Kusage(TotalKpnts))

  Do i=1, TotalCubes             !** Init Corner and Tetra Orders
    Do j=1, TotalBands
      Do k=1, 8                     !** Init CornerOrder
         CornerOrder(k,j,i) = k
	 InvOrder(k,j,i) = k
      End Do

      Do k=1, 6
	Corner => TetraCorners(k)%Corners
        Sum = 0
	Do n=1, 4
	  Sum = Sum + Power2(CornerOrder(Corner(n), j, i))
	End Do

	TetraOrder(k,j,i) = Sum
      End Do      
    End Do
  End Do         

  Return
End Subroutine


!****************************************************************************
!
! t_DetKUsage - Determines how the K points are used in the various
!               Cubes and Tetrahedra
!
!****************************************************************************

Subroutine t_DetKUsage

  Integer                    :: i, j, k
  Integer, Pointer           :: Corner(:)
  Type (KUsageInfo), Pointer :: KU

  KUsage%NumCubes = 0

  Do i = 1, TotalCubes
    Corner => Cube(i,:)
 
    Do j = 1, 8
      KU => KUsage(Corner(j))
      
      KU%NumCubes = KU%NumCubes + 1
      KU%CubeList(1,KU%NumCubes) = i
      KU%CubeList(2,KU%NumCubes) = j
    End Do
  End Do

  Return
End Subroutine


!****************************************************************************
!
! t_Sort - Sorts the Cube energy levels in INCREASING energy
!
!****************************************************************************

Subroutine t_Sort

  Real, Pointer    :: KE(:)  
  Integer          :: i, j, k, n, a, b, t
  Integer, Pointer :: Corner(:), COrder(:), TOrder(:), TC(:), IOrder(:)
  Logical          :: Changed

  Do i=1, TotalCubes
    Corner => Cube(i,:)

    Do j = 1, TotalBands
      KE => KEnergy(j, :)
      COrder => CornerOrder(:,j, i)
      IOrder => InvOrder(:,j, i)     
      TOrder => TetraOrder(:, j, i)
      Changed = .FALSE.

      Do k=2, 8               !*** Sort Energies ****
	n = k
     
        Do While (n>1) 
	  if (KE(Corner(COrder(n-1)))>KE(Corner(COrder(n)))) then
	     Changed = .TRUE.
             a = COrder(n)
             b = COrder(n-1)
             
             COrder(n) =  b
             COrder(n-1) = a

             IOrder(b) = n
             IOrder(a) = n-1

             n = n - 1
          else
             n = 1
          End If
        End Do
      End Do

      if (Changed) then       !*** If needed Encode new Sort order ****
        Do k=1, 6	
	  TC => TetraCorners(k)%Corners
          t = 0
          Do n = 1, 4
	    t = t + Power2(IOrder(TC(n)))            
          End Do

          TOrder(k) = t
        End Do
      End If
    End Do      
  End Do

  Return
End Subroutine


!****************************************************************************
!
! t_BandNOS - Calculates the Number of states for a given Band
!
!             ****** NOTE : Does NOT init NOS to 0! ******
!
! NumPoints - Number of points to calculate
! Band      - Energy Band
! Energy    - Energy Levels to calculate NOS
! NOS       - Tabulated NOS 
!
!****************************************************************************

Subroutine t_BandNOS(NumPoints, Band, Energy, NOS)
  Integer, Intent(IN)    :: NumPoints
  Integer, Intent(IN)    :: Band
  Real,    Intent(IN)    :: Energy(:)
  Real,    Intent(INOUT) :: NOS(:)

  Integer       :: i, j, k, Map, n
  Real          :: CurrE
  Real          :: e1, e2, e3, e4, tc
  Real          :: t1, t2, t3, t4, denom
  Real, Pointer :: KE(:)
  Integer, Pointer :: CC(:), TOrder(:), COrder(:)

  tc = TetraConst

  Do i=1, TotalCubes
    CC => Cube(i,:)

    TOrder => TetraOrder(:, Band, i)
    COrder => CornerOrder(:, Band, i)
    KE => Kenergy(Band, :)

    Do k = 1, 6
      Map = TOrder(k)
      e4 = KE(CC(COrder(CornerMap(Map))))
      Map = Map - NextMap(Map)
      e3 = KE(CC(COrder(CornerMap(Map))))
      Map = Map - NextMap(Map)
      e2 = KE(CC(COrder(CornerMap(Map))))
      Map = Map - NextMap(Map)
      e1 = KE(CC(COrder(CornerMap(Map))))
    
      Do n=1, NumPoints 
        CurrE = Energy(n)
        if (CurrE>e4) then           !** E>e4
           NOS(n) = NOS(n) + tc
        else if (CurrE>e3) then      !** e4>E>e3
           t1 = e4 - CurrE
           NOS(n) = NOS(n) + tc - tc*t1*t1*t1 / ((e4-e1)*(e4-e2)*(e4-e3))
        else if (CurrE>e2) then      !** e3>E>e2
           t1 = CurrE- e2
           t2 = e2 - e1
           t3 = e3 - e1
           t4 = e4 - e2
           t1 = t2*t2 + t1 * (3*t2 + t1*(3 - t1*(t3+t4)/(t4*(e3-e2))))
           NOS(n) = NOS(n) + tc*t1/(t3*(e4-e1))
        else if (CurrE>e1) then      !** e2>e>e1
           t1 = CurrE - e1
           NOS(n) = NOS(n) + tc*t1*t1*t1/((e2-e1)*(e3-e1)*(e4-e1))
        End If
      End Do
    End Do
  End Do
End Subroutine


!****************************************************************************
!
! t_NOS - Calculates the Number of states
!         Does init NOS to 0!
!          
!
! NumPoints - Number of points to calculate
! Energy    - Energy Levels to calculate NOS
! NOS       - Tabulated NOS 
!
!****************************************************************************

Subroutine t_NOS(NumPoints, Energy, NOS)
  Integer, Intent(IN)    :: NumPoints
  Real,    Intent(IN)    :: Energy(:)
  Real,    Intent(INOUT) :: NOS(:)

  Integer :: i

  NOS = 0

  Do i = 1, TotalBands
    Call t_BandNOS(NumPoints, i, Energy, NOS)
  End Do

End Subroutine  


!****************************************************************************
!
! t_TetraDOS - Calculates the Density of states for a SINGLE TetraHedra
!
! Ef          - Fermi Energy
! e1,e2,e3,e4 - Corner Energies in ascending order
! tc          - Scale Constant
!
! Return Values
!   Returns the DOS for the given Tetrahedra
!
!****************************************************************************

Real Function t_TetraDOS(Ef, e1,e2,e3,e4, tc)
  Real, Intent(IN) :: Ef
  Real, Intent(IN) :: e1,e2,e3,e4
  Real, Intent(IN) :: tc

  Real :: t1,t2,t3,t4
  Real :: DOS

  if ((Ef>=e4) .OR. (Ef<=e1)) then           !** E>e4 or E<e1
     DOS = 0
  else if (Ef>e3) then      !** e4>E>e3
    t1 = e4 - Ef
    DOS = 3*tc*t1*t1 / ((e4-e1)*(e4-e2)*(e4-e3))
  else if (Ef>e2) then      !** e3>E>e2
    t1 = Ef- e2
    t2 = e2 - e1
    t3 = e3 - e1
    t4 = e4 - e2
    t1 = t2 + t1 * (2 - t1*(t3+t4)/(t4*(e3-e2)))
    DOS =  3*tc*t1/(t3*(e4-e1))
  else                      !** e2>e>e1
    t1 = Ef - e1
    DOS = 3*tc*t1*t1/((e2-e1)*(e3-e1)*(e4-e1))
 End If

 t_TetraDOS = DOS

 Return
End Function

!****************************************************************************
!
! t_BandDOS - Calculates the density of states for a single Band
!
!        ***** NOTE : Does NOT init DOS to 0! *****
!
! NumPoints - Number of points to calculate
! Band      - Energy Band
! Energy    - Energy Levels to calculate DOS
! DOS       - Tabulated DOS 
!
!****************************************************************************

Subroutine t_BandDOS(NumPoints, Band, Energy, DOS)
  Integer, Intent(IN)    :: NumPoints
  Integer, Intent(IN)    :: Band
  Real,    Intent(IN)    :: Energy(:)
  Real,    Intent(INOUT) :: DOS(:)

  Integer          :: i, j, k, Map, n
  Real             :: CurrE
  Real             :: e1, e2, e3, e4
  Real, Pointer    :: KE(:)
  Integer, Pointer :: CC(:), TOrder(:), COrder(:)


  Do i=1, TotalCubes
    CC => Cube(i,:)

    TOrder => TetraOrder(:, Band, i)
    COrder => CornerOrder(:, Band, i)
    KE => KEnergy(Band, :)

    Do k = 1, 6
      Map = TOrder(k)
      e4 = KE(CC(COrder(CornerMap(Map))))
      Map = Map - NextMap(Map)
      e3 = KE(CC(COrder(CornerMap(Map))))
      Map = Map - NextMap(Map)
      e2 = KE(CC(COrder(CornerMap(Map))))
      Map = Map - NextMap(Map)
      e1 = KE(CC(COrder(CornerMap(Map))))
    
      Do n=1, NumPoints 
        DOS(n) = DOS(n) + t_TetraDOS(Energy(n), e1,e2,e3,e4, TetraConst)
      End Do
    End Do
  End Do
End Subroutine


!****************************************************************************
!
! t_DOS - Calculates the density of states
!         Does init DOS to 0!
!          
!
! NumPoints - Number of points to calculate
! Energy    - Energy Levels to calculate DOS
! DOS       - Tabulated DOS 
!
!****************************************************************************

Subroutine t_DOS(NumPoints, Energy, DOS)
  Integer, Intent(IN)    :: NumPoints
  Real,    Intent(IN)    :: Energy(:)
  Real,    Intent(INOUT) :: DOS(:)

  Integer :: i

  DOS = 0

  Do i = 1, TotalBands
    Call t_BandDOS(NumPoints, i, Energy, DOS)
  End Do

End Subroutine  


!****************************************************************************
!
! t_CalcBaseWeights - Calculates the BASE Weight Factors. 
!                     These are the UNCORRECTED weights
!
!           NOTE : Does NOT init WEIGHTS to 0!
!
! Ef - Fermi Energy Level
!
!****************************************************************************

Subroutine t_CalcBaseWeights(Ef)
  Real          :: Ef

  Integer       :: i, j, k, Map
  Integer       :: k1, k2, k3, k4
  Real          :: CurrE
  Real          :: e1, e2, e3, e4
  Real          :: t1, t2, t3, t4, t5, c1, c2, c3, c4, tc, denom
  Real          :: a1, a2, a3, a4
  Real, Pointer :: KE(:), Wt(:)
  Integer, Pointer :: CC(:), TOrder(:), COrder(:)

  Tc = TetraConst / 4.0

  Do i=1, TotalCubes
    CC => Cube(i,:)

    Do j = 1, TotalBands
      TOrder => TetraOrder(:, j, i)
      COrder => CornerOrder(:, j, i)
      KE => Kenergy(j, :)
      Wt => Weight(j, :)

      Do k = 1, 6
	Map = TOrder(k)
        k4 = CC(COrder(CornerMap(Map)))
        Map = Map - NextMap(Map)
        k3 = CC(COrder(CornerMap(Map)))
        Map = Map - NextMap(Map)
        k2 = CC(COrder(CornerMap(Map)))
        Map = Map - NextMap(Map)
        k1 = CC(COrder(CornerMap(Map)))

        e4 = KE(k4)
        e3 = KE(k3)
        e2 = KE(k2)
        e1 = KE(k1)

        if (Ef<e1) then             !** Ef<e1
           !** no Contribution **
        else if (Ef<e2) then        !** e1<Ef<e2
           t1 = Ef - e1
           t2 = e2 - e1
           t3 = e3 - e1
           t4 = e4 - e1 
           c1 = tc * (t1**3) / (t2*t3*t4)

           Wt(k1) = Wt(k1) + c1 *(4 - t1 * (1/t2 + 1/t3 + 1/t4))
           Wt(k2) = Wt(k2) + c1 * t1/t2
           Wt(k3) = Wt(k3) + c1 * t1/t3
           Wt(k4) = Wt(k4) + c1 * t1/t4
        else if (Ef<e3) then        !** e2<Ef<e3
           t1 = Ef - e1
           t2 = Ef - e2
           t3 = e3 - Ef
           t4 = e4 - Ef
           a4 = e4 - e1
           a3 = e3 - e1
           a2 = e3 - e2
           a1 = e4 - e2
 
           c1 = tc * t1*t1 / (a4*a3)
           c2 = tc * t1*t2*t3 / (a4*a2*a3)
           c3 = tc * t2*t2*t4 / (a1*a2*a4)

           Wt(k1) = Wt(k1) + c1 + (c1+c2)*t3/a3 + (c1+c2+c3)*t4/a4
           Wt(k2) = Wt(k2) + c1+c2+c3 + (c2+c3)*t3/a2 + c3*t4/a1
           Wt(k3) = Wt(k3) + (c1+c2)*t1/a3 + (c2+c3)*t2/a2
           Wt(k4) = Wt(k4) + (c1+c2+c3)*t1/a4 + c3*t2/a1
        else if (Ef<e4) then        !** e3<Ef<e4
           t1 = e4 - e1
           t2 = e4 - e2
           t3 = e4 - e3
           t4 = e4 - Ef

           c1 = tc * (t4**3)/(t1*t2*t3)

           Wt(k1) = Wt(k1) + tc - c1*t4/t1
           Wt(k2) = Wt(k2) + tc - c1*t4/t2
           Wt(k3) = Wt(k3) + tc - c1*t4/t3
           Wt(k4) = Wt(k4) + tc - c1*(4 - (1/t1 + 1/t2 + 1/t3)*t4)
        else                        !** Ef>e4
           Wt(k1) = Wt(k1) + tc
           Wt(k2) = Wt(k2) + tc
           Wt(k3) = Wt(k3) + tc
           Wt(k4) = Wt(k4) + tc
        end IF
      End Do
    End Do
  End Do
End Subroutine


!****************************************************************************
!
! t_CalcCorrectedWeights - Calculates the CORRECTION to the Weight Factors. 
!                          Does not calculate the Base weight factors!
!
!           NOTE : Does not init Weights to 0!
!
! Ef - Fermi Energy Level
!
!****************************************************************************

Subroutine t_CalcCorrectedWeights(Ef)
  Real, Intent(IN) :: Ef

  Type (KUsageInfo), Pointer :: Kpnt
  Type (Corner2TetraRec), Pointer :: C2T
  Integer                  :: i,j,k, c
  Real                     :: DOS(TotalBands), de
  Real                     :: e1, e2, e3, e4, CurrE
  Real, Pointer            :: KE(:), Wt(:)
  Integer, Pointer         :: CC(:), TOrder(:), COrder(:)
  Integer                  :: k1,k2,k3,k4, t, map

  Do i=1, TotalKpnts
    KPnt => KUsage(i)

    Do j=1, KPnt%NumCubes
      c = KPnt%CubeList(1,j)
      C2T => Corner2Tetra(KPnt%CubeList(2,j))
      CC => Cube(c,:)


      Do k=1, TotalBands
        CurrE = KEnergy(k, i)
        COrder => CornerOrder(:,k,c) 

        Do t=1, C2T%NumTetra 
          Map = TetraOrder(C2T%Tetra(t), k, c)
          e4 = Kenergy(k, CC(COrder(CornerMap(Map))))
          Map = Map - NextMap(Map)
          e3 = Kenergy(k, CC(COrder(CornerMap(Map))))
          Map = Map - NextMap(Map)
          e2 = Kenergy(k, CC(COrder(CornerMap(Map))))
          Map = Map - NextMap(Map)
          e1 = Kenergy(k, CC(COrder(CornerMap(Map))))

          de = e1+e2+e3+e4 - 4*CurrE
          Weight(k,i) = Weight(k,i) +              & 
                        0.025*t_TetraDOS(Ef,e1,e2,e3,e4,TetraConst)* de
        End Do
      End Do
    End Do
  End Do

  Return
End Subroutine


!****************************************************************************
!
! t_CalcWeights - Calculates the the Weight Factors. 
!
! Ef - Fermi Energy Level
! Mode      - Type of NOS to calculate
!              TETRA_BASE_ONLY - Calc using BASE weight factors only
!              TETRA_CORR_ONLY -  "     "   Correction weight factors only
!              TETRA_BOTH      - "      "   both base and correction weights
!
!
!****************************************************************************

Subroutine t_CalcWeights(Ef, Mode)
  Real,    Intent(IN) :: Ef
  Integer, Intent(IN) :: Mode

  Weight = 0

  Select Case (Mode)
    Case (TETRA_BASE_ONLY)
       Call t_CalcBaseWeights(Ef)
    Case (TETRA_CORR_ONLY)
       Call t_CalcCorrectedWeights(Ef)
    Case (TETRA_BOTH)
       Call t_CalcBaseWeights(Ef)
       Call t_CalcCorrectedWeights(Ef)
  End Select
       
  Return
End Subroutine


!****************************************************************************
!
! t_CorrectedNOS - Calculates the number of states with the 
!                  corrected integration Weights
!
!       NOTE : Assumes the Kpnts Energies are SORTED!
!
! NumPoints - Number of points to calculate
! Energy    - Energy Levels to calculate DOS
! DOS       - Tabulated DOS
! Mode      - Type of NOS to calculate
!              TETRA_BASE_ONLY - Calc using BASE weight factors only
!              TETRA_CORR_ONLY -  "     "   Correction weight factors only
!              TETRA_BOTH      - "      "   both base and correction weights
!
!****************************************************************************

Subroutine t_CorrectedNOS(NumPoints, Energy, NOS, Mode)
  Integer,  Intent(IN)    :: NumPoints
  Real,     Intent(IN)    :: Energy(:)
  Real,     Intent(INOUT) :: NOS(:)
  Integer,  Intent(IN)    :: Mode


  Integer :: i
integer ::j
Real :: s

  NOS = 0
  Weight = 0

  Do i=1, NumPoints
    Call t_CalcWeights(Energy(i), Mode)

    NOS(i) = Sum(Weight)
  End Do

  Return
End Subroutine


!****************************************************************************
!
! t_Integral - Calculates the integral of a function. Assumes that the
!              integration weights have already been calculated.
!
!  Fn - Function evaluated at each k pnt and band, must have same dim as KE
!
!****************************************************************************

Real Function t_Integral(Fn)
  Real, Intent(IN) :: Fn(:,:)

  t_Integral = Sum(Fn(1:TotalBands,1:TotalKPnts)*Weight(1:totalBands,1:totalKpnts))

  Return
end Function

End Module





