!******************************************************************************
!
!  File : mathlib.f90
!    by : Alan Tackett
!    on : 7/27/95
!   for : PAW Method
!
!  Initializes values of global constants such as pi, machine_precision, etc.
!  Contains various math routines for creating a crystal cartesian basis from
!  spherical basis, calculating reciprocal lattice vectors, etc...
!  The following is a list of the routines currently implemented:
!
!  Calc3x3Inverse - Calculates the 3x3 inverse of a 3x3 symmetric matrix
!  CrossProduct - Calculates the Cross Product of 2 vectors      T
!  Outer_Product - Calculates the outer product of 2 vectors-> AB
!  XtalBasis    - Converts the Length and Angle information into a XYZ basis
!  hkl2xyz      - Converts xtal coordinates (h,k,l) to cart. coords (x,y,z)
!  hkl2xyz      - Converts cart. coords (x,y,z) to xtal coordinates (h,k,l)
!  RecipBasis   - Calculates the Reciprocal Basis from the xtal basis
!  NearestIntWithFactors - Calculates the nearest integer that has a
!                 given set of factors.
!  IsInteger    - Determines if a number is an integer within a given epsilon
!  Factorial    - Calculates n!
!  DoulbeFac    - Calculates n!!
!  Permutations - Calculates N!/(N-k)!
!  GaussLeg     - Calculates the Gauss-Legendre weight factors and points
!  IntSimpsons  - Integrates a 1-D function using Simpsons Rule(Real/Complex)
!  IsOdd        - Detemines if an integer is odd
!  IsEven       - Detemines if an integer is even
!  itotheL      - returns i**L
!  SphereBessel - returns spherical Bessel function
!  DerivSphereBessel - Calculates the Derivative of Spherical Bessel function
!  RadialFourier - calculates the radial integral for Fourier transform      
!  AssosP       - returns associated Legendre function
!  Der_Theta_P  - returns theta derivative of AssosP
!  Der_Phi_P    - returns phi derivative of spherical harmonic function
!  Lininterp    - performs linear interpolation of a numeric function
!  Extrapto0    - performs a 2-point extrapolation to zero
!  EvalPoly     - Evaluates a polynomial
!  Checkbounds - Checks to make sure the xmin and xmax values bound a root
!  brak_grow - Expands an inital interval until a root is bracketed or 
!  brak_shrink - Subdivides an inital interval and searches the new list 
!  rt_Brent - Finds the root of a function using Brent's method given
!  rt_Newton - Newton-Raphson with Bisection.  Returns the fn root .
!  xyztosphericalpolar -- converts from cartesian to spherical polar coords.
!  nderiv      - evaluates numerical derivative of a function on a regular grid
!
!******************************************************************************

Module mathlib


Implicit NONE!!!


!Real, Parameter :: PI = 3.1415926535897932D0 
!Real, Parameter :: R2D = 180.0/PI          !Conversion for radians to degrees
Real :: Pi, Two_Pi, Four_Pi
Real :: R2D
Real :: machine_precision , machine_zero , machine_infinity

!  Data structure needed for logarithmic grid where r(i)=f(u(i)) and
!    u is uniformly tabulated  (for example f(i)=r0*exp(h*(i-1))-1))
  Type GridInfo
     Logical :: islog
     Real, pointer :: drdu(:)   ! needed only if islog==.true.
  End type

Real*8, private :: minlog,maxlog,minexp,maxexp
Real*8, private :: minlogarg,maxlogarg,minexparg,maxexparg

Interface IntSimpson
  Module Procedure Complex_Simpson
  Module Procedure Real_Simpson
End Interface

Interface nderiv
  Module Procedure Complex_nderiv
  Module Procedure Real_nderiv
End Interface

!******************************************************************************
Contains
!******************************************************************************

!*****************************************************************************
!
! Subroutine initconstants -- calculates Pi and related constants
!*****************************************************************************
  subroutine initconstants
     implicit none
      Integer :: i
      Real    :: tmp , a1,a2,a3

      pi=acos(-1.d0)
      two_pi=2*Pi
      four_pi=4*Pi
      r2d=pi/180

   ! Calculate machine accuracy
   machine_precision = 0
   a1 = 4.0/3.0
   do while (machine_precision == 0.0)
      a2 = a1 - 1.0
      a3 = a2 + a2 + a2
      machine_precision = ABS(a3 - 1.0)
   enddo

      machine_zero= machine_precision**4
      machine_infinity = 1.d0/machine_zero

   minlogarg=machine_precision; minlog=log(minlogarg)
   maxlogarg=1.d0/machine_precision; maxlog=log(maxlogarg)
   minexparg=log(machine_precision);  minexp=0.d0
   maxexparg=-log(machine_precision);  maxexp=exp(maxexparg)

 return
 end subroutine initconstants

!****************************************************
 Function ddlog(arg)
   Real*8 :: arg,ddlog

   If (arg>maxlogarg) then
      ddlog=maxlog
   Else if (arg<minlogarg) then
      ddlog=minlog
   Else
      ddlog=log(arg)
   EndIf

 Return
 End Function
 
!****************************************************
 Function ddexp(arg)
   Real*8 :: arg,ddexp

   If (arg>maxexparg) then
      ddexp=maxexp
   Else if (arg<minexparg) then
      ddexp=minexp
   Else
      ddexp=exp(arg)
   EndIf

 Return
 End Function
 
!******************************************************************************
!
! File : complex_simpson.f90
!   by : Alan Tackett
!   on : 04/21/99
!  for : Mathlib
!
!  Generic template for performing 1-D Simpson's rule.  This code is based on 
!  that given to me by N.A.W. Holzwarth.  The number of points does not
!  have to be even.
!
!  Additional parameter needed for loggrid 
!******************************************************************************


!******************************************************************************
!
!  COMPLEX_simpson  - Returns integral over tabulated function f, with 
!        n tabulated integrands at interval h.
!
!     N  - Number of points(Doesn't have to be even)
!     H  - Spacing between points
!     fn - Tabulated function values
!
!******************************************************************************

COMPLEX FUNCTION COMPLEX_Simpson(n, h, fn, Grid)
  Integer,  Intent(IN) :: n
  Real,     Intent(IN)  ::  h
  Complex , Intent(IN) :: fn(n)
  Type(GridInfo), INTENT(IN) :: Grid
 
  Real :: fac
  Integer :: i, j
  Real, pointer :: d(:)
  
  COMPLEX_Simpson = 0.d0
  
  j = 2*((n-1)/2) + 1
  If (Grid%islog) then
     d =>Grid%drdu
     If  (n == 2) THEN      !** use trapazoidal rule
        COMPLEX_simpson = h * (fn(1)*d(1)+fn(2)*d(2)) / 2
     else            !** Use Simpson's rule ** 
        fac = h / 3
        COMPLEX_Simpson = fn(1)*d(1) + 4*fn(2)*d(2) + fn(3)*d(3)
        
        IF (j >= 5) THEN
           DO i = 5, j, 2
              COMPLEX_Simpson = COMPLEX_Simpson + &
                fn(i-2)*d(i-2) + 4*fn(i-1)*d(i-1) + fn(i)*d(i)
           END DO
        END IF
        
        COMPLEX_Simpson = fac * COMPLEX_Simpson
        
        !** trapezoidal rule for end if N is odd **
        IF (n > j) COMPLEX_simpson = COMPLEX_simpson + &
                  h * (fn(n-1)*d(n-1)+fn(n)*d(n)) / 2 
     End If

  Else
     If  (n == 2) THEN      !** use trapazoidal rule
        COMPLEX_simpson = h * (fn(1)+fn(2)) / 2
     else            !** Use Simpson's rule ** 
        fac = h / 3
        COMPLEX_Simpson = fn(1) + 4*fn(2) + fn(3)
        
        IF (j >= 5) THEN
           DO i = 5, j, 2
              COMPLEX_Simpson = COMPLEX_Simpson + fn(i-2) + 4*fn(i-1) + fn(i)
           END DO
        END IF
        
        COMPLEX_Simpson = fac * COMPLEX_Simpson
        
        !** trapezoidal rule for end if N is odd **
        IF (n > j) COMPLEX_simpson = COMPLEX_simpson + h * (fn(n-1)+fn(n)) / 2 
     End If

   Endif
  RETURN
END FUNCTION

!******************************************************************************
!
!  REAL_simpson  - Returns integral over tabulated function f, with 
!        n tabulated integrands at interval h.
!
!     N  - Number of points(Doesn't have to be even)
!     H  - Spacing between points
!     fn - Tabulated function values
!
!******************************************************************************

REAL FUNCTION REAL_Simpson(n, h, fn, Grid)
  Integer,  Intent(IN) :: n
  Real,     Intent(IN)  ::  h
  Real    , Intent(IN) :: fn(n)
  Type(GridInfo), INTENT(IN) :: Grid
 
  Real :: fac
  Integer :: i, j
  Real   , pointer :: d(:)
  
  REAL_Simpson = 0.d0
  
     j = 2*((n-1)/2) + 1
  If (Grid%islog) then
     d =>Grid%drdu
     If  (n == 2) THEN      !** use trapazoidal rule
        REAL_simpson = h * (fn(1)*d(1)+fn(2)*d(2)) / 2
     else            !** Use Simpson's rule ** 
        fac = h / 3
        REAL_Simpson = fn(1)*d(1) + 4*fn(2)*d(2) + fn(3)*d(3)
        
        IF (j >= 5) THEN
           DO i = 5, j, 2
              REAL_Simpson = REAL_Simpson + &
                fn(i-2)*d(i-2) + 4*fn(i-1)*d(i-1) + fn(i)*d(i)
           END DO
        END IF
        
        REAL_Simpson = fac * REAL_Simpson
        
        !** trapezoidal rule for end if N is odd **
        IF (n > j) REAL_simpson = REAL_simpson + &
                  h * (fn(n-1)*d(n-1)+fn(n)*d(n)) / 2 
     End If

  Else
     If  (n == 2) THEN      !** use trapazoidal rule
        REAL_simpson = h * (fn(1)+fn(2)) / 2
     else            !** Use Simpson's rule ** 
        fac = h / 3
        REAL_Simpson = fn(1) + 4*fn(2) + fn(3)
        
        IF (j >= 5) THEN
           DO i = 5, j, 2
              REAL_Simpson = REAL_Simpson + fn(i-2) + 4*fn(i-1) + fn(i)
           END DO
        END IF
        
        REAL_Simpson = fac * REAL_Simpson
        
        !** trapezoidal rule for end if N is odd **
        IF (n > j) REAL_simpson = REAL_simpson + h * (fn(n-1)+fn(n)) / 2 
     End If

   Endif
  RETURN
END FUNCTION

!******************************************************************************
!
! Calc3x3Inverse -- Calculates the 3x3 inverse of a 3x3 symmetric matrix
!
!   Input matrix M
!   Returns matrix IM , ierr=0 if successful inverse  
!                       ierr=1 if M is not symmetric
!                       ierr=-1 if M has zero determinant
!******************************************************************************
Subroutine Calc3x3Inverse(M,IM,ierr)

 Real, Intent(IN) :: M(3,3)
 Real, Intent(OUT) :: IM(3,3)
 Integer, Intent(OUT) :: ierr
 Real  :: det

 ierr=0
 IM = 0

 if (ABS(M(1,2)-M(2,1))+ABS(M(2,3)-M(3,2))+ABS(M(1,3)-M(3,1)).gt.1.d-10) then
    ierr=1 
    return
 endif

 det= M(1,1)*M(2,2)*M(3,3)-M(1,1)*(M(2,3)**2) &
                          -M(2,2)*(M(1,3)**2) &
                          -M(3,3)*(M(1,2)**2) &
                          +2*M(1,2)*M(1,3)*M(2,3)

 if (ABS(det).lt.1.d-10) then
    ierr=-1
    return
 endif

 IM(1,1)=(M(2,2)*M(3,3)-M(2,3)**2)/det
 IM(1,2)=(M(1,3)*M(2,3)-M(1,2)*M(3,3))/det
 IM(2,1)=IM(1,2)
 IM(1,3)=(M(1,2)*M(2,3)-M(1,3)*M(2,2))/det
 IM(3,1)=IM(1,3)
 IM(2,2)=(M(1,1)*M(3,3)-M(1,3)**2)/det
 IM(2,3)=(M(1,2)*M(1,3)-M(1,1)*M(2,3))/det
 IM(3,2)=IM(2,3)
 IM(3,3)=(M(1,1)*M(2,2)-M(1,2)**2)/det

 Return
 End Subroutine


!******************************************************************************
!
! Calc3x3InverseN -- Calculates the 3x3 inverse of a 3x3 general matrix
!
!   Input matrix M
!   Returns matrix IM , ierr=0 if successful inverse  
!                       ierr=-1 if M has zero determinant
!******************************************************************************
Subroutine Calc3x3InverseN(M,IM,ierr)

 Real, Intent(IN) :: M(3,3)
 Real, Intent(OUT) :: IM(3,3)
 Integer, Intent(OUT) :: ierr
 Real  :: det

 ierr=0
 IM = 0


 det= M(1,1)*M(2,2)*M(3,3)-M(1,1)*M(2,3)*M(3,2) &
                          -M(2,2)*M(1,3)*M(3,1) &
                          -M(3,3)*M(1,2)*M(2,1) &
                          +M(1,2)*M(3,1)*M(2,3) &
                          +M(2,1)*M(1,3)*M(3,2) 

 if (ABS(det).lt.1.d-10) then
    ierr=-1
    return
 endif

 IM(1,1)=(M(2,2)*M(3,3)-M(2,3)*M(3,2))/det
 IM(1,2)=(M(1,3)*M(3,2)-M(1,2)*M(3,3))/det
 IM(1,3)=(M(1,2)*M(2,3)-M(1,3)*M(2,2))/det
 IM(2,1)=(M(3,1)*M(2,3)-M(2,1)*M(3,3))/det
 IM(2,2)=(M(1,1)*M(3,3)-M(1,3)*M(3,1))/det
 IM(2,3)=(M(2,1)*M(1,3)-M(1,1)*M(2,3))/det
 IM(3,1)=(M(2,1)*M(3,2)-M(3,1)*M(2,2))/det
 IM(3,2)=(M(3,1)*M(1,2)-M(1,1)*M(3,2))/det
 IM(3,3)=(M(1,1)*M(2,2)-M(1,2)*M(2,1))/det

 Return
 End Subroutine


!******************************************************************************
!
! CrossProduct - Calculates the Cross Product of 2 vectors
!
!   A,B - Vectors
!
! Return Values
!    The cross product of AxB is returned
!
!******************************************************************************

Function CrossProduct(A,B)
   Real             :: CrossProduct(3)
   Real, Intent(IN) :: A(3)
   Real, Intent(IN) :: B(3)
   
   Real :: C(3)

   C(1) = A(2)*B(3) - A(3)*B(2)
   C(2) = A(3)*B(1) - A(1)*B(3)
   C(3) = A(1)*B(2) - A(2)*B(1)

   CrossProduct = C
   Return
End Function

!******************************************************************************
!                                                              T
! Outer_Product - Calculates the Outer Product of 2 vectors, AB
!
!   A,B - Vectors
!   n   - Length of Vectors
!
! Return Values             T
!    The outer product of AB  is returned
!
!******************************************************************************

Function Outer_Product(A,B, n)
   Integer, Intent(IN) :: n
   Real, Intent(IN)    :: A(n)
   Real, Intent(IN)    :: B(n)
   Real                :: Outer_Product(n,n)

   
   Integer :: i, j

   Do i=1, n
      Outer_Product(:,i) = A*B(i)
   End Do

   Return
End Function

!******************************************************************************
!
!  XtalBasis - Converts the length and Angle information into a cartesian
!              basis.  The A-axis lies along the X-Axis and the B-axis lies
!              in the XY plane.
!
!   basis   - Contains the the resulting 3x3 matrix of the 3 cartesian vectors
!   lengths - Crystal unit cell lengths
!   angles  - Crystal unit cell angles in DEGREES
!
!   Return Values
!      Upon return BASIS contains the xyz basis vectors for the unit cell.
!      Each COLUMN of Basis is a basis vector.
!
!******************************************************************************

Function XtalBasis(basis, lengths, angles)
   Real              :: XtalBasis(3,3)
   Real, Intent(IN)  :: Lengths(3)
   Real, Intent(IN)  :: Angles(3)
   
   Real :: Basis(3,3)
 
        !*** Calculate unit basis vectors ***
   Basis(1,1) = 1 
   Basis(2,1) = 0
   Basis(3,1) = 0

   Basis(1,2) = cos(angles(3)*R2D)
   Basis(2,2) = sin(angles(3)*R2D)
   Basis(3,2) = 0

   Basis(1,3) = cos(angles(2)*R2D)
   Basis(2,3) = (cos(angles(1)*R2D)-cos(angles(2)*R2D)*cos(angles(3)*R2D)) &
                / sin(angles(3)*R2D)
   Basis(3,3) = sqrt(1 - Basis(1,3)*Basis(1,3) - Basis(2,3)*Basis(2,3))

          !*** Scale basis ***
   Basis(1:3,1) = Basis(1:3,1)*lengths(1)
   Basis(1:3,2) = Basis(1:3,2)*lengths(2)
   Basis(1:3,3) = Basis(1:3,3)*lengths(3)

   XtalBasis = Basis
   Return
End Function


!******************************************************************************
!
!  hkl2xyz - Converts crystallographic coordinates(h,k,l) to 
!            cartesian coordinates(x,y,z) based on the given crstal basis.
!
!   Basis - Crystal Basis, 3x3 Matrix
!   hkl   - Crystal Coordinates
!
!  Return Values
!      The Cartesian coordinates are returned in PNT 
!  
!******************************************************************************

Function hkl2xyz(Basis, hkl)
   Real              :: hkl2xyz(3)
   Real, Intent(IN)  :: Basis(3,3)
   Real, Intent(IN)  :: hkl(3)

   hkl2xyz = MATMUL(Basis, hkl)

   Return
End Function

!******************************************************************************
!
!  xyz2hkl - Converts cartesian coordinates(x,y,z) to crystallographic
!            coordinates(h,k,l) to based on the given crstal basis.
!
!   Basis - Crystal Basis, 3x3 Matrix
!   xyz   - Crystal Coordinates
!
!  Return Values
!      The crystallographic coordinates are returned in PNT 
!  
!******************************************************************************

Function xyz2hkl(Basis, xyz)
   Real              :: xyz2hkl(3)
   Real, Intent(IN)  :: Basis(3,3)
   Real, Intent(IN)  :: xyz(3)


   xyz2hkl = MATMUL(TRANSPOSE(RecipBasis(Basis))/(2*PI), xyz)

   Return
End Function

!******************************************************************************
!
! RecipBasis - Calculates the reciprocal lattice vectors given the
!              cartesian basis calculated in XtalBasis 
!
!   Basis - Crystal Basis (3x3 matrix)
!   Recip - Reciprocal Crystal Basis of above (3x3 matrix)
!
! Return Values
!    RECIP contains the reciprocal basis with each COLUMN being a
!    reciprocal lattice vector.
!
!******************************************************************************

Function RecipBasis(Basis)
  Real              :: RecipBasis(3,3)
  Real, Intent(IN)  :: Basis(3,3)

  Real :: Recip(3,3)
  Real :: t, C(3)
  
  
  C = CrossProduct(basis(1:3,2),basis(1:3,3))
  t = DOT_PRODUCT(Basis(1:3,1), C)
  Recip(1:3,1) = 2*PI * C / t

  C = CrossProduct(basis(1:3,3),basis(1:3,1))
  t = DOT_PRODUCT(Basis(1:3,2), C)
  Recip(1:3,2) = 2*PI * C / t

  C = CrossProduct(basis(1:3,1),basis(1:3,2))
  t = DOT_PRODUCT(Basis(1:3,3), C)
  Recip(1:3,3) = 2*PI * C / t
  
!Do i=1,3
!   Write(*,*) i,'REC=',Recip(:,i)
!End Do

  RecipBasis = Recip
  Return
End Function


!******************************************************************************
!
!  NearestIntWithFactors - Calculates the nearest integer from the specified N
!            in either direction that is composed of the specified Factors.
!
!  N         - Starting Number
!  Factor    - List of Factors
!  NumFact   - Number of factors
!  Direction - Direction to search: +1=up, -1=down
!              NOTE: This is actually the INCREMENT between successive guesses 
!
!  Return Values
!     Returns the number or 0 if Direction=0.    
!
!  Examples
!     N = NearestIntWithFactors(79, (/2,3,5/), 3, +1)
!                              4
!     Returns N=80 since 80 = 2 * 5
!
!******************************************************************************

Integer Function NearestIntWithFactors(N, Factor, NumFact, Direction)
  Integer, Intent(IN) :: N
  Integer, Intent(IN) :: Factor(:)
  Integer, Intent(IN) :: NumFact
  Integer, Intent(IN) :: Direction

  Integer :: M, Num, i

  If (Direction == 0) then
    Num = 1
    M = 0
  else
    Num =  0
    M = N - Direction
  End If

  Do While (Num /= 1)
    M = M + Direction
    
    i = 1
    Num = M
    Do While ((Num > 1) .AND. (i<= NumFact))
      Do While (MOD(Num, Factor(i)) == 0)
        Num = Num / Factor(i)
      End Do
      
      i = i + 1
    End Do 
  End Do

  NearestIntWithFactors = M

  Return
End Function

!******************************************************************************
!
!  IsInteger - Determines if a number is an integer within a given epsilon
!
!  Num - Number to check
!  epsilon - Precision
!
!  Return Values
!     Returns .TRUE. if INT(num)-epsilon < num < INT(num)+epsilon
!     and .FALSE. otherwise.
!
!******************************************************************************

Logical Function IsInteger(Num, epsilon)
  Real, Intent(IN) :: Num
  Real, Intent(IN) :: epsilon

  Logical :: Answer

  If (((Int(num)-epsilon) < num) .AND. (num < (Int(num)+epsilon))) then
    Answer = .TRUE.
  else
    Answer = .FALSE.
  End If

  IsInteger = Answer
  Return
End Function

!******************************************************************************
!
! Factorial - Calculates N!
!
! N  - Number to use
!
! Return Value
!   Returns N!
!
!******************************************************************************

Real Function Factorial(N)
  Integer, Intent(IN) :: N

  Integer :: i
  
  Factorial = 1
  If (N<2) return
  Do i = 2, N
    Factorial = Factorial * i
  End Do

  Return
End Function

!******************************************************************************
!
! DoubleFac - Calculates N!!
!
! N  - Number to use
!
! Return Value
!   Returns N!!
!
!******************************************************************************

Real Function DoubleFac(N)
  Integer, Intent(IN) :: N

  Integer :: i
  
  DoubleFac = 1
  Do i = N, 1, -2
    DoubleFac = DoubleFac * i
  End Do

  Return
End Function

!******************************************************************************
!
! Permutations - Calculates N!/(N-k)!
!
! N, K  - Numbers to use
!
! Return Value
!   Returns N!/(N-k)! if N>=0 and N-k>0 otherwise 0 is returned
!
!******************************************************************************

Real Function Permutations(N, K)
  Integer, Intent(IN) :: N
  Integer, Intent(IN) :: k

  Integer :: i

  if ((N>=0) .AND. ((N-k)>=0)) then
    Permutations = 1
    Do i = N-K+1, N
       Permutations = Permutations * i
    End Do
  Else
    Permutations = 0
  End If

  Return
End Function

!******************************************************************************
!
! GaussLeg - Calculates the Gauss-Legendre integration weights and points
!            on the interval (-1, 1)
!
! n   - Number of G-L points and weights to calculate
! X   - G-L Points
! W   - G-L weights
!
! NOTE: This routine is based on the routine from Numerical Recipes in Fortran
!
!******************************************************************************

SUBROUTINE GaussLeg(x1,x2,n,x,w)
  Real, Intent(IN) :: x1,x2
  Integer, Intent(IN) :: n
  Real,    Intent(OUT) :: X(:)
  Real,    Intent(OUT) :: W(:)

  Real, PARAMETER :: EPS=3.d-14  !** EPS is the relative precision.

  Real ::  p1,p2,p3,pp,xl,xm,z,z1,pi
  Integer :: i,j,m

  pi=acos(-1.d0)
                     !** High precision is a good idea for this routine.
  m=(n+1)/2          !** The roots are symmetric in the interval, so we
  xl=0.5*(x2-x1)     !** only have to nd half of them. 
  xm=0.5d0*(x2+x1)
  
  do i=1,m           !** Loop over the desired roots.
     z=cos(pi*(i-.25d0)/(n+.5d0))
     !** Starting with the above approximation to the ith root,
     !** we enter the main loop of refinement by Newton's method.
     z1 = z+10*EPS               !** Cheat to enter loop

     Do While (ABS(z-z1) > EPS)
        p1=1.0d0
        p2=0.0d0

        do j=1,n        !** Loop up the recurrence relation to get the Legendre
           p3=p2        !** polynomial evaluated at z. 
           p2=p1
           p1=((2.0*j-1.d0)*z*p2-(j-1.0)*p3)/j
        end do

        !** p1 is now the desired Legendre polynomial. We next compute pp,
        !** its derivative, by a standard relation involving also p2, 
        !** the polynomial of one lower order.
        pp=n*(z*p1-p2)/(z*z-1.d0)
        z1=z
        z=z1-p1/pp  !**Newton's method.
     End Do

     x(i)=xm-xl*z                   !** Scale the root to the desired interval,
     x(n+1-i)=xm+xl*z               !** and put in its symmetric counterpart.
     w(i)=2.0*xl/((1.0-z*z)*pp*pp) !** Compute the weight
     w(n+1-i)=w(i)                  !**and its symmetric counterpart.
  end do
  return
End Subroutine


!******************************************************************************
!
!   RadialFourier - calculates the radial integral necessary to evaluate the 
!             Fourier transform of f(r)/r*Ylm
!
!     L  - L value of Ylm
!     k  - Spherical Bessel argument scale factor(see below)
!     h  - Radial spacing between grid points
!     N  - Number of grid points
!     F  - tabulated Function
!
!   input : l, k, h, n, f(:) , where f(r) is tabulated at n points 
!      in intervals of h;   f(1)==f(r=0)=0
!
!   The radial integral is returned:
!        ftf == int(r*f(r)*j_l(k*r),r=0..(n-1)*h)
!
!  NOTE: This code was taken from N.A.W. Holzwarth 
!
!******************************************************************************

Real Function RadialFourier(l,k,h,n,f,Grid)
  Integer, Intent(IN) :: L
  Real,    Intent(IN) :: k
  Real,    Intent(IN) :: h
  Integer, Intent(IN) :: n
  Real,    Intent(IN) :: f(n)
  Type(GridInfo), Intent(IN) :: Grid

  Integer :: i
  Real, pointer :: x(:),y(:)
 
  Allocate(x(n),y(n))

  If (Grid%islog) then
      Do i=1,n
         x(i) = Grid%drdu(i) - Grid%drdu(1)
      End Do
  Else
      Do i=1,n
         x(i) = (i-1)*h
      End Do
  Endif

  y=k*x

  Call SphereBessel(l,n,y)

  Do i=1,n
     y(i)=y(i)*f(i)*(x(i))
  End Do


  RadialFourier = IntSimpson(n,h,y,Grid)        
  
  DeAllocate(x,y)
  return
End Function


!******************************************************************************
!
!  SphereBessel - Calculates the Sphereical Bessel function of order Lval, 
!        jl(x), at the given x(i) positions
!
!  Ndim - Number of grid points
!  X - On input this contains the positions where to calculate jl(x)
!      On output x contains the tabulated jl
!  
!******************************************************************************

Subroutine SphereBessel(lval,ndim,x)
  Integer, Intent(IN)    :: lval
  Integer, Intent(IN)    :: ndim
  Real,    Intent(INOUT) :: x(ndim)

  Integer :: it,lll,l,n,il
  Real :: z,bs,bs0,bs1,xx,z2,arg,sum
  Real, parameter :: tol=1.e-11

  do it=1,ndim
     z=x(it)
     
     if(z.le.0.5) then ! series expansion for small arguments
        if (z.le.tol) then ! - special treatment if z = 0
           bs=0.d0
           if(lval.eq.0) bs=1.d0

        else     
           lll=2*lval+1
           xx=1.d0
           do l=1,lll,2
              xx=xx*l
           end do

           z2=0.5d0*z*z
           lll=lll+2
           arg=-z2/lll
           sum=1.d0+arg

           do  n=2,500
              lll=lll+2
              arg=-arg*z2/(n*lll)
              sum=sum+arg
              if (abs(arg).le.tol) exit
           end do
  
           bs=(z**lval)*sum/xx
        end if
     else        ! -- trigometric form
        bs=sin(z)/z

        if (lval.gt.0) then    
           bs0=bs
           bs=(bs0-cos(z))/z
           if (lval.gt.1) then
              bs1=bs
              
              do  il=2,lval
                 bs=(2*il-1)*bs1/z-bs0
                 bs0=bs1
                 bs1=bs
              end do
           end if
        end if
     end if       

     if (ABS(bs)<1D-50) bs = 0
     x(it)=bs

  end do 

 Return
End  Subroutine



!******************************************************************************
!
!  DerivSphereBessel - Calculates the Derivative of Spherical Bessel 
!        function of order Lval, d jl(x), at the given x(i) positions
!                                -------
!                                 dx
!
!  Ndim - Number of grid points
!  X - On input this contains the positions where to calculate d jl(x)/dx
!      On output x contains the tabulated d jl/dx
!  
!
!******************************************************************************

Subroutine DerivSphereBessel(lval,ndim,x)
  Integer, Intent(IN)    :: lval
  Integer, Intent(IN)    :: ndim
  Real,    Intent(INOUT) :: x(ndim)

  Integer :: it,lll,l,n,il
  Real :: z,bs,bs0,bs1,xx,z2,arg,sum,db,bsm
  Real, parameter :: tol=1.e-11

  do it=1,ndim
     z=x(it)
     
     if(z.le.0.5) then ! series expansion for small arguments
        if (z.le.tol) then ! - special treatment if z = 0
           db=0.d0
           if(lval.eq.0) db=1.d0/3

        else     
           lll=2*lval+1
           xx=1.d0
           do l=1,lll,2
              xx=xx*l
           end do
           
           If (lval == 0) then
              sum=0
           Else
              sum=lval*(z**(lval-1))/xx
           endIf
           lll=lll+2
           xx=xx*lll
           arg=-(z**(lval+1))/(2*xx)
           sum=sum+(lval+2)*arg

           z2=z*z
           do  n=2,500
              lll=lll+2
              arg=-arg*z2/(2*n*lll)
              sum=sum+(lval+2*n)*arg
              if ((lval+2*n)*abs(arg).le.tol) exit
           end do
  
           db=sum               
        end if
     else        ! -- trigometric form
        bs=sin(z)/z
        bsm=cos(z)/z


        if (lval.gt.0) then    
           bsm=bs
           bs=(bsm-cos(z))/z
           if (lval.gt.1) then
              bs0=bsm
              bsm=bs
              
              do  il=2,lval
                 bs=(2*il-1)*bsm/z-bs0
                 bs0=bsm
                 bsm=bs
              end do
                 bsm=bs0
           end if
        end if

      db=bsm-(lval+1)*bs/z
     end if       

     if (ABS(db)<tol*tol) db = 0
     x(it)=db

  end do 

 Return
End  Subroutine


!******************************************************************************
!
!  IsOdd - Determines if an integer is odd
!
!     n - Integer
!
!******************************************************************************

Logical Function IsOdd(n)
  Integer, Intent(IN) :: n

  If (2*INT(n/2.0) == n) then
     IsOdd = .FALSE.
  else
     IsOdd = .TRUE.
  End If

  Return
End Function

!******************************************************************************
!
!  IsEven - Determines if an integer is even
!
!     n - Integer
!
!******************************************************************************

Logical Function IsEven(n)
  Integer, Intent(IN) :: n

  If (2*INT(n/2.0) == n) then
     IsEven = .TRUE.
  else
     IsEven = .FALSE.
  End If

  Return
End Function

!******************************************************************************
!
!  itotheL - Returns i**L                    
!
!     L - Integer
!
!******************************************************************************
 
Complex Function itotheL(L)
  Integer, Intent(IN) :: L
 
  itotheL=1
  if(L==0) return

  If (2*INT(L/2.0) == L) then
     itotheL=(-1)**(L/2)
  else
     itotheL=CMPLX(0,1)*(-1)**((L-1)/2)
  End If
 
  Return
End Function

!******************************************************************************
! Function AssosP(L,m,x)
! From Numerical Recipes, 2nd. Edition, pg. 247
!
!*****************************************************************************
 Real Function AssosP(L,m,x,Error_Unit)
  Implicit None
 Integer, Intent(IN) :: L,m
 Real, Intent(IN) :: x
 Integer, Optional, Intent(IN) :: Error_Unit

 Real :: fact,pll,pmm,pmmp1,somx2
 Integer :: i, ll

 If (m.lt.0.or.m.gt.L.or.(abs(x)-1.0).gt.1.e-8) then
  If (present(Error_Unit)) then
     write(Error_Unit,*) 'error in AssosP',m,x
  Else
     write(*,*) 'error in AssosP',m,x
  EndIf
  stop
 Endif

 pmm=1
 if (m.gt.0) then
   somx2=sqrt((1-x)*(1+x))
   fact=1
   do i=1,m
      pmm=-pmm*fact*somx2
      fact=fact+2
   enddo
 Endif
 If(L.eq.m) then 
   AssosP=pmm
 Else
   pmmp1=x*(2*m+1)*pmm
   If(L.eq.m+1) then
     AssosP=pmmp1
   else
     do ll=m+2,L 
       pll=(x*(2*ll-1)*pmmp1-(ll+m-1)*pmm)/(ll-m)
       pmm=pmmp1
       pmmp1=pll
     Enddo
   AssosP=pll
   Endif
 Endif
 End function AssosP
 

!******************************************************************************
! Function Der_Theta_P(L,m,x)
! From Numerical Recipes, 2nd. Edition, pg. 247
! Implimented by Yonas Abraham
! This function is modified to return Diff(P^m_l(x),x),theta)
!*****************************************************************************
 Real Function Der_Theta_P(L,m,x,Error_Unit)
  Implicit None
 Integer, Intent(IN) :: L,m
 Real, Intent(IN) :: x
 Integer, Optional, Intent(IN) :: Error_Unit

 Real :: fact,dpll,pll,dpmm,pmm,dpmmp1,pmmp1,somx2
 REAL :: dosomx2
 Integer :: i, ll

 If (m.lt.0.or.m.gt.L.or.abs(x).gt.1.0) then
    If (Present(Error_Unit)) then
       write(Error_Unit,*) 'error in AssosP',m,x
    Else
       write(*,*) 'error in AssosP',m,x
    EndIf
  stop
 Endif

 pmm=1
 dpmm=1
 dosomx2=1
 somx2=sqrt((1-x)*(1+x))
 IF(m==0)THEN
   dpmm=0.
 ELSE IF (m > 0) THEN
   fact=1
   do i=1,m
      pmm=-pmm*fact*somx2
      dpmm=-dpmm*fact
      fact=fact+2
   ENDDO
   IF(m>1)THEN
     DO i=2,m
       dosomx2=dosomx2*somx2
     END DO
   END IF   
   dpmm= dpmm*m*x*dosomx2 
 ENDIF
 IF(l==m) THEN 
   Der_Theta_P=dpmm
 ELSE
   pmmp1=x*(2*m+1)*pmm
   dpmmp1=-(2*m+1)*somx2*pmm+x*(2*m+1)*dpmm
   IF(L==m+1) THEN
     Der_Theta_P=dpmmp1
   ELSE
     do ll=m+2,L
      pll=(x*(2*ll-1)*pmmp1-(ll+m-1)*pmm)/(ll-m) 
      dpll=(-somx2*(2*ll-1)*pmmp1+(x*(2*ll-1)*dpmmp1-(ll+m-1)*dpmm))/(ll-m)
       pmm=pmmp1
       pmmp1=pll
       dpmm=dpmmp1
       dpmmp1=dpll
     ENDDO
   Der_Theta_P=dpll
   ENDIF
 ENDIF
 END function Der_Theta_P

!******************************************************************************
! Function Der_Phi_P(L,m,x)
! From Numerical Recipes, 2nd. Edition, pg. 247
! Implimented by Yonas Abraham
!                                         m
!  This function is modified to return m*P_l(x)/sqrt((1-x^2))
!*****************************************************************************
 REAL Function Der_Phi_P(L,m,x,Error_Unit)
 IMPLICIT None
 INTEGER, Intent(IN) :: L,m
 Real, Intent(IN) :: x
 INTEGER, Optional, Intent(IN) :: Error_Unit

 REAL :: fact,pll,pmm,pmmp1,somx2,dosomx2
 INTEGER :: i, ll

 IF (m.lt.0.or.m.gt.L.or.abs(x).gt.1.0) THEN
     If (Present(Error_Unit)) then
        WRITE(Error_Unit,*) 'error in Der_Phi_P',m,x
     Else
        WRITE(*,*) 'error in Der_Phi_P',m,x
     EndIf
  STOP
 ENDIF

 Der_Phi_P=0
 If (m==0) return

 pmm=1
 dosomx2=1
 IF (m > 0) then
   somx2=sqrt((1-x)*(1+x))
   fact=1
   DO i=1,m
      pmm=-pmm*fact
      fact=fact+2
   ENDDO
   If (m > 1) then
   DO i=2,m
      dosomx2=somx2*dosomx2
   END DO
   EndIf
   pmm=pmm*dosomx2 !due to one more term (-1^M) 
 ENDIF
 IF(L==m) THEN
   Der_Phi_P=pmm*m
 ELSE
   pmmp1=x*(2*m+1)*pmm
   IF(L==m+1) THEN
     Der_Phi_P=pmmp1*m
   ELSE
     DO ll=m+2,L
       pll=(x*(2*ll-1)*pmmp1-(ll+m-1)*pmm)/(ll-m)
       pmm=pmmp1
       pmmp1=pll
     ENDDO
   Der_Phi_P=pll*m
   ENDIF
 ENDIF
 END function Der_Phi_P
 
!******************************************************************************
!
! subroutine Lininterp -- performs linear interpolation of numerical
!   function on uniform grid
!
!  fgrid(i) is tabulated at n points corresponding to argument i*h
!  It is assumed that fgrid(1)==fgrid(0) and fofx is returned as 0 for x>n*h
!
!   Modified for possible case of logarithmic grid
!******************************************************************************
Subroutine Lininterp(n,h,fgrid,x,fofx,Grid)
Integer, intent(IN) :: n
Real, intent(IN)    :: h,fgrid(n),x
Real, intent(OUT)   :: fofx
Type(GridInfo), Intent(IN) :: Grid

Real                :: x1 
Integer             :: i

    if (Grid%islog) then
      x1=LOG(1.d0+x/Grid%drdu(1))/h
    else
      x1=x/h
    endif
    i=x1+1
    if (i.lt.1) then
      fofx=fgrid(1)
     else if (i.gt.n-1) then
      fofx=0.d0
     else
      fofx=(x1+1-i)*fgrid(i+1)+(i-x1)*fgrid(i)
    endif

End Subroutine Lininterp


!******************************************************************************
!  Extrapto0    - performs a 2-point extrapolation to zero
!******************************************************************************

Real Function Extrapto0(f1,f2)
  Real,  Intent(IN) :: f1,f2   !! f(h), f(2h)

  Extrapto0 = 2*f1 - f2

End Function Extrapto0

!******************************************************************************
!  
!  EvalPoly - Evaluates a polynomial at the given position
!
!     N    - Degree of polynomial
!     Poly - List of coefficients.  Poly(1) = constant, Poly(2)=c1*x, etc
!     x    - Position
!
!******************************************************************************

Real Function EvalPoly(n, poly, x)
  Integer, Intent(IN) :: n
  Real,    Intent(IN) :: Poly(:)
  Real,    Intent(IN) :: x
  
  Integer :: i
  Real    :: fn, term

  fn = poly(n+1)
  Do i=n, 1, -1
     fn = fn*x + poly(i)
  End Do

  EvalPoly = Fn

  Return
End Function

!******************************************************************************
!
!  File : roots.f90
!    on : 7/31/95
!    by : Alan Tackett
!   for : Misc. Math routines
!
!
!  This module contains several routines for finding and bracketing a root of
!  of a 1-D function. The two main methods are:
!
!        rt_Brent  - Which doesn't use derivatives and combines 
!                    Secant and Bisection methods so give ~quad convergence
!        rt_newton - Does use derivatives and combines Newton-Raphson and
!                    Bisection. 
!
!  Two routines exist for bracketing a root :
!
!        brak_grow   - Takes an initial interval and expands it until an
!                      interval containing a root is located
!        brak_Shrink - Takes an initial interval and subdivides it into
!                      smaller intervals and checks them for roots.
!
!******************************************************************************

!******************************************************************************
!
! Checkbounds - Checks to make sure the xmin and xmax values bound a root
!
!    FitFunc - External, user-supplied routine that calculates the value
!              of the function and its 1st derivative at a point.  It
!              has the following form 
!                     Double Precision  FitFunc(x, fn, deriv) 
!              where all arguments are double precision. The Return value
!              is ignored.
!    xmin, xmax - Root bounds.  Automatically swaps xmin and xmax if
!              f(xmin)>f(xmax)
!
! Return Values
!    Returns .TRUE. if a root is bounded and .FALSE. otherwise.
!
!******************************************************************************

Logical Function CheckBounds(FitFunc, xmin, xmax)
  Real, External                :: FitFunc
  Real,           Intent(INOUT) :: xmin
  Real,           Intent(INOUT) :: xmax

  Logical :: ok
  Real    :: junk
  Real    :: f1, f2
  Real    :: df1, df2

  junk = FitFunc(xmin, f1, df1)
  junk = FitFunc(xmax, f2, df2)

  if ((f1<0) .AND. (f2>0)) then
    ok = .TRUE.
  else if ((f1>0) .AND. (f2<0)) then
    ok = .TRUE.
    junk = xmin
    xmin = xmax
    xmax = junk
  else 
    ok = .FALSE.
  end if

  CheckBounds = ok
  Return
End Function


!******************************************************************************
!
!  brak_grow - Expands an inital interval until a root is bracketed or 
!              the range becomes unacceptably large.
!
!  Func - Function for finding the root. Has the form
!             Double Precision Func(Double Precision x)
!  x1, x2 - Initial interval, Root Brackets(RETURNED)
!  NumEval - (OPTIONAL) Returns the number of function evaluations
!
!  Return Values
!     brak_grow returns 1 if a root was successfully found and 0 otherwise
!
!
!******************************************************************************

Integer Function Brak_Grow(Func, x1, x2, NumEval)
   Real, External                :: Func
   Real,           Intent(INOUT) :: x1
   Real,           Intent(INOUT) :: x2
   Integer, Optional, Intent(OUT) :: NumEval

   Integer, PARAMETER :: NTRY   = 50
   Real,    PARAMETER :: FACTOR = 1.6

   integer :: j, NE
   Real    :: f1, f2

   f1 = Func(x1)
   f2 = Func(x2)
 
   NE = 2

   j = 1
   Do While ((f1*f2>=0) .AND. (j<NTRY))
      if (ABS(f1) < ABS(f2)) then
         x1 = x1 + FACTOR*(x1-x2)
         f1 = Func(x1)
         Ne = NE + 1
      else 
         x2 = x2 + FACTOR*(x2-x1)
         f2 = Func(x2)
         NE = NE + 1
      End If

      j = j + 1
   End Do

   if (f1*f2<0) then
     j = 1
   else
     j = 0
     write(*,*) 'Brak_Grow: X1=',X1, ' X2=',X2
     write(*,*) 'Brak_Grow: F1=',F1, ' f2=',f2
   End If

   Brak_Grow = j

   if (PRESENT(NumEval)) NumEval = NE

   Return   
End Function


!******************************************************************************
!
!  brak_shrink - Subdivides an inital interval and searches the new list 
!                of intervals for roots.
!
!  Func - Function for finding the root. Has the form
!             Double Precision Func(Double Precision x)
!  x1, x2 - Initial Interval to search
!  n      - Number of subintervals to check
!  xb1    - Array containing the min points of bracket containing roots
!  xb2    - Array containing the max points of bracket containing roots
!  nb     - Number of roots to bracket upon entry and upon return it contains
!           the number of roots bracketed
!  NumEval- (OPTIONAL) Returns the number of function evaluations
!
!******************************************************************************

Subroutine brak_shrink(Func, x1, x2, n, xb1, xb2, nb, NumEval)
  Real, External                :: Func
  Real,           Intent(IN)    :: x1
  Real,           Intent(IN)    :: x2
  Integer,        Intent(IN)    :: n
  Real,           Intent(OUT)   :: xb1(:)
  Real,           Intent(OUT)   :: xb2(:)
  Integer,        Intent(INOUT) :: nb
  Integer, Optional, Intent(OUT):: NumEval

  Integer :: nbb, i, NE
  Real    :: x, fp, fc, dx

  dx = (x2-x1)/n
  x = x1
  fp = Func(x)
  nbb = 0

  NE = 1

  Do i=1, n
    x = x + dx
    fc = Func(x)
    NE = NE + 1
    if ((fc*fp<0) .AND. (nbb<nb)) then
      nbb = nbb + 1
      xb1(nbb) = x - dx
      xb2(nbb) = x
    End If

    fp = fc
  End Do

  nb = nbb

  if (PRESENT(NumEval)) NumEval = NE

  Return
End Subroutine


!******************************************************************************
!
!  rt_Brent - Finds the root of a function using Brent's method given
!             an initially bracketed root. No derivatives are required.
!
!    Func    - Function to find the root. Must have the form
!                 Double Precision Func(Double Precision x)
!    x1,x2   - Bracketed interval
!    tol     - Accuracy of root
!    MaxIter - Max number of iterations to allow
!    NumEval - (OPTIONAL) Returns the number of function evaluations
!
!
!  Return Values
!     Returns the root found
!
!******************************************************************************

Real Function rt_Brent(Func, x1, x2, tol, MAXITER, NumEval)
  Real, External                :: Func
  Real,              Intent(IN) :: x1
  Real,              Intent(IN) :: x2
  Real,              Intent(IN) :: tol
  Integer,           Intent(IN) :: MAXITER
  Integer, Optional, Intent(OUT):: NumEval

  Real :: a, b, c, d, e, min1, min2, eps, rt
  Real :: fa, fb, fc, p, q, r, s, tol1, xm
  Integer :: i, NE
  Logical :: Finished

  eps = Epsilon(a)

  a = x1; b = x2; c = x2;
  fa = Func(a); fb = Func(b)

  NE = 2

  if (fa*fb>0) then
    write(*,*) 'Root Not Bracketed in [',a,b, ']'
    write(*,*) '    with function values [',fa,fb,']'
    
    rt_Brent = x1
    Return
  End If

  fc = fb  
  Finished = .FALSE.
  i = 0
  Do While ((.NOT. Finished) .AND. (i<=MaxIter))
    i = i + 1

    if (fb*fc>0) then                         !*** Rename a,b,c and adjust
      c = a                                   !*** Bounding interval d
      fc = fa
      e = b - a;  d = e;
    End If

    if (ABS(fc) < ABS(fb)) then
       a=b;   b=c;   c=a;
       fa=fb; fb=fc; fc=fa;
    End If

    tol1 = 2*eps + 0.5 * tol
    xm = 0.5 * (c-b)
    if ((ABS(xm) <= tol1) .OR. fb == 0.0) then    !*** Check Convergence ***
       Finished = .TRUE.
       rt = b
    End If
 
    if ((ABS(e) >= tol1) .AND. (ABS(fa)>ABS(fb))) then !*** Try inv quad interp
       s = fb / fa

       if (a == c) then
          p = 2 * xm * s
          q = 1 - s
       else 
          q = fa/fc
          r = fb/fc
          p = s * (2*xm*q*(q-r) - (b-a)*(r-1))
          q = (q-1) * (r-1) * (s-1)
       End If

       if (p>0) q = -q             !*** Check whether in bounds

       p = abs(p)
       min1 = 3*xm*q - ABS(tol1*q)
       min2 = ABS(e*q)

       if (2*p < MIN(min1, min2)) then              
          e = d                   !*** Accept Step
          d = p/q
       else
          d = xm                  !** Step failed, use bisection
          e = d
       End If
    else                          !** Bounds decreasing slowly, use bisection
       d=xm 
       e=d 
    End If

    a=b; fa=fb;                   !** Move last guess to a

    if (ABS(d)>tol1) then         !** Evaluate new trial root
       b = b + d
    else
       b = b + SIGN(tol1, xm)
    End If

    fb = Func(b)
    NE = NE + 1
  End Do

  rt_Brent = rt

  if (PRESENT(NumEval)) NumEval = NE

  Return
End Function

       
!******************************************************************************
!  rt_Newton - Newton-Raphson with Bisection.  Returns the fn root .
!
!    FitFunc - External, user-supplied routine that calculates the value
!              of the function and its 1st derivative at a point.  It
!              has the following form 
!                     Double Precision  FitFunc(x, fn, deriv) 
!              where all arguments are double precision. The Return value is
!              ignored.
!
!    xmim,
!    xman    - Min and Max values of x that bracket the root. Automatically
!              swapped if F(xmin)>F(xmax)
!    xacc    - Accuracy of root requested
!    MAXITER - Max number of iterations to perform
!    NumEval - (OPTIONAL) Returns the number of function evaluations
!   
!******************************************************************************

Real Function rt_Newton(FitFunc, xmin, xmax, xacc, MaxIter, NumEval)
  Real, External                :: FitFunc
  Real,           Intent(INOUT) :: xmin
  Real,           Intent(INOUT) :: xmax
  Real,           Intent(IN)    :: xacc
  Integer,        Intent(IN)    :: MaxIter
  Integer, Optional, Intent(OUT):: NumEval

  Logical :: Finished
  Integer :: j, k, NE
  Real    :: df, dx, dxold, f, fhigh, flow
  Real    :: temp, xhigh, xlow, rts


  Finished = CheckBounds(FitFunc, Xmin, Xmax)
  
  if (.NOT. Finished) then
    write(*,*) 'Root Not Bracketed in [',Xmin, Xmax, ']'
    
    rt_Newton = Xmin
    Return
  End If
     
  xlow = xmin
  xhigh = xmax
  rts = 0.5 * (xlow+xhigh)
  dxold = ABS(xhigh-xlow)

  k = FitFunc(rts, f, df)
  NE = 1

  j = 0
  Finished = .FALSE.
  Do While ((j<MaxIter) .AND. .NOT. Finished)
    if (((((rts-xhigh)*df-f)*(rts-xlow)*df-f)>=0) .OR.  &
       (abs(2.0*f)>abs(dxold*df))) then       !** Bisect if out of range
       dxold = dx
       dx = 0.5 * (xhigh-xlow)
       rts = xlow + dx
     
       if (xlow == rts) Finished = .TRUE.
    else                                      !** Do a Newton step **
       dxold = dx
       dx = f / df
       temp = rts
       rts = rts - dx
       
       if (temp == rts) Finished = .TRUE.
    End If

    if (abs(dx) < xacc) Finished = .TRUE.

    k = FitFunc(rts, f, df)
    NE = NE + 1

    if (f < 0) then                     !** Maintain Root bracket
      xlow = rts
    else
      xhigh = rts
    End If
  End Do

  rt_Newton = rts

  if (PRESENT(NumEval)) NumEval = NE

  Return
End Function


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!  Subroutine xyztosphericalpolar(x,y,z,r,theta,phi)
!
!  subroutine to transform between cartesian and spherical polar coordinates
!
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

Subroutine xyztosphericalpolar(x,y,z,r,theta,phi)

  Real , intent(IN) :: x,y,z
  Real , intent(OUT) :: r,theta,phi

  Real, parameter :: tol=1.e-13
  Real   :: Pi

  Pi=acos(-1.0)

  r=0
  theta=0
  phi=0
  r=sqrt(x*x+y*y+z*z)

  If (r < tol) return

  theta=ACOS(z/r)
  
  If (ABS(x) < tol .AND. ABS(y) < tol) then
      Phi = 0
  ElseIf (ABS(x)< tol .AND. ABS(y) >= tol) then
      If ( y >= 0 ) Phi=Pi/2
      If ( y < 0 ) Phi=3*Pi/2
  ElseIf (ABS(y)< tol .AND. ABS(x) >= tol) then
      If (x >=0 ) Phi=0
      If (x < 0 ) Phi=Pi
  Else
      Phi=ATAN(y/x)
      If (Phi < 0 ) Phi=Phi+Pi
      If (y < 0 )   Phi=Phi+Pi
  EndIf

  Return
  End subroutine
  
!***********************************************************************************************
!     subroutine ddet5(h,y,z,ndim)
!      ssp routine modified by nawh 6/8/76
      !!!!  implicit real*8 (a-h,o-z)
!***********************************************************************************************

      subroutine Real_nderiv(h,y,z,ndim,Grid)
         integer, intent(IN) :: ndim
         real, intent(IN) ::  h,y(ndim)
         real, intent(OUT) ::  z(ndim)
         Type(GridInfo), INTENT(IN) :: Grid

         real :: hh,yy,a,b,c
         integer :: i,ier
!        prepare differentiation loop
         hh=.08333333333333333d0/h
         yy=y(ndim-4)
         b=hh*(-25.d0*y(1)+48.d0*y(2)-36.d0*y(3)+16.d0*y(4)-3.d0*y(5))
         c=hh*(-3.d0*y(1)-10.d0*y(2)+18.d0*y(3)-6.d0*y(4)+y(5))
!
!        start differentiation loop
         do  i=5,ndim
            a=b
            b=c
            c=hh*(y(i-4)-y(i)+8.d0*(y(i-1)-y(i-3)))
            z(i-4)=a
         enddo
!        end of differentiation loop
!
!        normal exit
         ier=0                                                              
         a=hh*(-yy+6.d0*y(ndim-3)-18.d0*y(ndim-2)+10.d0*y(ndim-1)          &
            +3.d0*y(ndim))                                               
         z(ndim)=hh*(3.d0*yy-16.d0*y(ndim-3)+36.d0*y(ndim-2)               &
                  -48.d0*y(ndim-1)+25.d0*y(ndim))
         z(ndim-1)=a
         z(ndim-2)=c
         z(ndim-3)=b
!
         if (Grid%islog) z(1:ndim)=z(1:ndim)/Grid%drdu(1:ndim)
       end subroutine
!***********************************************************************************************
!     subroutine ddet5(h,y,z,ndim)
!      ssp routine modified by nawh 6/8/76
      !!!!  implicit real*8 (a-h,o-z)
!***********************************************************************************************

      subroutine Complex_nderiv(h,y,z,ndim,Grid)
         integer, INTENT(IN) :: ndim
         real, INTENT(IN) ::  h
         complex, INTENT(IN) :: y(ndim)
         complex, INTENT(OUT) :: z(ndim)
         Type(GridInfo), INTENT(IN) :: Grid

         complex :: yy,a,b,c
         real :: hh
         integer :: i,ier

!        prepare differentiation loop
         hh=.08333333333333333d0/h
         yy=y(ndim-4)
         b=hh*(-25.d0*y(1)+48.d0*y(2)-36.d0*y(3)+16.d0*y(4)-3.d0*y(5))
         c=hh*(-3.d0*y(1)-10.d0*y(2)+18.d0*y(3)-6.d0*y(4)+y(5))
!
!        start differentiation loop
         do   i=5,ndim
            a=b
            b=c
            c=hh*(y(i-4)-y(i)+8.d0*(y(i-1)-y(i-3)))
            z(i-4)=a
         enddo
!        end of differentiation loop
!
!        normal exit
         ier=0                                                              
         a=hh*(-yy+6.d0*y(ndim-3)-18.d0*y(ndim-2)+10.d0*y(ndim-1)          &
            +3.d0*y(ndim))                                               
         z(ndim)=hh*(3.d0*yy-16.d0*y(ndim-3)+36.d0*y(ndim-2)               &
                  -48.d0*y(ndim-1)+25.d0*y(ndim))
         z(ndim-1)=a
         z(ndim-2)=c
         z(ndim-3)=b
!
         if (Grid%islog) z(1:ndim)=z(1:ndim)/Grid%drdu(1:ndim)
       end subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 Subroutine MatrixInverse123(MM,IM)
     Real, INTENT(IN) :: MM(:,:)
     REAL, INTENT(OUT) :: IM(:,:)

     Integer :: i,j,k
     Real :: det

     i=SIZE(MM,1)
     j=SIZE(MM,2)

     if (i /= j) then
        write(*,*) '123MatrixInverse:  Error -- matrix not square',&
              i,j
        stop
     endif

     If (i == 1) then
        IM(1,1)=1.0/MM(1,1)
     else if (i==2) then
        det=MM(1,1)*MM(2,2)-MM(1,2)*MM(2,1)
        IM(1,1)=MM(2,2)/det
        IM(2,2)=MM(1,1)/det
        IM(1,2)=-MM(1,2)/det
        IM(2,1)=-MM(2,1)/det
     else if (i==3) then
        det=MM(1,1)*(MM(2,2)*MM(3,3)-MM(2,3)*MM(3,2))+&
            MM(1,2)*(MM(2,3)*MM(3,1)-MM(2,1)*MM(3,3))+&
            MM(1,3)*(MM(2,1)*MM(3,2)-MM(2,2)*MM(3,1))
         IM(1,1)=(MM(2,2)*MM(3,3)-MM(2,3)*MM(3,2))/det
         IM(1,2)=(MM(1,3)*MM(3,2)-MM(1,2)*MM(3,3))/det
         IM(1,3)=(MM(1,2)*MM(2,3)-MM(1,3)*MM(2,2))/det
         IM(2,1)=(MM(2,3)*MM(3,1)-MM(2,1)*MM(3,3))/det
         IM(2,2)=(MM(1,1)*MM(3,3)-MM(1,3)*MM(3,1))/det
         IM(2,3)=(MM(1,3)*MM(2,1)-MM(1,1)*MM(2,3))/det
         IM(3,1)=(MM(2,1)*MM(3,2)-MM(2,2)*MM(3,1))/det
         IM(3,2)=(MM(1,2)*MM(3,1)-MM(1,1)*MM(3,2))/det
         IM(3,3)=(MM(1,1)*MM(2,2)-MM(1,2)*MM(2,1))/det
      else
         Write(*,*) 'MatrixInverse123:  Error -- matrix  too large', i
         stop
      endif
 End Subroutine MatrixInverse123

! subroutine to input number f and output shifted value such that
!     -1/2 < f <= 1/2
   subroutine fixhalfrange(f)
     real(8), INTENT(INOUT) :: f

     do
       if ( f < -0.5d0) then
          f=f+1.d0
       else
          exit
       endif
     enddo

     do
       if ( f >= 0.5d0) then
          f=f-1.d0
       else
          exit
       endif

     enddo
   end subroutine fixhalfrange

!********************************************************
! NIsEven(i)
!   Logical function which is true if i is even
!*******************************************************

 function NIsEven(i)
    Integer, INTENT(IN) :: i
    logical :: NIsEven
    
    NIsEven=.false.
    if ((i/2)*2==i) NIsEven=.true.
 End function

!********************************************************
! NIsOdd(i)
!   Logical function which is true if i is odd
!*******************************************************

 function NIsOdd(i)
    Integer, INTENT(IN) :: i
    logical :: NIsOdd
    
    NIsOdd=.true.
    if ((i/2)*2==i) NIsOdd=.false.
 End function

End Module
