!******************************************************************************
!
! File : mem_data.f90
!   by : Alan Tackett
!   on : 04/27/97
!  for : PAW Project
!
!  Contains the Data structures and constants for use by the MemMgr, Projector,
!  and Psi modules
!
! Updated for spin by Ping Tang and N. A. W. Holzwarth
!    Last change 5/23/05
!******************************************************************************

Module mem_data


Use misc
Use paw_inout
Use anderson_mixing

Implicit NONE!!!!!!!!!!!!!!!!!!!

Integer, PARAMETER :: SIZEOF_C    = 16
Integer, PARAMETER :: SIZEOF_COMPLEX = 16
Integer, PARAMETER :: SIZEOF_REAL = 8

Integer,      PARAMETER  :: BG_KPNT     = -3 !** Special processing index     

!*Mem_Handle constants for *_toProcess, *_Processsed, *_UnProcessed variables *
Integer,      PARAMETER  :: MH_PROCESSED   = 0  !** Vec processed
Integer,      PARAMETER  :: MH_toPROCESS   = 1  !** To be Processed
Integer,      PARAMETER  :: MH_SKIP        = 0  !** Ignore or skip vec
Integer,      PARAMETER  :: MH_UnProcessed = 1  !** Vec currently not processed
Integer,      PARAMETER  :: Mem_Available =  1  !** Memory slot available
Integer,      PARAMETER  :: Mem_Used      =  0  !** Memory slot used     

Real             :: MemSize        !** Amount of buffer size in Mbytes
Integer            :: MemWords       !** Size of buf in Complex*16 units
Integer            :: MemFree        !** Amount of free space
Integer            :: BaseFree       !** Size of buffer w/o V 

Type Cache_Handle   !** Cache data structure
   Integer :: Hits
   Integer :: Misses
   Integer :: Total
End Type

Type Mem_Cache     !** Cache Data structure
   Type (Cache_handle) :: Psi_Read
   Type (Cache_handle) :: Psi_Write
End Type

Type(Mem_Cache), Pointer  :: MemCache

Type Mem_handle       !*** Memory Allocation Handle ***
   Integer                  :: Index       !** Data Index, PsiInfo number
   Complex, Pointer :: Ptr(:)      !** Pointer to data
End Type

Type global_handle     !*** Memory Allocation for set of data ***
   Integer                    :: NumPsi          !** Number of Psi's alloc
   Integer                    :: PsiLoaded      !** Number of Psi's loaded
   Type (Mem_Handle), Pointer :: PSI_handle(:)   !** PSI memory handles
End Type

Type (global_Handle), Pointer  :: Globalmap

Integer, Pointer :: Psi_Processed(:)   !**LUT denoting which psi have been used
Integer, Pointer :: Psi_toProcess(:)   !** LUT of Psi's to process(=0)
Integer, Pointer :: PsiBand_toProcess(:)

Integer, Pointer :: ClusterMap(:)      !** LUT determining eigenvalue clusters
Integer, Pointer :: ClusterSize(:)     !** Size of each cluster
Integer          :: LastCLuster        !** Last cluster used

Integer :: Mem_Kpnt

Type Potential_handle      !** Potential Handle **
   Complex,  Pointer :: Ve(:),Vespin(:)      !** Veff in real space
   Complex,  Pointer :: Work(:)      !** Work array
   Complex,  Pointer :: V_local(:) !** Local Potential
   Complex,  Pointer  :: Ve_Fourier(:),Ve_Fourierspin(:) !** Veff in Fourier Space
   Complex,  Pointer  ::cVe_F(:),cVe_Fspin(:) !** current Veff in Fourier Space
                ! not from Anderson mix
   Complex,  Pointer :: RhoSmooth(:),RhoSmoothspin(:)    !** Smooth, n~, charge density
   Complex,  Pointer :: RhoHat(:)       !** n^ charge density
   Complex,  Pointer :: CoreTail(:)     !** Coretail density
   Complex,  Pointer :: VXC(:),VXCSPIN(:)     !** VXC in Fourier Space
   Type (Anderson_Context), Pointer :: AC, ACspin
End Type

Type (Potential_handle), Pointer,save  ::   SCFvalues !** Potential ptr's

Type BufferPtr         !*** Temp Buffer Ptr 
   Complex, Pointer :: Ptr(:)
End Type

Type (BufferPtr), Pointer :: WorkPtr(:)


Type PsiEigen_Handle  !** Contains info about how the eigen maps to the BZ data
   Integer      :: Kpnt        !** Kpnt the Psi maps to
   Logical      :: Spinup      !** used for spin polarized calculation
   Integer      :: DoSave      !** 1 if wavefunction should be saved to disk
   Integer      :: Available   !** Tracks availability of virtual memory slot
   Integer      :: CorrIndex   !** Correction index
   Integer      :: OxIndex     !** Ox disk index
   Integer      :: MemBufIndex  !** Memory buffer index
   Real :: Occupancy   !** Original occupancy
   REal :: Energy      !** Energy eigen value of the Psi
   Real :: KE          !** Kinetic energy for band
   Real :: Error       !** Error in Hx-eOx
   Logical :: PDOT_Stored 
   Logical :: KE_Stored 
   Complex, Pointer :: PDOT(:)
End Type

Type (PsiEigen_Handle), Pointer :: PsiInfo(:)  !** LUT for psi info

Integer :: TotalPsi           !** Total Number of Psi's (Psi_bands*Psi_Kpnts)
Integer :: Mem_MaxPsi         !** Max number of Psi's per K-pnt.
Integer :: ExtraPsi           !** Min number of extra Psi's
Integer :: MinBands           !** Min Num of bands to have(NumBands+ExtraPsi)
Integer :: Mem_MapSize        !** LUT, and Map allocated size
Integer :: PsiArraySize       !** Array size for each Psi(G)

!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
!  GetNextIndex_toProcess - Returns the next index to be processed.
!
!  Index     - The value of Index on entry determines the first index
!              for looking for a valid index to process. On exit INDEX
!              contains the next index to process or 0           if no
!              more Indices were found.
!  toProcess - Array containing 1's for Indices to process and 0's for
!              indices to skip. NOTE: This array is not modified in any way!!
!
!******************************************************************************

Subroutine GetNextIndex_toProcess(Index, toProcess)
  Integer, Intent(INOUT) :: Index
  Integer, Intent(INOUT)    :: toProcess(:)

  Index=1
  Do While ((toProcess(Index) /= MH_toProcess) .AND. (Index <= Mem_MapSize))
     index = Index + 1
  End Do

  If (Index > Mem_MapSize) Index=0

  Return
End Subroutine

!******************************************************************************
!
!   Swap_Mem_Handle - Swaps two memory handles
!
!   MH1, MH2 - The 2 handles to swap
!
!******************************************************************************

Subroutine Swap_Mem_Handles(MH1, MH2)
  Type (Mem_Handle), Intent(INOUT) :: MH1
  Type (Mem_Handle), Intent(INOUT) :: MH2

  Type (Mem_Handle) :: Temp


  Temp%Index = MH1%Index;    Temp%Ptr => MH1%Ptr

  MH1%Index = MH2%Index;     MH1%Ptr => MH2%Ptr

  MH2%Index = Temp%Index;    MH2%Ptr => Temp%Ptr

  Return
End Subroutine


!******************************************************************************
!
!  CanAlloc - checks to see if it is possible to allocate the extra space
!          for a Psi
!
!  LSize     - Ptr size
!
!******************************************************************************

Logical Function CanAlloc( LSize)
  Integer, Intent(IN) :: LSize

  Logical :: Ok

  Ok = .FALSE.
  
        If (((MemFree - LSize) > 0) .AND. &
                (Globalmap%Numpsi<Mem_MapSize)) Ok = .TRUE.

  CanAlloc = Ok

  Return
End Function


!******************************************************************************
!
!  AllocMemHandle - Allocates a data pointer to a memory handle
!
!  LSize - Ptr size
!  MH1   - Pointer to the Memory handle returned
!
!******************************************************************************

Subroutine AllocMemHandle(  LSize, slot, MH1)
  Integer,                  Intent(IN) :: LSize
  Integer,                  Intent(OUT) :: Slot
  Type (Mem_Handle), Pointer           :: MH1
  
  Integer :: i, k
  Type (Global_Handle), Pointer :: LH
  Character*100 :: msg


  LH => Globalmap

        LH%NumPsi = LH%NumPsi+1;
        MH1 => LH%Psi_handle(LH%NumPsi)

        Allocate(MH1%Ptr(LSize), STAT=k)
        Write(msg,*) 'AllocMemHandle: Could Not allocate Psi buffer!',  &
              ' * Size:',LSize
        Call Check_Error(k, msg, Error_Unit, .TRUE.,paw_wc, "AllocMemHandle:")
        

        MH1%Index = 0            
        
        MemFree = MemFree - LSize

  slot = LH%NumPsi

  Return
End Subroutine

End Module

