!******************************************************************************
!
! File : oinverse.f90
!   by : Alan Tackett
!   on : 09/20/99
!  for : PWPAW
!
!  Routines for calculating O^(-1), ie the inverse of the overlap matrix
!
!******************************************************************************

Module oinverse

Use lrulib
Use gpoints
Use work_mgr
Use atom_data
Use crystal_data
Use spherical_harmonic
Use structfact
Use ylm_fact
Use projectors
Use doijmatrix

Implicit NONE!!!!!

Type RLE_Oi_type
   Integer          :: N
   Integer, Pointer :: RLE(:,:)
   Complex, Pointer :: Oinv(:)
End Type

Type (RLE_Oi_Type), Pointer :: RS_Oinv(:,:)

Type (COMPLEX_LRU_Context) :: LRU_GS_Oinv
Type (COMPLEX_LRU_Context) :: LRU_RS_Oinv
Complex, Pointer ::  GS_Oinv_Scale(:,:)
Complex, Pointer ::  RS_Oinv_Scale(:,:)
Complex, PRIVATE, Pointer :: Oi_Work1(:)
Complex, PRIVATE, Pointer :: Oi_Work2(:)

Integer, Pointer, PRIVATE :: Group(:)
Integer, Pointer, PRIVATE :: Group_Nmax(:)
Integer,          PRIVATE :: NumGroups
Integer,          PRIVATE :: NumRecsPerKpnt

!******************************************************************************
Contains
!******************************************************************************


!******************************************************************************
!
!  RS_Oinverse - Calculates O**(-1) times V in R-space
!
!******************************************************************************

Subroutine RS_Oinverse(V, kpnt, MaxAtom, MaxN)
  Complex,           Intent(INOUT) :: V(:)
  Integer,           Intent(IN)    :: Kpnt
  Integer, OPTIONAL, Intent(IN)    :: MaxAtom
  Integer, OPTIONAL, Intent(IN)    :: MaxN
  
  Integer :: i,j,k,dn, Base_DI,  atom, n, N_rle, DI
  Integer :: LastAtom, LastN
  Complex, Pointer :: Proj_R(:), Oi(:)
  Complex :: dot, dot2, dot3
  Real :: DV
  Integer, Pointer :: RLE(:,:)

!*V=V/(FFT_Grid(4,G_PROJ))**2
!*V=0
!RETURN

!*  DV = Xtal%Volume/(1.0*FFT_Grid(4,G_PROJ))
!DV = 1.0/(1.0*FFT_Grid(4,G_PROJ))
DV = 1

  DI = PLM_Max*(Kpnt-1)
!Write(Log_Unit,*) 'RS_Oinv: DI=',DI, ' * Kpnt=',Kpnt


  LastAtom = Specific_Atoms
  If (Present(MaxAtom)) LastAtom = MaxAtom
  Do atom=1, LastAtom
     Base_DI = RS_PLMMax(atom)*(Kpnt-1) 

     LastN = RS_PLMMax(atom)
     If (atom==LastAtom) then
        If (Present(MaxAtom)) LastN = MaxN
     End If

     Do n=1, LastN
        dot = 0
        Call LRU_GetRec(LRU_RS_Proj(atom),  Base_DI + n, Proj_R)
        i = 0;  j = 0;
        Do k=1, RS_RLE(atom)%N
           j = RS_RLE(atom)%RLE(1,k)-1
           dn = RS_RLE(atom)%RLE(2,k)
           dot = dot + DOT_PRODUCT(Proj_R(i+1:i+dn), V(j+1:j+dn))
           
           i = i + dn
        End Do
        
        DI = DI + 1

        N_rle = RS_Oinv(RS_PLM(atom,4,n),kpnt)%N
        RLE => RS_Oinv(RS_PLM(atom,4,n),kpnt)%RLE
        Oi => RS_Oinv(RS_PLM(atom,4,n),kpnt)%Oinv

        dot = RS_Oinv_Scale(RS_PLM(atom,4,n),Kpnt)*dot
        i = 0;  j = 0;
!Write(Log_Unit,*) 'RS_Oinv: DOT=',dot, ' Oi=',Oi, ' * Size(Oi)=',size(Oi)
        Do k=1, N_rle
           j = RLE(1,k)-1
           dn = RLE(2,k)
!write(Log_unit,*) 'k=',k, ' ij=',i,j,' dn=',dn, ' * V=',V(j+1:j+dn), ' * Oi=',Oi(i+1:i+dn)
           V(j+1:j+dn) = V(j+1:j+dn) - dot*Oi(i+1:i+dn)
           
           i = i + dn
        End Do


!**        V = V - RS_Oinv_Scale(RS_PLM(atom,4,n),Kpnt)*dot*Oi
     End Do !**n
  End Do !** atom

  Return
End Subroutine


!******************************************************************************
!
!  GS_Oinverse_TIME - Applies the inverse of the overlap matrix in G-space to
!                the given vector
!
!  V    - On Input the vector to multiply by. On output V=O**-1 * V
!  MaxN - Optional Max inverse vector to apply.  Used during construction of
!         the inverse vectors.
!
!*TIME*
!******************************************************************************

Subroutine GS_Oinverse_TIME(V, Kpnt, MaxAtom, MaxN)
  Complex, Intent(INOUT) :: V(:)
  Integer, Intent(IN)    :: Kpnt
  Integer, OPTIONAL, Intent(IN) :: MaxAtom
  Integer, OPTIONAL, Intent(IN) :: MaxN

  Integer :: i,j,k,level, m,am, L, n, t, G_Size, G_Half, atom
  Integer :: Base_DI, DI, RadP, PLM_Index
  Integer :: NG, atype, LastN, LastAtom
  Complex :: dot
  Complex, Pointer :: Oinv(:), Proj(:)


  LastAtom = Specific_Atoms
  If (Present(MaxAtom)) LastAtom = MaxAtom

  NG = Gpnt_Size(G_Low) - 1

  DI = PLM_Max*(Kpnt-1)

  Do atom=1, LastAtom
     atype = Atom_List(atom)%typeIndex
     !G_Half = Atomtype_Info(Atom_List(atom)%typeIndex)%Gpnt_size
     !G_Size = 2*G_Half - 1

     G_Half = Gpnt_Size(G_LOW)
     G_Size = Gall_Size(G_LOW)
        
     Base_DI = RS_PLMMax(atom) * (Kpnt-1)  !** Calc starting DI for Kpnt

     LastN = RS_PLMMax(atom)
     If (atom==LastAtom) then
        If (Present(MaxAtom)) LastN = MaxN
     End If
     Do n=1, LastN
        RadP = RS_PLM(atom,1,n)
        L = RS_PLM(atom,2,n)
        m = RS_PLM(atom,3,n)
        am = ABS(m)
        PLM_Index = RS_PLM(atom,4,n)

        Call LRU_GetRec(LRU_Proj(atom),  Base_DI + n, Proj)
        i = 1
        dot = Proj(i)*V(i)
        
        Do i=2, G_Half
           dot = dot + Proj(i)*V(i)
           dot = dot + Proj(i+G_Half-1) * V(i+ng)
        end Do

        DI = DI + 1
        Call LRU_GetRec(LRU_GS_Oinv,  DI, Oinv)

!Write(Log_Unit,*) 'GS_Oinverse: atom=',atom, ' * n=',n, ' * DOT=',dot, ' * Gscale=',GS_Oinv_Scale(PLM_Index), ' * Both=',dot*GS_Oinv_Scale(PLM_Index)
        V = V - dot*Oinv*GS_Oinv_Scale(PLM_Index, Kpnt)
     End Do
  End Do !*atom

!Write(Log_Unit,*) 'GS_Oinverse_TIME:...........'
  Return
End subroutine


!******************************************************************************
!
!  GS_Oinverse_MEM - Applies the inverse of the overlap matrix in G-space to
!                the given vector
!
!  V    - On Input the vector to multiply by. On output V=O**-1 * V
!  MaxN - Optional Max inverse vector to apply.  Used during construction of
!         the inverse vectors.
!
!*MEMORY*
!******************************************************************************

Subroutine GS_Oinverse_MEM(V, Kpnt, MaxAtom, MaxN)
  Complex, Intent(INOUT) :: V(:)
  Integer, Intent(IN)    :: Kpnt
  Integer, OPTIONAL, Intent(IN) :: MaxAtom
  Integer, OPTIONAL, Intent(IN) :: MaxN

  Integer :: i,j,k,level, m,am, L, n, t, G_Size, G_Half, atom
  Integer :: Base_RAD, Base_DI, G_Index, DI, RadP, PLM_Index
  Integer :: Basis_size, NG, atype, LastN, LastAtom
  Complex :: c1,c2,c3, Cij, dot
  Complex, Pointer :: Oinv(:), Phase(:), Ylm(:)
  Real, Pointer :: Proj(:)


  LastAtom = Specific_Atoms
  If (Present(MaxAtom)) LastAtom = MaxAtom

  NG = Gpnt_Size(G_Low) - 1
  c1 = 4*Pi / sqrt(xtal%Volume)

  DI = PLM_Max*(Kpnt-1)

  Do atom=1, LastAtom
     atype = Atom_List(atom)%typeIndex
     Basis_Size = AtomType_Info(atype)%Basis_Size
        
     !G_Half = Atomtype_Info(Atom_List(atom)%typeIndex)%Gpnt_size
     !G_Size = 2*G_Half - 1

     G_Half = Gpnt_Size(G_LOW)
     G_Size = Gall_Size(G_LOW)
        
     Call GetStructFactor( atom, Phase)
        
     Base_DI = Basis_Size * (Kpnt-1)  !** Calc starting DI for Kpnt

     LastN = RS_PLMMax(atom)
     If (atom==LastAtom) then
        If (Present(MaxAtom)) LastN = MaxN
     End If
     Do n=1, LastN
        RadP = RS_PLM(atom,1,n)
        L = RS_PLM(atom,2,n)
        m = RS_PLM(atom,3,n)
        am = ABS(m)
        PLM_Index = RS_PLM(atom,4,n)

        Call LRU_GetRec(LRU_RadProj(atype),  Base_DI + radP, Proj)
        Call GetYlm( Kpnt, L, am, Ylm)        
        
        c2 = CMPLX(0,1)**L
        
        dot = 0
        If (m>=0) then
           i = 1
           dot = CONJG(Ylm(i))*Proj(i)*Phase(i)*V(i)
           
           Do i=2, G_Half
              dot = dot + CONJG(Ylm(i))*Proj(i)*Phase(i)*V(i)
              dot = dot + CONJG(Ylm(i+ng)) * Proj(i+G_Half-1) * &
                   Phase(i+ng)*V(i+ng)
           end Do
           dot = c1*c2*dot
        Else             
           i = 1
           dot = Ylm(i)*Proj(i)*Phase(i)*V(i)
           
           Do i=2, G_Half
              dot = dot + Ylm(i)*Proj(i)*Phase(i)*V(i)
              dot = dot + Ylm(i+ng) * Proj(i+G_Half-1) * &
                   Phase(i+ng)*V(i+ng)
           end Do
           
           c3 = (-1)**am
           dot = c1*c2*c3*dot
        End If
        

        DI = DI + 1
        Call LRU_GetRec(LRU_GS_Oinv,  DI, Oinv)

!Write(Log_Unit,*) 'GS_Oinverse: atom=',atom, ' * n=',n, ' * DOT=',dot, ' * Gscale=',GS_Oinv_Scale(PLM_Index), ' * Both=',dot*GS_Oinv_Scale(PLM_Index)
        V = V - dot*Oinv*GS_Oinv_Scale(PLM_Index, Kpnt)
     End Do
  End Do !*atom

  Return
End subroutine


!******************************************************************************
!
!  Oinverse - Applies the inverse of the overlap matrix in G-space to
!                the given vector
!
!  V    - On Input the vector to multiply by. On output V=O**-1 * V
!  MaxN - Optional Max inverse vector to apply.  Used during construction of
!         the inverse vectors.
!
!*DRIVER*
!******************************************************************************

Subroutine CalcOinverse(V, Kpnt)
  Complex, Intent(INOUT) :: V(:)
  Integer, Intent(IN)    :: Kpnt

!RETURN
  Select Case (Proj_Mode)
     Case (MIN_TIME)
        Call GS_Oinverse_TIME(V, Kpnt)
     Case (MIN_MEMORY)
        Call GS_Oinverse_MEM(V, Kpnt)
     Case (PROJ_RS)
!        RS_Work1 = 0
!        Call RS_toR(V, RS_Work1)
!        RS_Work1 = RS_Work1/(1.0*FFT_Grid(4,G_PROJ))
!        Call RS_Oinverse(RS_Work1, Kpnt)
        Call RS_Oinverse(V, Kpnt)

!        V = 0
!        Call RS_toG(RS_Work1, V)
!        V = V * (1.0*FFT_Grid(4,G_PROJ))
  End Select

  Return
End Subroutine

!******************************************************************************
!
!  RS_Psum - Sums the projectors for the given group
!
!******************************************************************************

Subroutine RS_Psum(V, Kpnt, atom, njlj, Lj, mj)
  Complex, Intent(OUT) :: V(:)
  Integer, Intent(IN)  :: Kpnt
  Integer, Intent(IN)  :: atom
  Integer, Intent(IN)  :: njlj
  Integer, Intent(IN)  :: Lj
  Integer, Intent(IN)  :: mj

  Integer :: i,j,k,dn, Base_DI,  n, DI
  Integer :: nili, Li, mi
  Complex, Pointer :: Proj_R(:)
  Complex :: dot, Oij
  Real :: DV

!*  DV = Xtal%Volume/(1.0*FFT_Grid(4,G_PROJ))
DV = 1.0*(1.0*FFT_Grid(4,G_PROJ))
!DV = 1

  Base_DI = RS_PLMMax(atom)*(Kpnt-1) 
  V = 0
  Do n=1, RS_PLMMax(atom)
     nili = RS_PLM(atom, 1, n)
     Li = RS_PLM(atom,2,n)
     mi = RS_PLM(atom,3,n)
     
     Oij = OijMatrix(Atom, nili, njlj, Li, lj, mi, mj)

     Call LRU_GetRec(LRU_RS_Proj(atom),  Base_DI + n, Proj_R)
     i = 0;  j = 0;
     Do k=1, RS_RLE(atom)%N
        j = RS_RLE(atom)%RLE(1,k)-1
        dn = RS_RLE(atom)%RLE(2,k)
        V(j+1:j+dn) = V(j+1:j+dn) + Proj_R(i+1:i+dn)*Oij
        
        i = i + dn
     End Do
  End Do !**n

Write(Log_Unit, *) 'RS_Psum: atom=',atom, ' n=', n, ' * Kpnt=',kpnt, ' * DOT=',DOT_PRODUCT(V,V)    
  Return
End Subroutine



!******************************************************************************
!
!  GS_Psum - Sums the projectors for the given atom
!
!  PSum - Output vector
!  Kpnt - K-point
!  atom - atom for sum
!  njlj, Lj, mj - Indices of "other" vector for Oij factor in sum 
!
!******************************************************************************

Subroutine GS_Psum(Psum_G, Kpnt, atom, njlj, Lj, mj)
  Complex, Intent(OUT) :: Psum_G(:)
  Integer, Intent(IN)  :: Kpnt
  Integer, Intent(IN)  :: atom
  Integer, Intent(IN)  :: njlj
  Integer, Intent(IN)  :: Lj
  Integer, Intent(IN)  :: mj

  Integer :: i,j,k,level, m,am, L, n, t, G_Size, G_Half
  Integer :: Base_RAD, Base_DI, G_Index, DI, RadP
  Integer :: Basis_size, NG, atype
  Complex :: c1,c2,c3, Oij
  Complex, Pointer :: Proj_G(:), Ylm(:), Phase(:)
  Real, Pointer :: Proj(:)



  c1 = 4*Pi / sqrt(xtal%Volume)

  Call Getbuffer( Proj_G)

  NG = Gpnt_size(G_LOW) - 1

  Psum_G = 0

  atype = Atom_List(atom)%typeIndex
  Basis_Size = AtomType_Info(atype)%Basis_Size
  
  !G_Half = Atomtype_Info(Atom_List(atom)%typeIndex)%Gpnt_size
  !G_Size = 2*G_Half - 1

  G_Half = Gpnt_Size(G_LOW)
  G_Size = Gall_Size(G_LOW)
  
  Call GetStructFactor( atom, Phase)
  
  Base_DI = Basis_Size * (Kpnt-1)  !** Calc starting DI for Kpnt
  
  Do n=1, RS_PLMMax(atom)
     RadP = RS_PLM(atom,1,n)
     L = RS_PLM(atom,2,n)
     m = RS_PLM(atom,3,n)
     am = ABS(m)
     
     Call LRU_GetRec(LRU_RadProj(atype),  Base_DI + radP, Proj)
     Call GetYlm( Kpnt, L, am, Ylm)        
     
     c2 = CMPLX(0,1)**L
     
     Proj_G = 0
     If (m>=0) then
        i = 1
        Proj_G(i) = CONJG(Ylm(i))*Proj(i)*Phase(i)
        
        Do i=2, G_Half
           Proj_G(i) = CONJG(Ylm(i))*Proj(i)*Phase(i)
           Proj_G(i+ng) = CONJG(Ylm(i+ng)) * Proj(i+G_Half-1) * &
                Phase(i+ng)
        end Do
        Proj_G = c1*c2*Proj_G
     Else             
        i = 1
        Proj_G(i) = Ylm(i)*Proj(i)*Phase(i)
        
        Do i=2, G_Half
           Proj_G(i) = Ylm(i)*Proj(i)*Phase(i)
           Proj_G(i+ng) = Ylm(i+ng) * Proj(i+G_Half-1) * &
                Phase(i+ng)
        end Do
        
        c3 = (-1)**am
        Proj_G = c1*c2*c3*Proj_G
     End If

Write(Log_Unit, *) 'GS_Psum: atom=',atom, ' n=', n, ' * Kpnt=',kpnt, ' * DOT=',DOT_PRODUCT(Proj_G, Proj_G)
     
     Oij = OijMatrix(Atom, radP, njlj, L, lj, m, mj)

     Psum_G = Psum_G + CONJG(Proj_G)*Oij
  End Do
  
  Call Freebuffer(Proj_G)

  Return
End Subroutine


!**********************************************************************
!
! RS_RLE_Vector - Stores the vector in compressed form
!
!**********************************************************************

subroutine RS_RLE_Vector(DoStore, rle_size, data_size, RLE, Proj_RAW, Proj_RLE)
  Logical, Intent(IN) :: DoStore
  Integer, Intent(OUT) :: rle_size
  Integer, Intent(OUT) :: data_size
  Integer, Intent(OUT) :: RLE(:,:)
  Complex, Intent(IN)  :: Proj_RAW(:)
  Complex, Intent(OUT) :: Proj_RLE(:)

  
  Integer :: t, i,j,k
  Integer :: npnts, run_Start
  Logical :: Ok, InRun
  Real    :: num


  If (DoStore) RLE = 0
  rle_size = 0
  data_size = 0
  npnts = 0
  t = 0
  Inrun = .FALSE.
  Do t=1, FFT_Grid(4,G_PROJ)
     num = SQRT(ABS(REAL(CONJG(Proj_Raw(t))*Proj_raw(t))))
     If (num > 1E-20) then
        If (.NOT. InRun) then
           Run_Start = t
           InRun = .TRUE.
        End If

        npnts = npnts + 1
!Write(Log_Unit,*) 'RS_RLE: t=',t, ' * n=',npnts, ' * RLE=',rle_size
        If (DoStore) Proj_RLE(npnts) = Proj_RAW(t)
     else If (Inrun) then
        rle_size = rle_size + 1
        data_size = data_size + t-Run_Start
        If (DoStore) RLE(:,rle_size) = (/Run_start, (t-Run_Start) /)
        InRun = .FALSE.
!Write(Log_Unit,*) 'RS_RLE: Start=',run_Start, ' * n=',npnts, ' * RLE=',rle_size, ' * data_size=',data_size
     End If
  End Do 

  If (InRun) then
     t = FFT_Grid(4,G_Proj)
     rle_size = rle_size + 1
        data_size = data_size + t-Run_Start+1
     If (DoStore) RLE(:,rle_size) = (/Run_start, (t-Run_Start+1) /)
!Write(Log_Unit,*) 'RS_RLE: Start=',run_Start, ' * n=',data_size, ' * RLE=',rle_size
  end If
  If (DoStore) Write(Log_Unit,*) 'RS_RLE: Total points:',npnts, &
                  ' RLE Size:',rle_size, ' * Data_size=',data_size

  data_size = npnts

  Return
end Subroutine



!******************************************************************************
!
!  RS_CreateOinverse - Creates the Oinverse vectors in R-space
!
!******************************************************************************

Subroutine RS_CreateOinverse
  Integer :: i,j,k,dn, Base_DI,  atom, n, Kpnt, grp, DI
  Complex, Pointer :: Proj_R(:), Work1(:), Work2(:)
  Complex :: dot
  Real :: DV, r

!*  DV = Xtal%Volume/(1.0*FFT_Grid(4,G_PROJ))
DV = 1.0/(1.0*FFT_Grid(4,G_PROJ))
!DV = 1


  Allocate(Work1(FFT_Grid(4,G_PROJ)), Work2(FFT_Grid(4,G_PROJ)))

!Write(Log_Unit,*) 'CPP_RS: Base_DI=',Base_DI, ' * Kpnt=',Mem_Kpnt

  DI = 0
  Do Kpnt=1, Size(BZ%Ku(1,:))
     Do atom=1, Specific_Atoms
        Base_DI = RS_PLMMax(atom)*(Kpnt-1) 

        Do n=1, RS_PLMMax(atom)
           Work1 = 0

           Call LRU_GetRec(LRU_RS_Proj(atom),  Base_DI + n, Proj_R)
           i = 0;  j = 0;
           Do k=1, RS_RLE(atom)%N
              j = RS_RLE(atom)%RLE(1,k)-1
              dn = RS_RLE(atom)%RLE(2,k)
              Work1(j+1:j+dn) = Proj_R(i+1:i+dn)
              
              i = i + dn
           End Do

           DI = DI + 1

           Call RS_Psum(Work2, Kpnt, atom, RS_PLM(atom,1,n), &
                RS_PLM(atom,2,n), RS_PLM(atom,3,n))

Work2 = Work2 * DV           
           Call RS_Oinverse(Work2, kpnt, atom, n-1)

           dot = DOT_PRODUCT(Work1, Work2) !* DV !??????
           RS_Oinv_Scale(RS_PLM(atom,4,n), Kpnt) = 1.0/(1.0+dot)

Write(Log_Unit, *) 'RS_CreateOinv: atom=',atom, ' n=', n, ' * Kpnt=',kpnt, ' * WORK1 DOT=',DOT_PRODUCT(Work1, Work1)
Write(Log_Unit,*) '    dot=',dot,' * scale=',RS_Oinv_Scale(RS_PLM(atom,4,n),kpnt), ' * DI=',DI

j=0
Do i=1, FFT_Grid(4,G_PROJ)
   r = CONJG(Work2(i))*Work2(i)
   r = SQRT(ABS(r))
   If (ABS(r)>1E-20) j = j + 1
end Do
Write(log_Unit,*) 'RS_CreateOinv: DI=',DI, ' * NonZero=',j

           k = RS_PLM(atom,4,n)
           If (RS_Oinv(k,kpnt)%N>0) then
              DeAllocate(RS_Oinv(k,kpnt)%RLE, RS_Oinv(k,kpnt)%Oinv)
           End If

           Call RS_RLE_Vector(.FALSE., i, j, RS_Oinv(k,kpnt)%RLE, Work2, Work1)

           RS_Oinv(k,kpnt)%N = i
           Allocate(RS_Oinv(k, Kpnt)%RLE(2,i), RS_Oinv(k,kpnt)%Oinv(j))
           Call RS_RLE_Vector(.TRUE., i, j, RS_Oinv(k,kpnt)%RLE, &
                Work2, RS_Oinv(k,kpnt)%Oinv)
        End do
     End Do !** atom
  End Do

  DeAllocate(Work1, Work2)

  Return
End Subroutine

!******************************************************************************
!
!  GS_CreateOinverse - Creates the Oinverse vectors in G-space
!
!******************************************************************************

Subroutine GS_CreateOinverse
  Integer :: i,j,k,atom, level, m,am, L, n, kpnt, t, G_Size, G_Half
  Integer :: Base_RAD, Base_DI, G_Index, MaxRec, RS_Size(0:0), DI, RadP
  Integer :: Basis_size, NG, atype
  Integer, Pointer :: Map(:)
  Real, Pointer :: Proj(:), Wt(:)
  Complex, Pointer :: Work(:), Ylm(:), Phase(:), Proj_G(:), Proj_R(:)
  Complex, Pointer :: Buffer(:), Psum_G(:)
  Complex :: c1,c2,c3, c4, dot
  Real :: DV, R(3), R_xtal(3,27)
Integer :: x,y,z,xo,yo,zo, rle_size, data_size
Real :: r1,r2

  DV = Xtal%Volume/(1.0*FFT_Grid(4,G_PROJ))


  Call Getbuffer( Proj_G)
  Call Getbuffer( Psum_G)
  Call GetBuffer( Work)

  c1 = 4*Pi / sqrt(xtal%Volume)


  NG = Gpnt_size(G_LOW) - 1

  DI = 0
  Do Kpnt = 1, Size(BZ%Ku(1,:))
     Do atom=1, Specific_Atoms
        atype = Atom_List(atom)%typeIndex
        Basis_Size = AtomType_Info(atype)%Basis_Size
        
        !G_Half = Atomtype_Info(Atom_List(atom)%typeIndex)%Gpnt_size
        !G_Size = 2*G_Half - 1

        G_Half = Gpnt_Size(G_LOW)
        G_Size = Gall_Size(G_LOW)
        
        Call GetStructFactor( atom, Phase)
        
        Base_DI = Basis_Size * (Kpnt-1)  !** Calc starting DI for Kpnt

        Do n=1, RS_PLMMax(atom)
           RadP = RS_PLM(atom,1,n)
           L = RS_PLM(atom,2,n)
           m = RS_PLM(atom,3,n)
           am = ABS(m)
           
           Call LRU_GetRec(LRU_RadProj(atype),  Base_DI + radP, Proj)
           Call GetYlm( Kpnt, L, am, Ylm)        
           
           c2 = CMPLX(0,1)**L

           Proj_G = 0
           If (m>=0) then
              i = 1
              Proj_G(i) = CONJG(Ylm(i))*Proj(i)*Phase(i)
              
              Do i=2, G_Half
                 Proj_G(i) = CONJG(Ylm(i))*Proj(i)*Phase(i)
                 Proj_G(i+ng) = CONJG(Ylm(i+ng)) * Proj(i+G_Half-1) * &
                      Phase(i+ng)
              end Do
              Proj_G = c1*c2*Proj_G
           Else             
              i = 1
              Proj_G(i) = Ylm(i)*Proj(i)*Phase(i)
              
              Do i=2, G_Half
                 Proj_G(i) = Ylm(i)*Proj(i)*Phase(i)
                 Proj_G(i+ng) = Ylm(i+ng) * Proj(i+G_Half-1) * &
                      Phase(i+ng)
              end Do
              
              c3 = (-1)**am
              Proj_G = c1*c2*c3*Proj_G
           End If

!Proj_G = CONJG(Proj_G)
Write(Log_Unit, *) 'GS_CreateOinv: atom=',atom, ' n=', n, ' * Kpnt=',kpnt, ' * DOT=',DOT_PRODUCT(Proj_G, Proj_G)
           Call GS_Psum(Psum_G, Kpnt, atom, radp, L, m)
!Psum_G = CONJG(PSum_G)
           Work = PSum_G
           Call GS_Oinverse_MEM(Work, kpnt, atom, n-1)


           dot = SUM(Proj_G*Work)
           GS_Oinv_Scale(RS_PLM(atom,4,n),Kpnt) = 1.0/(1.0+dot)
           DI = DI + 1

Write(Log_Unit, *) 'GS_CreateOinv: atom=',atom, ' n=', n, ' * Kpnt=',kpnt, ' * WORK DOT=',DOT_PRODUCT(Work, Work)
Write(Log_Unit,*) '    dot=',dot,' * scale=',GS_Oinv_Scale(RS_PLM(atom,4,n),Kpnt), ' * DI=',DI
           Call LRU_PutRec(LRU_GS_Oinv,  DI, Work)
        End Do
     End Do !*atom
  End do  !* kpnt


  Call FreeBuffer(Proj_G)
  Call FreeBuffer(Psum_G)
  Call freeBuffer(Work)

  Return
End Subroutine




!******************************************************************************
!
! GS_Test_Oinverse - Tests the Oinverse code in G-space
!
!******************************************************************************

Subroutine GS_Test_Oinverse
  Complex, Pointer :: O(:), Oinv(:), V(:)
 !Complex :: dot, dot2, PDOT(PLM_Max)
  Complex :: dot, dot2
  Complex ,allocatable ::  PDOT(:)
  Integer :: i,j, Kpnt,  N
  Real    :: r, dv
Complex, Pointer :: Omat(:,:), Oinvmat(:,:)
Write(Log_Unit,*) 'GS_Test_Oinverse: Start!!!!!!!!!!!!!!!!!!! PLM_Max=',PLM_Max
  write(Log_unit,*) 'GS_Test_Oinvers which no longer works'
  stop
  Allocate(PDOT(PLM_Max))

  N = FFT_Grid(4,G_PROJ)
!  N = Gall_Size(G_LOW)
  Allocate(O(N), Oinv(N), V(N))

Mem_Kpnt = 1
dv = 1.0/FFT_Grid(4,G_PROJ)


!Omat = 0
!Do i=1,n
!   Omat(i,i) = Omat(i,i) + 1
!   Do j=1, n
!      Omat(i,j) = Omat(i,j) + V(i)*CONJG(V(j))
!   End Do
!end Do
!Call RS_Oinverse(V, 1)


  Do Kpnt=1, NumKpnts  
     Write(Log_Unit,*) 'GS_Test_Oinverse: checking Kpnt=',kpnt
     Mem_Kpnt = Kpnt

     Do i=1, N
        V = 0
        V(i) = 1
        PDOT = 0
        Call CalcProjProducts(V, PDOT)
        
        O = V
        Call AccumProj(O, PDOT, OijMatrix)
!Write(Log_Unit,*) 'i=',i, ' * DOT=',DOT_PRODUCT(O,O)
!Do j=1, N
!  dot = O(j) - Omat(j,i)
!  r = CONJG(dot)*dot
!  if (ABS(r) > 1E-9) then
!     Write(Log_Unit,*) 'Omat: ij=',i,j, ' * Oj=',O(j), ' * Oji=',Omat(j,i), ' * ERR=',r
!  End if
!End Do
        Do j=1, i
           Oinv = 0
           Oinv(j) = 1
           Call CalcOinverse(Oinv, Kpnt)
           

           dot = DOT_PRODUCT(O, Oinv)
           
           r = CONJG(dot)*dot
           r = SQRT(r)
           If (i==j) then
              if (ABS(r-1) > 1E-9) then
                 write(Log_Unit,*) 'ij=',i,j,' * dot=',dot
              end If
           else
              if (ABS(r) > 1E-9) then
                 write(Log_Unit,*) 'ij=',i,j,' * dot=',dot
              end If
           End If
        End Do
     End Do
  End do


Write(Log_Unit,*) 'GS_Test: Completed!'
  DeAllocate(O,Oinv,V)
  DeAllocate(PDOT)

!  Call FreeBuffer(O)
!  Call FreeBuffer(Oinv)
!  Call FreeBuffer(V)

  Return
End Subroutine


!******************************************************************************
!
!  Update_Oinverse - Updates the Oinverse vectors when the atoms move
!
!******************************************************************************

Subroutine Update_Oinverse(FD_Base, Mem)
  Integer, Intent(IN) :: FD_Base
  Real,    Intent(IN) :: Mem

  Integer :: MaxRec, Lsize

  MaxRec = Size(BZ%Ku(1,:)) * PLM_Max  

  If (Proj_Mode == PROJ_RS) then
!*     Call DetGroups


     Call RS_CreateOinverse
  else
     LSize = Gall_size(G_LOW)
     Call LRU_InitContext(LRU_GS_Oinv, 'GS_Oinv:', FD_Base, Mem, &
          MaxRec,  LSize, .FALSE.)

     Call GS_CreateOinverse
  End if

  Return
End Subroutine

!******************************************************************************
!
!  InitOinverse - Initializes the Oinverse routines
!
!******************************************************************************


Subroutine InitOinverse(FD_Base, Mem)
  Integer, Intent(IN) :: FD_Base
  Real,    Intent(IN) :: Mem

  Integer :: i,j,MaxRec, LSize(0:0)

!RETURN

  If (Proj_Mode == PROJ_RS) then
     Allocate(RS_Oinv_Scale(PLM_Max, Size(BZ%Ku(1,:))))
     Allocate(RS_Oinv(PLM_Max, Size(BZ%Ku(1,:))))
     Allocate(Group(Specific_Atoms), Oi_Work1(FFT_Grid(4,G_PROJ)))

     RS_Oinv(:,:)%N = 0

     !** Dummy allocs to get things started ***
!*     Allocate(group_nmax(1), RS_Oinv_Scale(1))
  else
     Allocate(GS_Oinv_Scale(PLM_Max, Size(BZ%Ku(1,:))))
  End if

  Call Update_Oinverse(FD_Base, Mem)


!  Call GS_Test_Oinverse
!STOP
  Return
end Subroutine



End Module
