!******************************************************************************
!
! File : orbital_matrix.f90
!   by : Alan Tackett
!   on : 11/09/95
!  for : PAW project
!
! Orbital_Matrix - Initializes the aqlm, avlm, and anlm arrays for the given
!                  fixed atom data.
!
! atom      - Fixed atom data structure to use
! ArraySize - If this variable is set to -1 then no data is stored
!             and the number of nonzero elements are determined and 
!             returned in ArraySize
!
!
! The form of the aqlm and avlm cooeficients are defined below.
!
!            LM        aL
! aqlm(:) = G         N                   (see WN46)
!            l m l m   n l n l
!             i i j j   i i j j
!
!               M L-M      ^ aL
! avlm(:) = (-1) G         v                   (see WN47)
!                 l m l m   n l n l
!                  i i j j   i i j j
!
! Where the G's are Gaunt coefficients and N is the fixed atom density.
! The various l's amd m's are determined by the atomic orbitals defined
! for the specific atom. Capital L is a value in the range of
!
!      |l - l | <= L <= (l + l )
!        i   j            i   j
!
! Based on the symmetry properties of the Gaunt coefficients you know that
!
!    LM   
!   G        != 0    if (l + l + L) is EVEN and M = m - m  and |M| <= L
!    l m l m              i   j                      j   i
!     i i j j
!
!
! For this reason the indices that have to be stored are :
!
!                (n l ) (n l )     m      m     L
!                  i i    j j       i      j
!
! For avlm the m's need to be negated in order to get the proper indices,ie,
! 
! if m = 1 and m = 2 then m = -1 and m = -2 for avlm 
!     i         j          i          j
!
! This helps minimive storage space.
!
! Based on the symmetry of the Gaunt coefficients, n's and v's only
! half of the values need to be stored, li < lj, to get the flipped value
! you just multiply by (-1)^M (I think?? Ask Natalie)
!  
! 
! NOTE : This routine assumes that the V_Hat and Density variables contained
!        in the atom data structure are encoded based on the above idea and
!        that the orbitals are stored in increasing order.
!
!        The LUT must use 4 byte or greater integers for the index encoding!
!        
!******************************************************************************

Subroutine Orbital_Matrix(atom, Array_Size)
  Use atom_data
  Use spherical_harmonic
  Use paw_inout
  Use word
  Use orbital_pack
  Use denvhat_pack

  Implicit None

  Type(Atom_Info_Fixed), Intent(INOUT) :: atom
  Integer,               Intent(INOUT) :: Array_Size

  Integer          :: cnt, Num_Orbitals, denvhat_cnt
  Integer          :: Lmin, Lmax, i, j
  Integer          :: li, lj, mi, mj, L, m
  Integer          :: nili, njlj 
  Integer, Pointer :: L_Value(:), LUT_orb(:), LUT_denvhat(:)
  Real, Pointer    :: Vhat(:), den(:), aqlm(:), avlm(:)
  Real             :: g

!OPEN(29,file='orb.out')
!write(29,*) 'Ni Nj Li Lj Mi Mj  L  M' 
  Vhat => atom%V_Hat
  Den  => atom%Density
  L_Value => Atom%L_Value
  aqlm => Atom%aqlm
  avlm => Atom%avlm
  LUT_orb => Atom%LUT_orb
  LUT_denvhat => Atom%LUT_denvhat

  Num_Orbitals = Atom%Basis_Size

  cnt = 0

  Do denvhat_cnt = 1, Atom%DenVhat_Size
    Call DenVhat_Decode(LUT_denvhat(denvhat_cnt), nili, njlj, L)
    li = L_value(nili);  lj = L_Value(njlj);

    If (nili == njlj) then
       j = 1      !** No mirror image
    Else
       j = 2      !** Indices are different so do it twice
    End If

!write(Log_Unit,*) 'Orbital_Matrix: DenVHat_Cnt=',denvhat_cnt, ' * J=',j, ' nili=',nili, ' * njlj=', njlj
    Do i=1, j          !** Possibly do it twice
       If (i==2) then    !** If 2nd time through then swap nili, njlj
!write(Log_Unit,*) 'Orbital_MAtrix: Swapping nili and njlj'
          m = nili;  nili=njlj;  njlj = m;
          m = Li;    Li = Lj;    Lj = m;
       End If

!write(Log_Unit,*) 'Orbital_Matrix: nili,njlj=',nili,njlj
       Do mi = -li, li
          Do mj = -lj, lj
             M = mj - mi
!write(Log_Unit,*) 'Orbital_Matrix: mi,mj,L,M=',mi,mj,l,m             
             if (abs(M) <= L) then
                cnt = cnt + 1
                
                if (Array_Size > 0) then      !** Check to see if we store it
                   LUT_Orb(cnt) = Orbital_Encode(nili, njlj, mi, mj, L)
                   
                   aqlm(cnt) = Den(denvhat_cnt) * Gaunt(L, M, li, mi, lj, mj)
                   avlm(cnt) = Vhat(denvhat_cnt) * Gaunt(L,M,li, mi, lj, mj)

                   if (Mod(ABS(M),2) == 1) avlm(cnt) = -avlm(cnt)
!write(29, '(8i3,2f)') nili,njlj,li,lj,mi,mj,L,M, avlm(cnt), Gaunt(L,M,li,mi,lj,mj)

!write(Log_Unit,*) 'orbital_Matrix: cnt=',cnt, '* den=',den(denvhat_cnt)
!write(Log_Unit,*) 'orbital_Matrix:   nili,njlj,mi,mj,L=',nili,njlj,mi,mj,l
!Write(Log_Unit,*) 'Orbital_Matrix:   avlm=',avlm(cnt), ' * aqlm=',aqlm(cnt)
                End If
             End If
          End Do
       End Do
    End Do
  End Do

close(29)
  Array_Size = cnt
  Return
End Subroutine


