!******************************************************************************
!
! File : orbital_pack.f90
!   by : Alan Tackett
!   on : 11/08/95
!  for : PAW Method 
!
! Contains 2 routines to encode and decode the orbital matrix indices that are
! used in the atom_data module for the fields aqlm, and avlm.
!
!            LM        aL
! aqlm(:) = G         N                        (see WN46)
!            l m l m   n l n l
!             i i j j   i i j j
!
!               M L-M      ^ aL
! avlm(:) = (-1) G         v                   (see WN47)
!                 l m l m   n l n l
!                  i i j j   i i j j
!
!
! Where the G's are Gaunt coefficients and N is the fixed atom density.
! The various l's amd m's are determined by the atomic orbitals defined
! for the specific atom. Capital L is a value in the range of
!
!      |l - l | <= L <= (l + l )
!        i   j            i   j
!
! Based on the symmetry properties of the Gaunt coefficients you know that
!
!    LM   
!   G        != 0    if (l + l + L) is EVEN and M = m - m
!    l m l m              i   j                      j   i
!     i i j j
!
!
! For this reason the indices that have to be stored are :
!
!                (n l )  (n l )    m      m     L
!                  i i     j j      i      j
!
! Bits Required:   (5)     (5)    (3)    (3)   (3)   = 19 bits total
!
! Each l value is in the range on 0..3 and the corresponding m is in the 
! range of -l <= m <= l for 2*l + 1 possible values.  This gives the range
! for L = 0..6.  The (nl) indices are stored as pairs which correspond
! to the actual basis function index.  Since only 5 bits are used the max
! number of basis functions is 32 for each atom. The l's can be determined
! from the (nl)'s and hence are not stored.  
!
! For avlm the m's need to be negated in order to get the proper indices,ie,
! 
! if m = 1 and m = 2 then m = -1 and m = -2 for avlm 
!     i         j          i          j
!
! This helps minimive storage space.
!
! Also note that since 19 bits are required each integer should 
! be 4 byte integer. The 19 LSB's are used all other bits are set to 0.
!
!    *******  NOTE: No Error checking is done!!!!!! ********
!
!******************************************************************************

Module orbital_pack

 implicit none

!******************************************************************************
Contains
!******************************************************************************


!******************************************************************************
!
!  Orbital_Encode - Encodes the given nili,njlj, li, lj, mi, mj, and L supplied
!                   assuming the above restrictions on l's, m's, and L.
!
!
!  NOTE : NO error checking is done!!!!
!         The return value must be a 4 byte or greater integer!!
!
!******************************************************************************

Integer Function Orbital_Encode(nili, njlj, mi, mj, L)
   Integer, Intent(IN) :: nili
   Integer, Intent(IN) :: njlj
   Integer, Intent(IN) :: mi
   Integer, Intent(IN) :: mj
   Integer, Intent(IN) :: L

   Integer :: N

   N = nili
   N = ISHFT(N, 5) + njlj
   N = ISHFT(N, 3) + mi+3
   N = ISHFT(N, 3) + mj+3
   N = ISHFT(N, 3) + L

   Orbital_Encode = N
   Return
End Function
         
!******************************************************************************
!
!  Orbital_Decode - Decodes the given value(N) into the separate nili, njlj,
!                   li, lj, mi, mj, and L assuming the above restrictions on 
!                   (nl)'s, l's, m's, and L.
!
!
!  NOTE : NO error checking is done!!!!
!         The default integer must be a 4 byte or greater integer!!
!
!******************************************************************************

Subroutine Orbital_Decode(N, nili, njlj, mi, mj, L)
   Integer, Intent(IN)  :: N
   Integer, Intent(OUT) :: nili
   Integer, Intent(OUT) :: njlj
   Integer, Intent(OUT) :: mi
   Integer, Intent(OUT) :: mj
   Integer, Intent(OUT) :: L

   nili = ISHFT(IAND(31*(2**14), N), -14)
   njlj = ISHFT(IAND(31*(2**9), N), -9)
   mi = ISHFT(IAND(7*(2**6), N), -6) - 3
   mj = ISHFT(IAND(7*(2**3), N), -3) - 3
   L  = IAND(7, N) 

   Return
End Subroutine
    

End Module
