!*****************************************************************************
!
! File : paw_inout.f90
!   by : Alan Tackett
!   on : 8/29/95
!  for : PAW method
!
! Contains input/output constants and definitions needed for the MAE method
! along with several routines for handling user I/O.
!
!*****************************************************************************


Module paw_inout

Use strings
Use Units
Use Word

implicit none

              !*** Define Constants that control output level ****
Integer, PARAMETER :: PRINT_TERSE    = 4   !** Print only results
Integer, PARAMETER :: PRINT_COMMANDS = 3   !** Print Commands and results only
Integer, PARAMETER :: PRINT_NORMAL   = 2   !** Normal output
Integer, PARAMETER :: PRINT_VERBOSE  = 1   !** Everything output(usually debug)

Integer :: Print_Level        !** Variable that controls output info


                  !*** Define Reserved output unit numbers ***
!Integer, PARAMETER:: SAVE_HAT_UNIT   =  9 !* Reserved for self-energy dump
                   ! not used
Integer, PARAMETER:: LOG_BASE_UNIT   = 10 !* Reserved log Unit if NOT screen
Integer, PARAMETER:: OUTPUT_BASE_UNIT= 11 !* Reserved output unit if NOT screen
Integer, PARAMETER:: ERROR_BASE_UNIT = 12 !* Reserved error Unit if NOT screen
Integer, PARAMETER:: BAND_IN         = 13 !** Reserved for Band Struct Kpnts
Integer, PARAMETER:: BAND_OUT        = 14 !** Reserved for Band Struct Output
Integer, PARAMETER:: SCRATCH_UNIT    = 15 !* Reserved unit for scratch work
Integer, PARAMETER:: BANDOUT_UNIT       = 16 ! Output for band results
Integer, PARAMETER:: CHARGEOUT_UNIT     = 17 ! Output for charge results
Integer, PARAMETER:: FORCE_UNIT      = 18 ! Output for force results
Integer, PARAMETER:: TOTALCURRENT_UNIT = 19 ! Total current unit
Integer, PARAMETER:: E_UNIT            = 20 ! Electric field unit
Integer, PARAMETER:: POSITIONS_UNIT    = 21 ! Atomic positions and forces unit

              !**** Define Output Units and their filenames ****
Integer :: Log_Unit               !** Log file unit
Integer :: Output_Unit            !** Output file unit for results
Integer :: Error_Unit             !** Unit for error output

Character*200 :: Log_Name         !** Log Filename
Character*200 :: Output_Name      !** Output Filename
Character*200 :: Error_Name       !** Error Filename
Character*200 :: Band_Name        ! Filename for band output
Character*200 :: Charge_Name      ! Filename for charge output
Character*200 :: Sym_Name         ! Filename for symmetry output
Character*200 :: Force_Name       ! Filename for force output
Character*200 :: DOS_Name         ! Filename for DOS output
Character*200 :: PLOT_Name        ! Filename for 2d or 3d output data
Character*200 :: TotalCurrent_Name! Filename for Total current output
Character*200 :: E_Name           ! Filename for electric field
Character*200 :: Pos_Name         ! Filename atom positions and forces
Character*200 :: Best_Name        ! Filename for storing lowest energy results


              !**** Define Input configuration ****
Integer,      PARAMETER :: INPUT_BASE_UNIT = 40   !** Base input unit
Integer,      PARAMETER :: MAX_INCLUDE = 10       !** Max Level of Include file
Character*7,  PARAMETER :: INCLUDE_STR = "INCLUDE"!** Include String
Character*8,  PARAMETER :: DELIMS  = " ()[]{},"   !** System Delimiters
Character*1,  PARAMETER :: COMMENT = "#"          !** Comment Character
Character*1,  PARAMETER :: LIT_CHAR = "'"         !** Literal Character

Integer,      PARAMETER :: MAX_FILE_UNIT = 200

Type(Word_context), TARGET :: paw_WC                 !** Main input context
 
              !**** Define system Units ****
Integer, PARAMETER :: Length_Unit = U_BOHR
Integer, PARAMETER :: Energy_Unit = U_EV

              !**** Define Variables for Output Units *****
Integer :: Length_Output_Unit 
Integer :: Energy_Output_Unit 

              !*** Define File units for the different modules *****
Integer, PARAMETER  :: FD_BASE = 110

Character*(*), PARAMETER :: FILENAME_PSI = "psi.wfn"  !** PSi file

!!!Integer :: XC_Type               !*** Exchange Correlation Type - NOT USED
Logical :: PAW_FirstTime

Interface GetRealsWithUnit
   Module Procedure SNGLE_GetRealsWithUnit
   Module Procedure ARRAY_GetRealsWithUnit
End Interface

Contains

!*****************************************************************************
!
! SNGLE_GetRealswithUnit - Gets a real number from the input with an optional
!                   unit and converts it to the requested unit.
!                   If the next token is not a unit then the token is returned.
!
! Number - Where to store the number
! Output_Unit - Output unit
! token       - Token or next word in the input if not a unit
!
! Return Values
!     If the next 2 tokens comprise a number, unit pair then FALSE is
!     returned otherwise TRUE is returned along with the token.
!
!*****************************************************************************

Logical Function SNGLE_GetRealsWithUnit(wc, Number, Output_Unit, Token, InUnit)
   Type (Word_Context), Intent(INOUT) :: WC   
   Real,          Intent(OUT) :: Number
   Integer,       Intent(IN)  :: Output_Unit
   Character*(*), Intent(OUT) :: Token
   Integer, Optional, Intent(OUT) :: InUnit

   Integer :: j, tlen
   Logical :: ExtraToken

   Call GetNumber(WC, Number)
   Call GetNextWord(WC, token, tlen)
   Call UpperCase(token)

   j = Str2Unit(token)
   if (j /= -1) then
      if (Present(INUnit)) InUnit = j

      ExtraToken = .FALSE.
      Number = Number * ConvertFactor(j, OUTPUT_Unit)
   else 
      ExtraToken = .TRUE.
   End If

   SNGLE_GetRealsWithUnit = ExtraToken
   Return
End Function

!*****************************************************************************
!
! ARRAY_GetRealswithUnit - Gets a real vector from the input with an optional
!                   unit and converts it to the requested unit.
!                   If the next token is not a unit then the token is returned.
!
! Number - Where to store the number
! Output_Unit - Output unit
! token       - Token or next word in the input if not a unit
! VecLen      - Vector Length
!
! Return Values
!     If the next 2 tokens comprise a number, unit pair then FALSE is
!     returned otherwise TRUE is returned along with the token.
!
!*****************************************************************************

Logical Function ARRAY_GetRealsWithUnit(WC, Number, Output_Unit, Token, VecLen)
   Type (Word_Context), Intent(INOUT) :: WC   
   Real,          Intent(OUT) :: Number(:)
   Integer,       Intent(IN)  :: Output_Unit
   Character*(*), Intent(OUT) :: Token
   Integer, Optional, Intent(IN) :: VecLen


   Integer :: j, tlen, k
   Logical :: ExtraToken

   if (PRESENT(VecLen)) then
      k = VecLen
   else
      k = Size(Number)
   End If

   if (k > 1) Call GetNumbers(WC, Number, k-1)

   ExtraToken = SNGLE_GetRealsWithUnit(wc, Number(k), Output_Unit, token, j)
   
   if ((.NOT. ExtraToken) .AND. (k>1)) then
      Number(1:k-1) = Number(1:k-1) * ConvertFactor(j, OUTPUT_Unit)
   End If
 
   ARRAY_GetRealsWithUnit = ExtraToken
   Return
End Function


!*****************************************************************************
!
! EOF_Error - Prints a message telling the user an unexpected EOF occured
!
! IDText - Identification string telling where in the program the error
!          occured.
!
!*****************************************************************************

Subroutine EOF_Error(WC, IDText)
   Use word

   Type (Word_Context), Intent(INOUT) :: WC   
   Character*(*) :: IDText
   Character*100 :: msg

   Write(Error_Unit, *) Trim(IDText), " Unexpected Error encountered!"
   !Call GError(msg)   !not available on the SP2
   Write(Error_Unit, *) Trim(IDText), ' ', Trim(msg)

   Call Word_GetAndPrint(WC, Error_Unit, IDText)

   STOP
   Return
End Subroutine

End Module

