!******************************************************************************
!
! File : projectors.f90
!   by : Alan Tackett
!   on : 04/20/1999
!  for : PW-PAW
!
!  Contains the routines for creating and storing the radial projectors
!  and also routines for perfomring the <p|Psi> calculations
!
!******************************************************************************

Module projectors

Use lrulib
Use gpoints
Use work_mgr
Use atom_data
Use crystal_data
Use spherical_harmonic
Use structfact
Use ylm_fact
Use timing

Implicit NONE!!!!!

Type (REAL_LRU_Context), Pointer :: LRU_RadProj(:)
Type (COMPLEX_LRU_Context), Pointer :: LRU_Proj(:)

Integer  :: PLM_Max, Proj_Max
Integer  :: PDOT_Max
Integer  :: RadProj_Max

Integer, Pointer :: PLM(:,:)
Integer, POINTER :: PLM_AtomRange(:,:)
Integer, Pointer :: PDOT_Map(:,:)

Integer :: PDOT_Dir  !** Previous direction when processing PDOTS

Integer, PRIVATE :: Count !** Counter for g-points list
Integer, PRIVATE :: Glist_Half
Real, Pointer, PRIVATE :: GList(:,:)
Integer, Pointer, Private :: GMap(:)
Real, PRIVATE :: Kvec(3)

!--------------------------------
!Integer, PARAMETER :: G_PROJ = G_HIGH

Type RLE_type
   Integer :: N
   Integer, Pointer :: RLE(:,:)
end Type

Type (RLE_Type), Pointer :: RS_RLE(:)

Integer, Pointer :: RS_PLMMax(:)
Integer, Pointer :: RS_ProjMap(:)
Integer, Pointer :: RS_PLM(:,:,:)
!Integer, Pointer :: FFTMap_Proj(:)
Type (COMPLEX_LRU_Context), ALLOCATABLE :: LRU_RS_Proj(:)
Type (COMPLEX_LRU_Context) :: LRU_RS_dummy
Complex, Pointer :: RS_Work1(:), RS_Work2(:)
Real, Pointer :: RS_Wt(:)


Type RLE_RUN_Type
   Integer :: N
   Integer, Pointer :: Table(:,:)
   Complex, Pointer :: RLE_Data(:)
End Type

Type (RLE_Run_Type), Pointer :: RLE_Run(:)

!--------------------------------
!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
!  FindNext_RLE - Finds the next set of RLE's to process
!
!      pos - Position in Wave function space
!      Table - Merge table
!      next_atom - Next atom to process
!      next      - info about the next merge operation
!
!******************************************************************************

Subroutine FindNext_RLE(pos, Table, Next_Atom, next)
  Integer, Intent(OUT)   :: pos
  Integer, Intent(INOUT) :: Table(:,:)
  Integer, Intent(OUT)   :: Next_atom
  Integer, Intent(OUT)   :: NExt(:)

  Integer :: atom, n, k, next_plm

Write(Error_Unit,*) 'Entered FindNext_RLE which is not debugged'
stop

  Pos = FFT_Grid(4,G_PROJ)+1
  next_atom = -1

  Do atom=1, Specific_Atoms
     If (pos > Table(2,atom)) then
        pos = Table(2,atom)
        next_atom = atom
     End IF
  End Do

  If (next_atom > 0) then
     next = Table(:,next_atom)
     k = Table(1,next_atom)
     Table(1,next_atom) = Table(1,next_atom) + 1
     If (k >= RS_RLE(next_atom)%N) then
        Table(2,next_atom) = FFT_Grid(4,G_PROJ) + 2
     else
        Table(2,next_atom) = RS_RLE(next_atom)%RLE(1,k+1)
        Table(3,next_atom) = Table(3,next_atom) + RS_RLE(next_atom)%RLE(2,k)
     End If
!Write(*,*) 'FindNExt_RLE: Update Table=',Table(:,next_atom)
  End if

  Return
End Subroutine

!******************************************************************************
!
!  Merge_Proj_RLE - Merges the separate projector's for a given k-point
!         If RunSize<=0 then the data is NOT stored and the size of the
!         table and rle_data arrays are calculated and returned
!
!    RLE_Run - Data structure where to store the merged RLE data
!    Kpnt    - K-point index
!    RunSize - Number of RLE runs
!    DataSize- Size of data array
!
!******************************************************************************

Subroutine Merge_Proj_RLE(RLE_Run, Kpnt, RunSize, DataSize)
  Type (RLE_RUN_Type), Intent(OUT)   :: RLE_Run
  Integer,             Intent(IN)    :: Kpnt
  Integer,             Intent(INOUT) :: RunSize
  Integer,             Intent(INOUT) :: DataSize

  Complex, Pointer :: Proj_R(:)
 !Integer :: Table(3, Specific_Atoms), atom, Base_DI, next(3), pos
  Integer ::  atom, Base_DI, next(3), pos
  Integer ,allocatable :: Table(:,:)
  Integer :: PLM_Index, dn, i,j
  Logical :: DoStore
  
  Allocate(Table(3,Specific_Atoms))
Write(Error_Unit,*) 'Entered Merge_Proj_RLE which is not debugged'
stop
  DoStore = .TRUE.
  If (RunSize <= 0) DoStore = .FALSE.

  pos = 0
  RunSize = 1
  DataSize = 0
  Table = 0

  Do atom=1, Specific_Atoms
     Table(1,atom) = 1
     Table(2,atom) = RS_RLE(atom)%RLE(1,1)
     Table(3,atom) = 0
  End Do

  Call FindNext_RLE(pos, Table, atom, next)
  Do While (atom > 0)
!Write(Log_Unit,*) 'Merge_Proj_RLE: Atom=',atom, ' * next=',next, ' * Pos=',pos, ' *RLE%N=',RS_RLE(atom)%N
     Base_DI = RS_PLMMax(atom)*(Kpnt-1)
     
     Do i=1, RS_PLMMax(atom)
        Call LRU_GetRec(LRU_RS_Proj(atom),  Base_DI + i, Proj_R)

        PLM_Index = RS_PLM(atom,4,i)
        dn = RS_RLE(atom)%RLE(2,Next(1))
        If (DoStore) then
           RLE_Run%Table(:,RunSize) = (/PLM_Index, Next(2), dn/)

           j = Next(3)
           RLE_Run%RLE_Data(DataSize+1:DataSize+dn) = Proj_R(j+1:j+dn)
        End If

        DataSize = DataSize + dn
        RunSize = RunSize + 1
     End Do

     Call FindNext_RLE(pos, Table, atom, next)
  End Do

  RunSize = RunSize - 1
  DeAllocate(Table)

  Return
End Subroutine

!******************************************************************************
!
!  RS_MakeGlobal_RLE - Makes the global RLE data structure
!
!******************************************************************************

Subroutine RS_MakeGlobal_RLE
  Integer :: kpnt, RunSize, DataSize

Write(Error_Unit,*) 'Entered RS_MakeGlobal_RLE which is not debugged'
stop

  Do Kpnt=1, NumKpnts
     RunSize = -1
     Call Merge_Proj_RLE(RLE_Run(kpnt), Kpnt, RunSize, DataSize)

     If (RLE_Run(Kpnt)%N > 0) then
        DeAllocate(RLE_Run(kpnt)%Table, RLE_Run(kpnt)%RLE_Data)
     End if

Write(Log_Unit,*) 'RS_MakeGlobal: Kpnt=',Kpnt, ' * RunSize=',RunSize, &
     ' * DataSize=',DataSize

     RLE_Run(kpnt)%N = RunSize
     Allocate(RLE_Run(kpnt)%Table(3,Runsize), RLE_Run(kpnt)%RLE_Data(DataSize))

     Call Merge_Proj_RLE(RLE_Run(kpnt), Kpnt, RunSize, DataSize)
  End Do

  Return
End Subroutine

!******************************************************************************
!
!  MakeIntWts - Makes the integration weights  using Bodes
!          5-point rule.
!
!******************************************************************************

Subroutine MakeIntWts_Bode(Wt)
  Real,         Intent(OUT) :: Wt(:)

  Integer :: x,y,z,k, xw,yw,zw
  Integer :: BodeWt(5)

Write(Error_Unit,*) 'Entered MakeIntWts_Bode which is not debugged'
stop
  BodeWt = (/28, 64, 24, 64, 28/)

  k = 0
  Do z=1, FFT_Grid(3,G_PROJ)
     zw = BodeWt(Mod(z-1,4)+1)
     Do y=1, FFT_Grid(2,G_PROJ)
        yw = BodeWt(Mod(y-1,4)+1)

        Do x=1, FFT_Grid(1,G_PROJ)
           xw = BodeWt(Mod(x-1,4)+1)
           k = k + 1

           Wt(k) = xw*yw*zw
        End Do
     End Do
  End Do

  Wt = Wt/(45.0)**3
!*  Wt = 1.0

  Return
End Subroutine


!******************************************************************************
!
!  MakeIntWts_Simpson - Makes the integration weights 
!         using Simpson's 3-pnt rule
!
!******************************************************************************

Subroutine MakeIntWts_Simpson(Wt)
  Real,         Intent(OUT) :: Wt(:)

  Integer :: x,y,z,k, xw,yw,zw

Write(Error_Unit,*) 'Entered MakeIntWts_Simpson which is not debugged'
stop
  k = 0
  Do z=1, FFT_Grid(3,G_PROJ)
     If (Mod(z,2)==1) then
        zw = 2
     else
        zw = 4
     End If
     Do y=1, FFT_Grid(2,G_PROJ)
        If (Mod(y,2)==1) then
           yw = 2
        else
           yw = 4
        End If

        Do x=1, FFT_Grid(1,G_PROJ)
           k = k + 1
           
           If (Mod(x,2)==1) then
              xw = 2
           else
              xw = 4
           End If

           Wt(k) = xw*yw*zw
        End Do
     End Do
  End Do

  Wt = Wt/27.0
!*  Wt = 1.0

  Return
End Subroutine


!******************************************************************************
!
!  MakeIntWeights - Makes the integration weights
!
!******************************************************************************

Subroutine MakeIntWeights(Wt)
  Real,     Intent(OUT) :: Wt(:)

  Integer :: i,j,k
  Logical :: DoBode

Write(Error_Unit,*) 'Entered MakeIntWts which is not debugged'
stop
  DoBode = .TRUE. 
  i = FFT_Grid(1,G_PROJ)
  j = FFT_Grid(2,G_PROJ)
  k = FFT_Grid(3,G_PROJ)

  If ((Mod(i,4)==0) .AND. (Mod(j,4)==0) .AND. (Mod(k,4)==0) .AND. DoBode) then 
Write(Log_Unit,*) 'MakeIntWeights: BODE'
     Call MakeIntWts_BODE(Wt)
  else
Write(Log_Unit,*) 'MakeIntWeights: SIMPSON'
     Call MakeIntWts_Simpson(Wt)
  End If


  Return
End Subroutine




!**********************************************************************
!
! RS_CreateProjMap - Creates the Projector Maps
!
!**********************************************************************

subroutine RS_CreateProjMap(atom, Map, R_xtal, rle_size, data_size)
  Integer, Intent(IN) :: atom
  Integer, Intent(OUT) :: Map(:)
  Real,    Intent(IN)  :: R_xtal(:,:)
  Integer, Intent(OUT) :: rle_size
  Integer, Intent(OUT) :: data_size

  !Rick Matthews, 2/9/97
  !Create the projector map
  !
  !(See module declarations for explanations of these.)
  
  integer :: ix,iy,iz  !Integer coordinates on curvilinear grid.
  !Range from 1 to Grid(1), Grid(2),Grid(3)
  integer :: PM(3)     !Vector with elements ix, iy, iz
  real    :: PC(3)     !Actual curvilinear coords corresponding to PM
  real    :: R(3)      !Real space equivalent of PM, PC.
  real    :: Offset(3)
  Integer :: t, x,y,z
  Real    :: Rc2       ! Rc^2
  Integer :: Offset_xtal
  integer :: i         !Index of specific atoms
  integer :: k         !=1,2,3 for curv coord x,y,z
  integer :: L,m       !Local l and m of projector.
  Type(Atom_Info_Fixed),Pointer :: At_info !Structure on atom type info
  Type(Specific_Atom),  Pointer :: AT      !Structure on specific atom info
  Real :: DR(3), mag, rsmall
  Character*100 :: msg
  Integer :: npnts, Run_Start
  Logical :: Ok, InRun

Write(Error_Unit,*) 'Entered RS_CreateProjMap which is not debugged'
stop


  Map = 0

  At=>Atom_List(atom)                 !This specific atom in atom list
  AT_Info=>AtomType_Info(At%TypeIndex)!AtomTypeInfo for the atom type
  Rc2 = (AT_Info%Rc_RS_scale*AT_Info%Rc)**2

  rle_size = 0
  data_size = 0
  npnts = 0
  t = 0
  Inrun = .FALSE.
  Do z=1, FFT_Grid(3, G_PROJ)
     Do y=1, FFT_Grid(2, G_PROJ)
        Do x=1, FFT_Grid(1, G_PROJ)
           t = t + 1
           
           PM = (/x,y,z/)
           R = MATMUL(xtal%Basis, (PM-1.0)/(FFT_grid(1:3,G_PROJ)))

!Write(Log_Unit,*) ' xyz=',x,y,z
!Write(Log_Unit,*) '  R=',R


           offset_xtal = 0
           Ok = .FALSE.
           Do While ((.NOT. Ok) .AND. (offset_xtal<27))
              Offset_xtal = offset_xtal + 1

              DR = R - R_xtal(:,offset_xtal) - AT%Pos
              mag = DOT_PRODUCT(DR,DR)
              If (mag <= Rc2) Ok = .TRUE.
           end Do

           If (Ok) then
              If (.NOT. InRun) then
                 Run_Start = t
                 InRun = .TRUE.
              End If

              Map(t) = offset_xtal
              npnts = npnts + 1
!Write(Log_Unit,*) 'RS_CPM: t=',t, ' * n=',npnts, ' * RLE=',rle_size
           else If (Inrun) then
              data_size = data_size + t-Run_Start
              rle_size = rle_size + 1
              InRun = .FALSE.
!Write(Log_Unit,*) 'RS_CPM: Start=',run_Start, ' * n=',data_size, ' * RLE=',rle_size
           End If
        End Do !*x
     End Do !*y
  End Do !*z

  If (InRun) then
     data_size = data_size + t-Run_Start + 1
     rle_size = rle_size + 1
     InRun = .FALSE.
!Write(Log_Unit,*) 'RS_CPM: Start=',run_Start, ' * n=',data_size, ' * RLE=',rle_size
  end If
  Write(Log_Unit,*) 'RS_CreateProjMap: Total points:',npnts, ' RLE Size:',rle_size
  
  Return
end subroutine


!****************************************************************
!
!  RS_Find_PLMMAX - Calculates the RS_PLMMax constant
!
!****************************************************************

Subroutine RS_Find_PLMMax
  !Rick Matthews, 2/9/97
  !Find Proj_Max,the maximum number of projector functions of any atom
  !and L_Max, the max L_Value of any atom
  !There are Basis_Size radial projector functions.
  !For each radial projector, the L_value(projector number) tells the 
  !    corresponding quantum number L.
  !For each L, their are 2L+1 projectors (one for each quantum number m).
  
  integer :: atom,atype           !Which atom type
  integer :: projector            !which radial projector
  integer :: sum                  !Count of proj functions for each atom
  Type(Atom_Info_fixed),pointer:: AT !Shortcut or alias to AtomType_Info(atom)

  do atom=1, Specific_Atoms
     atype = Atom_List(atom)%TypeIndex
     AT=>AtomType_Info(atype)
     sum=0
     do projector=1,AT%Basis_Size
        sum= sum+2*AT%L_Value(projector)+1
     end do
     
     RS_PLMMax(atom) = sum
  end do
  
  atom = MAXVAL(RS_PLMMax)
  Allocate(RS_PLM(Specific_Atoms, 5, atom))
  Write(Log_Unit,*) 'RS_Find_PLMmax: RS_PLMMax = ', RS_PLMMax

  Return
end subroutine


!******************************************************************************
!
!  RS_FindPLM - Creates the RS PLM array.
!
!******************************************************************************

Subroutine RS_FindPLM 
  !Rick Matthews,created 2/11/97
  !Finds PLM(atom type, quantum number, grid index)
  
  integer :: atom      !atom index
  integer :: rad_proj  !index of radial projector
  integer :: L         !quantum number L
  integer :: m         !quantum number m
  integer :: i         !projector grid index
  Integer :: j, err, V(4), DV(4), RadP
  Type (Atom_Info_Fixed), Pointer :: AT_Info
  
  RS_PLM = 0
  
  do atom=1,Specific_Atoms
     i=0
     j = Atom_List(atom)%TypeIndex
     AT_Info => AtomType_Info(j)

     do rad_proj=1, AT_Info%Basis_Size
        L=AT_Info%L_Value(rad_proj)
        do m= -L,L 
           i = i + 1
           RS_PLM(atom,1,i) = Rad_Proj
           RS_PLM(atom,2,i) = L
           RS_PLM(atom,3,i) = m
           
           V = (/atom,Rad_Proj,L,m/)
           j=1
           DV = V - PLM(:,j)
           err = DOT_PRODUCT(DV,DV)
           Do while (err>0.5)
              j = j + 1
              DV = V - PLM(:,j)
              err = DOT_PRODUCT(DV,DV)
           End Do

           RS_PLM(atom,4,i) = j   !** PLM Index
           RS_PLM(atom,5,i) = 1   !atom is present in this grid
        end do
     end do
     
!**     Proj_AtomMax(atom) = i
  end do
  
!  Write(Log_Unit, *) 'RS_FindPLM: Proj_Max=', RS_PLMMax
!  Do atom=1, Specific_Atoms
!     Do i=1, rS_PLMMax(atom)
!        Write(Log_Unit, *) 'RS_FindPLM:    Atom=',Atom, &
!             ' * NLM=',RS_PLM(atom,1:3,i)
!     End Do
!  End Do
  
End Subroutine


!**********************************************************************
!
! RS_StoreProj - Stores the Projector for the atom
!
!**********************************************************************

subroutine RS_StoreProj(atom, Map, R_xtal, RLE, Proj_G, Proj_R)
  Integer, Intent(IN) :: atom
  Integer, Intent(IN) :: Map(:)
  Real,    Intent(IN)  :: R_xtal(:,:)
  Integer, Intent(OUT) :: RLE(:,:)
  Complex, Intent(IN)  :: Proj_G(:)
  Complex, Intent(OUT) :: Proj_R(:)

  !Rick Matthews, 2/9/97
  !Create the projector map
  !
  !(See module declarations for explanations of these.)
  
  integer :: ix,iy,iz  !Integer coordinates on curvilinear grid.
  !Range from 1 to Grid(1), Grid(2),Grid(3)
  Integer :: t, x,y,z
  integer :: i         !Index of specific atoms
  integer :: k         !=1,2,3 for curv coord x,y,z
  integer :: L,m       !Local l and m of projector.
  Type(Atom_Info_Fixed),Pointer :: At_info !Structure on atom type info
  Type(Specific_Atom),  Pointer :: AT      !Structure on specific atom info
  Integer :: npnts, run_Start
  Logical :: Ok, InRun
  Integer :: data_size, rle_size


Write(Error_Unit,*) 'Entered RS_StoreProj which is not debugged'
stop
  At=>Atom_List(atom)                 !This specific atom in atom list
  AT_Info=>AtomType_Info(At%TypeIndex)!AtomTypeInfo for the atom type

  rle_size = 0
  RLE = 0
  data_size = 0
  npnts = 0
  t = 0
  Inrun = .FALSE.
  Do z=1, FFT_Grid(3, G_PROJ)
     Do y=1, FFT_Grid(2, G_PROJ)
        Do x=1, FFT_Grid(1, G_PROJ)
           t = t + 1

           If (Map(t) > 0) then
              If (.NOT. InRun) then
                 Run_Start = t
                 InRun = .TRUE.
              End If

              npnts = npnts + 1
!Write(Log_Unit,*) 'RS_SPM: t=',t, ' * n=',npnts, ' * RLE=',rle_size
              Proj_R(npnts) = Proj_G(t)
           else If (Inrun) then
              rle_size = rle_size + 1
              RLE(:,rle_size) = (/Run_start, (t-Run_Start) /)
              InRun = .FALSE.
!Write(Log_Unit,*) 'RS_SPM: Start=',run_Start, ' * n=',npnts, ' * RLE=',rle_size
           End If
        End Do !*x
     End Do !*y
  End Do !*z

  If (InRun) then
     rle_size = rle_size + 1
     RLE(:,rle_size) = (/Run_start, (t-Run_Start+1) /)
!Write(Log_Unit,*) 'RS_SPM: Start=',run_Start, ' * n=',data_size, ' * RLE=',rle_size
  end If
  Write(Log_Unit,*) 'RS_CreateProjMap: Total points:',npnts, ' RLE Size:',rle_size
  
  Return
end Subroutine


!******************************************************************************
!
!  RS_CreateProj - Creates the Real-space projectors
!
!******************************************************************************

Subroutine RS_CreateProj

  Integer :: i,j,k,atom,  m,am, L, n, kpnt, t, G_Size, G_Half
  Integer :: Base_RAD, Base_DI, G_Index, MaxRec, RS_Size, DI, RadP
  Integer :: Basis_size, NG, atype
  Integer, Pointer :: Map(:)
  Real, Pointer :: Proj(:), Wt(:)
  Complex, Pointer :: P(:), Work(:), Ylm(:), Phase(:), Proj_G(:), Proj_R(:)
  Complex, Pointer :: Buffer(:)
  Complex :: c1,c2,c3, c4
  Real :: DV, R(3), R_xtal(3,27)
Integer :: x,y,z,xo,yo,zo, rle_size, data_size
Real :: r1,r2

Write(Error_Unit,*) 'Entered RS_CreateProj which is not debugged'
stop
  DV = Xtal%Volume/(1.0*FFT_Grid(4,G_PROJ))

  Allocate(P(FFT_Grid(4,G_PROJ)), buffer(FFT_Grid(4,G_PROJ)), &
       Map(FFT_Grid(4,G_PROJ)))

  Call Getbuffer( Proj_G)

  c1 = 4*Pi / sqrt(xtal%Volume)


!**  FD_Proj = 60

  i = 0
  Do z=-1, 1
     Do y= -1, 1
        Do x=-1, 1
           i = i + 1
           R_xtal(:,i) = MatMul(xtal%Basis, (/x,y,z/))
!Write(Log_Unit,*) 'RS_CreateProjMap: i=',i, '* DR=',R_xtal(:,i)
        End Do
     End Do
  End Do
  R_xtal(:,14) = R_xtal(:,1);  R_xtal(:,1) = 0  !** Move R=0 to front for speed


  NG = Gpnt_size(G_LOW) - 1

  Do atom=1, Specific_Atoms
     atype = Atom_List(atom)%typeIndex
     Basis_Size = AtomType_Info(atype)%Basis_Size
     Call RS_CreateProjMap(atom, Map, R_xtal, rle_size, data_size)

     G_Half = Atomtype_Info(Atom_List(atom)%typeIndex)%Gpnt_size
     G_Size = 2*G_Half - 1

     Call GetStructFactor( atom, Phase)

Write(Log_Unit,*) 'RS_CreateProj: atom=',atom, ' * RLE%N=',rle_size, ' * Data=',Data_Size
     RS_RLE(atom)%N = RLE_Size
     Allocate(Proj_R(data_size), RS_RLE(atom)%RLE(2,RLE_Size), &
          Work(G_Size))

     MaxRec = Size(BZ%Ku(1,:)) * RS_PLMMax(atom)
     RS_Size = Data_Size

     i = FD_RS_Proj+atom-1; R(1) = 50
     Call LRU_InitContext(LRU_RS_Proj(atom), 'RS_Proj:', i, 50.0, &
          MaxRec,  RS_Size, .TRUE.)
     

!-----
     DI = 0
     Do Kpnt = 1, Size(BZ%Ku(1,:))
        Base_DI = Basis_Size * (Kpnt-1)  !** Calc starting DI for Kpnt

        Do n=1, RS_PLMMax(atom)
           RadP = RS_PLM(atom,1,n)
           L = RS_PLM(atom,2,n)
           m = RS_PLM(atom,3,n)
           am = ABS(m)
           
           Call LRU_GetRec(LRU_RadProj(atype),  Base_DI + radP, Proj)
           Call GetYlm( Kpnt, L, am, Ylm)        
           
           c2 = CMPLX(0,1)**L

           Proj_G = 0
           If (m>=0) then
              i = 1
              Proj_G(i) = CONJG(Ylm(i))*Proj(i)*Phase(i)
              
              Do i=2, G_Half
                 Proj_G(i) = CONJG(Ylm(i))*Proj(i)*Phase(i)
                 Proj_G(i+ng) = CONJG(Ylm(i+ng)) * Proj(i+G_Half-1) * &
                      Phase(i+ng)
              end Do
              Proj_G = c1*c2*Proj_G
           Else             
              i = 1
              Proj_G(i) = Ylm(i)*Proj(i)*Phase(i)
              
              Do i=2, G_Half
                 Proj_G(i) = Ylm(i)*Proj(i)*Phase(i)
                 Proj_G(i+ng) = Ylm(i+ng) * Proj(i+G_Half-1) * &
                      Phase(i+ng)
              end Do
              
              c3 = (-1)**am
              Proj_G = c1*c2*c3*Proj_G
           End If

           Proj_G = CONJG(Proj_G)
!Write(Log_Unit,*) 'RS_CreateProj: nlm=',n,l,m,' * a=',atom, ' DOT=',DOT_PRODUCT(Proj_G, Proj_G)

           Call RS_toR(Proj_G, Buffer)

           Proj_R = 0
           Call RS_StoreProj(atom, Map, R_xtal, RS_RLE(atom)%RLE, &
                Buffer, Proj_R)
           DI = DI + 1
           Call LRU_PutRec(LRU_RS_Proj(atom), DI, Proj_R)

P = 0           
Where (Map > 0) P = Buffer
           
c4=DOT_PRODUCT(Buffer, buffer)
!c4 = SQRT(REAL(c4))
Where (Map > 0) Buffer = 0
c3=DOT_PRODUCT(Buffer, buffer)
!c3 = SQRT(REAL(c3))
Buffer = 0
where(Map > 0) Buffer = P
c2=DOT_PRODUCT(Buffer, Buffer)
!c2 = SQRT(REAL(c2))

c3=SQRT(REAL(c3/c4))
c2 = SQRT(REAL(c2/c4))
Write(Log_Unit,*) 'RS_CreateProj:2 nlm=',n,l,m,' * a=',atom
Write(Log_Unit,*) '   INSIDE==',REAL(C2), ' * OUTSIDE=',REAL(C3)

Write(LOG_Unit,*) '  DOT=',DOT_PRODUCT(Proj_R, Proj_R)
        End Do
     End Do

     DeAllocate(Work, Proj_R)
  End do  !*atom


  DeAllocate(P, Buffer, Map)
  Call FreeBuffer(Proj_G)

  Return
End Subroutine

!******************************************************************************
!
!  RS_toR - transforms the function to RS
!
!  Fg - function in G-space
!  Fr - Function in real space
!
!******************************************************************************

Subroutine RS_toR(Fg, Fr)
  Complex, Intent(IN)  :: Fg(:)
  Complex, Intent(OUT) :: Fr(:)

  Integer :: j,k,t
Write(Error_Unit,*) 'Entered RS_toR which is not debugged'
stop

  Fr = 0
  Fr(FFTMap_PROJ(1:Gpnt_Size(G_Low))) = Fr(FFTMap_PROJ(1:Gpnt_Size(G_Low))) + &
       Fg(1:Gpnt_Size(G_Low))
  t = Gpnt_Size(G_Low)+1
  j = Gpnt_Size(G_PROJ)+1;   k = j + (Gpnt_Size(G_Low)-1) - 1
  Fr(FFTMap_Proj(j:k)) = Fr(FFTMap_Proj(j:k)) + Fg(t:)
  
  Call PerformFFT(FFT_TO_R, G_PROJ, Fr)

  Return
end Subroutine


!******************************************************************************
!
!  RS_toG - transforms the function from RS to 1-d G-space
!
!  Fr - Function in real space
!  Fg - function in G-space - accumulated!
!
!******************************************************************************

Subroutine RS_toG(Fr, Fg)
  Complex, Intent(INOUT) :: Fr(:)
  Complex, Intent(OUT)   :: Fg(:)

  Integer :: j,k,t
Write(Error_Unit,*) 'Entered RS_toG which is not debugged'
stop

  Call PerformFFT(FFT_TO_G, G_PROJ, Fr)

  Fg(1:Gpnt_Size(G_Low)) = Fg(1:Gpnt_Size(G_Low)) + &
       Fr(FFTMap_PROJ(1:Gpnt_Size(G_Low))) 
  t = Gpnt_Size(G_Low)+1
  j = Gpnt_Size(G_PROJ)+1;   k = j + (Gpnt_Size(G_Low)-1) - 1
  Fg(t:) = Fg(t:) + Fr(FFTMap_Proj(j:k))
  
  Return
end Subroutine


!******************************************************************************
!
!  CalcProjProducts_RS - Calcuylates the PDOT's in real space
!
!   Psi  - Wave function in R-SPACE
!   PDOT - Output array containing the <p|Psi> products
!
!*RS*
!******************************************************************************

Subroutine CalcProjProducts_RS(psi, PDOT)
  Complex, Intent(IN)  :: Psi(:)
  Complex, Intent(OUT) :: PDot(:)

  Integer :: i,j,k,dn, n, PLM_Index
  Integer, Pointer :: Table(:,:)
  Complex, Pointer :: Proj_R(:)
  Real :: DV
Write(Error_Unit,*) 'Entered CalcProjProducts_RS which is not debugged'
stop

  DV = 1.0/(1.0*FFT_Grid(4,G_PROJ))

  Proj_R => RLE_Run(Mem_Kpnt)%RLE_Data
  Table => RLE_Run(Mem_Kpnt)%Table

  PDOT = 0
  i = 0
  Do n=1, RLE_Run(Mem_Kpnt)%N
     PLM_Index = Table(1,n)
     j = Table(2,n) - 1
     dn = Table(3,n)
     Pdot(PLM_Index) = PDOT(PLM_Index) + &
          SUM(CONJG(Proj_R(i+1:i+dn)) * Psi(j+1:j+dn))

     i = i + dn
  End do

  PDOT = PDOT * DV

!Do i=1, PLM_Max
!  Write(Log_Unit,*) ' CPP_RS: i=',i, ' * PDOT=',PDOT(i)
!End Do

  Return
end Subroutine


!******************************************************************************
!
!  CalcProjProducts_RS - Calcuylates the PDOT's in real space
!
!   Psi  - Wave function in R-SPACE
!   PDOT - Output array containing the <p|Psi> products
!
!*RS*OLD
!******************************************************************************

Subroutine CalcProjProducts_RS_OLD(psi, PDOT)
  Complex, Intent(IN)  :: Psi(:)
  Complex, Intent(OUT) :: PDot(:)

  Integer :: i,j,k,dn, Base_DI,  atom, n
  Complex, Pointer :: Proj_R(:)
  Complex :: dot
  Real :: DV
  Real, Pointer :: Wt(:)

Write(Error_Unit,*) 'Entered CalcProjProducts_RS_OLD which is not debugged'
stop
!*  DV = Xtal%Volume/(1.0*FFT_Grid(4,G_PROJ))
DV = 1.0/(1.0*FFT_Grid(4,G_PROJ))
!DV = 1


!*  Allocate(Wt(FFT_Grid(4,G_PROJ)))
!*  Call MakeIntWeights(Wt)
!*Wt = 1

!Write(Log_Unit,*) 'CPP_RS: Base_DI=',Base_DI, ' * Kpnt=',Mem_Kpnt


!**FIXME
!DV=1
!RS_Work1 = Psi
!****

  Do atom=1, Specific_Atoms
     Base_DI = RS_PLMMax(atom)*(Mem_Kpnt-1)  !** Calc starting DI for Kpnt

     Do n=1, RS_PLMMax(atom)
        Call LRU_GetRec(LRU_RS_Proj(atom),  Base_DI + n, Proj_R)
        i = 0;  j = 0;
        dot = 0
        Do k=1, RS_RLE(atom)%N
           j = RS_RLE(atom)%RLE(1,k)-1
           dn = RS_RLE(atom)%RLE(2,k)
!write(*,*) 'atom=',atom, ' * i,j,dn=',i,j,dn, ' * dot=',dot
!Write(*,*) '  Proj=',Proj_R(i+1:i+dn)
!Write(*,*) '  Psi =',RS_Work1(j+1:j+dn)
!Write(*,*) '  Wt  =',RS_Wt(j+1:j+dn)

           dot = dot + SUM(CONJG(Proj_R(i+1:i+dn)) * Psi(j+1:j+dn))
           
           i = i + dn
        End Do

        PDOT(RS_PLM(atom,4,n)) = dot*DV
     End Do
  End Do


!Do i=1, PLM_Max
!  Write(Log_Unit,*) ' CPP_RS: i=',i, ' * PDOT=',PDOT(i)
!End Do

  Return
end Subroutine

!******************************************************************************
!
!  AccumBothProj_RS - Accumulates the summed |pi>Dij<pj|Psi> and
!                       and |pi>Oij<pj|Psi> to the given vectors
!                       using a real space representation
!
!     Hx - Output vector for Dij sum. NOTE: The vector is not initialized to 0!
!     Ox - Output vector for Oij sum. NOTE: The vector is not initialized to 0!
!     PDOT - List of <p|Psi> products
!     Dij  - External function for calculaint matrix elements
!     Oij  - External function for calculaint matrix elements
!*RS*
!******************************************************************************

Subroutine AccumBothProj_RS(Hx, Ox, PDot, Dij, Oij)
  Complex, Intent(INOUT) :: Hx(:)
  Complex, Intent(INOUT) :: Ox(:)
  Complex, Intent(IN)    :: PDOT(:)
  Complex, EXTERNAL      :: Dij
  Complex, EXTERNAL      :: Oij

 !Complex :: Ch(PLM_Max), Co(PLM_Max)
  Complex ,allocatable :: Ch(:), Co(:)
  Integer :: atom,i, j, k, dn, Base_DI,  PLM_Index, n
  Integer, Pointer :: Table(:,:)
  Complex, Pointer :: Proj_R(:)
  Complex :: c1, c2


  Allocate(Ch(PLM_Max), Co(PLM_Max))
Write(Error_Unit,*) 'Entered AccumBothProj_RS which is not debugged'
stop

  Call CalcAccumVector(Ch, PDOT, Dij)
  Call CalcAccumVector(Co, PDOT, Oij)

  Proj_R => RLE_Run(Mem_Kpnt)%RLE_Data
  Table => RLE_Run(Mem_Kpnt)%Table

  i = 0
  Do n=1, RLE_Run(Mem_Kpnt)%N
     PLM_Index = Table(1,n)
     j = Table(2,n) - 1
     dn = Table(3,n)
     Hx(j+1:j+dn) = Hx(j+1:j+dn) + Proj_R(i+1:i+dn)*Ch(PLM_Index)
     Ox(j+1:j+dn) = Ox(j+1:j+dn) + Proj_R(i+1:i+dn)*Co(PLM_Index)

     i = i + dn
  End do
  DeAllocate(Ch, Co)

  Return
End Subroutine


!******************************************************************************
!
!  AccumBothProj_RS - Accumulates the summed |pi>Dij<pj|Psi> and
!                       and |pi>Oij<pj|Psi> to the given vectors
!                       using a real space representation
!
!     Hx - Output vector for Dij sum. NOTE: The vector is not initialized to 0!
!     Ox - Output vector for Oij sum. NOTE: The vector is not initialized to 0!
!     PDOT - List of <p|Psi> products
!     Dij  - External function for calculaint matrix elements
!     Oij  - External function for calculaint matrix elements
!*RS*OLD
!******************************************************************************

Subroutine AccumBothProj_RS_OLD(Hx, Ox, PDot, Dij, Oij)
  Complex, Intent(INOUT) :: Hx(:)
  Complex, Intent(INOUT) :: Ox(:)
  Complex, Intent(IN)    :: PDOT(:)
  Complex, EXTERNAL      :: Dij
  Complex, EXTERNAL      :: Oij

 !Complex :: Ch(PLM_Max), Co(PLM_Max)
  Complex ,allocatable :: Ch(:), Co(:)
  Integer :: atom,i, j, k, dn, Base_DI,  PLM_Index, n
  Complex, Pointer :: Proj(:)
  Complex :: c1, c2


  Allocate( Ch(PLM_Max), Co(PLM_Max))
Write(Error_Unit,*) 'Entered AccumBothProj_RS_OLD which is not debugged'
stop

  Call CalcAccumVector(Ch, PDOT, Dij)
  Call CalcAccumVector(Co, PDOT, Oij)

  Do atom=1, Specific_Atoms
     Base_DI = RS_PLMMax(atom)*(Mem_Kpnt-1)  !** Calc starting DI for Kpnt

     Do n=1, RS_PLMMax(atom)
        Call LRU_GetRec(LRU_RS_Proj(atom),  Base_DI + n, Proj)
        PLM_Index = RS_PLM(atom,4,n)
        i = 0;  j = 0;
        Do k=1, RS_RLE(atom)%N
           j = RS_RLE(atom)%RLE(1,k)-1
           dn = RS_RLE(atom)%RLE(2,k)
           Hx(j+1:j+dn) = Hx(j+1:j+dn) + Proj(i+1:i+dn)*Ch(PLM_Index)
           Ox(j+1:j+dn) = Ox(j+1:j+dn) + Proj(i+1:i+dn)*Co(PLM_Index)
           
           i = i + dn
        End Do
     End Do
  End Do
  DeAllocate( Ch, Co)

  Return
End Subroutine

!******************************************************************************
!
!  AccumProj_RS - Accumulates the summed |pi>Dij<pj|Psi> and
!                       and |pi>Oij<pj|Psi> to the given vectors
!                       using a real space representation
!
!     Hx - Output vector for Dij sum. NOTE: The vector is not initialized to 0!
!     PDOT - List of <p|Psi> products
!     Dij  - External function for calculaint matrix elements
!*RS*
!******************************************************************************

Subroutine AccumProj_RS(Hx, PDot, Dij)
  Complex, Intent(INOUT) :: Hx(:)
  Complex, Intent(IN)    :: PDOT(:)
  Complex, EXTERNAL      :: Dij

 !Complex :: Ch(PLM_Max)
  Complex ,allocatable:: Ch(:)
  Integer :: i, j, k, Base_DI,  atom, PLM_Index, n, dn
  Complex, Pointer :: Proj_R(:)
  Integer, Pointer :: Table(:,:)
  Complex :: c1, c2


  Allocate( Ch(PLM_Max))
Write(Error_Unit,*) 'Entered AccumProj_RS which is not debugged'
stop
!RETURN

  Call CalcAccumVector(Ch, PDOT, Dij)

  Proj_R => RLE_Run(Mem_Kpnt)%RLE_Data
  Table => RLE_Run(Mem_Kpnt)%Table

  i = 0
  Do n=1, RLE_Run(Mem_Kpnt)%N
     PLM_Index = Table(1,n)
     j = Table(2,n) - 1
     dn = Table(3,n)
     Hx(j+1:j+dn) = Hx(j+1:j+dn) + Proj_R(i+1:i+dn)*Ch(PLM_Index)

     i = i + dn
  End do
  DeAllocate( Ch)

  Return
End Subroutine

!******************************************************************************
!
!  AccumProj_RS - Accumulates the summed |pi>Dij<pj|Psi> and
!                       and |pi>Oij<pj|Psi> to the given vectors
!                       using a real space representation
!
!     Hx - Output vector for Dij sum. NOTE: The vector is not initialized to 0!
!     PDOT - List of <p|Psi> products
!     Dij  - External function for calculaint matrix elements
!*RS*OLD
!******************************************************************************

Subroutine AccumProj_RS_OLD(Hx, PDot, Dij)
  Complex, Intent(INOUT) :: Hx(:)
  Complex, Intent(IN)    :: PDOT(:)
  Complex, EXTERNAL      :: Dij

 !Complex :: Ch(PLM_Max)
  Complex ,allocatable:: Ch(:)
  Integer :: i, j, k, Base_DI,  atom, PLM_Index, n, dn
  Complex, Pointer :: Proj(:)
  Complex :: c1, c2


  Allocate(Ch(PLM_Max))
Write(Error_Unit,*) 'Entered AccumProj_RS_OLD which is not debugged'
stop
!RETURN

  Call CalcAccumVector(Ch, PDOT, Dij)

  Do atom=1, Specific_Atoms
     Base_DI = RS_PLMMax(atom)*(Mem_Kpnt-1)  !** Calc starting DI for Kpnt

     Do n=1, RS_PLMMax(atom)
        Call LRU_GetRec(LRU_RS_Proj(atom),  Base_DI + n, Proj)
        PLM_Index = RS_PLM(atom,4,n)
        i = 0;  j = 0;
        Do k=1, RS_RLE(atom)%N
           j = RS_RLE(atom)%RLE(1,k)-1
           dn = RS_RLE(atom)%RLE(2,k)
           Hx(j+1:j+dn) = Hx(j+1:j+dn) + Proj(i+1:i+dn)*Ch(PLM_Index)
           
           i = i + dn
        End Do
     End Do
  End Do
  DeAllocate(Ch)

  Return
End Subroutine



!-------------------------------------------------------------------------
!-------------------------------------------------------------------------

!******************************************************************************
!
!  Create_PLM - Creates the PLM map
!
!    DoStore - Determines if info is stored in PLM or just calculate PLM_Max
!
!******************************************************************************

Subroutine Create_PLM(DoStore)
  Logical, Intent(IN) :: DoStore

  Integer :: atype, atom , n, l, m, lut
  Type (Atom_Info_Fixed), Pointer :: AT

  PLM_Max = 0
  
  Do atype = 1, Atom_types
     AT => AtomType_Info(atype)

     Do lut = AtomType_Range(1,atype), AtomType_Range(2,atype)
        atom = AtomType_MAP(lut)

        If (DoStore) PLM_AtomRange(1,atom) = PLM_Max + 1
        Do n=1, AT%Basis_Size
           L = AT%L_Value(n)
           Do m=-L, L
              PLM_Max = PLM_Max + 1
              If (DoStore) PLM(:,PLM_Max) = (/atom, n, L, m/)
           End Do
        End Do
        If (DoStore) PLM_AtomRange(2,atom) = PLM_Max
     End Do
  End Do
  
  Return
End Subroutine


!******************************************************************************
!
!  Create_PDOT_Map - Creates the Process Map for doing PDOT's
!
!  DoStore - Determines if the array is filled or just it's size is calculated
!
!******************************************************************************

Subroutine Create_PDOT_Map(DoStore)
  Logical, Intent(IN) :: DoStore
  Integer :: atype, atom , n, l, m, i, j, RadP, err, Diff(4), lut
  Type (Atom_Info_Fixed), Pointer :: AT

  i = 0
  Do atype = 1, Atom_types
     AT => AtomType_Info(atype)

     Do n=1, AT%Basis_Size
        L = AT%L_Value(n)
        Do m=-L, L
           Do lut = AtomType_Range(1,atype), AtomType_Range(2,atype)
              atom = AtomType_MAP(lut)

              i = i + 1

              If (DoStore) then
                 !** Now find the PLM index
                 j = 1
                 Diff = PLM(:,j) - (/atom, n, l, m/)
                 err = DOT_PRODUCT(Diff,Diff)
                 Do While (err /= 0)
                    j = j + 1
                    Diff = PLM(:,j) - (/atom, n, l, m/)
                    err = DOT_PRODUCT(Diff,Diff)
                 End Do
                 
                 !*** Finally store the result ***
                 PDOT_Map(:,i) = (/n, L, m, atom, j/)
              End If
           End Do
        End Do
     End Do
  End Do

  PDOT_Max = i
  
  Return
End Subroutine


!******************************************************************************
!
!  ConvertGpoints - Converts the G-points list to the 1-D array
!
!******************************************************************************

Logical Function ConvertProjGpoints(Node)
  Type (BinaryTreeData) :: Node

  Integer :: n, ng
Write(Error_Unit,*) 'Entered ConvertGpoints which is not debugged'
stop

  Count = Count + 1

!Write(Log_Unit,*) 'ConvertGpoints:  Count=',count, ' * Size(Glist)=',Size(Glist(1,:))
!Write(Log_Unit,*) 'ConvertGpoints:  Gmag=',Node%Gmag, ' * N=',Node%Npnt(1)
!write(Log_Unit,*) 'ConvertGpoints : Recip=',G_recip
  
  n = Node%Npnt(1)
  ng = Glist_Half - 1
!  ng = Gpnt_Size(G_LOW) - 1

  If (n<=Glist_Half) then
!  If (n<=Gpnt_Size(G_LOW)) then
     Glist(1:3, Count) = Kvec + Gpnt(1:3,n) 
  else
     Glist(1:3, Count) = Kvec - Gpnt(1:3,n-ng) 
  End if

  Glist(4, Count) = Node%Gmag
  GMap(count) = n
  
  ConvertProjGpoints = .TRUE.

  Return
End Function


!******************************************************************************
!
! CreateProjGlist - create the projector g-points list
!
!    Kvec  - K-point 
!    Glist - Returned list of g-points
!
!******************************************************************************

Subroutine CreateProjGlist(Kvec, Gmax)
  Real, Intent(IN)  :: Kvec(:)
  Integer, Intent(IN) :: Gmax

  Integer :: i, j, Ng
  Real    :: G(3), Gmag
  Type (BinaryTreeNode), Pointer     :: Root, Match
  Type (BinaryTreeData), Pointer     :: Item, Test
  Logical :: Ok
  
Write(Error_Unit,*) 'Entered CreateProjGlist which is not debugged'
stop
  Nullify(Root)
  
  Ng = Gmax - 1
  
  !*** Generate the list ****
  Allocate(Item)
  Gmag = DOT_PRODUCT(Kvec, Kvec)
  Item%Gmag = Gmag
  Item%Npnt = 1
  Call InsertNode(Root, Item, Ok)
  
  Do j=2, Gmax
     G = Kvec + Gpnt(1:3,j)
     Gmag = SQRT(DOT_PRODUCT(G,G))
     Allocate(Item)
     Item%Gmag = Gmag
     Item%Npnt = j
     Call InsertNode(Root, Item, Ok)

     G = Kvec - Gpnt(1:3,j)
     Gmag = SQRT(DOT_PRODUCT(G,G))
     Allocate(Item)
     Item%Gmag = Gmag
     Item%Npnt = Ng + j
     Call InsertNode(Root, Item, Ok)
  End Do

  !*** Now make the 1-d sorted list ***
  Count = 0
  Glist_Half = Gmax
  Call InOrderTraversal(Root, ConvertProjGpoints)
  Call FreeTree(Root, .TRUE.)


  Return
End Subroutine

!******************************************************************************
!
!  StoreRadProj - Stores the Radial projector grids to disk
!
!******************************************************************************

Subroutine StoreRadProj

   Integer       :: atype, n, L, m, j, Kpnt,  DI, Ng, G_Index
   Integer :: G_Size, G_Half, dummy
   Real  :: G(3), Kvec(3),  Gmag , Gmag1, diff, RF
   Real, Pointer :: RP(:)
   Character*200 :: token
   Type (Atom_Info_Fixed), Pointer :: AT
   
   Write(Log_Unit,*) 'StoreRadProj: Start!!!!!!!!!!!!!!!!'
   Call Flush(Log_Unit)

!RETURN !**REMOVE**



   Ng = Gpnt_Size(G_LOW) - 1
   G_half = Gpnt_Size(G_LOW)
   G_Size=Gall_Size(G_LOW)
   Allocate(RP(G_Size))


   

   Do atype = 1, Atom_types
      AT => AtomType_Info(atype)

      n = AT%Basis_Size*Size(BZ%KU(1,:))
      dummy = G_Size
      Call LRU_InitContext(LRU_RadProj(atype), 'RS_RadProj:', &
           FD_RadProj + atype - 1, 50.0, n,  dummy, .TRUE.)

      DI = 0
      Do kpnt=1, Size(BZ%Ku(1,:))
         Kvec = BZ%Ku(:,kpnt)

         Do n = 1, AT%Basis_Size
            L=AT%L_value(n)

            RP = 0

            count = 1
            G = Kvec
            Gmag1 = SQRT(DOT_PRODUCT(G,G))
            RF = RadialFourier(L, Gmag1, AT%Mesh_Step, &
                 AT%Mesh_Size, AT%TP(:,n),AT%RadGrid)
            RP(1) = RF
            Do j=2, G_Half
                  G=Kvec+Gpnt(1:3,j)
                  Gmag1 = SQRT(DOT_PRODUCT(G,G))
                  count = count + 1
                  RF = RadialFourier(L, Gmag1, AT%Mesh_Step, &
                       AT%Mesh_Size, AT%TP(:,n),AT%RadGrid)
                  RP(j) = RF
                  G=Kvec-Gpnt(1:3,j)
                  Gmag1 = SQRT(DOT_PRODUCT(G,G))
                  count = count + 1
                  RF = RadialFourier(L, Gmag1, AT%Mesh_Step, &
                       AT%Mesh_Size, AT%TP(:,n),AT%RadGrid)
                  RP(j+ng) = RF
            End Do


            DI = DI + 1


!Write(Log_Unit,*) 'StoreRadProj:  Writing DI=',DI, ' * Atype=',atype, ' * n=',n, ' * Kpnt=',Kpnt, ' * DOT=',DOT_PRODUCT(RP,RP)

           Where (ABS(RP)<Filter_Min) RP = 0

           Call LRU_PutRec(LRU_RadProj(atype),  DI, RP)

         !   Write(token,*) 'StoreRadProj: Error Writing Data! ', &
         !        ' * Kpnt=',Kpnt, ' * AType=',atype, ' * n=',n, ' * DI=',DI
         !   Write(Log_Unit,*) 'StoreRadProj: Writing Projector Data ', &
         !        ' * Kpnt=',Kpnt, ' * AType=',atype, ' * n=',n, ' * DI=',DI
         !   Call WriteFile_LOW(LRU_RadProj%FD_Base, DI,  &
         !         RP(1:Gall_Size(G_LOW)), token)
!Write(Log_Unit,*) 'StoreRadProj: After Write!!!!!!!!!!!!!!!!!'
         call flush(Log_unit)
         End Do  !** n
      End Do  !**Kpnt

   End Do  !** atype
   DeAllocate(RP)

   Write(Log_Unit,*) 'StoreRadProj: Completed store!'
         call flush(Log_unit)

   Return
End Subroutine


!******************************************************************************
!
! StoreRawProj - Stores the full procectors to disk
!
!******************************************************************************

Subroutine StoreRawProj

  Integer :: Kpnt, Base_Raw, Basis_Size, ng, G_Half, G_Size
  Integer :: Prange(2), dummy, DI, atype
  Integer :: i,j, Base_DI,  atom, L, m, am, n, RadP, PLM_Index
  Complex, Pointer :: Ylm(:), Phase(:)
  Complex, Pointer :: Work(:)
  Real, Pointer :: Proj(:)
  Real :: c1
  Complex :: Pos_M, Neg_M, c2, c3
Complex, Pointer :: Temp2(:), Temp3(:)

!  Write(Log_Unit,*) 'Storing Unfactored projectors....'

  c1 = 4*Pi / sqrt(xtal%Volume)

  PDOT_Dir = -PDOT_Dir
  Call SetStartEnd(Prange)

  ng = Gpnt_Size(G_LOW) - 1
  G_Half=Gpnt_Size(G_LOW)
  G_Size=Gall_Size(G_LOW)

!----
  Do atom=1, Specific_Atoms
     atype = Atom_List(atom)%typeIndex
     Basis_Size = AtomType_Info(atype)%Basis_Size
     Call GetStructFactor( atom, Phase)

     Allocate(Work(G_Size))

     n = RS_PLMMax(atom)*Size(BZ%Ku(1,:))
     dummy = G_Size
     Call LRU_InitContext(LRU_Proj(atom), 'RS_Proj:', &
          FD_Proj + atom - 1, 50.0, n,  dummy, .TRUE.)

     DI = 0
     Do Kpnt = 1, Size(BZ%Ku(1,:))
        Base_DI = Basis_Size*(Kpnt-1)   !** Calc starting DI for Kpnt

        Do n=1, RS_PLMMax(atom)
           RadP = RS_PLM(atom,1,n)
           L = RS_PLM(atom,2,n)
           m = RS_PLM(atom,3,n)
           am = ABS(m)
           
           Call LRU_GetRec(LRU_RadProj(atype),  Base_DI + radP, Proj)
           Call GetYlm( Kpnt, L, am, Ylm)        
           
           c2 = CMPLX(0,1)**L
           
           If (m>=0) then
              !i = 1
              !Work(i) = CONJG(Ylm(i))*Proj(i)*Phase(i)
              
              !Do i=2, G_Half
              !   work(i) = CONJG(Ylm(i))*Proj(i)*Phase(i)
              !   work(i+G_Half-1) = CONJG(Ylm(i+ng)) * Proj(i+G_Half-1) * &
              !        Phase(i+ng)
              !end Do

              Work(:) = CONJG(Ylm(:))*Proj(:)*Phase(:)
              work = c1*c2*work
           Else             
              !i = 1
              !work(i) = Ylm(i)*Proj(i)*Phase(i)
              
              !Do i=2, G_Half
              !   work(i) = Ylm(i)*Proj(i)*Phase(i)
              !   work(i+g_Half-1) = Ylm(i+ng) * Proj(i+G_Half-1) * &
              !        Phase(i+ng)
              !end Do
              
              Work(:) = Ylm(:)*Proj(:)*Phase(:)
              c3 = (-1)**am
              Work = c1*c2*c3*Work
           End If

           DI = DI + 1
           Call FilterValue(Work)
           Call LRU_PutRec(LRU_Proj(atom),  DI, Work)
!Write(Log_Unit, *) 'StoreRawProj: atom=',atom, ' n=', n, ' * Kpnt=',kpnt, ' * DOT=',DOT_PRODUCT(Work,Work)
        End Do
     End Do

     DeAllocate(Work)
  End do

  Return
End Subroutine


!******************************************************************************
!
!  SetStartEnd - Sets the starting and ending points for processing the
!      PDOTS
!
!  P - Projector Start/end range
!
!******************************************************************************

Subroutine SetStartEnd(P)
  Integer, Intent(OUT) :: P(:)
  
  If (PDOT_Dir > 0) then
     P(1) = 1;  P(2) = PLM_Max
  else
     P(1) = PLM_Max;   P(2) = 1
  End If

  Return
End Subroutine


!******************************************************************************
!
! CalcProjProducts_TIME-Calculates ALL the <p|Psi> products and stores them in 
!           the PDOT array.
!
!   Kpnt - Bloch Phase factor
!   Psi  - Wave function
!   PDOT - Output array containing the <p|Psi> products
!*TIME*
!******************************************************************************

Subroutine CalcProjProducts_TIME(psi, PDOT)
  Complex, Intent(IN)  :: Psi(:)
  Complex, Intent(OUT) :: PDot(:)

  Integer :: Kpnt, gmax, ng, G_half, G_size
  Integer :: Prange(2)
  Integer :: i,j, Base_DI,  atom, L, m, am, n, RadP, PLM_Index
  Complex, Pointer :: Proj(:)
  Real :: c1, c3
  Complex :: Pos_M, Neg_M, c2, dot
  Complex, EXTERNAL :: ZDOTC

!Complex :: PD(PLM_Max)

!PDOT = 0
!RETURN  !**REMOVE** Debug Only!!!
  
  Kpnt = Mem_Kpnt
  
  gmax = Gall_Size(G_LOW)

  PDOT_Dir = -PDOT_Dir
  Call SetStartEnd(Prange)

  Base_DI = PLM_Max*(Kpnt-1)  !** Calc starting disk index for Kpnt


!------
  ng = Gpnt_Size(G_LOW) - 1
  G_Half=Gpnt_Size(G_LOW)
  G_Size=Gall_Size(G_LOW)

  Do atom=1, Specific_Atoms
     Base_DI =  RS_PLMMax(atom)*(Mem_Kpnt-1)

     Do n=1, RS_PLMMax(atom)
        RadP = RS_PLM(atom,1,n)
        L = RS_PLM(atom,2,n)
        m = RS_PLM(atom,3,n)
        am = ABS(m)

        Call LRU_GetRec(LRU_Proj(atom),  Base_DI + n, Proj)
        !i = 1
        !dot = Proj(i)*Psi(i)
        
        !Do i=2, G_Half
        !   dot = dot + Proj(i)*Psi(i)
        !   dot = dot + Proj(i+G_Half-1) * Psi(i+ng)
        !end Do

        dot=SUM(Proj(:)*Psi(:))

        PDOT(RS_PLM(atom,4,n)) = dot
     End Do
  End Do


!Do i=1, PLM_Max
!  Write(Log_Unit,*) ' CPP_TIME: i=',i, ' * PDOT=',PDOT(i)
!End Do

  Return
End Subroutine


!******************************************************************************
!
! CalcProjProducts_MEM -Calculates ALL the <p|Psi> products and stores them in 
!           the PDOT array.
!
!   Kpnt - Bloch Phase factor
!   Psi  - Wave function
!   PDOT - Output array containing the <p|Psi> products
!*MEMORY*
!******************************************************************************

Subroutine CalcProjProducts_MEM(psi, PDOT)
  Complex, Intent(IN)  :: Psi(:)
  Complex, Intent(OUT)  :: PDot(:)

  Integer :: Kpnt, G_half, G_size, ng, atype
  Integer :: Prange(2)
  Integer :: i,j, Base_DI,  atom, L, m, am, n, RadP, PLM_Index
  Complex, Pointer :: Ylm(:), Phase(:)
  Complex, Pointer :: Work(:)
  Real, Pointer :: Proj(:)
  Complex :: dot, c1, c2, c3

!PDOT = 0
!RETURN  !**REMOVE** Debug Only!!!
  

  Kpnt = Mem_Kpnt
  
  c1 = 4*Pi / sqrt(xtal%Volume)

  PDOT_Dir = -PDOT_Dir
  Call SetStartEnd(Prange)

  ng = Gpnt_Size(G_LOW) - 1
  G_Half=Gpnt_Size(G_LOW)
  G_Size=Gall_Size(G_LOW)

!----
  Do atom=1, Specific_Atoms
     Call GetStructFactor( atom, Phase)

     atype = Atom_List(atom)%typeIndex
     Base_DI = Atomtype_Info(atype)%Basis_Size * &
          (Mem_Kpnt-1)  !** Calc starting DI for Kpnt

     Do n=1, RS_PLMMax(atom)
        RadP = RS_PLM(atom,1,n)
        L = RS_PLM(atom,2,n)
        m = RS_PLM(atom,3,n)
        am = ABS(m)

        Call LRU_GetRec(LRU_RadProj(atype),  Base_DI + radP, Proj)
        Call GetYlm( Mem_Kpnt, L, am, Ylm)        

        c2 = CMPLX(0,1)**L

       If (m>=0) then
          !i = 1
          !dot = CONJG(Ylm(i))*Proj(i)*Phase(i)*Psi(i)
          
          !Do i=2, G_Half
          !   dot = dot + CONJG(Ylm(i))*Proj(i)*Phase(i)*Psi(i)
          !   dot = dot + CONJG(Ylm(i+ng)) * Proj(i+G_Half-1) * &
          !        Phase(i+ng) * Psi(i+ng)
          !end Do

          dot=SUM(CONJG(Ylm(:))*Proj(:)*Phase(:)*Psi(:))
          dot = c1*c2*dot
        Else             
          !i = 1
          !dot = Ylm(i)*Proj(i)*Phase(i)*Psi(i)
          
          !Do i=2, G_Half
          !   dot = dot + Ylm(i)*Proj(i)*Phase(i)*Psi(i)
          !   dot = dot + Ylm(i+ng) * Proj(i+G_Half-1) * &
          !        Phase(i+ng) * Psi(i+ng)
          !end Do
          
          dot=SUM(Ylm(:)*Proj(:)*Phase(:)*Psi(:))
          c3 = (-1)**am
          dot = c1*c2*c3*dot
        End If

        PDOT(RS_PLM(atom,4,n)) = dot
     End Do
  End Do
!-----

!Write(Log_Unit,*) 'CalcProjProducts: PDOT=',PDOT/sqrt(4*Pi/xtal%Volume)
!Write(*,*) 'CalcProjProducts: PDOT=',PDOT/sqrt(4*Pi/xtal%Volume)

!Write(Log_Unit,*) 'CPP_MEM: PDOT=',PDOT
!Do i=1, PLM_Max
!   Write(Log_Unit,*) 'CPP_MEM: i=',i, ' * PDOT=',PDOT(i)
!end Do

  Return
End Subroutine

!******************************************************************************
!
! CalcProjProducts - Calculates ALL the <p|Psi> products and stores them in 
!           the PDOT array.
!
!   Kpnt - Bloch Phase factor
!   Psi  - Wave function
!   PDOT - Output array containing the <p|Psi> products
!
!******************************************************************************

Subroutine CalcProjProducts(psi, PDOT)
  Complex, Intent(IN) :: Psi(:)
  Complex, Intent(OUT)   :: PDot(:)

  Call Start_Timer(Timer(CalcProjProd_Timer))

  Select Case (Proj_Mode)
     Case (MIN_TIME)
        Call CalcProjProducts_TIME(Psi, PDot)
     Case (MIN_MEMORY)
        Call CalcProjProducts_MEM(Psi, PDot)
     Case (PROJ_RS)
        Call CalcProjProducts_RS(Psi, PDot)
!**        Call CalcProjProducts_TIME(Psi, PDot)
  End Select

  Call Stop_Timer(Timer(CalcProjProd_Timer))

  Return
End Subroutine

  
!******************************************************************************
!
!  CalcAccumVector - Calculates the AccumProj vector
!
!              ----
!        a     \      a    a
!       C  =  /     D   <p | Psi>
!        i     ----   ij   j
!                j
!
!     Ci   - Returned array contining the above sum
!     PDOT - Array containsing the <p|Psi> products
!     Dij  - External function to calculate the Dij matrix elements
!     
! 
!******************************************************************************

Subroutine CalcAccumVector(Ci, PDOT, Dij)
  Complex, Intent(OUT) :: Ci(:)
  Complex, Intent(IN)  :: PDOT(:)
  Complex, EXTERNAL    :: Dij

  Integer :: atom, nili, njlj, Li, Lj, mi, mj, n, RadP, PLM_Index, i, j
  Complex :: ans

  Do i=1, PLM_Max
     atom = PLM(1,i)
     nili = PLM(2,i)
     Li   = PLM(3,i)
     mi   = PLM(4,i)

     ans = 0
     Do j=PLM_AtomRange(1,atom), PLM_AtomRange(2,Atom)
        njlj = PLM(2,j)
        Lj   = PLM(3,j)
        mj   = PLM(4,j)

        Ans = Ans + Dij(atom,nili,njlj,li,lj,mi,mj)*PDOT(j)        
     End Do

     Ci(i) = ans
  End Do

  Return
End Subroutine

!******************************************************************************
!
!  AccumBothProj_TIME - Accumulates the summed |pi>Dij<pj|Psi> and
!                       and |pi>Oij<pj|Psi> to the given vectors
!
!     Hx - Output vector for Dij sum. NOTE: The vector is not initialized to 0!
!     Ox - Output vector for Oij sum. NOTE: The vector is not initialized to 0!
!     PDOT - List of <p|Psi> products
!     Dij  - External function for calculaint matrix elements
!     Oij  - External function for calculaint matrix elements
!*TIME*
!******************************************************************************

Subroutine AccumBothProj_TIME(Hx, Ox, PDot, Dij, Oij)
  Complex, Intent(INOUT) :: Hx(:)
  Complex, Intent(INOUT) :: Ox(:)
  Complex, Intent(IN)    :: PDOT(:)
  Complex, EXTERNAL      :: Dij
  Complex, EXTERNAL      :: Oij

 !Complex :: Ch(PLM_Max), Co(PLM_Max)
  Complex ,allocatable :: Ch(:), Co(:)
  Integer :: gmax, ng, G_Half, G_Size
  Integer :: Prange(2)
  Integer :: i, Base_DI,  atom, L, m, n, am, RadP, PLM_Index
  Complex, Pointer :: Ylm(:), Phase(:)
  Complex, Pointer :: Work(:), Proj(:)
  Real :: c1
  Complex :: c2, c3

  Allocate( Ch(PLM_Max), Co(PLM_Max))
!V=0
!RETURN  !**REMOVE** Debug Only

!Call AccumBothProj_MEM(Hx, Ox, PDot, Dij, Oij)
!RETURN


  Call CalcAccumVector(Ch, PDOT, Dij)
  Call CalcAccumVector(Co, PDOT, Oij)

!------
  ng = Gpnt_Size(G_LOW) - 1
  G_Half=Gpnt_Size(G_LOW)
  G_Size=Gall_Size(G_LOW)

  Do atom=1, Specific_Atoms
     Base_DI =  RS_PLMMax(atom)*(Mem_Kpnt-1)

     Do n=1, RS_PLMMax(atom)
        RadP = RS_PLM(atom,1,n)
        L = RS_PLM(atom,2,n)
        m = RS_PLM(atom,3,n)
        am = ABS(m)

        Call LRU_GetRec(LRU_Proj(atom),  Base_DI + n, Proj)

        c2 = Ch(RS_PLM(atom,4,n))
        c3 = Co(RS_PLM(atom,4,n))
        !i = 1
        !Hx(i) = Hx(i) + CONJG(Proj(i))*c2
        !Ox(i) = Ox(i) + CONJG(Proj(i))*c3
        
        !Do i=2, G_Half
        !   Hx(i) = Hx(i) + CONJG(Proj(i))*c2
        !   Hx(i+ng) = Hx(i+ng) + CONJG(Proj(i+G_Half-1)) * c2
        !   Ox(i) = Ox(i) + CONJG(Proj(i))*c3
        !   Ox(i+ng) = Ox(i+ng) + CONJG(Proj(i+G_Half-1)) * c3
        !end Do

        Hx(:)=Hx(:)+CONJG(Proj(:))*c2
        Ox(:)=Ox(:)+CONJG(Proj(:))*c3
     End Do
  End Do

  DeAllocate( Ch, Co)
!------

  Return
End Subroutine

!******************************************************************************
!
!  AccumBothProj_MEM - Accumulates the summed |pi>Dij<pj|Psi> and
!                       and |pi>Oij<pj|Psi> to the given vectors
!
!     Hx - Output vector for Dij sum. NOTE: The vector is not initialized to 0!
!     Ox - Output vector for Oij sum. NOTE: The vector is not initialized to 0!
!     PDOT - List of <p|Psi> products
!     Dij  - External function for calculaint matrix elements
!     Oij  - External function for calculaint matrix elements
!
!*MEMORY*
!******************************************************************************

Subroutine AccumBothProj_MEM(Hx, Ox, PDot, Dij, Oij)
  Complex, Intent(INOUT) :: Hx(:)
  Complex, Intent(INOUT) :: Ox(:)
  Complex, Intent(IN)    :: PDOT(:)
  Complex, EXTERNAL      :: Dij
  Complex, EXTERNAL      :: Oij

 !Complex :: Ch(PLM_Max), Co(PLM_Max), Hc, Oc
  Complex ::  Hc, Oc
  Complex ,allocatable :: Ch(:), Co(:)
  Integer :: Kpnt, G_Half, G_Size, NG
  Integer :: Prange(2), atype
  Integer :: i, Base_DI,  atom, L, m, n, am, RadP, PLM_Index
  Complex, Pointer :: Ylm(:), Phase(:)
  Complex, Pointer :: Work(:)
  Real, Pointer :: Proj(:)
  Real :: c1
  Complex :: Pos_M, Neg_M, c2, c3


  Allocate(Ch(PLM_Max), Co(PLM_Max))

  Call CalcAccumVector(Ch, PDOT, Dij)
  Call CalcAccumVector(Co, PDOT, Oij)

  Kpnt = Mem_Kpnt

  c1 = 4*Pi / sqrt(xtal%Volume)

  ng = Gpnt_Size(G_LOW) - 1
  G_Half = Gpnt_Size(G_LOW)
  G_Size = Gall_Size(G_LOW)

  Do atom=1, Specific_Atoms
     Call GetStructFactor( atom, Phase)

     atype = Atom_List(atom)%typeIndex
     Base_DI = Atomtype_Info(atype)%Basis_Size * &
          (Mem_Kpnt-1)  !** Calc starting DI for Kpnt

     Do n=1, RS_PLMMax(atom)
        RadP = RS_PLM(atom,1,n)
        L = RS_PLM(atom,2,n)
        m = RS_PLM(atom,3,n)
        PLM_Index = RS_PLM(atom,4,n)

        am = ABS(m)
        
        Call LRU_GetRec(LRU_RadProj(atype),  Base_DI + radP, Proj)
        Call GetYlm( Mem_Kpnt, L, am, Ylm)        

        c2 = CMPLX(0,-1)**L

        If (m>=0) then
           Hc = c1*c2*Ch(PLM_Index)
           Oc = c1*c2*Co(PLM_Index)
           !i = 1
           !Hx(i) = Hx(i) + Hc*Ylm(i)*Proj(i)*CONJG(Phase(i))
           !Ox(i) = Ox(i) + Oc*Ylm(i)*Proj(i)*CONJG(Phase(i))
           
           !Do i=2, G_Half
           !   Hx(i) = Hx(i) + Hc*Ylm(i)*Proj(i)*CONJG(Phase(i))
           !   Hx(i+ng) = Hx(i+ng) + Hc*Ylm(i+ng)*Proj(i+G_half-1) * &
           !        CONJG(Phase(i+ng))
           !   Ox(i) = Ox(i) + Oc*Ylm(i)*Proj(i)*CONJG(Phase(i))
           !   Ox(i+ng) = Ox(i+ng) + Oc*Ylm(i+ng)*Proj(i+G_half-1) * &
           !        CONJG(Phase(i+ng))
           !end Do

           Hx(:)=Hx(:)+Hc*Ylm(:)*Proj(:)*CONJG(Phase(:))
           Ox(:)=Ox(:)+Oc*Ylm(:)*Proj(:)*CONJG(Phase(:))
        Else    
           Hc = c1*c2*Ch(PLM_Index)*(-1)**am
           Oc = c1*c2*Co(PLM_Index)*(-1)**am
           !i = 1
           !Hx(i) = Hx(i) + Hc*Proj(i)*CONJG(Ylm(i)*Phase(i))
           !Ox(i) = Ox(i) + Oc*Proj(i)*CONJG(Ylm(i)*Phase(i))
           
           !Do i=2, G_Half
           !   Hx(i) = Hx(i) + Hc*Proj(i)*CONJG(Ylm(i)*Phase(i))
           !   Hx(i+ng) = Hx(i+ng) + Hc*Proj(i+G_half-1)*CONJG(Ylm(i+ng) * &
           !        Phase(i+ng))
           !   Ox(i) = Ox(i) + Oc*Proj(i)*CONJG(Ylm(i)*Phase(i))
           !   Ox(i+ng) = Ox(i+ng) + Oc*Proj(i+G_half-1)*CONJG(Ylm(i+ng) * &
           !        Phase(i+ng))
           !end Do

           Hx(:)=Hx(:)+Hc*Proj(:)*CONJG(Ylm(:)*Phase(:))
           Ox(:)=Ox(:)+Oc*Proj(:)*CONJG(Ylm(:)*Phase(:))
        End If
     End Do
  End Do
  DeAllocate(Ch, Co)

  Return
End Subroutine


!******************************************************************************
!
!  AccumBothProj - Accumulates the summed |pi>Dij<pj|Psi> and
!                       and |pi>Oij<pj|Psi> to the given vectors
!
!     Hx - Output vector for Dij sum. NOTE: The vector is not initialized to 0!
!     Ox - Output vector for Oij sum. NOTE: The vector is not initialized to 0!
!     PDOT - List of <p|Psi> products
!     Dij  - External function for calculaint matrix elements
!     Oij  - External function for calculaint matrix elements
!
!******************************************************************************

Subroutine AccumBothProj(Hx, Ox, PDot, Dij, Oij)
  Complex, Intent(INOUT) :: Hx(:)
  Complex, Intent(INOUT) :: Ox(:)
  Complex, Intent(IN)    :: PDOT(:)
  Complex, EXTERNAL      :: Dij
  Complex, EXTERNAL      :: Oij

  Call Start_Timer(Timer(AccumProj_Timer))

  Select Case (Proj_Mode)
     Case (MIN_TIME)
        Call AccumBothProj_TIME(Hx, Ox, PDot, Dij, Oij)
     Case (MIN_MEMORY)
        Call AccumBothProj_MEM(Hx, Ox, PDot, Dij, Oij)
     Case (PROJ_RS)
        Call AccumBothProj_RS(Hx, Ox, PDot, Dij, Oij)
!**        Call AccumBothProj_MEM(Hx, Ox, PDot, Dij, Oij)
  End Select

  Call Stop_Timer(Timer(AccumProj_Timer))
  Return
End Subroutine

!******************************************************************************
!
!  AccumProj_TIME - Accumulates the summed |pi>Dij<pj|Psi> to the given vector
!
!     V - Output vector. NOTE: The vector is not initialized to 0!
!     PDOT - List of <p|Psi> products
!     Dij  - External function for calculaint matrix elements
!*TIME*
!******************************************************************************

Subroutine AccumProj_TIME(V, PDot, Dij)
  Complex, Intent(INOUT) :: V(:)
  Complex, Intent(IN)    :: PDOT(:)
  Complex, EXTERNAL      :: Dij

 !Complex :: Ci(PLM_Max)
  Complex ,allocatable :: Ci(:)
  Integer :: Kpnt, gmax, ng, G_half, G_Size
  Integer :: Prange(2)
  Integer :: i, Base_DI,  atom, L, m, n, am, RadP, PLM_Index
  Complex, Pointer :: Ylm(:), Phase(:)
  Complex, Pointer :: Work(:), Proj(:)
  Real :: c1
  Complex :: c3

  Allocate( Ci(PLM_Max))
!V=0
!RETURN  !**REMOVE** Debug Only


  Call CalcAccumVector(Ci, PDOT, Dij)


!------
  ng = Gpnt_Size(G_LOW) - 1
  G_Half = Gpnt_Size(G_LOW)
  G_Size = Gall_Size(G_LOW)

  Do atom=1, Specific_Atoms
     Call GetStructFactor( atom, Phase)

     Base_DI =  RS_PLMMax(atom)*(Mem_Kpnt-1)

     Do n=1, RS_PLMMax(atom)
        RadP = RS_PLM(atom,1,n)
        L = RS_PLM(atom,2,n)
        m = RS_PLM(atom,3,n)
        am = ABS(m)

        Call LRU_GetRec(LRU_Proj(atom),  Base_DI + n, Proj)

        c3 = Ci(RS_PLM(atom,4,n))
        !i = 1
        !V(i) = V(i) + CONJG(Proj(i))*c3
        
        !Do i=2, G_Half
        !   V(i) = V(i) + CONJG(Proj(i))*c3
        !   V(i+ng) = V(i+ng) + CONJG(Proj(i+G_Half-1)) * c3
        !end Do
        V(:)=V(:) + CONJG(Proj(:))*c3
     End Do
  End Do
  DeAllocate( Ci)

!------

  Return
End Subroutine

!******************************************************************************
!
!  AccumProj_MEM - Accumulates the summed |pi>Dij<pj|Psi> to the given vector
!
!     V - Output vector. NOTE: The vector is not initialized to 0!
!     PDOT - List of <p|Psi> products
!     Dij  - External function for calculaint matrix elements
!*MEMORY*
!******************************************************************************

Subroutine AccumProj_MEM(V, PDot, Dij)
  Complex, Intent(INOUT) :: V(:)
  Complex, Intent(IN)    :: PDOT(:)
  Complex, EXTERNAL      :: Dij

 !Complex :: Ci(PLM_Max)
  Complex ,allocatable :: Ci(:)
  Integer :: Kpnt, G_Half, G_Size, ng
  Integer :: Prange(2), atype
  Integer :: i, Base_DI,  atom, L, m, n, am, RadP, PLM_Index
  Complex, Pointer :: Ylm(:), Phase(:)
  Complex, Pointer :: Work(:)
  Real, Pointer :: Proj(:)
  Real :: c1
  Complex :: Pos_M, Neg_M, c2, c3

  Allocate( Ci(PLM_Max))
!V=0
!RETURN  !**REMOVE** Debug Only


  Call CalcAccumVector(Ci, PDOT, Dij)

  Kpnt = Mem_Kpnt

  c1 = 4*Pi / sqrt(xtal%Volume)

  ng = Gpnt_Size(G_LOW) - 1
  G_Half = Gpnt_Size(G_LOW)
  G_Size = Gall_Size(G_LOW)

  Do atom=1, Specific_Atoms
     Call GetStructFactor( atom, Phase)
     atype = Atom_List(atom)%typeIndex
     Base_DI = Atomtype_Info(atype)%Basis_Size * &
          (Mem_Kpnt-1)  !** Calc starting DI for Kpnt

     Do n=1, RS_PLMMax(atom)
        RadP = RS_PLM(atom,1,n)
        L = RS_PLM(atom,2,n)
        m = RS_PLM(atom,3,n)
        PLM_Index = RS_PLM(atom,4,n)

        am = ABS(m)
        
        Call LRU_GetRec(LRU_RadProj(atype),  Base_DI + radP, Proj)
        Call GetYlm( Mem_Kpnt, L, am, Ylm)        

        c2 = CMPLX(0,-1)**L

        If (m>=0) then
           c3 = c1*c2*Ci(PLM_Index)
           !i = 1
           !V(i) = V(i) + c3*Ylm(i)*Proj(i)*CONJG(Phase(i))
           
           !Do i=2, G_Half
           !   V(i) = V(i) + c3*Ylm(i)*Proj(i)*CONJG(Phase(i))
           !   V(i+ng) = V(i+ng) + c3*Ylm(i+ng)*Proj(i+G_half-1) * &
           !        CONJG(Phase(i+ng))
           !end Do

           V(:)=V(:) +c3*Ylm(:)*Proj(:)*CONJG(Phase(:))
        Else    
           c3 = c1*c2*Ci(PLM_Index)*(-1)**am
           !i = 1
           !V(i) = V(i) + c3 * Proj(i)*CONJG(Ylm(i)*Phase(i))
           
           !Do i=2, G_Half
           !   V(i) = V(i) + c3 * Proj(i)*CONJG(Ylm(i)*Phase(i))
           !   V(i+ng) = V(i+ng) + c3*Proj(i+G_half-1)*CONJG(Ylm(i+ng) * &
           !        Phase(i+ng))
           !end Do
           V(:)=V(:)+c3*Proj(:)*CONJG(Ylm(:)*Phase(:))
        End If
     End Do
  End Do
  DeAllocate( Ci)

  Return
End Subroutine


!******************************************************************************
!
!  AccumProj - Accumulates the summed |pi>Dij<pj|Psi> to the given vector
!
!     V - Output vector. NOTE: The vector is not initialized to 0!
!     PDOT - List of <p|Psi> products
!     Dij  - External function for calculaint matrix elements
!
!******************************************************************************

Subroutine AccumProj(V, PDot, Dij)
  Complex, Intent(INOUT) :: V(:)
  Complex, Intent(IN)    :: PDOT(:)
  Complex, EXTERNAL      :: Dij


  Call Start_Timer(Timer(AccumProj_Timer))
  Select Case (Proj_Mode)
     Case (MIN_TIME)
        Call AccumProj_TIME(V, PDot, Dij)
     Case (MIN_MEMORY)
        Call AccumProj_MEM(V, PDot, Dij)
     Case (PROJ_RS)
        Call AccumProj_RS(V, PDot, Dij)
!**        Call AccumProj_TIME(V, PDot, Dij)
  End Select

  Call Stop_Timer(Timer(AccumProj_Timer))
  Return
End Subroutine


!******************************************************************************
!
!  RS_Update - Frees all the RS proj data and updates the data to the new
!       atom positions
!
!******************************************************************************

Subroutine RS_Update
  Integer :: atom

  Do atom=1, Specific_Atoms
     Call LRU_FreeAll(LRU_RS_Proj(atom))
     Close(LRU_RS_Proj(atom)%FD_Base)

     DeAllocate(RS_RLE(atom)%RLE)
  End do

  Call RS_CreateProj

  Return
End Subroutine
!******************************************************************************
!
! InitProjectors - Initializes the Projector routines
!
!   FD_RadProj - Base File unit for the radial projectors
!   FD_Proj - Base File unit for the complete projectors
!   MaxMem  - Max amout of memory to use for structure factors(Mbytes)
!
!******************************************************************************

Subroutine InitProjectors(FD_RadProj, FD_Proj, MaxMem)
  Integer,       Intent(IN) :: FD_RadProj
  Integer,       Intent(IN) :: FD_Proj
  Real,          Intent(IN) :: MaxMem

  Integer :: L, m, MaxRec, RS_Size(0:0), i

  PDOT_Dir = -1  !** Init previous direction of processing the pdots

!RETURN !**REMOVE**

  Allocate(LRU_RadProj(Atom_Types), LRU_Proj(Specific_Atoms))
  Do i=1, Atom_types
     If (run_mode == Proj_RS) then
        If (Atomtype_Info(i)%Gcut_Proj > PW_Gcut(G_HIGH)) then
           AtomType_Info(i)%Gcut_Proj = PW_Gcut(G_High)
        else If (Atomtype_Info(i)%Gcut_Proj <= 0) then
           AtomType_Info(i)%Gcut_Proj = PW_Gcut(G_LOW)
        End If
     else
        AtomType_Info(i)%Gcut_Proj = PW_Gcut(G_LOW)
     End If
    
     Atomtype_Info(i)%Gpnt_Size = FindGcut(Atomtype_Info(i)%Gcut_Proj, &
          Gpnt,Gpnt_Size(G_HIGH))
  End Do

!**REMOVE** Uncomment the next 2 lines 
!**  Call Create_RadProjMap  !** Create the radial projector map
!**  Proj_Max = Size(BZ%Ku(1,:)) * RadProj_Max

  Call Create_PLM(.FALSE.)   !** Determine the PLM and PDOT array sizes
  Call Create_PDOT_Map(.FALSE.)

  Allocate(LRU_RS_Proj(Specific_Atoms), RS_PLMMax(Specific_Atoms), &
       RS_RLE(Specific_Atoms))


  Proj_Max = PLM_Max

  !*** Allocate the space ***
  Allocate(PLM(4,PLM_Max), PLM_AtomRange(2,Specific_Atoms), &
       PDOT_Map(5,PDOT_Max))

  Call Create_PLM(.TRUE.)   !** Now fill the arrays
  Call Create_PDOT_Map(.TRUE.)

  Call RS_Find_PLMMax
  Call RS_FindPLM

!RS_PLMMax = 1
!RS_PLM(:,2,1) = 1
!RS_PLM(:,3,1) = -1
!RS_PLM(:,2,2) = 1
!RS_PLM(:,3,2) = -1

!RETURN !**REMOVE**

  MaxRec = Size(BZ%Ku(1,:)) * RadProj_Max

Write(Log_Unit,*) 'InitProjectors: Size(BZ%Ku))=',Size(BZ%Ku(1,:)), &
     ' * RadProj_Max=',RadProj_Max, ' * MaxRec=',MaxRec, &
     ' * PLM_Max=',PLM_Max, ' * PDOT_Max=',PDOT_Max

Call Flush(Log_Unit)
  Call StoreRadProj  !** Store the radial projectors on disk

  If (Run_Mode == MIN_TIME) Call StoreRawProj


Write(log_unit,*) 'Proj_mode = ',Proj_mode
Call Flush(Log_Unit)
!Call StoreRawProj
If (Proj_Mode == Proj_RS) then
  Allocate(RS_wt(FFT_Grid(4,G_PROJ)))
  Call MakeIntWeights(RS_Wt)
RS_Wt = 1

!**  Call RS_CreateProjMap

  Allocate(RS_Work1(FFT_Grid(4,G_PROJ)), RS_Work2(FFT_Grid(4,G_PROJ)))
  Call RS_CreateProj

  Allocate(RLE_Run(NumKpnts))
  Do i=1, NumKpnts
     RLE_Run(i)%N = 0
  End do
  Call RS_MakeGlobal_RLE
End If
 Write(Log_Unit,*) ' Completed InitProjectors '
Call Flush(Log_Unit)

  Return
End Subroutine


End Module
