!***************************************************************************
!
! File : qtbzi.f90
!   by : Alan Tackett
!   on : 7/26/95
!  for : PAW Method
!
!
! qtbzi - Quadratic Tetrahedron Brillouin Zone Integration
!
! This file contains routines to perform BZ integration using a
! Quadratic Tetrahedron Method. Currently all integration is done
! numberically.  Hopefully it can be replaced at a later date with an
! analytic version.
!
!**************************************************************************

Module qtbzi

implicit none

Type TetraVertices
   Integer         :: Corners(10)       !** Tetrahedron Corners
End Type



Type (TetraVertices), Private, Target  :: TetraCorners(6)
Real,                 Private, Pointer :: Klist(:,:)    !** K points list
Real,                 Private, Pointer :: Kenergy(:,:)
Integer,              Private          :: TotalKpnts  !** Num K points
Integer,              Private, Pointer :: TetraList(:,:) !** Tetrahedron list
Integer,              Private, Pointer :: Cube(:,:)   !** Cubelist
Integer,              Private          :: TotalTetra  !** Num Tetrahedra
Integer,              Private          :: TotalCubes  !** Num Cubes
Integer,              Private          :: TotalBands  !** Num Energy Bands
Integer,              Private          :: TotalElectrons !** # electrons
Real,                 Private          :: TetraVolume !** Tetrahedron Volume
Real,                 Private          :: FermiEnergy !** Fermi Energy Level
Real,                 Private          :: TetraConst  !** Volume Const
Real,                 Private, Pointer :: Weights(:,:)!** Integration Weights 
Integer,              Public           :: qtbzi_INTERVALS !**Num Int. intervals


Contains

!****************************************************************************
!
! InitQTBZI - Initialize the QTBZI routines for use.
!
!  Num_Bands- # of Energy Bands
!  Num_Electrons - # of electrons
!  KSize    - # of K points
!  KPnts    - Pointer to K Points table
!  KE       - Pointer to Energy Table for K points
!  Wt       - Pointer to the Integration Weights Table
!  CubeNum  - # of Cubes
!  CubeList - Pointer to Cube info
!  CubeVol  - Volume of a cube in reciprocal space
!  UnitCellVol  - Volume of a Unit Cell
!
!****************************************************************************

Subroutine InitQTBZI(Num_Bands, Num_Electrons, KSize, KPnts, Ke, Wt, &
                     CubeNum, CubeList, CubeVol, UnitCellVol)
  Integer,                 Intent(IN) :: Num_Bands
  Integer,                 Intent(IN) :: Num_Electrons  
  Integer,                 Intent(IN) :: KSize
  Real,           Pointer             :: KPnts(:,:)
  Real,           Pointer             :: Ke(:,:)
  Real,           Pointer             :: Wt(:,:)
  Integer,                 Intent(IN) :: CubeNum
  Integer,        Pointer             :: CubeList(:,:)
  Real,                    Intent(IN) :: CubeVol
  Real,                    Intent(IN) :: UnitCellVol


  Integer :: i, j, k, n, Sum
  Integer, Pointer :: Corner(:)

            !*********** Initialize the Cube to Tetra LUT***********
  TetraCorners(1)%Corners = (/ 3,  5,  7,  2,  1, 12, 21, 14, 11,  4/)
  TetraCorners(2)%Corners = (/19, 20, 21, 10,  1, 13,  7, 14,  4, 11/)
  TetraCorners(3)%Corners = (/ 3,  6,  9,  5,  7, 12, 21, 15, 14,  8/) 
  TetraCorners(4)%Corners = (/19, 22, 25, 20, 21, 13,  7, 16, 14, 23/)
  TetraCorners(5)%Corners = (/ 9, 15, 21, 18, 27,  8,  7, 14, 17, 24/)
  TetraCorners(6)%corners = (/25, 26, 27, 23, 21, 16,  7, 17, 14, 24/)

            !*********** Initialize the tables ***********
  TotalKpnts = KSize
  Klist => Kpnts
  KEnergy => Ke
  Weights => Wt

  TotalCubes = CubeNum
  Cube => CubeList

  TotalBands = Num_Bands
  TetraConst = CubeVol / UnitCellVol 
  TotalElectrons = Num_Electrons

  qtbzi_INTERVALS = 15

  TotalTetra = 6*TotalCubes
  Allocate(TetraList(TotalTetra, 10))

  n = 0
  Do i=1, Totalcubes
    Do j=1, 6
       n = n + 1
       TetraList(n,1:10) = Cube(i,TetraCorners(j)%Corners(1:10))
    End Do
  End Do

  Return
End Subroutine


!****************************************************************************
!
!  qt_TetraNOS - Integrates over the tetrahedron the NOS function up to
!                the Fermi surface.
!
!  NOS - Upon Return contains the NOS for the Tetrahedron 
!  N - Number of intervals for subdivision.
!  ef - Fermi Energy
!  e1,..,e10 - Tetrahedron Energy Values
!
!  Return Values
!     Returns the integral of the function below the Fermi Surface in NOS
!
!****************************************************************************

Real Function qt_TetraNOS(N,ef,e1,e2,e3,e4,e5,e6,e7,e8,e9,e10)
   Integer, Intent(IN) :: N
   Real,    Intent(IN) :: ef,e1,e2,e3,e4,e5,e6,e7,e8, e9, e10


   Real :: delta, dv, t, ans, x, y, z, energy
   Real :: w1,w2,w3,w4,w5,w6,w7,w8,w9,w10
   Integer :: i, j, k   


   delta = 1.0/N
   dv = delta*delta*delta

   ans = 0

   do i=1, N
     x = delta * (i-0.5)
     Do j=1, N-i+1
       y = delta * (j-0.5)
       Do k=1, N-i-j+1
         z = delta * (k-0.5)

         t = x + y + z - 1
         w1 = 2 * (t + 0.5) * t 
         w2 = -4*x*t
         w3 = x * (2*x-1)
         w4 = -4*y*t
         w5 = y * (2*y-1)
         w6 = -4*z*t
         w7 = z*(2*z-1)
         w8 = 4*x*z
         w9 = 4*y*z
         w10 = 4*x*y

         energy = w1*e1+w2*e2+w3*e3+w4*e4+w5*e5+ &
                  w6*e6+w7*e7+w8*e8+w9*e9+w10*e10
         if (energy<=ef) ans = ans + dv
       End Do
     End Do
   End Do

   qt_TetraNOS = ans * TetraConst

   Return
End Function


!****************************************************************************
!
!  qt_TetraWeights - Integrates over the tetrahedron the Weight functions up to
!                    the Fermi surface.
!
!  W - Array containing the int of the Weight Functions
!  N - Number of intervals for subdivision.
!  ef - Fermi Energy
!  e1,..,e10 - Tetrahedron Energy Values
!
!  Return Values
!     Returns the integral of the weight functions below the Fermi Surface
!
!***************************************************************************

Subroutine qt_TetraWeights(W,N,ef,e1,e2,e3,e4,e5,e6,e7,e8,e9,e10)
   Real,   Intent(OUT) :: W(:)
   Integer, Intent(IN) :: N
   Real,    Intent(IN) :: ef,e1,e2,e3,e4,e5,e6,e7,e8, e9, e10


   Real :: delta, dv, t, x, y, z, energy
   Real :: w1,w2,w3,w4,w5,w6,w7,w8,w9,w10
   Integer :: i, j, k   
   
   delta = 1.0/N
   dv = delta*delta*delta*TetraConst
   W = 0

   do i=1, N
     x = delta * (i-0.5)
     Do j=1, N-i+1
       y = delta * (j-0.5)
       Do k=1, N-i-j+1
         z = delta * (k-0.5)

         t = x + y + z - 1
         w1 = 2 * (t + 0.5) * t 
         w2 = -4*x*t
         w3 = x * (2*x-1)
         w4 = -4*y*t
         w5 = y * (2*y-1)
         w6 = -4*z*t
         w7 = z*(2*z-1)
         w8 = 4*x*z
         w9 = 4*y*z
         w10 = 4*x*y

         energy = w1*e1+w2*e2+w3*e3+w4*e4+w5*e5+w6*e6+w7*e7+w8*e8+w9*e9+w10*e10
         if (energy<=ef) then 
            w(1) = w(1) + w1
            w(2) = w(2) + w2
            w(3) = w(3) + w3
            w(4) = w(4) + w4
            w(5) = w(5) + w5
            w(6) = w(6) + w6
            w(7) = w(7) + w7
            w(8) = w(8) + w8
            w(9) = w(9) + w9
            w(10) = w(10) + w10
          End If

       End Do
     End Do
   End Do

   w(1) = w(1)*dv
   w(2) = w(2)*dv
   w(3) = w(3)*dv
   w(4) = w(4)*dv
   w(5) = w(5)*dv
   w(6) = w(6)*dv
   w(7) = w(7)*dv
   w(8) = w(8)*dv
   w(9) = w(9)*dv
   w(10) = w(10)*dv

   Return
End Subroutine


!****************************************************************************
!
! qt_BandNOS - Calculates the NOS for a given Band
!
!   Band - Band to use
!   ef   - Fermi Energy
!
! Return Values
!     Returns the Bands NOS 
!
!****************************************************************************

Real Function qt_BandNOS(Band, ef)
  Integer, Intent(IN) :: Band
  Real,    Intent(IN) :: ef

  Integer          :: i, j, k
  Real             :: NOS, t
  Real,    Pointer :: KE(:)
  Integer, Pointer :: CC(:), TC(:)

  KE => KEnergy(Band, :)
  NOS = 0

  Do i = 1, TotalTetra
    TC => TetraList(i,:) 

    t = qt_TetraNOS(qtbzi_INTERVALS, ef,                &
                    KE(TC(1)),KE(TC(2)), KE(TC(3)),KE(TC(4)),   &
                    KE(TC(5)),KE(TC(6)), KE(TC(7)),KE(TC(8)),   &
                    KE(TC(9)),KE(TC(10)))
    NOS = NOS + t
  End Do

  qt_BandNOS = NOS
  Return
End Function


!****************************************************************************
!
! qt_BandWeights - Calculates the Weights for a given Band
!
!   Band - Band to use
!   ef   - Fermi Energy
!
! Return Values
!     Stores the Weight in the global Weight array
!
!****************************************************************************

Subroutine qt_BandWeights(Band, ef)
  Integer, Intent(IN) :: Band
  Real,    Intent(IN) :: ef

  Integer          :: i, j, k
  Real             :: Wt(10)
  Real,    Pointer :: KE(:)
  Integer, Pointer :: CC(:), TC(:)

  KE => KEnergy(Band, :)

  Do i = 1, TotalTetra
    TC => TetraList(i,:) 

    Call qt_TetraWeights(Wt, qtbzi_INTERVALS, ef,                 &
                    KE(TC(1)),KE(TC(2)), KE(TC(3)),KE(TC(4)),    &
                    KE(TC(5)),KE(TC(6)), KE(TC(7)),KE(TC(8)),    &
                    KE(TC(9)),KE(TC(10)))

!    Weights(Band, TC(1:10)) = Weights(Band, TC(1:10)) + Wt(1:10)

     Do j = 1, 10
       Weights(Band,TC(j)) = Weights(Band,TC(j)) + Wt(j)
     End Do
  End Do

  Return
End Subroutine


!****************************************************************************
!
! qt_NOS - Calculates the NOS for all bands
!
!   ef   - Fermi Energy
!
! Return Values
!     Returns the NOS 
!
!****************************************************************************

Real Function qt_NOS(ef)
   Real, Intent(IN) :: ef

   Integer :: i
   Real    :: NOS

!Write(*,*) 'qt_NOS : Ef=',ef, ' * TotalBands=',TotalBands
 
   NOS = 0
   Do i = 1, TotalBands
     NOS = NOS + qt_BandNOS(i, ef)
   End Do
!Write(*,*) 'qt_NOS : NOS=',NOS
   qt_NOS = NOS
   Return
End Function

!****************************************************************************
!     
! qt_Weights - Calculates the Weights for all bands
!
!   ef   - Fermi Energy
!
! Return Values
!     Stores all weights in the the global WEIGHTS
!
!****************************************************************************

Subroutine qt_Weights(ef)
   Real, Intent(IN) :: ef

   Integer :: i

   Weights = 0

   Do i = 1, TotalBands
     Call qt_BandWeights(i, ef)
   End Do

   Return
End Subroutine

End Module
   
