!******************************************************************************
!
! File : REAL_lru.inc
!   by : Alan Tackett
!   on : 04/19/1999
!  for : PW-PAW
!
!  This is a generic module implementing a Least Recently Used algorithm for
!  I/O management for REAL arrays.
!
!******************************************************************************

Module real_lru

Use paw_inout
Use fileio

Implicit NONE!!!!

Integer, PRIVATE, PARAMETER :: COUNT_START = 0           !** Start counter
Integer, PRIVATE, PARAMETER :: COUNT_MAX   = 2147483640  !** Reset counter val

Type REAL_LRUGrid
   Integer                  :: Index   !** Disk Index
   Integer                  :: Count   !** Least recently used counter
   Logical                  :: DoSave  !** Should save before purging
   REAL, Pointer :: Ptr(:)  !** Pointer to Grid
End Type

Type REAL_LRUData          !*** Data Structure for holding the info
   Type (REAL_LRUGrid), Pointer :: Buffer(:)
   Integer, Pointer :: DiskMap(:)  
End Type

Type REAL_LRU_Cache   !** Cache data structure
   Integer :: Hits
   Integer :: Misses
   Integer :: Total
End Type

Type REAL_LRU_Context    !** Context containing all LRU data structures needed
   Character*100 :: Id    !** User-supplied text Id for debugging
   Integer :: Count       !** LRU count
   Integer :: NumRecs     !** Number of Records loaded
   Integer :: TotalRecs       !** Total number of Records
   Integer :: MaxMem      !** MAx Amount of memory to use for buffers
   Integer :: MemUsed     !** Amount of memory currently used
   Integer :: MemFree     !** Amount of Free memory
   Integer :: LastIndex       !** Last index
   Integer :: FD_Base         !** Base unit
   Logical :: ReplaceOldest   !** Replacement mode
   Integer, Pointer :: Data_Size  !** Record Size
   Type(REAL_LRU_Cache), Pointer :: Cache
   Type (REAL_LRUData), Pointer :: LRUstruc  !**Data structure of LRU module
End Type


!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
!  REAL_LRU_Write - Writes a Record to disk
!
!  LRU     - LRU context
!  Index   - K point index
!  Ptr     - Data to store
!
!******************************************************************************

Subroutine REAL_LRU_Write(LRU,  Index, Ptr)
  Type (REAL_LRU_Context), Intent(INOUT) :: LRU
  Integer,         Intent(IN) :: Index
  REAL,        Intent(IN) :: Ptr(:)

  Integer       :: N
  Character*100 :: Token


  Write(token,*) 'REAL_WriteLRU: Error Writing Data! :', &
       ' * Index:',index, ' * ID=',TRIM(LRU%Id)

  N = LRU%Data_Size
  Call WriteFile_LOW(LRU%FD_Base, Index,  Ptr(1:N), token)

  Return
End Subroutine


!******************************************************************************
!
!  REAL_LRU_Read - Reads a record from disk
!
!  LRU     - LRU context
!  Index   - K point index
!  Ptr     - Data to store
!
!******************************************************************************

Subroutine REAL_LRU_Read(LRU,  Index, Ptr)
  Type (REAL_LRU_Context), Intent(INOUT) :: LRU
  Integer,            Intent(IN)    :: Index
  REAL,           Intent(OUT)   :: Ptr(:)

  Character*100 :: Token
  Integer :: N

  Write(token,*) 'REAL_LRU_Read: Error reading data! :', &
	 ' * Index:',index, ' * ID=',TRIM(LRU%Id)

  N = LRU%Data_Size
  Call ReadFile_LOW(LRU%FD_Base, Index,  Ptr(1:N), token)

  Return
End Subroutine


!******************************************************************************
!
! REAL_LRU_GetRec - Gets the requested record from either memory or disk.
!
!
!  LRU     - LRU context
!  Index   - K point index
!  Ptr     - Data to store
!
!******************************************************************************

Subroutine REAL_LRU_GetRec(LRU,  Index, Ptr)
  Type (REAL_LRU_Context), Intent(INOUT) :: LRU
  Integer,            Intent(IN)    :: Index
  REAL,           Pointer       :: Ptr(:)

  Integer :: Rec, BufPos, Last,  dir, i
  Type (REAL_LRUGrid), Pointer :: LG(:)
  Integer, Pointer :: DiskMap(:)

  LG => LRU%LRUstruc%Buffer
  DiskMap => LRU%LRUstruc%DiskMap

  bufPos = diskMap(Index)
  If (BufPos == 0) then  !**** Have to load it from disk *****
     LRU%Cache%Misses = LRU%Cache%Misses + 1

     If (REAL_LRU_CanAlloc(LRU)) then   !*** Can allocate space
        BufPos = REAL_LRU_Alloc(LRU) !*** Alloc space and get  Buf
     else
        BufPos = REAL_LRU_MakeSpace(LRU)!**Free other bufs/alloc new
     End If

     Call REAL_LRU_Read(LRU,  Index, LG(BufPos)%Ptr)

     LG(BufPos)%Index = Index
     LG(BufPos)%DoSave = .FALSE.
     DiskMap(Index) = BufPos
  else
     LRU%Cache%Hits = LRU%Cache%Hits + 1
  End If


  Ptr => LG(BufPos)%Ptr

  LG(BufPos)%Count = LRU%Count
  LRU%Count = LRU%Count + 1;
  LRU%LastIndex = BufPos

  If (LRU%Count >= COUNT_MAX) Call REAL_LRU_ResetCount(LRU)

  Return
End Subroutine


!******************************************************************************
!
! REAL_LRU_PutRec - Puts the requested record to either memory or disk.
!
!
!  LRU     - LRU context
!  Index   - K point index
!  Ptr     - Data to store
!
!******************************************************************************

Subroutine REAL_LRU_PutRec(LRU,  Index, Ptr)
  Type (REAL_LRU_Context), Intent(INOUT) :: LRU
  Integer,            Intent(IN)    :: Index
  REAL,           Intent(IN)    :: Ptr(:)
  
  Integer :: Rec, BufPos, Last
  Integer, Pointer :: DiskMap(:)
  Type (REAL_LRUGrid), Pointer :: LG(:)


  LG => LRU%LRUstruc%Buffer
  DiskMap => LRU%LRUstruc%DiskMap

  LRU%Cache%Hits = LRU%Cache%Hits + 1

  BufPos = DiskMap(Index)

  If (BufPos == 0) then  !**** Not currently loaded *****
     If (REAL_LRU_CanAlloc(LRU)) then   !*** Can allocate space
        BufPos = REAL_LRU_Alloc(LRU) !*** Alloc space and get  Buf
     else
        BufPos = REAL_LRU_MakeSpace(LRU)!**Free other bufs/alloc new
     End If

     LG(BufPos)%Index = Index
     DiskMap(Index) = BufPos
  End If

  LG(BufPos)%Ptr = Ptr
  LG(BufPos)%DoSave = .TRUE.

  LG(BufPos)%Count = LRU%Count
  LRU%Count = LRU%Count + 1;

  If (LRU%Count >= COUNT_MAX) Call REAL_LRU_ResetCount(LRU)

  Return
End Subroutine



!******************************************************************************
!
!  REAL_LRU_CanAlloc - Checks to see if free space is available for
!             buffer allocation
!
!  LRU   - LRU context
!
!******************************************************************************

Logical Function REAL_LRU_CanAlloc(LRU)
  Type (REAL_LRU_Context), Intent(INOUT) :: LRU

  If (LRU%MemFree > LRU%Data_Size) then
     REAL_LRU_CanAlloc = .TRUE.
  else
     REAL_LRU_CanAlloc = .FALSE.
  End If

  Return
End Function

!******************************************************************************
!
!   REAL_LRU_Alloc - Allocates a buffer  and returns
!            the new buffer position
!
!   LRU   - LRU context
!
!******************************************************************************

Integer Function REAL_LRU_Alloc(LRU)
  Type (REAL_LRU_Context), Intent(INOUT) :: LRU

  Type (REAL_LRUGrid), Pointer :: LG(:)
  Integer       :: i, BufPos
  Character*100 :: msg

  LG => LRU%LRUstruc%Buffer

  BufPos = 1
  Do While ((LG(BufPos)%Index > 0) .AND. (BufPos < LRU%TotalRecs))
     BufPos = BufPos + 1
  End Do

  Allocate(LG(BufPos)%Ptr(LRU%Data_Size), STAT=i)
  Write(msg,*) 'LRU_Alloc: Could Not allocate Buffer array!',  &
       i, ' * BufPos=',BufPos, ' * ID=',TRIM(LRU%ID)
  Call Check_Error(i, msg, Error_Unit, .TRUE.,paw_wc, "LRU_Alloc:")

  LRU%MemUsed = LRU%MemUsed + LRU%Data_Size
  LRU%MemFree = LRU%MemFree - LRU%Data_Size

  REAL_LRU_Alloc = BufPos

  Return
End Function


!******************************************************************************
!
!  REAL_LRU__FindOldest - Finds the oldest Phase factor and returns 
!         and buffer position
!
!  BufPos - Buffer position
!
!******************************************************************************

Subroutine REAL_LRU_FindOldest(LRU,  BufPos)
  Type (REAL_LRU_Context), Intent(INOUT) :: LRU
  Integer, intent(OUT) :: BufPos

  Type (REAL_LRUGrid), Pointer :: LG(:)
  Integer :: i, j, OldCount
  Logical :: ReplaceOldest

  ReplaceOldest = LRU%ReplaceOldest

  If (ReplaceOldest) then
     OldCount = COUNT_MAX+1
  else
     OldCount = -1
  End If

     LG => LRU%LRUstruc%Buffer
     
     If (ReplaceOldest) then
        Do j=1, LRU%TotalRecs
           If ((LG(j)%Index > 0) .AND. (OldCount > LG(j)%Count)) then
              OldCount = LG(j)%Count
              BufPos = j
           End If
        End Do
     else
        Do j=1, LRU%TotalRecs
           If ((LG(j)%Index > 0) .AND. (OldCount < LG(j)%Count)) then
              OldCount = LG(j)%Count
              BufPos = j
           End If
        End Do
     End If

  Return
End Subroutine


!******************************************************************************
!
! REAL_LRU_MakeSpace - Deletes the oldest buffers until enough room is made
!            to Allocate a buffer  and returns
!            the new buffer position
!
!   LRU   - LRU context
!
!******************************************************************************

Integer Function REAL_LRU_MakeSpace(LRU)
  Type (REAL_LRU_Context), Intent(INOUT) :: LRU

  Integer   ::  OldPos, temp
  Logical   :: ReUse

  ReUse = .FALSE.

  Do While ((LRU%MemFree < LRU%Data_Size) .AND. (.NOT. ReUse))
     Call REAL_LRU_FindOldest(LRU,  OldPos)

     !*** check to see if it should be saved first ***
     If (LRU%LRUstruc%Buffer(OldPos)%DoSave) then
        Call REAL_LRU_Write(LRU, &
             LRU%LRUstruc%Buffer(OldPos)%Index, &
             LRU%LRUstruc%Buffer(OldPos)%Ptr)
        LRU%LRUstruc%Buffer(OldPos)%DoSave = .FALSE.
        LRU%Cache%Misses = LRU%Cache%Misses + 1
     End If

     temp = LRU%MemFree + LRU%Data_Size

     LRU%LRUstruc%DiskMap( &
          LRU%LRUstruc%Buffer(OldPos)%Index) = 0

     If ((temp < LRU%Data_Size) ) then
        ReUse = .TRUE.
     else
        DeAllocate(LRU%LRUstruc%buffer(OldPos)%Ptr)
        
        LRU%LRUstruc%buffer(OldPos)%Index = -1
        LRU%MemFree = LRU%MemFree + LRU%Data_Size
        LRU%MemUsed = LRU%MemUsed - LRU%Data_Size
     End If
  End Do

  If (ReUse) then
     REAL_LRU_MakeSpace = OldPos
  Else
     REAL_LRU_MakeSpace = REAL_LRU_Alloc(LRU)
  End If

  Return
End Function

!******************************************************************************
!
!  REAL_LRU_ResetCount - Resets all the counts to COUNT_START
!
!   LRU   - LRU context
!
!******************************************************************************

Subroutine REAL_LRU_ResetCount(LRU)
  Type (REAL_LRU_Context), Intent(INOUT) :: LRU

     LRU%LRUstruc%buffer(:)%Count = COUNT_START

  LRU%Count = COUNT_START

  Return
End Subroutine

!******************************************************************************
!
!  REAL_LRU_FreeALL - Frees all the memory associated with the LRU buffers.
!
!   LRU   - LRU context
!
!******************************************************************************

Subroutine REAL_LRU_FreeAll(LRU)
  Type (REAL_LRU_Context), Intent(INOUT) :: LRU

  Integer :: i

     LRU%LRUstruc%DiskMap = 0

     Do i=1, LRU%TotalRecs
        If (LRU%LRUstruc%buffer(i)%Index>0) then
           DeAllocate(LRU%LRUstruc%Buffer(i)%Ptr)
           LRU%LRUstruc%Buffer(i)%Index = -1
           Nullify(LRU%LRUstruc%Buffer(i)%Ptr)
        End If
     End Do

  LRU%MemUsed = 0
  LRU%MemFree = LRU%MaxMem

  Return
End Subroutine


!******************************************************************************
!
! REAL_DisplayBufferInfo - Displays the Buffer Information to 
!          the specified device
!
!   LRU   - LRU context
!   Output - Output unit
!
!******************************************************************************

Subroutine REAL_LRU_DisplayBufferInfo(LRU, Output)
  Type (REAL_LRU_Context), Intent(INOUT) :: LRU
  Integer,            Intent(IN) :: OutPut

  Integer :: i, j, LSize, Used
  Integer ::  Mem_Proj, Mem_Psi,  Calc_Proj,Calc_Psi
  Integer :: TotalMem, CacheTotal, CacheHits, CacheMisses
  Type (REAL_LRUGrid), Pointer :: LG(:)
  Real :: Hits, miss, total

  Write(output,*) 'LRU_PRINT:  Printing Buffer information*******'
  Write(output,*) 'ID=',Trim(LRU%ID), ' * Max Record:',LRU%TotalRecs

  TotalMem = 0
     LG => LRU%LRUstruc%Buffer
     LSize = LRU%Data_Size
     
     Used = 0
     Do j=1, LRU%TotalRecs
        If (LG(j)%Index > 0) Used = Used + 1
     End Do

     TotalMem = TotalMem + Used*LSize
     Write(Output, *) 'Buffer info  * Base Size:',LSize
     Write(Output, *) '   Allocated :',Used, ' *  Mem:',Used*LSize

     CacheTotal = LRU%Cache%Hits + LRU%Cache%Misses
     Write(Output,*) ' Cache  Stats: Total:',CacheTotal, &
          ' * Hits:',LRU%Cache%Hits, ' * Misses:',LRU%Cache%Misses
     if (CacheTotal > 0) then
        total = CacheTotal
        hits = (1.0*LRU%Cache%Hits)/total
        miss = (1.0*LRU%Cache%Misses)/total
        Write(Output,*) ' Cache  Stats:  %Hits:',Hits, ' * %Misses:',Miss
     End If
     Write(Output, *) '   Data : (Buffer, Disk, Size, Count)'
     Do j=1, LRU%TotalRecs
        If (LG(j)%Index > 0) then
           Write(Output,*) j,' *',LG(j)%Index,' *',Size(LG(j)%Ptr), &
                ' *', LG(j)%Count
        End If
     End Do

  Write(Output,*) 'Total Memory-->  Max:', LRU%MaxMem, ' * Free:', LRU%MemFree, ' * Used:', LRU%MemUsed, ' * Calc Used:',TotalMem

  Write(Output,*) ' Overal Cache Statistics:'
  CacheHits = (LRU%Cache%Hits)
  CacheMisses = (LRU%Cache%Misses)
  CacheTotal = CacheHits + CacheMisses
  Write(Output,*) '    (Unweighted) Total:',CacheTotal, &
       ' * Hits:',CacheHits, ' * Misses:',CacheMisses
  total = CacheTotal
  If (total>0) then
     hits = (1.0*CacheHits)/total
     miss = (1.0*CacheMisses)/total
     Write(Output,*) '            %Hits:',Hits, ' * %Misses:',Miss
  End If


  hits = (LRU%Data_Size)
hits = 1

     miss = LRU%Data_Size/hits
     CacheHits = CacheHits + LRU%Cache%Hits * hits
     CacheMisses = CacheMisses + LRU%Cache%Misses * hits
     j = j / 8
  CacheTotal = CacheHits + CacheMisses
  Write(Output,*) '    (Weighted) Total:',CacheTotal, &
       ' * Hits:',CacheHits, ' * Misses:',CacheMisses
  IF (CacheTotal > 0) then
     total = CacheTotal
     hits = (1.0*CacheHits)/total
     miss = (1.0*CacheMisses)/total
     Write(Output,*) '            %Hits:',Hits, ' * %Misses:',Miss
  End If

  If (LRU%MemFree < 0) then
     Write(Error_Unit, *) 'LRU_DisplayBufferInfo: Memory Error!!! Memory Free:',LRU%MemFree, ' * ID:',TRim(LRU%Id)
  End If

  Write(Output,*) '*********************************************************'
  Call Flush(OutPut)

  Return
End Subroutine


!******************************************************************************
!
!  REAL_LRU_InitAlloc - Allocates the LRU data structures and 
!                    initializes them for use.
!
!   LRU   - LRU context
!
!******************************************************************************

Subroutine REAL_LRU_InitAlloc(LRU)
  Type (REAL_LRU_Context), Intent(INOUT) :: LRU

  Integer :: i, j, num
  Integer :: k, Offset, mem
  Character*100 :: msg

  Allocate(LRU%LRUStruc, LRU%Cache, STAT=i)
  Write(msg, *)'LRU_InitAlloc: Could Not allocate LRUStruc Array!', &
     TRIM(LRU%Id)
  Call Check_Error(i, msg, Error_Unit, .TRUE.,paw_wc, "LRU_InitAlloc:")

  LRU%Cache%Hits = 0
  LRU%Cache%Misses = 0
  LRU%Cache%Total = 0

     k = LRU%TotalRecs

     Allocate(LRU%LRUstruc%Buffer(k), LRU%LRUstruc%DiskMap(k), STAT=j)
     Write(msg,*) 'LRI_InitAlloc: Could Not allocate  Buffer array!',  &
           ' * ID=',Trim(LRU%Id)
     Call Check_Error(j, msg, Error_Unit, .TRUE.,paw_wc, "LRU_InitAlloc:")

     LRU%LRUstruc%DiskMap = 0 

     Do j=1, k
        LRU%LRUstruc%buffer(j)%Index = -1
        LRU%LRUstruc%Buffer(j)%Count = COUNT_START
     End Do

  Return
End Subroutine


!******************************************************************************
!
! REAL_LRU_InitContext - Initializes the LRU Context for use
!
!   LRU      - LRU context to initialize
!   Id       - Unique text Id for data
!   FD_Base  - Base Unit
!   MaxMem   - Max amount of memory to use for buffers
!   MaxRec   - Max data record
!   RecSize  - Array containing the number of data elem for each record
!
!******************************************************************************

Subroutine REAL_LRU_InitContext(LRU, id, FD_Base, MaxMem, MaxRec, &
	 DataSize, ReplaceOldest)
  Type (REAL_LRU_Context), Intent(INOUT) :: LRU
  Character*(*),      Intent(IN)    :: Id
  Integer,            Intent(IN)    :: FD_Base
  Real,               Intent(IN)    :: MaxMem
  Integer,            Intent(IN)    :: MaxRec  
  Integer,            Intent(IN)    :: DataSize
  Logical,            Intent(IN)    :: ReplaceOldest

  Integer :: i, err, RecSize
  Character*100 :: msg


  LRU%Id = Id
  LRU%FD_Base = FD_Base

  i = SizeOf_C
  LRU%MaxMem = (MaxMem*(2**20))/i
  LRU%MemUsed = 0
  LRU%MemFree = LRU%MaxMem
  LRU%TotalRecs = MaxRec
  LRU%ReplaceOldest = ReplaceOldest
  LRU%LastIndex = 1

  Allocate(LRU%Data_Size)
  LRU%Data_Size = DataSize
  
  Nullify(LRU%LRUstruc)

     RecSize = SizeOf_REAL*LRU%Data_Size

     Write(Log_Unit, *) 'LRU_InitContext: Unit=',FD_Base, &
           ' * Size=',LRU%data_size
     Open(FD_base, Form="UNFORMATTED", status="SCRATCH", &
          Access="DIRECT", RECL=REcSize, IOSTAT=err)
     Write(msg,*) ' Error opening file. Rec Size=',RecSize
     Msg = TRIM(LRU%Id) // ':' // Msg
     Call Check_Error(err, msg, Error_Unit, .TRUE., PAW_wc, &
	"REAL_LRU_InitContext:")


  Call REAL_LRU_InitAlloc(LRU)
  
  Return
End Subroutine


End Module
