!******************************************************************************
!
! File : relax.f90
!   by : Alan Tackett
!   on : 11/13/98
!  for : PAW
!
! Contains the routines for performing the different smoothing or relaxation
! operators.
!
!******************************************************************************

Module relax

Use atom_data
Use mem_data
Use projectors
Use psilib
Use hamfunc
Use projectors
Use options_data
Use work_mgr
Use debug
Use basis_lib
Use laplacian
!Use metric
Implicit NONE!!!!!!!!!!!!!

Integer, PARAMETER :: CORR_APPROX = 1    !** Approx the JD correction eq.
Integer, PARAMETER :: CORR_BICGSTABL = 2 !** Solve JD corr eq using BiCGSTABL

Type Prec_List
   Complex, Pointer :: InvMat(:,:)
   Integer          :: N
End Type

Type (Prec_List), Pointer :: PrecMatrix(:)

Type PrecVec_List
   Complex, Pointer :: V(:)
End Type

Type (PrecVec_List), Pointer :: PreVec(:)

Complex, Pointer          :: Prec_Diag(:)

Complex, Pointer :: JD_Work(:,:)  !** Kq and q orthog vectors for JD solver
Complex, Pointer :: JD_Matrix(:,:) !** qKz matrix for JD Solver
Complex, Pointer :: JD_f(:)        !** Work array needed for BiCGSTAB
Integer, Pointer :: JD_Pivot(:)    !** Pivot array for jd_matrix
Integer          :: JD_N        !** Number of JD vectors

!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
!  Prec_DiagProj - Calculates the Diagonal matrix in R-space times the 
!          specified projector.
!
!     index - PLM index of the projector
!     Diag  - Diag vector in R-space
!     DP    - Result of diag*Proj
!*RS*
!******************************************************************************

Subroutine Prec_DiagProj_RS(index, Diag, DP)
  Integer, Intent(IN)  :: index
  Complex, intent(IN)  :: Diag(:)
  Complex, intent(OUT) :: DP(:)

  Integer :: atom,  i,j,k, dn
  Complex, Pointer :: Proj(:)

  atom = PLM(1,Index)

  Proj => PreVec(index)%V
  DP = 0
  i = 0;  j = 0;
  Do k=1, RS_RLE(atom)%N
     j = RS_RLE(atom)%RLE(1,k)-1
     dn = RS_RLE(atom)%RLE(2,k)
     DP(j+1:j+dn) = Diag(j+1:j+dn)*Proj(i+1:i+dn)
     
     i = i + dn
  End Do

  Return
End Subroutine

!******************************************************************************
!
!  Prec_DiagProj - Calculates the Diagonal matrix in R-space times the 
!          specified projector.
!
!     index - PLM index of the projector
!     Diag  - Diag vector in R-space
!     DP    - Result of diag*Proj
!*TIME*
!******************************************************************************

Subroutine Prec_DiagProj_TIME(index, Diag, DP)
  Integer, Intent(IN)  :: index
  Complex, intent(IN)  :: Diag(:)
  Complex, intent(OUT) :: DP(:)
 
  Integer :: Kpnt, gmax, ng, G_half, G_size
  Integer :: Prange(2)
  Integer :: i,j,k, Base_DI,  atom, L, m, am, n, RadP, PLM_Index
  Complex, Pointer :: Proj(:), Work(:)
  Real :: c1, c3
  Complex :: Pos_M, Neg_M, c2, dot


!Complex :: PD(PLM_Max)

!PDOT = 0
!RETURN  !**REMOVE** Debug Only!!!
  
  atom = PLM(1,Index)

  Kpnt = Mem_Kpnt
  
  gmax = Gall_Size(G_LOW)

  Base_DI = PLM_Max*(Kpnt-1)  !** Calc starting disk index for Kpnt
  ng = Gpnt_Size(G_LOW) - 1

  Base_DI =  RS_PLMMax(atom)*(Mem_Kpnt-1)
  !G_Half = Atomtype_Info(Atom_List(atom)%typeIndex)%Gpnt_size
  !G_Size = 2*G_Half - 1
  G_Half = Gpnt_Size(G_LOW)
  G_Size = Gall_Size(G_LOW)

  Work => SCFvalues%Work
  Work = 0
  Work(FFTMap_PROJ(1:G_Half)) = PreVec(index)%V(1:G_Half)
  j = Gpnt_Size(G_PROJ)+1;   k = j + G_Half-2
  Work(FFTMap_PROJ(j:k)) = PreVec(index)%V(G_Half+1:)

  Call PerformFFT(FFT_TO_R, G_PROJ, Work)

  Work = Diag*Work

  DP = 0
  Call RS_toG(Work, DP)

!DP(1:G_Half) = PreVec(index)%V(1:G_Half)  !**QWERT
!DP(j:k) = PreVec(index)%V(G_Half+1:)     !QWERT
  Return
End Subroutine

!******************************************************************************
!
!  Prec_DiagProj - Calculates the Diagonal matrix in R-space times the 
!          specified projector.
!
!     index - PLM index of the projector
!     Diag  - Diag vector in R-space
!     DP    - Result of diag*Proj
!*MEM*
!******************************************************************************

Subroutine Prec_DiagProj_MEM(index, Diag, DP)
  Integer, Intent(IN)  :: index
  Complex, intent(IN)  :: Diag(:)
  Complex, intent(OUT) :: DP(:)

  STOP

  Return
End Subroutine



!******************************************************************************
!
!  Prec_DiagProj - Calculates the Diagonal matrix in R-space times the 
!          specified projector.
!
!     index - PLM index of the projector
!     Diag  - Diag vector in R-space
!     DP    - Result of diag*Proj
!******************************************************************************

Subroutine Prec_DiagProj(index, Diag, DP)
  Integer, Intent(IN)  :: index
  Complex, intent(IN)  :: Diag(:)
  Complex, intent(OUT) :: DP(:)
 
  Select Case (Proj_Mode)
     Case (MIN_TIME)
        Call Prec_diagProj_TIME(index, Diag, DP)
     Case (MIN_MEMORY)
        Call Prec_DiagProj_MEM(index, Diag, DP)
     Case (PROJ_RS)
        Call Prec_DiagProj_RS(Index, Diag, DP)
   End Select

  Return
End subroutine

!******************************************************************************
!
!  Prec_ProjDiagProj - Calculates the Proj*D*Proj in R-space times the 
!          specified projector.
!
!     index - PLM index of the projector
!     DP    - Diag*Proj
!*RS*
!******************************************************************************

Complex Function Prec_ProjDiagProj_RS(index, DP)
  Integer, Intent(IN)  :: index
  Complex, intent(IN) :: DP(:)

  Integer :: atom,  i,j,k, dn, Base_DI
  Complex, Pointer :: Proj(:)
  Complex :: dot

  atom = PLM(1,Index)

  Base_DI = RS_PLMMax(atom)*(Mem_Kpnt-1)  !** Calc starting DI for Kpnt
  i = index - PLM_AtomRange(1,atom)+1
  Call LRU_GetRec(LRU_RS_Proj(atom),  Base_DI + i, Proj)

  dot = 0
  i = 0;  j = 0;
  Do k=1, RS_RLE(atom)%N
     j = RS_RLE(atom)%RLE(1,k)-1
     dn = RS_RLE(atom)%RLE(2,k)
     dot = dot + SUM(Proj(i+1:i+dn)*DP(j+1:j+dn))
     
     i = i + dn
  End Do

  Prec_ProjDiagProj_RS = dot/FFT_Grid(4,G_PROJ)

  Return
end function

!******************************************************************************
!
!  Prec_ProjDiagProj - Calculates the Proj*D*Proj in R-space times the 
!          specified projector.
!
!     index - PLM index of the projector
!     DP    - Diag*Proj
!*MEM*
!******************************************************************************

Complex Function Prec_ProjDiagProj_MEM(index, DP)
  Integer, Intent(IN)  :: index
  Complex, intent(IN) :: DP(:)


  STOP

  Prec_ProjDiagProj_MEM = 0

  Return
end Function

!******************************************************************************
!
!  Prec_ProjDiagProj - Calculates the Proj*D*Proj in R-space times the 
!          specified projector.
!
!     index - PLM index of the projector
!     DP    - Diag*Proj
!*TIME*
!******************************************************************************

Complex Function Prec_ProjDiagProj_TIME(index, DP)
  Integer, Intent(IN)  :: index
  Complex, intent(IN) :: DP(:)

  Integer :: Kpnt, gmax, ng, G_half, G_size
  Integer :: Prange(2)
  Integer :: i,j,k, Base_DI,  atom, L, m, am, n, RadP, PLM_Index
  Complex, Pointer :: Proj(:), Work(:)
  Real :: c1, c3, dv
  Complex :: Pos_M, Neg_M, c2, dot
  Complex, EXTERNAL :: ZDOTC

  DV = 1.0/(1.0*FFT_Grid(4,G_PROJ))

!Complex :: PD(PLM_Max)

!PDOT = 0
!RETURN  !**REMOVE** Debug Only!!!
  
  atom = PLM(1,Index)

  Kpnt = Mem_Kpnt
  
  gmax = Gall_Size(G_LOW)

  Base_DI = PLM_Max*(Kpnt-1)  !** Calc starting disk index for Kpnt
  ng = Gpnt_Size(G_LOW) - 1

  Base_DI =  RS_PLMMax(atom)*(Mem_Kpnt-1)
  !G_Half = Atomtype_Info(Atom_List(atom)%typeIndex)%Gpnt_size
  !G_Size = 2*G_Half - 1

  G_Half = Gpnt_Size(G_LOW)
  G_Size = Gall_Size(G_LOW)

  RadP = RS_PLM(atom,1,n)
  n = Index - PLM_AtomRange(1,atom) + 1
  Call LRU_GetRec(LRU_Proj(atom),  Base_DI + n, Proj)

  
!  Allocate(Work(FFT_grid(4,G_PROJ)))
!  Work = 0
!  Work(FFTMap_PROJ(1:G_Half)) = Proj(1:G_Half)
!  j = Gpnt_Size(G_PROJ)+1;   k = j + G_Half-2
!  Work(FFTMap_PROJ(j:k)) = Proj(G_Half+1:)

!  Call PerformFFT(FFT_TO_R, G_PROJ, Work)

!  Prec_ProjDiagProj_TIME = DV*SUM(Work*DP)

!  DeAllocate(Work)

  Dot = SUM(Proj(1:G_Half)*DP(1:G_HAlf))
  j = Gpnt_Size(G_PROJ)+1;   k = j + G_Half-2
  Prec_ProjDiagProj_Time = dot + SUM(Proj(G_Half+1:)*DP(j:k))

  Return
end Function


!******************************************************************************
!
!  Prec_ProjDiagProj - Calculates the Proj*D*Proj in R-space times the 
!          specified projector.
!
!     index - PLM index of the projector
!     DP    - Diag*Proj
!******************************************************************************

Complex Function Prec_ProjDiagProj(index, DP)
  Integer, Intent(IN)  :: index
  Complex, intent(IN) :: DP(:)
 
  Select Case (Proj_Mode)
     Case (MIN_TIME)
        Prec_ProjDiagProj = Prec_ProjdiagProj_TIME(index, DP)
     Case (MIN_MEMORY)
        Prec_ProjDiagProj = Prec_ProjDiagProj_MEM(index, DP)
     Case (PROJ_RS)
        Prec_ProjDiagProj = Prec_ProjDiagProj_RS(Index, DP)
   End Select

  Return
End Function

!******************************************************************************
!
!  MakePrecondVects - Makes the prconditioned vectors
!
!******************************************************************************

Subroutine MakePrecondVects(Energy)
  Real, Intent(IN) :: Energy

  Integer :: nili,li,mi, njlj,lj,mj, i, j, atom, Base_DI,  n
  Complex, Pointer :: Proj1(:), Proj2(:), V(:)
  Complex :: term
  Real    :: dot


  Do atom=1, Specific_Atoms
     Base_DI =  RS_PLMMax(atom)*(Mem_Kpnt-1)

     Do i=PLM_AtomRange(1,atom), PLM_AtomRange(2,Atom)
        nili = PLM(2,i); li = PLM(3,i);  mi = PLM(4,i);
!        n = i - PLM_AtomRange(1,atom)+1
!        Call LRU_GetRec(LRU_Proj(atom),  Base_DI + n, Proj1)

        V => PreVec(i)%V
        V = 0
        Do j=PLM_AtomRange(1,atom), PLM_AtomRange(2,Atom)
           njlj = PLM(2,j); lj = PLM(3,j);  mj = PLM(4,j);
           term = DijMatrix(atom,njlj,nili,lj,li,mj,mi) -  &
                Energy*OijMatrix(atom,njlj,nili,lj,li,mj,mi)
           term = CONJG(term)
           dot = term*CONJG(term)
           if (abs(dot)>1E-20) then
!write(Log_Unit,*) 'MPV: ij=',i,j,' * term=',term
              n = j - PLM_AtomRange(1,atom)+1
              If (Proj_Mode == PROJ_RS) then
                 Call LRU_GetRec(LRU_RS_Proj(atom),  Base_DI + n, Proj2)
              else
                 Call LRU_GetRec(LRU_Proj(atom),  Base_DI + n, Proj2)
              End If
              V = V + CONJG(term)*Proj2
           End if
        End do

        V = CONJG(V)
     End do
  End Do

  Return
End Subroutine


!******************************************************************************
!
!  KE_PreCond - Constructs the KE preconditioner
!
!  Diag - KE preconditioner returned
!  KE   - Ke for band
!  BlochK - K-point
!
!******************************************************************************

Subroutine KE_Precond(Diag, KE, BlochK)
  Complex, Intent(OUT) :: Diag(:)
  Real,    Intent(IN)  :: KE
  Real,    Intent(IN)  :: BlochK(:)

  Integer :: gi, N
  Real    :: G(3), G2, x, poly

  N = Gpnt_Size(G_LOW) - 1

  Do gi=1, Gpnt_Size(G_LOW)          
     G = BlochK + Gpnt(1:3,gi)
     G2 = DOT_PRODUCT(G,G)
     
     x = G2/KE
     poly = 8*x**3 + 12*x**2 + 18*x + 27
     Diag(gi) = poly/(poly + 16*x**4)
  End Do
        
  Do gi=2, Gpnt_Size(G_LOW)
     G = BlochK - Gpnt(1:3,gi)
     G2 = DOT_PRODUCT(G,G)
     
     x = G2/KE
     poly = 8*x**3 + 12*x**2 + 18*x + 27
     Diag(gi+N) = poly/(poly + 16*x**4)
  End Do

  Return
End Subroutine

!******************************************************************************
!
!  CalcCorr - Calculates a correction to the Psi's  using a 
!       Block-Davidson step
!
!***NEW Process a single vec***
!******************************************************************************

Subroutine CalcCorr( Psi, Hx, Ox, PsiIndex, Kpnt, CorrType)
  Complex,          Intent(IN) :: Psi(:)
  Complex,          Intent(IN) :: Hx(:)
  Complex,          Intent(IN) :: Ox(:)
  Integer,          Intent(IN) :: PsiIndex
  Integer,          Intent(IN) :: Kpnt
  integer,          Intent(IN) :: CorrType

  Complex, Pointer :: Diag(:), Work(:), Work2(:), OWork2(:)
  Real :: Energy, BlochK(3), G(3), G2, KE, deps
  Complex  :: dot, zalpha, zbeta
  Integer :: N, BG_Index, gi, L, mxmv
  Logical :: PDOT_Stored
  Complex :: PDOT(PLM_Max)


  Call Start_Timer(Timer(CalcCorr_Timer))
  N = Gpnt_Size(G_LOW) - 1

  Call GetBuffer( Diag)
  Call GetBuffer( Work)

  BlochK = BZ%Ku(:,Kpnt)

  Call GetNextFreeIndex(BG_Index)
  PsiInfo(BG_Index)%Kpnt = BG_KPNT !special index for temp vectors 
  PsiInfo(PsiIndex)%CorrIndex = BG_Index

  Energy = PSiInfo(PsiIndex)%Energy

  Select Case (CorrType)
     Case (CORR_APPROX)
        KE = CalcKineticenergy(BZ%Ku(1:3,Kpnt), Psi)
        Call KE_Precond(Diag, KE, BlochK)
        Work = Diag*(Hx - Energy*Ox)


!*3*-------   ** Working Scheme **
        If (Eigen_Mode == EIGEN_DAVIDSON) then
           Diag = Diag*Psi
           dot = Basis_DotProd( Ox, Work, .FALSE.)
           dot = dot / Basis_DotProd( Ox, Diag, .FALSE.)
           Work = dot*Diag - Work
        End If
!--------
  End Select

   call kplusGfilter(Work,Kpnt)
  Call SavePsi_toBuffer( PSiInfo(PsiIndex)%CorrIndex, Work)

 !write(Log_Unit,*)'in JD_C -- psiindex,corrindex', psiindex,&
 !psiinfo(psiindex)%corrindex,psiinfo(psiinfo(psiindex)%corrindex)%Available,&
 !psiinfo(psiinfo(psiindex)%corrindex)%membufindex
  

  Call FreeBuffer(Work)
  Call FreeBuffer(Diag)

  Call Stop_Timer(Timer(CalcCorr_Timer))  
  Return
End Subroutine


!******************************************************************************
!
!  JD_CalcCorr_Partition - Calculates a correction to the Psi's  using a 
!       Block-Davidson step for the partition energy scheme
!
!***NEW Process a single vec***
!******************************************************************************

Subroutine JD_CalcCorr_Partition( Psi, Hx, Ox, PsiIndex, Kpnt, Energy)
  Complex,          Intent(IN) :: Psi(:)
  Complex,          Intent(IN) :: Hx(:)
  Complex,          Intent(IN) :: Ox(:)
  Integer,          Intent(IN) :: PsiIndex
  Integer,          Intent(IN) :: Kpnt
  Real,             Intent(IN) :: Energy

  Complex, Pointer :: Diag(:), Work(:), W2(:), DiagPsi(:)
 !Complex :: PDOT_HOiHx(PLM_Max), dot
  Complex ::  dot
  Complex ,allocatable :: PDOT_HOiHx(:)
  Logical :: PDOT_Stored
  Real :: BlochK(3), G(3), G2, x, poly, KE
!*  Complex  :: KE
  Integer :: N, BG_Index, gi


  Allocate(PDOT_HOiHx(PLM_Max))
  N = Gpnt_Size(G_LOW) - 1

  Call GetBuffer( DiagPsi)
  Call GetBuffer( Diag)
  Call GetBuffer( Work)

  BlochK = BZ%Ku(:,Kpnt)

  Call GetNextFreeIndex(BG_Index)
  PsiInfo(BG_Index)%Kpnt = BG_KPNT
  PsiInfo(BG_Index)%DoSave = 1      
  PsiInfo(BG_Index)%Available=Mem_Used
  PsiInfo(PsiIndex)%CorrIndex = BG_Index

  KE = CalcKineticenergy(BZ%Ku(1:3,Kpnt), Psi)

  Call KE_Precond(Diag, KE, BlochK)
   
  Work = Diag*(Hx - Energy*Ox)
  DiagPsi = Diag*Psi

  PDOT_Stored = .FALSE.
  Diag = 0
  Call OPsi( Work, Diag, PDOT_HoiHx, Kpnt, PDOT_Stored)
  Work = Diag

  PDOT_Stored = .FALSE.
  Diag = 0
  Call OPsi( DiagPsi, Diag, PDOT_HoiHx, Kpnt, PDOT_Stored)
  DIagPsi = Diag



  KE = CalcKineticenergy(BZ%Ku(1:3,Kpnt), Work)
  Call KE_Precond(Diag, KE, BlochK)
  Work = Diag*Work             !*Old Working scheme
!**  Work = Diag*Work + Energy*Work
  DiagPsi = Diag*DiagPsi

  

!*3*-------
!**  Diag = Diag*Psi
  dot = Basis_DotProd( Ox, Work, .FALSE.)
  dot = dot / Basis_DotProd( Ox, DiagPsi, .FALSE.)
  Work = dot*DiagPsi - Work
!--------

  Call SavePsi_toBuffer( PSiInfo(PsiIndex)%CorrIndex, Work)
  

  Call FreeBuffer(Work)
  Call FreeBuffer(Diag)
  Call FreeBuffer(DiagPsi)
  DeAllocate(PDOT_HOiHx)
  
  Return
End Subroutine

!******************************************************************************
!
!  AllocPrecVec - Allocatesthe precondition vectors
!
!******************************************************************************

Subroutine AllocPrecVec
  Integer :: i,j,atom

  Do i=1, PLM_Max
     atom = PLM(1,i)
     If (Proj_Mode == PROJ_RS) then
        j =LRU_RS_Proj(atom)%Data_Size
     else
        !j = 2*Atomtype_Info(Atom_List(atom)%typeIndex)%Gpnt_size - 1
        j = Gall_Size(G_LOW)
     End If

     If (JD_N>0) DeAllocate(PreVec(i)%V)
     Allocate(PreVec(i)%V(j))
  End do

  Return
End Subroutine

!******************************************************************************
!
!  Init_relax - Initializes the relax routines for use
!
!******************************************************************************

!!!Subroutine Init_relax
!!!  Integer :: i, j, atom
!!!
!!!  If (Atomic_Mode == SIM_SEPM) then
!!!     Allocate(Prec_Diag(FFT_Grid(4,G_PROJ)))
!!!  else
!!!     Allocate(PrecMatrix(Specific_Atoms), PreVec(PLM_Max), &
!!!          Prec_Diag(FFT_Grid(4,G_PROJ)))
!!!
!!!     Do i=1, Specific_Atoms
!!!        j = PLM_AtomRange(2,i) - PLM_AtomRange(1,i)+1
!!!        PrecMatrix(i)%N = j
!!!        Allocate(PrecMatrix(i)%InvMat(j,j))
!!!     End Do
!!!     
!!!     JD_N = -1
!!!
!!!     Call AllocPrecVec
!!!  End If
!!!
!!!  Call CalcGij(xtal%Basis)
!!!
!!!  Return
!!!end subroutine

End Module
