!******************************************************************************
!
! File : relaxsys.f90
!   by : Alan Tackett
!   on : 7/27/98
!  for : PAW project
!
!  This module contains the routines for relaxing the electrons and 
!  also the atoms.  
!
!  Updated for spin by Ping Tang and N. A. W. Holzwarth
!    Last change 5/23/05
!******************************************************************************

Module relaxsys

Use paw_inout
Use misc
Use options_data
Use word
Use atom_data
Use crystal_data
Use crystal_symmetry
Use psilib
Use solver
Use mem_data
Use memmgr
Use forces
Use work_mgr
Use storedata
!TD Use timedependent
Use anderson_mixing
Use gradenk
Use hamiltonian

Implicit NONE

Integer, PARAMETER :: RELAX_MD       = 0
Integer, PARAMETER :: RELAX_GEOMETRY = 1

!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
!  CheckAtomOverlap - Checks to make sure that none of the atomic spheres
!      overlap
!
!******************************************************************************

Subroutine CheckAtomOverlap
  Integer :: i, j, k, a1, a2
  Real    :: DX(3), R(3), Rc, R0(3), T(3), Rct, r2

  Do a1=1, Specific_Atoms
     Rc = AtomType_Info(Atom_List(a1)%TypeIndex)%Rc
     R0 = Atom_List(a1)%Pos

     Do i=-1, 1
        Do j=-1, 1
           Do k=-1, 1
              T = MatMul(xtal%Basis, (/i,j,k/))
              Do a2=a1, Specific_Atoms  
                If (Abs(i)+Abs(j)+Abs(k)+a2-a1>0) then
                 Rct = Rc + AtomType_Info(Atom_List(a2)%TypeIndex)%Rc
                 R = R0 - (Atom_List(a2)%Pos + T)
                 r2 = SQRT(DOT_PRODUCT(R, R))
                 If (r2 < RcT) then
                    Write(Error_Unit,*) 'CAO: A1=',&
                         TRIM(Atom_List(a1)%Name), ' overlaps ', &
                         TRIM(Atom_List(a2)%Name), ' * Rc=',Rc,  &
                         ' * RcT=',RcT,' * R=', &
                         R2, ' * ijk=',i,j,k
                 End If
               Endif
              End Do
           End Do
        End Do
     End Do
  End Do

  Return
End Subroutine


!******************************************************************************
!
!  UpdateSystem - Performs system update each time the atoms move.
!
!******************************************************************************

Subroutine UpdateSystem

  Integer :: i,j,Lmax

  If (xtal%Auto_Symmetry) then
      !******* Initialize the Crystal_Symmetry routines **********
     Lmax = -1
     Do i=1, Atom_Types
        j = MaxVal(AtomType_Info(i)%L_Value)
        Lmax = MAX(Lmax, j)
     End Do
     
     Call StoreSymmetry('RUNTIME')
     If (xtal%Rot_Size > 0) Call SetSym(Lmax)
  End If


  Call LRU_FreeAll(LRU_Phase)
  Call StoreStructFactors
  Call CalcVlocal(SCFvalues%V_local)
  Call Updatecoretail(SCFvalues%coretail)

  If (Run_Mode == MIN_TIME) then
     Do i=1, Specific_Atoms
        Call LRU_FreeAll(LRU_Proj(i))
        Close(LRU_Proj(i)%FD_Base)
     End Do
     Call StoreRawProj
  end If

  If (Proj_Mode == PROJ_RS)  Call RS_Update

  !Call AllocPrecVec  !** Allocate space for the JDQZ preconditioner --inactive

  Return
End Subroutine


!******************************************************************************
!
!  UpdateKpoint - Updates the system set of K-points for band structure
!                 calculations
!
!    NewK - New list of kpoints should be the same size as BZ%Ku.
!
!******************************************************************************

Subroutine UpdateKpoint(NewK)
   Real, Intent(IN) :: NewK(:)

   Integer :: i

   BZ%Ku(:,1) = NewK  !** Assign New Kpoint set
   i=1
   call InitkplusGfilter(i)


   PsiInfo(:)%PDOT_Stored = .FALSE.
   PsiInfo(:)%KE_Stored = .FALSE.
 
Write(Log_Unit,*) 'UpadteKpoint K1=',NewK(:) ; call flush(Log_unit)
   Do i=1, Atom_Types
      Call LRU_FreeAll(LRU_RadProj(i))
      Close(LRU_RadProj(i)%FD_Base)
   End Do

   Call StoreRadProj

   Call LRU_FreeAll(LRU_Ylm)
   Call StoreYlm

   Call UpdateSystem
RETURN

!   !*** Purge the Buffers ***
!   Do i=1, Atom_Types
!      Call LRU_FreeAll(LRU_RadProj(i))
!      Close(LRU_RadProj(i)%FD_Base)
!   end Do
!
!   Call LRU_FreeAll(LRU_Phase)
!   Call LRU_FreeAll(LRU_Ylm)
!
!   Do i=1, Specific_Atoms
!      Call LRU_FreeAll(LRU_Proj(i))
!      Close(LRU_Proj(i)%FD_Base)
!   End Do

!   Call StoreRadProj         !** Update radial projectors
!   Call StoreStructFactors   !** Update Structure Factors
!   Call StoreYlm             !** Update the Ylm's

!   If (Run_Mode == MIN_TIME) Call StoreRawProj

End Subroutine

!******************************************************************************
!
!  PrintOccAndEig - Prints to the Log_Unit the Old, new, and diff 
!                   occupancies, eigenvalues, and fermi energy
!
!  OldOcc   - Old occupancies
!  OldEn    - Old energy Eigenvalues
!  OldFermi - Old Fermi energy level
!
!******************************************************************************

Subroutine PrintOccAndEig(OldOcc, OldEn, OldFermi)
  Real, Intent(IN) :: OldOcc(:,:)
  Real, Intent(IN) :: OldEn(:,:)
  Real, Intent(IN) :: OldFermi

  Integer :: i,j,k
  Real    :: docc, den, df
  Character*100 :: Msg

  Write(Log_Unit, *) '----------------------------------------------'
  Write(Log_Unit, *) 'New Fermi:',BZ%Fermi, ' * Old Fermi:',OldFermi, &
                     'Diff:',BZ%Fermi-OldFermi

  msg = '(2i5," ",3e15.7)'
  Write(Log_Unit, *) 'Occupancy-----------------------------------------------'
  Write(Log_Unit,*) 'k-pnt  Band     NewOcc         OldOcc        DiffOcc'
  Do k=1, NumKpnts
     Do i=1, NumBands
        Write(Log_Unit, msg) k,i, Occupancy(i,k), OldOcc(i,k), &
             (Occupancy(i,k)-OldOcc(i,k))
     End Do
  End Do

  msg = '(2i5," ",3e15.7)'
  Write(Log_Unit, *) 'Energy Levels------------------------------------------'
  Write(Log_Unit,*) 'k-pnt  Band     NewEig         OldEig        DiffEig'
  Do k=1, NumKpnts
     Do i=1, NumBands
        Write(Log_Unit,msg) k, i, BZ%Keu(i,k), OldEn(i,k), &
             (BZ%Keu(i,k)-OldEn(i,k))
     End Do
  End Do

  Write(Log_Unit, *) '-----------------------------------------------------'
  Return
End Subroutine

!******************************************************************************
!
!  RelaxElectrons - Relax's the wave functions until equilibrium is met or 
!                   the maximum number of iterations is reached.
!
!  MaxIter - Max number of SCF iterations
!  Tol     - Error tolerance for the Energy
!
!******************************************************************************

Subroutine RelaxElectrons(MaxIter, Tol, Etol, Converged)
  Integer,       Intent(IN) :: MaxIter
  Real,          Intent(IN) :: Tol
  Real,          Intent(OUT) :: Etol
  Logical,       Intent(OUT)       :: Converged

  Integer       :: Iter, i,j, Lmax
  Character*100 :: msg
  Real          :: Countup, Countdn

  If (.not.bandstructure_mode) then
  Call Printdate(Output_Unit,'#  Starting SCF on date ')
  Write(Output_Unit,*) ' '
  Write(Output_Unit,*) '# Current Cartesian Positions in Bohr units'
  Write(Output_Unit,*) '#Atom_List CART_POSITION'
  Do i=1, Specific_Atoms
     j=Atom_List(i)%TypeIndex
     If (Atom_List(i)%Freeze) then
        write(Output_Unit,'(" # ", A8,"  ",A8, 1p3e15.7, " # FROZEN")') &
             Atom_List(i)%Name, &
             AtomType_Info(j)%Atom_Name, &
             Atom_List(i)%Pos
     Else If (Anchor.and.i==AnchorIndex) then
        write(Output_Unit,'(" # ", A8,"  ",A8, 1p3e15.7, " # ANCHORED")') &
             Atom_List(i)%Name, &
             AtomType_Info(j)%Atom_Name, &
             Atom_List(i)%Pos
     else
        write(Output_Unit,'(" # ", A8,"  ",A8, 1p3e15.7)') &
             Atom_List(i)%Name, &
             AtomType_Info(j)%Atom_Name, &
             Atom_List(i)%Pos
     EndIf
   Enddo
   Call Flush(Output_Unit)
  EndIf

  Call CheckAtomOverlap
  CohesiveEnergy=1E30  !***REMOVE******
  Prev_CohesiveEnergy = 1E30

  Etol = Tol
  write(log_unit,*) 'RelaxElectrons:  Calling SCF_Solver '
  call flush(log_unit)
  Call SCF_Solver( MaxIter, ETol, iter)

  Converged = .FALSE.
  If (ETol < Tol) Converged = .TRUE.

  Call Printdate(Output_Unit,'#  SCF results generated on date ')

  If (Converged) then
     Write(Log_Unit, *) 'RelaxElectrons: Converged in ',iter, &
          'iterations with a final Energy error of ', Etol
     Write(Output_Unit, *) 'RelaxElectrons: Converged in ',iter, &
          'iterations with a final Energy error of ', Etol
  else
     Write(Log_Unit, *) 'RelaxElectrons: Did Not converged in ',iter, &
          'iterations with a final energy error of ', Etol
     Write(Output_Unit, *) 'RelaxElectrons: Did Not converged in ',iter, &
          'iterations with a final energy error of ', Etol
  End If

  Write(Output_Unit,*) 'RelaxElectrons: Cohesive Energy:',CohesiveEnergy
  Write(Output_Unit,*) 'RelaxElectrons: # iterations   :',iter ,'/',MaxIter
  Write(Output_Unit,*) 'RelaxElectrons: Tolerance      :',Etol ,'/',Tol
  Write(Output_Unit,*) ' '
  If (spindependence) then
    Call DisplayEnergy(Output_Unit,Countup,Countdn)
    Write(Log_Unit,*) 'RelaxElectrons:  Nup+Ndn = ', Countup+Countdn,' Nup-Ndn = ', Countup-Countdn
    Write(Output_Unit,*) 'RelaxElectrons:  Nup+Ndn = ', Countup+Countdn,' Nup-Ndn = ', Countup-Countdn
  Else
    Call DisplayEnergy(Output_Unit,Countup)
    Write(Log_Unit,*) 'RelaxElectrons:  Ntot = ', Countup
    Write(Output_Unit,*) 'RelaxElectrons:  Ntot = ', Countup
  Endif
                                                                                               
  Call Flush(Output_Unit)

  If (.NOT. Converged) then
     Write(Msg, *) 'RelaxElectrons: Convergence NOT satisfied but exiting!'
     Write(Error_Unit, *) Msg
     IF (Error_Unit /= Log_Unit) Write(Log_Unit, *) Msg
     If ((Error_Unit /= Output_Unit) .AND. (Log_Unit /= Output_Unit)) then
        Write(Error_Unit, *) Msg
     End IF
  End If

  Return
End Subroutine


!******************************************************************************
!
!   MakeGeometryMixing - Initializes the Geometry relaxation data structure
!
!******************************************************************************

Subroutine MakeGeometryMixing
  Integer :: i, j, err, s,t
  Character*100 :: Msg
  
  j = 0  !*** First determine how many atoms are allowed to move
  Do i=1, Specific_Atoms
     If (.NOT. Atom_List(i)%Freeze) j = j + 1
  End Do

  If (Associated(Move_AC)) then
     Call FreeAnderson(Move_AC)
  End If

  i = 3*j
  Call InitAnderson(Move_AC, Error_Unit, Mix_Size, i, 1E-5, &
                Geometry_NewMix, AndersonConditionNo)

  Return
End Subroutine

!******************************************************************************
!
!  GeometryMixingStep - Updates the atom positions
!
!******************************************************************************

Subroutine GeometryMixingStep
  Integer :: i, j, k
  Complex :: OldPos(3*Specific_Atoms), DiffPos(3*Specific_Atoms)
  Real :: mag
  
  j = 1
  Do i=1, Specific_Atoms
    If (.NOT. Atom_List(i)%Freeze) then
       OldPos(j:j+2) = Atom_List(i)%Pos
       Atom_List(i)%Force%Pos(:,1) = Atom_List(i)%Force%Pos(:,1) + &
            Atom_List(i)%Force%CurrForce
       DiffPos(j:j+2) = Atom_List(i)%Force%Pos(:,1)  !** New Position

       j = j + 3
    End If
  End Do

  j = j - 1
  DiffPos(1:j) = DiffPos(1:j) - OldPos(1:j)
  Call Anderson_Mix(Move_AC, OldPos(1:j), DiffPos(1:j))

  j = 1
  Do i=1, Specific_Atoms
    If (.NOT. Atom_List(i)%Freeze) then
       Atom_List(i)%Force%Pos(:,1) = OldPos(j:j+2)  !** Now OldPos=MixedPos

       j = j + 3
    End If
  End Do

  Return
End Subroutine


!******************************************************************************
!
! MoveAtom - Calculates the new atom position given the Verlet data 
!
! Force - Verlet data structure containg the pos, Vel, and acc information 
!         for the atom
! Mass  - Atoms mass
! Friction - Friction coeff
!
!******************************************************************************

Subroutine MoveAtom(Force, mass, Friction)
  Type (Verlet_Data), Intent(INOUT) :: Force
  Real,       Intent(IN)    :: Mass
  Real,       Intent(IN)    :: Friction

  Integer :: i, j
  Real :: DT, DT2

  DT = TimeStep
  DT2 = DT*DT

  j = Force%NumMoves    !*** Copy R,V, and A to make room for new data
  If (j>2) j = 2
  Do i=j, 1, -1
     Force%Pos(:,i+1) = Force%Pos(:,i)
     Force%Velocity(:,i+1) = Force%Velocity(:,i)
     Force%Acceleration(:,i+1) = Force%Acceleration(:,i)
  End Do

  Force%Acceleration(:,1) = Force%CurrForce/Mass - Friction*Force%Velocity(:,2)

  Where (ABS(Force%Acceleration) < Force_Zero) Force%Acceleration = 0

  If (Force%NumMoves < 3) then  !*** First few steps are Euler steps
     Force%Pos(:,1) = Force%Pos(:,2) + Force%Velocity(:,2)*DT + &
          0.5*Force%Acceleration(:,1)*DT2
     Force%Velocity(:,1) = Force%Velocity(:,2) + Force%Acceleration(:,1)*DT
     Force%NumMoves = Force%NumMoves + 1
     Force%Predictor_Phase = .TRUE.

  else           !**** Can now take Beeman-Verlet steps *****
     Force%Pos(:,1) = 2*Force%Pos(:,2) - Force%Pos(:,3) + &
          Force%Acceleration(:,1)*DT2
     Force%Velocity(:,1) = (Force%Pos(:,1) - Force%Pos(:,2))/DT

!     Force%Velocity(:,1) = Force%Pos(:,1) - Force%Pos(:,2) + &
!          (2*Force%Acceleration(:,1)+Force%Acceleration(:,2))*DT2/6.0
!     Force%Velocity(:,1) = Force%Velocity(:,1)/DT
!     
!     Force%Pos(:,1) = Force%Pos(:,2) + DT*Force%Velocity(:,1) + &
!          (4*Force%Acceleration(:,2)-Force%Acceleration(:,3))*DT2/6.0
  End If

  Return
End Subroutine


!******************************************************************************
!
!  MoveAllAtoms - Moves the atoms using the standard Beeman-Verlet method
!
!  Mode - Type of move to perform, ie, MD or Broyden step for geometry
!
!******************************************************************************

Subroutine MoveAllAtoms(Mode)
  Integer, Intent(IN) :: Mode

  Real :: CurrForce(3), FrictionForce(3)
  Real :: Inpos(3,specific_atoms),Outpos(3,specific_atoms)
  Integer :: i, j

  Write(Log_Unit,*) 'MoveAllAtoms:  Printing Updated Atom Positions and Forces'

  !If (Mode == RELAX_GEOMETRY) Call GeometryMixingStep -- now done in relaxatoms

  Do i=1, Specific_Atoms 
           !** Calculate the new position *
     If (Mode == RELAX_MD) Atom_List(i)%Force%Pos(:,1) = Atom_List(i)%Pos 

     Write(Log_Unit, &
          '("MoveAllAtoms: DT = ",1e10.2, " * Friction = ", 1e10.2, " * Mass = ",1e10.2)') &
          TimeStep, Atom_List(i)%Friction, &
          Atomtype_Info(Atom_List(i)%TypeIndex)%Mass

!Write(Log_Unit,*) 'MoveAllAtoms: DT=',TimeStep, ' * Friction=',&
!     Atom_List(i)%Friction, ' * Mass=', &
!     Atomtype_Info(Atom_List(i)%TypeIndex)%Mass

     Write(Log_Unit,'("MoveAllAtoms: ", A10," R(old)  =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Pos
     Write(Log_Unit,'("MoveAllAtoms: ", A10," F(old)  =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Force%CurrForce
     Write(Log_Unit,'("MoveAllAtoms: ", A10," V(old)  =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Force%Velocity(:,1)


     If ((Mode == RELAX_MD) .AND. (.NOT. Atom_List(i)%Freeze)) then
        Call MoveAtom(Atom_List(i)%Force, &
             AtomType_Info(Atom_List(i)%TypeIndex)%Mass, Atom_List(i)%Friction)
     End If

     Atom_List(i)%Pos = Atom_List(i)%Force%Pos(:,1)  !*** Copy it 
     Atom_List(i)%Frac_Pos = MATMUL(xtal%InverseBasis, Atom_List(i)%Pos)  

     Write(Log_Unit,'("MoveAllAtoms: ", A10," R(new)  =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Pos     
     Write(Log_Unit,'("MoveAllAtoms: ", A10," R(xtal) =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Frac_Pos     
  End Do

  Call Flush(Log_Unit)


 If (Xtal%Rot_Size > 1) then
    Do i=1,Specific_Atoms
       Inpos(:,i)=Atom_List(i)%Frac_Pos
    Enddo
    Call Symmetrize_Pos(Inpos,Outpos)
    Do i=1,Specific_Atoms
       Atom_List(i)%Frac_Pos=Outpos(:,i)
       Atom_List(i)%Pos=MATMUL(Xtal%Basis,Atom_List(i)%Frac_Pos)
    Enddo
 EndIf

  Return
End Subroutine

!******************************************************************************
!
!  RelaxGeometry - Moves the atoms in the direction of the forces on them  
!
!  ForceTol, MaxMove -- determine the scale of the move
!  ConditionNo  -- determines parameters for FitMove
!
!  This version does not use constraints: "anchored atoms" just do not move
!
!******************************************************************************
Subroutine RelaxGeometry(ForceTol, MaxMove, ConditionNo)
  Real, Intent(IN) :: ForceTol, MaxMove, ConditionNo

  Real :: CurrForce(3), scale,R(3),F(3),energy,Efit,vec(3)
  Real :: Inpos(3,specific_atoms),Outpos(3,specific_atoms)
  Real, allocatable :: Edata(:),Rdata(:,:),Fdata(:,:),Rfit(:)
  Integer, allocatable :: Map(:)
  Integer :: i, j, k, a,ier,index, nmove,ndim , Nfreedom , p, n, many
  character*4 :: label
  character*9 :: moveoption
  logical :: MatrixMethod,Success


  Write(Log_Unit,*) 'RelaxGeometry: Printing Updated Atom Positions and Forces'

! Read Name_Pos file 
! If there are more than 2 entries ,  call FitMove
  
   Success=.FALSE.

  open(unit=Positions_Unit,file=Trim(Pos_Name),form='formatted', &
          action='read',iostat=ier)
  MatrixMethod=.false.
  index=0
  nmove=0

  If (ier/=0) then
     Write(Error_Unit,*)' Error in RelaxGeometry -- no force/position file',&
             Trim(Pos_Name)
     Stop
  EndIf

    Do 
       read(Positions_Unit,'(a4)',iostat=ier) label
       If (ier/=0) exit
       !write(Log_Unit,*) label
       If (label=="STEP") then
          index=index+1
          read(Positions_Unit,*,iostat=ier)  ! energy
          Do i=1,Specific_Atoms
             read(Positions_Unit,'(68x,a9)',iostat=ier) moveoption  ! R
       !      write(Log_Unit,*) index,i,moveoption
             read(Positions_Unit,*,iostat=ier) ! F
             If (index==1.AND.trim(moveoption)=="") nmove=nmove+1
          Enddo
       EndIf
       !write(Log_Unit,*) index,nmove
    Enddo

  Write(Log_Unit,*) ' RelaxGeometry: index,nmove = ', index,nmove
             
  If (index>2) then
     MatrixMethod=.true.
     If (.not.UseShells) Nfreedom=3*nmove   ! this will change for "shell" mode
     If (UseShells) Nfreedom=Shell%NPARAMS
     write(Log_Unit,*) 'RelaxGeometry: Nfreedom =', Nfreedom 
     Allocate(Edata(index),Rdata(Nfreedom,index),Fdata(Nfreedom,index),&
         Map(Nfreedom),Rfit(Nfreedom))
     Rewind(Positions_Unit)
     Do i=1,index
        Read(Positions_Unit,*)
        Read(Positions_Unit,'(30x,1pe20.10)') energy
         Edata(i)=-energy
         Rdata(:,i)=0
         Fdata(:,i)=0
         Map(:) = 0
          ! write(Log_Unit,*) 'Energy',Energy
           k=0
           Do  a=1,Specific_Atoms
             read(Positions_Unit,'(14x,3e18.10,a9)') R(:),moveoption
          !   write(Log_Unit,*) R(:),moveoption
             read(Positions_Unit,'(14x,3e18.10)') F(:)
          !   write(Log_Unit,*) F(:)
             If (.not.UseShells) then
              If (trim(moveoption)=="") then
               Rdata(k+1:k+3,i)=R(:)
               Fdata(k+1:k+3,i)=F(:)
               Map(k+1:k+3)=a
               k=k+3
              EndIf  
             EndIf
             If (UseShells) then
                many=Shell%AtomMap(a)%MapParams
                If (many>0) then
                   Do n=1,many
                      p=Shell%AtomMap(a)%Map(n)
                      Map(p)=Map(p)+1
                      vec=Shell%AtomMap(a)%V(:,n)
                      Rdata(p,i)=Rdata(p,i) + &
                         Dot_Product(vec,(R-Shell%AtomMap(a)%Origin))
                      Fdata(p,i)=Fdata(p,i) + Dot_Product(vec,F)
                   Enddo
                 EndIf
             EndIF

           Enddo   !a

           If (UseShells) then
              Do p=1,Shell%NPARAMS
                 Rdata(p,i)=Rdata(p,i)/Map(p)
              EndDo
           EndIf

     Enddo

     Close(Positions_Unit)

      Call LSmove(Index,Nfreedom,Edata,Fdata,Rdata,Efit,Rfit,  &
           ForceTol,MaxMove,ConditionNo,Success)
 
      Write (Log_Unit,*) ' Completed LSMove with Success', success
      Write(Log_Unit,'(1p3e15.7)') (Rfit(i),i=1,nfreedom)
      Write(Log_Unit,*)'RelaxGeometry: E0 =', Efit
    
   If (success) then
   Do a=1,Specific_atoms
      Inpos(:,a) = Atom_List(a)%Pos
      Outpos(:,a) = Atom_List(a)%Pos
   Enddo

   If (UseShells) then
      Do j=1,Nfreedom
         Shell%C(j)=Rfit(j)
      Enddo
      Call ShellPos
      Do a=1,Specific_atoms
         Outpos(:,a)=Atom_List(a)%Pos
         Atom_List(a)%Pos=Inpos(:,a)
      Enddo
    Else
      Do j=1,Nfreedom,3
         a = Map(j)
         Outpos(:,a) = Rfit(j:j+2)
      Enddo

   EndIf

   Write(Log_Unit,*) &
         'RelaxGeometry RESULTS -  New                          Dif'
   Do a=1,Specific_atoms
     Write(Log_Unit,'(i5,2x,3f8.3,2x,3f8.3)') & 
               a,Outpos(:,a),InPos(:,a)-OutPos(:,a)
   Enddo
  

   EndIf
  EndIf
 
  If (Success) then
  Write(Log_Unit,*) &
        'RelaxGeometry Converged new positions -- '
      Do a=1,Specific_atoms
         Atom_List(a)%Pos = Outpos(:,a)
         Atom_List(a)%Force%Pos(:,1) = Atom_List(a)%Pos
         Atom_List(a)%Frac_Pos = MATMUL(xtal%InverseBasis,Outpos(:,a))
         Write(Log_Unit,'(i5,2x,3f10.5)') a,Atom_List(a)%Pos
      Enddo


 Else  

    Close(Positions_Unit)
   
 EndIf

 If (.NOT.Success)  then
     Do a=1,Specific_atoms
         Inpos(:,a)=Atom_List(a)%Pos
      Enddo

       If(.not.UseShells)  Call ForceMove(ForceTol,MaxMove)
       If (UseShells)      Call ShellForceMove(ForceTol,MaxMove)
       Write(Log_Unit,*) &
         'RelaxGeometry Moved new positions along forces          Dif'
       Do a=1,Specific_atoms
          Outpos(:,a)=Atom_List(a)%Pos
          Write(Log_Unit,'(i5,2x,3f8.3,2x,3f8.3)') &
             a,Outpos(:,a),InPos(:,a)-OutPos(:,a)
       Enddo
 EndIf

  
  Call Flush(Log_Unit)

!Symmetrize positions taken out for now!!!
! If (Xtal%Rot_Size > 1) then
!    Do i=1,Specific_Atoms
!       Inpos(:,i)=Atom_List(i)%Frac_Pos
!    Enddo
!    Call Symmetrize_Pos(Inpos,Outpos)
!    Do i=1,Specific_Atoms
!       Atom_List(i)%Frac_Pos=Outpos(:,i)
!       Atom_List(i)%Pos=MATMUL(Xtal%Basis,Atom_List(i)%Frac_Pos)
!    Enddo
! EndIf

  Return
End Subroutine

!******************************************************************************
!
!  ForceMove - Moves the atoms in the direction of the forces on them  
!
!  ForceTol, MaxMove -- determine the scale of the move
!
!******************************************************************************

Subroutine ForceMove(ForceTol, MaxMove)
  Real, Intent(IN) :: ForceTol, MaxMove

  Real :: CurrForce(3), scale ,x(3*specific_atoms), u
  Real :: Inpos(3,specific_atoms),Outpos(3,specific_atoms)
  Integer :: i, j

  Write(Log_Unit,*) 'ForceMove: Moving atoms in direction of forces'

   scale=0
  Do i=1, Specific_Atoms
   CurrForce=Atom_List(i)%Force%CurrForce
   If (Anchor) CurrForce=CurrForce-Atom_List(Anchorindex)%Force%CurrForce
   scale=MAX(scale,ABS(CurrForce(1)),ABS(CurrForce(2)),ABS(CurrForce(3)))
  EndDo

  scale= Max(scale,MaxMove)
  scale= MIN(MaxMove/scale,1.0)

    Write(Log_unit,*) 'ForceMove -- move scale = ', scale
  
  CurrForce = 0
  If (Anchor) CurrForce=Atom_List(Anchorindex)%Force%CurrForce
  
  Do i=1, Specific_Atoms 
           !** Calculate the new position *
     Atom_List(i)%Force%Pos(:,1) = Atom_List(i)%Pos  &
           +  (Atom_List(i)%Force%CurrForce-CurrForce)*scale
     Write(Log_Unit,'("ForceMove: ", A10," R(old)  =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Pos
     Write(Log_Unit,'("ForceMove: ", A10," F(old)  =",1p3e13.5)') &
             Atom_List(i)%Name, (Atom_List(i)%Force%CurrForce-CurrForce)

     Atom_List(i)%Pos = Atom_List(i)%Force%Pos(:,1)  !*** Copy it 
     Atom_List(i)%Frac_Pos = MATMUL(xtal%InverseBasis, Atom_List(i)%Pos)  

     Write(Log_Unit,'("ForceMove: ", A10," R(new)  =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Pos     
     Write(Log_Unit,'("ForceMove: ", A10," R(frac) =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Frac_Pos     
  End Do

  Call Flush(Log_Unit)


 If (Xtal%Rot_Size > 1) then
    Do i=1,Specific_Atoms
       Inpos(:,i)=Atom_List(i)%Frac_Pos
    Enddo
    Call Symmetrize_Pos(Inpos,Outpos)
    Do i=1,Specific_Atoms
       Atom_List(i)%Frac_Pos=Outpos(:,i)
       Atom_List(i)%Pos=MATMUL(Xtal%Basis,Atom_List(i)%Frac_Pos)
    Enddo
 EndIf

  Return
End Subroutine

!***************************************************************************
!
! ShellForce  Calculates - d E/ d Cp   for Shell parameters
!
!   This subroutine should be called iff UseShells = .true.
!
!***************************************************************************

 Subroutine ShellForce

 Integer :: p,a,i,j,k, NPARAMS ,many
 Real ::  term,vec(3),force(3)

 NPARAMS = Shell%NPARAMS

 Write(Log_Unit,*)'ShellForce: NPARAMS',NPARAMS
 
    Shell%ShellForce(1:NPARAMS) = 0

 Do a=1,Specific_Atoms
    many = Shell%AtomMap(a)%MapParams
    Write(Log_Unit,*)'ShellForce: a,many',a,many
    If (many > 0) then
       force=Atom_List(a)%Force%CurrForce
       Do i=1,many
          p=Shell%AtomMap(a)%Map(i)
          vec=Shell%AtomMap(a)%V(:,i)
          term = Dot_Product(vec,force)
          Write(Log_Unit,*) 'ShellForce: p,i,term',p,i,term
          Shell%ShellForce(p)=Shell%ShellForce(p) + term
       Enddo
     EndIf
  Enddo

  Do p=1,NPARAMS
     Write(Log_Unit,&
      '("SHELLFORCE: ",i4," COEF ",1pe15.7," FORCE ",1pe15.7)') &
           p,Shell%C(p),Shell%ShellForce(p)
  Enddo

  Return
  End Subroutine
!***************************************************************************
! 
! CheckShellForces  -- Checks convergence of Shellforces
!
!  Should only be called if UseShells = .true.
!
!***************************************************************************
    
                                                             
Subroutine CheckShellForces(ForceTol,Converged)                   
  Real    , Intent(IN):: ForceTol                            
  Logical , Intent(OUT):: Converged                          
                                                             
  Integer :: p  , NPARAMS                                             
  Real :: mag                                                
                                                             
     mag = 0
  NPARAMS = Shell%NPARAMS
  Converged = .TRUE.                                         
  Do p=1, NPARAMS
     mag =  MAX(mag, ABS(Shell%ShellForce(p)))
  EndDo
                                                             
     If (mag>ForceTol) Converged = .FALSE.                   
 
 return                                                      
End Subroutine                                               

!***************************************************************************
!
! ShellPos    Calculates - Atomic postions given Cp Shell parameters
!
!   This subroutine should be called iff UseShells = .true.
!
!***************************************************************************
 Subroutine ShellPos

 Integer :: p,a,i,j,k, NPARAMS ,many
 Real ::  term,vec(3),pos(3)

 NPARAMS = Shell%NPARAMS
 
 Do a=1,Specific_Atoms
    many = Shell%AtomMap(a)%MapParams
    If (many > 0) then
       pos = Shell%AtomMap(a)%Origin
       Do i=1,many
          p=Shell%AtomMap(a)%Map(i)
          vec=Shell%AtomMap(a)%V(:,i)
          pos = pos + Shell%C(p)*vec    
       Enddo
       Atom_List(a)%Force%Pos(:,1) = pos
       Atom_List(a)%Pos = pos
       Atom_List(a)%Frac_Pos = MATMUL(xtal%InverseBasis,pos)
     EndIf
  Enddo

  Return
  End Subroutine
    
!******************************************************************************
!
!  ShellForceMove - Moves the atoms in the direction 
!                   of generalized forces  
!
!  ForceTol, MaxMove -- determine the scale of the move
!
!  Should be used only iff UseShells = .true. 
!******************************************************************************

Subroutine ShellForceMove(ForceTol, MaxMove)
  Real, Intent(IN) :: ForceTol, MaxMove

  Real :: scale 
  Real, allocatable :: C(:)
  Integer :: i, j , p, NPARAMS,ier

  Write(Log_Unit,*) 'ShellForceMove: Moving atoms with Shellforces'

  !  Call ShellForce   (assume that this has been already called)

  NPARAMS = Shell%NPARAMS
  Allocate(C(NPARAMS),stat=ier)
    If (ier > 0 ) then
       Write(Error_Unit,*) 'Error in ShellForceMove allocations',ier
       Stop
    EndIf

   scale=0
  Do p=1, NPARAMS
   scale=MAX(scale,ABS(Shell%ShellForce(p)))
  EndDo

  scale= Max(scale,MaxMove)
  scale= MIN(MaxMove/scale,1.0)

    Write(Log_unit,*) 'ShellForceMove -- move scale = ', scale

  
  Do p=1, NPARAMS
     C(p) = Shell%C(p) + Shell%ShellForce(p)*scale
     Write(Log_Unit,'("ShellForceMove: ",i5," Cold ", 1pe15.7)')&
             p,Shell%C(p)
     Write(Log_Unit,'("ShellForceMove: ",i5," Cnew ", 1pe15.7)')p,C(p)
     Write(Log_Unit,'("ShellForceMove: ",i5," Fp   ", 1pe15.7)')&
             p,Shell%ShellForce(p)

     Shell%C(p) = C(p)   ! Replace old value
  End Do

  Call Flush(Log_Unit)

  Call ShellPos

  Return
End Subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! LSMove
!
!   Least Squares fit --
!   Estimates equilibrium values of generalized positions given          
!     Niter values of E, generalized forces, and generalized positions
!
!   Edata is an array of the total energies (or negative of cohesive E)
!      for each iteration
!   Updated 05-15-02 NAWH
!   Effectively disabled by setting success=.false. 05-20-02 NAWH
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
Subroutine LSMove(Niter,Nfreedom,Edata,Fdata,Rdata,Efit,Rfit,&
           ForceTol,Maxmove,ConditionNo,Success)

 Integer, Intent(IN) :: Niter, Nfreedom
 Real,    Intent(IN) :: ForceTol,Maxmove,ConditionNo
 Real,Intent(INOUT)  :: Rdata(Nfreedom,Niter),Fdata(Nfreedom,Niter),Edata(Niter)
 Real,    Intent(OUT):: Efit,Rfit(Nfreedom)
 Logical, Intent(OUT):: Success

  Real, allocatable  :: A(:,:),U(:,:),VT(:,:),S(:),Work(:),Rref(:),B(:)
  Integer, allocatable :: IWork(:),Index(:)
  Integer :: N, M, Lwork, Liwork, Info, i,j,k,icount ,p ,pp
  Real :: x,weight,Emin

  Success = .false.
  return      ! Program generally does not work so this
              !  effectively minimizes the damage   5-20-02 NAWH
  Efit=0;Rfit=0

  If (Niter.lt.2) Return

  M=Niter
  N=Nfreedom+1

  Lwork=8*N*N+12*N
  Liwork=8*N

  Allocate(A(N,N),S(N),U(N,N),VT(N,N),Work(Lwork),IWork(Liwork),&
           Rref(Nfreedom),B(N),Index(Niter),stat=info)
    If (info /= 0) then
       write(Error_Unit,*) 'Error in Fitmove allocation ', info
       call flush(Error_Unit)
       stop
    Endif

  Call Insertion_Sort(Edata(1:Niter),Index,.TRUE.)  !Order in increasing E

  ! Write(Log_Unit,*) 'Index',Index

  ! Use lowest energy iteration as reference

  Rref(1:Nfreedom)=Rdata(1:Nfreedom,Index(1))
  ! Write(Log_Unit,*)'Rref',Rref
  Emin = Edata(Index(1))
  Edata(1:Niter) = Edata(1:Niter)-Emin

  A=0 ; B=0 ; S=0

  Do i=1,M
       j=Index(i)
       Rdata(1:Nfreedom,j)=Rdata(1:Nfreedom,j)-Rref(1:Nfreedom)
       weight = 1   !  weight= 1.0/(abs(Edata(j)/ForceTol) + 1)
       Edata(j)=Edata(j)+&
        0.5d0*Dot_Product(Rdata(1:Nfreedom,j),Fdata(1:Nfreedom,j))
       S(1)=S(1)+weight
       A(1,1)=A(1,1)+weight
       B(1)=B(1)+weight*Edata(j)

       write(Log_Unit,*) 'LSMOVE',i,j,Edata(j),weight,A(1,1),B(1)
        Do p=1,Nfreedom
           x=Fdata(p,j)*weight
           A(1,p+1)=A(1,p+1) + x/2
           A(p+1,1)=A(p+1,1) + x/2
           B(p+1)=B(p+1) + x*Edata(j)

           Do pp=1,Nfreedom
              A(p+1,pp+1)=A(p+1,pp+1) + x*Fdata(pp,j)/2
           EndDo
        EndDo

  EndDo

        Do p=1,Nfreedom+1
           B(p)=B(p)/S(1)
           write(Log_unit,*) 'B',p,B(p)
           Do pp=1,Nfreedom+1
            A(p,pp)=A(p,pp)/S(1)
           Enddo
           write(Log_Unit,*) 'A',(A(p,pp),pp=1,Nfreedom+1)
        Enddo

  ! Write(Log_Unit,*) 'A',A
  ! Write(Log_Unit,*) 'B',B
  ! Write(Log_Unit,*) 'S',S
  S = 0
  Call DGESDD('A',N,N,A,N,S,U,N,VT,N,Work,Lwork,Iwork,info)
     Write(Log_Unit,*) 'completed DGESDD with info = ',info
     Write(Log_Unit,*) 'Singular Values'
     Write(Log_Unit,'(1p5e15.7)') (S(i),i=1,N)
     If (info /= 0) then
       write(Error_Unit,*) 'Error in Fitmove RDGESDD ', info
       call flush(Error_Unit)
       stop
     Endif
  
  Work=0
  x=ABS(S(1))/ConditionNo
  Do i=1,N
     If (ABS(S(i)).gt.x) then
        Work(1:N)=Work(1:N)+VT(i,1:N)*DOT_PRODUCT(U(1:N,i),B)/S(i)
     EndIf
  Enddo

   Write(Log_Unit,*) 'Work',(Work(i),i=1,N)

  Efit=Work(1)+Emin
  Rfit(1:Nfreedom)=Rref(1:Nfreedom)+Work(2:Nfreedom+1)      

   write(Log_Unit,*) ' Efit = ',Efit

    write(Log_Unit,*) ' Rref = ',Rref
    write(Log_Unit,*) ' Rfit = ',Rfit

 ! Check MaxMove:
      x=-100
      Do i=1,Nfreedom
         x=Max(x,ABS(Rref(i)-Rfit(i)))
      Enddo

    Write(Log_Unit,*) 'error measure = ',x

   if (x <= MaxMove) Success=.true.
   Deallocate(A,S,U,VT,Work,IWork,Rref,B,Index)
 
   Return

End subroutine LSMove

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! LSMove -- old version -- no longer used
!
!   Least Squares fit --
!   Estimates equilibrium values of generalized positions given          
!     Niter values of E, generalized forces, and generalized positions
!
!   Edata is an array of the total energies (or negative of cohesive E)
!      for each iteration
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
Subroutine LSMove_old(Niter,Nfreedom,Edata,Fdata,Rdata,Efit,Rfit,&
           ForceTol,Maxmove,ConditionNo,Success)

 Integer, Intent(IN) :: Niter, Nfreedom
 Real,    Intent(IN) :: ForceTol,Maxmove,ConditionNo
 Real,Intent(INOUT)  :: Rdata(Nfreedom,Niter),Fdata(Nfreedom,Niter),Edata(Niter)
 Real,    Intent(OUT):: Efit,Rfit(Nfreedom)
 Logical, Intent(OUT):: Success

  Real, allocatable  :: A(:,:),U(:,:),VT(:,:),S(:),Work(:),Rref(:),B(:)
  Integer, allocatable :: IWork(:),Index(:)
  Integer :: N, M, Lwork, Liwork, Info, i,j,k,icount ,p ,pp
  Real :: x,weight,Emin

  Success = .false.
  Efit=0;Rfit=0

  If (Niter.lt.2) Return

  M=Niter
  N=Nfreedom+1

  Lwork=8*N*N+12*N
  Liwork=8*N

  Allocate(A(N,N),S(N),U(N,N),VT(N,N),Work(Lwork),IWork(Liwork),&
           Rref(Nfreedom),B(N),Index(Niter),stat=info)
    If (info /= 0) then
       write(Error_Unit,*) 'Error in Fitmove allocation ', info
       call flush(Error_Unit)
       stop
    Endif

  Call Insertion_Sort(Edata(1:Niter),Index,.TRUE.)  !Order in increasing E

  ! Write(Log_Unit,*) 'Index',Index

  ! Use lowest energy iteration as reference

  Rref(1:Nfreedom)=Rdata(1:Nfreedom,Index(1))
  ! Write(Log_Unit,*)'Rref',Rref
  Emin = Edata(Index(1))
  Edata(1:Niter) = Edata(1:Niter)-Emin

  A=0 ; B=0 ; S=0

  Do i=1,M
       j=Index(i)
       Rdata(1:Nfreedom,j)=Rdata(1:Nfreedom,j)-Rref(1:Nfreedom)
       weight= 1.0/((Edata(j)/ForceTol)**2 + 1)
       S(1)=S(1)+weight
       A(1,1)=A(1,1)+weight
       B(1)=B(1)+weight*Edata(j)

        Do p=1,Nfreedom
           x=Fdata(p,j)*weight
           S(p+1)=S(p+1) + x
           A(1,p+1)=A(1,p+1) + x
           A(p+1,1)=A(p+1,1) + x
           B(1)=B(1) + 0.5*x*Rdata(p,j)
           B(p+1)=B(p+1) + x*Edata(j)

           Do pp=1,Nfreedom
              A(p+1,pp+1)=A(p+1,pp+1) + x*Fdata(pp,j)
              B(p+1)=B(p+1) + 0.5*x*Fdata(pp,j)*Rdata(pp,j)
           EndDo
        EndDo

  EndDo

        A(1,1)=A(1,1)/S(1)
        B(1)=B(1)/S(1)
        Do p=1,Nfreedom
           A(1,p+1)=0.5*A(1,p+1)/S(1)
           B(p+1)=B(p+1)/S(p+1)
           A(p+1,1:N)=0.5*A(p+1,1:N)/S(p+1)
        EndDo

  ! Write(Log_Unit,*) 'A',A
  ! Write(Log_Unit,*) 'B',B
  ! Write(Log_Unit,*) 'S',S
  S = 0
  Call DGESDD('A',N,N,A,N,S,U,N,VT,N,Work,Lwork,Iwork,info)
     Write(Log_Unit,*) 'completed DGESDD with info = ',info
     Write(Log_Unit,*) 'Singular Values'
     Write(Log_Unit,'(1p5e15.7)') (S(i),i=1,N)
     If (info /= 0) then
       write(Error_Unit,*) 'Error in Fitmove RDGESDD ', info
       call flush(Error_Unit)
       stop
     Endif
  
  Work=0
  x=ABS(S(1))/ConditionNo
  Do i=1,N
     If (ABS(S(i)).gt.x) then
        Work(1:N)=Work(1:N)+VT(i,1:N)*DOT_PRODUCT(U(1:N,i),B)/S(i)
     EndIf
  Enddo

  ! Write(Log_Unit,*) 'Work',(Work(i),i=1,N)

  Efit=Work(1)+Emin
  Rfit(1:Nfreedom)=Rref(1:Nfreedom)+Work(2:Nfreedom+1)      

   write(Log_Unit,*) ' Efit = ',Efit

   ! write(Log_Unit,*) ' Rref = ',Rref
   ! write(Log_Unit,*) ' Rfit = ',Rfit

 ! Check MaxMove:
      x=-100
      Do i=1,Nfreedom
         x=Max(x,ABS(Rref(i)-Rfit(i)))
      Enddo

    Write(Log_Unit,*) 'error measure = ',x

   if (x <= MaxMove) Success=.true.
   Deallocate(A,S,U,VT,Work,IWork,Rref,B,Index)
 
   Return

End subroutine LSMove_old

!******************************************************************************
!
!  PrepPrevious - Prepares the previous set of Psi's for use with the
!      new positions
!
!******************************************************************************

Subroutine PrepPrevious
  Integer :: i, Psi_toUse(Mem_MapSize), VecSize,  j, Flag ,k
  Complex, Pointer  :: Vec(:,:), Work(:)
 !Real     :: Eigen(TotalPsi),  Curr(2), mag, Energy
  Real     ::  Curr(2), mag, Energy
  Real  ,allocatable :: Eigen(:)
 !Integer  :: NewPsi_Map(Mem_MapSize), NewSize
  Integer  :: NewSize,spin,s
  Integer,allocatable  :: NewPsi_Map(:), LUT(:,:)
 !Integer  :: LUT(2,Mem_MapSize)
  Type (Mem_Handle), Pointer :: Psi


  Allocate(Eigen(TotalPsi),NewPsi_Map(Mem_MapSize),LUT(2,Mem_MapSize))
  Call GetBuffer( Work)

  DoLCAO = 0
  PsiInfo(:)%PDOT_Stored = .FALSE.  !** Clear the PDOT's
  PsiInfo(:)%KE_Stored = .FALSE.    !** and KE's
  
  curr = (/V_NewMix, Dij_NewMix/)
  V_NewMix = 1;  Dij_NewMix = 1;

  !*** First Normalize the Psi's to Qt+Qh=0 ***
  
  NewPsi_Map = MH_Skip
 
  If (spindependence) spin=2
  If (.not.spindependence) spin=1

  Do  k=1, NumKpnts
    do s=1,spin
        if (s==1) Global_spinup=.true.
        if (s==2) Global_spinup=.false.

     Psi_toUse = MH_Skip
     Do i=1,Mem_MapSize
       If ((PsiInfo(i)%Kpnt == k) .AND. &
               (PsiInfo(i)%spinup.eqv.Global_spinup) .AND.  &
               (PsiInfo(i)%Available == Mem_Used)) then
        Psi_toUse(i)=MH_toProcess
        NewPsi_Map(i)=MH_toProcess
       EndIf
     EndDo

     Call Phase_Generic( Psi_toUse,  k)

     Call GetNextPsi(Flag, Psi)
     Do While (Flag > 0)
        call kplusGfilter(Psi%ptr,k)
        j = Psi%Index
        PsiInfo(j)%DoSave=1
        PsiInfo(j)%Available=Mem_Used
        PsiInfo(j)%PDOT_Stored = .FALSE.

        Call Opsi( Psi%Ptr, Work, &
             PsiInfo(j)%PDOT, i, PsiInfo(j)%PDOT_Stored)
              
        mag = Basis_DotProd( Psi%Ptr, Work, .FALSE.)
        mag = SQRT(mag)
!Write(Log_Unit,*) 'PrepPrev: Kpnt=',i, ' * DI=',Psi%Index, ' * DOT=',mag
        Psi%Ptr = Psi%Ptr / mag
        PsiInfo(Psi%Index)%PDOT = PsiInfo(Psi%Index)%PDOT/mag

        Call SavePsi_toBuffer(j,Psi%Ptr)
        Call GetNextPsi(Flag, Psi)
     End Do
   Enddo
  End Do

! Clean up the remaining stored wavefunctions

  Do j=1,Mem_MapSize
     If(NewPsi_Map(j) == MH_Skip) then
       PsiInfo(j)%Kpnt=0
       PsiInfo(j)%spinup=.true.
       PsiInfo(j)%DoSave=0
       PsiInfo(j)%Available=Mem_Available
       i=PsiInfo(j)%MemBufIndex
       If(i>0) Globalmap%Psi_Handle(i)%Index=0
       PsiInfo(j)%MemBufIndex=0
     EndIf
  Enddo

  Call CalcMeritFunction(  Energy)

  Do i=1, NumKpnts
    do s=1,spin
        if (s==1) Global_spinup=.true.
        if (s==2) Global_spinup=.false.
     Psi_toUSe = MH_Skip
     Where (PsiInfo(:)%Kpnt == i .and. &
          (PsiInfo(:)%spinup.eqv.Global_spinup)) Psi_toUse = MH_toProcess
     VecSize = SUM(Psi_toUse)
     Allocate(Vec(VecSize, VecSize))


     Call PerformGRRP_ALL( Psi_toUse, VecSize, &
       Vec, Eigen, LUT, NewSize, i)
     Call MakeNewPsi( VecSize, NewSize, Vec, Eigen, LUT, i)

     DeAllocate(Vec)
    enddo
  End Do
Write(Log_Unit,*) 'PrepPrevious   : After Orthog'

!ReCalculate Occupancies
  Call MakeBandMap
  Call CalcNewOccAndFermi
  Call ReCluster_ALL

  PsiInfo(:)%PDOT_Stored = .FALSE.  !** Clear the PDOT's
  PsiInfo(:)%KE_Stored = .FALSE.    !** and KE's
  Call CalcMeritFunction(  Energy)

  V_NewMix = curr(1);  Dij_NewMix = curr(2); 
  Call CalcEigenvalues

  Call FreeBuffer(Work)
  DeAllocate(Eigen,NewPsi_Map,LUT)

  Return
End Subroutine

!******************************************************************************
!
!  PrepPrevious_band bandstructure_mode  - Prepares the previous set of Psi's
!      next iteration
!
!******************************************************************************

Subroutine PrepPrevious_band(NumB)
  Integer, Intent(IN) :: NumB
  Integer :: i, Psi_toUse(Mem_MapSize), VecSize,  j, Flag,n,maxindex
  Complex, Pointer  :: Vec(:,:), Work(:)
!  Real     :: Eigen(TotalPsi), Curr(2), mag, Energy
  Real     ::  Curr(2), mag, Energy
  Real, allocatable :: Eigen(:)
  Integer  :: NewSize,spin,s
! Integer  :: LUT(2,Mem_MapSize)
  Integer, allocatable  :: LUT(:,:)
  Type (Mem_Handle), Pointer :: Psi

  Allocate(LUT(2,Mem_MapSize),Eigen(TotalPsi))

  MinBands=NumB+4
  TotalPsi=MinBands+8
  Numbands=NumB

  Call GetBuffer( Work)

  DoLCAO = 0
  PsiInfo(:)%PDOT_Stored = .FALSE.  !** Clear the PDOT's
  PsiInfo(:)%KE_Stored = .FALSE.    !** and KE's
  
  If (spindependence) spin=2
  If (.not.spindependence) spin=1
 


! Find set of bands from previous run with energies less than Eigen_Max

  Psi_toUse = MH_Skip

  n=0; maxindex=Maxval(PsiInfo(:)%Kpnt)

  Kloop: Do i=1,maxindex
    Do  j=1,Mem_MapSize
     If (PsiInfo(j)%Available == Mem_Used) then
        If ((PsiInfo(j)%Kpnt == i) .AND. (PsiInfo(j)%Energy <= Eigen_max)) then
           Psi_toUse(j) = MH_toProcess
           n=n+1
           If (n > NumB*spin) Exit Kloop
        EndIf
     EndIf
    ENddo
  End Do  Kloop
     
  If (n < NumB*spin) then
    DoLcao=1
    return
  Endif

  i=1
  !  Clean up the remaining stored wavefunctions
  Do j = 1,Mem_MapSize
     If (Psi_toUse(j)== MH_toProcess) then
        PsiInfo(j)%Kpnt=i
        PsiInfo(j)%DoSave=1
        PsiInfo(j)%Available=Mem_Used
     Else
        PsiInfo(j)%Kpnt=0
        PsiInfo(j)%spinup=.true.
        PsiInfo(j)%DoSave=0
        PsiInfo(j)%Available=Mem_Available
        n=PsiInfo(j)%MemBufIndex
        If (n > 0) then
           Globalmap%Psi_Handle(n)%Index=0
        EndIf
        PsiInfo(j)%MemBufIndex=0
     EndIf
  Enddo

     Call Phase_Generic( Psi_toUse,  i)

     Call GetNextPsi(Flag, Psi)
     Do While (Flag > 0)
        call kplusGfilter(Psi%ptr,i)
        j = Psi%Index
        PsiInfo(j)%DoSave=1
        PsiInfo(j)%Available=Mem_Used
        PsiInfo(j)%PDOT_Stored = .FALSE.

        Call Opsi( Psi%Ptr, Work, &
             PsiInfo(j)%PDOT, i, PsiInfo(j)%PDOT_Stored)
              
        mag = Basis_DotProd( Psi%Ptr, Work, .FALSE.)
        mag = SQRT(mag)
Write(Log_Unit,*) 'PrepPrev: Kpnt=',i, ' * DI=',Psi%Index, ' * DOT=',mag
Call flush(log_unit)
        Psi%Ptr = Psi%Ptr / mag
        PsiInfo(Psi%Index)%PDOT = PsiInfo(Psi%Index)%PDOT/mag

        Call SavePsi_toBuffer(j,Psi%Ptr)
        Call GetNextPsi(Flag, Psi)
     End Do

  Call CalcMeritFunction(  Energy)

  Do i=1, NumKpnts
   do s=1,spin
        if (s==1) Global_spinup=.true.
        if (s==2) Global_spinup=.false.

     Psi_toUSe = MH_Skip
     Where (PsiInfo(:)%Kpnt == i .and. &
        (PsiInfo(:)%spinup.eqv.Global_spinup)) Psi_toUse = MH_toProcess
     VecSize = SUM(Psi_toUse)
     Allocate(Vec(VecSize, VecSize),stat=j)

     Call PerformGRRP_ALL( Psi_toUse, VecSize, &
       Vec, Eigen, LUT, NewSize, i)
     Call MakeNewPsi( VecSize, NewSize, Vec, Eigen, LUT, i)

     DeAllocate(Vec)
   enddo
  End Do
Write(Log_Unit,*) 'PrepPrevious_band   : After Orthog'

!ReCalculate Occupancies
  Call MakeBandMap
  Call CalcNewOccAndFermi
  Call ReCluster_ALL

  PsiInfo(:)%PDOT_Stored = .FALSE.  !** Clear the PDOT's
  PsiInfo(:)%KE_Stored = .FALSE.    !** and KE's
  Call CalcMeritFunction(  Energy)

  V_NewMix = curr(1);  Dij_NewMix = curr(2); 
  Call CalcEigenvalues

  Call FreeBuffer(Work)
  Deallocate(LUT,Eigen)

  Return
End Subroutine

!******************************************************************************
!
! RelaxAtoms - Relaxes the Atoms to equilibrium or until the max number of
!       allowed steps are made.
!
!  Mode      - Relaxation mode 0=MD, 1=Geometry
!  MaxStep   - Max number of allowed steps
!  ChargeBase - Ouput base filename containing the Psi's and V for each 
!               step.  If ChargeBASE="NULL" then nothing is stored   
!  MaxSCF     - Max number of SCF steps to perform/atom iteration
!  ChargeTol  - Cohesive Energy tolerance for RelaxElectrons
!  ForceTol   - Force Magnitude tolerance to determine equilibrium
!  MaxMove    - Maximum move distance allowed for Geometry optimization
!  ConditionNo -- Parameter for Fitmove in Geometry optimization
!
!******************************************************************************

Subroutine RelaxAtoms(Mode,MaxStep,ChargeBase,MaxSCF,ChargeTol,ForceTol,&
                             MaxMove,ConditionNo)
  Integer,       Intent(IN) :: Mode
  Integer,       Intent(IN) :: MaxStep
  Character*(*), Intent(IN) :: ChargeBase
  Integer,       Intent(IN) :: MaxSCF
  Real,  Intent(IN) :: ChargeTol
  Real,  Intent(IN) :: ForceTol
  Real,  Intent(IN) :: MaxMove , ConditionNo 

  Character*100 :: FileName,message
  Logical       :: Converged, EConverged
  Integer       :: Fstep, i
  Real          :: TotalE, KE, curr(2) , Etol
  

  Converged = .FALSE.
  Fstep = 1

  Call RelaxElectrons(MaxSCF, ChargeTol, Etol, EConverged) 
  !If (Trim(ChargeBase) /= "NULL") then
  !   Write(Filename ,*) FStep
  !   FileName = Trim(ChargeBase) // Trim(ADJUSTL(Filename)) // '.rsgrid'
  !   Call StoreSolution_Binary( FileName)
  !End If
     
  write(log_unit,*) 'RelaxAtoms: after RelaxElectrons', EConverged
  call flush(Log_unit)
  If (EConverged) message='SCF step converged'
  If (.not.EConverged) message='SCF step not converged'
  Call CalcForces ; If (UseShells) Call ShellForce
  Call Store_ForcesAndPositions(Trim(Force_Name),message)
  If (EConverged) Call Position_Archive(Etol)
  Call CheckForces(ForceTol,Converged)

  !If (Mode == RELAX_GEOMETRY) Call MakeGeometryMixing

  Do While ((.NOT. Converged) .AND. (Fstep < MaxStep))
     Fstep = Fstep + 1

     !*** Calculate the KE of the nuclei **
     KE = 0
     Do i=1, Specific_Atoms
        KE = KE + 0.5*AtomType_Info(Atom_List(i)%TypeIndex)%Mass* &
             SUM(Atom_list(i)%Force%Velocity(:,1)**2)
     End Do

     TotalE = CohesiveEnergy - KE
     Write(Log_Unit,*) 'RelaxAtoms: Total energy:',TotalE, ' * Cohesive=', &
          CohesiveEnergy, ' * KE=',KE

     If (Mode == RELAX_MD) Call MoveAllAtoms(Mode)
! For relax_geometry mode, do not move if SCF step has not converged
     If (Mode == RELAX_GEOMETRY .AND. EConverged) then
               Call RelaxGeometry(ForceTol,MaxMove,ConditionNo)
               Write(Output_Unit,*)
               Write(Output_Unit,*) 'Atoms moved to new positions'
               Write(Output_Unit,*)
               Call Flush(Output_Unit)
     EndIf
     If (Mode == RELAX_GEOMETRY .AND. .not.EConverged) then
               Write(Output_Unit,*)
               Write(Output_Unit,*) 'Atoms left in same positions'
               Write(Output_Unit,*)
               Call Flush(Output_Unit)
     EndIf
     !If(.not.bandstructure_mode)Call StoreSolution_Binary('paw.wfn')
     !Call StoreSolution_Binary('paw.wfn')   !for safety

     Call UpdateSystem

     If (MD_Guess == MD_PREVIOUS) then
        Call PrepPrevious
!Call LoadGuess('paw.wfn')
     else  
        DoLCAO = Initial_guess
        !*** Make sure ALL wave functions are free ***
        PsiInfo(:)%Kpnt = 0
        PsiInfo(:)%spinup = .true.
        PsiInfo(:)%DoSave = 0
        PsiInfo(:)%Available = Mem_Available
        PsiInfo(:)%PDOT_Stored = .FALSE.
        PsiInfo(:)%KE_Stored = .FALSE.    
     End If
     
     !** Reset the mixing to NULL ***
     Call Anderson_ResetMix(SCFvalues%AC)
     If (spindependence) Call Anderson_ResetMix(SCFvalues%ACspin)
     Do i=1, Specific_Atoms
        Call Anderson_ResetMix(Atom_List(i)%Dij_AC)
        If (spindependence) Call Anderson_ResetMix(Atom_List(i)%Dijspin_AC)
     End Do
     
     
     Call RelaxElectrons(MaxSCF, ChargeTol, Etol, EConverged) 
     !If (Trim(ChargeBase) /= "NULL") then
     !   Write(Filename ,*) FStep
     !   FileName = Trim(ChargeBase) // Trim(ADJUSTL(Filename)) // '.wfn'
     !   Call StoreSolution_Binary( FileName)
     !End If

     If (EConverged) message='SCF step converged'
     If (.not.EConverged) message='SCF step not converged'
     Call CalcForces ; If (UseShells) Call ShellForce
     Call Store_ForcesAndPositions(Trim(Force_Name),message)
     If (EConverged) Call Position_Archive(Etol)
     If(.not.UseShells)Call CheckForces(ForceTol,Converged)
     If(UseShells)Call CheckShellForces(ForceTol,Converged)
  End Do

  Write(Log_Unit,*) 'RelaxAtoms: Finished ',FStep,&
      ' iterations with convergence = ',Converged
  Write(Output_Unit,*) 'RelaxAtoms: Finished ',FStep,&
      ' iterations with convergence = ',Converged
  
  Return
End Subroutine


!TD!******************************************************************************
!TD!
!TD! RelaxTime - Loads a time-dependent run fron the input file
!TD!
!TD!   NumSteps - Number of steps to take
!TD!
!TD!   The calling format is :
!TD!
!TD!       SET_NAME TOTAL_CURRENT_NAME 'current_filename'  # optional
!TD!       SET_NAME ELECTRIC_FIELD_NAME 'e_field_name'     # optional
!TD!       TIME_STEP  dt                                   # optional
!TD!
!TD!       RELAX TIME Num_steps    # This is the actual command
!TD!
!TD!******************************************************************************
!TD
!TDSubroutine RelaxTime(NumSteps)
  !TDInteger, Intent(IN) :: NumSteps
!TD
  !TDInteger :: n, Psi_toUSe(Mem_MapSize), i, Kpnt,Mode
  !TDType (Mem_Handle), Pointer :: Psi
  !TDComplex, Pointer :: NewPsi(:)
  !TDReal :: backup(2), time, Current(3), Elec(3)
  !TDLogical :: AtomsMoved
!TD
  !TDWrite(*,*) "RelaxTime --  Mode has not been defined"
!TD
  !TDAtomsMoved = .FALSE.
  !TDDo i=1, Specific_Atoms
     !TDIf (.NOT. Atom_list(i)%Freeze) AtomsMoved = .TRUE.
  !TDEnd Do
!TD
  !TDWrite(Log_Unit,*) 'RelaxTime: Initial settings for time-dependent run'
  !TDWrite(Log_Unit,*) '   Total Steps:',Numsteps, ' * DT=',TimeStep, ' fs'
  !TDWrite(Log_Unit,*) '   Total simulation time :',NumSteps*TimeStep, ' fs'
  !TDWrite(log_Unit,*) ' '
  !TDWrite(Log_Unit,*) '   Force Output:', TRIM(Force_Name)
  !TDWrite(Log_Unit,*) '   Total Current Name:', TRIM(TotalCurrent_Name)
  !TDWrite(Log_Unit,*) '   Electric Field Name:', TRIM(E_Name)
  !TDWrite(Log_Unit,*) '   ** NOTE: Appending ALL results to the above files! **'
  !TDWrite(log_Unit,*) ' '
!TD
  !TDopen(unit=TotalCurrent_Unit,file=Trim(totalCurrent_Name), RECL=1024, &
       !TDposition='append')
  !TDWrite(Totalcurrent_Unit,*)
  !TDCall Printdate(Totalcurrent_Unit,'#  Current results generated on date ')
  !TDWrite(TotalCurrent_Unit, *) '# 1.Time  2.Jx   3.Jy  4.Jz'  
!TD
  !TDopen(unit=E_Unit,file=Trim(E_Name), RECL=1024, &
       !TDposition='append')
  !TDWrite(E_Unit,*)
  !TDCall Printdate(E_Unit,'#  E-field results generated on date ')
  !TDWrite(E_Unit, *) '# 1.Time  2.Ex   3.Ey  4.Ez  5.Ax  6.Ay  7.Az'  
!TD
  !TDbackup = (/V_newMix, Dij_NewMix/)
  !TDV_newMix=1
  !TDDij_NewMix=1
!TD
  !TDtime = 0
  !TDCall CalcVectorPotential(time, Elec, Vector_Pot)
  !TDCall CalcTotalCurrent(Current)  
!TDWrite(Log_Unit,*) 'RelaxTime: Elec=',Elec, ' * VecPot=',Vector_Pot
!TDWrite(Log_Unit,*) 'RelaxTime: Atomic_Mode=',Atomic_Mode
!TDCall CalcEigenvalues
!TD
  !TDAtomic_Mode = SIM_TIME
!TDWrite(Log_Unit,*) 'RelaxTime: SIM_TIME * AM=',Atomic_Mode
!TDCall CalcEigenvalues
!TD
  !TDDo n=1, NumSteps
     !TDWrite(Log_Unit,*) 'RelaxTime:  Iter=',n
     !TDCall PrintDate(Log_Unit,'RelaxTime: Start')
!TD
     !TD!** Update all Loaded bands **
     !TDCall GetBuffer( NewPsi)
     !TDDo Kpnt = 1, NumKpnts
        !TDPsi_toUse = MH_Skip
        !TDWhere (PsiInfo(:)%Kpnt == Kpnt) Psi_toUse = MH_toProcess
!TD
        !TDCall Phase_Generic( Psi_toUse,  kpnt)
       !TD 
        !TDCall GetNextPsi(i, Psi)
        !TDDo While (i /= 0)
           !TDCall TIME_PropagatePsi(Psi%Index, Psi%Ptr, NewPsi, Kpnt)        
           !TDCall SavePsi_toBuffer( Psi%Index, NewPsi)
!TD
           !TDCall GetNextPsi(i, Psi)
        !TDEnd Do
     !TDEnd Do
     !TDCall FreeBuffer(NewPsi)
    !TD 
     !TD!** Move the atoms **     
     !TDIf (AtomsMoved) then
        !TDCall CalcForces ; If (UseShells) Call ShellForce
        !TDCall Store_ForcesAndPositions(Force_Name,'Called from RelaxTime')
        !TDCall MoveAllAtoms(Mode)
        !TDCall UpdateSystem
     !TDEnd If
!TD
     !TDtime = TimeStep*(n-1)
     !TDCall CalcVectorPotential(time, Elec, Vector_Pot)
     !TDCall CalcTotalCurrent(Current)
!TD
     !TDWrite(TotalCurrent_Unit,*) time, Current
     !TDWrite(E_Unit,*) time, Elec, Vector_Pot
    !TD 
     !TD!** Update the hamiltonian **
     !TDCall CalcHam(CohesiveEnergy)
     !TDPsiInfo(:)%PDOT_Stored = .FALSE.
     !TDCall CalcEigenvalues
  !TDEnd Do
!TD
  !TDV_newMix = Backup(1)
  !TDDij_NewMix = Backup(2)
!TD
  !TDClose(TotalCurrent_Unit)
  !TDClose(E_Unit)
!TD
  !TDWrite(Log_Unit,*) 'RelaxTime: Completed Run!'
!TD
  !TDAtomic_Mode = SIM_PAW
!TD
  !TDReturn
!TDEnd Subroutine


!******************************************************************************
!
! LDRelax - Processes the RELAX statment which has the form
!
!  RELAX CHARGE  MaxIter Tol  ---- Relax the electrons
!  RELAX [ATOMS|GEOMETRY] MaxSteps ChargeBase MaxIter ChargeTol ForceTol &
!                        MaxMove ConditionNo
!  RELAX TIME Num_Steps 
!
!******************************************************************************

Subroutine LdRelax(WC)
  Type (Word_Context), Intent(INOUT) :: WC

  Character*100 :: Token, FileName, ChargeBase
  Integer       :: tlen, MaxIter, MaxSteps, mode
  Real  :: Tol, Etol, ForceTol, Bo, MaxMove, ConditionNo
  Logical :: EConverged

  Call GetNextWord(WC, Token, tlen)
  If (W_Error /= W_OK) then
     Write(Error_Unit, *) 'LdRelax: Error occured while processing command!'
     Call Word_GetAndPrint(WC, Error_Unit, "LdRelax:")
     STOP
  End If

  Call UpperCase(token)
  If (trim(Token) == 'CHARGE') then
     Call GetNumbers(WC, MaxIter)
     Call GetNumbers(WC, Tol)

     if (Print_Level == PRINT_VERBOSE) then
        write(LOG_Unit, *) "ldRelax: Relaxing Charge:  MaxIter=",MaxIter, &
             ' * Tolerance=',Tol
        Write(LOG_Unit,*) 'ldRelax: Starting Cohesive Energy :', CohesiveEnergy
     End IF

     If (PAW_FirstTime) then
        PAW_FirstTime = .FALSE.
     End if

     write(log_unit,*) 'LdRelax: starting SCF interations', MaxIter,Tol
     call flush(Log_unit)
     Call RelaxElectrons(MaxIter, Tol, Etol, EConverged) 

     if (Print_Level == PRINT_VERBOSE) then
        write(LOG_Unit, *) "ldRelax: Completed Charge relaxation."
     End IF
  else if ((Trim(Token) == 'ATOMS') .OR. (Trim(Token) == 'GEOMETRY')) then
     If (trim(Token) == 'ATOMS') then
        Mode = RELAX_MD
     else
        Mode = RELAX_GEOMETRY
     End If

     Call GetNumbers(WC, MaxSteps)
!*     Call GetNextWord(WC, Filename, tlen)
     Call GetNextWord(WC, ChargeBase, tlen)
     Call GetNumbers(WC, MaxIter)
     Call GetNumbers(WC, Tol)
     Call GetNumbers(WC, ForceTol)
     Call GetNumbers(WC, MaxMove)
     Call GetNumbers(WC, ConditionNo)

     if (Print_Level == PRINT_VERBOSE) then
        write(LOG_Unit, *) "ldRelax: Relaxing Atoms:  MaxSteps=", MaxSteps, &
             ' * MaxChargeITer=',MaxIter, ' * Force Tolerance=', ForceTol,  &
             ' * Charge Tolerance=',Tol, ' * Max Move=', MaxMove, &
             ' * Condition Number = ', ConditionNo
        Write(LOG_Unit,*) 'ldRelax: Starting Cohesive Energy :', CohesiveEnergy
     End IF


     Call RelaxAtoms(mode,MaxSteps, TRIM(ChargeBase), MaxIter, Tol,ForceTol,&
         MaxMove,ConditionNo)
  !TDelse if (TRIM(Token) == "TIME") then
     !TDCall GetNumber(WC, MaxSteps)
     !TDCall RelaxTime(MaxSteps)
  else
     Write(Error_Unit,*) 'LdRelax: Unknown RELAX option: ',TRim(Token)
  End If

  Return
End Subroutine

!******************************************************************************
!
!  LdBandStruct - Loads the Band structure information from disk and
!        processes it.
!
!******************************************************************************

Subroutine LdBandStruct(WC)
  Type (Word_Context), Intent(INOUT) :: WC

  Character*100 :: Token, FileName, ChargeBase
  Integer       :: tlen, MaxIter, i,j,k, Kindex, NumK, NumB
  Real  :: Tol, Etol, Kpnt(3), FracKpnt(3),ENK  
  Type (Word_Context) :: Kpnt_WC
  Logical :: EConverged

  bandstructure_mode=.true.
  ReCalculate_Occupancies=.true.

  Call GetNextWord(WC, Filename, tlen)
  If (W_Error /= W_OK) then
     Write(Error_Unit, *) 'LdBandStruct: Error occured while processing command!'
     Call Word_GetAndPrint(WC, Error_Unit, "LdBandStruct:")
     STOP
  End If

  Band_Name=Trim(FileName)//'.band'

  open(unit=Band_Out,file=Band_name,form='formatted',position='append')

  Call GetNumbers(WC, MaxIter)
  Call GetNumbers(WC, Tol)
  
  if (Print_Level == PRINT_VERBOSE) then
     write(LOG_Unit, *) "ldBandStruct: Relaxing Charge:  MaxIter=",MaxIter, &
          '* Tolerance=',Tol, ' * K-point File:', TRIM(FileName)
  End IF

  If (.not.HamLoaded) then
     write(Error_Unit,*) ' Error in Band Structure Calculation -- ', &
       'Must first calculate or load Hamiltonian'
     stop
  EndIf

  Call InitWord(Kpnt_WC, BAND_IN, Filename, DELIMS, COMMENT, LIT_CHAR, &
       1, INCLUDE_STR)


  !*** Create the New K-points storage space ***
  NumKpnts = 1  ; NumK=1
  BZ%TotalUniq = NumKpnts
  BZ%TotalKpnts = NumKpnts

write(log_unit,*) 'ldbandstruct -- before allocations';call flush(Log_Unit)
  If (Associated(bz%ku)) Nullify(Bz%ku) 
  If (Associated(bz%keu)) Nullify(Bz%keu) 
  If (Associated(bz%wtuniq)) Nullify(Bz%wtuniq) 
  If (Associated(bz%weight)) Nullify(Bz%weight)
  If (Associated(xtal%wt)) Nullify(xtal%wt) 
  If (Associated(occupancy)) Nullify(occupancy) 
  If (Associated(User_Kpoints)) Nullify(User_Kpoints) 
 if(spindependence) then
  Allocate(BZ%Ku(3,1), xtal%Wt(1), BZ%Keu(2*NumBands,1), &
       BZ%WtUniq(2*NumBands, 1), BZ%Weight(2*NumBands, 1), &
       Occupancy(2*NumBands, NumKpnts),  User_Kpoints(3,1), stat=i)
 else
  Allocate(BZ%Ku(3,1), xtal%Wt(1), BZ%Keu(NumBands,1), &
       BZ%WtUniq(NumBands, 1), BZ%Weight(NumBands, 1), &
       Occupancy(NumBands, NumKpnts),  User_Kpoints(3,1), stat=i)
 endif
write(log_unit,*) 'ldbandstruct -- after allocations',i;call flush(Log_Unit)
  If (i /= 0 ) then
      write(Error_Unit,*) 'Error in LdBandStructure --' ,&    
        ' Cannot reallocate kpoint arrays ',i
      stop
  EndIf
  BZ%Kp => BZ%Ku;  
  BZ%Weight(:,1) = 1;  !***BZ%Weight(:,2) = 0;
  BZ%WtUniq = BZ%Weight

  BZ%Kp = 0
  User_Kpoints = 0


  !*** Now Get the maximum number of bands for processing
  Call GetNumbers(Kpnt_WC, NumB)      ! Maximum number of bands to calculate

  Write(Log_Unit,*) 'LdBandStruct: Processing ',NumK, ' k-points',NumB,' bands'
  call flush(Log_Unit)

  Call PrepPrevious_band(NumB)

  !*** Finally process the K's ****
  i=1
  Kpnt = 0
  Kindex=1
  Do 
     Call GetNumbers(Kpnt_WC, FracKpnt)
     If (W_ERROR /= W_OK) exit

     Write(Log_Unit,*) 'LdBandStruct: Loop=',i, ' * FracKpnt=',FracKpnt
     call flush(Log_Unit)

     Kpnt(:) = MATMUL(xtal%Recip, FracKpnt)
     User_Kpoints(:,Kindex) = FracKpnt
     Call UpdateKpoint(Kpnt)
     
     Call RelaxElectrons(MaxIter, Tol, Etol, EConverged) 

     !*** Write the info to the OUTPUT file ***
     Write(Output_Unit,*) 'LdBandStruct: Loop:',i, ' * FracKpnt=',FracKpnt

     k = 0
     Do j=1, Mem_MapSize
        If (PsiInfo(j)%Kpnt>0) then
           k = k + 1
           ENK=PsiInfo(j)%Energy
           Write(Output_Unit,'(2i10,1pe15.7,1x,l1,1x,1pe15.7)')&
               k, j, ENK, PsiInfo(j)%spinup,PsiInfo(j)%Error
           If (ENK <= Eigen_Max) & 
           Write(Band_Out,'(1p3e15.7,2x,1pe15.7,1x,l1)')&
                 Frackpnt,ENK, PsiInfo(j)%spinup
        End If
     End Do

     Call Flush(Band_Out)
 
     i=i+1

  End Do

   Write(Log_Unit,*) 'Ending band structure calculation after processing ', &
     i-1,' k-points'
   Call Flush(Log_Unit)
    

  Call CloseWord(Kpnt_WC)
  bandstructure_mode=.false.

  Return
End Subroutine

!******************************************************************************
!
!  LdSurfBands - Loads the Surface bandstructure info and 
!                  Calculates surface bands
!
!******************************************************************************

Subroutine LdSurfBands(WC)
  Type (Word_Context), Intent(INOUT) :: WC

  Character*100 :: Token, FileName, ChargeBase
  Integer       :: tlen, MaxIter, i,j,k, Kindex, NumK, NumB,Flag
  Integer :: Nnormal,kp,Psi_toUse(Mem_MapSize)
  Real  :: Tol,Etol,  Kpnt(3), FracKpnt(3),ENK  ,GradEnkv(3),Gperp(3)
  Real :: deltaGp(3),chk,Ctof(3,3),deltaEnk,Gmag,deltaGmag
  Type (Word_Context) :: Kpnt_WC
  Logical :: EConverged
  Type (Mem_handle), Pointer :: Psi

  bandstructure_mode=.true.
  ReCalculate_Occupancies=.true.

  Ctof=TRANSPOSE(xtal%Basis)/(2*Pi)  ! for calculating fractional k's

  Call GetNextWord(WC, Filename, tlen)
  If (W_Error /= W_OK) then
     Write(Error_Unit, *) 'LdSurfBands: Error occured while processing command!'
     Call Word_GetAndPrint(WC, Error_Unit, "LdSurfBands:")
     STOP
  End If
  Band_Name='RECIP_VECTOR'
  open(unit=Band_Out,file=Band_name,form='formatted')
  Write(Band_OUT,'("GA   ",1p3e20.12)') xtal%Recip(1:3,1)
  Write(Band_OUT,'("GB   ",1p3e20.12)') xtal%Recip(1:3,2)
  Write(Band_OUT,'("GC   ",1p3e20.12)') xtal%Recip(1:3,3)
  close(Band_OUT)

  Band_Name=Trim(FileName)//'.surfband'

  open(unit=Band_Out,file=Band_name,form='formatted',position='append')

  Call GetNumbers(WC, MaxIter)
  Call GetNumbers(WC, Tol)
  
  if (Print_Level == PRINT_VERBOSE) then
     write(LOG_Unit, *) "ldBandStruct: Relaxing Charge:  MaxIter=",MaxIter, &
          '* Tolerance=',Tol, ' * K-point File:', TRIM(FileName)
  End IF

  If (.not.HamLoaded) then
     write(Error_Unit,*) ' Error in Surface Band Calculation -- ', &
       'Must first calculate or load Hamiltonian'
     stop
  EndIf

  Call InitWord(Kpnt_WC, BAND_IN, Filename, DELIMS, COMMENT, LIT_CHAR, &
       1, INCLUDE_STR)


  !*** Create the New K-points storage space ***
  NumKpnts = 1
  BZ%TotalUniq = NumKpnts
  BZ%TotalKpnts = NumKpnts

write(log_unit,*) 'ldbandstruct -- before allocations';call flush(Log_Unit)
  If (Associated(bz%ku)) Nullify(Bz%ku) 
  If (Associated(bz%keu)) Nullify(Bz%keu) 
  If (Associated(bz%wtuniq)) Nullify(Bz%wtuniq) 
  If (Associated(bz%weight)) Nullify(Bz%weight)
  If (Associated(xtal%wt)) Nullify(xtal%wt) 
  If (Associated(occupancy)) Nullify(occupancy) 
  If (Associated(User_Kpoints)) Nullify(User_Kpoints) 


 If (spindependence) then
  Allocate(BZ%Ku(3,1), xtal%Wt(1), BZ%Keu(2*NumBands,1), &
       BZ%WtUniq(2*NumBands, 1), BZ%Weight(2*NumBands, 1), &
       Occupancy(2*NumBands, 2*NumKpnts),  User_Kpoints(3,1), stat=i)
 else
  Allocate(BZ%Ku(3,1), xtal%Wt(1), BZ%Keu(NumBands,1), &
       BZ%WtUniq(NumBands, 1), BZ%Weight(NumBands, 1), &
       Occupancy(NumBands, NumKpnts),  User_Kpoints(3,1), stat=i)
 endif

write(log_unit,*) 'ldsurfbands -- after allocations',i;call flush(Log_Unit)
  If (i /= 0 ) then
      write(Error_Unit,*) 'Error in LdSurfBands --' ,&    
        ' Cannot reallocate kpoint arrays ',i
      stop
  EndIf
  BZ%Kp => BZ%Ku;  
  BZ%Weight(:,1) = 1;  !***BZ%Weight(:,2) = 0;
  BZ%WtUniq = BZ%Weight

  BZ%Kp = 0
  User_Kpoints = 0

 !Define Gperp -- read in Fractional units
     Call GetNumbers(Kpnt_WC, FracKpnt)
     If (W_ERROR /= W_OK) then
        Write(Error_Unit,*) 'error in ldsurfband: Need Gperp'
        stop
     Endif
 !Define Nnormal -- # number of perpendicular values of K for each K||
     Call GetNumber(Kpnt_WC, Nnormal)
     If ((W_ERROR /= W_OK) .OR. (Nnormal < 1)) then
        Write(Error_Unit,*) 'error in ldsurfband: Need Nnormal'
        stop
     Endif

     Gperp(:) = MATMUL(xtal%Recip, FracKpnt)
     DeltaGp(:) = Gperp(:)/Nnormal

  Write(Log_Unit,*) 'ldSurfBands: Nnormal = ',Nnormal
  Write(Log_Unit,*) 'ldSurfBands: Gperp   = ',Gperp  
  Write(Log_Unit,*) 'ldSurfBands: DeltaGp = ',DeltaGp
     deltaGmag=SQRT(DOT_PRODUCT(DeltaGp,DeltaGp))
  Write(Log_Unit,*) 'ldSurfBands: DeltaGmag = ',DeltaGmag
  Call Flush(Log_Unit)

  !*** Now Get the maximum number of bands for processing
  Call GetNumbers(Kpnt_WC, NumB)      ! Maximum number of bands to calculate

  Write(Log_Unit,*) 'LdSurfBands: Processing ',NumB,' bands'
  call flush(Log_Unit)

  Call PrepPrevious_band(NumB)

  !!!********Initialize Gradient module**************!!!
  Call InitGradEnk

  !*** Finally process the K||'s ****
  i=1
  Kpnt = 0
  Kindex=1
  Do 
     Call GetNumbers(Kpnt_WC, FracKpnt)
     If (W_ERROR /= W_OK) exit

     Write(Log_Unit,*) 'LdSurfBands: Loop=',i, ' * FracKpnt=',FracKpnt
     call flush(Log_Unit)
     Write(Output_Unit,'("LdSurfBands: i, * FracK||=",i5,3f10.5)') i,FracKpnt
     call flush(Output_Unit)

     Kpnt(:) = MATMUL(xtal%Recip, FracKpnt)
! Check that K|| is perpendicular to Gperp
     chk=DOT_PRODUCT(Kpnt,Gperp)
      If (ABS(chk).gt.1.e-5) then
         write(Error_Unit,*) &
            'Error in SurfBands -- K|| ,Gperp invalid',&
             Kpnt,Gperp,chk
         stop
      EndIf

      Kpnt(:) = Kpnt(:) - DeltaGp(:)/2
      Gmag = -deltaGmag/2

      Do  kp=1,Nnormal
          Kpnt(:) = Kpnt(:) + DeltaGp(:)
          Gmag = Gmag + deltaGmag
   
          User_Kpoints(:,Kindex) = MATMUL(Ctof,Kpnt)

          write(Log_Unit,*) 'SurfBand -- ', FracKpnt
          write(Log_Unit,*) 'SurfBand -- ', kp, Kpnt     
          write(Log_Unit,*) 'SurfBand -- ', kp, User_Kpoints(:,Kindex)

          Call UpdateKpoint(Kpnt)
     
          Call RelaxElectrons(MaxIter, Tol, Etol, EConverged) 

          !*** Write the info to the OUTPUT file ***
          Write(Output_Unit,'("    kp, kpnt= ", i5,3f10.5)') kp,Kpnt

          Call NewKforGradEnk(Kpnt)
          Psi_toUse=MH_Skip

          Where ((PsiInfo(:)%Kpnt > 0)) Psi_toUse=MH_toProcess
          Call Phase_Generic(Psi_toUse,Kindex)
          Call GetNextPsi(Flag,Psi)

          k = 0
          Do while (Flag > 0)
             j=Psi%Index
             ENK=PsiInfo(j)%Energy
             If (ENK <= Eigen_Max) then
                k = k + 1
                Write(Output_Unit,&
                 '("     k j ENK spin ERR",2i5,1p2e15.7,1x,l1,1x,1pe15.7)')&
                     k, j, ENK, PsiInfo(j)%spinup, PsiInfo(j)%Error
                Call CalcGradEnk(ENK, Psi%ptr, PsiInfo(j)%PDOT,GradEnkv)
                deltaEnk=DOT_PRODUCT(GradEnkv,DeltaGp)
                Write(Output_Unit, '(" Grad, DeltaE = ",1p4e15.7,1x,l1)') &
                       GradEnkv, DeltaEnk,PsiInfo(j)%spinup
                Write(Output_Unit,'(">>>>",1p3e15.7,1x,l1)')&
                       Gmag,Enk,deltaEnk,PsiInfo(j)%spinup
                Write(Band_Out,'(5f10.5,1x,l1)')& 
                     Frackpnt,ENK-DeltaEnk/2,ENK+DeltaEnk/2,&
                               PsiInfo(j)%spinup
                Call Flush(Band_Out)
                Call Flush(Output_Unit)
             End If
          Call GetNextPsi(Flag,Psi)
          End Do

       EndDo    !kp
 
     i=i+1

  End Do

   Write(Log_Unit,*) 'Ending surface band calculation after processing ', &
     i-1,' k-points'
   Call Flush(Log_Unit)
    

  Call FinishGradEnk
  Call CloseWord(Kpnt_WC)
  bandstructure_mode=.false.

  Return
End Subroutine
End Module
