!******************************************************************************
!
! File : search_sort.f90
!   by : Alan Tackett
!   on : 11/17/95
!  for : General Routines for Searching and Sorting arrays
!
! Searching Routines
!      Linear_Search - Performs a linear search
!      Binary_Search - Performs a binary search
!
! Sorting Routines
!      Insertion_Sort - Sorts an array using the Insertion Sort method
!
!
!******************************************************************************

Module search_sort

implicit none

Interface Insertion_Sort
   Module Procedure Real_InsSort
   Module Procedure Integer_InsSort
End Interface


Contains

!******************************************************************************
!
!  Linear_Search - Performs a linear search of an array and returns
!                  the index of the closest matching element
!
!  A          - Array of values to search
!  MatchValue - Value to match
!  Exact      - Upon return it contains TRUE if an exact match was found
!               and FALSE otherwise. (OPTIONAL)
!
!  Return Values
!       The closest match, index is returned. If EXACT is
!       supplied then it determines whether an exact match was found.
!
!******************************************************************************

Integer Function Linear_Search(A, MatchValue, Exact)
  Integer,           Intent(IN)  :: A(:)
  Integer,           Intent(IN)  :: MatchValue
  Logical, OPTIONAL, Intent(OUT) :: Exact

  Logical  :: Found
  Integer  :: i, BestIndex, A_Size
  Integer  :: BestValue, BestMiss

  A_Size = Size(A)
  BestValue = A(1)
  BestIndex = 1
  BestMiss = abs(BestValue - MatchValue)
  i = 2
  Found = .FALSE.

  Do While ((i<=A_Size) .AND. (BestValue /= MatchValue))
    if (abs(A(i) - MatchValue) < BestMiss) then
       BestIndex = i
       BestValue = A(i)
       BestMiss = Abs(BestValue - MatchValue)
    End IF

    i = i + 1
  End Do

  if (Present(Exact)) then
    If (BestValue == MatchValue) then
       Exact = .TRUE.
    Else
       Exact = .FALSE.
    End If
  End If

  Linear_Search = BestIndex
  Return
End Function  


!******************************************************************************
!
!  Linear_Search - Performs a binary search of an array and returns
!                  the index of the closest matching element that
!                  is LESS THAN OR EQUAL to the element.
!
!  A          - Array of values to search
!  MatchValue - Value to match
!  Exact      - Upon return it contains TRUE if an exact match was found
!               and FALSE otherwise. (OPTIONAL)
!
!  Return Values
!       The closest match, index is returned. If EXACT is
!       supplied then it determines whether an exact match was found.
!
!******************************************************************************

Integer Function Binary_Search(A, MatchValue, Exact)
  Integer,           Intent(IN)  :: A(:)
  Integer,           Intent(IN)  :: MatchValue
  Logical, OPTIONAL, Intent(OUT) :: Exact

  Integer  :: A_Size
  Integer  :: HiIndex, LoIndex, i

  A_Size = Size(A)
  If (A(1) > A(A_Size)) then
     HiIndex = 1
     LoIndex = A_Size
  else
     HiIndex = A_Size
     LoIndex = 1
  End If

  i = (HiIndex + LoIndex) / 2
  if (A(HiIndex) <= MatchValue) then
     i = HiIndex
     LoIndex = i
  else if (A(LoIndex) >= MAtchValue) then
     i = LoIndex
     HiIndex = i
  End If
  
  Do While ((HiIndex /= LoIndex) .AND. (A(i) /= MAtchValue))
    if (A(i) > MatchValue) then
       HiIndex = i
    else
       LoIndex = i
    End If

    i = (HiIndex + LoIndex) / 2
  End Do

  if (Present(Exact)) Exact = .FALSE.

  if (A(i) == MatchValue) then
    if (Present(Exact)) Exact = .TRUE.
  End If

  Binary_Search = i
  Return
End Function

!******************************************************************************
!
!  Insertion_Sort - Sorts an Array using an Insertion sort method.
!                   Best used for small arrrays or arrays that are almost
!                   sorted.
!
!  A         - Array to Sort
!  LUT       - Index table containing the new sorted order(RETURNED)
!  Ascending - Sort Direction.  If TRUE then sorted in ascending order
!              otherwise the array is sorted in descending order.
!
!
!******************************************************************************

Subroutine Integer_InsSort(A, LUT, Ascending)
  Integer, Intent(IN)    :: A(:)
  Integer, Intent(INOUT) :: LUT(:)
  Logical, Intent(IN)    :: Ascending

  Integer :: i, j, k, A_Size
  Integer :: CurrVal

  A_Size = Size(A)
  
  Do i=1, A_Size
    LUT(i) = i
  End Do

  Do i = 2, A_Size
    CurrVal = A(LUT(i))
    j = i - 1

    Do While ((CurrVal < A(LUT(j))) .AND. (j>1))
      LUT(j+1) = LUT(j)
      j = j - 1
    End Do

    If (CurrVal < A(LUT(j))) then
       LUT(j+1) = LUT(j)
       j = j - 1
    End If

    LUT(j+1) = i
  End Do

  If (.NOT. Ascending) then
    j = A_Size / 2

    Do i = 1, j
       k = LUT(i)
       LUT(i) = LUT(A_Size - i + 1)
       LUT(A_Size - i + 1) = k 
    End Do
  End If

  Return
End Subroutine

!*******

Subroutine Real_InsSort(A, LUT, Ascending)
  Real,    Intent(IN)    :: A(:)
  Integer, Intent(INOUT) :: LUT(:)
  Logical, Intent(IN)    :: Ascending

  Integer :: i, j, k, A_Size
  Real    :: CurrVal

  A_Size = Size(A)

  Do i=1, A_Size
    LUT(i) = i
  End Do

  Do i = 2, A_Size
    CurrVal = A(LUT(i))
    j = i - 1

    Do While ((CurrVal < A(LUT(j))) .AND. (j>1))
      LUT(j+1) = LUT(j)
      j = j - 1
    End Do

    If (CurrVal < A(LUT(j))) then
       LUT(j+1) = LUT(j)
       j = j - 1
    End If

    LUT(j+1) = i
  End Do

  If (.NOT. Ascending) then
    j = A_Size / 2

    Do i = 1, j
       k = LUT(i)
       LUT(i) = LUT(A_Size - i + 1)
       LUT(A_Size - i + 1) = k 
    End Do
  End If

  Return
End Subroutine

End Module


