!******************************************************************************
!
! File : spherical_harmonic.f90
!   by : Natalie Holzwarth)
!   on : 6/14/99     
!  for : generic fuctions; will be used in PAW method
!
!  This file contains several generic functions used when working with 
!  spherical harmonics.
!
!  Modified on 2/17/97 by Alan Tackett so that Spharm returns the complex Ylm  
!
!******************************************************************************

module spherical_harmonic

Use Misc
Use mathlib

implicit none

  integer, private :: nang

Contains

!******************************************************************************
!
! anginit: Generates coordinates and weight factors for spherical integration
!          points generated on the unit sphere. The integral is over a unit
!          sphere normalized to one.
!
! IN : ntheta     - Number of sample points in the theta dir
!      nphi       -   "     "   "      "     "  "  phi    "
!      vol_crystal- the volume of the crystal
!      Error_Unit - Output unit if an error occrs(OPTIONAL)
!
! OUT: anginit-     nang x 4 array, first three rows were formely the nang-
!                   dimensional arrays ax, ay, and az; the fourth row was
!                   foremly the nang-dimensional array aw. 
!
!*****************************************************************************

function anginit(ntheta, nphi, vol_crystal, Error_Unit)
  implicit none
  Real, pointer       :: anginit(:,:)

! Dummy argument declarations

  integer, intent(in)           :: ntheta, nphi
  Real,    intent(in)           :: vol_crystal
  Integer, Optional, Intent(IN) :: Error_Unit

! Declarations

  integer             :: it, ip, npoints, index
  Real                :: wth(ntheta), th(ntheta)
  Real                :: enpt(2)
  Real                :: temp_ax(nang)
  Real                :: con, cos_theta, sin_theta, ang, cos_phi, sin_phi
  Real                :: pref
! (defined in mathlib) Real, parameter     :: two_pi = 2*pi, four_pi = 4*pi

! Initialize nang

  nang = ntheta * nphi

  Allocate(AngInit(4, Nang), STAT=it)

  if (Present(Error_Unit)) then
    Call Check_Error(it, "ERROR allocating anginit", Error_Unit, .TRUE.)
  else
    Call Check_Error(it, "ERROR allocating anginit", 7, .TRUE.)
  End If
! The following call to ~natalie/debug/paw/gaussq.f provides us with enpt, th,
! wth, and temp_ax. In Natalie's original code, temp_ax was ax, an nang-dim.
! array. Here we receive temp_ax, which is then copied into anginit via a do-
! loop. This may not be the most efficient way, although it is the one that 
! immediately presents itself.

 Call GaussLeg(-1.0, 1.0, ntheta, th, wth)

! Initialize npoints, pref, and con
  
  npoints = 0

  pref = four_pi / vol_crystal

  con = two_pi / nphi

! We now open two nested do-loops. The first loops through the number of theta
! angles, the second through the number of phi angles (?). The two together 
! initialize anginit.

  do it = 1, ntheta

     cos_theta = th(it)
     sin_theta = sqrt(1.0 - cos_theta*cos_theta)

     do ip = 1, nphi

        ang = con * (ip-1)        
        cos_phi = cos(ang)
        sin_phi = sin(ang)
        
        npoints = npoints + 1

        anginit(1, npoints) = sin_theta * cos_phi
        anginit(2, npoints) = sin_theta * sin_phi
        anginit(3, npoints) = cos_theta

! Normalization requires dividing by 4-pi.

        anginit(4, npoints) = wth(it) / (2 * nphi)

!**        anginit(4, npoints) = wth(it) * four_pi

!** Divide anginit by an extra factor of 4-pi/vol_crystal.
!**        anginit(4, npoints) = pref * anginit(4, npoints)

!Write(*,*) 'anginit: i=',npoints, ' * Wt=',anginit(4,npoints)
!Write(*,*) 'anginit:    Ra=',anginit(1:3,npoints)
        end do
     end do

! The following is an error statement that will be generated if npoints exceeds
! nang

     if(npoints > nang) write(6, *) "ERROR: anginit-- npoints = ",npoints,&
                                     ", nang = ", nang

end function anginit


!******************************************************************************
!
! spharm: Returns spherical harmonic functions for l in the order
!         m=-l,...0,...l in Condon-Shortley convention
!
! IN : x, y, z- These are the x, y, and z cartesian coordinates            
!      l-       This  is the normal l-value in quantum mechanics
!      DoNorm-  If .false., spharm returns r**l times spherical harmonic
!
! OUT: spharm-  A thirteen-member array containing the spherical harmonic
!               values.
!
!
!******************************************************************************

function spharm(x, y, z, L, DoNorm,Error_Unit)
  implicit none

  Complex :: spharm(13)

! Dummy argument declarations

  integer, intent(IN):: L
  Real, intent(IN)   :: x, y, z
  Logical, Intent(IN):: DoNorm
  integer, optional, Intent(IN) :: Error_Unit

! Local declarations

  Real               :: Ylm_Norm, r, costheta, cosphi, sinphi , tmp
  Real, parameter    :: tol = 1.e-13
  Integer    :: m,im
  Complex    :: phase

! Initialize spharm

  Ylm_Norm= sqrt((2*L+1)/Four_Pi)
  spharm = 0

  r=sqrt(x*x+y*y+z*z)

  If (R.lt.tol) then
     If (L==0) then
        Spharm(1) = Ylm_norm
     else
        Spharm = 0
     End If
     
     RETURN
  End If

  costheta=z/r
  cosphi=1
  sinphi=0
  tmp=sqrt((1-costheta)*(1+costheta))
  if (tmp.gt.tol) then
     cosphi=x/(r*tmp)
     sinphi=y/(r*tmp)
  end if
  
  select case (L)
    case (0)
       spharm(1)=Ylm_Norm
    case (1:6)
       spharm(L+1)=Ylm_Norm*AssosP(L,0,costheta)
       phase=CMPLX(cosphi,sinphi)
       tmp=1.0/(L*(L+1))
       Do m=1,L
          spharm(L+1+m)=Ylm_Norm*sqrt(tmp)*AssosP(L,m,costheta)*(phase**m)
          spharm(L+1-m)=((-1)**m)*CONJG(spharm(L+1+m))
          if (m.lt.L) tmp=tmp/((L+m+1)*(L-m))
       End do
    case default
       if (present(Error_Unit)) write(Error_Unit,*) 'error in spharm -- L =',L
       stop       
  end select

  if (.not.donorm) then
     tmp=r**L
     spharm=spharm*tmp
  endif
  
end function spharm

!******************************************************************************
!
! GradYlm:  Returns angular components of gradient of 
!          spherical harmonic functions for l in the order
!         m=-l,...0,...l in Condon-Shortley convention
!
! IN : x, y, z- These are the x, y, and z cartesian coordinates            
!      l-       This  is the normal l-value in quantum mechanics
! OUT: dYlmdtheta, DYlmdphi  Two  thirteen-member arrays
!              containing the spherical harmonic derivatives
!
!******************************************************************************

Subroutine GradYlm(x, y, z, L,dYlmdtheta,dYlmdphi, Error_Unit)
  implicit none

  Complex , intent(OUT):: dYlmdtheta(13),dYlmdphi(13)
  integer, intent(IN):: L
  Real, intent(IN)   :: x, y, z
  integer, optional, Intent(IN) :: Error_Unit

! Local declarations

  Real               :: Ylm_Norm, r, costheta, cosphi, sinphi , tmp
  Real, parameter    :: tol = 1.e-13
  Integer    :: m,im
  Complex    :: phase,iphase

! Initialize

  Ylm_Norm= sqrt((2*L+1)/Four_Pi)
  dYlmdtheta=0 ; dYlmdphi=0

  r=sqrt(x*x+y*y+z*z)

  If (L == 0 .or. r < tol) return

  costheta=z/r
  cosphi=1
  sinphi=0
  tmp=sqrt((1-costheta)*(1+costheta))
  if (tmp.gt.tol) then
     cosphi=x/(r*tmp)
     sinphi=y/(r*tmp)
  end if
  
  select case (L)
    case (1:6)
       dYlmdtheta(L+1)=Ylm_Norm*Der_Theta_P(L,0,costheta)
       dYlmdphi(L+1)=0
       phase=CMPLX(cosphi,sinphi)
       iphase=CMPLX(-sinphi,cosphi)
       tmp=1.0/(L*(L+1))
       Do m=1,L
          dylmdtheta(L+1+m)=&
            Ylm_Norm*sqrt(tmp)*Der_Theta_P(L,m,costheta)*(phase**m)
          dylmdtheta(L+1-m)=((-1)**m)*CONJG(dylmdtheta(L+1+m))
          dylmdphi(L+1+m)=&
            Ylm_Norm*sqrt(tmp)*Der_Phi_P(L,m,costheta)*iphase*(phase**(m-1))
          dylmdphi(L+1-m)=((-1)**m)*CONJG(dylmdphi(L+1+m))
          if (m.lt.L) tmp=tmp/((L+m+1)*(L-m))
       End do
    case default
       if (present(Error_Unit)) write(Error_Unit,*) 'error in spharm -- L =',L
       stop       
  end select

end subroutine GradYlm
!******************************************************************************
!
! ylmlist: Returns spherical harmonic functions for l in the order
!         m=-l,...0,...l in Condon-Shortley convention and for 
!         l=0,.. lmax
!
! IN : x, y, z- These are the x, y, and z cartesian coordinates
!      lmax
!      error_unit
!
! OUT: Ylm(13,lmax+1) -array containing the spherical harmonic
!               values.
!
!******************************************************************************
Subroutine Ylmlist(x, y, z, lmax, ylm,error_unit)
  implicit none

  Real, intent(IN)   :: x, y, z
  integer, intent(IN):: lmax,error_unit
  Complex, INTENT(OUT) :: ylm(:,:)

! Local declarations

  Real               :: Ylm_Norm, r, costheta, cosphi, sinphi , tmp
  Real, parameter    :: tol = 1.e-13
  Integer    :: m,im ,L
  Complex    :: phase

  ylm = 0

  if (lmax.lt.0.or.lmax.gt.6) then   
   write(error_unit,*) 'error in Ylmlist',lmax,x,y,z
   stop
  endif

  ylm(1,1)=sqrt(1/Four_Pi)

  r=sqrt(x*x+y*y+z*z)

  If (r.lt.tol.or.lmax.eq.0) return

  costheta=z/r
  cosphi=1
  sinphi=0
  tmp=sqrt((1-costheta)*(1+costheta))
  if (tmp.gt.tol) then
     cosphi=x/(r*tmp)
     sinphi=y/(r*tmp)
  end if

  phase=CMPLX(cosphi,sinphi)

  do L=1,lmax

  Ylm_Norm= sqrt((2*L+1)/Four_Pi)
     ylm(L+1,l+1)=Ylm_Norm*AssosP(L,0,costheta)
     tmp=1.0/(L*(L+1))
     Do m=1,L
        ylm(L+1+m,l+1)=Ylm_Norm*sqrt(tmp)*AssosP(L,m,costheta)*(phase**m)
        ylm(L+1-m,l+1)=((-1)**m)*CONJG(ylm(L+1+m,l+1))
        if (m.lt.L) tmp=tmp/((L+m+1)*(L-m))
     End do

  Enddo
  
end subroutine ylmlist

!******************************************************************************
!
! gaunt- Returns the integral of Y*lm Y*llmm Yl'm' x sqrt(4pi). See the 3-j and
!        6-j symbols by Rotenberg, etc., (Technology Press, 1959), pg.5.
!
! IN : L, M, l1, l2, m1, m2- quantu numbers
!
! OUT: gaunt- the value of the integral.
!
!******************************************************************************
 
Real function gaunt(L,M,l1,m1,l2,m2)

  implicit none

  integer, intent(IN) :: L, M, l1, l2, m1, m2

  integer             :: j, j1, j2, j3, j1half, j2half, j3half, j_half
  integer             :: k, k1, k2, i1, i2, n1, n2
  Real                :: xx,yy, sum, sign, argument
  logical             :: ok
  
  gaunt = 0
  sum = 0

  ok = .true.

  if((-m1-M+m2) /= 0) ok = .false.

  if(abs(m1) > l1) ok = .false.
  if(abs(M) > L) ok = .false.
  if(abs(m2) > l2) ok = .false.

  j = l1 + L + l2

  if(mod(j,2) /= 0) ok = .false.

  j1 = j-2*L2
  j2 = j-2*L
  j3 = j-2*L1
  
  if(j1 < 0 .or. j2 < 0 .or. j3 < 0) ok = .false.

  if (ok) then

     xx = (2 * l1 + 1) * (2 * L + 1) * (2 * l2 + 1)

     j1half = j1/2
     j2half = j2/2
     j3half = j3/2
     j_half = j/2

     gaunt = (-1)**j1half * sqrt(xx)
     gaunt = gaunt * factorial(j2) * factorial(j3) / factorial(j+1)
     gaunt = gaunt * factorial(j_half)/(factorial(j1half)&
          &*factorial(j2half)*factorial(j3half))

     yy = factorial(l2 + m2) * factorial(l2 - m2)

     if (M >= 0) then
        yy = yy * Permutations(L+M, 2*M)
     else
        yy = yy / Permutations(L-M, -2*M)
     End if

     if (m1 >= 0) then
        yy = yy / Permutations(l1+m1, 2*m1)
     else
        yy = yy * Permutations(l1-m1, -2*m1)
     End if

     gaunt = gaunt * sqrt(yy)

     i1 = l2 - L - m1
     i2 = l2 - l1 + M
     k1 = -min(0, i1, i2)
  
     n1 = l1 + m1
     n2 = L - M
     k2 = min(j1, n1, n2)
  
     sign = 1.0
     if(k1 > 0) sign = (-1)**k1

     argument = sign * Permutations(n1, k1)/factorial(k1)
     argument = argument * Permutations(n2, k1)/factorial(i1 + k1)
     argument = argument * Permutations(j1, k1)/factorial(i2 + k1)
     sum = sum + argument
     
     sign = -sign
     k1 = k1 + 1

     do while(k1 <= k2)
        argument = sign * Permutations(n1, k1)/factorial(k1)
        argument = argument * Permutations(n2, k1)/factorial(i1 + k1)
        argument = argument * Permutations(j1, k1)/factorial(i2 + k1)
        sum = sum + argument

        sign = -sign
        k1 = k1 + 1
     end do
                 
  end if

  gaunt = gaunt * sum

end function gaunt

end module spherical_harmonic
