!******************************************************************************
!
! File : strings.f90
!   by : Alan Tackett
!   on : 8/23/95
!  for : String Manipulation
!
!  This module contains routines to manipulate strings and lists of strings.
!
!******************************************************************************

Module strings

implicit none

Integer, PARAMETER :: STR_LENGTH = 25   !** String length

Integer, PARAMETER :: STR_IGNORE = 0  !** Ignore Case when matching
Integer, PARAMETER :: STR_EXACT  = 1  !** Perform exact matches
Integer, PARAMETER :: STR_NONE   = 0  !** Perform no conversion when adding
Integer, PARAMETER :: STR_UPPER  = 1  !** Perform upper case conversion
Integer, PARAMETER :: STR_LOWER  = 2  !** Perform lower case conversion



Type StringList_struct     !** String List structure
   Integer               :: MaxSize         !** Max Array size
   Integer               :: Used            !** Num of strings currently used
   Integer               :: CaseType        !** Search Type
   Integer               :: ConvertType     !** List Type ie, exact,upcase,..
   Character*(STR_LENGTH), Pointer :: strlist(:)  !** Pointer to string list
End Type


Interface SearchList                 !** Set up generic routine for 
  Module Procedure SearchList_Manual !** searching the list
  Module Procedure SearchList_struct
end Interface


Contains


!*****************************************************************************
!
! UpperCase - Converts a string to Upper Case
!
! str     - String to convert
!
!*****************************************************************************

Subroutine UpperCase(str)
  Character(*), Intent(INOUT) :: str

  Integer  :: i, j, k

  j = len(Str)

  Do i=1, j
    k = IACHAR(str(i:i))
    if ((k>96) .AND. (k<123)) str(i:i) = ACHAR(k-32)
  End Do

  Return
End Subroutine

!*****************************************************************************
!
! LowerCase - Converts a string to Lower Case
!
! str     - String to convert
!
!*****************************************************************************

Subroutine LowerCase(str)
  Character(*), Intent(INOUT) :: str

  Integer  :: i, j, k

  j = len(Str)

  Do i=1, j
    k = IACHAR(str(i:i))
    if ((k>64) .AND. (k<91)) str(i:i) = ACHAR(k+32)
  End Do

  Return
End Subroutine


!*****************************************************************************
!
! SearchList_manual - Search a user supplied list manually for a match
!     
! Str - String to match
! StrList - List of strings to search
! ListSize - Size of list
! CaseType - Determines how a match is determined
!             STR_EXACT  - Perform an exact match using case 
!             STR_IGNORE - Ignore case when doing a match
!
! Return Values
!    Returns the location in the list where STR is stored or -1 if
!    STR was not found
! 
!*****************************************************************************

Integer Function SearchList_manual(Str, StrList, ListSize, CaseType)
   Character(*), Intent(IN) :: Str
   Character(*), Intent(IN) :: StrList(:)
   Integer,      Intent(IN) :: ListSize
   Integer,      Intent(IN) :: CaseType

   Integer :: i, j
   Character*(100) :: txt, Match

   i = 1
  
   Match = Str;  txt = Strlist(i)

   If (CaseType == STR_IGNORE) then
     Call UpperCase(Match)
     Call UpperCase(txt)
   end IF
    
        
   
   Do While ((TRIM(Match) /= TRim(txt)) .AND. (i<ListSize))
      i = i + 1
      txt = StrList(i)
      If (CaseType == STR_IGNORE) Call UpperCase(txt)  
   End Do

!***write(*,*) 'SLM: i=',i, ' * used=',ListSize, ' MAtch=',TRim(Match), ' * TXT=!',Trim(Txt),'!'
   if (Trim(Match) /= Trim(txt)) i = -1

   SearchList_Manual = i
   Return
End Function


!*****************************************************************************
!
! SearchList_struct - Search for a string using a list structure
!     
! Str - String to match      
! SL  - String List structure to search
!
! Return Values
!    Returns the location in the list where STR is stored or -1 if
!    STR was not found
! 
!*****************************************************************************

Integer Function SearchList_Struct(Str, SL)
   Character(*),             Intent(IN) :: Str
   Type (StringList_Struct), Intent(IN) :: SL

   
   SearchList_Struct = SearchList_Manual(Str, SL%Strlist, SL%Used, SL%CaseType)
   Return
End Function


!*****************************************************************************
!
! SearchAddList - Search for a string using a list structure and ADD the
!                 string if it does not currently exist.
!     
! Str - String to match      
! SL  - String List structure to search
! IsNew - If the string is added to the list then
!         IsNew=.TRUE. else IsNew is .FALSE. (RETURNED, OPTIONAL)
!
! Return Values
!    Returns the location in the list where STR is stored or -1 if
!    STR was not found.  ISNEW is also returned
! 
!*****************************************************************************

Integer Function SearchAddList(Str, SL, IsNew)
   Character(*),             Intent(IN)    :: Str
   Type (StringList_Struct), Intent(INOUT) :: SL
   Logical,   Optional,      Intent(OUT)   :: IsNew

   Integer i

   i = SearchList_Manual(Str, SL%Strlist, SL%Used, SL%CaseType)

   if (i == -1) then
      SL%Used = SL%Used + 1
      i = SL%USed
      SL%StrList(i) = Str   
   
      Select Case (SL%ConvertType)
        Case (STR_UPPER)
           Call UpperCase(SL%StrList(i))
        Case (STR_LOWER)      
           Call UpperCase(SL%StrList(i))
      End Select

      If (PRESENT(IsNew)) IsNew = .TRUE. 
   else 
      If (PRESENT(IsNew)) IsNew = .FALSE.      
   End If


   SearchAddList = i
   Return
End Function

!*****************************************************************************
!
!  InitStringList - Initializes a string search list
!
!  SL       - StringList structure to initialize
!  Array    - String array to store list information
!  MaxList  - Max size of list
!  ListUsed - Number of elements currently used
!  CaseType - Determines how a match is determined
!             STR_EXACT  - Perform an exact match using case 
!             STR_IGNORE - Ignore case when doing a match
!  ConvertType - Determines how the strings are stored in the list.
!             STR_NONE - No case conversion
!             STR_UPPER - Convert to Upper case
!             STR_LOWER - Convert to lower case
!
!*****************************************************************************

Subroutine InitStringList(SL, Array, MAxList, ListUsed, CaseType, ConvertType)
  Type (StringList_Struct), Intent(OUT) :: SL
  Character(*),             Pointer     :: Array(:)
  Integer,                  Intent(IN)  :: MAxList
  Integer,                  Intent(IN)  :: ListUsed
  Integer,                  Intent(IN)  :: CaseType
  Integer,                  Intent(IN)  :: ConvertType

  integer :: i

  SL%StrList => Array
  SL%MaxSize = MaxList
  SL%Used = ListUsed
  SL%CaseType = CaseType
  SL%ConvertType = ConvertType

  Do i=ListUsed+1, MaxList
     Array(i) = ' '
  End Do

  Return
End Subroutine

End Module

