!******************************************************************************
!
! File : timing.f90
!   by : Alan Tackett
!   on : 01/27/00
!  for : PWPAW
!
!  This module contains the stopwatch definitions and initializes them
!
!******************************************************************************

Module timing

Use stopwatch

implicit NONE!!!!!!!!

Type Timer_Type                 !** Timer definition
   Type (WatchType) :: Watch    !** The watch
   Integer          :: N        !** Number of invocations
   Character(len=30):: Name     !** Timer name (Duplicated in watch)
End Type
 
Integer, PARAMETER :: Total_Timer        = 1  !** Total CPU Time for program
Integer, PARAMETER :: CalcHam_Timer      = 2  !** Time for all CalcHam calls
Integer, PARAMETER :: CalcSmoothH_Timer  = 3  !** CalcSmoothH calls
Integer, PARAMETER :: DenHat_Timer       = 4  !** Time for calculating n^
Integer, PARAMETER :: AtomVxc_Timer      = 5  !** Atomic Vxc terms
Integer, PARAMETER :: SymRho_Timer       = 6  !** Symmetrize n~
Integer, PARAMETER :: MakeNewPsi_Timer   = 7  !** Timer for NewPsi step
Integer, PARAMETER :: CalcProjProd_Timer = 8  !** CalcProjProducts  
Integer, PARAMETER :: AccumProj_Timer    = 9  !** AccumProjProducts
Integer, PARAMETER :: FFT_Timer          = 10  !** Total time spent in FFT's
Integer, PARAMETER :: Setup_Timer        = 11 !** Time spent in initialization
Integer, PARAMETER :: Iter_timer         = 12 !** Time for each loop
Integer, PARAMETER :: JDQZSetupCorr_Timer= 13 !** Setup time for JDQZ solver
Integer, PARAMETER :: JDQZSetupPrec_Timer= 14 !** Setup time for JDQZ solver
Integer, PARAMETER :: CalcCorr_Timer     = 15 !** Time for calc'ing the corr v
Integer, PARAMETER :: Force_Timer        = 16 !** Time for forces
Integer, PARAMETER :: HPsi_Timer         = 17 !** H*Psi timer
Integer, PARAMETER :: OPsi_Timer         = 18 !** O*Psi Timer
Integer, PARAMETER :: HandOPsi_Timer     = 19 !** O*Psi Timer

Integer, PARAMETER :: MAX_TIMERS = 19   !** Max different timers
 
Type (Timer_Type) :: Timer(MAX_TIMERS)

!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
!  Print_timers - Prints the timing inforation to the specified unit
!      
!       Unit - Output unit
!
!******************************************************************************

Subroutine Print_Timers(unit)
  Integer, Intent(IN) :: unit

  Integer :: i
  Real    :: wall, cpu, user, sys, walliter

  Call Pause_ALL_Timers 

  Write(unit,*) ' '
  Write(unit,'(A20, 1X, A8, 1X, 4(A11,1X))') '    Timer', 'Count', &
       'Wall', 'CPU', 'User', 'Wall/call'
  Write(unit,'(A20, 1X, A8, 1X, 5(A11,1X))') '--------------------', &
       '--------', '-----------', '-----------', '-----------','-----------'

  Do i=1, MAX_TIMERS
     Call Read_Watch(wall, Timer(i)%Watch, 'wall')
     Call Read_Watch(cpu, Timer(i)%Watch, 'cpu ')
     Call Read_Watch(user, Timer(i)%Watch, 'user')
!**     Call Read_Watch(sys, Timer(i)%Watch, 'sys ')

     walliter = 0
     If (timer(i)%N > 0) then
        walliter = wall/timer(i)%N
     End if
     Write(unit, '(A20, 1X, I8, 1X, 4(F11.2,1X))') timer(i)%Name, timer(i)%N, &
          wall, cpu, user, walliter
  End Do

  Write(unit,*) ' '

  Call Resume_ALL_Timers 
  Return
End Subroutine

!******************************************************************************
! Pause_ALL_timers - Pauses all the timers in the system
!******************************************************************************

Subroutine Pause_All_Timers
  Integer :: i

  Do i=1, Max_Timers
     Call Pause_Timer(Timer(i))
 End Do

 Return
End Subroutine

!******************************************************************************
! Resume_ALL_timers - Resumes all the timers in the system after a pause.
!******************************************************************************

Subroutine Resume_All_Timers
  Integer :: i

  Do i=1, Max_Timers
     Call Resume_Timer(Timer(i))
 End Do

 Return
End Subroutine

!******************************************************************************
!  Start_Timer - Starts the timer
!******************************************************************************

Subroutine Start_Timer(Timer)
  Type (Timer_Type), Intent(INOUT) :: Timer

  Call Start_Watch(Timer%Watch)
  timer%N = timer%N + 1    !** Increment the calls

  Return
End Subroutine

!******************************************************************************
!  Stop_Timer - Stops the timer
!******************************************************************************

Subroutine Stop_Timer(Timer)
  Type (Timer_Type), Intent(INOUT) :: Timer

  Call Stop_Watch(Timer%Watch)

  Return
End Subroutine


!******************************************************************************
!  Pause_Timer - Pauses the timer
!******************************************************************************

Subroutine Pause_Timer(Timer)
  Type (Timer_Type), Intent(INOUT) :: Timer

!*  Call Stop_Watch(Timer%Watch)
  Call Pause_Watch(Timer%Watch)

  Return
End Subroutine

!******************************************************************************
!  Resume_Timer - Resumes the timer after a pause
!******************************************************************************

Subroutine Resume_Timer(Timer)
  Type (Timer_Type), Intent(INOUT) :: Timer

!*  Call Start_Watch(Timer%Watch)
  Call End_Pause_Watch(Timer%Watch)

  Return
End Subroutine

!******************************************************************************
!  Reset_Timer - Resets the timer
!******************************************************************************

Subroutine Reset_Timer(Timer)
  Type (Timer_Type), Intent(INOUT) :: Timer

  Call Reset_Watch(Timer%Watch)
  timer%N = 0

  Return
End Subroutine

!******************************************************************************
!  Create_Timer - Create the timer
!******************************************************************************

Subroutine Create_Timer(Timer, timer_name)
  Type (Timer_Type), Intent(INOUT) :: Timer
  Character(len=*),  Intent(IN)    :: timer_name

  Call Create_Watch(Timer%Watch, NAME=timer_name)
  timer%N = 0
  timer%Name = Timer_Name
  Return
End Subroutine

!******************************************************************************
!
!  InitTimers - Initializes all the timers for use.
!
!******************************************************************************

Subroutine InitTimers
  Integer :: i

  Call Option_StopWatch(default_clock=(/'wall','cpu ','user'/))

  Call Create_Timer(Timer(Total_timer), 'Total')
  Call Create_Timer(timer(CalcHam_timer), 'CalcHam')
  Call Create_Timer(timer(CalcSmoothH_timer), '   CalcSmoothH')
  Call Create_Timer(timer(DenHat_timer), '   DenHat')
  Call Create_Timer(timer(AtomVxc_timer), '   Atomic Vxc')
  Call Create_Timer(timer(SymRho_timer), '   Symmtrize Rho')
  Call Create_Timer(Timer(MakeNewPsi_timer), 'MakeNewPsi')
  Call Create_Timer(timer(CalcProjProd_timer), 'CalcProjProducts')
  Call Create_Timer(Timer(AccumProj_Timer), 'AccumProj')
  Call Create_Timer(Timer(FFT_timer), 'FFT')
  Call Create_Timer(Timer(Setup_timer), 'Setup')
  Call Create_Timer(Timer(Iter_timer), 'Iteration')
  Call Create_Timer(Timer(JDQZSetupCorr_timer), 'JDQZ Setup Corr')
  Call Create_Timer(Timer(JDQZSetupPrec_timer), 'JDQZ Setup Precond')
  Call Create_Timer(Timer(CalcCorr_timer), 'Calc Corr Eqn')
  Call Create_Timer(Timer(Force_timer), 'Forces')
  Call Create_Timer(Timer(HPsi_timer), 'HPsi')
  Call Create_Timer(Timer(OPsi_timer), 'OPsi')
  Call Create_Timer(Timer(HandOpsi_timer), 'HPsi and OPsi')

  Return
End Subroutine

End Module
