!******************************************************************************
!
! File : units.f90
!   by : Alan Tackett
!   on : 8/23/95
!  for : Unit conversion
!
!  This unit contains routines to convert numbers back and forth between
!  various different units.
!
!******************************************************************************

Module units

Use strings

implicit none

Integer, PARAMETER :: U_MAX_ABBREV = 10       !** Max unit string

         !** Supported Length units **
Integer, PARAMETER :: U_M        =   1        !** Meters
Integer, PARAMETER :: U_MM       =   2        !** millimeters 
Integer, PARAMETER :: U_UM       =   3        !** Micrometers
Integer, PARAMETER :: U_NM       =   4        !** nanometers
Integer, PARAMETER :: U_PM       =   5        !** picometers
Integer, PARAMETER :: U_FM       =   6        !** femtometers
Integer, PARAMETER :: U_ANGSTOMS =   7        !** angstroms
Integer, PARAMETER :: U_BOHR     =   8        !** Bohr

Integer, PARAMETER, PRIVATE :: Length_SIZE = 7

Real,  TARGET, PRIVATE :: Length(LENGTH_SIZE) = (/    &
             1.0,                & !** Meters
            1E-3,                & !** millimeters
            1E-6,                & !** micrometers
            1E-9,                & !** nanometers
            1E-12,               & !** femtometers
            1E-10,               & !** Angstoms
            0.529177249E-10      & !** Bohr 
/)                               

Character*(U_MAX_ABBREV),TARGET,PRIVATE:: Length_Text(LENGTH_SIZE)=(/ &
            "m   ",              & !** Meters
            "mm  ",              & !** millimeters
            "um  ",              & !** micrometers
            "nm  ",              & !** nanometers
            "fm  ",              & !** femtometers
            "ang ",              & !** Angstoms
            "Bohr"               & !** Bohr 
/)                               

         !** Supported Energy units **
Integer, PARAMETER :: U_J        = 101        !** Joules
Integer, PARAMETER :: U_EV       = 102        !** Electron-Volts
Integer, PARAMETER :: U_MEV      = 103        !** million Electron-Volts
Integer, PARAMETER :: U_HARTREE  = 104        !** Hartree
Integer, PARAMETER :: U_RYDBERGS = 105        !** Rydbergs

Integer, PARAMETER, PRIVATE :: ENERGY_SIZE = 5
Real, TARGET, PRIVATE :: Energy(ENERGY_SIZE) = (/    &
              1.0,               & !** Joules
              1.60217733E-19,    & !** Electron-Volts
              1.60217733E-13,    & !** Million Electron-Volts      
              4.3597482E-18,     & !** Hartree
              2.17989043E-18     & !** Rydbergs 
/)

Character*(U_MAX_ABBREV),TARGET,PRIVATE:: Energy_Text(ENERGY_SIZE)=(/ &
             "J  ",              & !** Joules
             "eV ",              & !** Electron-Volts
             "meV",              & !** Million Electron-Volts      
             "H  ",              & !** Hartree
             "Ryd"               & !** Rydbergs
/)


Contains


!****************************************************************************
!
! Str2Unit - Determines if a text string is a unit and if so converts 
!             it to the appropriate unit constant.
!
! txt - Text to convert
!
! Return Values
!     If TXT corresponds to a supported unit abbreviation then the
!     corresponding unit constant is returned.  If no unit is found then
!     a value of -1 is returned.
!
!****************************************************************************

Integer Function Str2Unit(txt)
   Character*(*), Intent(IN) :: txt

   Integer :: i, ans

   ans = SearchList_Manual(txt, Length_Text, Length_Size, STR_IGNORE)
   if (ans == -1) then
     ans = SearchList_Manual(txt, Energy_Text, Energy_Size, STR_IGNORE)
   End IF

   Str2Unit = ans

   Return   
End Function


!****************************************************************************
!
! Unit2Str - Converts a unit constant to a string
!
! Unit - Unit to convert
!
! Return Values
!     Returns the corresping unit text abbreviation
!
!****************************************************************************

Character*(U_MAX_ABBREV) Function Unit2Str(Unit)
   Integer, Intent(IN) :: Unit

   Character*10 :: ans

   ans = " "

   If (Unit < 100) then
      If (Unit <= Length_SIZE)  ans = Length_Text(Unit)
   else if (Unit < 200) then
      If ((Unit-100) <= ENERGY_SIZE)  ans = Energy_Text(Unit-100)
   End If

   Unit2Str = ans

   Return   
End Function


!****************************************************************************
!
!  ConvertFactor - Determines which unit type the From and To units are from
!                 and returns a conversion factor for the units.
!
!  From_Unit - Input or Starting Unit
!  To_Unit   - Output unit
!  ConvertFactor - Conversion Factor(RETURNED)
!
!  Return Values
!    The conversion factor is returned.  If FROM_UNIT and TO_UNIT are different
!    and an error occured a unitary, 1, conversion factor is returned.
!
!****************************************************************************

Real Function ConvertFactor(From_Unit, To_Unit)
  Integer, Intent(IN)  :: From_Unit
  Integer, Intent(IN)  :: To_Unit

  Real          :: Factor
  Integer       :: base
  Real, Pointer :: Units(:)


  if (ABS(From_Unit-To_unit) >= 99) then
    Factor = 1
  else 
    if (From_unit < 100) then
       Units => Length
       base = 0
    else
       Units => Energy
       base = 100
    End If

    Factor = Units(From_Unit - base) / Units(To_Unit - base)
  end If

  ConvertFactor = Factor
  Return 
End Function


End Module
