!******************************************************************************
!
! File : Vhartree_pack.f90
!   by : Alan Tackett
!   on : 11/15/95
!  for : PAW Method 
!
! Contains 2 routines to encode and decode the V_Hartree matrix indices that
! are used in the atom_data module for the field V_Hartree.
!
!               aL
! Density(:) = V  
!               n l n l ; n l n l
!                i i j j   k k l l
!
!
! The indices have to be stored are :
!
!                (n l )  (n l )   (n l )  (n l )   L
!                  i i     j j      k k     l l
!
! Bits Required:   (5)     (5)     (5)     (5)    (3)   = 23 bits total
!
! Each l value is in the range on 0..3 and the corresponding m is in the 
! range of -l <= m <= l for 2*l + 1 possible values.  This gives the range
! for L = 0..6.  The (nl) indices are stored as pairs which correspond
! to the actual basis function index.  Since only 5 bits are used the max
! number of basis functions is 32 for each atom. The l's can be determined
! from the (nl)'s and hence are not stored.  
!
! Also note that since 23 bits are required each integer should 
! be 4 byte integer. The 23 LSB's are used all other bits are set to 0.
!
!    *******  NOTE: No Error checking is done!!!!!! ********
!
!******************************************************************************

Module vhartree_pack

implicit none

!******************************************************************************
Contains
!******************************************************************************


!******************************************************************************
!
!  VHartree_Encode - Encodes the given nili, njlj,  nklk, nlll, and L supplied
!                    assuming the above restrictions on indices.
!
!
!  NOTE : NO error checking is done!!!!
!         The return value must be a 4 byte or greater integer!!

!
!******************************************************************************

Integer Function Vhartree_Encode(nili, njlj, nklk, nlll, L)
   Integer, Intent(IN) :: nili
   Integer, Intent(IN) :: njlj
   Integer, Intent(IN) :: nklk
   Integer, Intent(IN) :: nlll
   Integer, Intent(IN) :: L

   Integer :: N

   N = nili
   N = ISHFT(N, 5) + njlj
   N = ISHFT(N, 5) + nklk
   N = ISHFT(N, 5) + nlll
   N = ISHFT(N, 3) + L

   Vhartree_Encode = N
   Return
End Function
         
!******************************************************************************
!
!  Vhartree_Decode - Decodes the given value(N) into the separate nili, njlj,
!                    nklk, nlll, and L assuming the above restrictions on
!                    (nl)'s, l's, m's, and L.
!
!
!  NOTE : NO error checking is done!!!!
!         The value for N must be a 4 byte or greater integer!!
!
!******************************************************************************

Subroutine VHartree_Decode(N, nili, njlj, nklk, nlll, L)
   Integer, Intent(IN)  :: N
   Integer, Intent(OUT) :: nili
   Integer, Intent(OUT) :: njlj
   Integer, Intent(OUT) :: nklk
   Integer, Intent(OUT) :: nlll
   Integer, Intent(OUT) :: L

   nili = ISHFT(IAND(31*(2**18), N), -18)
   njlj = ISHFT(IAND(31*(2**13), N), -13)
   nklk = ISHFT(IAND(31*(2**8), N), -8)
   nlll = ISHFT(IAND(31*(2**3), N), -3)
   L    = IAND(7, N) 

   Return
End Subroutine
    

End Module
