!******************************************************************************
!
! File : work_mgr.f90
!   by : Alan Tackett
!   on : 9/13/98
!  for : PAW Project
!
!  Routines for manipulating the work buffers by "checking" them "in"
!  and "out".
!
!******************************************************************************

Module work_mgr

Use paw_inout
Use mem_data
Use misc

Implicit NONE!!!!!!!


Integer, PARAMETER, PRIVATE :: USED_BUFFER = 1
Integer, PARAMETER, PRIVATE :: FREE_BUFFER = 0

Type Work_Struct
   Complex, Pointer :: BasePtr(:)
   Complex, Pointer :: Ptr(:)
   Integer                  :: Used
End Type

Type (Work_struct), PRIVATE, Pointer :: Work(:)

Integer  :: Work_Size  !** Size of work buffers in c*16
Integer  :: Num_Work   !** Number of work buffers
Integer  :: Max_Buffers_Used  !** Max buffers used at any given time

!******************************************************************************
Contains
!******************************************************************************


!******************************************************************************
!
!  FreeBuffer - Frees  the buffer for use
!
!     Ptr - pointer to buffer for freeing
!
!******************************************************************************

Subroutine FreeBuffer(Ptr)
  Complex, Pointer :: Ptr(:)

  Integer :: i

  i = 1
  Do While ((i<Num_Work) .AND. (.NOT. Associated(Work(i)%Ptr, Ptr)))
    i = i + 1
  End Do

!Write(*,*) 'FreeBuffer: checked in index=',i

  Work(i)%Used = FREE_BUFFER
  Nullify(Work(i)%Ptr)
  Nullify(Ptr)

  Return
End Subroutine

!******************************************************************************
!
!  GetBuffer - Gets the next free buffer and returns its pointer or dies
!     if none are available!
!
!
!******************************************************************************


Subroutine GetBuffer( Ptr)
  Complex,   Pointer    :: Ptr(:)

  Integer :: i
  real :: tmp

  i=1
  Do While ((i < Num_Work) .AND. (Work(i)%Used==USED_BUFFER))
    i = i + 1
  End Do

  If (Work(i)%Used==USED_BUFFER) then
	Write(Error_Unit, *) 'GetBuffer: No Free Buffers!!!'
	Write(Error_Unit, *) 'GetBuffer:   Work(:)%Used=',Work(:)%Used
	tmp = 0
	tmp = 1.0/tmp
  else
	Work(i)%Ptr => Work(i)%BasePtr(1:PsiArraySize)
	Ptr => Work(i)%Ptr
	Work(i)%Used = USED_BUFFER

        Max_Buffers_Used = MAX(Max_Buffers_Used, SUM(Work(:)%Used))
  End if

  
  Return
End Subroutine

!******************************************************************************
!
!  GetBuffer_Size - Gets the next free buffer and returns its pointer or dies
!     if none are available!
!
!     NOTE :  Returns a buffer of the requested grid size. 
!
!******************************************************************************


Subroutine GetBuffer_Size(GridSize,  Ptr)
  Integer,           Intent(IN) :: GridSize
  Complex,   Pointer    :: Ptr(:)
  Integer :: i
  real :: tmp

  i=1
  Do While ((i < Num_Work) .AND. (Work(i)%Used==USED_BUFFER))
    i = i + 1
  End Do

  If (Work(i)%Used==USED_BUFFER) then
	Write(Error_Unit, *) 'GetBuffer: No Free Buffers!!!'
	Write(Error_Unit, *) 'GetBuffer:   Work(:)%Used=',Work(:)%Used
	tmp = 0
	tmp = 1.0/tmp
  else
	Work(i)%Ptr => Work(i)%BasePtr(1:GridSize)
	Ptr => Work(i)%Ptr
	Work(i)%Used = USED_BUFFER
  End if

  Return
End Subroutine


!******************************************************************************
!
!  Print_WorkStatus - Prints the work buffer status to the unit
!
!     Fd - Output unit
!
!******************************************************************************

Subroutine Print_WorkStatus(Fd)
  Integer, Intent(IN) :: Fd
 
  Character*(NUM_WORK) :: string
  Integer :: i

  Do i=1, NUM_WORK
     If (Work(i)%Used==USED_BUFFER) then
        string(i:i) = '1'
     else
        string(i:i) = '0'
     End If
  End Do

  Write(fd, *) 'Print_WorkStatus: Work(:)%Used: ',string
  Write(fd, *) 'Print_WorkStatus: Max Buffers Used:',Max_Buffers_Used

  Return
End Subroutine

!******************************************************************************
!
!  ReAllocate_Work - reallocates the work arrays to a different size
!
!      W_Size - Size of buffers
!     
!******************************************************************************

Subroutine ReAllocate_Work(W_Size)
  Integer, Intent(IN) :: W_Size

  Integer :: i,j
  Character*100 :: msg

  Work_size = W_size

Write(Log_Unit,*) 'ReAllocate_Work: New Size = ',Work_Size
Write(*,*) 'ReAllocate_Work: New Size = ',Work_Size

  Work(:)%Used = FREE_BUFFER
Write(*,*) FREE_BUFFER,Num_Work

  !*** First deallocate ALL old buffers so don't frag memory **
  Do i=1, Num_Work
    If (Associated(Work(i)%BasePtr)) DeAllocate(Work(i)%BasePtr)
    Nullify(Work(i)%Ptr)
  End Do

  !** Now ReAllocate **
  Do i=1, Num_Work
     Allocate(Work(i)%BasePtr(Work_size), STAT=j)
     msg = 'InitWorkArrays: Could Not allocate buffer!'
     Call Check_Error(j, msg, Error_Unit, .TRUE.,paw_wc, "InitWorkArrays:")
  End Do
  
  return
End Subroutine


!******************************************************************************
!
!  InitWorkArrays - Initializes the work arrays for use.
!
!      N      - Number of work buffers to allocate
!      W_Size - Size of buffers
!     
!******************************************************************************

Subroutine InitWorkArrays(N)
  Integer, Intent(IN) :: N

  Integer :: i,j
  Character*100 :: msg

Write(*,*) 'InitWorkArrays:  N=',N !**, ' * W_size=',W_Size

  Max_Buffers_Used = 0

  Work_size = 0
  Num_Work = N

  Allocate(Work(num_Work), STAT=j)
  msg = 'InitWorkArrays: Could Not allocate Work Array!'
  Call Check_Error(j, msg, Error_Unit, .TRUE.,paw_wc, "InitWorkArrays:")
  
  Work(:)%Used = FREE_BUFFER

  Do i=1, Num_Work
     Nullify(Work(i)%BasePtr)
  End Do
  
  Return
End Subroutine


End Module
